/* appli.c => linked to main.c */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/*
 ** Thanks to: Theodore Tso <tytso@mit.edu> for the e2fs library.
 ** Thanks to: Jean-loup Gailly <jloup@gzip.org> and
 **            Mark Adler <madler@alumni.caltech.edu> for the zlib library.
 **              http://www.freesoftware.com/pub/infozip/zlib/
 ** Thanks to: Colin Plumb <colin@nyx.net> for his guide to use GCC/asm()
 **		sent in the list linux-kernel@vger.rutgers.edu .
 **            See also: "info GCC"
 **              "Assembler Instructions with C Expression Operands"
 ** Thanks to: Ralf Brown <ralf@pobox.com> for his list of interrupts
 **		INTER60 (http://www.pobox.com/~ralf) and all contributers.
 ** Thanks to: The GAS developper team.
 ** Thanks to: opcode databases like http://www.sandpile.org/80x86
 ** Thanks to: "flat" real mode description from John Fine
 **		(johnfine@erols.com , http://users.erols.com/johnfine/)
 **            and Robert Collins
 **		(rcollins@x86.org , http://www.x86.org/ftp/articles/)
 ** Thanks to: Bart Lagerweij for saving me time to understand that
 **            I will not be able to call the Bootable CDROM BIOS
 **            services (see the list of really implemented calls)
 **            http://www.nu2.nu/bootcd/
 */

#include "make.h"
#include "library.h"
#include "libextra.h"	/* detect_processor(), get_number() */
#include "messages.h"
#include "boot.h"
#include "debug.h"
#include "user.h"	/* define UI which is the user interface */
#include "disk.h"
#include "bios.h"	/* _DOSEMU_check() */
#include "malloc.h"	/* init_malloc(), report_malloc() */
#include "dos.h"	/* DOS_version() */
#include "xms.h"

#include "vesabios.h"	/* _VESA_setmode() */
#include "vgabios.h"	/* _VGA_autoload_default_palette()... */

#include "util.h"
#include "vmlinuz.h"


static void before_exit (void)
  {
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
  if (BOOT1_COM_port() < 0) {
      _VGA_autoload_default_palette (1);
      if (VESA_ACTIVE() && !UI.parameter.BIOS_compatible) {
	  /* some BIOS need VESA before VGA setmode: */
	  if (_VESA_setmode (3) != 0)
	      DBG ((" [Last _VESA_setmode failed!] "));
	  _VGA_setmode (3);
	  }
	else {
	  unsigned char nbcol, page;
	  _VGA_setmode (_VGA_getmode (&nbcol, &page) & 0x7F);
	  }
      }
#endif

#ifdef TREAT_EXCEPTION
  restore_vectors();
#endif

  DBG_END();
  }

#if USER_SUPPORT != 0
struct {
    unsigned short curpage, totalpage, nbperpage, nbtotal, interline;
    enum { kernel_bottom_menu, setup_bottom_menu } type;
    } Menu; /* init to BSS */

#if USER_SUPPORT & (BIOS_MOUSE_SUPPORT | SERIAL_MOUSE_SUPPORT | JOYSTICK_SUPPORT)
#include "mouse.h"

#define MOUSE_declare_attribute(varname) struct mousefieldattr_str varname = {}
#define MOUSE_attrset_mainfield(varname, val)	 varname.mainfield = val
#define MOUSE_attrset_upperfield(varname, val)	 varname.upperfield = val
#define MOUSE_attrset_pressedfield(varname, val) varname.pressed = val
#define MOUSE_attrset_chkboxfield(varname, val)	 varname.checkbox = val
#define MOUSE_attrset_activefield(varname, val)	 varname.active = val
#define MOUSE_attrset_greyfield(varname, val)	 varname.greyed = val

#define RESET_FIELDS(fieldmask)		MOUSE_reset_field (fieldmask)
#define PUT_FIELD(name, attr, key, row, col)	MOUSE_print_field (name, attr, key, row, col)

#else  /* BIOS_MOUSE_SUPPORT | SERIAL_MOUSE_SUPPORT | JOYSTICK_SUPPORT */
/*
 * Here the mouse is disabled probably to save code space,
 * so we do not do 3D button: no mouse.c::drawbutton()
 */
static void NOMOUSE_print (const char *str, unsigned active)
  {
  union {
    unsigned		 color;
    struct attribute_str attr;
    } save = {0}; /* warning */

  if (active) {
      if (UI.attributes.valid.underline) {
	  save.attr = UI.attributes.current;
	  UI_setattribute (attr_underline);
	  }
	else if (UI.attributes.valid.brighter) {
	  save.attr = UI.attributes.current;
	  UI_setattribute (attr_brighter);
	  }
	else {
	  save.color = UI.fgcolor;
	  UI_setfgcolor (MOUSE_SELECTABLE_CHARS);
	  }
      }
  UI_putstr (str);
  if (active) {
      if (   UI.attributes.valid.transparent
	  || UI.attributes.valid.underline
	  || UI.attributes.valid.brighter)
	  UI_setattribute (save.attr);
	else
	  UI_setfgcolor (save.color);
      }
  UI_putstr (" ");
  }

#define MOUSE_declare_attribute(varname)	 unsigned char varname = 0
#define MOUSE_attrset_mainfield(varname, val)	 /* */
#define MOUSE_attrset_upperfield(varname, val)	 /* */
#define MOUSE_attrset_pressedfield(varname, val) /* */
#define MOUSE_attrset_chkboxfield(varname, val)	 /* */
#define MOUSE_attrset_activefield(varname, val)	 varname = val
#define MOUSE_attrset_greyfield(varname, val)	 /* */

#define RESET_FIELDS(fieldmask)		/* */
#define PUT_FIELD(name, active, key, row, col)	NOMOUSE_print (name, active)
#endif /* BIOS_MOUSE_SUPPORT | SERIAL_MOUSE_SUPPORT | JOYSTICK_SUPPORT */
#endif /* USER_SUPPORT != 0 */

static inline void detect_environment (struct processor_str *processor)
  {
#if !(SETUP & QUITE)
  unsigned short dosemu_version, dosemu_patchlevel;

  if (BOOT1_DOS_running()) {
      PRINT (" " MSG_ON " DOS");
      processor->DOS_version = DOS_version ();
      PRINTF (" %u.%u",
		processor->DOS_version >> 8,
		processor->DOS_version & 0xFF);
      }
    else {
      PRINT (" " MSG_ON " BIOS");
      processor->DOS_version = 0;
      }

  if (_DOSEMU_check (&dosemu_version, &dosemu_patchlevel) == 0)
      PRINTF (", DOSEMU: V%u.%u-%u",
		dosemu_version >> 8,
		dosemu_version & 0xFF, dosemu_patchlevel);
#endif
  }

static inline void detect_keyboard (void)
  {
  if (BOOT1_COM_port() < 0) {
      if (_BIOS_is_extended_keyboard()) {
	  if (_BIOS_is_122key_keyboard()) {
	      processor.keyboard = ext122kbd;
	      PRINT (KEYBOARD_122);
	      }
	    else {
	      processor.keyboard = extkbd;
	      PRINT (KEYBOARD_EXT);
	      }
	  }
	else {
	  PRINT (KEYBOARD_STD);
	  processor.keyboard = stdkbd;
	  }
      }
    else
      processor.keyboard = serialkbd;
  }

static inline void list_disks (void)
  {
#if !(SETUP & QUITE)
  unsigned short number;
  char disk_name[SIZE_DISKNAME];

  for (number = 0; number < DI.nbdisk; number++) {
#if 0
      unsigned long long longsize = DI.param[number].nbtotalsector;
      unsigned size;

      longsize *= DI.param[number].bytepersector;
      size = longsize / 1024;
#else
      unsigned nbsector_in_16k = (16 * 1024) / DI.param[number].bytepersector;
      unsigned size = (DI.param[number].nbtotalsector / nbsector_in_16k) * 16;
#endif

      diskname (&DI.param[number], disk_name);
      PRINTF (DISK_NAME_SIZE_NBPARTITION,
		number, disk_name,
		(size > 9*1024) ? size/1024 : size,
		(size > 9*1024) ? "Mb" : "Kb",
		DI.param[number].nbpartition
		);
      }
#endif
  }

static inline void detect_serial_timeout (void)
  {
  unsigned char number, tmp, max;

  DBG (("Timeout for "));
  max = _BIOS_equipment_flags().nb_serial_port;

  PRINTF (U_SERIAL_PORT, max, (max > 1) ? "s" : "");

  for (number = 0; number < max; number++) {
      tmp = peekb (0x0040007C + number);
      DBG (("COM%u: 0x%X", number+1, tmp));
      if (tmp == 0 || tmp > 5) {
	  DBG ((" set to 1"));
	  pokeb (0x0040007C + number, 1);
	  }
      if (number < max - 1)
	  DBG ((", "));
	else
	  DBG (("\r\n"));
      }
  }

#if USER_SUPPORT & VGA_SUPPORT
static inline void listmode (struct user_interface_str *ui)
  {
#if !(SETUP & QUITE)
  unsigned short start = 0, cpt, firsttime = 1,
		 valid = VIDEO_mode_is_valid(0);

  for (cpt = 1; cpt <= VIDEO_mode_max(); cpt++) {
      if (cpt == VIDEO_mode_max() || VIDEO_mode_is_valid(cpt) != valid) {
	  if (valid == 0)
	      start = cpt;
	    else {
	      const char *format;
	      if (!firsttime)
		  PRINT (", ");
	      firsttime = 0;
	      if (start <= cpt-2) {
		  if (start == cpt-2)
		      format = "0x%X, 0x%X";
		    else
		      format = "0x%X..0x%X";
		  }
		else
		  format = "0x%X";
	      PRINTF (format, start, (cpt == VIDEO_mode_max())? cpt : cpt-1);
	      }
	  valid = !valid;
	  }
      }
  if (firsttime)
      PRINT (MSG_NONE);
  PRINT (DOT_CRLF);
#endif
  }
#endif /* VGA_SUPPORT */

#if USER_SUPPORT & JOYSTICK_SUPPORT
extern inline unsigned enable_joystick_when_no_move (void)
  {
  unsigned short Ax, Ay, Bx, By;
  if (JoystickPosition (&Ax, &Ay, &Bx, &By) == 0) {
      JoyStick_init_released (Ax, Ay);
      processor.joystick_present = 1;
      return 1;
      }
    else {
      processor.joystick_present = 0;
      return 0;
      }
  }
#endif

static inline void detect_mouse (void)
  {
#if !(SETUP & QUITE)
#if USER_SUPPORT & (BIOS_MOUSE_SUPPORT | SERIAL_MOUSE_SUPPORT)
  const char *msg;
  unsigned nb = 0;

  if (MOUSE_init(_BIOS_equipment_flags().nb_serial_port) == 0) {
      enum MOUSE_type type = MOUSE_gettype();

      if (type == MOUSE_PS2)
	  msg = BIOS_PS2_MOUSE;
	else if ((type & MOUSE_SERIAL_MASK) == MOUSE_SERIAL_COM1) {
	  msg = SERIAL_ON_COMX_MOUSE;
	  nb = type - MOUSE_SERIAL_COM1;
	  }
	else if ((type & MOUSE_SERIAL_MASK) == MOUSE_SERIAL3_COM1) {
	  msg = SERIAL_3_BUTTON_ON_COMX_MOUSE;
	  nb = type - MOUSE_SERIAL3_COM1;
	  }
	else if ((type & MOUSE_SERIAL_MASK) == MOUSE_SERIALM_COM1) {
	  msg = SERIAL_MOUSESYSTEM_ON_COMX_MOUSE;
	  nb = type - MOUSE_SERIALM_COM1;
	  }
	else
	  msg = ERROR_MOUSE;
      }
    else
      msg = NO_MOUSE;
  PRINTF (msg, nb + 1);
#if USER_SUPPORT & JOYSTICK_SUPPORT
  if (get_gujin_param_attrib().enable_joystick)
      if (enable_joystick_when_no_move())
	  PRINT (JOYSTICK_PRESENT);
#endif
#endif /* MOUSE_SUPPORT */
  PRINT (".\r\n");
#else /* QUITE */
#if USER_SUPPORT & (BIOS_MOUSE_SUPPORT | SERIAL_MOUSE_SUPPORT)
  MOUSE_init(_BIOS_equipment_flags().nb_serial_port);
#endif
#if USER_SUPPORT & JOYSTICK_SUPPORT
  if (get_gujin_param_attrib().enable_joystick)
      enable_joystick_when_no_move();
#endif
#endif /* QUITE */
  }

static inline void detect_interface (void)
  {
  if (BOOT1_COM_port() >= 0) {
#if USER_SUPPORT & SERIAL_SUPPORT
      _BIOS_serialstatus status;
      PRINT (SERIAL_TERMINAL);
      status = _BIOS_getserialstatus(BOOT1_COM_port());
      DBG ((" [ COM%u status: CTS=%u , DSR=%u, RI=%u, CD=%u ] ",
		BOOT1_COM_port() + 1,
		status.al.modem.cts, status.al.modem.dsr,
		status.al.modem.ri, status.al.modem.cd));
      if (SERIAL_init() == 0) {
	  PRINT (UI.info.serial.terminalname);
	  PRINTF (TERMINAL_MODES, UI.info.serial.max_mode);
	  }
	else
	  PUTS (SERIAL_INIT_FAILED);
#endif /* SERIAL_SUPPORT */
      }
    else {
#if USER_SUPPORT & VGA_SUPPORT
      if (VGA_init() == 0) {
	  UI_setmode (UI_getmode() | 0x8000);
//        UI_getsize (UI_getmode(), &UI.parameter);
	  PRINT (VGA_CARD_PRESENT_MODES);
	  listmode (&UI);
	  if (get_gujin_param_attrib().verbose) {
	      unsigned char row, col, cpt;

	      /* Change the bgcolor of the end of the screen: */
	      UI_getcursor (&row, &col);
	      for (cpt = row + 1; cpt < UI.parameter.nbrow; cpt++)
		  puts ("");
	      UI_setcursor (row, col);
	      }
	  }
#endif
#if USER_SUPPORT & VESA_SUPPORT
      if (VESA_init() == 0) {
	  PRINTF (S_CARD_VESA_VERSION_U_U,
		UI.info.vesa.cardname,
		UI.info.vesa.version >> 8,
		UI.info.vesa.version & 0xFF);
#if USER_SUPPORT & VESA_PMINFO
	  if (UI.ProtectedModeIO[0] != 0xFFFF) {
	      unsigned short cpt;

	      PRINT (VESA_USING_IOPORTS);
	      for (cpt = 0; UI.ProtectedModeIO[cpt] != 0xFFFF; cpt ++)
		  PRINTF ("0x%X, ", UI.ProtectedModeIO[cpt]);
	      PRINT ("\010\010.\r\n");
	      }
#endif /* VESA_PMINFO */
	  }
#endif /* VESA_SUPPORT */
      }
  }

static inline unsigned largest_free_himem (void)
  {
  farptr entrypoint;

  DBG ((" [XMS: "));
  entrypoint = _XMS_get_entry_point ();
  DBG (("(at 0x%X) ", entrypoint));
  if ((entrypoint >> 16) != 0) { /* segment not null */
      unsigned short version_bcd, revision, flags, free, largest;
      unsigned char  error1;

      version_bcd = _XMS_get_version (entrypoint, &revision, &flags);
      DBG (("v%u.%u, rev 0x%X, HMA %s; ",
		version_bcd >> 8, version_bcd & 0xFF,
		revision, (flags & 1)? "present" : "abscent"));
      free = _XMS_query_free_memory (entrypoint, &largest, &error1);
      DBG (("free mem: %u Kb, largest: %u Kb, error: 0x%X; ",
		free, largest, error1));
      if ((version_bcd >> 8) >= 3) {
	  unsigned largest_KB, highest_memory_address, total_KB_free;
	  unsigned char error2;

	  total_KB_free = _XMSv3_query_free_memory (entrypoint,
			&largest_KB, &highest_memory_address, &error2);
	  DBG (("V3: total free: %u Kb, largest: %u Kb,"
		" highest address: 0x%X, error: 0x%X, ",
		total_KB_free, largest_KB,
		highest_memory_address, error2));
	  if (error2 == 0) {
	      DBG (("use V3 %u Kb] ", largest_KB));
	      return largest_KB;
	      }
	    else if (error1 == 0) {
	      DBG (("V3 error happend, use short value: %u Kb] ",
			largest));
	      return largest;
	      }
	    else {
	      DBG (("two error happened, returns 0] "));
	      return 0;
	      }
	  }
	else {
	  DBG (("old version, use short value: %u Kb] ", largest));
	  return largest;
	  }
      }
    else {
      DBG (("invalid address] "));
      return 0;
      }
  }

#if USER_SUPPORT != 0

static inline void APM_power_OFF (void)
  {
  unsigned short BCD_version, flags, tmp;

  tmp = APM_installation_check (&BCD_version, &flags);
  DBG (("\r\n%s: check result 0x%X, version 0x%X, flags 0x%X\r\n",
	__FUNCTION__, tmp, BCD_version, flags));
  if (tmp != 0 || BCD_version < 0x102)
      DBG (("So cannot Power Down (works only with APM v1.2+)\r\n"));
    else {
      tmp = APM_connect_realmode();
      if (tmp != 0)
	  DBG (("APM_connect_realmode failed 0x%X\r\n", tmp));
	else {
	  tmp = APM_declare_driver_version (0x102);
	  if (tmp != 0)
	      DBG (("APM_declare_driver_version(0x102) refused: 0x%X\r\n",
			tmp));
	    else {
	      tmp = APM_state (APM_alldevice, APM_off);
	      if (tmp != 0)
		  DBG (("APM_state(OFF) failed 0x%X\r\n", tmp));
	      /* If we return there for any reason: */
	      tmp = APM_state (APM_alldevice, APM_ready);
	      if (tmp != 0)
		  DBG (("APM_state(ON) failed 0x%X\r\n", tmp));
	      }
	  tmp = APM_disconnect();
	  if (tmp != 0)
	      DBG (("APM_disconnect failed: 0x%X\r\n", tmp));
	  }
      }
  }

#if USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT)
static inline unsigned get_refresh_freq (void)
  {
  unsigned timeout = 100000;
  unsigned char ioval;
  unsigned short addr = 0x3DA;	/* 0x3BA for MDA */

  if (processor.calibrate_rdtsc == 0)
      return 0;
  ioval = inb (addr) & 0x08;
  while (--timeout && ioval == (inb (addr) & 0x08)) ;
  if (timeout == 0)
      return 0;
    else {
      unsigned long long rdtsc_start;
      unsigned total_rdtsc;

      ioval = inb (addr) & 0x08;
      while (ioval == (inb (addr) & 0x08)) ;
      rdtsc_start = rdtsc();
      while (ioval != (inb (addr) & 0x08)) ;
      while (ioval == (inb (addr) & 0x08)) ;
      total_rdtsc = rdtsc() - rdtsc_start;
      return (processor.calibrate_rdtsc + total_rdtsc/2) / total_rdtsc;
      }
  }

static inline void detect_screen(void)
  {
#if !(SETUP & QUITE)
  const char *msg;

  PRINT (SCREEN_INTERFACE);
  UI.monitor = BIOS_probemonitor();

  if (UI.monitor == 0)
      msg = UNKNOWN_SCREEN;
    else if (UI.monitor == 1)
      msg = BLACKWHITE_SCREEN;
    else if (UI.monitor == 2)
      msg = MONOCHROME_SCREEN;
    else if (UI.monitor == 3)
      msg = COLOR_SCREEN;
    else
      msg = COLOR_SCREEN_TYPE_X;
  PRINTF (msg, UI.monitor);

#if USER_SUPPORT & VESA_EDID
  if (UI.edid.horizontal_resolution != 0) {
      PRINTF (EDID_TYPE,
		UI.edid.horizontal_resolution,
		UI.edid.vertical_resolution,
		UI.edid.frequency,
		(UI.edid.Vsize_cm + 3 * UI.edid.Hsize_cm / 4 + 2) / 3
		);
      }
#endif /* VESA_EDID */
#else /* QUITE */
  UI.monitor = BIOS_probemonitor();
#endif /* QUITE */
  }
#endif /* VGA_SUPPORT | VESA_SUPPORT */

#if !(SETUP & QUITE)
static const char * __attribute__ ((const)) PartName (unsigned char parttype)
  {
  const char *partname;

#define P(ident, nb, str) if (parttype == nb) partname = str; else
	  PARTITIONLIST()	partname = "unknown";
#undef P
  return partname;
  }
#endif /* QUITE */

/**
 ** The MINIDEBUG functions:
 ** All messages stay in English here.
 **/
#if SETUP & MINIDEBUG
static inline void menu_list_memory (void)
  {
  unsigned i, total;

  report_malloc();
  total = DI.nbdisk * sizeof (struct diskparam_str);
  printf ("Disk structure: %u bytes at 0x%X\r\n",
		total, (unsigned)DI.param);
  for (i = 0; i < DI.nbdisk; i++) {
#if defined (NB_FREELIST) && (NB_FREELIST != 0)
      int tmp = DI.param[i].nbpartition * sizeof (struct partition_str)
		+ DI.param[i].nbfreelist * sizeof (struct freelist_str);
      printf ("disk %u: %u partition at 0x%X, %u freelist at 0x%X"
		" => %u bytes\r\n",
		i,
		DI.param[i].nbpartition,
		(unsigned)DI.param[i].partition,
		DI.param[i].nbfreelist,
		(unsigned)DI.param[i].freelist,
		tmp);
#else /* NB_FREELIST */
      int tmp = DI.param[i].nbpartition * sizeof (struct partition_str);
      printf ("disk %u: %u partition at 0x%X => %u bytes\r\n",
		i,
		DI.param[i].nbpartition,
		(unsigned)DI.param[i].partition,
		tmp);
#endif /* NB_FREELIST */
      total += tmp;
      }
  printf ("total used for disk: %u bytes\r\n", total);
  printf ("total used for boot_system: %u bytes at 0x%X\r\n",
		(boot_system.nb + boot_system.nb_initrd)
			* sizeof(struct desc_str),
		(unsigned)boot_system.desc);
#if USER_SUPPORT != 0
  printf ("total used for VGA modes: %u bytes at 0x%X\r\n",
		UI.nbmode * sizeof (struct videomode_str), UI.mode);
#endif
  }


#if USER_SUPPORT & VESA_HARDWINDOW
static inline void
display_hardwin (unsigned short *offsetarray, struct hardwindow_str *hardwindow)
  {
  unsigned short i, j;

  if (offsetarray)
      printf ("Win0 [%u]: ", offsetarray[0]);
    else
      print ("AllWin: ");
  if (hardwindow == 0) {
      print ("NO hardwin.\r\nEnd.");
      return;
      }
  for (i = j = 0;; i++) {
      if (   hardwindow[i].instruction == 0
	  && hardwindow[i].data == 0
	  && hardwindow[i].port == 0) {
	  j++;
	  if (offsetarray == 0 || offsetarray[j+1] == 0) {
	      /* end array kept, end marker written after */
	      print ("\r\nEnd.");
	      break;
	      }
	    else
	      printf ("\r\nWin%u [%u]: ", j, offsetarray[j]);
	  }
	else {
	  unsigned short instruction = hardwindow[i].instruction;
	  unsigned char tmp;

#ifdef COMPRESS_HARDWIN
	  const char *str = 0;

	  if (instruction != 0xEF66 && (tmp = (instruction >> 8)) != 0) {
	      if (tmp == 1)
		  str = "(page & 0x0F)";
		else if (tmp == 3)
		  str = "((page & 0x0F) << 4)";
		else if (tmp == 19)
		  str = "page";
		else if (tmp == 20)
		  str = "(page >> 4)";
		else if (tmp == 28)
		  str = "(page << 2)";
		else if (tmp == 67)
		  str = "((page & 0xF0) >> 2)";
#ifdef COMPRESS_EXTENDED
		else if ((tmp == 37) && ((instruction & 0xFF) == 0xEE))
		  str = "((inb(0x%X) & 0xF0) ^ (page & 0x0F))";
		else if ((tmp == 11) && ((instruction & 0xFF) == 0xEE))
		  str = "((inb(0x%X) & 0x0F) ^ ((page & 0x0F) << 4))";
#endif
		else
		  str = "ERROR ";
	      instruction &= 0xFF;
	      }
#endif /* COMPRESS_HARDWIN */

	  if (instruction == 0xEF66)
	      tmp = 'l';
	    else if (instruction == 0xEE)
	      tmp = 'b';
	    else if (instruction == 0xEF)
	      tmp = 'w';
	    else
	      tmp = '?';

	  printf ("out%c (0x%X, 0x%X",
		  tmp, hardwindow[i].port, hardwindow[i].data);
#ifdef COMPRESS_HARDWIN
	  if (str) {
	      printf (" ^ ");
	      if (tmp == 'b')
#ifdef COMPRESS_EXTENDED
		  printf (str, hardwindow[i].port);
#else
		  print (str);
#endif
		else
		  printf ("(%s << 8)", str);
	      }
#endif /* COMPRESS_HARDWIN */
	  print ("); ");
	  }
      }
  }

static inline void menu_list_hardwin (void)
  {
  unsigned short *offsetarray = UI.parameter.offsetarray;
  struct hardwindow_str *hardwindow = UI.parameter.hardwindow;

  for (;;) {
#if USER_SUPPORT & VESA_2WINDOWS
      if (UI.parameter.readhardwindow != 0) {
	  if (hardwindow == UI.parameter.hardwindow)
	      print ("Write:\r\n");
	    else /* if (hardwindow == UI.parameter.readhardwindow) */
	      print ("\r\nRead:\r\n");
	  }
	else
#endif
	  print ("Read/Write:\r\n");

      /* call display_hardwin() only once, inline fct */
      display_hardwin (offsetarray, hardwindow);

#if USER_SUPPORT & VESA_2WINDOWS
      if (UI.parameter.readhardwindow != 0) {
	  if (hardwindow == UI.parameter.hardwindow) {
	      offsetarray = UI.parameter.readoffsetarray;
	      hardwindow = UI.parameter.readhardwindow;
	      }
	    else /* if (hardwindow == UI.parameter.readhardwindow) */
	      return;
	  }
	else
#endif
	  return;
      }
  }

#else /* VESA_HARDWINDOW */

static inline void menu_list_hardwin (void)
  {
  print ("Hardwin not compiled-in.\r\n");
  }

#endif /* VESA_HARDWINDOW */

static inline void menu_list_valid_disk (void)
  {
  unsigned short cpt;

#if DISK_SUPPORT & IDE_SUPPORT
  for (cpt = 0; DI.IDE_IObase_found[cpt] != 0; cpt++) ;
  printf ("Found %u IDE interfaces: ", cpt);
  {
  unsigned short i;
  for (i = 0; i < cpt; i++)
      printf ("0x%X, ", DI.IDE_IObase_found[i]);
  }
  puts ("end.");
#else
  puts ("No IDE found: not compiled in.");
#endif

  printf ("Found %u valid disks:\r\n\r\n", DI.nbdisk);
  for (cpt = 0; cpt < DI.nbdisk; cpt ++) {
      unsigned i;
      struct diskparam_str *dps = &DI.param[cpt];
      char disk_name[SIZE_DISKNAME];

      diskname (dps, disk_name);
      printf ("\r%s", disk_name);

      if (dps->ebiosversion != 0) {
	  printf (" [EBIOS V0x%X", dps->ebiosversion);
	  if (dps->ideIOadr != 0 && dps->irq != 0)
	      printf (" I/O: 0x%X, IRQ: %u, %s] ", dps->ideIOadr, dps->irq,
			dps->ismaster ? "master" : "slave");
	    else
	      print ("] ");
	  }
      if (dps->access == dos_part)
	  printf (", SectorPerCluster: %u, ClustersOnDisk: %u, "
		  "MediaDescriptorByte: 0x%X",
		  dps->nbsectorpertrack, dps->nbcylinder, dps->nbhead);
	else
	  printf (", C/H/S: %u/%u/%u",
		  dps->nbcylinder, dps->nbhead, dps->nbsectorpertrack);
      printf (", total sector: %u, sector size: %u",
		(unsigned)dps->nbtotalsector, /* long long */
		dps->bytepersector);

      if (!dps->bootable) {
	  print (", Disk not bootable.\r\n");
	  continue;
	  }

      printf (", bootable, %u partitions:", dps->nbpartition);
      for (i = 0; i < dps->nbpartition; i++) {
	  struct partition_str *part = &dps->partition[i];

	  printf (" %u/OS%u: ", part->misc.order, part->OSnumber);
	  if (   part->start == 0
	      && part->length == 0
	      && part->type == EMPTY) {
	      print ("[empty]");
	      continue;
	      }
	  if (part->active == part_active)
	      print ("[active, ");
	    else if (part->active == part_inactive)
	      print ("[inactive, ");
	    else {
	      print ("[invalid]");
	      continue;
	      }
	  printf ("%u/%u, %s]", part->start, part->length,
#if USER_SUPPORT != 0
			PartName (part->type)
#else
			""
#endif
			);
	  }
      print ("\r\n");

#if defined (NB_FREELIST) && (NB_FREELIST != 0)
      if (dps->nbfreelist != 0) {
	  print ("Free list: ");
	  for (i = 0; i < dps->nbfreelist; i++)
	      printf (" [%u sectors at %u] ",
			dps->freelist[i].length, dps->freelist[i].start);
	  print ("\r\n");
	  }
#endif /* NB_FREELIST */
      print ("\r\n");
      }
  }

static inline void menu_list_video_modes (void)
  {
#if USER_SUPPORT != 0
  unsigned short cpt;
  printf ("%u video modes:\r\n", UI.nbmode);
  for (cpt = 0; cpt < UI.nbmode; cpt++)
      printf ("mode 0x%X: %s %ux%u, %u bpp, %s, %sattributes\r\n",
		UI.mode[cpt].number,
		UI.mode[cpt].text? "text" : "graphic",
		UI.mode[cpt].width,
		UI.mode[cpt].heigh,
		UI.mode[cpt].bpp,
		UI.mode[cpt].vesa? "VESA" : "VGA/SERIAL",
		UI.mode[cpt].attr? "" : "no ");
#endif
  }

static inline void menu_list_boot_images (void)
  {
  unsigned short cpt;
  printf ("%u boot images:\r\n", boot_system.nb);
  for (cpt = 0; cpt < boot_system.nb; cpt++)
      printf ("%s, size %u, disk %u, partition %u, name_offset %u, attr %c%c%c\r\n",
		boot_system.desc[cpt].filename,
		boot_system.desc[cpt].filesize,
		boot_system.desc[cpt].disk,
		boot_system.desc[cpt].partition,
		boot_system.desc[cpt].name_offset,
		boot_system.desc[cpt].in_slashboot? 'B' : ' ',
		boot_system.desc[cpt].is_kernel? 'K' : ' ',
		boot_system.desc[cpt].has_vmlinuz_header? 'H' : ' '
		);
  printf ("%u initrd:\r\n", boot_system.nb_initrd);
  for (cpt = 0; cpt < boot_system.nb_initrd; cpt++)
      printf ("%s, size %u, disk %u, partition %u, name_offset %u, attr %c%c%c\r\n",
		boot_system.desc[boot_system.nb + cpt].filename,
		boot_system.desc[boot_system.nb + cpt].filesize,
		boot_system.desc[boot_system.nb + cpt].disk,
		boot_system.desc[boot_system.nb + cpt].partition,
		boot_system.desc[boot_system.nb + cpt].name_offset,
		boot_system.desc[boot_system.nb + cpt].in_slashboot? 'B' : ' ',
		boot_system.desc[boot_system.nb + cpt].is_kernel? 'K' : ' ',
		boot_system.desc[boot_system.nb + cpt].has_vmlinuz_header? 'H' : ' '
		);
  }

static inline void tstCDbootcatalog (void)
  {
  unsigned boot_catalog[800 / 4] = {};
  CDROMcatalog catal = {
	packetsize:	8,
	nbsector:	1,
	buffer:		boot_catalog, /* ds:si */
	firstsector:	0
	};
  unsigned short returned_code;
  unsigned char ret;

  printf ("_BIOSCDROM_getcatalog at 0x%X: ", catal.buffer);
  ret = _BIOSCDROM_getcatalog (&catal, &returned_code);
  printf ("returns 0x%X, code: 0x%X\r\n", ret, returned_code);
  printf ("packetsize: %u, nbsector: %u, buffer at 0x%X, "
	  "firstsector at 0x%X\r\n",
	  catal.packetsize, catal.nbsector, catal.buffer,
	  catal.firstsector);
  printf ("boot_catalog[0..3]: 0x%X, 0x%X, 0x%X, 0x%X.\r\n",
		boot_catalog[0], boot_catalog[1],
		boot_catalog[2], boot_catalog[3]);
  printf ("boot_catalog[4..7]: 0x%X, 0x%X, 0x%X, 0x%X.\r\n",
		boot_catalog[4], boot_catalog[5],
		boot_catalog[6], boot_catalog[7]);
  printf ("boot_catalog[8..11]: 0x%X, 0x%X, 0x%X, 0x%X.\r\n",
		boot_catalog[8], boot_catalog[9],
		boot_catalog[10], boot_catalog[11]);
  }

static void tstATIvideo (unsigned char mode)
  {
  struct ATI_mode_table mode_table;
  farptr mode_table_farptr = _ATI_alt_funct_select_get_mode_table (mode);
  unsigned bpp, width, heigh;

  if ((mode_table_farptr & 0xFFFF) == 0xFFFF)
      printf ("_ATI_alt_funct_select_get_mode_table(0x%X) error\r\n", mode);
    else {
      lmemcpy (&mode_table, mode_table_farptr, sizeof (mode_table));
      asm volatile ("");
//      if (mode_table.graphic_controller.mode.enable_256_color)
      if (mode_table.attribute_controller.mode.PELCLK_div_2)
	  bpp = 8;
	else if (mode_table.attribute_controller.mode.monochrome)
	  bpp = 0;
	else if (mode_table.graphic_controller.mode.shift_register_mode)
	  bpp = 2; /* CGA graphic */
	else if (   mode_table.attribute_controller.color_enable.enable_plane_3
		 && mode_table.crtc.mode_control.address_wrap)
	  bpp = 4; /* could check also graphic_controller.color_dont_care == 0x0F */
	else
	  bpp = 1; /* could check palette, most at 0x3F */
      width = (mode_table.crtc.horizontal_displayed_chars + 1) * 8;
      if (mode_table.attribute_controller.mode.PELCLK_div_2)
	  width /= 2;
      heigh = mode_table.crtc.vertical_display_end + 1
		+ mode_table.crtc.overflow.vertical_display_end_8 * 256
		+ mode_table.crtc.overflow.vertical_display_end_9 * 512;
      if (   mode_table.crtc.maximum_row_address.enable_double_scan
	  || mode_table.attribute_controller.mode.PELCLK_div_2)
	  heigh /= 2;
      if (!mode_table.graphic_controller.miscellaneous.graphic) {
	  width = mode_table.number_of_columns;
	  heigh = mode_table.maximum_row + 1;
	  }
      printf ("get_mode_table 0x%X success, %s, %ux%u*%u, charheigh = %u",
		mode, mode_table.attribute_controller.mode.graphic
			? "graphic" : "text",
		width, heigh, bpp, mode_table.scan_line_per_row);
      printf (" [%u,%u,%u,%u] ",
		mode_table.attribute_controller.color_enable.enable_plane_0,
		mode_table.attribute_controller.color_enable.enable_plane_1,
		mode_table.attribute_controller.color_enable.enable_plane_2,
		mode_table.attribute_controller.color_enable.enable_plane_3);
      puts ("");
      }
  }

#endif /* MINIDEBUG */

/**
 ** Video mode selection with "intuitive" interface.
 **/
static inline int
prev_next_heigh (unsigned next, unsigned char currbpp,
		 struct gujin_param_attrib gujin_attr)
  {
  struct videomode_str *array;
  unsigned char cpt;
  int found = -1;

  for (array = UI.mode, cpt = 0; cpt < UI.nbmode; cpt++, array++) {
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
      if (array->vesa? !gujin_attr.VESA_interface
		     : !gujin_attr.VGA_interface)
	  continue;
#endif
      if (gujin_attr.refuse_8x14font && array->font8x14)
	  continue;
      if (   gujin_attr.lock_bpp
	  && array->bpp != currbpp)
	  continue;
      if (   gujin_attr.lock_text_graphic
	  && array->text != !UI.parameter.attr.isgraphic)
	  continue;
      if (array->width != UI.parameter.width)
	  continue;
      if (array->heigh == UI.parameter.heigh)
	  continue;
      if ((array->heigh < UI.parameter.heigh) ^ !next)
	  continue;
      if (found == -1)
	  found = cpt;
	else if (array->heigh == UI.mode[found].heigh) {
	  if (ABS((int)currbpp - array->bpp)
				< ABS((int)currbpp - UI.mode[found].bpp))
	      found = cpt;
	  }
	else if ((array->heigh < UI.mode[found].heigh) ^ !next)
	  found = cpt;
      }
  return found;
  }

static inline int
prev_next_width (unsigned next, unsigned char currbpp,
		 struct gujin_param_attrib gujin_attr)
  {
  struct videomode_str *array;
  unsigned char cpt;
  int found = -1;

  for (array = UI.mode, cpt = 0; cpt < UI.nbmode; cpt++, array++) {
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
      if (array->vesa? !gujin_attr.VESA_interface
		     : !gujin_attr.VGA_interface)
	  continue;
#endif
      if (gujin_attr.refuse_8x14font && array->font8x14)
	  continue;
      if (   gujin_attr.lock_bpp
	  && array->bpp != currbpp)
	  continue;
      if (   gujin_attr.lock_text_graphic
	  && array->text != !UI.parameter.attr.isgraphic)
	  continue;
      if (array->width == UI.parameter.width)
	  continue;
      if ((array->width < UI.parameter.width) ^ !next)
	  continue;
      if (found == -1)
	  found = cpt;
	else if (array->width == UI.mode[found].width) {
	  if (array->heigh == UI.mode[found].heigh) {
	      if (ABS((int)currbpp - array->bpp)
			< ABS((int)currbpp - UI.mode[found].bpp))
		  found = cpt;
	      }
	    else if ((array->heigh < UI.mode[found].heigh) ^ !next)
	      found = cpt;
	  }
	else if ((array->width < UI.mode[found].width) ^ !next)
	  found = cpt;
      }
  return found;
  }

static int
prev_next_definition (unsigned next, struct gujin_param_attrib gujin_attr)
  {bound_stack();{
  unsigned currbpp = get_bpp (&UI.parameter);
  int found;

  found = prev_next_heigh (next, currbpp, gujin_attr);
  if (found == -1)
      found = prev_next_width (next, currbpp, gujin_attr);
  return found;
  }}

static int
prev_next_color (unsigned next, struct gujin_param_attrib gujin_attr)
  {bound_stack();{
  struct videomode_str *array;
  unsigned currbpp = get_bpp (&UI.parameter);
  unsigned char cpt;
  int found = -1;

  for (array = UI.mode, cpt = 0; cpt < UI.nbmode; cpt++, array++) {
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
      if (array->vesa? !gujin_attr.VESA_interface
		     : !gujin_attr.VGA_interface)
	  continue;
#endif
      if (gujin_attr.refuse_8x14font && array->font8x14)
	  continue;
      if (array->width != UI.parameter.width)
	  continue;
      if (array->heigh != UI.parameter.heigh)
	  continue;
      if (array->bpp == currbpp)
	  continue;
      if ((array->bpp < currbpp) ^ !next)
	  continue;
      if (found == -1)
	  found = cpt;
	else if (array->bpp == UI.mode[found].bpp)
	  continue; /* take only the first one */
	else if ((array->bpp < UI.mode[found].bpp) ^ !next)
	  found = cpt;
      }
  return found;
  }}

static int
prev_next_alternate (struct gujin_param_attrib gujin_attr)
  {bound_stack();{
  struct videomode_str *array;
  unsigned currbpp = get_bpp (&UI.parameter);
  unsigned char cpt;
  int found = -2, first= -1;

  for (array = UI.mode, cpt = 0; cpt < UI.nbmode; cpt++, array++) {
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
      if (array->vesa? !gujin_attr.VESA_interface
		     : !gujin_attr.VGA_interface)
	  continue;
#endif
      if (gujin_attr.refuse_8x14font && array->font8x14)
	  continue;
      if (array->width != UI.parameter.width)
	  continue;
      if (array->heigh != UI.parameter.heigh)
	  continue;
      if (array->bpp != currbpp)
	  continue;
      if (array->number == UI.parameter.identification)
	  found = -1;
	else if (first == -1)
	  first = cpt;	/* if none after, start with first */
	else if (found == -1)
	  return cpt;	/* return the following mode */
      }
  return first;
  }}

static inline int best_max_width (unsigned short maxwidth,
				  struct gujin_param_attrib gujin_attr)
  {
  struct videomode_str *array;
  unsigned char cpt;
  int found = -1;

  for (array = UI.mode, cpt = 0; cpt < UI.nbmode; cpt++, array++) {
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
      if (array->vesa? !gujin_attr.VESA_interface
		     : !gujin_attr.VGA_interface)
	  continue;
#endif
      if (gujin_attr.refuse_8x14font && array->font8x14)
	  continue;
      if (array->width > maxwidth)
	  continue;
      if (found == -1)
	  found = cpt;
	else if (array->width >= UI.mode[found].width) {
	  if (   array->heigh > UI.mode[found].heigh
	      || array->bpp > UI.mode[found].bpp)
	      found = cpt;
	  }
      }
  return found;
  }

static inline int next_lower_quality (unsigned short maxheigh,
				      unsigned short maxwidth,
				      unsigned short maxbpp,
				      struct gujin_param_attrib gujin_attr)
  {
  struct videomode_str *array;
  unsigned char cpt;
  int found = -1;

  for (array = UI.mode, cpt = 0; cpt < UI.nbmode; cpt++, array++) {
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
      if (array->vesa? !gujin_attr.VESA_interface
		     : !gujin_attr.VGA_interface)
	  continue;
#endif
      if (gujin_attr.refuse_8x14font && array->font8x14)
	  continue;
      if (array->heigh > maxheigh)
	  continue;
      if (array->width > maxwidth)
	  continue;
      if (array->bpp > maxbpp)
	  continue;
      if (found == -1)
	  found = cpt;
	else if (   array->heigh >= UI.mode[found].heigh
		 || array->width >= UI.mode[found].width
		 || array->bpp >= UI.mode[found].bpp)
	  found = cpt;
      }
  return found;
  }

static inline int switch_text_graphic (struct gujin_param_attrib gujin_attr)
  {
  unsigned short tmp;
  unsigned char cpt;

  if (UI.parameter.attr.isgraphic)
      tmp = getDefaultTextVideoMode();
    else
      tmp = getDefaultGraphicVideoMode();
  if (tmp != 0xFFFF) {
      for (cpt = 0; cpt < UI.nbmode; cpt++) {
	  if (gujin_attr.refuse_8x14font && UI.mode[cpt].font8x14)
	      continue;
	  /* exclude clear screen and linear bits: */
	  if (((UI.mode[cpt].number ^ tmp) & 0x3FF) == 0) {
	      return cpt;
	      }
	  }
      }
  for (cpt = 0; cpt < UI.nbmode; cpt++) {
      if (gujin_attr.refuse_8x14font && UI.mode[cpt].font8x14)
	  continue;
      if (UI.mode[cpt].text ^ !UI.parameter.attr.isgraphic)
	  break;
      }
  if (cpt < UI.nbmode)
      return cpt;
  return -1;
  }

static unsigned
change_video_mode (unsigned char key, unsigned update_bootparam)
  {bound_stack();{
  int found;
  unsigned result, mode;
  struct gujin_param_attrib gujin_attr = get_gujin_param_attrib ();

#if USER_SUPPORT & VESA_SUPPORT
  unsigned short linear_vesa;

  if ((UI.info.vesa.version >> 8) >= 2 && gujin_attr.VESA2_interface)
      linear_vesa = 0x4000;
    else
      linear_vesa = 0;
#else /* VESA_SUPPORT */
  /* inline */ const unsigned short linear_vesa = 0;
#endif /* VESA_SUPPORT */

#if USER_SUPPORT & SERIAL_SUPPORT
  if (key == ',')
      key = '+'; /* VTxxx keyboard has no '+' on numeric pad */
#endif

  if (key == '+' || key == '-')
      found = prev_next_definition (key == '+', gujin_attr);
    else if (key == '*' || key == '/')
      found = prev_next_color (key == '*', gujin_attr);
    else if (key == '\t')
      found = prev_next_alternate (gujin_attr);
    else /* key == '.' */
      found = switch_text_graphic (gujin_attr);

  if (found < 0) {
      PRINT (NO_MODE_FOUND);
      return 1;
      }
#if USER_SUPPORT & VESA_EDID
  if (   (   UI.edid.horizontal_resolution
	  && UI.edid.vertical_resolution)
      && (   UI.mode[found].width > UI.edid.horizontal_resolution
	  || UI.mode[found].heigh > UI.edid.vertical_resolution)) {
      PRINTF (MODE_X_U_U_OVER_MONITOR_SPEC,
		UI.mode[found].number,
		UI.mode[found].width,
		UI.mode[found].heigh);
      return 2;
      }
#endif /* VESA_EDID */
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
  if (UI.mode[found].vesa) {
      if (VGA_ACTIVE()) {
	  if (VESA_init() != 0) {
	      PUTS (ERROR_VESA_SWITCH_FAILED);
	      return 3;
	      }
	  }
      }
    else {
      if (VESA_ACTIVE() && BOOT1_COM_port() < 0) {
	  if (VGA_init() != 0) {
	      PUTS (ERROR_VGA_SWITCH_FAILED);
	      return 4;
	      }
	  }
      linear_vesa = 0;
      }
#endif /* VESA_SUPPORT && VGA_SUPPORT */

  /* do not clear the screen : */
  mode = UI.mode[found].number | linear_vesa | 0x8000;
  PRINTF (SETTING_MODE_X, UI.mode[found].number | linear_vesa);

  result = UI_setmode (mode);
  if (result == 0) {
      mode &= ~0x8000;
      if (update_bootparam) {
	  setDefaultVideoMode (mode);
	  if (UI.parameter.attr.isgraphic)
	      setDefaultGraphicVideoMode (mode);
	    else
	      setDefaultTextVideoMode (mode);
	  }
      return 0;
      }
    else {
      PRINTF (ERROR_SETTING_X, result);
      return 5;
      }
  }}

static void menu_reset_saved_parameters (void)
  {
  struct gujin_param_attrib gujin_attr;

  gujin_attr = get_gujin_param_attrib();
#define Apply(field, deflt, grey)	gujin_attr.field = deflt;
	APPLY_ATTRIBUTE_ORDER(Apply)
#undef Apply
  set_gujin_param_attrib (gujin_attr);
  set_timeout_autoload(0);

#if USER_SUPPORT & VGA_SUPPORT
  VIDEO_mode_reset();
#endif
  }

/**
 ** Decide which mode to set initially, never re-used after init
 **/
static void set_initial_mode (struct gujin_param_attrib gujin_attr)
  {bound_stack();{
  unsigned short mode, tmp, nb_inch;

  if ((mode = getDefaultVideoMode()) != 0xFFFF) {
      if ((tmp = UI_setmode (mode)) == 0) {
#if (SETUP & UPDATE_BOOTPARAM) && (USER_SUPPORT & VGA_SUPPORT)
	  restore_gujin_param_vga_mode();
#endif
	  return;
	  }
      DBG (("Setting default mode 0x%X failed: 0x%X\r\n", mode, tmp));
      }

  if (BOOT1_COM_port() >= 0) {
      tmp = UI_setmode (0);
      if (tmp != 0)
	  DBG (("Error setting serial mode 0: 0x%X\r\n", tmp));
      return;
      }

#if (USER_SUPPORT & VGA_SUPPORT) && (USER_SUPPORT & VESA_SUPPORT)
  if (/* BOOT1_COM_port() < 0 && */
      mode < 0x100 && VIDEO_mode_is_valid(mode)
      /* && mode != 0xFFFF && UI_setmode == VESA_setmode */) {
      DBG (("Switching back to VGA\r\n"));
      VGA_init();
      if ((tmp = UI_setmode (mode)) == 0) {
#if (SETUP & UPDATE_BOOTPARAM) && (USER_SUPPORT & VGA_SUPPORT)
	  restore_gujin_param_vga_mode();
#endif
	  return;
	  }
      DBG (("Setting default VGA mode 0x%X failed: 0x%X\r\n"
	    "Switching back to VESA\r\n", mode, tmp));
      VESA_init();
      }
#endif /* (USER_SUPPORT & VGA_SUPPORT) && (USER_SUPPORT & VESA_SUPPORT) */

  /* That is the first power-on mode selection: */

#if USER_SUPPORT & VESA_EDID
  if (UI.edid.Hsize_cm != 0 && UI.edid.Vsize_cm != 0) {
      /*
       * After asking to a professionnal mathematician, we
       * suppose a display size approx 4/3 :
       *  sqrt(sqr(Vsize_cm) + sqr(Hsize_cm)) == 5/3*Vsize_cm
       * 50 inches = 50 * 2.54 = 127 cm ; and 50 * 5 == 2 * 127
       * We round up the result.
       */
      nb_inch = (UI.edid.Vsize_cm + 3 * UI.edid.Hsize_cm / 4 + 2) / 3;
      }
    else
#endif /* VESA_EDID */
      {
      puts (NON_EDID_SCREEN);
      print (PRESS_KEY_TO_CONTINUE);
      UI_getkey (30 * TICKS_PER_SECOND);
      puts ("\r");
      nb_inch = 13;
      }

#if USER_SUPPORT & VESA_SUPPORT
  if (UI.info.vesa.version != 0) {
      unsigned short maxwidth;
      int found;

      if (nb_inch < 14)
	  maxwidth = 640;
	else if (nb_inch < 15)
	  maxwidth = 800;
	else if (nb_inch < 16)
	  maxwidth = 1024;
	else if (nb_inch < 18)
	  maxwidth = 1152;
	else
	  maxwidth = 1280;

#if USER_SUPPORT & VESA_EDID
      if (   UI.edid.Hsize_cm != 0
	  && UI.edid.Vsize_cm != 0
	  && maxwidth > UI.edid.horizontal_resolution)
	  maxwidth = UI.edid.horizontal_resolution;
#endif /* VESA_EDID */

      found = best_max_width (maxwidth, gujin_attr);
      while (found != -1) {
	  unsigned short mode = UI.mode[found].number;
	  /* UI.mode[] updated if setmode error: */
	  unsigned short maxheigh = UI.mode[found].heigh,
			 maxbpp = UI.mode[found].bpp;
	  maxwidth = UI.mode[found].width;

	  if ((UI.info.vesa.version >> 8) >= 2)
	      mode |= 0x4000;
	  tmp = UI_setmode (mode);
	  if (tmp) {
	      DBG (("Setting mode 0x%X failed: 0x%X\r\n", mode, tmp));
	      if (mode & 0x4000) {
		  mode &= ~0x4000;
		  tmp = UI_setmode (mode);
		  if (tmp)
		      DBG (("Setting mode 0x%X failed: 0x%X\r\n", mode, tmp));
		  }
	      }
	  if (tmp == 0) {
	      setDefaultVideoMode (mode);
	      if (UI.parameter.attr.isgraphic)
		  setDefaultGraphicVideoMode (UI.mode[found].number);
		else
		  setDefaultTextVideoMode (UI.mode[found].number);
	      return;
	      }
	  found = next_lower_quality (maxheigh, maxwidth, maxbpp, gujin_attr);
	  }
      DBG (("No even one good VESA mode found!\r\n"));
#if USER_SUPPORT & VGA_SUPPORT
      VGA_init();
#else
      print ("PANIC: recompile with VGA_SUPPORT!\r\n");
#endif
      }
#endif /* VESA_SUPPORT */

  /* No default mode selected and no working VESA */
  UI_setmode (3);

#if (SETUP & UPDATE_BOOTPARAM) && (USER_SUPPORT & VGA_SUPPORT)
  setDefaultVideoMode (3);
  setup_gujin_param_vga_mode ();
#endif
  }}

/**
 ** Menu management:
 **/
static void centered_print (const char *str, unsigned row)
  {
  unsigned nbcol = UI.parameter.nbcol;
  char line[nbcol + 1];
  char *const end = &line[nbcol];

  while (*str != '\0') {
      char *ptr = line;
      while (*str != '\0') {
	  if (*str == '\r') {
	      str++;
	      continue;
	      }
	  if (*str == '\n') {
	      str++;
	      break;
	      }
	  if (ptr >= end)
	      return; /* too long message */
	  *ptr++ = *str++;
	  }
      *ptr = '\0';
      UI_setcursor (row++, (nbcol - strlen(line)) / 2);
      print (line);
      }
  }

static
#if USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT)
inline
#endif
void draw_box (unsigned baserow, unsigned heigh)
  {
  if (UI.parameter.attr.graphicbgcolor) {
      unsigned short xend = UI.parameter.width - UI.parameter.charheigh/2;
      coord xy = {x: UI.parameter.charheigh / 2,
		  y: baserow * UI.parameter.charheigh
			 - UI.parameter.charheigh/2};

      if (baserow == 1) {
	  /* The top, the half line over the first line: */
	  coord xytmp = { x: 0, y: 0 };
	  while (xytmp.y < UI.parameter.charheigh / 2) {
	      UI_plotHline (xytmp, UI.parameter.width,
					UI.stdcolor[black]);
	      xytmp.y++;
	      }
	  }
      do {
	  coord xytmp = { x: 0, y: xy.y };
	  UI_plotHline (xytmp, xy.x, UI.stdcolor[black]);
	  UI_plotHline (xy, xend, UI.bgcolor);
	  xytmp.x = xend;
	  UI_plotHline (xytmp, UI.parameter.width,
					UI.stdcolor[black]);
	  xy.x--; xy.y++; xend++;
	  } while (xy.x != 0);
      if (VESA_ACTIVE()) {
	  heigh *= UI.parameter.charheigh;
	  if (UI.parameter.nbcolor == 16)
	      heigh --; /* one black line as separation */
	  do {
	      UI_plotHline (xy, xend, UI.bgcolor);
	      xy.y++;
	      } while (--heigh);
	  }
	else {
	  /* A lot faster in VGA graphic, here only VGA 256 colors */
	  xy.y += heigh * UI.parameter.charheigh;
	  UI_setcursor (baserow, 0);
	  while (heigh--)
	      puts ("");
	  }
      do {
	  coord xytmp = { x: 0, y: xy.y };
	  UI_plotHline (xytmp, xy.x, UI.stdcolor[black]);
	  UI_plotHline (xy, xend, UI.bgcolor);
	  xytmp.x = xend;
	  UI_plotHline (xytmp, UI.parameter.width,
					UI.stdcolor[black]);
	  xy.x++; xy.y++; xend--;
	  } while (   xy.x != UI.parameter.charheigh / 2
		   && xy.y < UI.parameter.heigh);
      if (UI.parameter.nbcolor == 16 && xy.y < UI.parameter.heigh) {
	  /* one black line as separation */
	  coord xytmp = { x: 0, y: xy.y };
	  UI_plotHline (xytmp, UI.parameter.width,
					UI.stdcolor[black]);
	  }
      }
    else {
      unsigned short i;

      if (baserow == 1) {
	  baserow = 0;
	  heigh++;
	  }
      if (baserow + heigh < UI.parameter.heigh) {
	  i = UI.parameter.width;
	  if (UI.parameter.attr.isgraphic)
	      i /= UI.parameter.charwidth;
	  heigh++;
	  }
	else
	  i = 0;
      UI_setcursor (baserow, 0);
      while (--heigh)
	  puts ("");
      while (i--)
	  print ("-");
      }
  }

#if USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT)
static void
draw_bg_box (unsigned baserow, unsigned heigh, unsigned color)
  {
  if (UI.parameter.attr.graphicbgcolor && UI.parameter.nbcolor >= 256)
      UI_setbgcolor (color);
    /* else the grey color (due to the mode initialisation) is used */
  draw_box (baserow, heigh);
  }
#else
/* do not write an inline function, the third parameter does not exist: */
#define draw_bg_box(baserow, heigh, color) \
  draw_box (baserow, heigh)

#endif

static inline void redraw_first_lines (unsigned refresh)
  {
  const char *msg2 = "", *msg1;
  char topline[80];

  draw_bg_box (1, 2, FIRST_LINES_BACKGROUND);

  if (BOOT1_COM_port() >= 0)
      msg1 = "";
    else
#if USER_SUPPORT & VGA_SUPPORT
	 if (VGA_ACTIVE())
      msg1 = "[VGA]";
    else
#endif
#if USER_SUPPORT & VESA_SUPPORT
	 if (!VESA_ACTIVE())
      msg1 = "[]";
    else
#if USER_SUPPORT & VESA_LINEAR
	 if (UI.parameter.winsize == VESA2_MARKER)
      msg1 = "[lin]"; /* using VESA2 linear framebuffer */
    else
#endif
	 if ((unsigned) UI.parameter.linelength * UI.parameter.heigh < 64*1024)
      msg1 = "[nowin]"; /* no window switching needed */
    else
#if USER_SUPPORT & VESA_WINFUNCTION
	 if (UI.parameter.addrpositionfarfct == 0)
	  msg1 = "[int]"; /* using VESA interrupt to change windows */
    else
#if USER_SUPPORT & VESA_HARDWINDOW
	 if (UI.parameter.hardwindow) {
      if (UI.parameter.offsetarray)
	  msg1 = "[hard]"; /* using hard method to change windows */
	else
	  msg1 = "[hardc]"; /* using hard compressed method to change windows */
      }
    else
#endif /* VESA_HARDWINDOW */
      msg1 = "[fct]"; /* using VESA function to change windows */
#else /* VESA_WINFUNCTION */
      msg1 = "[int]"; /* using VESA interrupt to change windows */
#endif /* VESA_WINFUNCTION */

#if USER_SUPPORT & VESA_2WINDOWS
  if (   UI.parameter.NbWin == 2
      && (unsigned) UI.parameter.linelength * UI.parameter.heigh >= 64*1024) {
#if USER_SUPPORT & VESA_HARDWINDOW
      if (UI.parameter.readhardwindow) {
	  if (UI.parameter.readoffsetarray)
	      msg2 = "[hard]"; /* using hard method to change windows */
	    else
	      msg2 = "[hardc]"; /* using hard compressed method to do so */
	  }
	else
#endif /* VESA_HARDWINDOW */
	  msg2 = msg1;
      }
#endif /* VESA_2WINDOWS */
#else /* VESA_SUPPORT */
      msg1 = "[]";
#endif /* VESA_SUPPORT */

  sprintf (topline,
	   "%$" TOPLINE_XSSSU,
	   &topline[MIN(sizeof(topline), UI.parameter.nbcol)],
	   UI.parameter.identification,
	   UI.parameter.attr.isgraphic? MSG_GRAPHIC : MSG_TEXT,
	   msg1, msg2, UI.nbmode);

  centered_print (topline, 1);

  if (refresh) {
#if SETUP & UPDATE_BOOTPARAM
      if (get_gujin_param_attrib().read_only_param)
	  sprintf (topline, "%$ refresh %u Hz, %s",
		&topline[MIN(sizeof(topline), UI.parameter.nbcol) - 1],
		refresh,
		BOOT_DEVICE_READ_ONLY);
	else
#endif
	  sprintf (topline, "%$ refresh %u Hz",
		&topline[MIN(sizeof(topline), UI.parameter.nbcol) - 1],
		refresh);
	}
      else {
#if SETUP & UPDATE_BOOTPARAM
      if (get_gujin_param_attrib().read_only_param)
	  strcpy (topline, BOOT_DEVICE_READ_ONLY);
	else
#endif
	  topline[0] = '\0';
	}

  if (strlen (topline))
      centered_print (topline, 2);
  }

static unsigned redraw_top_field (int field, unsigned short row, unsigned short col)
  {
  char middle_buffer[40];
  const char *startstr = 0, *middlestr = 0;
  const char *field1name;
  const char *field2name = 0;
  unsigned short field1key, field2key = 0; /* warning */
  struct gujin_param_attrib gujin_attr = get_gujin_param_attrib ();

  MOUSE_declare_attribute (mouse_attr1);
  MOUSE_declare_attribute (mouse_attr2);
  MOUSE_attrset_upperfield (mouse_attr1, 1);
  MOUSE_attrset_upperfield (mouse_attr2, 1);

  switch ((field >= 0)? field : -field) {
      case 1:
	startstr = MENU_RESOLUTION;
	field1name = "<->";
	MOUSE_attrset_activefield (mouse_attr1, prev_next_definition (0, gujin_attr) != -1);
	field1key  = MINUS_KEYCODE();
	/* With our implementation, UI.parameter.{width,heigh} is
	   equal to UI.parameter.{nbcol,nbrow} in text modes */
	sprintf (middle_buffer, MENU_DISPLAY_RESOLUTION,
			UI.parameter.width, UI.parameter.heigh);
	middlestr = middle_buffer;
	field2name = "<+>";
	MOUSE_attrset_activefield (mouse_attr2, prev_next_definition (1, gujin_attr) != -1);
	field2key  = PLUS_KEYCODE();
	break;
      case 2:
	startstr = MENU_COLORS;
	field1name = "</>";
	MOUSE_attrset_activefield (mouse_attr1, prev_next_color (0, gujin_attr) != -1);
	field1key  = DIVIDE_KEYCODE();
	sprintf (middle_buffer, MENU_DISPLAY_COLORS, get_bpp (&UI.parameter));
	middlestr = middle_buffer;
	field2name = "<*>";
	MOUSE_attrset_activefield (mouse_attr2, prev_next_color (1, gujin_attr) != -1);
	field2key  = MULTIPLY_KEYCODE();
	break;
      case 3: {
	unsigned short cpt, cptgraphic, cpttext;

	for (cpt = cpttext = cptgraphic = 0; cpt < UI.nbmode; cpt++)
	    if (UI.mode[cpt].text)
		cpttext++;
	      else
		cptgraphic++;
	if (UI.parameter.attr.isgraphic) {
	    field1name = MENU_TO_TEXT;
	    MOUSE_attrset_activefield (mouse_attr1, cpttext > 0);
	    }
	  else {
	    field1name = MENU_TO_GRAPHIC;
	    MOUSE_attrset_activefield (mouse_attr1, cptgraphic > 0);
	    }
	field1key  = DOT_KEYCODE();
	}
	break;
      case 4:
	field1name = MENU_ALTERNATE;
	MOUSE_attrset_activefield (mouse_attr1, prev_next_alternate (gujin_attr) != -1);
	field1key  = TAB_KEYCODE();
	break;
      case 5:
	if (UI.attributes.valid.transparent) {
	    /* can do 3D buttons */
	    MOUSE_attrset_pressedfield(mouse_attr1,
				       (Menu.type != kernel_bottom_menu));
	    field1name = MENU_SETUP;
	    }
	  else {
	    if (Menu.type == kernel_bottom_menu)
		field1name = MENU_SETUP;
	      else
		field1name = MENU_APPLICATION;
	    }
	MOUSE_attrset_activefield (mouse_attr1, 1);
	field1key  = SPACE_KEYCODE();
	break;
      case 0:
	return 5; /* number of fields */
      default:
	return 0;
      }

  if (field > 0) {
      if (startstr) {
	  print (startstr);
	  col += strlen (startstr);
	  }
      PUT_FIELD (field1name, mouse_attr1, field1key, row, col);
      col += strlen (field1name) + 1; /* PUT_FIELD() advances to next char */
      if (middlestr) {
	  print (middlestr);
	  col += strlen (middlestr);
	  }
      if (field2name) {
	  PUT_FIELD (field2name, mouse_attr2, field2key, row, col);
	  }
      return 0;
      }
    else {
      unsigned len = strlen (field1name) + 1;
      if (startstr)
	  len += strlen (startstr);
      if (middlestr)
	  len += strlen (middlestr);
      if (field2name)
	  len += strlen (field2name) + 1;
      return len + 1; /* at least one space after */
      }
  }

static unsigned short
redraw_top_menu (int curline)
  {
  unsigned space[redraw_top_field(0, 0, 0)];
  unsigned curfield, just_report, oneperline = 0;

  for (curfield = 0; curfield < nbof (space); curfield++)
      space[curfield] = redraw_top_field (-(curfield + 1), 0, 0);

  if (curline != 0) {
      unsigned nblines = redraw_top_menu (0);
      just_report = 0;
      draw_bg_box (curline, nblines, TOP_MENU_BACKGROUND);
      }
    else
      just_report = 1;

  for (curfield = 0; curfield < nbof (space); curline ++) {
      unsigned field = curfield, interspace, curcol;
      unsigned freespace = UI.parameter.nbcol;

      while (field < nbof (space)) {
	  if (freespace < space[field])
	      break;
	  freespace -= space[field++];
	  if (oneperline)
	      break;
	  }
      if (field == 1)
	  oneperline = 1;
      interspace = freespace / (field - curfield + 1);
      curcol = interspace + (freespace % (field - curfield + 1))/2;
      while (curfield < field) {
	  if (!just_report) {
		/* setcursor: home at 0,0 => add 1 to curcol */
	      UI_setcursor (curline, curcol + 1);
	      redraw_top_field (curfield + 1, curline, curcol + 1);
	      curcol += space[curfield] + interspace;
	      }
	  curfield ++;
	  }
      }
  return curline;
  }

static inline unsigned
begin_line_kernel_field (char *buffer, struct diskparam_str *dp,
			 struct desc_str *ptr)
  {
  struct gujin_param_attrib attrib = get_gujin_param_attrib ();
  char *buf = buffer;

  *buf = 0;
  if (!attrib.menu_with_disk && !attrib.menu_with_partition)
      return 0;
  if (attrib.menu_with_disk) {
      diskname (dp, buf);
      buf += strlen(buf);
      }
  if (!attrib.menu_with_partition) {
      if (dp->nbpartition != 0) /* not after "DOS C:" */
	  buf = STRINIT (buf, ": ");
	else
	  buf = STRINIT (buf, " ");
      }
    else {
      if (attrib.menu_with_disk)
	  buf = STRINIT (buf, "/");
      if (ptr->partition == 0xFF)
	  buf = STRINIT (buf, "MBR: ");
	else if (dp->nbpartition == 0) { /* DOS disk */
#if defined (MAX_DISKNAMESIZE) && (MAX_DISKNAMESIZE != 0)
	  /*
	   * Dirty if "MAX_DISKNAMESIZE != 0" and access == dos_part,
	   * the "partition" pointer is used as the DOS name pointer
	   * because nbpartition == 0.
	   */
	  if (dp->partition != 0) {
	      buf = STRINIT (buf, "\"");
	      buf = stradd (buf, (char *)dp->partition);
	      buf = STRINIT (buf, "\": ");
	      }
#endif
	  }
	else {
	  struct partition_str *curpart = &dp->partition[ptr->partition];

#if defined (MAX_DISKNAMESIZE) && (MAX_DISKNAMESIZE != 0)
	  if (curpart->name[0] != 0) {
	      buf = STRINIT (buf, "\"");
	      buf = stradd (buf, curpart->name);
	      buf = STRINIT (buf, "\": ");
	      }
	    else
#endif
	      {
	      buf = STRINIT (buf, "part ");
	      if (curpart->OSnumber >= 10) /* max 99 - 4 partitions */
		  *buf++ = '0' + curpart->OSnumber / 10;
	      *buf++ = '0' + curpart->OSnumber % 10;
#if !(SETUP & QUITE)
	      buf = STRINIT (buf, " (");
	      buf = stradd (buf, PartName (curpart->type));
	      buf = STRINIT (buf, ")");
#endif
	      buf = STRINIT (buf, ": ");
	      }
	  }
      }
  return strlen (buffer);
  }

static inline int
redraw_kernel_field (int field, unsigned short row, unsigned short col)
  {
  /* This assumes that initrd's are at the end of the array */
  struct desc_str *ptr = &boot_system.desc[ABS(field) - 1];
  char buffer[128]; /* Hopes it is enought */
  unsigned buflen;

  MOUSE_declare_attribute (mouse_attr);

  buflen = begin_line_kernel_field (buffer, &DI.param[ptr->disk], ptr);
  if (buflen < UI.parameter.nbcol - sizeof (" F10: ""/boot/vmli...")) {
      if (field < 0)
	  return buflen + strlen (ptr->filename) + 1 + 1;
      print (buffer);
      col += buflen;
      }
    else {
      if (field < 0)
	  return sizeof ("... ") - 1 + strlen (ptr->filename) + 1 + 1;
		/* clickable + space needed after */
      print ("... ");
      col += sizeof("... ") - 1;
      }

  MOUSE_attrset_mainfield (mouse_attr, 1);
  MOUSE_attrset_activefield (mouse_attr, 1);
  PUT_FIELD (ptr->filename, mouse_attr,
	     FCT_KEYCODE(ABS(field) - Menu.curpage * Menu.nbperpage),
	     row, col);
  return 0;
  }

static inline int
redraw_setup_attribute_field (int field, unsigned short keyNo,
			      unsigned short row, unsigned short col)
  {
  const char *startstr = 0;
  char active = 0, greyed = 0;
  unsigned absnb = ABS(field) - 1;

#if 0 /* code size optimise */
  struct gujin_param_attrib params = get_gujin_param_attrib ();

  switch (absnb) {
#define APPLY_REDRAW(field, deflt, grey) \
	case enum_ ## field: startstr = SMENU_ ## field ; \
		active = params.field ; greyed = grey ; break;

      APPLY_ATTRIBUTE_ORDER(APPLY_REDRAW)
#undef APPLY_REDRAW
      }

#else /* code size optimise */
  const char *Smenu[] = {
#define APPLY_STRING(field, deflt, grey)	SMENU_ ## field,
      APPLY_ATTRIBUTE_ORDER(APPLY_STRING)
#undef APPLY_STRING
      };
  union { unsigned value; struct gujin_param_attrib params; } tmp;

  startstr = Smenu[absnb];
  tmp.params = get_gujin_param_attrib ();
  active = ((tmp.value & (1U << absnb)) != 0);

#define APPLY_GREYED(field, deflt, grey) \
	if ((result = (grey)) != 0 && enum_ ## field == absnb) \
	    {greyed = result; break;}
  do {
      unsigned char result;
      APPLY_ATTRIBUTE_ORDER(APPLY_GREYED)
      greyed = 0;
      } while (0);
#undef APPLY_GREYED


#endif /* code size optimise */

  if (startstr) {
      char field[2];
      unsigned oldbgcolor = 0;
      MOUSE_declare_attribute (mouse_attr);

      if (field < 0)
	  return strlen (startstr) + sizeof("[X]") - 1
			+ 1 + 1;	/* clickable + space needed after */
      if (UI.attributes.valid.transparent) {
	  MOUSE_attrset_pressedfield (mouse_attr, 1);
	  oldbgcolor = UI.bgcolor;
	  UI_setbgcolor (UI.stdcolor[white]);
	  field[0] = '\373'; /* square root */
	  }
	else {
	  print ("[ ]");
	  col += 1; /* "[" */
	  UI_setcursor (row, col); /* \010 do not always work */
	  field[0] = 'X';
	  }
      if (!active)
	  field[0] = ' ';
      field[1] = '\0';

      MOUSE_attrset_chkboxfield(mouse_attr, 1);
      MOUSE_attrset_mainfield (mouse_attr, 1);
      MOUSE_attrset_activefield (mouse_attr, 1);
      MOUSE_attrset_greyfield (mouse_attr, greyed);
      PUT_FIELD (field, mouse_attr, FCT_KEYCODE(keyNo), row, col);
      if (UI.attributes.valid.transparent)
	  UI_setbgcolor (oldbgcolor);
      if (!UI.attributes.valid.transparent)
	  col += 1; /* "]" */
      UI_setcursor (row, col + 2);
      print (startstr);
      }
  return 0;
  }

static inline int
redraw_setup_action_field (int field, unsigned short keyNo,
			   unsigned short row, unsigned short col)
  {
  const char *fieldname;
  MOUSE_declare_attribute (mouse_attr);

  MOUSE_attrset_mainfield (mouse_attr, 1);

  switch ((field >= 0)? field : -field) {
      case 1:
	fieldname = SMENU_PROBE_VESA;
	MOUSE_attrset_activefield (mouse_attr, !!(USER_SUPPORT & VESA_SUPPORT));
	break;
      case 2:
	fieldname = SMENU_PROBE_VGA_0_13;
	MOUSE_attrset_activefield (mouse_attr, !!(USER_SUPPORT & VGA_SUPPORT));
	break;
      case 3:
	fieldname = SMENU_PROBE_VGA_14_127;
	MOUSE_attrset_activefield (mouse_attr, !!(USER_SUPPORT & VGA_SUPPORT));
	break;
      case 4:
	fieldname = SMENU_RESET_VGA_MODES;
	MOUSE_attrset_activefield (mouse_attr, !!(USER_SUPPORT & VGA_SUPPORT));
	break;
      case 5:
	fieldname = SMENU_RESET_ALL_PARAMS;
	MOUSE_attrset_activefield (mouse_attr, 1);
	break;
      case 6:
	fieldname = SMENU_POWER_DOWN;
	MOUSE_attrset_activefield (mouse_attr, !BOOT1_DOS_running());
	break;
      case 7:
	fieldname = SMENU_FAIL_BOOT;
	MOUSE_attrset_activefield (mouse_attr, !BOOT1_DOS_running());
	break;
      case 8:
	fieldname = SMENU_SET_TIMEOUT_AUTOLOAD;
	MOUSE_attrset_activefield (mouse_attr, 1);
	break;
      case 0:
	return 8; /* number of fields */
      default:
	return 0;
      }

  if (field > 0) {
      PUT_FIELD (fieldname, mouse_attr, FCT_KEYCODE(keyNo), row, col);
      return 0;
      }
    else {
      return strlen (fieldname) + 1 + 1; /* at least one space after */
      }
  }

static inline int
redraw_setup_field (int field, unsigned short row, unsigned short col)
  {
  unsigned short keyNo = ABS(field) - Menu.curpage * Menu.nbperpage;

  if (field == 0)
      return TOTAL_ATTRIBUTE + redraw_setup_action_field (0, 0, 0, 0);
    else if (ABS(field) <= TOTAL_ATTRIBUTE)
      return redraw_setup_attribute_field (field, keyNo, row, col);
    else
      return redraw_setup_action_field (field + ((field > 0)
						 ? - TOTAL_ATTRIBUTE
						 : TOTAL_ATTRIBUTE),
					keyNo, row, col);
  }

static inline void invert_setup_attribute_field (unsigned absnb)
  {
#if 0 /* code size optimise */
  struct gujin_param_attrib params = get_gujin_param_attrib ();

  switch (absnb) {
#define APPLY_INVERT(field, deflt, grey) \
      case enum_ ## field: param.field = !params.field ; break;
      APPLY_ATTRIBUTE_ORDER(APPLY_INVERT)
#undef APPLY_INVERT
      }
  set_gujin_param_attrib (params);
#if USER_SUPPORT & JOYSTICK_SUPPORT
  if (absnb == enum_enable_joystick && params.enable_joystick)
      enable_joystick_when_no_move();
#endif
#else /* code size optimise */
  union { unsigned value; struct gujin_param_attrib params; } tmp;

  tmp.params = get_gujin_param_attrib ();
  tmp.value ^= 1U << absnb;
  set_gujin_param_attrib (tmp.params);
#if USER_SUPPORT & JOYSTICK_SUPPORT
  if (absnb == enum_enable_joystick && tmp.params.enable_joystick)
      enable_joystick_when_no_move();
#endif
#endif /* code size optimise */
  }

struct todraw_str {
    unsigned char firstlines		: 1;
    unsigned char topmenu		: 1;
    unsigned char bottommenu		: 1;
    unsigned char advertisement		: 1;
    unsigned char copyright		: 1;
    unsigned char update_bootparam	: 1;
    unsigned char refresh		: 1;
    unsigned char presskey		: 1;
    };

static inline struct todraw_str
activate_setup_action_field (unsigned absnb, unsigned short advline)
  {
  unsigned char min, max;

  switch (absnb) {
      case 1:
#if USER_SUPPORT & VESA_SUPPORT
	  if (BOOT1_COM_port() < 0) {
	      unsigned short i;

		/* FIXME: Is it needed to switch to VESA? */
		/* Call of VESA_getmode in VGA */
#if USER_SUPPORT & VGA_SUPPORT
	      unsigned in_vesa = VESA_ACTIVE();
	      unsigned short current_mode = 0; /* warning */

	      if (!in_vesa) {
		  current_mode = UI_getmode() & ~0x8000;
		  if (VESA_init() != 0)
		      break;
		  }
#endif

	      for (i = 0; i < 0x80; i++) {
		  struct video_parameter_str param;
		  UI_getsize (i, &param);
		  }

#if USER_SUPPORT & VGA_SUPPORT
	      if (!in_vesa) {
		  VGA_init();
		  UI_setmode (current_mode);
		  }
#endif
	      }
#endif
	  /* if we have switch VGA->VESA->VGA: */
	  return (struct todraw_str) {
		firstlines:	  1,
		topmenu:	  1,
		bottommenu:	  1,
		advertisement:	  1,
		copyright:	  1,
		update_bootparam: 0,
		refresh:	  0,
		presskey:	  0
		};
      case 2:
	  min = 0; max = 0x13;
	  goto case_2_3;
      case 3: {
	  unsigned short key;

	  centered_print (TURN_OFF_MONITOR, advline);
	  print (PRESS_KEY_TO_CONTINUE);
	  key = UI_getkey (30 * TICKS_PER_SECOND);
	  if (key == TIMEOUT_KEYCODE() || (key & 0xFF) == '\033')
	      return (struct todraw_str) {
		firstlines:	  0,
		topmenu:	  0,
		bottommenu:	  1,
		advertisement:	  1,
		copyright:	  0,
		update_bootparam: 0,
		refresh:	  0,
		presskey:	  0
		};
	  min = 0x14; max = 0x7F;
	  }
	  goto case_2_3;
      case_2_3:
#if USER_SUPPORT & VGA_SUPPORT
	  if (BOOT1_COM_port() < 0) {
	      struct video_parameter_str param;
	      unsigned char i;
	      unsigned short current_mode = UI_getmode() & ~0x8000;

#if USER_SUPPORT & VESA_SUPPORT
	      unsigned in_vesa = VESA_ACTIVE();

	      if (in_vesa && VGA_init() != 0)
		  break;
#endif
	      UI_setmode (3);
	      UI_setbgcolor (0); /* due to  probing */
	      UI_setfgcolor (UI.stdcolor[lightgray]);
	      UI_clearscreen();

	      for (i = min; i <= max; i++) {
		  unsigned err;
		  unsigned char j;

		  for (j = 0; j < UI.nbmode; j++)
		      if (i == UI.mode[j].number && UI.mode[j].vesa)
			  break;
		  if (j < UI.nbmode)
		      continue;
		  printf (SETTING_MODE_X, i);
		  err = UI_getsize (i, &param);
		  if (err)
		      printf (ERROR_SETTING_X, err);
#ifndef KEEP_ALL_VGA_MODE
		    else if (param.identification > 0x37 && param.nbcolor <= 4)
		      print (": <= 2 BPP");
#endif
		    else
		      print (MSG_COLON_OK);
		  puts ("");
		  }
	      print (MSG_BEEP);
#if SETUP & UPDATE_BOOTPARAM
	      setup_gujin_param_vga_mode ();
#endif

#if USER_SUPPORT & VESA_SUPPORT
	      if (in_vesa)
		  VESA_init();
#endif
	      UI_setmode (current_mode);
	      }
#endif
	  /* if we have switch VESA->VGA->VESA: */
	  return (struct todraw_str) {
		firstlines:	  1,
		topmenu:	  1,
		bottommenu:	  1,
		advertisement:	  1,
		copyright:	  1,
		update_bootparam: 1,
		refresh:	  0,
		presskey:	  0
		};
      case 4: {
#if USER_SUPPORT & VGA_SUPPORT
	  VIDEO_mode_reset();
#endif
	  reset_low_valid_mode(128);
#if (SETUP & UPDATE_BOOTPARAM) && (USER_SUPPORT & VGA_SUPPORT)
	  setup_gujin_param_vga_mode ();
#endif
	  }
	  return (struct todraw_str) {
		firstlines:	  1,
		topmenu:	  1,
		bottommenu:	  1,
		advertisement:	  1,
		copyright:	  0,
		update_bootparam: 1,
		refresh:	  0,
		presskey:	  0
		};
      case 5:
	  menu_reset_saved_parameters();
	  return (struct todraw_str) {
		firstlines:	  0,
		topmenu:	  0,
		bottommenu:	  1, /* redraw the pressed button */
		advertisement:	  1,
		copyright:	  0,
		update_bootparam: 1,
		refresh:	  0,
		presskey:	  0
		};
      case 6:
	  if (!BOOT1_DOS_running()) {
	      PRINT (POWERING_OFF);
	      APM_power_OFF();
	      PRINT (MSG_COLON_FAILED);
	      }
	  return (struct todraw_str) {
		firstlines:	  1,
		topmenu:	  1,
		bottommenu:	  1,
		advertisement:	  1,
		copyright:	  1,
		update_bootparam: 0,
		refresh:	  0,
		presskey:	  1
		};
      case 7:
	  if (!BOOT1_DOS_running()) {
	      before_exit ();
	      _BIOS_failedboot();
	      }
	  return (struct todraw_str) {
		firstlines:	  1,
		topmenu:	  1,
		bottommenu:	  1,
		advertisement:	  1,
		copyright:	  1,
		update_bootparam: 0,
		refresh:	  0,
		presskey:	  1
		};
      case 8: {
	  unsigned bgcolor = UI.bgcolor;
	  printf ("Timeout autoload [0..99]:    s (0:none)"
		  "\010\010\010\010\010\010\010\010\010\010\010\010\010");
	  UI_setbgcolor (UI.stdcolor[lightgray]);
	  print ("99 \010\010\010");
	  set_timeout_autoload (get_number (get_timeout_autoload(), 99));
	  UI_setbgcolor (bgcolor);
	  }
	  return (struct todraw_str) {
		firstlines:	  0,
		topmenu:	  0,
		bottommenu:	  1,
		advertisement:	  1,
		copyright:	  1, /* 40 cols mode */
		update_bootparam: 1,
		refresh:	  0,
		presskey:	  0
		};
      }
  /* VESA_init() failed */
  return (struct todraw_str) {
	firstlines:	  0,
	topmenu:	  0,
	bottommenu:	  1, /* redraw the pressed button */
	advertisement:	  1,
	copyright:	  0,
	update_bootparam: 0,
	refresh:	  0,
	presskey:	  0
	};
  }

static inline struct todraw_str
activate_setup_field (unsigned absnb, unsigned short advline)
  {
  struct todraw_str returned;

  if (absnb < TOTAL_ATTRIBUTE) {
      invert_setup_attribute_field (absnb);
	/* number of modes, +/- through BPP... : */
      returned = (struct todraw_str) {
	firstlines:	  1,
	topmenu:	  1,
	bottommenu:	  1,
	advertisement:	  1,
	copyright:	  0,
	update_bootparam: 1,
	refresh:	  0,
	presskey:	  0
	};
      }
    else
      returned = activate_setup_action_field (absnb - TOTAL_ATTRIBUTE + 1,
						advline);

  return returned;
  }

static inline void
redraw_bottom_menu (unsigned nbline, unsigned curline)
  {
  unsigned char fctkeyno;
  unsigned endline = curline + nbline - 1;

  if (Menu.totalpage > 1) {
      MOUSE_declare_attribute (mouse_attr);
      MOUSE_attrset_mainfield (mouse_attr, 1);
      MOUSE_attrset_activefield (mouse_attr, (Menu.curpage != 0));

      UI_setcursor (curline, UI.parameter.nbcol / 4 - 4);
      PUT_FIELD (PAGEUP_KEYNAME, mouse_attr, PAGEUP_KEYCODE(),
			curline, UI.parameter.nbcol / 4 - 4);
      UI_setcursor (curline, UI.parameter.nbcol / 2);
      PUT_FIELD (HOME_KEYNAME, mouse_attr, HOME_KEYCODE(),
			curline, UI.parameter.nbcol / 2);
      curline++;
      }

  curline -= Menu.interline / 2;

  for (fctkeyno = 1;
	  fctkeyno <= Menu.nbperpage
       && fctkeyno <= Menu.nbtotal - Menu.curpage * Menu.nbperpage;
       fctkeyno++) {
      const char *format;
      curline += Menu.interline;
      /* We could also do 'Menu.interline' times '\r\n', it would
	 not need a draw_box()... inteface would be quicker,
	 but it may be better if we want to display something
	 on the right part of the screen ... */
      UI_setcursor (curline, 0);
      if (BOOT1_COM_port() >= 0 && fctkeyno <= 4) {
	  /* Here, on a real VT, if you cannot have F5, use the 'Do' key */
	  format = " PF%u: ";
	  }
	else if (fctkeyno <= 9)
	  format = " F%u:  ";
	else
	  format = " F%u: ";
      printf (format, fctkeyno);
      if (Menu.type == kernel_bottom_menu)
	  redraw_kernel_field (Menu.curpage * Menu.nbperpage + fctkeyno,
				curline, sizeof (" F%u: ") - 1);
	else
	  redraw_setup_field (Menu.curpage * Menu.nbperpage + fctkeyno,
				curline, sizeof (" F%u: ") - 1);
      }

  if (Menu.totalpage > 1) {
      MOUSE_declare_attribute (mouse_attr);
      MOUSE_attrset_mainfield (mouse_attr, 1);
      MOUSE_attrset_activefield (mouse_attr,
		 (Menu.totalpage > 1 && Menu.curpage < Menu.totalpage - 1));

      UI_setcursor (endline, UI.parameter.nbcol / 4 - 4);
      PUT_FIELD (PAGEDOWN_KEYNAME, mouse_attr, PAGEDOWN_KEYCODE(),
			endline, UI.parameter.nbcol / 4 - 4);
      UI_setcursor (endline, UI.parameter.nbcol / 2);
      PUT_FIELD (END_KEYNAME, mouse_attr, END_KEYCODE(),
			endline, UI.parameter.nbcol / 2);
      }
  }

static void show_ladders (unsigned bgcolor)
  {
#if USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT)
  unsigned short endy = UI.parameter.nbrow - 1;
  coord xy;

  if (UI.parameter.attr.isgraphic) {
      endy *= UI.parameter.charheigh;
      xy.y = endy - UI.parameter.charheigh / 2;

      if (UI.parameter.nbcolor > 16) {
	  const unsigned dacsize = 6; /* 6 bit DAC */
	  unsigned char cpt = (1 << dacsize) - 1;
	  unsigned color[3];
	  unsigned endx = UI.parameter.width - 32 * UI.parameter.charwidth;
	  unsigned short increment = (endx / (1 << dacsize))
					? (endx / (1 << dacsize)) : 1;
	  unsigned char i, j = DIVROUNDUP (UI.parameter.charheigh, 3);

	  for (xy.x = 0; xy.y < endy; xy.y++)
	      UI_plotHline (xy, UI.parameter.width, bgcolor);

	  do {
	      color[0] = VESA_color (cpt,   0,   0);
	      color[1] = VESA_color (0,   cpt,   0);
	      color[2] = VESA_color (0,     0, cpt);
	      for (i = UI.parameter.charheigh; i != 0; i--, xy.y++)
		  UI_plotHline (xy, xy.x + increment, color[i / j]);
	      xy.x += increment;
	      xy.y = endy;
	      } while (cpt-- != 0);
	  for (i = UI.parameter.charheigh; i-- > 0; xy.y++)
	      UI_plotHline (xy, UI.parameter.width, bgcolor);
	  }
      }
    else {
      xy.y = endy;
      bgcolor <<= 8;
      bgcolor |= ' ';
      }

  for (xy.x = 0; xy.y < UI.parameter.heigh; xy.y++)
      UI_plotHline(xy, UI.parameter.width, bgcolor);
#endif /* VGA_SUPPORT | VESA_SUPPORT */
  }

static inline void init_Menu (unsigned nblines)
  {
  unsigned max_Fkey;

  if (processor.keyboard == serialkbd)
      max_Fkey = 12; /* see is_valid_fct_keycode() just after */
    else if (processor.keyboard == stdkbd)
      max_Fkey = 10;
    else
      max_Fkey = 12;
  /* minimum one empty line in between two menu lines: */
  if (Menu.nbtotal < nblines / 2 && Menu.nbtotal <= max_Fkey) {
      Menu.nbperpage = Menu.nbtotal;
      Menu.totalpage = 1;
      Menu.interline = nblines / (Menu.nbperpage + 1);
      }
    else {
      /* 2 lines for the PgUp/Home menu, one for PgDn/end: */
      Menu.nbperpage = (nblines - 3) / 2;
      if (Menu.nbperpage > max_Fkey)
	  Menu.nbperpage = max_Fkey;
      Menu.totalpage = DIVROUNDUP (Menu.nbtotal, Menu.nbperpage);
      Menu.interline = (nblines - 3) / Menu.nbperpage;
      }
  }

static inline unsigned short
redraw_screen (struct todraw_str toredraw, unsigned refresh, unsigned *timeout)
  {
  unsigned line;
  unsigned short advline = UI.parameter.nbrow - 1;
  struct autoload_str autoload = get_last_boot_default();

  if (UI.parameter.nbcol >= ADVERTISEMENT_BANNER_MINWIDTH)
      advline -= ADVERTISEMENT_BANNER_NBLINES + 1;

  if (toredraw.firstlines)
      redraw_first_lines (refresh);

  RESET_FIELDS (((struct mousefieldattr_str) {
			upperfield: toredraw.topmenu,
			mainfield: toredraw.bottommenu
			}));

  if (toredraw.topmenu)
      line = redraw_top_menu (4) + 1;
    else
      line = redraw_top_menu (0) + 5; /* still get line */

  if (toredraw.bottommenu /* FIXME: && advline > line */) {
      unsigned nblines = advline - 1 - line;

      draw_bg_box (line, nblines, BOTTOM_MENU_BACKGROUND);
      if (Menu.nbtotal == 0)
	  centered_print (NOT_EVEN_ONE_SELECTABLE_ITEM, line + 3);
	else {
	  if (Menu.nbperpage == 0) {
	      init_Menu (nblines);
	      if (autoload.total_page == Menu.totalpage)
		  Menu.curpage = autoload.init_page;
		else {
		  Menu.curpage = 0;
		  *timeout = 0;
		  }
	      }
	  redraw_bottom_menu (nblines, line);
	  }
      }

  if (toredraw.copyright) {
      unsigned savebgcolor, savefgcolor;

      if (BOOT1_COM_port() < 0) {
	  show_ladders (UI.stdcolor[black]);
	  UI_setcursor (UI.parameter.nbrow - 1, UI.parameter.nbcol - 31);
	  savebgcolor = UI.bgcolor;
	  UI_setbgcolor (UI.stdcolor[black]);
	  }
	else {
	  unsigned short cpt;

	  savebgcolor = UI.bgcolor;
	  UI_setbgcolor (UI.stdcolor[black]);
	  UI_setcursor (UI.parameter.nbrow - 1, 0);
	  cpt = UI.parameter.nbcol - 31 + 1;
	  while (--cpt)
	      print (" ");
	  }
      savefgcolor = UI.fgcolor;
      UI_setfgcolor (UI.stdcolor[lightgray]);
      print ((char *)0xFFB0);
      UI_setbgcolor (savebgcolor);
      UI_setfgcolor (savefgcolor);
      }

  if (autoload.total_loadable != boot_system.nb)
      *timeout = 0;

  if (   toredraw.advertisement
      && UI.parameter.nbcol >= ADVERTISEMENT_BANNER_MINWIDTH) {
      char advert[ADVERTISEMENT_BANNER_NBLINES * ADVERTISEMENT_BANNER_MINWIDTH];

      if (*timeout == 0)
	  strcpy (advert, ADVERTISEMENT_BANNER);
	else
	  sprintf (advert, IN_U_SECOND_AUTOLOAD_S, *timeout,
		boot_system.desc[autoload.last_loaded].filename);

      draw_bg_box (advline, ADVERTISEMENT_BANNER_NBLINES, ADVERT_BACKGROUND);
      centered_print (advert, advline);
      }
  return advline;
  }

/*
 * This function treat the key pressed / mouse clics, it is completely
 * independant of what is displayed on the screen:
 */

/* simple helpers */
static inline unsigned
is_valid_fct_keycode (unsigned short key, unsigned short nbfctkey)
  {
  if ((key & 0xFF) != 0 || key < FCT_KEYCODE(1))
      return 0; /* surely not an FCT[1-12] key */
  if (key == FCT_KEYCODE(12) && nbfctkey >= 12)
      return 1;
  if (key == FCT_KEYCODE(11) && nbfctkey >= 11)
      return 1;
  if (nbfctkey > 10)
      return (key <= FCT_KEYCODE(10));
    else
      return (key <= FCT_KEYCODE(nbfctkey));
  }

static inline unsigned
is_valid_chgmode_keycode (unsigned short key)
  {
  /* We include the comma if serial because on a real VT,
     the comma replaces the plus on the numeric keypad */
  if (BOOT1_COM_port() >= 0)
#if 0 /* I have a bug here, with constant strings? */
      return strchr ("+-*/.,\t", key & 0xFF) != 0;
#else
      return    (key & 0xFF) == '+' || (key & 0xFF) == '-'
	     || (key & 0xFF) == '*' || (key & 0xFF) == '/'
	     || (key & 0xFF) == '.' || (key & 0xFF) == ','
	     || (key & 0xFF) == '\t';
#endif
  if (   key == PLUS_KEYCODE()
      || key == MINUS_KEYCODE()
      || key == MULTIPLY_KEYCODE()
      || key == DIVIDE_KEYCODE()
      || key == DOT_KEYCODE()
      || key == TAB_KEYCODE())
      return 1;
  /* only extkbd: */
  if (   key == KP_DIVIDE_KEYCODE()
      || key == NUMPAD_DOT_KEYCODE())
      return 1;
  return 0;
  }

static inline unsigned
is_valid_chgmenu_keycode (unsigned short key)
  {
  if (BOOT1_COM_port() >= 0)
      return (key & 0xFF) == ' ';
    else
      return key == SPACE_KEYCODE();
  }

static inline unsigned
is_valid_ONOFF_keycode (unsigned short key)
  {
  if (key == ON_OFF_KEYCODE())
      return 1;
  if (key == POWER_KEYCODE())
      return 1;
  return 0;
  }

static inline unsigned
is_PgUp_keycode (unsigned short key)
  {
  if (key == PAGEUP_KEYCODE())
      return 1;
  /* only extkbd: */
  if (key == (PAGEUP_KEYCODE() | 0xE0))
      return 1;
  return 0;
  }

static inline unsigned
is_PgDn_keycode (unsigned short key)
  {
  if (key == PAGEDOWN_KEYCODE())
      return 1;
  /* only extkbd: */
  if (key == (PAGEDOWN_KEYCODE() | 0xE0))
      return 1;
  return 0;
  }

static inline unsigned
is_Home_keycode (unsigned short key)
  {
  if (key == HOME_KEYCODE())
      return 1;
  /* only extkbd: */
  if (key == (HOME_KEYCODE() | 0xE0))
      return 1;
  return 0;
  }

static inline unsigned
is_End_keycode (unsigned short key)
  {
  if (key == END_KEYCODE())
      return 1;
  /* only extkbd: */
  if (key == (END_KEYCODE() | 0xE0))
      return 1;
  return 0;
  }


static unsigned
menu (struct registers *regs, unsigned extendmem, unsigned freemem,
	struct gujin_param_attrib gujin_attr)
  {bound_stack();{
  unsigned refresh_freq = 0; /* warning */
  unsigned timeout = get_timeout_autoload();
  struct todraw_str toredraw = { 1,1,1,1,1, refresh: 1, presskey: 0};
  unsigned short advline;

  set_initial_mode (gujin_attr);
  toredraw.update_bootparam = 1; /* to save the initial mode */
  Menu.nbtotal = boot_system.nb;
  Menu.nbperpage = 0;
  Menu.type = kernel_bottom_menu;

  for (;;) {
      unsigned short key;

      if (toredraw.refresh == 1) {
#if USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT)
	  refresh_freq = get_refresh_freq ();
#endif
	  toredraw.refresh = 0; /* refresh_freq updated */
	  }

#if SETUP & UPDATE_BOOTPARAM
      if (toredraw.update_bootparam)
	  BOOT1_update_bootparam ();
#endif

      advline = redraw_screen (toredraw, refresh_freq, &timeout);
      toredraw = (struct todraw_str) {0, 0, 0, 0, 0, presskey: 0};

      key = UI_getkey(timeout? TICKS_PER_SECOND : 0);

      if (timeout != 0) {
	  if (   key != TIMEOUT_KEYCODE()
#if USER_SUPPORT & (BIOS_MOUSE_SUPPORT | SERIAL_MOUSE_SUPPORT)
	      || MOUSE_has_moved()
#endif
	      ) {
	      /* Cancel autoboot & redraw advert: */
	      timeout = 0;
	      toredraw.advertisement = 1;
	      continue;
	      }
	    else if (--timeout != 0) {
	      toredraw.advertisement = 1;
	      continue;
	      }
	  }

      /*
       * Choices which will not fail, so will not
       * display any messages in the "advert" window:
       */
      if (is_PgUp_keycode (key) && Menu.curpage != 0) {
	  Menu.curpage -= 1;
	  toredraw.bottommenu = 1;
	  continue;
	  }
	else if (is_PgDn_keycode (key) && Menu.curpage < Menu.totalpage - 1) {
	  Menu.curpage += 1;
	  toredraw.bottommenu = 1;
	  continue;
	  }
	else if (is_Home_keycode (key)) {
	  Menu.curpage = 0;
	  toredraw.bottommenu = 1;
	  continue;
	  }
	else if (is_End_keycode (key)) {
	  Menu.curpage = Menu.totalpage - 1;
	  toredraw.bottommenu = 1;
	  continue;
	  }

      if (is_valid_chgmenu_keycode (key) && Menu.type == kernel_bottom_menu) {
	  Menu.type = setup_bottom_menu;
	  Menu.nbtotal = redraw_setup_field (0, 0, 0);
	  toredraw = (struct todraw_str) {0,1,1,0,0, presskey: 0};
	  Menu.nbperpage = 0; /* recalculate nbpage */
	  continue;
	  }

      if (UI.parameter.nbcol >= ADVERTISEMENT_BANNER_MINWIDTH) {
	  unsigned short row = UI.parameter.nbrow
				- ADVERTISEMENT_BANNER_NBLINES - 2;
	  draw_bg_box (row, ADVERTISEMENT_BANNER_NBLINES,
				WORK_ADVERT_BACKGROUND);
	  toredraw.advertisement = 1;
	  UI_setcursor (row, 5);
	  }
	else
	  UI_setcursor (UI.parameter.nbrow - 2, 0);

      /*
       * Choices which can fail:
       */
      if (is_valid_chgmode_keycode (key)) {
	  if (change_video_mode (key & 0xFF, 1) == 0) {
	      Menu.nbperpage = 0; /* recalculate nbpage */
	      toredraw = (struct todraw_str) {1,1,1,1,1,
						update_bootparam: 1,
						refresh: 1};
	      }
	    else
	      toredraw.presskey = 1;
	  }
	else if (   is_valid_chgmenu_keycode (key)
		 && Menu.type == setup_bottom_menu) {
	  struct gujin_param_attrib params = get_gujin_param_attrib();

	  /* This may display few lines of warning from
	     the disk/partition analysis... */
	  if (diskconfig_changed (gujin_attr, params)) {
#ifdef UTIL_USE_MALLOC
	      /* To not create too many holes in our malloc-able
		 memory - with our simple malloc management, we
		 release the memory malloced by disk_analyse()
		 before the call to probedisk() : */
	      if (boot_system.desc != 0) {
		  free (boot_system.desc);
		  boot_system.desc = 0;
		  }
#endif
	      probedisk (regs->edx & 0xFF, params);
	      disk_analyse (params);
	      }
	    else if (searchconfig_changed (gujin_attr, params))
	      disk_analyse (params);
	  gujin_attr = params;
	  Menu.type = kernel_bottom_menu;
	  Menu.nbtotal = boot_system.nb;
	  Menu.nbperpage = 0; /* recalculate nbpage */
	  toredraw = (struct todraw_str) {0,1,1,1,0, presskey: 0};
	  }
	else if (   (key == TIMEOUT_KEYCODE() && Menu.type == kernel_bottom_menu)
		 || is_valid_fct_keycode (key,
			MIN (Menu.nbperpage,
			     Menu.nbtotal - (Menu.curpage * Menu.nbperpage)
			     ) )) {
	  unsigned index;
	  if (key != TIMEOUT_KEYCODE()) {
	      index = Menu.curpage * Menu.nbperpage;
	      if (key == FCT_KEYCODE(12))
		  index += 11;
		else if (key == FCT_KEYCODE(11))
		  index += 10;
		else
		  index += ((key - FCT_KEYCODE(1)) >> 8);
	      }
	    else
	      index = get_last_boot_default().last_loaded;

	  if (Menu.type == kernel_bottom_menu) {
	      if (key != TIMEOUT_KEYCODE()) {
		  setup_last_boot_default (index, boot_system.nb,
					Menu.curpage, Menu.totalpage);
#if SETUP & UPDATE_BOOTPARAM
		  BOOT1_update_bootparam();
#endif
		  }
	      if (boot_system.desc[index].filesize == 0) {
		  unsigned char ret;
		  ret = DISK_readsector (
				&DI.param[boot_system.desc[index].disk],
				boot_system.desc[index].inode, 1, 0x7C00);
		  if (ret == 0) {
		      /* complete restart! */
		      return 0;
		      }
		  }
		else {
		  extern const struct instboot_info_str instboot_info;

		  menu_load_system (index, extendmem, freemem,
				    get_gujin_param_extra_cmdline(),
				    farptr2linear(code_adr(&instboot_info)),
				    getDefaultTextVideoMode(),
				    gujin_attr);
		  }
	      toredraw = (struct todraw_str) {1,1,1,1,1,
						presskey: 1,
						update_bootparam: 1};
	      }
	    else
	      toredraw = activate_setup_field (index, advline);
	  }
	else if (!BOOT1_DOS_running() && is_valid_ONOFF_keycode (key)) {
	  PRINT (POWERING_OFF);
	  APM_power_OFF();
	  PRINT (MSG_COLON_FAILED);
	  toredraw = (struct todraw_str) {1,1,1,1,1, presskey: 1};
	  }
#if SETUP & MINIDEBUG
	else if (gujin_attr.verbose && (key & 0xFF) == '1') {
	  menu_list_memory();
	  puts(""); puts(""); puts("");
	  toredraw = (struct todraw_str) {1,1,1,1,1, presskey: 1};
	  }
	else if (gujin_attr.verbose && (key & 0xFF) == '2') {
	  menu_list_hardwin();
	  puts(""); puts(""); puts("");
	  toredraw = (struct todraw_str) {1,1,1,1,1, presskey: 1};
	  }
	else if (gujin_attr.verbose && (key & 0xFF) == '3') {
	  menu_list_valid_disk();
	  puts(""); puts(""); puts("");
	  toredraw = (struct todraw_str) {1,1,1,1,1, presskey: 1};
	  }
	else if (gujin_attr.verbose && (key & 0xFF) == '4') {
	  menu_list_video_modes();
	  puts(""); puts(""); puts("");
	  toredraw = (struct todraw_str) {1,1,1,1,1, presskey: 1};
	  }
	else if (gujin_attr.verbose && (key & 0xFF) == '5') {
	  menu_list_boot_images();
	  puts(""); puts(""); puts("");
	  toredraw = (struct todraw_str) {1,1,1,1,1, presskey: 1};
	  }
	else if (gujin_attr.verbose && (key & 0xFF) == '9') {
	  int i;
	  for (i = 0; i < 0x15; i++)
	      tstATIvideo(i);
	  puts(""); puts(""); puts("");
	  toredraw = (struct todraw_str) {1,1,1,1,1, presskey: 1};
	  }
#if 0
	else if (gujin_attr.verbose && (key & 0xFF) == '8') {
	  tstCDbootcatalog();
	  puts(""); puts(""); puts("");
	  toredraw = (struct todraw_str) {1,1,1,1,1, presskey: 1};
	  }
#endif
#endif /* MINIDEBUG */
	else {
	  PRINTF (MENU_UNKNOWN_KEY, key);
	  toredraw = (struct todraw_str) {0,0,0,1,0, presskey: 1};
	  }

      /* gujin_attr.verbose may not be up to date if we are in the
	 "setup" screen, use get_gujin_param_attrib() value: */
      if (toredraw.presskey && get_gujin_param_attrib().verbose) {
	  print (PRESS_KEY_TO_CONTINUE);
	  UI_getkey (30 * TICKS_PER_SECOND);
	  }

      { /* If we have written too many lines, we may have scrolled: */
      unsigned char row, col;

      if (   UI_getcursor (&row, &col) != 0
	  || row == UI.parameter.nbrow - 1)
	  toredraw = (struct todraw_str) {1,1,1,1,1};
      }
      } /* for (;;) */
  }}

static inline void treat_startup_keys (void)
  {
  unsigned something_changed = 0;

  if (_BIOS_get_shift_flags().caps_lock_active) {
      /*
       * give a way to reset saved parameters - else it may
       * be impossible to boot after changing the video card...
       */
      PUTS (IMPORTANT_PARAMETERS_RESETED);
      menu_reset_saved_parameters();
      something_changed = 1;
      }

  if (_BIOS_get_shift_flags().control_key_pressed) {
      PUTS (FORCE_VIDEO_MODE_3);
      setDefaultVideoMode (3);
      something_changed = 1;
      }

#if SETUP & UPDATE_BOOTPARAM
  if (something_changed)
      BOOT1_update_bootparam();
#endif
  }

#else /* USER_SUPPORT != 0 */

/*
 * Functions to read a key pressed are not even linked-in,
 * so decide alone what to load:
 */
static inline unsigned
menu (struct registers *regs, unsigned extendmem, unsigned freemem,
	struct gujin_param_attrib gujin_attr)
  {
  unsigned short cpt;

  for (cpt = 0; cpt < boot_system.nb; cpt++) {
      extern const struct instboot_info_str instboot_info;

      menu_load_system (cpt, extendmem, freemem,
			get_gujin_param_extra_cmdline(),
			farptr2linear(code_adr(&instboot_info)),
			0, gujin_attr);
      puts (""); puts ("");
      }
  _BIOS_failedboot();
  return 1;
  }

#endif /* USER_SUPPORT != 0 */

#if !(SETUP & QUITE)
#if ASSEMBLY_TYPE == ASSEMBLY_DSES
static inline unsigned
check_cached_memsize (void)
  {
  unsigned long long rdtsc_start, rdtsc_val1, rdtsc_val2;
  unsigned val1, val2;

  DBG (("%s: checking memory over 64 Mb cached: ", __FUNCTION__));
  rdtsc();
  force_cache ((void *)linear2dsrelative(0x00070000), 16*1024);
  rdtsc_start = rdtsc();
  force_cache ((void *)linear2dsrelative(0x00070000), 16*1024);
  rdtsc_val1 = rdtsc() - rdtsc_start;
  DBG (("below 64 Mb: 0x%X, ", (unsigned)rdtsc_val1));
  force_cache ((void *)linear2dsrelative(64*1024*1024), 16*1024);
  rdtsc_start = rdtsc();
  force_cache ((void *)linear2dsrelative(64*1024*1024), 16*1024);
  rdtsc_val2 = rdtsc() - rdtsc_start;
  DBG (("at 64 Mb: 0x%X, ", (unsigned)rdtsc_val2));
  while (rdtsc_val2 >> 32) { /* squale to long */
      rdtsc_val1 >>= 8;
      rdtsc_val2 >>= 8;
      }
  val1 = (unsigned)rdtsc_val1;
  val2 = (unsigned)rdtsc_val2;
  DBG (("after squaling: 0x%X to 0x%X, ", val1, val2));
  if (val2 > 2 * val1) {
      DBG (("NOT cached, memory ignored!\r\n"));
      printf (" [Memory over 64Mb not cached, ignored] ");
      return 1;
      }
    else
      DBG (("cache OK.\r\n"));
  return 0;
  }
#endif
#endif

__attribute__ ((cdecl, regparm (0)))
int main (struct registers *regs)
  {
  struct gujin_param_attrib gujin_attr;
  int tmp;
  unsigned basemem, extendmem, freemem;
  enum detect_memory_enum index;

  init_malloc();
  DBG_INIT();

#ifdef SERIAL_INIT_MODEM_STRING
  if (BOOT1_COM_port() >= 0)
      BOOT1_putstr (SERIAL_INIT_MODEM_STRING);
#endif

#if USER_SUPPORT != 0
  UI_init (FIRST_BOOT1_COM_port() >= 0);
#endif

  gujin_attr = get_gujin_param_attrib ();

#if USER_SUPPORT != 0 /* else save space - would still be an error */
  /* Check that - if error there will be memory corruptions... */
#if (USER_SUPPORT & VGA_SUPPORT)
  if (!gujin_attr.vga_support)
#else
  if (gujin_attr.vga_support)
#endif
      print (PANIC_VGA_SUPPORT_NOT_EQUAL);
#endif /* USER_SUPPORT != 0 */

#if USER_SUPPORT != 0
  treat_startup_keys ();
#endif

  if (!BOOT1_DOS_running())
      gujin_attr.read_only_param = 0;
  set_gujin_param_attrib (gujin_attr);

  detect_processor (&processor);
  /* before serial init, if serial interface: */
  detect_environment (&processor);
  detect_keyboard ();
  PRINT (DOT_CRLF);
  detect_serial_timeout ();
  detect_mouse ();
#if !(SETUP & BIOS_ONLY)
  if (BOOT1_DOS_running()) {
      freemem = largest_free_himem();
      if (freemem)
	  PRINTF (HIMEM_MAXFREE, freemem);
      }
    else
#endif
      freemem = 0;
  index = detect_memory (&basemem, &extendmem);

#if !(SETUP & QUITE)
#if ASSEMBLY_TYPE == ASSEMBLY_DSES
  if (is_80486() && has_cpuid () && cpuid1().tsc && extendmem >= 63 * 1024 + 16)
      if (check_cached_memsize () != 0)
	  extendmem = 63 * 1024;
#endif
#endif

  PRINTF (BASE_MEMORY_U_EXTENDED_MEMORY_U_S,
	   basemem, extendmem, mem_name[index]);

#if USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT)
  if (BOOT1_COM_port() < 0)
      detect_screen();
#endif

  detect_interface();

  PRINT (DETECTING_DISKS);
  probedisk (regs->edx & 0xFF, gujin_attr);
  PRINTF (HAVE_FOUND_U_DISKS, DI.nbdisk);
  list_disks();

  disk_analyse (gujin_attr);
  PRINTF (FOUND_D_WAY_TO_BOOT_U_INITRD,
		boot_system.nb, boot_system.nb_initrd);

#if !(SETUP & QUITE)
  if (gujin_attr.verbose) {
      print (PRESS_PAUSE_TO_STOP_DISPLAY);
#if USER_SUPPORT != 0
      UI_getkey (4 * TICKS_PER_SECOND);
#else
      _BIOS_wait (4 * 1000 * 1000);
#endif
      }
#endif

  tmp = menu (regs, extendmem, freemem, gujin_attr);
  before_exit();
  return tmp;
  }

