/* boot.h */
#ifndef BOOT_H
#define BOOT_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/**
 ** The structures used in boot level 1:
 **/

/*
 * The "diskread" command structure for disk interface:
 */
typedef struct {
    struct {
	unsigned short next;
	unsigned short checksum;	/* sum of sector + checksum = 0 */
	unsigned short nbword;		/* size to calculate checksum */
	} __attribute__ ((packed)) header;

    union diskcmd {
	struct diskcmd_bios {	/* used if reg_ah != 0xF0 */
	    unsigned short	reg_dx;
	    unsigned short	reg_ax;
	    unsigned short	reg_cx;
	    /* no need to initialise the following, just a remainder: */
	    /* but checksum of boot1 is recalculated if first attempt failed! */
	    unsigned short	dummy;
	    unsigned long	will_be_set_to_esbx;
	    } __attribute__ ((packed)) int1302;
	struct diskcmd_ebios {	/* used (logically) if reg_ah = 0x42, 0x43... */
	    unsigned short	reg_dx;
	    unsigned short	reg_ax;
	    unsigned char	cst_16;
	    unsigned char	cst_0;		/* has to be zero */
	    unsigned char	nb_block;	/* max 127 */
	    unsigned char	cst2_0;		/* has to be zero */
	    unsigned short	offset, segment;
	    unsigned long long	lba;
	    } __attribute__ ((packed)) int1342;
	struct diskcmd_hard {	/* used if reg_ah == 0xF0 */
	    unsigned short	ide_dcr_address; /* usually base_ide+0x206 */
	    unsigned short	reg_ax;		/* %al has to be 0x0A */
	    unsigned short	base_ide_plus7;	/* added by 7, i.e. 1F7, 3F7 */
	    unsigned char	lba_head;
	    unsigned char	cylinder_high;
	    unsigned char	cylinder_low;
	    unsigned char	sector;
	    unsigned char	nb_sect;
	    unsigned char       ide_command;    /* to read with retry, 0x20 */
	    unsigned short	nb_sect_read;	/* keep it equal to nb_sect */
	    unsigned short	sector_size_div2;
	    } __attribute__ ((packed)) hardide;
	} bootloader2_cmd;
    } __attribute__ ((packed, aligned(1))) bootloader2_type;

/*
 * To use the serial (BIOS) interface, you need hardware flow control,
 * i.e. either a true VT connected or a null-modem cable (DTR-> DCD, RI,
 * CTS, DSR), and fill in the following structure (serialconf) with
 * the configuration you need.
 * If serialconf is null, more precisely the first "configuration
 * byte, i.e. 8 bits", then the video BIOS is used; and "port" will
 * not be copied in %bp, %bp will stay greater than 0xFF00.
 */
typedef struct {
    unsigned char	eightbit   :1;	/* else seven bits */
    unsigned char	constant_1 :1;
    unsigned char	doublestop :1;	/* else one stop bit */
    unsigned char	parity     :1;	/* else no parity generation/check */
    unsigned char	even       :1;	/* else odd */
    enum { B110 = 0, B150, B300, B600, B1200, B2400, B4800, B9600 }
			speed      :3;
    unsigned char	constant_0;
    signed short	port;		/* <0: screen, 0: COM1, 1: COM2 ... */
    } __attribute__ ((packed, aligned(1))) serialconf_type;

/*
 * The parameter block used by bootloader 1 to get
 * bootloader 2 in memory. Its size can not be modified !
 */
typedef struct {
    unsigned short	boot1_checksum;
    serialconf_type	serialconf;
    bootloader2_type	bootloader2;
    } __attribute__ ((packed, aligned(1))) boot1param_type;

/**
 ** The description of a boot sector:
 **/

/*
 * The list of partition kind:
 */
#define PARTITIONLIST() \
	P (EMPTY,		0x0,	"empty")		\
	P (DOS_FAT12,		0x1,	"Dos FAT12")		\
	P (XENIX_root,		0x2,	"XENIX root")		\
	P (XENIX_usr,		0x3,	"XENIX user")		\
	P (DOS_FAT16,		0x4,	"Dos FAT16")		\
	P (extended,		0x5,	"extended")		\
	P (BIG_DOS_FAT16,	0x6,	"Dos big FAT16")	\
	P (OS2_HPFS_NTFS,	0x7,	"OS2 HPFS NTFS")	\
	P (AIX,			0x8,	"AIX")			\
	P (AIX_BOOTABLE,	0x9,	"AIX bootable")		\
	P (OS2_BOOTMANAGER,	0x0A,	"OS2 bootmanager")	\
	P (Win95_FAT32,		0x0B,	"Win95 FAT32")		\
	P (Win95_FAT32LBA,	0x0C,	"Win95 FAT32 LBA")	\
	P (Win95_FAT16LBA,	0x0E,	"Win95 FAT16 LBA")	\
	P (Win95_extendedLBA,	0x0F,	"Win95 extended LBA")	\
	P (HiddenDOS_FAT12,	0x11,	"Hidden Dos FAT12")	\
	P (HiddenDOS_FAT16,	0x14,	"Hidden Dos FAT16")	\
	P (HiddenDOS_FAT32,	0x16,	"Hidden Dos FAT32")	\
	P (Novell,		0x51,	"Novell")		\
	P (Microport,		0x52,	"Microport")		\
	P (GNU_HURD,		0x63,	"GNU HURD")		\
	P (Novell_Netware2,	0x64,	"Novell Netware2")	\
	P (Novell_Netware3,	0x65,	"Novell Netware3")	\
	P (Linux_MINIX,		0x81,	"Linux MINIX")		\
	P (Linux_Swap,		0x82,	"Linux swap")		\
	P (Linux,		0x83,	"Linux")		\
	P (Linux_extended,	0x85,	"Linux extended")	\
	P (BSD_386,		0xA5,	"BSD 386")		\
	P (BSDI_fs,		0xB7,	"BSDI fs")		\
	P (BSDI_swap,		0xB8,	"BSDI swap")		\
	P (CP_M,		0xDB,	"CP/M")			\
	P (DOS_access,		0xE1,	"DOS access")		\
	P (DOS_R_O,		0xE3,	"DOS R/O")		\
	P (DOS_secondary,	0xF2,	"DOS secondary")	\
	P (LinuxAutoRaid,	0xFD,	"Linux RAID autodetect")

/*
 * The boot sector contains this structure at its beginning:
 * (after the initial jump, i.e. first 3 bytes)
 */
typedef struct {
    unsigned char	String[8];	/* blank padded ! */
    unsigned short	Bytepersector;
    unsigned char	Sectorpercluster;
    unsigned short	Reservedsector;	/* not usefull or not working ? */
    unsigned char	NbFAT;
    unsigned short	NbRootdirentry;
    unsigned short	NbTotalsector;
    unsigned char	Mediadescriptor;
    unsigned short	NbSectorperFAT;
    /* DOS 3+ : */
    unsigned short	NbSectorpertrack;
    unsigned short	NbHead;
    unsigned long	NbHiddensector;	/* not usefull or not working ? */
    /* DOS 4+ : */
    unsigned long	NbTotalsector2;
    unsigned char	PhysicaldriveNb;
    unsigned char	Reserved;
    unsigned char	Signaturebyte0x29;
    unsigned long	VolumeserialNb;
    unsigned char	Volumelabel[11];	/* blank padded ! */
    unsigned char	FileSysName[8];		/* blank padded ! */
    } __attribute__ ((packed, aligned(1))) bootbefore_type;

/*
 * The boot sector contains this structure at its end:
 */
typedef struct {
    unsigned long	WindowsNTmarker;
    unsigned short	Unknown;
    struct bootsect_partition_str {
	enum {non_bootable = 0, bootable = 0x80, invalid}
			indicator		:8;
	unsigned char	begin_head		:8;
	unsigned char	begin_cylinder_msb	:2;
	unsigned char	begin_sector		:6;
	unsigned char	begin_cylinder_lsb	:8;
	enum system_indicator_enum {
#define P(shortname, value, longname) shortname = value,
	    PARTITIONLIST()
#undef P /*(shortname, value, longname)*/
	    }	system_indicator		:8;
	unsigned char	end_head		:8;
	unsigned char	end_cylinder_msb	:2;
	unsigned char	end_sector		:6;
	unsigned char	end_cylinder_lsb	:8;
	unsigned long	nb_sector_before;
	unsigned long	nb_sector;
	} __attribute__ ((packed)) bootsect_partition[4];
    unsigned short	Signature0xAA55;
    } __attribute__ ((packed, aligned(1))) bootafter_type;

/*
 * A sample boot sector:
 */
typedef struct {
    char		jmpinstruction[3];
    bootbefore_type	before;
    unsigned char	code_part[512 - 3
				  - sizeof(bootbefore_type)
				  - sizeof(boot1param_type)
				  - sizeof(bootafter_type) ];
    boot1param_type	param;
    bootafter_type	after;
    } __attribute__ ((packed)) bootsector_type;

/**
 ** The interface to boot1 when in boot2:
 **/

/*
 * The "main()" function of boot level two is called
 * with this structure as parameter:
 * (These registers are those boot level one has found, but %%ss:%%esp !)
 */
struct registers {
	unsigned edi, esi, ebp, initsp;
	unsigned ebx, edx, ecx, eax;
	unsigned short gs, fs, aligngap, es, ds;
	unsigned short ip, cs, flags;
	} __attribute__ ((packed));

/*
 * Some references for instboot to patch executable:
 */
struct instboot_info_str {
    unsigned start_adr;
    unsigned boot1param_adr;
    unsigned bootend_adr;
    unsigned bootchain_adr;
    unsigned bootchain_nbblock;
    unsigned gujin_param_adr;
    unsigned patch1adr, patch2adr, patch3adr, patch4adr;
    unsigned sbss, edata, deltaseg, extraseg;
    };

/*
 * The variables of Gujin saved between boots:
 */
#define APPLY_ATTRIBUTE_ORDER(apply) \
	/* name                 default                             greyed */ \
    apply(verbose,		  1,					0) \
    apply(lock_bpp,		  1,					0) \
    apply(lock_text_graphic,	  0,					0) \
    apply(force_textmode,	  1,					0) \
    apply(menu_with_disk,	  1,					0) \
    apply(menu_with_partition,	  1,					0) \
    apply(VGA_interface,	  !!(USER_SUPPORT & VGA_SUPPORT),	0) \
    apply(refuse_8x14font,	  0,					UI.parameter.charheigh == 14) \
    apply(VESA_interface,	  !!(USER_SUPPORT & VESA_WINDOW),	0) \
    apply(enable_VESA_hardwin,	  !!(USER_SUPPORT & VESA_HARDWINDOW),	0) \
    apply(VESA2_interface,	  !!(USER_SUPPORT & VESA_LINEAR),	0) \
    apply(enable_joystick,	  !!(USER_SUPPORT & JOYSTICK_SUPPORT),	0) \
    apply(search_disk_mbr,	  1,					BOOT1_DOS_running()) \
    apply(prefer_enddisk_mbr,	  0,					BOOT1_DOS_running()) \
    apply(search_part_mbr,	  1,					BOOT1_DOS_running()) \
    apply(search_topdir_kernel,	  1,					0) \
    apply(search_subdir_kernel,	  1,					0) \
    apply(probe_bios_floppy_disk, !!(DISK_SUPPORT & (BIOS_SUPPORT | EBIOS_SUPPORT)), 0) \
    apply(probe_bios_hard_disk,   !!(DISK_SUPPORT & (BIOS_SUPPORT | EBIOS_SUPPORT)), 0) \
    apply(probe_ide_disk,	  !!(DISK_SUPPORT & IDE_SUPPORT),	BOOT1_DOS_running()) \
    apply(probe_atapi_cdrom,	  0,					1) \
    apply(probe_dos_disk,	  !!(DISK_SUPPORT & DOS_SUPPORT),	!BOOT1_DOS_running())

enum gujin_param_attrib_enum {
#define Apply(field, deflt, grey)	enum_ ## field,
    APPLY_ATTRIBUTE_ORDER (Apply)
#undef Apply
    enum_gujin_param_attrib_end
    };
#define TOTAL_ATTRIBUTE		enum_gujin_param_attrib_end

struct gujin_param_str {
    unsigned		magic;
    unsigned short	version;
    char		dot_msg[4];
    char		go_msg[8];
    unsigned short	default_video_mode;
    unsigned short	default_text_video_mode, default_graphic_video_mode;
    struct gujin_param_attrib {

#define Apply(field, deflt, grey)	unsigned field : 1;
	APPLY_ATTRIBUTE_ORDER(Apply)
#undef Apply

	unsigned available		: 32 - TOTAL_ATTRIBUTE - 4;

	unsigned read_only_param	: 1; /* for instance boot.exe */
	unsigned vga_support		: 1; /* constant */
	unsigned big_malloc		: 1; /* constant */
	unsigned bios_only		: 1; /* constant */
	} __attribute__ ((packed)) attrib;
#if (USER_SUPPORT & VGA_SUPPORT) || defined (INSTBOOT_C)
    unsigned char       cardname[32];	/* for autoreset, from VESA */
    unsigned char       VGA_valid_mode[16];
    struct vga_mode_str {
	unsigned isgraphic  : 1;
	unsigned font8x14   : 1;
	unsigned number     : 7;
	unsigned bpp        : 4; /* 0 to 15, 8bpp max for VGA BIOS */
	unsigned heigh      : 11; /* 0 to 2048 */
	unsigned width_div8 : 8; /* 0 to 2048, div8 only for graphic */
	} __attribute__ ((packed)) vga_mode[30];
#endif
    struct autoload_str {
	unsigned char last_loaded;
	unsigned char total_loadable;
	unsigned char init_page;
	unsigned char total_page;
	} __attribute__ ((packed)) autoload;
    unsigned char	timeout_autoload;
    unsigned char	align;
    char		extra_cmdline[64];
    unsigned short	partial_checksum;
    } __attribute__ ((packed));

/* while we are here, these few obvious things: */
static inline int
displayconfig_changed (struct gujin_param_attrib first,
		       struct gujin_param_attrib second)
  {
  union {
      struct gujin_param_attrib bits;
      unsigned all;
      } one, two, mask = { bits: {
	lock_text_graphic: 1,
	lock_bpp: 1,
	VGA_interface: 1,
	refuse_8x14font: 1,
	VESA_interface: 1,
	enable_VESA_hardwin : 1,
	VESA2_interface: 1,
	}};
  one.bits = first;
  two.bits = second;

  return (one.all & mask.all) != (two.all & mask.all);
  }

static inline int
diskconfig_changed (struct gujin_param_attrib first,
		    struct gujin_param_attrib second)
  {
  union {
      struct gujin_param_attrib bits;
      unsigned all;
      } one, two, mask = { bits: {
	probe_dos_disk: 1,
	probe_atapi_cdrom: 1,
	probe_ide_disk: 1,
	probe_bios_hard_disk: 1,
	probe_bios_floppy_disk: 1,
	}};
  one.bits = first;
  two.bits = second;

  return (one.all & mask.all) != (two.all & mask.all);
  }

static inline int
searchconfig_changed (struct gujin_param_attrib first,
		      struct gujin_param_attrib second)
  {
  union {
      struct gujin_param_attrib bits;
      unsigned all;
      } one, two, mask = { bits: {
	search_disk_mbr:	1,
	prefer_enddisk_mbr:	1,
	search_part_mbr:	1,
	search_topdir_kernel:	1,
	search_subdir_kernel:	1,
	}};
  one.bits = first;
  two.bits = second;

  return (one.all & mask.all) != (two.all & mask.all);
  }

/*
 * The structure of directories (BIOS/IDE access to FAT12/FAT16)
 */
typedef struct {
    char name[8]; /* offset 0: end_dir=0, firstcharE5=0x05, erased=0xE5 */
    char extension[3];
    struct file_attr_str {
	unsigned char	readonly	: 1;
	unsigned char	hidden		: 1;
	unsigned char	system		: 1;
	unsigned char	volumelabel	: 1;
	unsigned char	subdirectory	: 1;
	unsigned char	archive		: 1;
	unsigned char	unused		: 2;
	} __attribute__ ((packed)) attributes;
    char reserved[10];
    struct {
	unsigned seconddiv2	: 5;
	unsigned minute		: 6;
	unsigned hour		: 5;
	unsigned day		: 5;	/* 1..31 */
	unsigned month		: 4;	/* 1..12 */
	unsigned year		: 7;	/* 1980 + x */
	} __attribute__ ((packed)) date;
    unsigned short	cluster;
    unsigned long	size;
    } __attribute__ ((packed)) directory_type;

#ifdef SETUP	/* not for instboot.c */

/*
 * Interface functions:
 */
#if (SETUP & EXTRA_SEGMENT) && defined (CALLING_FROM_EXTRASEG)
#define ASMNAME(x)	asm (STRING(extraseg_ ## x))
#else
#define ASMNAME(x)	/* call in same segment */
#endif


void BOOT1_putstr (const char *str) ASMNAME(BOOT1_putstr);
void BOOT1_initdisks (unsigned char drive) ASMNAME(BOOT1_initdisks);

unsigned char BOOT1_diskread (union diskcmd *str, farptr buffer)
	ASMNAME(BOOT1_diskread);

#ifdef WRITE_ENABLE
unsigned char BOOT1_diskwrite (union diskcmd *str, farptr buffer)
	ASMNAME(BOOT1_diskwrite);
#endif

#if SETUP & UPDATE_BOOTPARAM
unsigned char BOOT1_update_bootparam (void) ASMNAME(BOOT1_update_bootparam);
#endif

#undef ASMNAME

/*
 * Few inlines, working even if called from extraseg,
 * because code_adr() references the initial %cs segment:
 */
#if 0 /* optimise them to reduce code size */

#if SETUP & BIOS_ONLY
#define BOOT1_DOS_running()     0
#else
extern inline const unsigned char BOOT1_DOS_running (void)
  {
  extern unsigned char dospatch4adr[];

  return peekl (code_adr(dospatch4adr)) == 0x21CD4CB4;
  }
#endif

extern inline const serialconf_type BOOT1_COM (void)
  {
  extern boot1param_type boot1param;

  return peekl (code_adr(&boot1param.serialconf));
  }

extern inline const signed short BOOT1_COM_port (void)
  {   /* < 0 if screen interface, always sign extend it ! */
  return BOOT1_COM().port;
  }

#else /* 0 */

#if SETUP & BIOS_ONLY
#define BOOT1_DOS_running()	0
#else
extern const unsigned char dos_running;
#define BOOT1_DOS_running()	dos_running
#endif

extern const serialconf_type copy_boot1param_serialconf;
#define BOOT1_COM()		copy_boot1param_serialconf
#define FIRST_BOOT1_COM_port()	copy_boot1param_serialconf.port

#if (USER_SUPPORT == 0) || (USER_SUPPORT & SERIAL_SUPPORT)
#define BOOT1_COM_port()	FIRST_BOOT1_COM_port()
#else
#define BOOT1_COM_port()	-1
#endif

#endif /* 0 */

extern inline void BOOT1_setcolor (unsigned char color)
  {  /* Only in graphic modes (BIOS) */
  extern unsigned char color_overwrite[];

  codeseg_pokeb (color_overwrite, color);
  }

extern inline unsigned getTFflagCounter (void)
  {
  extern unsigned TFflagCounter;

  return codeseg_peekl (&TFflagCounter);
  }

extern inline void setTFflagCounter (unsigned val)
  {
  extern unsigned TFflagCounter;

  codeseg_pokel (&TFflagCounter, val);
  }

extern inline unsigned short getDefaultVideoMode (void)
  {
  extern struct gujin_param_str gujin_param;

  return codeseg_peekw (&gujin_param.default_video_mode);
  }

extern inline void setDefaultVideoMode (unsigned short mode)
  {
  extern struct gujin_param_str gujin_param;

  codeseg_pokew (&gujin_param.default_video_mode, mode);
  }

extern inline unsigned short getDefaultTextVideoMode (void)
  {
  extern struct gujin_param_str gujin_param;

  return codeseg_peekw (&gujin_param.default_text_video_mode);
  }

extern inline void setDefaultTextVideoMode (unsigned short mode)
  {
  extern struct gujin_param_str gujin_param;

  codeseg_pokew (&gujin_param.default_text_video_mode, mode);
  }

extern inline unsigned short getDefaultGraphicVideoMode (void)
  {
  extern struct gujin_param_str gujin_param;

  return codeseg_peekw (&gujin_param.default_graphic_video_mode);
  }

extern inline void setDefaultGraphicVideoMode (unsigned short mode)
  {
  extern struct gujin_param_str gujin_param;

  codeseg_pokew (&gujin_param.default_graphic_video_mode, mode);
  }

extern inline struct gujin_param_attrib get_gujin_param_attrib (void)
  {
  extern struct gujin_param_str gujin_param;
  struct gujin_param_attrib returned;

#ifdef CALLING_FROM_EXTRASEG
  asm ASMREAD (
"	movw	2+codeseg_adr,%%fs				\n"
"	mov	%%fs:%1,%0	# automatic size decision	\n"
	: "=r" (returned) : "g" (gujin_param.attrib));
#else
  asm ASMREAD (" mov	%%cs:%1,%0      # automatic size decision "
	: "=r" (returned) : "g" (gujin_param.attrib));
#endif

  return returned;
  }

extern inline void set_gujin_param_attrib (struct gujin_param_attrib attr)
  {
  extern struct gujin_param_str gujin_param;

#ifdef CALLING_FROM_EXTRASEG
  asm volatile (
"	movw	2+codeseg_adr,%%fs				\n"
"	mov	%0,%%fs:%1	# automatic size decision	\n"
	: : "r" (attr), "g" (gujin_param.attrib));
#else
  asm volatile (" mov	%0,%%cs:%1	# automatic size decision "
	: : "r" (attr), "g" (gujin_param.attrib));
#endif
  }

extern inline farptr get_gujin_param_extra_cmdline (void)
  {
  extern struct gujin_param_str gujin_param;
  return code_adr(gujin_param.extra_cmdline);
  }

extern inline void
setup_last_boot_default(unsigned char last_loaded,
			unsigned char total_loadable,
			unsigned char init_page,
			unsigned char total_page)
  {
  extern struct gujin_param_str gujin_param;
  union {
      struct autoload_str str;
      unsigned data;
      } autoload = { str: {
	last_loaded:	last_loaded,
	total_loadable:	total_loadable,
	init_page:	init_page,
	total_page:	total_page
      }};

  codeseg_pokel (&gujin_param.autoload, autoload.data);
  }

extern inline struct autoload_str
get_last_boot_default(void)
  {
  extern struct gujin_param_str gujin_param;
  union {
      struct autoload_str str;
      unsigned data;
      } autoload;

  autoload.data = codeseg_peekl (&gujin_param.autoload);
  return autoload.str;
  }

extern inline void
set_timeout_autoload (unsigned char timeout)
  {
  extern struct gujin_param_str gujin_param;
  codeseg_pokeb (&gujin_param.timeout_autoload, timeout);
  }

extern inline unsigned char
get_timeout_autoload (void)
  {
  extern struct gujin_param_str gujin_param;
  return codeseg_peekb (&gujin_param.timeout_autoload);
  }

extern inline void disable_update_bootparam(void)
  {
  struct gujin_param_attrib tmp = get_gujin_param_attrib();
  tmp.read_only_param = 1;
  set_gujin_param_attrib (tmp);
  }


#if USER_SUPPORT & VGA_SUPPORT
extern inline unsigned VIDEO_mode_is_valid (unsigned char mode)
  {
  /* return (gujin_param.VGA_valid_mode[mode/8] & (1 << ((mode) % 8)) != 0),
     i.e. return 0 or 1 */
  extern struct gujin_param_str gujin_param;
  unsigned char shift;
  void *adr;

  adr = &gujin_param.VGA_valid_mode[divrem8 (mode, &shift)];
  return !!(codeseg_peekb (adr) & (1 << shift));
  }

extern inline void VIDEO_mode_invalidate (unsigned char mode)
  {
  /* gujin_param.VGA_valid_mode[mode/8] &= ~(1 << ((mode) % 8)) */
  extern struct gujin_param_str gujin_param;
  unsigned char shift;
  void *adr;

  adr = &gujin_param.VGA_valid_mode[divrem8 (mode, &shift)];
  codeseg_pokeb (adr, codeseg_peekb (adr) & ~ (1 << shift));
  }

extern inline void VIDEO_mode_revalidate (unsigned char mode)
  {
  /* gujin_param.VGA_valid_mode[mode/8] |= 1 << ((mode) % 8) */
  extern struct gujin_param_str gujin_param;
  unsigned char shift;
  void *adr;

  adr = &gujin_param.VGA_valid_mode[divrem8 (mode, &shift)];
  codeseg_pokeb (adr, codeseg_peekb (adr) | (1 << shift));
  }

extern inline unsigned VIDEO_mode_max (void)
  {
#if (USER_SUPPORT & VGA_SUPPORT)
  extern struct gujin_param_str gujin_param;

  return 8 * sizeof (gujin_param.VGA_valid_mode) - 1; /* 8 bits per byte */
#else
  return 0x7F;
#endif
  }

extern inline void VIDEO_mode_initname (farptr cardname)
  {
  extern struct gujin_param_str gujin_param;
  unsigned i;

  for (i = 0; i < nbof (gujin_param.cardname); i++) {
      unsigned char tmp = peekb (cardname++);
      codeseg_pokeb (&gujin_param.cardname[i], tmp);
      if (tmp == '\0')
	  break;
      }
  }

extern inline int VIDEO_mode_checkname (farptr cardname)
  {
  extern struct gujin_param_str gujin_param;
  unsigned i;

  for (i = 0; i < nbof (gujin_param.cardname); i++) {
      unsigned char tmp = peekb (cardname++);
      if (codeseg_peekb (&gujin_param.cardname[i]) != tmp)
	  return 1;
      if (tmp == '\0')
	  break;
      }
  return 0;
  }

/* GCC hates casts to long there: */
extern inline void
VGA_get_saved_mode (unsigned index, struct vga_mode_str *mode)
  {
  extern struct gujin_param_str gujin_param;

#ifdef CALLING_FROM_EXTRASEG
  asm ASMREAD (
"	movw	2+codeseg_adr,%%fs				\n"
"	mov	%%fs:%1,%0	# automatic size decision	\n"
	: "=r" (*mode) : "g" (gujin_param.vga_mode[index]));
#else
  asm ASMREAD (" mov	%%cs:%1,%0	# automatic size decision "
	: "=r" (*mode) : "g" (gujin_param.vga_mode[index]));
#endif
  }

extern inline void
VGA_set_saved_mode (unsigned index, struct vga_mode_str mode)
  {
  extern struct gujin_param_str gujin_param;

#ifdef CALLING_FROM_EXTRASEG
  asm volatile (
"	movw	2+codeseg_adr,%%fs				\n"
"	mov	%0,%%fs:%1	# automatic size decision	\n"
	: : "r" (mode), "g" (gujin_param.vga_mode[index]));
#else
  asm volatile (" mov	%0,%%cs:%1	# automatic size decision "
	: : "r" (mode), "g" (gujin_param.vga_mode[index]));
#endif
  }

extern inline void VIDEO_mode_reset (void)
  {
  extern struct gujin_param_str gujin_param;
  unsigned char *adr = gujin_param.VGA_valid_mode;
  unsigned char cpt;

  for (cpt = 0; cpt < sizeof (gujin_param.VGA_valid_mode); cpt++)
      codeseg_pokeb (adr++, 0xFF);
  for (cpt = 0; cpt < nbof (gujin_param.vga_mode); cpt++)
      VGA_set_saved_mode (cpt, ((struct vga_mode_str) {}));
  setDefaultGraphicVideoMode (0xFFFF);
  setDefaultTextVideoMode (0xFFFF);
  setDefaultVideoMode (0xFFFF);
  }

#else /* VGA_SUPPORT */
extern inline void VIDEO_mode_reset (void)
  {
  setDefaultGraphicVideoMode (0xFFFF);
  setDefaultTextVideoMode (0xFFFF);
  setDefaultVideoMode (0xFFFF);
  }
#endif /* VGA_SUPPORT */

#endif /* SETUP, instboot.c */
#endif /* BOOT_H */
