/* disk.c */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"
#if SETUP & EXTRA_SEGMENT	/* before other includes */
#define CALLING_FROM_EXTRASEG
#endif

#include "library.h"
#include "messages.h"
#include "boot.h"
#include "disk.h"
#include "bios.h"
#include "debug.h"

#ifdef DISK_USE_MALLOC
#include "malloc.h"
#endif

#if DISK_SUPPORT & DOS_SUPPORT
#include "dos.h"
#endif

#if DISK_SUPPORT & IDE_SUPPORT
#include "ide.h"

/*
 * Take care of the order here: Linux will get as
 * parameter the IDE I/O address with increasing numbers:
 * If the PC has IDE at 0x1F0, 0x170, 0x168 and 0xe820,
 * the command line will be:
 * ide0=0x1F0,... ide1=0x170,... ide2=0x168,... ide3=0xe820,...
 *
 * Update (gujin-0.5): The first IDE addresses are in the BIOS order,
 * if BIOS disks 0x80 is detected by BIOS at 0xe820, the previous
 * configuration will be:
 * ide0=0xe820,... ide1=0x1F0,... ide2=0x170,... ide3=,0x168...
 */
static const unsigned short idearray[] = {
    0x1F0, 0x170, 0x1E8, 0x168, 0x1E0, 0x160, 0xe820, 0xe828
    };

//#define COMMANDSETUP	0xF008	/* interrupts enabled, no driver present */
#define COMMANDSETUP    0xF00A

#endif

/*
 * Hack: display info on stdout when a read/write error happens.
 */
//#define DISPLAY_DISK_ERROR

/**
 ** These do not need to even exist when DEBUG is not defined:
 **/
#if defined (DEBUG) && (DEBUG & DEBUG_DISK)
SEGMENT_EXTRA(BIOSDISK_getstatus) static inline unsigned char
BIOSDISK_getstatus (unsigned char drive)
  {
  unsigned char status;

  if (_BIOSDISK_getstatus (drive, &status) && status == 0)
      status = LOCALERROR;
  return status;
  }

SEGMENT_EXTRA(BIOSDISK_error) static const char *
BIOSDISK_error (unsigned char status)
  {
  const char *str;

  switch (status) {
      case 0x00: str = ""; break;
      case 0x01: str = "invalid function or parameter"; break;
      case 0x02: str = "bad sector"; break;
      case 0x03: str = "write protect"; break;
      case 0x04: str = "sector not found"; break;
      case 0x05: str = "disk reset failed"; break;
      case 0x06: str = "disk changed or removed"; break;
      case 0x07: str = "bad disk parameter table"; break;
      case 0x08: str = "DMA overrun"; break;
      case 0x09: str = "DMA cross 64K boundary"; break;
      case 0x0A: str = "bad disk sector flag"; break;
      case 0x0B: str = "bad disk track"; break;
      case 0x0C: str = "unsupported track"; break;
      case 0x0D: str = "invalid sector nb"; break;
      case 0x0E: str = "address mark detected"; break;
      case 0x0F: str = "DMA arbitration problem"; break;
      case 0x10: str = "ECC/CRC error for read"; break;
      case 0x11: str = "ECC recovered data read"; break;
      case 0x20: str = "controller failure"; break;
      case 0x31: str = "no media in drive"; break;
      case 0x32: str = "invalid CMOS drive type"; break;
      case 0x40: str = "seek failure"; break;
      case 0x80: str = "timeout, drive not ready"; break;
      case 0xAA: str = "disk drive not ready"; break;
      case 0xB0: str = "volume not locked in drive"; break;
      case 0xB1: str = "volume locked in drive"; break;
      case 0xB2: str = "volume not removable"; break;
      case 0xB3: str = "volume in use"; break;
      case 0xB4: str = "volume lock count exceeded"; break;
      case 0xB5: str = "eject request failed"; break;
      case 0xBB: str = "disk undefined error"; break;
      case 0xCC: str = "disk write fault"; break;
      case 0xE0: str = "disk status error"; break;
      case 0xFF: str = "sense operation failed"; break;

      case LOCALERROR: str = "local software error"; break;
      default: str = "unknown error"; break;
      }
  return str;
  }

#if DISK_SUPPORT & IDE_SUPPORT
SEGMENT_EXTRA(IDE_requestsense) static char *
IDE_requestsense (unsigned char drive, unsigned short ideIOadr)
  {bound_stack();{
  unsigned short error_sense;
  unsigned char curdrive;
  static char buffer[100], *ptr, *end = &buffer[sizeof buffer - 1];
  const char *tmp;

  ptr = buffer + sprintf (buffer, "%$ [requestsense ", end);

  if (drive > 1) {
      STRINIT (ptr, "ERROR: drive neither master nor slave!] ");
      return buffer;
      }

  if ((_IDE_get_status (ideIOadr) & 1) == 0) {
      STRINIT (ptr, "ERROR: sense requested when no error in status!] ");
      return buffer;
      }

  curdrive = _IDE_get_drivehead (ideIOadr);
  if (((curdrive >> 4) & 1) != drive) {
      ptr += sprintf (ptr, "%$<swap> ", end);
      _IDE_set_drivehead (ideIOadr, curdrive ^ 0x10);
      }
  error_sense = _IDE_request_sense (ideIOadr);

  ptr += sprintf (ptr, "%$0x%X : %s%s%s%s%s ", end,
		  error_sense,
		  (error_sense & 0x8000)? "BBK " : "",
		  (error_sense & 0x4000)? "UNC " : "",
		  (error_sense & 0x1000)? "IDNF " : "",
		  (error_sense & 0x0400)? "ARRT " :"",
		  (error_sense & 0x0100)? "AMNF " : ""
		  );

  switch ((unsigned char) error_sense) {
    case 0x00: tmp = " no error"; break;
    case 0x01: tmp = " self test OK"; break;
    case 0x09: tmp = " misc error"; break;
    case 0x20: tmp = " invalid command"; break;
    case 0x21: tmp = " invalid address"; break;
    case 0x2F: tmp = " address overflow"; break;
    case 0x11: tmp = " uncorr ECC"; break;
    case 0x18: tmp = " corrected ECC"; break;
    default:   tmp = "???"; break;
    }
  ptr += sprintf (ptr, "%$ %s] ", end, tmp);

  return buffer;
  }}
#endif /* IDE_SUPPORT */

#endif /* DEBUG */

/**
 ** When WRITE_ENABLE not defined:
 **/
#ifndef WRITE_ENABLE
SEGMENT_EXTRA(BOOT1_diskwrite) static inline unsigned char
BOOT1_diskwrite (union diskcmd *str, farptr buffer)
  {
  return LOCALERROR;
  }
#endif

/**
 ** The low level BIOS functions:
 **/
#if DISK_SUPPORT & BIOS_SUPPORT

SEGMENT_EXTRA(BIOSDISK_gettype) static inline unsigned char
BIOSDISK_gettype (unsigned char disk, unsigned *nbsect)
  {
  unsigned char status;

  DDBG ((" Gettype: "));
  if (_BIOSDISK_gettype (disk, nbsect, &status)) {
      DDBG (("error. "));
      return 0;	/* no disk here */
      }

  DDBG (("%s", (status == 1)? "floppy w/o change detect" :
	(status == 2)? "floppy with change detect" :
	(status == 3)? "hard disk" :
	"unknown"));
  if (status >= 1 && status <= 3) {
      if (*nbsect != 0 && *nbsect != disk)	/* but buggy Award 486 BIOS! */
	  DDBG ((", probably nb sector = %u", *nbsect));
      DDBG (("\r\n"));
      }

  return status; /* return non-zero if present */
  }

SEGMENT_EXTRA(BIOSDISK_getparam) static inline unsigned char
BIOSDISK_getparam (unsigned char disk, struct diskparam_str *dp)
  {
  unsigned char CmosDriveType, status;
  unsigned short tmpcx, tmpdx;
  farptr DBTadr;

  if (_BIOSDISK_getparam (disk, &tmpcx, &tmpdx, &DBTadr,
			  &CmosDriveType, &status)) {
      DDBG (("BIOS getparam error.  "));
      return 1;
      }

  /* do not believe number of disks */
  dp->nbhead = (tmpdx >> 8) + 1;
  dp->nbsectorpertrack = tmpcx & 0x3F;
  dp->nbcylinder = (tmpcx >> 8) + ((tmpcx << 2) & 0x0300) + 1;

  DDBG (("BIOS reports: nbdisk: %u, CMOS drive type: %u, DBT adr: 0x%X",
	 tmpdx & 0xFF, CmosDriveType, DBTadr));

  if (DBTadr > 256*4) { /* check that it does not point in interrupt table ! */
      unsigned char bps = peekb (DBTadr
				 + offsetof (struct DBT_str, bytepersector));
      if (bps <= 3) {
	  dp->bytepersector = 128 << bps;
	  DDBG ((", %u byte/sect", dp->bytepersector));
	  }
	else {
	  dp->bytepersector = 512;
	  DDBG ((", 512 BYTE/SECT"));
	  }
      }
    else {
      dp->bytepersector = 512;
      DDBG ((", 512 bytes/sect"));
      }
  DDBG ((", nbhead: %u, nbsectorpertrack: %u, nbcylinder: %u; ",
	 dp->nbhead, dp->nbsectorpertrack, dp->nbcylinder ));
  return 0;
  }

/* Accept (unchecked) head up to 255 and number of sector to read up to 128 */
SEGMENT_EXTRA(biosRWsector) static inline unsigned char
biosRWsector (  unsigned char	drive,
		unsigned short	cylinder,
		unsigned char	head,
		unsigned char	sector,
		unsigned char	number,
		unsigned char	multiple, /* boolean */
		farptr		buffer,
		const unsigned	read)
  {
  union diskcmd str = { int1302: {
      reg_dx: ((unsigned short)head << 8) | drive,
      reg_ax: ((multiple? (read? 0x21 : 0x22) : (read? 0x02 : 0x03)) << 8)
		| (number & 0x7F),
      reg_cx: ((cylinder & 0xFF) << 8)  | ((cylinder & 0x300) >> 2)
					| (sector & 0x3F)
      }};

  DDBG ((" <BIOS %s drive %s 0x%X, %d sector(s) at C/H/S: %d/%d/%d> ",
	 (read) ? "reading" : "writing",
	 (multiple != 0)? "[multiple] " : "",
	 drive,
	 number & 0x7F,
	 cylinder,
	 head,
	 sector
	 ));

#ifndef DISPLAY_DISK_ERROR
  return read? BOOT1_diskread (&str, buffer) : BOOT1_diskwrite (&str, buffer);
#else
  {
  unsigned char tmp;

  tmp = read? BOOT1_diskread (&str, buffer) : BOOT1_diskwrite (&str, buffer)
  if (tmp) {
      printf ("%s error %u, C/H/S = %u/%u/%u, %u sector at 0x%X\r\n",
		__FUNCTION__, tmp, cylinder, head, sector, number, buffer);
      printf ("AX = 0x%X, CX = 0x%X, DX = 0x%X\r\n",
		str.int1302.reg_ax, str.int1302.reg_cx, str.int1302.reg_dx);
      }
  return tmp;
  }
#endif
  }

#if !(DISK_SUPPORT & NO_FLOPPY_SUPPORT)
/* variable size array: not declared in inline function:
static inline */ SEGMENT_EXTRA(biosFD_RWsector) static unsigned char
biosFD_RWsector (struct diskparam_str *dp,
		 unsigned char	drive,
		 unsigned short	cylinder,
		 unsigned char	head,
		 unsigned char	sector,
		 unsigned char	number,
		 farptr		buffer,
		 const unsigned	read)
  {
  /* the buffers may not be aligned on 512 bytes boundaries */
  /* one of "buffer" and "tmp_buffer" is not crossing a DMA fontier
	because the two fit in the 64 Kb stack: */
  unsigned char tmp_buffer[dp->bytepersector];

  DDBG ((" <FD BIOS %s %u sector starting at %u/%u/%u at 0x%X, ",
		read? "reading" : "writing", number,
		cylinder, head, sector, buffer));
  /* This set segment bit 4..11 to zero, each page is 64Kb: */
  buffer = linear2farptr (farptr2linear (buffer));
  DDBG (("adjusted address 0x%X> ", buffer));

  while (number != 0) {
      /* Three tries, because if the floppy disk has been changed, first
	 function call returns "disk changed" (after resetting BIOS),
	 and the motor of the floppy drive may be slow to start. */
      unsigned char maxtry = 3, tmp, partial, tmp_partial;
      unsigned max;
      farptr tmp_adr;

      partial = number;

      if (partial > 18) {	/* 1.68M floppy */
	  partial = 18;
	  DDBG ((", splitting to %u due to 18 sectors frontier, ", partial));
	  }

      max = dp->nbsectorpertrack - sector + 1;
      if (partial > max) {
	  partial = max;
	  DDBG ((", splitting to %u due to HEAD frontier", partial));
	  }

      max = (0x10000 - (buffer & 0xFFFF)) / dp->bytepersector;
      if (partial > max) {
	  partial = max;
	  DDBG ((", splitting to %u due to DMA frontier", partial));
	  }

      if (partial != 0) {
	  DDBG ((", using direct: "));
	  tmp_partial = partial;
	  tmp_adr = buffer;
	  }
	else { /* DMA inside the 512 bytes buffer */
	  DDBG ((", using indirect: "));
#ifdef WRITE_ENABLE
	  if (!read)
	      lmemcpy (tmp_buffer, smallestoffset (buffer),
			sizeof (tmp_buffer));
#endif /* WRITE_ENABLE */
	  tmp_partial = 1;
	  tmp_adr = stack_adr(tmp_buffer);
	  }

      do {
#ifdef WRITE_ENABLE
	  tmp = biosRWsector (drive, cylinder, head, sector, tmp_partial,
				0, tmp_adr, read);
#else /* WRITE_ENABLE */
	  tmp = biosRWsector (drive, cylinder, head, sector, tmp_partial,
				0, tmp_adr, 1);
#endif /* WRITE_ENABLE */
	  } while (tmp != 0 && tmp != 0x80 && tmp != 0xAA && --maxtry);

      if (tmp) {
	  DDBG (("error 0x%X] ", tmp));
	  return tmp;
	  }
	else
	  DDBG (("OK, "));

      if (partial == 0) {
#ifdef WRITE_ENABLE
	  if (read)
#endif /* WRITE_ENABLE */
	      smemcpy (smallestoffset (buffer), tmp_buffer,
			sizeof (tmp_buffer));
	  partial = 1;
	  }

      number -= partial;
      buffer += partial * dp->bytepersector;
      if (buffer & ~0xF000FFFFU) {
	  unsigned tmp = (buffer >> 16) & 0xFFF;
	  buffer &= 0xF000FFFFU;
	  buffer += tmp << 28;
	  }
      if ((unsigned short)sector + partial <= dp->nbsectorpertrack)
	  sector += partial;
	else {
	  sector = 1;
	  if (++head >= dp->nbhead) {
	      head = 0;
	      cylinder++;
	      }
	  }
      }
  return 0;
  }
#endif /* NO_FLOPPY_SUPPORT */

#endif /* BIOS_SUPPORT */

/**
 ** The low level extended BIOS functions:
 **/
#if DISK_SUPPORT & EBIOS_SUPPORT

SEGMENT_EXTRA(EBIOSDISK_getparam) static inline unsigned char
EBIOSDISK_getparam (unsigned char disk, struct diskparam_str *dp)
  {
  ebiosinfo_type ebiosinfo;
  ebiosPhoenix_type ebiosPhoenix;
  unsigned char status;

  ebiosinfo.buffersize = sizeof (ebiosinfo);

  /*
   * Note RedHat 5.2, DOSEMU, extended get geometry:
   * real extended BIOS returns C/H/S not translated for BIOS, i.e. head=16
   * DOSEMU extended BIOS returns C/H/S with head = 255
   * This geometry is anyway not used in this bootloader.
   */

  if (_EBIOSDISK_getparam (disk, &ebiosinfo, &status) || status != 0) {
      DDBG (("EBIOS getparam error.\r\n"));
      status = LOCALERROR;
      }
    else {
      DDBG (("EBIOS reports: nbhead: %u, nbsectorpertrack: %u"
	     ", nbcylinder+1: %u, bytepersector: %u, nbtotalsector: %u\r\n",
		ebiosinfo.nbhead,
		ebiosinfo.nbsectorpertrack,
		ebiosinfo.nbcylinder+1,
		ebiosinfo.bytepersector,
		(unsigned)ebiosinfo.nbtotalsector ));

      if (ebiosinfo.signature0xBEDD == 0xBEDD) {
	  DDBG (("V3.0 info: %s disk on %s bus, ", /* a lot more info */
		ebiosinfo.interface, ebiosinfo.bustype));
	  }
      DDBG (("EBIOS extension data at 0x%X", ebiosinfo.configptr));
      if (ebiosinfo.configptr && ebiosinfo.configptr != 0xFFFFFFFFU) {
	  int i;
	  unsigned char checksum = 0, *ptr = (unsigned char *)&ebiosPhoenix;

	  for (i = 0; i <= 14; i++)
	      checksum -= *ptr++ = peekb(ebiosinfo.configptr + i);
	  *ptr++ = peekb(ebiosinfo.configptr + 15);
	  *ptr++ = peekb(ebiosinfo.configptr + 16);

	  if (checksum == ebiosPhoenix.checksum) {
	      DDBG ((" Phoenix Bios compatible, ideIOadr 0x%X, multiple %u"
			", irq %u, lba_slave_mask= 0x%X\r\n",
			ebiosPhoenix.dataIOadr,
			ebiosPhoenix.multisectorsize,
			ebiosPhoenix.irq,
			ebiosPhoenix.lba_slave_mask ));

	      dp->ideIOadr = ebiosPhoenix.dataIOadr;
	      dp->multiplemode = ebiosPhoenix.multisectorsize;
	      dp->irq = ebiosPhoenix.irq;
	      if ((ebiosPhoenix.lba_slave_mask & 0x10) == 0)
		  dp->ismaster = 1;
		else
		  dp->ismaster = 0;
	      }
	    else {
	      DDBG ((" Non Phoenix Bios compatible.\r\n"));
	      dp->ideIOadr = 0;
	      dp->multiplemode = 0;
	      dp->irq = 0;
	      dp->ismaster = 0;
	      }
	  }

      dp->nbhead = ebiosinfo.nbhead;
      dp->nbsectorpertrack = ebiosinfo.nbsectorpertrack;
      dp->nbcylinder = ebiosinfo.nbcylinder + 1;
      dp->bytepersector = ebiosinfo.bytepersector;
      dp->nbtotalsector = ebiosinfo.nbtotalsector;
      }
  return status;
  }

SEGMENT_EXTRA(EBIOSDISK_probe) static inline int
EBIOSDISK_probe (unsigned char disk, struct diskparam_str *dp)
  {
  unsigned char version, extension;
  ebios_bitmap_type bitmap;

  dp->access = bios_chs;

  if (_EBIOSDISK_probe (disk, &version, &extension, &bitmap) == 0) {
      DDBG (("EBIOS detected: version 0x%X, extension 0x%X,"
		" features: %s%s%s\r\n",
		version,
		extension,
		bitmap.extended_fct? "extended_fct " : "",
		bitmap.removable? "removable " : "",
		bitmap.enhanced? "enhanced " : ""
		));

      dp->ebiosversion = version;
      if (bitmap.extended_fct)
	  dp->access = ebios_lba;
      return 0;
      }
  return 1;
  }

SEGMENT_EXTRA(ebiosRWsector) static inline unsigned char
ebiosRWsector ( unsigned char drive,
		lba_t lba,
		unsigned char number,
		farptr buffer,
		const unsigned read)
  {
  union diskcmd str = { int1342: {
      reg_dx:	drive,			/* reg_dl = drive number */
      reg_ax:	read? 0x4200 : 0x4300,	/* reg_al = 0 */
      cst_16:	16,			/* 0x10 */
      cst_0:	0,
      nb_block:	number,			/* max 127 */
      cst2_0:	0,
      /* segment/offset overwritten by boot1 */
      lba:	lba
      }};

  DDBG ((" <EBIOS %s drive 0x%X, %d sector(s) at LBA: %d> ",
	 read? "reading" : "writing", drive, number, (long)lba ));

  return read? BOOT1_diskread (&str, buffer) : BOOT1_diskwrite (&str, buffer);
  }

#endif /* EBIOS_SUPPORT */

/**
 ** The low level IDE functions:
 **/
#if DISK_SUPPORT & IDE_SUPPORT

SEGMENT_EXTRA(IDE_calibratedrive) static inline unsigned char
IDE_calibratedrive (unsigned char drive, unsigned short ideIOadr)
  {
  unsigned short timeout;
  unsigned char status;

  DDBG ((" [IDE calibrate %s @ 0x%X", drive? "slave" : "master", ideIOadr));

  status = _IDE_calibratedrive (drive, ideIOadr, &timeout);

  DDBG ((" status= 0x%X, %s %s] ",
	 status, (timeout == 0)? "timeout" : "", (status & 1)? "error" : "OK"));

  return status;
  }

SEGMENT_EXTRA(IDE_Initialisedrive) static inline unsigned char
IDE_Initialisedrive (unsigned char drive, unsigned short ideIOadr,
		     unsigned short nbcylinder,
		     unsigned char nbhead,
		     unsigned short nbsectorpertrack)
  {
  unsigned short timeout;
  unsigned char status;

  DDBG ((" [IDE Initialise %s @ 0x%X, cylinder = %d, nbhead = %d-1,"
	 " sectorpertrack = %d",
	 drive? "slave" : "master", ideIOadr, nbcylinder, nbhead & 0x0F,
	 nbsectorpertrack));

  status = _IDE_Initialisedrive (drive, ideIOadr, nbcylinder, nbhead,
				 nbsectorpertrack, &timeout);

  DDBG (("%s %s] ",
	 (timeout == 0)? "timeout" : "", (status & 1)? "error" : "OK"));

  return status;
  }

SEGMENT_EXTRA(ideIdentify_readbuff) static inline unsigned char
ideIdentify_readbuff (unsigned char  drive,
		      unsigned short ideIOadr,
		      unsigned char  readbuf, /* boolean */
		      farptr         buffer)
  {
  union diskcmd str = { hardide: {
      ide_dcr_address:	_IDE_guess_altadr (ideIOadr),
      reg_ax:		COMMANDSETUP,
      base_ide_plus7:	ideIOadr + 7, /* for instance 0x1F0 + 7 */
      lba_head:		0xA0 | ((drive)? 0x10 : 0),
      cylinder_high:	0,
      cylinder_low:	0,
      sector:		0,
      nb_sect:		0,
      ide_command:	(readbuf == 0)? 0xEC /* ID drv */: 0xE4 /* readbuf */,
      nb_sect_read:	1,
      sector_size_div2:	512/2
      }};

  DDBG ((" <IDE HARD %s %s drive @ 0x%X to address 0x%X> ",
	 (readbuf == 0)? "identify" : "read buffer",
	 (drive != 0)? "slave" : "master",
	 ideIOadr,
	 buffer
	 ));

  return BOOT1_diskread (&str, buffer);
  }

SEGMENT_EXTRA(ide_abscent) static inline unsigned
ide_abscent (unsigned short IOadr)
  {
  unsigned char tmp;

#if 0
  /* read 0 on some oldish IDE disks (220 Mbytes), if not BIOS initialised: */
  if (((tmp = _IDE_get_drivehead (IOadr)) & 0xA0) != 0xA0) {
      /* two bit set permanently */
      DDBG (("no IDE present, drive/head reg = 0x%X", tmp));
      return 1;
      }
#endif

  /* _IDE_get_alternate_status() do not clear a possible interrupt pending
     unlike _IDE_get_status(), but they should be the same... */
  if (_IDE_get_alternate_status (IOadr) != _IDE_get_status (IOadr)) {
      DDBG (("no IDE present, two status not equal (firsts IDE)"));
      return 1;
      }

  /* We test here the BUSY bit: */
  {
  unsigned timeout = 0x10000;

  while (  ((tmp = _IDE_get_alternate_status (IOadr)) & 0x80)
	 && --timeout) ;
  if (timeout == 0) {
      DDBG (("no IDE present, status register = 0x%X", tmp));
      return 1;
      }
  }

#if 0
  /* New UDMA disks/IDE do no more map this register in I/O space: */
  if (((tmp = _IDE_get_card_addr_reg (IOadr)) & 0x03) == 3) {
      /* at least 1 drive selected */
      DDBG (("no IDE present, Card Address Register = 0x%X", tmp));
      return 1;
      }
#endif

  DDBG (("seems present."));
  return 0;
  }

SEGMENT_EXTRA(idelbaRWsector) static inline unsigned char
idelbaRWsector (unsigned char	drive,
		unsigned short	ideIOadr,
		unsigned short	sectorsize,
		unsigned	lba,
		unsigned char	number,
		farptr		buffer,
		const unsigned	read)
  {
  union diskcmd str = { hardide: {
      ide_dcr_address:	_IDE_guess_altadr (ideIOadr),
      reg_ax:		COMMANDSETUP,
      base_ide_plus7:	ideIOadr + 7,	/* for instance 0x1F0 + 7 */
      lba_head:		0xE0 | ((drive)? 0x10 : 0) | ((lba >> 24) & 0x0F),
      cylinder_high:	(lba >> 16) & 0xFF,
      cylinder_low:	(lba >> 8) & 0xFF,
      sector:		lba & 0xFF,
      nb_sect:		number,
      ide_command:	read? 0x20 : 0x30,	/* with retry */
      nb_sect_read:	(number == 0)? 256 : number,
      sector_size_div2:	sectorsize/2
      }};

  DDBG ((" <IDE HARD %s %s drive @ 0x%X, %d sector(s) [%d] at LBA: %d> ",
	 read? "reading" : "writing",
	 (drive != 0)? "slave" : "master",
	 ideIOadr,
	 (number == 0)? 256 : number,
	 sectorsize,
	 lba & 0xFFFFFFF
	 ));

#ifndef DISPLAY_DISK_ERROR
  return read? BOOT1_diskread (&str, buffer) : BOOT1_diskwrite (&str, buffer);
#else
  {
  unsigned char returned;

  returned = read? BOOT1_diskread (&str, buffer)
		 : BOOT1_diskwrite (&str, buffer);
  if (returned != 0) {
      printf ("LBA(%u@0x%X) r/w error 0x%X, status 0x%X, ",
		number, lba, returned, _IDE_get_status (ideIOadr));
      printf ("lsb 0x%X, ", inb (ideIOadr+3));
      printf ("middle 0x%X, ", inw (ideIOadr+4));
      printf ("msb 0x%X.\r\n", inb (ideIOadr+6));
      }
  return returned;
  }
#endif
  }


SEGMENT_EXTRA(idechsRWsector) static inline unsigned char
idechsRWsector (unsigned char	drive,
		unsigned short	ideIOadr,
		unsigned short	sectorsize,
		unsigned short	cylinder,
		unsigned char	head,
		unsigned char	sector,
		unsigned char	number,
		farptr		buffer,
		const unsigned	read)
  {
  union diskcmd str = { hardide: {
      ide_dcr_address:	_IDE_guess_altadr (ideIOadr),
      reg_ax:		COMMANDSETUP,
      base_ide_plus7:	ideIOadr + 7,	/* for instance 0x1F0 + 7 */
      lba_head:		0xA0 | ((drive)? 0x10 : 0) | (head & 0x0F),
      cylinder_high:	(cylinder >> 8) & 0xFF,
      cylinder_low:	cylinder & 0xFF,
      sector:		sector,
      nb_sect:		number,
      ide_command:	read? 0x20 : 0x30,	/* with retry */
      nb_sect_read:	(number == 0)? 256 : number,
      sector_size_div2:	sectorsize/2
      }};

  DDBG ((" <IDE HARD %s %s drive @ 0x%X, %d sector(s) [%d] at C/H/S:> "
	 "%d/%d/%d ",
	 read? "reading" : "writing",
	 (drive != 0)? "slave" : "master",
	 ideIOadr,
	 (number == 0)? 256 : number,
	 sectorsize,
	 cylinder,
	 head & 0x0F,
	 sector
	 ));

#ifndef DISPLAY_DISK_ERROR
  return read? BOOT1_diskread (&str, buffer) : BOOT1_diskwrite (&str, buffer);
#else
  {
  unsigned char returned;

  returned = read? BOOT1_diskread (&str, buffer)
		 : BOOT1_diskwrite (&str, buffer);
  if (returned != 0) {
      printf ("CHS(%u@0x%x/0x%x/0x%x) r/w error 0x%X, status 0x%X, ",
		number, cylinder, head, sector,
		returned, _IDE_get_status (ideIOadr));
      printf ("lsb 0x%X, ", inb (ideIOadr+3));
      printf ("middle 0x%X, ", inw (ideIOadr+4));
      printf ("msb 0x%X.\r\n", inb (ideIOadr+6));
      }
  return returned;
  }
#endif
  }

#endif /* IDE_SUPPORT */

/**
 ** The generic function(s) to read or write disks:
 **/
SEGMENT_EXTRA(DISK_RWsector) static
#ifndef WRITE_ENABLE
inline
#endif
unsigned char
DISK_RWsector ( struct diskparam_str *dp, lba_t lba,
		unsigned char number, farptr buffer,
		const unsigned read)
  {
  unsigned short cylinder;
  unsigned char sector, head;

  if (lba > dp->nbtotalsector) {
	/* long long printf needed if lba_t */
      DDBG ((" %s 0x%X sector %u: too high LBA (max %u).\r\n", __FUNCTION__,
		dp->disknb, (unsigned)lba, (unsigned)dp->nbtotalsector));
      return 1;
      };

#if DISK_SUPPORT & EBIOS_SUPPORT
  if (dp->access == ebios_lba)
      return ebiosRWsector (dp->disknb, lba, number, buffer, read);
#endif /* EBIOS_SUPPORT */

#if DISK_SUPPORT & IDE_SUPPORT
  if (dp->access == hardide_lba)
      return idelbaRWsector (dp->disknb, dp->ideIOadr, dp->bytepersector,
				lba, number, buffer, read);
#endif /* IDE_SUPPORT */

  if (sizeof (lba_t) < 8) {
      sector = lba % dp->nbsectorpertrack + 1;
      lba = lba / dp->nbsectorpertrack;
      head = lba % dp->nbhead;
      cylinder = lba / dp->nbhead;
      }
    else {
      unsigned tmp, tmp1;
      tmp = ull_div_ul (lba, dp->nbsectorpertrack, &tmp1);
      sector = tmp1 + 1;
      head = tmp % dp->nbhead;
      cylinder = tmp / dp->nbhead;
      }

#if DISK_SUPPORT & IDE_SUPPORT
  if (dp->access == hardide_chs)
      return idechsRWsector (dp->disknb, dp->ideIOadr, dp->bytepersector,
				cylinder, head, sector, number, buffer, read);
#endif /* IDE_SUPPORT */

#if DISK_SUPPORT & BIOS_SUPPORT
#if !(DISK_SUPPORT & NO_FLOPPY_SUPPORT)
  if (dp->disknb < 0x80) {
      unsigned char tmp;
      DDBG ((" [BIOS FD read: "));

      if (cylinder >= 83) {
	  /* Floppy driver can physically break here... */
	  DBG (("Aborting read/write of floppy disk with "
		"too high cylinder !\r\n"));
	  return 1;
	  }

      tmp = biosFD_RWsector (dp, dp->disknb, cylinder, head, sector, number,
				buffer, read);
      if (tmp != 0)
	  DDBG (("error 0x%X] ", tmp));
	else
	  DDBG (("OK] "));
      return tmp;
      }
    else
#endif /* NO_FLOPPY_SUPPORT */
      {
      unsigned char multiplemode = !!dp->multiplemode;
      unsigned char tmp;

      if ((cylinder & 0xC00) != 0 && dp->nbhead <= 16) {
	  /* old bios with 4096 sectors, elsewhere compatible */
	  head |= ((cylinder & 0x0C00) >> 4);
	  cylinder &= 0x3FF;
	  }

      DDBG ((" [BIOS HD %s read: ", multiplemode? "multiple" : "single"));
      tmp = biosRWsector (dp->disknb, cylinder, head, sector, number,
				multiplemode, buffer, read);

      if (tmp == 0x80 || tmp == 0xAA)
	  DDBG (("%s permanent failure] ", __FUNCTION__));
	else if (tmp != 0)
	  DDBG (("error] "));
	else
	  DDBG (("OK] "));
      return tmp;
      }
#else /* BIOS_SUPPORT */
  DDBG (("no BIOS disk compiled in] "));
  return LOCALERROR;
#endif /* BIOS_SUPPORT */
  }

GENERATE_CODE2EXTRA_STUB(DISK_readsector, unsigned char,
	struct diskparam_str *dp, lba_t lba,
	unsigned char number, farptr buffer)

SEGMENT_EXTRA(DISK_readsector) unsigned char
DISK_readsector (struct diskparam_str *dp, lba_t lba,
		 unsigned char number, farptr buffer)
  {bound_stack();{
  return DISK_RWsector (dp, lba, number, buffer, 1);
  }}


#ifdef WRITE_ENABLE
GENERATE_CODE2EXTRA_STUB(DISK_writesector, unsigned char,
	struct diskparam_str *dp, lba_t lba,
	unsigned char number, farptr buffer)

SEGMENT_EXTRA(DISK_writesector) unsigned char
DISK_writesector (struct diskparam_str *dp, lba_t lba,
		  unsigned char number, farptr buffer)
  {bound_stack();{
  return DISK_RWsector (dp, lba, number, buffer, 0);
  }}
#endif

/*
 * If needed:
 */
GENERATE_VOID_CODE2EXTRA_STUB(diskname,
	struct diskparam_str *dp, char name[SIZE_DISKNAME])

SEGMENT_EXTRA(diskname) void
diskname (struct diskparam_str *dp, char name[SIZE_DISKNAME])
  {
#ifndef DISK_DISPLAY_FULLNAME
  unsigned char biosnb = dp->disknb;

  /* this will not need sprintf() */
  name[0] = '0';
  name[1] = 'x';
  name[2] = '0' + biosnb / 16;
  name[3] = '0' + biosnb % 16;
  name[4] = '\0';
#else
  const char *msg = 0;

  checkptr (dp, "diskname invalid parameter 1.");
  checkptr (name, "diskname invalid parameter 2.");

  if (dp->access == bios_chs)
      msg = BIOS_DISK_X;
    else if (dp->access == ebios_lba)
      msg = EBIOS_DISK_X;
    else if (dp->access == dos_part)
      msg = DOS_DISK_C;
  if (msg) {
      sprintf (name, msg, dp->disknb);
      return;
      }
  if (dp->access == hardide_chs)
      msg = IDECHS_DISK_X_s;
    else if (dp->access == hardide_lba)
      msg = IDELBA_DISK_X_s;
  if (msg) {
      sprintf (name, msg, dp->ismaster? IDE_MASTER : IDE_SLAVE, dp->ideIOadr);
      return;
      }
  STRINIT (name, DISKNAME_ERROR);
#endif
  }

/**
 ** The bit which understand partitions:
 **/
#if !(DISK_SUPPORT & NO_PROBE_PARTITIONS)
/* recursive pre-declaration: */
static unsigned
analysepartition (struct diskparam_str *dp,
		  lba_t lba,
		  struct partition_str *partition,
		  unsigned *nbpartition);

SEGMENT_EXTRA(analyse_extended_partition) static inline void
analyse_extended_partition (struct diskparam_str *dp,
			    struct partition_str *partition,
			    unsigned *nbpartition)
  {
  unsigned short extended_partition_index = *nbpartition - 1;
  lba_t trylba = partition[extended_partition_index].start;

  int retval = analysepartition (dp, trylba, partition, nbpartition);
  if (retval != 0 && retval != 3) {
		/* long long printf needed if lba_t ... */
      printf (WARNING_EXTENDED_PARTITION_BLOCK_NOT_FOUND_AT_U,
		(unsigned)trylba);
      if (*nbpartition < 2) {
	  trylba = RESERVEDSECTOR;
	  }
	else {
	  trylba = partition[*nbpartition - 2].start;
	  trylba += partition[*nbpartition - 2].length;
	  }
	/* EBIOS sometimes returns nbsectorpertrack = 0 ! */
      trylba += (dp->nbsectorpertrack != 0)? dp->nbsectorpertrack - 1 : 62;

      if (sizeof (lba_t) < 8) {
	  trylba = (trylba / dp->nbsectorpertrack) * dp->nbsectorpertrack;
	  }
	else {
	  unsigned tmp;

	  ull_div_ul (trylba, dp->nbsectorpertrack, &tmp);
	  trylba -= tmp;
	  }

      if (trylba != partition[extended_partition_index].start) {
		/* long long printf needed if lba_t ... */
	  printf (MSG_TRYING_COLON "%u", (unsigned)trylba);
	  }
	else {
	  puts (MSG_FULL_STOP);
	  return;
	  }

      if (analysepartition (dp, trylba, partition, nbpartition) == 0) {
	  puts (MSG_COLON_OK);
	  partition[extended_partition_index].start = trylba;
	  }
	else {
	  puts (MSG_COLON_FAILED);
	  return;
	  }
      }

   {
   lba_t extended_end = partition[extended_partition_index].start
			+ partition[extended_partition_index].length;
   /* take care: analysepartition() has changed *nbpartition: */
   lba_t lastlba = partition[*nbpartition - 1].start
			+ partition[*nbpartition - 1].length;

   if (extended_end != lastlba)
      DDBG (("\r\nInvalid length of extended partition, "
		"it is %u and should be %u, delta=%d.",
		partition[extended_partition_index].length,
		lastlba	- partition[extended_partition_index].start,
		extended_end - lastlba
		));
   }
   return;
   }

SEGMENT_EXTRA(dbg_check_chs2lba) static inline void
dbg_check_chs2lba (const struct diskparam_str *dp,
		   lba_t lba, unsigned partno,
		   const struct bootsect_partition_str *buffpart)
  {
  unsigned checklba;
  unsigned cylinder;

  cylinder = ((unsigned)buffpart->begin_cylinder_msb << 8)
		 | buffpart->begin_cylinder_lsb;
  checklba = (cylinder * dp->nbhead + buffpart->begin_head);
  checklba *= dp->nbsectorpertrack;
  checklba += buffpart->begin_sector - 1;

  if (   checklba != buffpart->nb_sector_before
      && buffpart->nb_sector_before < 1024L * 256 * 63) {
      DDBG (("\r\nInvalid start C/H/S [H0/S0]<->base+LBAstart "
		"in disk 0x%X, partition %u at LBA=%u:\r\n",
		dp->disknb, partno, lba));
      DDBG (("  %u/%u/%u [%u/%u] = %u <-> %u + %u = %u",
		cylinder,
		buffpart->begin_head,
		buffpart->begin_sector,
		dp->nbhead,
		dp->nbsectorpertrack,
		checklba,
		(unsigned)lba,
		buffpart->nb_sector_before,
		buffpart->nb_sector_before + lba ));
      }

  cylinder = ((unsigned)buffpart->end_cylinder_msb << 8)
		| buffpart->end_cylinder_lsb;
  checklba = (cylinder * dp->nbhead + buffpart->end_head);
  checklba *= dp->nbsectorpertrack;
  checklba += buffpart->end_sector - 1;

  if (   checklba != buffpart->nb_sector_before + lba
			+ buffpart->nb_sector - 1
      && buffpart->nb_sector_before + lba
			+ buffpart->nb_sector - 1 < 1024L * 256 * 63) {
      DDBG (("\r\nInvalid end C/H/S [H0/S0]<->base+LBAstart+length-1"
		" in disk 0x%X, partition %u at LBA=%u:\r\n",
		dp->disknb, partno, lba));
      DDBG (("  %u/%u/%u [%u/%u] = %u <-> %u + %u + %u - 1 = %u",
		cylinder,
		buffpart->end_head,
		buffpart->end_sector,
		dp->nbhead,
		dp->nbsectorpertrack,
		checklba,
		(unsigned)lba,
		buffpart->nb_sector_before,
		buffpart->nb_sector,
		lba + buffpart->nb_sector_before + buffpart->nb_sector - 1
		));
      }
  }

SEGMENT_EXTRA(analysepartition) static unsigned
analysepartition (struct diskparam_str *dp, lba_t lba,
		  struct partition_str *partition, unsigned *nbpartition)
  {bound_stack();{
  unsigned short i;
  bootsector_type buffer = {};

  if (DISK_readsector (dp, lba, 1, stack_adr(&buffer)) != 0) {
      DDBG (("READ FAILED! "));
#if DISK_SUPPORT & IDE_SUPPORT
      if (dp->access == hardide_lba || dp->access == hardide_chs)
	  DDBG ((IDE_requestsense (dp->disknb, dp->ideIOadr)));
	else
#endif
	  DDBG ((BIOSDISK_error (BIOSDISK_getstatus (dp->disknb)) ));
      DDBG (("\r\n"));

      return 1;
      }

  DDBG (("signature: 0x%X", buffer.after.Signature0xAA55));
  if (buffer.after.Signature0xAA55 != 0xAA55) {
      int error = 2;

      if (   buffer.after.Signature0xAA55 == 0xF6F6
	  && *(unsigned *)&buffer == lba) {
	  /*
	   * Happen with (at least) Seagate partitionning software, the
	   * extended sector are completely filled with 0xF6F6 but the first
	   * 4 bytes which are the LBA of the current sector. Fist, you
	   * cannot encode that it is not an MSDOS partition (i.e.
	   * bootafter.bootsect_partition.system_indicator), but
	   * you cannot know what is the size of the next partition
	   * without analysing its content - i.e. understanding the
	   * filesystem - and so you cannot look for the 3rd partition
	   * of the disk. I think I got the right name for this partition
	   * sheme... but it is also hostile to *BSD and everything else...
	   */
	  char disk_name[SIZE_DISKNAME];

	  diskname (dp, disk_name);
	  printf (WARNING_LINUX_HOSTILE_PARTITION_SCHEME_ON_S, disk_name);
	  error++;
	  }

      return error;
      }

  for (i = 0;
	  i < nbof(buffer.after.bootsect_partition)
       && *nbpartition < NB_PARTITION;
       i++) {
      struct bootsect_partition_str *buffpart
				 = &buffer.after.bootsect_partition[i];
      struct partition_str *part = &partition[*nbpartition];

      switch (buffpart->indicator) {
	case bootable:
	   part->active = part_active;
	   break;
	 case non_bootable:
	   part->active = part_inactive;
	   break;
	 default:
	   DDBG (("\r\n Invalid partition marker: neither active nor "
		 "inactive: 0x%X", buffpart->indicator));
	   part->active = part_invalid;
	   continue;
	 }

      if (buffpart->begin_sector != 0 && buffpart->nb_sector_before != 0)
	  dbg_check_chs2lba (dp, lba, i, buffpart);

      if (lba == 0 || buffpart->system_indicator != EMPTY) {
	  unsigned char is_extended;

	  part->start = buffpart->nb_sector_before + lba;
	  part->length = buffpart->nb_sector;
	  part->type = buffpart->system_indicator;
	  part->misc.order = *nbpartition;
#if defined (MAX_DISKNAMESIZE) && (MAX_DISKNAMESIZE != 0)
	  part->name[0] = '\0';
#endif
	  *nbpartition += 1;
	  if (   buffpart->system_indicator == extended
	      || buffpart->system_indicator == Win95_extendedLBA
	      || buffpart->system_indicator == Linux_extended)
	      is_extended = 1;
	    else
	      is_extended = 0;
	  if (lba == 0)
	      part->OSnumber = i + 1;
	    else if (!is_extended)
	      part->OSnumber = dp->nbOSnumber++;
	    else
	      part->OSnumber = 0;
#if NB_PARTITION > 4
	  if (is_extended)
	      analyse_extended_partition (dp, partition, nbpartition);
#endif
	  }
      }
  return 0;
  }}
#endif /* NO_PROBE_PARTITIONS */

#if DISK_SUPPORT & (BIOS_SUPPORT | EBIOS_SUPPORT | IDE_SUPPORT)

#if !(   (DISK_SUPPORT & (BIOS_SUPPORT | EBIOS_SUPPORT)) \
      && (DISK_SUPPORT & IDE_SUPPORT))
#define INLINE_analysebootsector
#endif

SEGMENT_EXTRA(analysebootsector) static
#ifdef INLINE_analysebootsector
inline
#endif
int analysebootsector (struct diskparam_str *dp)
#ifdef INLINE_analysebootsector
  {{
#else
  {bound_stack();{
#endif
  unsigned nbpartition = 0;
#ifdef DISK_USE_MALLOC
  struct partition_str partition[NB_PARTITION];
#else
  struct partition_str *partition = dp->partition;
#endif

  checkptr (dp, "analysebootsector invalid parameter.");

  {
  bootsector_type buffer;
  if (DISK_readsector (dp, 0, 1, stack_adr(&buffer)) != 0) {
      DDBG (("READ FAILED! "));
#if DISK_SUPPORT & IDE_SUPPORT
      if (dp->access == hardide_lba || dp->access == hardide_chs)
	  DDBG ((IDE_requestsense (dp->disknb, dp->ideIOadr)));
	else
#endif
	  DDBG ((BIOSDISK_error (BIOSDISK_getstatus (dp->disknb)) ));
      DDBG (("\r\n"));

      return 1;
      }

  DDBG (("signature: 0x%X", buffer.after.Signature0xAA55));

  if (buffer.after.Signature0xAA55 != 0xAA55)
      dp->bootable = 0;
    else
      dp->bootable = 1;

  /* To get floppy with non conventionnal size working, or even CDROM? : */
  if (buffer.before.Signaturebyte0x29 == 0x29) {
      nbsector_t NbTotalsector = (buffer.before.NbTotalsector)
					? buffer.before.NbTotalsector
					: buffer.before.NbTotalsector2;
      if (   dp->bytepersector != buffer.before.Bytepersector
	  || dp->nbsectorpertrack != buffer.before.NbSectorpertrack
	  || dp->nbhead != buffer.before.NbHead
	  || dp->nbtotalsector != NbTotalsector) {
	  char disk_name[SIZE_DISKNAME];

	  diskname (dp, disk_name);
	  printf (INFO_ADJUSTING_PARAMETERS_FROM_BOOTSECTOR_S, disk_name);
	  dp->bytepersector = buffer.before.Bytepersector;
	  dp->nbsectorpertrack = buffer.before.NbSectorpertrack;
	  dp->nbhead = buffer.before.NbHead;
	  dp->nbtotalsector = NbTotalsector;
	  dp->nbcylinder = NbTotalsector / dp->nbsectorpertrack / dp->nbhead;
	  dp->OSdisknumber = buffer.before.PhysicaldriveNb;
	  }
      }
  }

#if !(DISK_SUPPORT & NO_PROBE_PARTITIONS)
  dp->nbOSnumber = 5;
  /*
   * Here we will re-read sector at LBA = 0, but maybe
   * buffer.before.Bytepersector is higher than 512 and
   * the space added may be defined to contains new partitions
   * definitions:
   */
  analysepartition (dp, 0, partition, &nbpartition);

  dp->nbOSnumber -= 5;
#endif

  /*
   * Check if we have a floppy without a partition scheme, whe
   * then add a dummy partition to probe the files.
   */
  {
  unsigned short cpt;

  for (cpt = 0; cpt < nbpartition; cpt++)
      if (partition[cpt].type != EMPTY)
	  break;
  if (cpt >= nbpartition) {
      static const struct partition_str floppy_partition = {
	  start:	0,
	  length:	2880,
	  type:		DOS_FAT12,
	  OSnumber:	1,
	  misc:		{ order: 0, maybe_root: 1 },
#if defined (MAX_DISKNAMESIZE) && (MAX_DISKNAMESIZE != 0)
	  name:		"floppy",
#endif
	  active:	part_active
	  };
      *partition = floppy_partition;
      partition->length = dp->nbtotalsector;
      if (dp->nbtotalsector > 8 * 1024 * 1024 / 512) /* more than 8 Mb */
	  partition->type = DOS_FAT16;	/* Just in case someone uses that */
      nbpartition = 1;
      }
  }

#ifdef DISK_USE_MALLOC
  if (nbpartition != 0) {
      dp->partition = malloc (nbpartition * sizeof (struct partition_str));
      if (dp->partition == 0) {
	  dp->nbpartition = 0;
	  DDBG (("%s: malloc failed!\r\n", __FUNCTION__));
	  }
	else {
	  dp->nbpartition = nbpartition;
	  while (nbpartition --)
	      dp->partition[nbpartition] = partition[nbpartition];
	  }
      }
    else {
      dp->nbpartition = 0;
      dp->partition = 0;
      }
#else
  dp->nbpartition = nbpartition;
#endif
  return 0;
  }}
#endif /* BIOS_SUPPORT | EBIOS_SUPPORT | IDE_SUPPORT */

/*
 * The medium level functions:
 */
#if DISK_SUPPORT & (BIOS_SUPPORT | EBIOS_SUPPORT)
SEGMENT_EXTRA(diskcheck) static inline int
diskcheck (unsigned char disk, struct diskparam_str *dp)
  {
  unsigned nbtotalsector = 0;

#if DISK_SUPPORT & BIOS_SUPPORT
  /* On some PC, BIOSDISK_getparam returns values (probably
     the CMOS type of slave IDE) even for a non existant drive.
     EBIOS can also return 0 head, 0 sector, 0 track ... */

  if (BIOSDISK_gettype(disk, &nbtotalsector) == 0) {
      DDBG ((" [abscent] "));
      DDBG ((BIOSDISK_error (BIOSDISK_getstatus (disk)) ));
      return 1;
      }
#endif /* BIOS_SUPPORT */

#if DISK_SUPPORT & EBIOS_SUPPORT
  if (EBIOSDISK_probe (disk, dp) == 0 && EBIOSDISK_getparam (disk, dp) == 0) {
      if (dp->ebiosversion == 0x16 && dp->nbcylinder == 16384) {
	  DDBG (("COMPAQ DESKPRO 590/5120 bug detected, "
		 "will crash when read with BIOS OR EBIOS.\r\n"));
	  printf (INFO_DISK_X_BIOS_EBIOS_INTERFACE_BUGGY_EXCLUDED, disk);
	  return 1;
	  }
      if (dp->access == ebios_lba) {
	  unsigned total = dp->nbhead * dp->nbsectorpertrack * dp->nbcylinder;
	  if (dp->nbtotalsector == 0) {
	      if (nbtotalsector != 0 && (nbtotalsector * 10 / total) > 9) {
		  dp->nbtotalsector = nbtotalsector;
		  DDBG ((" [10% OK, use BIOSDISK_gettype nbtotalsector] "));
		  }
		else {
		  dp->nbtotalsector = total;
		  DDBG ((" [10% failed, use C*H*S] "));
		  }
	      }
	    else
	      DDBG ((" [use EBIOS nbtotalsector] "));
	  return 0;
	  }
	else {
	  unsigned char status;

	  printf (INFO_DISK_X_EXTENDED_BIOS_WITHOUT_EXTENDED_FCT, disk);
	  if (   dp->multiplemode != 0
	      && _BIOSDISK_setmultiple (disk, dp->multiplemode, &status) == 0
	      && status == 0)
	      printf (USING_MULT_U_BIOS, dp->multiplemode);
	    else {
	      dp->multiplemode = 0;
	      printf (USING_BIOS);
	      }
	  }
      }
#endif /* EBIOS_SUPPORT */

#if DISK_SUPPORT & BIOS_SUPPORT
  if (BIOSDISK_getparam (disk, dp) == 0) {
      unsigned total = dp->nbhead * dp->nbsectorpertrack * dp->nbcylinder;

      if (total == 0) {
	  /* at least one field (nbhead, nbsectorpertrack,
	   * or nbcylinder) equals 0:
	   */
	  DDBG (("; Invalid drive (C/H/S)!"));
	  return 1;
	  }
      if (dp->nbtotalsector == 0) {	/* keep EBIOS value : needed ? */
	  if (nbtotalsector != 0 && (nbtotalsector * 10 / total) > 9)
	      dp->nbtotalsector = nbtotalsector;
	    else
	      dp->nbtotalsector = total;
	  }
      return 0;
      }
  DDBG ((" Invalid drive (getparam)."));
#else /* BIOS_SUPPORT */
  DDBG (("; BIOS disk interface not compiled in!"));
#endif /* BIOS_SUPPORT */

  return 1;
  }

SEGMENT_EXTRA(probeBIOSdisk) static inline unsigned
probeBIOSdisk (unsigned disk, struct diskparam_str *dp)
  {
  DDBG (("\r\n%s disk 0x%X: ", (disk >= 0x80)? "Hard" : "Floppy", disk));
  checkptr (dp, "probeBIOSdisk invalid parameter.");
//  memset (dp, 0, sizeof (struct diskparam_str));
  *dp = (struct diskparam_str) {};
  dp->disknb = disk;
  if (diskcheck (disk, dp) != 0)
      return 1;
//  dp->nbpartition = 0;	// done by memset()
//  dp->partition = 0;
  if (analysebootsector (dp) == 1)
      return 1;
  DDBG ((", %u partitions.", dp->nbpartition));
  return 0;
  }

SEGMENT_EXTRA(probe_nbBIOS) static inline unsigned
probe_nbBIOS (unsigned BIOSdisk)
  {
  unsigned char CmosDriveType, status;
  unsigned short tmpcx, tmpdx;
  farptr DBTadr;

  /*
   * Unsafe because we ask to the manager of the _first_ disk
   * how many disk there is in the system - think about
   * removable disks with added drivers:
   */
  if (_BIOSDISK_getparam (BIOSdisk, &tmpcx, &tmpdx, &DBTadr,
				&CmosDriveType, &status))
      return 0;
    else
      return tmpdx & 0xFF;
  }

#endif /* BIOS_SUPPORT | EBIOS_SUPPORT */

#if DISK_SUPPORT & IDE_SUPPORT
SEGMENT_EXTRA(probeIDEdisk) static unsigned
probeIDEdisk (unsigned short IOadr, unsigned char slave,
		struct diskparam_str *dp)
  {bound_stack();{
  ATAstruct_type buff;
  unsigned char cpt;

  DDBG (("\r\n0x%X", IOadr));
  checkptr (dp, "probeIDEdisk invalid parameter.");
//  memset (dp, 0, sizeof (struct diskparam_str));
  *dp = (struct diskparam_str) {};

  DDBG (("[%s] ", slave? "slave " : "master"));

  if (ideIdentify_readbuff (slave, IOadr, 0, stack_adr (&buff)) != 0) {
      DDBG ((" failed."));
      return 1;
      }

  DDBG ((" reports: C/H/S=%u/%u/%u, RWmult=%u,"
	 " trans CHS %svalid: C/H/S=%u/%u/%u, total= %u,"
	 " BigDisk total: %u ",
	 buff.number_of_cylinder,
	 buff.number_of_head,
	 buff.number_of_sector_per_track,
	 buff.RWmultiple_max_sector,
	 buff.description.translation_fields_valid? "" : "in",
	 buff.number_of_current_cylinder,
	 buff.number_of_current_head,
	 buff.number_of_current_sector_per_track,
	 buff.current_capacity_in_sector,
	 buff.lba_capacity
	 ));

  if (   ((unsigned *)&buff)[0] == ((unsigned *)&buff)[1]
      && ((unsigned *)&buff)[0] == ((unsigned *)&buff)[2]) {
      DDBG ((" [recognised the f***ing CDROM] "));
      return 1;
      }

  dp->multiplemode = buff.RWmultiple_max_sector;

  DDBG (("%s supported, DMA%s supported; ",
	 buff.capabilities.lba_supported? "LBA" : "CHS only",
	 buff.capabilities.dma_supported? "" : " not"
	 ));

  if (buff.capabilities.lba_supported)
      dp->access = hardide_lba;
    else
      dp->access = hardide_chs;
  dp->disknb = slave;
  dp->ismaster = !slave;
  dp->bytepersector = 512;
  if (   buff.description.translation_fields_valid
      && ABS (buff.number_of_current_cylinder
		 * buff.number_of_current_head
	      - buff.current_capacity_in_sector
		 / buff.number_of_current_sector_per_track) <= 2) {
      /* accept it if error < 2*nb_track */
      dp->nbhead = buff.number_of_current_head;
      dp->nbcylinder = buff.number_of_current_cylinder;
      dp->nbsectorpertrack = buff.number_of_current_sector_per_track;
      dp->nbtotalsector = buff.current_capacity_in_sector;
      }
    else {
      dp->nbhead = buff.number_of_head;
      dp->nbcylinder = buff.number_of_cylinder;
      dp->nbsectorpertrack = buff.number_of_sector_per_track;
      dp->nbtotalsector = dp->nbhead * dp->nbsectorpertrack * dp->nbcylinder;
      }
  dp->ideIOadr = IOadr;

  /*extract from drivers/ide/ide-disk.c:
   *
   * The ATA spec tells large drives to return
   * C/H/S = 16383/16/63 independent of their size.
   * Some drives can be jumpered to use 15 heads instead of 16.
   * Some drives can be jumpered to use 4092 cyls instead of 16383.
   */
  if (   (   buff.number_of_cylinder == 16383
	  || (   buff.number_of_cylinder == 4092
	      && buff.description.translation_fields_valid
	      && buff.number_of_current_cylinder == 16383))
      && buff.number_of_sector_per_track == 63
      && (buff.number_of_head == 15 || buff.number_of_head == 16)
      && buff.lba_capacity >= 16383U*63U*buff.number_of_head) {
      dp->nbtotalsector = buff.lba_capacity;
      dp->nbcylinder = buff.lba_capacity / dp->nbhead
					 / dp->nbsectorpertrack;
      DDBG ((" [adjusting BIGDISK size %u sector, %u cylinders] ",
		dp->nbtotalsector, dp->nbcylinder));
      }

//  dp->nbpartition = 0;	// done by memset()
//  dp->partition = 0;
  cpt = 2;
  while (analysebootsector (dp) == 1) {
      IDE_calibratedrive (dp->disknb, IOadr);
      if (--cpt == 0)
	  break;
      IDE_Initialisedrive (dp->disknb, IOadr,
				dp->nbcylinder, dp->nbhead,
				dp->nbsectorpertrack);
      DDBG ((" retrying analysebootsector() "));
      }
  DDBG ((", %u partitions.", dp->nbpartition));
  return 0;
  }}
#endif /* IDE_SUPPORT */

/**
 ** This sort the partition in the start increasing order,
 ** to have "empty" partitions (i.e. start == end == 0) at
 ** end of the array. It is needed for freelist calculation.
 **/
SEGMENT_EXTRA(sort_partition) static inline void
sort_partition (struct partition_str *part, unsigned nbpart)
  {bound_stack();{
  unsigned i, j, min;

  checkptr (part, "sort_partition invalid parameter.");

  for (i = 0; i < nbpart; i++) {
      min = i;
      for (j = i+1; j < nbpart; j++) {
	  if (   part[j].start == 0
	      || part[j].length == 0
	      || part[j].active == part_invalid)
	      continue;
	  if (part[j].start < part[min].start)
	      min = j;
	  }
      if (min != i) {
	  struct partition_str tmp = part[i];
	  part[i] = part[min];
	  part[min] = tmp;
	  }
      }
  }}

/**
 ** Just to calculate the list of sector in any partition, so
 ** we may use them one day to install Gujin on a HD.
 ** It also checks if there is overlap of partition, or a partition
 ** over the limit of the disk.
 **/
#if defined (NB_FREELIST) && NB_FREELIST != 0
SEGMENT_EXTRA(calculatefreelist) static void
calculatefreelist (struct diskparam_str *dps)
  {bound_stack();{
  unsigned short	cpt_part;
  lba_t			curlba;
  unsigned		nbfreelist = 0;
#ifdef DISK_USE_MALLOC
  struct freelist_str	freelist[NB_FREELIST];
#else
  struct freelist_str	*freelist = dps->freelist;
#endif
  char disk_name[SIZE_DISKNAME];

  diskname (dps, disk_name);

  checkptr (dps, "calculatefreelist invalid parameter.");

  DDBG (("["));
  curlba = RESERVEDSECTOR; /* first sector of disk that can be in free list */
  for (cpt_part = 0;
       cpt_part < dps->nbpartition && nbfreelist < NB_FREELIST;
       cpt_part++) {
      struct partition_str *dpsp = &dps->partition[cpt_part];

      DDBG (("."));
      if (dpsp->type == EMPTY || dpsp->length == 0)
	  continue;
      if (dpsp->start == 0 && curlba == RESERVEDSECTOR) {
	  /* the dummy floppy partition, i.e. MBR in 1st partition: */
	  curlba = dpsp->length;
	  continue;
	  }
      if (dpsp->start != curlba) {
	  if (dpsp->start < curlba) {
	      DDBG ((" [ERROR %s: partition overlap!] ", __FUNCTION__));
	      printf (S_PARTITION_STARTING_AT_U_OVERLAP_DETECTED,
			disk_name, dpsp->start);
	      }
	    else {
	      freelist[nbfreelist].start = curlba;
	      freelist[nbfreelist].length = dpsp->start - curlba;
	      nbfreelist++;
	      }
	  }
      if (   dpsp->type == extended
	  || dpsp->type == Linux_extended
	  || dpsp->type == Win95_extendedLBA)
	  curlba = dpsp->start + RESERVEDSECTOR;
	else
	  curlba = dpsp->start + dpsp->length;
      }

  DDBG (("*"));

#if DISK_SUPPORT & DOS_SUPPORT
  if (dps->access != dos_part)
#endif
      {
      if (curlba > dps->nbtotalsector) {
	  DDBG ((" [%s: partition extend over disk limit] ", __FUNCTION__));
	  printf (S_S_LAST_PARTITION_OVER_DISK_LIMIT_DETECTED_U_KB,
		  (dps->access == bios_chs || dps->access == ebios_lba)
			? MSG_WARNING
			: MSG_ERROR,
		  disk_name,
		  (curlba - dps->nbtotalsector) / 2);
	  }
	else if (curlba < dps->nbtotalsector && nbfreelist < NB_FREELIST) {
	  /* space left at end of disk: */
	  freelist[nbfreelist].start = curlba;
	  freelist[nbfreelist].length = dps->nbtotalsector - curlba;
	  nbfreelist++;
	  }
      }

#ifdef DISK_USE_MALLOC
  if (nbfreelist != 0) {
      dps->freelist = malloc (nbfreelist * sizeof (struct freelist_str));
      if (dps->freelist == 0) {
	  dps->nbfreelist = 0;
	  DDBG (("%s: malloc failed!\r\n", __FUNCTION__));
	  }
	else {
	  dps->nbfreelist = nbfreelist;
	  while (nbfreelist--)
	      dps->freelist[nbfreelist] = freelist[nbfreelist];
	  }
      }
    else {
      dps->nbfreelist = 0;
      dps->freelist = 0;
      }
#else
  dps->nbfreelist = nbfreelist;
#endif
  DDBG (("]"));
  }}
#endif /* NB_FREELIST */

#if DISK_SUPPORT & EBIOS_SUPPORT
SEGMENT_EXTRA(get_ebios_ide_addresses) static inline void
get_ebios_ide_addresses (unsigned short *IDE_IObase_found,
			 unsigned nbof_IDE_IObase_found)
  {
  unsigned short disk, max_disk, max_disk1;
  struct diskparam_str param;

  max_disk = probe_nbBIOS (0x80);
  max_disk1 = peekb (BIOS_NBDISK_ADR);
  if (max_disk1 > max_disk)
      max_disk = max_disk1;

  for (disk = 0x80; disk < max_disk; disk++) {
      if (   EBIOSDISK_probe (disk, &param) == 0
	  && EBIOSDISK_getparam (disk, &param) == 0
	  && param.ideIOadr != 0) {
	  unsigned char cpt;
	  for (cpt = 0; cpt < nbof_IDE_IObase_found; cpt++)
	      if (   IDE_IObase_found[cpt] == 0
		  || IDE_IObase_found[cpt] == param.ideIOadr)
		  break;
	  if (cpt < nbof_IDE_IObase_found)
	      IDE_IObase_found[cpt] = param.ideIOadr;
	  }
      }
  }
#endif

/**
 ** Probe all BIOS Hard Drives, and then BIOS Floppy drives:
 **/
#if DISK_SUPPORT & (BIOS_SUPPORT | EBIOS_SUPPORT)
SEGMENT_EXTRA(probe_bios) static inline unsigned
probe_bios (unsigned nbdisk,
	    struct diskparam_str *param,
	    struct gujin_param_attrib attr)
  {
  signed short base_disk;

  for (base_disk = 0x80; base_disk >= 0; base_disk -= 0x80) {
      unsigned char nb_consecutive_fail = 0;
      unsigned char disk, max_disk, max_disk1;

      max_disk = probe_nbBIOS (base_disk);

      if (base_disk) {
	  if (!attr.probe_bios_hard_disk) {
	      DDBG (("Skiping HD probe due to current attributes.\r\n"));
	      continue;
	      }
	  max_disk1 = peekb (BIOS_NBDISK_ADR);
	  if (max_disk1 > max_disk)
	      max_disk = max_disk1;
	  DDBG (("Probing BIOS Hard drives (total = %u):", max_disk));
	  }
	else {
	  if (!attr.probe_bios_floppy_disk) {
	      DDBG (("Skiping floppy probe due to current attributes.\r\n"));
	      continue;
	      }
	  max_disk1 = _BIOS_equipment_flags().nb_floppy_less_1 + 1;
	  if (max_disk1 > max_disk)
	      max_disk = max_disk1;
#if !(DISK_SUPPORT & NO_FLOPPY_SUPPORT)
	  DDBG (("Probing BIOS/EBIOS floppy drives (total = %u):",
			max_disk));
#else
	  DDBG (("Probing EBIOS only floppy drives (total = %u):",
			max_disk));
#endif
	  }

      if (max_disk == 0 || max_disk > NB_DISK)
	  max_disk = NB_DISK;	/* the BIOS value is invalid */

      for (disk = 0;
	   disk < max_disk && nbdisk < NB_DISK && nb_consecutive_fail <= 3;
	   disk++) {
	  if (probeBIOSdisk (base_disk + disk, &param[nbdisk]) == 0) {
	      nbdisk++;
	      nb_consecutive_fail = 0;
	      }
	    else
	      nb_consecutive_fail++;
	  }
      DDBG ((" done\r\n"));
      }
  DDBG ((" %s end\r\n", __FUNCTION__));
  return nbdisk;
  }
#endif

/**
 ** Probe all DOS:
 **/
#if DISK_SUPPORT & DOS_SUPPORT
SEGMENT_EXTRA(probe_dos) static inline unsigned
probe_dos(unsigned nbdisk, struct diskparam_str	*param)
  {
  unsigned short nbDOSdisk;

  DDBG (("\r\nProbing DOS drives:"));
  nbDOSdisk = DOS_SelectDisk (DOS_GetCurrentDefaultDrive());
  DDBG ((" max drive %u", nbDOSdisk));
  if (nbDOSdisk > 'Z' - 'A') {
      DDBG ((" too high, I do not believe it (letter %c)\r\n",
			'A' + nbDOSdisk));
      }
    else {
      unsigned short cptdisk, BytesPerSector, ClustersOnDisk;
      unsigned char SectorPerCluster, MediaDescriptorByte;

      DDBG ((" = letter %c\r\n", 'A' + nbDOSdisk));
      for (cptdisk = 0;
	   nbdisk < NB_DISK && cptdisk < nbDOSdisk;
	   cptdisk++) {
	  if (   cptdisk == 1
	      && _BIOS_equipment_flags().nb_floppy_less_1 == 0) {
	      DDBG (("Do not probe 'B:' because only one floppy drive\r\n"));
	      continue;
	      }
	  DDBG (("Probing disk %c: ", 'A' + cptdisk));
	  MediaDescriptorByte = DOS_GetAllocationTableInformationForDrive (
					cptdisk + 1,
					&SectorPerCluster,
					&BytesPerSector,
					&ClustersOnDisk);
	  if (SectorPerCluster == 0xFF) {
	      DDBG (("cannot get allocation table, no drive there.\r\n"));
	      continue;
	      }
	    else
	      DDBG (("drive present.\r\n"));
	  param[nbdisk++] = (struct diskparam_str) {
		access: dos_part,
		disknb: 'A' + cptdisk,
		bytepersector: BytesPerSector,
		nbtotalsector: SectorPerCluster * ClustersOnDisk,
		nbsectorpertrack: SectorPerCluster,
		nbcylinder: ClustersOnDisk,
		nbhead: MediaDescriptorByte,
		nbpartition: 0,
		partition: 0,
		};
	  }
      }
  DDBG ((" %s end\r\n", __FUNCTION__));
  return nbdisk;
  }
#endif

/**
 ** Probe all IDE:
 **/
#if DISK_SUPPORT & IDE_SUPPORT
SEGMENT_EXTRA(probe_ide) static inline unsigned
probe_ide(unsigned nbdisk, struct diskparam_str	*param,
	  unsigned short *IDE_IObase_found,
	  unsigned nbof_IDE_IObase_found)
  {
  unsigned short cpt, index;

  DDBG (("\r\n\nKnow from BIOS the IDE interfaces:"));
  for (cpt = 0;
       cpt < nbof_IDE_IObase_found && IDE_IObase_found[cpt] != 0;
       cpt++)
      DDBG ((" 0x%X ", IDE_IObase_found[cpt]));

  DDBG (("\r\nAdding to-test IDE interfaces:"));
  for (index = 0; index < nbof (idearray); index++) {
      unsigned char i;
      for (i = 0; i < cpt; i++)
	  if (IDE_IObase_found[i] == idearray[index])
	      break;
      if (i >= cpt && i < nbof_IDE_IObase_found) {
	  DDBG ((" 0x%X ", idearray[index]));
	  IDE_IObase_found[cpt++] = idearray[index];
	  }
      }

  DDBG (("\r\nProbing IDE interfaces:\r\n"));
  for (index = cpt = 0;
       cpt < nbof_IDE_IObase_found && index < nbof (DI.IDE_IObase_found)
				     && IDE_IObase_found[cpt] != 0;
       cpt++) {
      DDBG (("testing 0x%X: ", IDE_IObase_found[cpt]));
      if (!ide_abscent (IDE_IObase_found[cpt]))
	  DI.IDE_IObase_found[index++] = IDE_IObase_found[cpt];
      DDBG (("\r\n"));
      }
  while (index < nbof (DI.IDE_IObase_found))
      DI.IDE_IObase_found[index++] = 0;

  if (DI.IDE_IObase_found[0] != 0) {
      unsigned short disk;

      DDBG (("\r\nProbing IDE drives:\r\n"));
      for (disk = 0;
	   DI.IDE_IObase_found[disk >> 1] != 0 && nbdisk < NB_DISK;
	   disk++) {
	  DDBG (("Probing IDE 0x%X drives %s:",
			DI.IDE_IObase_found[disk >> 1],
			(disk & 1)? "slave" : "master" ));
	  if (probeIDEdisk (DI.IDE_IObase_found[disk >> 1],
			    disk & 1, &param[nbdisk]) == 0) {
	      nbdisk++;
	      DDBG (("present\r\n"));
	      }
	    else
	      DDBG (("abscent\r\n"));
	  }
      }
  DDBG ((" %s end\r\n", __FUNCTION__));
  return nbdisk;
  }
#endif

/**
 ** The top-level functions:
 **/
GENERATE_VOID_CODE2EXTRA_STUB(probedisk,
				unsigned BIOSbootdevice,
				struct gujin_param_attrib attr)

SEGMENT_EXTRA(probedisk) void probedisk (unsigned BIOSbootdevice,
					 struct gujin_param_attrib attr)
  {bound_stack();{
  unsigned nbdisk = 0;
#ifdef DISK_USE_MALLOC
  struct diskparam_str	param[NB_DISK] = {};

  /* We need to free everything for the reprobe of the menu: */
  unsigned cptdisk;
  if (DI.nbdisk != 0 && DI.param == 0)
      DBG (("DI.nbdisk != 0 && DI.param == 0!\r\n"));

  for (cptdisk = 0; cptdisk < DI.nbdisk; cptdisk++) {
      struct diskparam_str *dp = &DI.param[cptdisk];
      if (dp->nbpartition != 0 && dp->partition == 0
#if defined (MAX_DISKNAMESIZE) && (MAX_DISKNAMESIZE != 0)
	  && dp->access != dos_part
#endif
	  )
	  DBG (("dp->nbpartition != 0 && dps->freelist == 0!\r\n"));
      if (dp->partition != 0)
	  free (dp->partition);
#if defined (NB_FREELIST) && (NB_FREELIST != 0)
      if (dp->nbfreelist != 0 && dp->freelist == 0)
	  DBG (("dp->nbfreelist != 0 && dp->freelist == 0!\r\n"));
      if (dp->freelist != 0)
	  free (dp->freelist);
#endif
      }
  if (DI.param != 0)
      free (DI.param);
#else
  struct diskparam_str	*param = DI.param;
#endif

  if (sizeof (bootsector_type) != 512)
      __ERROR();	/* the linker will produce an error here */

  if (attr.probe_bios_floppy_disk) {
      DDBG (("\r\nInitialising BIOS floppy disk system: "));
      BOOT1_initdisks (0x00);       /* init floppy only, for Phoenix BIOS */

      DDBG ((BIOSDISK_error (BIOSDISK_getstatus(0)) ));
      DDBG (("\r\n"));
      }

  DDBG (("Start probe disks:\r\n"));

#if DISK_SUPPORT & PROBE_ONLY_BIOSBOOTDEVICE
  {
  unsigned short cpt = 0xFFFF;

  nbdisk = 1;
  while (probeBIOSdisk (BIOSbootdevice, &param[0])) {
      // if BIOSbootdevice invalid, try floppies then HD:
      if (cpt == 0xFFFF) {
	  if (attr.probe_bios_floppy_disk)
	      cpt = 0;
	    else
	      cpt = 0x80;
	  }
	else if (cpt > 0x80 + peekb (BIOS_NBDISK_ADR)) {
	  nbdisk = 0;
	  break;
	  }
	else if (cpt != _BIOS_equipment_flags().nb_floppy_less_1 + 1)
	  cpt++;
	else
	  cpt = 0x80;
      }
  }
#else /* PROBE_ONLY_BIOSBOOTDEVICE */
  {
#if DISK_SUPPORT & (EBIOS_SUPPORT | IDE_SUPPORT)
  unsigned short IDE_IObase_found[2 * nbof (DI.IDE_IObase_found)] = {};

#if DISK_SUPPORT & EBIOS_SUPPORT
  get_ebios_ide_addresses (IDE_IObase_found, nbof (IDE_IObase_found));
#endif
#endif

#if DISK_SUPPORT & (BIOS_SUPPORT | EBIOS_SUPPORT)
  nbdisk = probe_bios (nbdisk, param, attr);
#else
  DDBG (("Probing BIOS HD/FD not compiled-in.\r\n"));
#endif /* DISK_SUPPORT & (BIOS_SUPPORT | EBIOS_SUPPORT) */

#if DISK_SUPPORT & DOS_SUPPORT
  if (!BOOT1_DOS_running())
      DDBG (("Skiping DOS probe in BIOS environment.\r\n"));
    else if (!attr.probe_dos_disk)
      DDBG (("Skiping DOS probe due to current attributes.\r\n"));
    else
      nbdisk = probe_dos (nbdisk, param);
#else
  DDBG (("Probing DOS not compiled-in.\r\n"));
#endif /* DISK_SUPPORT & DOS_SUPPORT */

#if DISK_SUPPORT & IDE_SUPPORT
#if !(DEBUG & DEBUG_DISK)	/* quick and dirty to get a log */
  if (BOOT1_DOS_running()) {
      DDBG (("Skiping IDE probe in DOS+ environment"));
      }
    else
#endif
	 if (!attr.probe_ide_disk)
      DDBG (("Skiping IDE probe due to current attributes.\r\n"));
    else
      nbdisk = probe_ide (nbdisk, param, IDE_IObase_found,
			  nbof (IDE_IObase_found));
#else
  DDBG (("Probing IDE not compiled-in.\r\n"));
#if DISK_SUPPORT & EBIOS_SUPPORT
  {
  unsigned short cpt;

  DDBG (("list of IDE interfaces found:"));
  for (cpt = 0;
       cpt < nbof(IDE_IObase_found) && IDE_IObase_found[cpt] != 0;
       cpt++) {
      DDBG ((" 0x%X ", IDE_IObase_found[cpt]));
      DI.IDE_IObase_found[cpt] = IDE_IObase_found[cpt];
      }
  while (cpt < nbof (DI.IDE_IObase_found))
      DI.IDE_IObase_found[cpt++] = 0;
  }
#endif /* DISK_SUPPORT & (BIOS_SUPPORT | EBIOS_SUPPORT) */
#endif /* DISK_SUPPORT & IDE_SUPPORT */
  }
#endif /* PROBE_ONLY_BIOSBOOTDEVICE */

  DDBG (("\r\nEnd probe disks.\r\n"));

#ifdef DISK_USE_MALLOC
  if (nbdisk != 0) {
      DI.param = malloc (nbdisk * sizeof (struct diskparam_str));
      if (DI.param == 0) {
	  DI.nbdisk = 0;
	  DDBG (("%s: malloc failed!\r\n", __FUNCTION__));
	  }
	else {
	  DI.nbdisk = nbdisk;
	  while (nbdisk--)
	      DI.param[nbdisk] = param[nbdisk];
	  }
      }
    else {
      DI.nbdisk = 0;
      DI.param = 0;
      }
#else
  DI.nbdisk = nbdisk;
#endif

#if NB_PARTITION > 1
  {
  unsigned disk;

  DDBG (("\r\nSort partition / Calculate free list: "));
#if NB_DISK > 1
  for (disk = 0; disk < DI.nbdisk; disk++) {
#else
  if (DI.nbdisk != 0) {
      disk = 0;
#endif
      if (DI.param[disk].nbpartition != 0) {
	  sort_partition (DI.param[disk].partition, DI.param[disk].nbpartition);
#if defined (NB_FREELIST) && NB_FREELIST != 0
	  calculatefreelist (&DI.param[disk]);
#endif /* NB_FREELIST */
	  }
      }
  DDBG (("\r\nEnd sort partition / free list.\r\n"));
  }
#endif

  }}

/**
 ** Main variable declaration:
 **/
struct disk_interface_str DI;

