/* disk.h */
#ifndef DISK_H
#define DISK_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"

/*
 * When I do not know which error to report:
 */
#define LOCALERROR 0x99

/*
 * Unfortunately, we need this data from the BIOS;
 * it is not implemented in DOSEMU, the probing
 * may fail then.
 */
#define BIOS_NBDISK_ADR 0x00400075

/*
 * Either "unsigned" or "unsigned long long",
 * but in fact we still do not have disks of 2048 Gb.
 */
typedef unsigned lba_t;
typedef unsigned nbsector_t;

/*
 * maximal number of disks, partition & freelist per disk (for boot2)
 * NB_DISK is set less or equal to 15 to not see the compaq bug
 *	(disk 0x80 = disk 0x90 = disk 0xA0 ...)
 * If you do not want the free list, just undefine NB_FREELIST.
 * If you do not want the name of the disk, just undefine MAX_DISKNAMESIZE.
 * Note that if NB_PARTITION is <= 4, extended partitions are not probed.
 */
#if DISK_SUPPORT & PROBE_ONLY_BIOSBOOTDEVICE
#define NB_DISK                1
#if DISK_SUPPORT & NO_PROBE_PARTITIONS
#define NB_PARTITION           1
#else
#define NB_PARTITION           4
#endif
#else
#define NB_DISK			15
#define NB_PARTITION		16
#define NB_FREELIST		8
#define MAX_DISKNAMESIZE	(32-12)
#endif

/*
 * This is the number of sector which are reserved at the beginning
 * of a partition, for instance at the beginning of a MSDOS partition,
 * one sector is reserved for the boot sector.
 * The "reserved" space is where we could put Gujin boot code.
 */
#define RESERVEDSECTOR 1

/*
 * The disk subsystem for boot2 :
 *
 * Dirty if "MAX_DISKNAMESIZE != 0" and access == dos_part,
 * the "partition" pointer is used as the DOS name pointer
 * because nbpartition == 0.
 */
extern struct disk_interface_str {
    unsigned			nbdisk;
    unsigned short		IDE_IObase_found[10];
    struct diskparam_str {
	enum { bios_chs, ebios_lba, hardide_chs, hardide_lba, dos_part }
			access : 8;
	unsigned char	disknb;	/* BIOS number or DOS letter
				   or 0=master, non zero=slave */
	unsigned short	bytepersector;
	unsigned short	nbhead;		  /* DOS: MediaDescriptorByte */
	unsigned	nbcylinder;       /* DOS: ClustersOnDisk */
	nbsector_t	nbtotalsector;
	unsigned short	nbsectorpertrack; /* DOS: SectorPerCluster */

	unsigned short	ideIOadr;	/* if known, else 0 */
	unsigned char	multiplemode;	/* int 0x13/0x21,0x22,0x24 or hardide */
	unsigned char	ebiosversion;	/* if present, used if access = ebios */
	unsigned char	irq;		/* just in case, unused now */
	unsigned char	ismaster;	/* what Phenix Bios compatible says */

	unsigned char	bootable;
	unsigned char	OSdisknumber;	/* as written in the boot block */

	unsigned short	nbpartition;
	unsigned short	nbOSnumber;
	struct partition_str {
	    unsigned	start, length;
	    unsigned char	type, OSnumber;
	    struct {
		unsigned char order      : 7;
		unsigned char maybe_root : 1;
		} __attribute__ ((packed)) misc;
	    /* OSnumber should be equal to the # in /dev/hda# */
	    enum part_state { part_invalid = 0, part_active, part_inactive }
			active;
#if defined (MAX_DISKNAMESIZE) && (MAX_DISKNAMESIZE != 0)
	    char	name[MAX_DISKNAMESIZE];
#endif
#ifdef DISK_USE_MALLOC
	    } *partition;
#else
	    } partition[NB_PARTITION];
#endif

#if defined (NB_FREELIST) && (NB_FREELIST != 0)
	unsigned	nbfreelist;
	struct freelist_str {
	    unsigned start, length;
#ifdef DISK_USE_MALLOC
	    } *freelist;
#else
	    } freelist[NB_FREELIST];
#endif
#endif
#ifdef DISK_USE_MALLOC
	} *param;
#else
	} param[NB_DISK];
#endif
    } DI;

/*
 * The functions:
 */
#if (SETUP & EXTRA_SEGMENT) && !defined (CALLING_FROM_EXTRASEG)
#define ASMNAME(x)	asm (STRING(codeseg_ ## x))
#else
#define ASMNAME(x)	/* call in same segment */
#endif

void probedisk (unsigned BIOSbootdevice,
		struct gujin_param_attrib attr) ASMNAME(probedisk);

unsigned char
DISK_readsector (struct diskparam_str *dp, lba_t lba,
		 unsigned char number, farptr buffer)
		ASMNAME(DISK_readsector);

#ifdef WRITE_ENABLE
unsigned char
DISK_writesector (struct diskparam_str *dp, lba_t lba,
		  unsigned char number, farptr buffer)
		ASMNAME(DISK_writesector);
#endif

#define SIZE_DISKNAME 32
void
diskname (struct diskparam_str *dp, char name[SIZE_DISKNAME])
		ASMNAME(diskname);

#undef ASMNAME

#endif /* DISK_H */
