/* dos.h */
#ifndef DOS_H
#define DOS_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

extern inline void
DOS_TerminateProgram (void)
  {
  /* %cs has to point to the PSP segment, deprecated */
  asm volatile (" int	$0x21 # DOS_TerminateProgram "
	: /* no output */
	: "a" ((unsigned short)0x0000)
	);
  }

extern inline unsigned char
DOS_StdinInputWithEcho (void) /* check keyboard ^C/^Break */
  {
  unsigned char returned;

  asm volatile (" int	$0x21 # DOS_StdinInputWithEcho "
	: "=a" (returned)
	: "a" ((unsigned short)0x0100)
	);
  return returned;
  }

extern inline void
DOS_StdoutOutput (unsigned char character) /* check keyboard ^C/^Break */
  {
  unsigned short code = 0x0200;
  asm volatile (" int	$0x21 # DOS_StdoutOutput "
	: "+a" (code)
	: "d" (character)
	);
  }

extern inline unsigned char
DOS_StdauxInput (void) /* check keyboard ^C/^Break */
  {
  unsigned char returned;

  asm volatile (" int   $0x21 # DOS_StdauxInput "
	: "=a" (returned)
	: "a" ((unsigned short)0x0300)
	);
  return returned;
  }

extern inline void
DOS_StdauxOutput (unsigned char character) /* check keyboard ^C/^Break */
  {
  asm volatile (" int	$0x21 # DOS_StdauxOutput "
	: /* no output */
	: "a" ((unsigned short)0x0400),
	  "d" (character)
	);
  }

extern inline void
DOS_StdprnOutput (unsigned char character) /* check keyboard ^C/^Break */
  {
  asm volatile (" int	$0x21 # DOS_StdprnOutput "
	: /* no output */
	: "a" ((unsigned short)0x0500),
	  "d" (character)
	);
  }

extern inline unsigned char
DOS_ConsoleInput (char *character) /* no ^C/^Break */
  {
  unsigned char error;

  asm volatile (
"	int	$0x21 # DOS_ConsoleInput	\n"
"	setz	%0				\n"
	: "=qm" (error), "=a" (*character)
	: "a" ((unsigned short)0x0600),
	  "d" ((unsigned char) 0xFF)
	);
  return error; /* != 0 if no char present */
  }

extern inline unsigned char
DOS_ConsoleOutput (char character) /* no ^C/^Break */
  {
  unsigned char error;
  unsigned short code = 0x0600;

  if (character == (char)0xFF)
      character = 0;

  asm volatile (
"	int	$0x21 # DOS_ConsoleOutput	\n"
"	setc	%0				\n"
	: "=qm" (error), "+a" (code)
	: "d" (character)
	);
  return error;
  }

extern inline unsigned char
DOS_DirectInputWithoutEcho (void) /* no ^C/^Break */
  {
  unsigned char character;

  asm volatile ("  int	$0x21 # DOS_DirectInputWithoutEcho "
	: "=a" (character)
	: "a" ((unsigned short)0x0700)
	);
  return character;
  }

extern inline unsigned char
DOS_StdinInput (void) /* check keyboard ^C/^Break */
  {
  unsigned char returned;

  asm volatile (" int   $0x21 # DOS_StdinInput "
	: "=a" (returned)
	: "a" ((unsigned short)0x0800)
	);
  return returned;
  }

/* check keyboard ^C/^Break: */
extern inline void
DOS_WriteStringStdout (const char *dollar_string)
  {
  unsigned short code = 0x0900;

  asm volatile (" int	$0x21 # DOS_WriteStringStdout "
	: "+a" (code)
	: "d" (dollar_string) /* %ds:%dx */
	);
  }

extern inline void
DOS_ResetDisk (void)
  {
  /* Flushes also files */
  asm volatile (" int	$0x21 # DOS_ResetDisk "
	: : "a" ((unsigned short)0x0D00)
	);
  }

extern inline unsigned char
DOS_SelectDisk (unsigned char disk)
  {
  unsigned char nb_logical_drive; /* LASTDRIVE */

  asm volatile (" int	$0x21 # DOS_SelectDisk "
	: "=a" (nb_logical_drive)
	: "a" ((unsigned short)0x0E00),
	  "d" ((unsigned char) disk) /* 0 for A: */
	);
  return nb_logical_drive;
  }

extern inline unsigned char
DOS_GetCurrentDefaultDrive (void)
  {
  unsigned char returned;

  asm ASMREAD (" int	$0x21 # DOS_GetCurrentDefaultDrive "
	: "=a" (returned)
	: "a" ((unsigned short)0x1900)
	);
  return returned; /* 0 for A: */
  }

extern inline unsigned char
DOS_GetAllocationTableInformation (unsigned char *SectorPerCluster,
				   unsigned short *BytesPerSector,
				   unsigned short *ClustersOnDisk)
  {
  farptr MediaDescriptorByte;
  asm ASMREAD (
"	pushl	%%ds						\n"
"	int	$0x21 # DOS_GetAllocationTableInformation	\n"
"	pushw	%%ds						\n"
"	pushw	%%bx						\n"
"	popl	%%ebx						\n"
"	popl	%%ds						\n"
	: "=a" (*SectorPerCluster), /* buggy Win95B, Win98 */
	  "=c" (*BytesPerSector),
	  "=d" (*ClustersOnDisk), /* buggy Win95B, Win98 */
	  "=b" (MediaDescriptorByte)
	: "a" ((unsigned short)0x1B00)
	);
  return peekb (MediaDescriptorByte);
  }

extern inline unsigned char
DOS_GetAllocationTableInformationForDrive (
		unsigned char drive, /* 0 for default, 1 for A: */
		unsigned char *SectorPerCluster,
		unsigned short *BytesPerSector,
		unsigned short *ClustersOnDisk)
  {
  farptr MediaDescriptorByte;
  asm ASMREAD (
"	pushl	%%ds						\n"
"	int	$0x21 # DOS_GetAllocationTableInformation	\n"
"	pushw	%%ds						\n"
"	pushw	%%bx						\n"
"	popl	%%ebx						\n"
"	popl	%%ds						\n"
	: "=a" (*SectorPerCluster), /* 0xFF if invalid drive */
	  "=c" (*BytesPerSector),
	  "=d" (*ClustersOnDisk), /* buggy Win95B, Win98 */
	  "=b" (MediaDescriptorByte)
	: "a" ((unsigned short)0x1C00),
	  "d" (drive)
	);
  return peekb (MediaDescriptorByte);
  }

extern inline void
DOS_SetInterruptVector (unsigned char vector_number, farptr address)
  {
  asm volatile (
"	pushl	%%ds					\n"
"	pushl	%%edx					\n"
"	popw	%%dx					\n"
"	popw	%%ds					\n"
"	int	$0x21 # DOS_SetInterruptVector		\n"
"	popl	%%ds					\n"
	: /* no output */
	: "a" ((unsigned short)0x2500 | vector_number),
	  "d" (address)
	);
  }

extern inline unsigned char
DOS_GetBootDrive (void)
  {
  unsigned char returned;

  asm ASMREAD (" int    $0x21 # DOS_GetBootDrive "
	: "=d" (returned)
	: "a" ((unsigned short)0x3305) /* DOS 4.x */
	);
  return returned; /* 1 for A: */
  }

extern inline farptr
DOS_GetInterruptVector (unsigned char vector_number)
  {
  farptr returned;

  asm ASMREAD (
"	pushl	%%es					\n"
"	int	$0x21 # DOS_GetInterruptVector		\n"
"	pushw	%%es					\n"
"	pushw	%%bx					\n"
"	popl	%%ebx					\n"
"	popl	%%es					\n"
	: "=b" (returned)
	: "a" ((unsigned short)0x3500 | vector_number)
	);
  return returned;
  }

extern inline unsigned
DOS_GetDiskFreeSpace (unsigned char drive, /* 0 for default, 1 for A: */
			unsigned short *SectorPerCluster,
			unsigned short *BytesPerSector,
			unsigned short *ClustersPerDrive,
			unsigned short *NbAvailableCluster)
  {
  asm ASMREAD ("  int	$0x21 # DOS_GetDiskFreeSpace "
	: "=a" (*SectorPerCluster),
	  "=c" (*BytesPerSector),
	  "=d" (*ClustersPerDrive),
	  "=b" (*NbAvailableCluster)
	: "a" ((unsigned short)0x3600),
	  "d" (drive)
	);
  if (*SectorPerCluster == 0xFFFF)
      return 1;
    else
      return 0;
  }

extern inline unsigned short DOS_old_version (void)
  {
  unsigned short version;

  asm ASMREAD (" int $0x21 # DOS_old_version"
	: "=a" (version)
	: "a" (0x3000)
	: "bx", "cx");

  return version;
  }

extern inline unsigned short DOS_new_version (void)
  {
  unsigned short version, result;

  asm ASMREAD (" int $0x21 # DOS_new_version (over 5.0)"
	: "=b" (version), "=a" (result)
	: "a" (0x3306), "b" (0) // version if failed: 0
	: "dx");
  if (   (result & 0xFF) != 0xFF
      && (version & 0xFF) >= 5
      && (version >> 8) < 100)
      return version;
    else
      return 0;
  }

extern inline unsigned short DOS_version (void)
  {
  unsigned short version;

  version = DOS_new_version ();
  if (version == 0)
      version = DOS_old_version ();
  /* return it in BCD to be simple, i.e. 0x301: */
  return (version >> 8) | (version << 8);
  }

/* All following are "Using Handle": */

struct DOS_attribute_str {
    unsigned short	read_only    : 1;
    unsigned short	hidden       : 1;
    unsigned short	system       : 1;
    unsigned short	volume_label : 1;
    unsigned short	directory    : 1;
    unsigned short	archive      : 1;
    unsigned short	reserved     : 10;
    } __attribute__ ((packed));

extern inline unsigned char
DOS_CreateFile (const char *name,
		struct DOS_attribute_str attribute,
		unsigned short *handle)
  {
  unsigned char error;

  asm volatile (
"	int	$0x21 # DOS_CreateFile, truncate if exists	\n"
"	setc	%0						\n"
	: "=qm" (error), "=a" (*handle)
	: "a" ((unsigned short)0x3C00),
	  "c" (attribute),
	  "d" (name)	/* in fact %ds:%dx */
	: "cc");
  return error;
  }

enum DOS_open_access {
    DOS_open_read_only = 0,
    DOS_open_write_only,
    DOS_open_read_write
    };

extern inline unsigned char
DOS_OpenFile (const char *name,
	      unsigned char attribute,
	      unsigned short *handle)
  {
  unsigned char error;

  asm volatile (
"	stc		# Error creation if DOS not running	\n"
"	int	$0x21	# DOS_OpenFile				\n"
"	setc	%0						\n"
	: "=qm" (error), "=a" (*handle)
	: "a" ((unsigned short)0x3D00 | attribute),
	  "d" (name)	/* in fact %ds:%dx */
	: "cc");
  return error;
  }

extern inline unsigned char
DOS_CloseFile (unsigned short handle, unsigned short *errorcode)
  {
  unsigned char error;

  asm volatile (
"	int	$0x21 # DOS_CloseFile		\n"
"	setc	%0				\n"
	: "=qm" (error), "=a" (*errorcode)
	: "a" ((unsigned short)0x3E00),
	  "b" (handle)
	: "cc");
  return error;
  }

extern inline unsigned char
DOS_ReadFile (unsigned short handle, void *buffer, unsigned short nb,
	      unsigned short *nbread_errorcode)
  {
  unsigned char error;

  asm volatile (
"	int	$0x21 # DOS_ReadFile		\n"
"	setc	%0				\n"
	: "=qm" (error), "=a" (*nbread_errorcode)
	: "a" ((unsigned short)0x3F00), "b" (handle),
	  "c" (nb), "d" (buffer) /* in fact %ds:%dx */
	: "cc");
  return error;
  }

extern inline unsigned char
DOS_WriteFile (unsigned short handle, const void *buffer,
	       unsigned short nb, unsigned short *nbwrite_errorcode)
  {
  unsigned char error;
  /* Writing 0 bytes truncate the file to current position */

  asm volatile (
"	int	$0x21 # DOS_WriteFile		\n"
"	setc	%0				\n"
	: "=qm" (error), "=a" (*nbwrite_errorcode)
	: "a" ((unsigned short)0x4000), "b" (handle),
	  "c" (nb), "d" (buffer) /* in fact %ds:%dx */
	: "cc");
  return error;
  }

extern inline unsigned char
DOS_DeleteFile (const char *name, unsigned short *errorcode)
  {
  unsigned char error;

  asm volatile (
"	int	$0x21 # DOS_CloseFile		\n"
"	setc	%0				\n"
	: "=qm" (error), "=a" (*errorcode)
	: "a" ((unsigned short)0x4100),
	  "d" (name) /* %ds:%dx */
	: "cc");
  return error;
  }

enum DOS_seek { DOS_SEEK_SET = 0, DOS_SEEK_CUR, DOS_SEEK_END };

extern inline unsigned char
DOS_SeekFile (unsigned short handle,
	      enum DOS_seek  origin,
	      signed int     pos,
	      unsigned       *newpos,
	      unsigned short *errorcode)
  {
  unsigned char error;

  asm volatile (
"	movw	%%cx,%%dx			\n"
"	sarl	$16,%%ecx			\n"
"	int	$0x21 # DOS_SeekFile		\n"
"	setc	%0				\n"
"	shll	$16,%%edx			\n"
"	mov	%%ax,%%dx			\n"
	: "=qm" (error), "=a" (*errorcode), "=d" (*newpos)
	: "a" ((unsigned short)0x4200 | origin),
	  "b" (handle),
	  "c" (pos)
	: "cc");
  return error;
  }

extern inline unsigned char
DOS_GetAttributeFile (const char *name,
		      struct DOS_attribute_str *attribute,
		      unsigned short *errorcode)
  {
  unsigned char error;

  asm volatile (
"	int	$0x21 # DOS_GetAttributeFile	\n"
"	setc	%0				\n"
	: "=qm" (error), "=a" (*errorcode), "=c" (*attribute)
	: "a" ((unsigned short)0x4300),
	  "d" (name)	/* in fact %ds:%dx */
	: "cc");
  return error;
  }

extern inline unsigned char
DOS_SetAttributeFile (const char *name,
		      struct DOS_attribute_str attribute,
		      unsigned short *errorcode)
  {
  unsigned char error;

  asm volatile (
"	int	$0x21 # DOS_SetAttributeFile	\n"
"	setc	%0				\n"
	: "=qm" (error), "=a" (*errorcode)
	: "a" ((unsigned short)0x4301),
	  "c" (attribute),
	  "d" (name)	/* in fact %ds:%dx */
	: "cc");
  return error;
  }

extern inline unsigned char
DOS_GetCwd (unsigned char disk, /* 0 for default, 1 for A: */
	    char name[64],	/* allocated by user */
	    unsigned short *errorcode)
  {
  unsigned char error;

  asm volatile (
"	int	$0x21 # DOS_GetCwd	\n"
"	setc	%0			\n"
	: "=qm" (error), "=a" (*errorcode)
	: "a" ((unsigned short)0x4700),
	  "d" (disk),
	  "S" (name) /* in fact %ds:%si */
	: "cc");
  return error;
  }


struct DOS_DTA_str {
    unsigned char		OSdepend[0x15];
    struct file_attr_str	attribute;
    unsigned short		time;
    unsigned short		date;
    unsigned			size;
    unsigned char		name_ext[13]; /* ASCIZ */
    } __attribute__ ((packed));

extern inline void
DOS_SetDTA (struct DOS_DTA_str *dta_atr)
  {
  asm volatile (" int	$0x21 # DOS_SetDTA "
	: /* no output */
	: "a" ((unsigned short)0x1A00),
	  "d" (dta_atr) /* in fact %ds:%dx */
	);
  }

extern inline farptr
DOS_GetDTA (void)
  {
  farptr dta_adr;

  asm volatile (
"	pushl	%%es			\n"
"	int	$0x21 # DOS_GetDTA	\n"
"	pushw	%%es			\n"
"	pushw	%%bx			\n"
"	popl	%%ebx			\n"
"	popl	%%es			\n"
	: "=b" (dta_adr)
	: "a" ((unsigned short)0x2F00)
	);
  return dta_adr;
  }

extern inline unsigned char
DOS_FindFirst (const char *pattern,
		struct DOS_attribute_str attribute,
		unsigned short *errorcode)
  {
  unsigned char error;

  asm volatile (
"	int	$0x21 # DOS_FindFirst		\n"
"	setc	%0				\n"
	: "=qm" (error), "=a" (*errorcode)
	: "a" ((unsigned short)0x4E00),
	  "c" (attribute),
	  "d" (pattern)	/* in fact %ds:%dx */
	: "memory", "cc");
  return error;
  }

extern inline unsigned char
DOS_FindNext (unsigned short *errorcode)
  {
  unsigned char error;

  asm volatile (
"	int	$0x21 # DOS_FindNext		\n"
"	setc	%0				\n"
	: "=qm" (error), "=a" (*errorcode)
	: "a" ((unsigned short)0x4F00)
	: "memory", "cc");
  return error;
  }

/* DOS 5+ : ALLOCATE HMA SPACE only valid if DOS=HIGH */
extern inline farptr
DOS_allocate_hma_space (unsigned short nbbytes, unsigned short *total_alloc)
  {
  farptr addr;
  asm volatile (
"	pushl	%%es		\n"
"	int	$0x2F		\n"
"	pushw	%%es		\n"
"	pushw	%%di		\n"
"	popl	%%edi		\n"
"	popl	%%es		\n"
	: "=D" (addr), "=b" (*total_alloc)
	: "a" (0x4A02), "b" (nbbytes)
	);
  return addr;
  }

extern inline unsigned DOS_loaded_HIGH (void)
  {
  unsigned short total_alloc;
  return DOS_allocate_hma_space (0, &total_alloc) != 0xFFFFFFFF
	 || total_alloc != 0;
  }

struct SMARTDRV_status { /* 0xFF if drive do not exist */
    unsigned char	drive         : 6;
    unsigned char	write_through : 1; /* i.e. not write-cached */
    unsigned char	not_cached    : 1;
    unsigned char	abscent;	/* non zero if no cache present */
    } __attribute__ ((packed));

enum SMARTDRV_enum {
    SMARTDRV_get_info = 0,
    SMARTDRV_read_cache_on,
    SMARTDRV_read_cache_off,
    SMARTDRV_write_cache_on,
    SMARTDRV_write_cache_off
    };

extern inline struct SMARTDRV_status
SMARTDRV_control (unsigned short drive, enum SMARTDRV_enum command)
  {
  /* SMARTDRV v4.00+ and PC-Cache v8.0,
     drive = 0 for A: ... */
  unsigned short ax = 0x4A10;
  struct SMARTDRV_status returned;

  asm volatile (
"	pushl	%%ebp		\n"
"	movw	%%bx,%%bp	\n"
"	movw	$3,%%bx		\n"
"	int	$0x2F		\n"
"	popl	%%ebp		\n"
	: "+a" (ax), "+b" (drive), "=d" (returned)
	: "d" (command)
	);
  if (ax != 0xBABE)
      return (struct SMARTDRV_status) { 0, 0, 0, 0xFF };
  return returned;
  }

extern inline void
SMARTDRV_flush_buffer (void)
  {
  /* SMARTDRV v4.00+ and PC-Cache v8.0, commit before flushing */
    asm volatile (" int	$0x2F "
	: : "a" (0x4A10), "b" (1)
	);
  }
#endif /* DOS_H */
