/* gzlib.c */

/*
  The original source is the zlib library - but it has been heavily
  modified to save code space (begin with gcc -E), bugs are mine.

 *********  My changes are covered by the GPL license  *********
 ***************************************************************
 * Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 ***************************************************************

  zlib.h -- interface of the 'zlib' general purpose compression library
  version 1.1.3, July 9th, 1998

  Copyright (C) 1995-1998 Jean-loup Gailly and Mark Adler

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  Jean-loup Gailly        Mark Adler
  jloup@gzip.org          madler@alumni.caltech.edu


  The data format used by the zlib library is described by RFCs (Request for
  Comments) 1950 to 1952 in the files ftp://ds.internic.net/rfc/rfc1950.txt
  (zlib format), rfc1951.txt (deflate format) and rfc1952.txt (gzip format).
*/

#include "make.h"
#if SETUP & EXTRA_SEGMENT		/* before other includes */
#define CALLING_FROM_EXTRASEG
#endif

#include "library.h"
#include "debug.h"
#include "boot.h"	/* struct gujin_param_attrib for vmlinuz.h */
#include "vmlinuz.h"	/* struct system_load_str */
#include "gzlib.h"

static const char inflate_copyright[] __attribute__ ((unused)) =
	" inflate 1.1.3 Copyright 1995-1998 Mark Adler,"
	" modified by Etienne Lorrain, GPL ONLY ";

#ifdef GZLIB_USE_INTERMEDIATE_WINDOW
#ifndef GZLIB_BIG_MALLOC
/*
 * For the TINY configuration, we have enought space because
 * E2FS library is not there. The stack is limited to 32Kb to keep
 * check_stack() working (signed bound limits), so we have:
 */
SEGMENT_EXTRA(memcpy_data2window) static inline void
memcpy_data2window (unsigned char *dest,
		    const unsigned char *src,
		    unsigned short size)
  {
  _memcpy (dest, src, size);
  }

SEGMENT_EXTRA(memcpy_window2window) static inline void
memcpy_window2window (unsigned char *dest,
		      const unsigned char *src,
		      unsigned short size)
  {
  _memcpy (dest, src, size);
  }

#define OUTBYTE(q, a)	{*q=(unsigned char)(a);}
#define CRCBYTE(q)	(*q++)

unsigned char       gzlib_window[1<<MAX_WBITS];
#define ZWINALLOC(w)	gzlib_window /* malloc (w) */
#define ZWINFREE(w)	w = 0 /* free (w) */
#define ZWINFARPTR(win)	data_adr(win)

#else /* GZLIB_BIG_MALLOC */

/*
 * Argggg... gzlib needs a 32 Kbytes buffer to decompress,
 * (and even more to compress) and if BIG_MALLOC is not
 * defined we will not have enough malloc memory to load
 * all that e2fs requires and that gzlib buffer... We just
 * have approx 45..50 Kbytes free without BIG_MALLOC.
 *
 * Requesting BIG_MALLOC means we cannot run as a real
 * DOS/other-OS bootloader.
 *
 * Fortunately, because (now) gzlib is only accessing that
 * big buffer using macro OUTBYTE() or with zmemcpy() -
 * we can provide it a "farptr" and new OUTBYTE()/zmemcpy()
 * which handle such a kind of pointer. Whatever calculus
 * gzlib does (inside a 64K segment) on the address, it will
 * behave the same with a farptr. If we ever forget one
 * read or write to this big buffer (content of a farptr),
 * we will get a "General Protection Exception": easy to debug!
 */
SEGMENT_EXTRA(memcpy_data2window) static inline void
memcpy_data2window (unsigned char *dest,
		    const unsigned char *src,
		    unsigned short size)
  {
  smemcpy ((farptr)dest, src, size);
  }

SEGMENT_EXTRA(memcpy_window2window) static inline void
memcpy_window2window (unsigned char *dest,
		      const unsigned char *src,
		      unsigned short size)
  {
  asm volatile (
"	pushw	%%es						\n"
"	popw	%%fs						\n"
"	pushl	%%edi						\n"
"	popw	%%di						\n"
"	popw	%%es						\n"
"	cld							\n"
"	rep movsb %%es:(%%si),%%es:(%%di)	# no macro	\n"
"	nop							\n"
"	pushw	%%fs						\n"
"	popw	%%es						\n"
	: "+S" (src), "+D" (dest), "+c" (size)
	: : "memory", "cc"
	);
  }

/* Here we _assume_ that "q" is a farptr ! */
#define OUTBYTE(q, a)	pokeb(((farptr)q), a)
#define CRCBYTE(q)	peekb(((farptr)q)++)

/*
 * The 32 Kb window malloc returns the address just following the top
 * of the stack, exactly %ss + 64 Kbytes because %sp has been cleared
 * at startup.
 * There is no re-entrancy check or malloc/free order check, take care!
 */
#define ZWINALLOC(w)	((void *)((0x1000U + getss()) << 16))
#define ZWINFREE(w)	w = 0
#define ZWINFARPTR(win)	((unsigned)win)

#endif /* GZLIB_BIG_MALLOC */
/*
 * If we were using an intermediate window when ASSEMBLY_TYPE == ASSEMBLY_DSES
 * we would have:
 * #if ASSEMBLY_TYPE != ASSEMBLY_DSES
 *  ...
 * #else
 *
 * SEGMENT_EXTRA(memcpy_data2window) static inline void
 * memcpy_data2window (unsigned char *dest,
 *	    const unsigned char *src,
 *	    unsigned short size)
 *   {
 *   memcpy_door_open (dest, src, size);
 *   }
 *
 * SEGMENT_EXTRA(memcpy_window2window) static inline void
 * memcpy_window2window (unsigned char *dest,
 *	      const unsigned char *src,
 *	      unsigned short size)
 *   {
 *   memcpy_door_open (dest, src, size);
 *   }
 *
 * #define OUTBYTE(q, a)   {*q=(unsigned char)(a);}
 * #define CRCBYTE(q)      (*q++)
 *
 * #define ZWINALLOC(w)    ((void *)0x10000U)
 * #define ZWINFREE(w)     w = 0
 * #define ZWINFARPTR(win) ((unsigned)win)
 *
 * SEGMENT_EXTRA(treat_data) static inline unsigned short
 * treat_data (z_stream *z, void *out, void *buffer, unsigned len)
 *   {
 *   memcpy_door_open (out, buffer, len);
 *   return 0;
 *   }
 * #endif
 */

#ifndef treat_data
/*
 * The treat_data() function which consumes the decompressed data:
 */
#include "xms.h"

/* Those two memory copier functions assume size != 0 */
SEGMENT_EXTRA(XMScopy) static inline unsigned
XMScopy (unsigned XMSentry,
	 unsigned short *KBalloced,
	 unsigned short handle,
	 unsigned offset,
	 farptr   windowfarptr,
	 unsigned size)
  {
  unsigned char error;
  XMS_EMM emm_struct = {
      size: size + (size & 1),	/* even number of bytes */
      src_handle: 0,		/* real mode address */
      src_offset: windowfarptr,
      dst_handle: handle,
      dst_offset: offset
      };

  for (;;) {
      if (_XMS_move_memory (XMSentry, &emm_struct, &error) == 0)
	  break;
      if (error == XMS_error_length_is_invalid) {
	  *KBalloced += 256; /* 256 Kbytes increment */
	  if (_XMS_realloc_memory (XMSentry, handle, *KBalloced, &error)) {
	      ZDBG (("_XMS_realloc_memory(%u KB) error 0x%X!\r\n",
			*KBalloced, error));
	      return 1;
	      }
	  ZDBG (("_XMS_realloc_memory(%u) OK\r\n", *KBalloced));
	  }
	else {
	  ZDBG (("_XMS_move_memory error 0x%X!\r\n", error));
	  return 2;
	  }
      }

  return 0;
  }

#if SETUP & USE_INT1587
#include "bios.h"

SEGMENT_EXTRA(BIOS_memcpy) static inline unsigned short
BIOS_memcpy (unsigned dest, farptr windowfarptr, unsigned size)
  {
  unsigned short res;
  res = _BIOS_copy_extended_memory (dest, farptr2linear(windowfarptr),
				    (size+1)/2);
  if (res != 0) {
      ZDBG ((" [_BIOS_copy_extended_memory returns 0x%X] ", res));
      return res;
      }
  return 0;
  }
#endif /* USE_INT1587 */

SEGMENT_EXTRA(treat_data) static inline unsigned short
treat_data (z_stream *z, void *out, void *buffer, unsigned len)
  {
  if (z->extra->XMSentry)
      return XMScopy (z->extra->XMSentry,
		      &z->extra->KBalloced,
		      z->extra->handle,
		      (unsigned)out,
		      ZWINFARPTR(buffer),
		      len);
    else {
#if SETUP & USE_INT1587
      return BIOS_memcpy ((unsigned)out, ZWINFARPTR(buffer), len) >> 8;
#else /* USE_INT1587 */
      memcpy_door_open (out, buffer, len);
      return 0;
#endif /* USE_INT1587 */
      }
  }
#endif /* treat_data */

#else /* GZLIB_USE_INTERMEDIATE_WINDOW */
#define CRCBYTE(q)	(*q++)
#endif /* GZLIB_USE_INTERMEDIATE_WINDOW */

/*
 * This is derived from the crc32() function in zlib, but to save
 * space I removed the precompiled array - so it is slower.
 * See zlib documentation for copyright of its own functions;
 * the following rewritten functions (added work) are GPL ONLY.
 * -----------------------------------------------------------
 */
#ifdef GZLIB_USE_PRECOMPILED_CRC_ARRAY
SEGMENT_EXTRA(crc32) static unsigned
crc32 (unsigned crc, const unsigned char *buf, unsigned len)
  {
  static unsigned crc_table[256];

  if (buf == 0) {
      unsigned short n;
      unsigned poly = 0xedb88320L;	/* polynomial exclusive-or pattern */

      for (n = 0; n < 256; n++) {
	  unsigned char k;
	  unsigned c = (unsigned) n;
	  for (k = 0; k < 8; k++)
	      c = c & 1 ? poly ^ (c >> 1) : c >> 1;
	  crc_table[n] = c;
	  }
      return 0L;
      }

  crc = crc ^ 0xffffffffL;
  if (len != 0) {
      unsigned nb = ((len - 1) / 8) + 1;

      switch (len % 8) { do {
	  case 0: crc = crc_table[(crc ^ CRCBYTE(buf)) & 0xff] ^ (crc >> 8);
	  case 7: crc = crc_table[(crc ^ CRCBYTE(buf)) & 0xff] ^ (crc >> 8);
	  case 6: crc = crc_table[(crc ^ CRCBYTE(buf)) & 0xff] ^ (crc >> 8);
	  case 5: crc = crc_table[(crc ^ CRCBYTE(buf)) & 0xff] ^ (crc >> 8);
	  case 4: crc = crc_table[(crc ^ CRCBYTE(buf)) & 0xff] ^ (crc >> 8);
	  case 3: crc = crc_table[(crc ^ CRCBYTE(buf)) & 0xff] ^ (crc >> 8);
	  case 2: crc = crc_table[(crc ^ CRCBYTE(buf)) & 0xff] ^ (crc >> 8);
	  case 1: crc = crc_table[(crc ^ CRCBYTE(buf)) & 0xff] ^ (crc >> 8);
	  } while (--nb); }
      }
  return crc ^ 0xffffffffL;
  }
#else /* GZLIB_USE_PRECOMPILED_CRC_ARRAY */
SEGMENT_EXTRA(crc32) static inline unsigned
crc32 (unsigned crc, const unsigned char *buf, unsigned len)
  {
  const unsigned char *end = buf + len;

  if (buf == 0)
	return 0L;

  crc = ~crc;
  while (buf < end) {
      unsigned c = (crc ^ CRCBYTE(buf)) & 0xff;
      unsigned short k;
#if ASSEMBLY_TYPE == ASSEMBLY_NONE
#define POLY	0xedb88320L /* polynomial exclusive-or pattern */
      for (k = 8; k != 0; k--)
	  c = c & 1 ? (c >> 1) ^ POLY : c >> 1;
#elif 0	// Optimised in assembler - 1st step
      asm (
"	movw $8,%%cx		\n"
"	1:			\n"
"	clc			\n"
"	rcrl %0			\n"
"	jnc 2f			\n"
"	xorl $0xedb88320,%0	\n"
"	2:			\n"
"	loopw 1b		\n"
	: "+r" (c), "=c" (k));
#else	// I think I cannot do faster without precompiled array
      unsigned short tmp;
      asm (
"	movw $8,%w2		\n"
"	1:			\n"
"	bsfw %w0,%%cx		\n"
"	jz 2f			\n"
"	incw %%cx		\n"
"	subw %%cx,%w2		\n"
"	jc 3f			\n"
"	shrl %%cl,%0		\n"
"	xorl $0xedb88320,%0	\n"
"	jmp 1b			\n"
"	2:			\n"
"	xorw %%cx,%%cx		\n"
"	3:			\n"
"	addw %w2,%%cx		\n"
"	shrl %%cl,%0		\n"
	: "+r" (c), "=c" (k), "=r" (tmp));
#endif
      crc = c ^ (crc >> 8);
      }
  return ~crc;
  }
#endif /* GZLIB_USE_PRECOMPILED_CRC_ARRAY */

/*
 * These are helpers for decompression:
 */
SEGMENT_EXTRA(readbits) static inline unsigned
readbits (z_stream *z, unsigned short nb)
  {
  while (z->bitk < nb) {
      if (UNUSUAL(z->avail_in == 0))
	  return 1;
      z->bitb |= ((unsigned) (z->avail_in--, *z->next_in++)) << z->bitk;
      z->bitk += 8;
      }
  return 0;
  }

SEGMENT_EXTRA(maskbits) static inline unsigned short
maskbits (z_stream *z, unsigned short nb)
  {
  return z->bitb & ~(0xFFFFFFFFU << nb);
  }

SEGMENT_EXTRA(flushbits) static inline void
flushbits (z_stream *z, unsigned short nb)
  {
  z->bitb >>= nb;
  z->bitk -= nb;
  }

/*
   Huffman code decoding is performed using a multi-level table lookup.
   The fastest way to decode is to simply build a lookup table whose
   size is determined by the longest code.  However, the time it takes
   to build this table can also be a factor if the data being decoded
   is not very long.  The most common codes are necessarily the
   shortest codes, so those codes dominate the decoding time, and hence
   the speed.  The idea is you can have a shorter table that decodes the
   shorter, more probable codes, and then point to subsidiary tables for
   the longer codes.  The time it costs to decode the longer codes is
   then traded against the time it takes to make longer tables.

   This results of this trade are in the variables lbits and dbits
   below.  lbits is the number of bits the first level table for literal/
   length codes can decode in one step, and dbits is the same thing for
   the distance codes.  Subsequent tables are also less than or equal to
   those sizes.  These values may be adjusted either when all of the
   codes are shorter than that, in which case the longest code length in
   bits is used, or when the shortest code is *longer* than the requested
   table size, in which case the length of the shortest code in bits is
   used.

   There are two different values for the two tables, since they code a
   different number of possibilities each.  The literal/length table
   codes 286 possible values, or in a flat code, a little over eight
   bits.  The distance table codes 30 possible values, or a little less
   than five bits, flat.  The optimum values for speed end up being
   about one bit more than those, so lbits is 8+1 and dbits is 5+1.
   The optimum values may differ though from machine to machine, and
   possibly even between compilers.  Your mileage may vary.
 */

/* Given a list of code lengths and a maximum table size, make a set of
   tables to decode that set of codes.
   Return Z_OK on success, Z_BUF_ERROR if the given code set is incomplete,
   Z_DATA_ERROR if the input is invalid (an over-subscribed set of lengths),
   or Z_MEM_ERROR if not enough memory. */
SEGMENT_EXTRA(huft_build) static int huft_build (
	unsigned short *b,	/* code lengths in bits (all assumed <= BMAX) */
	unsigned short n,	/* number of codes (assumed <= 288) */
	unsigned short s,	/* number of simple-valued codes (0..s-1) */
	const unsigned short *d,/* list of base values for non-simple codes */
	const unsigned char *e,	/* list of extra bits for non-simple codes */
	inflate_huft_size *result,	/* starting table +
					  maximum lookup bits, returns actual */
	inflate_huft *hp,	/* space for trees */
	unsigned short *hn	/* hufts used in space */
	)
{
  /* work area, either 19 or 288: */
  unsigned short v[288] = {[0 ... 287] = 0 };
  unsigned short bit_length_count[BMAX + 1] = {[0 ... BMAX] = 0 };
  unsigned short maximum_code_length, current_code_length;
  int returned;

  unsigned short x[BMAX + 1];	/* bit offsets, then code stack */

  /* Generate counts for each bit length */
  {
  unsigned short i = n;
  unsigned short *p = b;

  do {
      bit_length_count[*p++]++;	/* assume all entries <= BMAX */
      } while (--i);
  }

  /* Test for null input--all zero length codes */
  if (bit_length_count[0] == n) {
      result->tree = (inflate_huft *) 0;
      result->need = 0;
      return Z_OK;
      }

  /* Find minimum and maximum length, bound "result->need" by those */
  for (current_code_length = 1;
       current_code_length <= BMAX;
       current_code_length++)
      if (bit_length_count[current_code_length])
	  break;
  if (result->need < current_code_length)	/* minimum code length */
      result->need = current_code_length;

  for (maximum_code_length = BMAX;
       maximum_code_length != 0;
       maximum_code_length--)
      if (bit_length_count[maximum_code_length])
	  break;
  if (result->need > maximum_code_length)
      result->need = maximum_code_length;

  /* Adjust last length count to fill out codes, if needed */
  {
  unsigned short j = current_code_length;
  int number_of_dummy_codes_added = 1 << current_code_length;

  for (;;) {
      number_of_dummy_codes_added -= bit_length_count[j];
      if (number_of_dummy_codes_added < 0)
	  return Z_DATA_ERROR;
      if (j <= maximum_code_length) {
	  if (j == maximum_code_length)
	      break;
	  number_of_dummy_codes_added <<= 1;
	  }
      j++;
      }

  /* Return Z_BUF_ERROR if we were given an incomplete table */
  if (number_of_dummy_codes_added != 0 && maximum_code_length != 1) {
      /* inflate_trees_fixed() will finish with this error (ignored),
	 but wait the table (distance table) to be filled in! */
      returned = Z_BUF_ERROR;
      }
    else
      returned = Z_OK;

  bit_length_count[maximum_code_length] += number_of_dummy_codes_added;
  }

  /* Generate starting offsets into the value table for each length */
  {
  unsigned short i = maximum_code_length + 1;
  unsigned short *p = &bit_length_count[1];
  unsigned short *xp = &x[1];
  unsigned short accumulate = 0;

  while (--i) {
      *xp++ = accumulate;
      accumulate += *p++;
      }
  }

  /* Make a table of values in order of bit lengths */
  {
  unsigned short i = 0;

  for (i = 0; i < n; ++i) {
      if (b[i] != 0) {
	  v[x[b[i]]] = i;
	  x[b[i]]++;
	  }
      }
  }

  /* Generate the Huffman codes and for each, make the table entries */
  {
  int table_level;
  int bits_before_this_table;	/* == (result->need * table_level) */
  inflate_huft *table_stack[BMAX];
  /* the Huffman code of length current_code_length bits for value *p: */
  unsigned short current_code;
  unsigned short *p;		/* pointer into v[] */
  unsigned short nb_entry_current_table;
  unsigned short *end_of_v = &v[x[maximum_code_length]];

  /* just to keep compilers happy: */
  table_stack[0] = (inflate_huft *) 0;
  nb_entry_current_table = 0;

  /* go through the bit lengths
    (current_code_length already is bits in shortest code) */
  for (x[0] = current_code = 0,	/* first Huffman code is zero */
       p = v,			/* grab values in bit order */
       table_level = -1,	/* no tables yet--level -1 */
       bits_before_this_table = -result->need;	/* bits decoded */
       current_code_length <= maximum_code_length;
       current_code_length++) {
      /* counter for codes of length current_code_length: */
      unsigned short a = bit_length_count[current_code_length];

      /* make tables up to required level */
      while (a--) {
	  while (current_code_length > bits_before_this_table + result->need) {
	      unsigned short bit_in_this_table;

	      table_level++;
	      /* previous table always "result->need" bits: */
	      bits_before_this_table += result->need;

	      /* compute minimum size table less than
		 or equal to "result->need" bits: */
	      nb_entry_current_table = maximum_code_length
					- bits_before_this_table;
	      /* table size upper limit: */
	      if (nb_entry_current_table > result->need)
		  nb_entry_current_table = result->need;

	      bit_in_this_table = current_code_length - bits_before_this_table;

	      if (1 << bit_in_this_table > a + 1) {
		  /* too few codes for k-w bit table */
		  /* deduct codes from patterns left */
		  unsigned short f = (1 << bit_in_this_table) - a + 1;
		  unsigned short *xp = &bit_length_count[current_code_length];

		  if (bit_in_this_table < nb_entry_current_table) {
		      /* try smaller tables upto nb_entry_current_table bits: */
		      while (++bit_in_this_table < nb_entry_current_table) {
			  f <<= 1;
			  xp++;
			  if (f <= *xp)
			      break; /* enough codes to use up
					bit_in_this_table bits */
			  f -= *xp;	 /* else deduct codes from patterns */
			  }
		      }
		  }
	      nb_entry_current_table = 1 << bit_in_this_table;

	      /* allocate new table */
	      table_stack[table_level] = hp + *hn;
	      *hn += nb_entry_current_table;
	      if (*hn > MANY)
		  return Z_MEM_ERROR;	/* not enough memory */

	      /* connect to last table, if there is one */
	      if (!table_level) {
		  /* first table is returned result: */
		  result->tree = table_stack[table_level];
		  }
		else {
		  unsigned short j = current_code >> (bits_before_this_table
							- result->need);
		  /* connect to last table: */
		  inflate_huft *ptr = &table_stack[table_level - 1][j];

		  /* save pattern for backing up: */
		  x[table_level] = current_code;
		  ptr->exop = bit_in_this_table;
		  ptr->bits = result->need; /* bits to dump before this table */
		  ptr->base = (table_stack[table_level]
				 - table_stack[table_level - 1]
				 - j);	    /* offset to this table */
		  }
	      }

	  {
	  /* set up table entry in tmp */
	  inflate_huft tmp;

	  tmp.bits = current_code_length - bits_before_this_table;
	  if (p >= end_of_v) {
	      /* out of values--invalid code: */
	      tmp.exop = 128 + 64;
	      tmp.base = 0;
	      }
	    else if (*p < s) {
	      /* 256 is end-of-block: */
	      tmp.exop = *p < 256 ? 0 : 32 + 64;
	      /* simple code is just the value: */
	      tmp.base = *p;
	      p++;
	      }
	    else {
	      /* non-simple--look up in lists: */
	      tmp.exop = e[*p - s] + 16 + 64;
	      tmp.base = d[*p - s];
	      p++;
	      }

	  /* fill code-like entries with tmp */
	  {
	  /* current_code repeats in table every (1 << tmp.bits) entries: */
	  unsigned short j;

	  for (j = current_code >> bits_before_this_table;
	       j < nb_entry_current_table;
	       j += (1 << tmp.bits))
	      table_stack[table_level][j] = tmp;
	  }
	  }

	  /* backwards increment the k-bit code current_code */
	  {
	  unsigned short j = 1 << current_code_length;
	  for (;;) {
	      unsigned short old_code = current_code;

	      j >>= 1;
	      current_code ^= j;
	      if ((old_code & j) == 0)
		  break;
	      }
	  }

	  /* backup over finished tables */
	  while ((current_code & ((1U << bits_before_this_table) - 1))
		 != x[table_level]) {
	      table_level--;
	      bits_before_this_table -= result->need;
	      }
	  }
      }
  }
  return returned;
  }

/*
 * At first entry, z->mode == INIT == 0 and all other
 * fields have to be null, but z->extra.
 * To abort, set "z->mode = BAD" and call inflate().
 */
SEGMENT_EXTRA(inflate) zret_t
inflate (z_stream *z)
  {
  /* it would take too much code to generate cplext[] and cpdext[] on line */
  static const unsigned short cplens[31] = {
      3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
      35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258, 0, 0
      };
  static const unsigned char cplext[31] = {
      0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2,
      3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 0, 112, 112
      };

  static const unsigned short cpdist[30] = {
      1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193,
      257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145,
      8193, 12289, 16385, 24577
      };
  static const unsigned char cpdext[30] = {
      0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,
      7, 7, 8, 8, 9, 9, 10, 10, 11, 11,
      12, 12, 13, 13
      };
#ifdef GZLIB_USE_INTERMEDIATE_WINDOW
  unsigned short m;
#endif

  if (z == 0)
      return Z_STREAM_ERROR;

#ifdef GZLIB_USE_INTERMEDIATE_WINDOW
  m = (1U << MAX_WBITS) - z->write;
#endif

  for (;;) {
      unsigned short utmp;
      inflate_huft *htmp;
      static const signed char nbbitarray[] = {
	[INIT]		= 0,
	[SIGNATURE]	= 16,
	[METHOD]	= 8,
	[FLAGS]		= 8,
	[MODIF_TIME1]	= 8,
	[MODIF_TIME2]	= 8,
	[MODIF_TIME3]	= 8,
	[MODIF_TIME4]	= 8,
	[EXTRAFLAGS]	= 8,
	[OPERATING_SYSTEM]	= 8,
	[EXTRAFIELDLEN]	= 16,
	[EXTRAFIELD]	= 8,
	[ORIGNAME]	= 8,
	[COMMENT]	= 8,
	[HEADERCRC]	= 16,
	[TYPE]		= 3,
	[STORED_LENS]	= 16,
	[STORED_NOTLENS]= 16,
	[STORED_COPY]	= 0,
	[TABLE]		= 14,
	[BTREE]		= 3,
	[DTREE1]	= -1,
	[DTREE2]		= -2,
	[DTREE3]	= 0,
	[CODES_START]	= 0,
	[CODES_LEN]	= -1,
	[CODES_LENEXT]		= -2,
	[CODES_DIST]	= -1,
	[CODES_DISTEXT]		= -2,
	[CODES_COPY]	= 0,
	[DONE]		= 0,
	[CRCCHK]	= 16,
	[BAD]		= 0
	  };

      {
      unsigned short nb;
      unsigned short read;
#if 0
      const char *statename[] = {
	  "INIT", "SIGNATURE", "METHOD", "FLAGS",
	  "MODIF_TIME1", "MODIF_TIME2", "MODIF_TIME3", "MODIF_TIME4",
	  "EXTRAFLAGS", "OPERATING_SYSTEM",
	  "EXTRAFIELDLEN", "EXTRAFIELD", "ORIGNAME", "COMMENT", "HEADERCRC",
	  "TYPE",
	  "STORED_LENS", "STORED_NOTLENS", "STORED_COPY",
	  "TABLE", "BTREE", "DTREE1", "DTREE2", "DTREE3",
	  "CODES_START", "CODES_LEN", "CODES_LENEXT", "CODES_DIST",
	  "CODES_DISTEXT", "CODES_COPY",
	  "DONE", "CRCCHK",
	  "BAD"
	  };
#define PRINTSTATE
#endif

      if (nbbitarray[z->mode] == -1)
	  nb = z->sub.cdtr.curhuft.need;
	else if (nbbitarray[z->mode] == -2)
	  nb = z->sub.cdtr.sub.copy.get; /* or sub.trees.nbbit */
	else
	  nb = nbbitarray[z->mode];

//if (nb != 0) {
      if (readbits (z, nb))
	  return Z_OK;
      read = maskbits (z, nb);
      if (nbbitarray[z->mode] == -1) {
	  htmp = z->sub.cdtr.curhuft.tree + read;
	  nb = htmp->bits;
	  utmp = 0; /* I can not leave a warning */
#ifdef PRINTSTATE
	  printf ("mode %s, htmp->bits %u, htmp->exop = 0x%X\n",
			statename[z->mode], htmp->bits, htmp->exop);
#endif
	  }
	else {
	  utmp = read;
	  htmp = 0; /* I can not leave a warning */
#ifdef PRINTSTATE
	  printf ("mode %s, utmp = 0x%X\n", statename[z->mode], utmp);
#endif
	  }

      flushbits (z, nb);
//}
      }

      switch (z->mode) {
	  case INIT:
#if defined (GZLIB_USE_INTERMEDIATE_WINDOW)
	      if ((z->window = ZWINALLOC (1U << MAX_WBITS)) == 0)
		  return Z_MEM_ERROR;
#endif
#if defined (GZLIB_USE_INTERMEDIATE_WINDOW) \
 || defined (GZLIB_USE_PRECOMPILED_CRC_ARRAY)
	      z->crc32 = crc32 (z->crc32, 0, 0);
#endif
	      z->mode++;	/* = SIGNATURE */
	      continue;
	  case SIGNATURE:
	      if (utmp != 0x8B1F) {
		  ZDBG (("invalid Gzip signature"));
		  z->mode = BAD;
		  continue;
		  }
	      z->mode++;	/* = METHOD */
	      continue;
	  case METHOD:
	      if (utmp != Z_DEFLATED) {
		  ZDBG (("unknown method"));
		  z->mode = BAD;
		  continue;
		  }
	      z->mode++;	/* = FLAGS */
	      continue;
	  case FLAGS:
	      z->sub.gzip_header.sub.data = utmp;
	      if (z->sub.gzip_header.sub.flags.reserved) {
		  ZDBG (("unknown flags"));
		  z->mode = BAD;
		  continue;
		  }
	      z->mode++;	/* = MODIF_TIME1 */
	      continue;
	  case MODIF_TIME1:
	  case MODIF_TIME2:
	  case MODIF_TIME3:
	  case MODIF_TIME4:
	  case EXTRAFLAGS:
	      z->mode++;	/* = next */
	      continue;
	  case OPERATING_SYSTEM:
	      z->sub.gzip_header.len = 0;	/* for ORIGNAME or COMMENT */
	      if (z->sub.gzip_header.sub.flags.extra_field)
		  z->mode++;	/* = EXTRAFIELDLEN */
		else if (z->sub.gzip_header.sub.flags.orig_name)
		  z->mode = ORIGNAME;
		else if (z->sub.gzip_header.sub.flags.comment)
		  z->mode = COMMENT;
		else if (z->sub.gzip_header.sub.flags.header_crc)
		  z->mode = HEADERCRC;
		else
		  z->mode = TYPE;
	      continue;
	  case EXTRAFIELDLEN:
	      z->sub.gzip_header.len = utmp;
	      if (z->sub.gzip_header.len) /* Is that needed? */
		  z->mode++;	/* = EXTRAFIELD */
		else if (z->sub.gzip_header.sub.flags.orig_name)
		  z->mode = ORIGNAME;
		else if (z->sub.gzip_header.sub.flags.comment)
		  z->mode = COMMENT;
		else if (z->sub.gzip_header.sub.flags.header_crc)
		  z->mode = HEADERCRC;
		else
		  z->mode = TYPE;
	      continue;
	  case EXTRAFIELD:
	      if (-- z->sub.gzip_header.len)
		  continue;
	      z->sub.gzip_header.len = 0;	/* for ORIGNAME or COMMENT */
	      if (z->sub.gzip_header.sub.flags.orig_name)
		  z->mode++;	/* = ORIGNAME */
		else if (z->sub.gzip_header.sub.flags.comment)
		  z->mode = COMMENT;
		else if (z->sub.gzip_header.sub.flags.header_crc)
		  z->mode = HEADERCRC;
		else
		  z->mode = TYPE;
	      continue;
	  case ORIGNAME:
	      if (z->sub.gzip_header.len <= GZLIB_MAXBUF) {
		  if (z->sub.gzip_header.len == GZLIB_MAXBUF)
		      z->sub.gzip_header.buffer[GZLIB_MAXBUF] = 0;
		    else
		      z->sub.gzip_header.buffer[z->sub.gzip_header.len]	= utmp;
		  }
	      z->sub.gzip_header.len++;
	      if (utmp != 0)
		  continue;
	      ZDBG (("%s: gzip name: %s\r\n",
			__FUNCTION__, z->sub.gzip_header.buffer));
#ifdef GZLIB_TREAT_NAME
	      GZLIB_TREAT_NAME (z->extra, z->sub.gzip_header.buffer);
#endif
	      z->sub.gzip_header.len = 0;	/* for COMMENT */
	      if (z->sub.gzip_header.sub.flags.comment)
		  z->mode++; /* = COMMENT */
		else if (z->sub.gzip_header.sub.flags.header_crc)
		  z->mode = HEADERCRC;
		else
		  z->mode = TYPE;
	      continue;
	  case COMMENT:
	      if (z->sub.gzip_header.len <= GZLIB_MAXBUF) {
		  if (z->sub.gzip_header.len == GZLIB_MAXBUF)
		      z->sub.gzip_header.buffer[GZLIB_MAXBUF] = 0;
		    else
		      z->sub.gzip_header.buffer[z->sub.gzip_header.len]	= utmp;
		  }
	      z->sub.gzip_header.len++;
	      if (utmp != 0)
		  continue;
	      ZDBG (("%s: gzip comment: %s\r\n",
			__FUNCTION__, z->sub.gzip_header.buffer));
#ifdef GZLIB_TREAT_COMMENT
	      GZLIB_TREAT_COMMENT (z->extra, z->sub.gzip_header.buffer);
#endif
	      if (z->sub.gzip_header.sub.flags.header_crc)
		  z->mode++; /* = HEADERCRC */
		else
		  z->mode = TYPE;
	      continue;
	  case HEADERCRC:
	      z->mode++;	/* = TYPE */
	      continue;
	  case TYPE:
	      z->last = utmp & 1;
	      utmp >>= 1;
	      if (utmp == 3) {
		  ZDBG (("invalid block type"));
		  z->mode = BAD;
		  }
		else if (utmp == 2) {
		  z->mode = TABLE;
		  }
		else if (utmp == 1) { /* inflate_trees_fixed() */
		  int k;	/* temporary variable */
		  /* number of hufts used in fixed_mem: */
		  unsigned short f = 0;
		  /* length list for huft_build: */
		  unsigned short c[288];

		  /* literal table */
		  for (k = 0; k < 144; k++)
		      c[k] = 8;
		  for (; k < 256; k++)
		      c[k] = 9;
		  for (; k < 280; k++)
		      c[k] = 7;
		  for (; k < 288; k++)
		      c[k] = 8;
		  z->sub.cdtr.lbitstree.need = 9;
		  huft_build (c,
			      288,
			      257,
			      cplens,
			      cplext,
			      &z->sub.cdtr.lbitstree,
			      z->hufts,
			      &f);
#if 0
for (k = 0; k < 512; k++) {
    static const inflate_huft fixed_tl[512] = {
	{96,7,256},	{0,8,80},	{0,8,16},	{84,8,115},
	{82,7,31},	{0,8,112},	{0,8,48},	{0,9,192},
	{80,7,10},	{0,8,96},	{0,8,32},	{0,9,160},
	{0,8,0},	{0,8,128},	{0,8,64},	{0,9,224},
	{80,7,6},	{0,8,88},	{0,8,24},	{0,9,144},
	{83,7,59},	{0,8,120},	{0,8,56},	{0,9,208},
	{81,7,17},	{0,8,104},	{0,8,40},	{0,9,176},
	{0,8,8},	{0,8,136},	{0,8,72},	{0,9,240},
	{80,7,4},	{0,8,84},	{0,8,20},	{85,8,227},
	{83,7,43},	{0,8,116},	{0,8,52},	{0,9,200},
	{81,7,13},	{0,8,100},	{0,8,36},	{0,9,168},
	{0,8,4},	{0,8,132},	{0,8,68},	{0,9,232},
	{80,7,8},	{0,8,92},	{0,8,28},	{0,9,152},
	{84,7,83},	{0,8,124},	{0,8,60},	{0,9,216},
	{82,7,23},	{0,8,108},	{0,8,44},	{0,9,184},
	{0,8,12},	{0,8,140},	{0,8,76},	{0,9,248},
	{80,7,3},	{0,8,82},	{0,8,18},	{85,8,163},
	{83,7,35},	{0,8,114},	{0,8,50},	{0,9,196},
	{81,7,11},	{0,8,98},	{0,8,34},	{0,9,164},
	{0,8,2},	{0,8,130},	{0,8,66},	{0,9,228},
	{80,7,7},	{0,8,90},	{0,8,26},	{0,9,148},
	{84,7,67},	{0,8,122},	{0,8,58},	{0,9,212},
	{82,7,19},	{0,8,106},	{0,8,42},	{0,9,180},
	{0,8,10},	{0,8,138},	{0,8,74},	{0,9,244},
	{80,7,5},	{0,8,86},	{0,8,22},	{192,8,0},
	{83,7,51},	{0,8,118},	{0,8,54},	{0,9,204},
	{81,7,15},	{0,8,102},	{0,8,38},	{0,9,172},
	{0,8,6},	{0,8,134},	{0,8,70},	{0,9,236},
	{80,7,9},	{0,8,94},	{0,8,30},	{0,9,156},
	{84,7,99},	{0,8,126},	{0,8,62},	{0,9,220},
	{82,7,27},	{0,8,110},	{0,8,46},	{0,9,188},
	{0,8,14},	{0,8,142},	{0,8,78},	{0,9,252},
	{96,7,256},	{0,8,81},	{0,8,17},	{85,8,131},
	{82,7,31},	{0,8,113},	{0,8,49},	{0,9,194},
	{80,7,10},	{0,8,97},	{0,8,33},	{0,9,162},
	{0,8,1},	{0,8,129},	{0,8,65},	{0,9,226},
	{80,7,6},	{0,8,89},	{0,8,25},	{0,9,146},
	{83,7,59},	{0,8,121},	{0,8,57},	{0,9,210},
	{81,7,17},	{0,8,105},	{0,8,41},	{0,9,178},
	{0,8,9},	{0,8,137},	{0,8,73},	{0,9,242},
	{80,7,4},	{0,8,85},	{0,8,21},	{80,8,258},
	{83,7,43},	{0,8,117},	{0,8,53},	{0,9,202},
	{81,7,13},	{0,8,101},	{0,8,37},	{0,9,170},
	{0,8,5},	{0,8,133},	{0,8,69},	{0,9,234},
	{80,7,8},	{0,8,93},	{0,8,29},	{0,9,154},
	{84,7,83},	{0,8,125},	{0,8,61},	{0,9,218},
	{82,7,23},	{0,8,109},	{0,8,45},	{0,9,186},
	{0,8,13},	{0,8,141},	{0,8,77},	{0,9,250},
	{80,7,3},	{0,8,83},	{0,8,19},	{85,8,195},
	{83,7,35},	{0,8,115},	{0,8,51},	{0,9,198},
	{81,7,11},	{0,8,99},	{0,8,35},	{0,9,166},
	{0,8,3},	{0,8,131},	{0,8,67},	{0,9,230},
	{80,7,7},	{0,8,91},	{0,8,27},	{0,9,150},
	{84,7,67},	{0,8,123},	{0,8,59},	{0,9,214},
	{82,7,19},	{0,8,107},	{0,8,43},	{0,9,182},
	{0,8,11},	{0,8,139},	{0,8,75},	{0,9,246},
	{80,7,5},	{0,8,87},	{0,8,23},	{192,8,0},
	{83,7,51},	{0,8,119},	{0,8,55},	{0,9,206},
	{81,7,15},	{0,8,103},	{0,8,39},	{0,9,174},
	{0,8,7},	{0,8,135},	{0,8,71},	{0,9,238},
	{80,7,9},	{0,8,95},	{0,8,31},	{0,9,158},
	{84,7,99},	{0,8,127},	{0,8,63},	{0,9,222},
	{82,7,27},	{0,8,111},	{0,8,47},	{0,9,190},
	{0,8,15},	{0,8,143},	{0,8,79},	{0,9,254},
	{96,7,256},	{0,8,80},	{0,8,16},	{84,8,115},
	{82,7,31},	{0,8,112},	{0,8,48},	{0,9,193},
	{80,7,10},	{0,8,96},	{0,8,32},	{0,9,161},
	{0,8,0},	{0,8,128},	{0,8,64},	{0,9,225},
	{80,7,6},	{0,8,88},	{0,8,24},	{0,9,145},
	{83,7,59},	{0,8,120},	{0,8,56},	{0,9,209},
	{81,7,17},	{0,8,104},	{0,8,40},	{0,9,177},
	{0,8,8},	{0,8,136},	{0,8,72},	{0,9,241},
	{80,7,4},	{0,8,84},	{0,8,20},	{85,8,227},
	{83,7,43},	{0,8,116},	{0,8,52},	{0,9,201},
	{81,7,13},	{0,8,100},	{0,8,36},	{0,9,169},
	{0,8,4},	{0,8,132},	{0,8,68},	{0,9,233},
	{80,7,8},	{0,8,92},	{0,8,28},	{0,9,153},
	{84,7,83},	{0,8,124},	{0,8,60},	{0,9,217},
	{82,7,23},	{0,8,108},	{0,8,44},	{0,9,185},
	{0,8,12},	{0,8,140},	{0,8,76},	{0,9,249},
	{80,7,3},	{0,8,82},	{0,8,18},	{85,8,163},
	{83,7,35},	{0,8,114},	{0,8,50},	{0,9,197},
	{81,7,11},	{0,8,98},	{0,8,34},	{0,9,165},
	{0,8,2},	{0,8,130},	{0,8,66},	{0,9,229},
	{80,7,7},	{0,8,90},	{0,8,26},	{0,9,149},
	{84,7,67},	{0,8,122},	{0,8,58},	{0,9,213},
	{82,7,19},	{0,8,106},	{0,8,42},	{0,9,181},
	{0,8,10},	{0,8,138},	{0,8,74},	{0,9,245},
	{80,7,5},	{0,8,86},	{0,8,22},	{192,8,0},
	{83,7,51},	{0,8,118},	{0,8,54},	{0,9,205},
	{81,7,15},	{0,8,102},	{0,8,38},	{0,9,173},
	{0,8,6},	{0,8,134},	{0,8,70},	{0,9,237},
	{80,7,9},	{0,8,94},	{0,8,30},	{0,9,157},
	{84,7,99},	{0,8,126},	{0,8,62},	{0,9,221},
	{82,7,27},	{0,8,110},	{0,8,46},	{0,9,189},
	{0,8,14},	{0,8,142},	{0,8,78},	{0,9,253},
	{96,7,256},	{0,8,81},	{0,8,17},	{85,8,131},
	{82,7,31},	{0,8,113},	{0,8,49},	{0,9,195},
	{80,7,10},	{0,8,97},	{0,8,33},	{0,9,163},
	{0,8,1},	{0,8,129},	{0,8,65},	{0,9,227},
	{80,7,6},	{0,8,89},	{0,8,25},	{0,9,147},
	{83,7,59},	{0,8,121},	{0,8,57},	{0,9,211},
	{81,7,17},	{0,8,105},	{0,8,41},	{0,9,179},
	{0,8,9},	{0,8,137},	{0,8,73},	{0,9,243},
	{80,7,4},	{0,8,85},	{0,8,21},	{80,8,258},
	{83,7,43},	{0,8,117},	{0,8,53},	{0,9,203},
	{81,7,13},	{0,8,101},	{0,8,37},	{0,9,171},
	{0,8,5},	{0,8,133},	{0,8,69},	{0,9,235},
	{80,7,8},	{0,8,93},	{0,8,29},	{0,9,155},
	{84,7,83},	{0,8,125},	{0,8,61},	{0,9,219},
	{82,7,23},	{0,8,109},	{0,8,45},	{0,9,187},
	{0,8,13},	{0,8,141},	{0,8,77},	{0,9,251},
	{80,7,3},	{0,8,83},	{0,8,19},	{85,8,195},
	{83,7,35},	{0,8,115},	{0,8,51},	{0,9,199},
	{81,7,11},	{0,8,99},	{0,8,35},	{0,9,167},
	{0,8,3},	{0,8,131},	{0,8,67},	{0,9,231},
	{80,7,7},	{0,8,91},	{0,8,27},	{0,9,151},
	{84,7,67},	{0,8,123},	{0,8,59},	{0,9,215},
	{82,7,19},	{0,8,107},	{0,8,43},	{0,9,183},
	{0,8,11},	{0,8,139},	{0,8,75},	{0,9,247},
	{80,7,5},	{0,8,87},	{0,8,23},	{192,8,0},
	{83,7,51},	{0,8,119},	{0,8,55},	{0,9,207},
	{81,7,15},	{0,8,103},	{0,8,39},	{0,9,175},
	{0,8,7},	{0,8,135},	{0,8,71},	{0,9,239},
	{80,7,9},	{0,8,95},	{0,8,31},	{0,9,159},
	{84,7,99},	{0,8,127},	{0,8,63},	{0,9,223},
	{82,7,27},	{0,8,111},	{0,8,47},	{0,9,191},
	{0,8,15},	{0,8,143},	{0,8,79},	{0,9,255}
	};
    inflate_huft *hufts = z->sub.cdtr.lbitstree.tree;
    if (z->sub.cdtr.lbitstree.need != 9)
	printf ("z->sub.cdtr.lbitstree.need = %u != 9\n",
		 z->sub.cdtr.lbitstree.need);
    if (   hufts[k].exop != fixed_tl[k].exop
	|| hufts[k].bits != fixed_tl[k].bits
	|| hufts[k].base != fixed_tl[k].base) {
	printf ("hufts[%u] = {%u, %u, %u} != fixed_tl[%u] = {%u, %u, %u}\n",
		k, hufts[k].exop, hufts[k].bits, hufts[k].base,
		k, fixed_tl[k].exop, fixed_tl[k].bits, fixed_tl[k].base);
	}
    }
#endif

		  /* distance table */
		  for (k = 0; k < 30; k++)
		      c[k] = 5;
		  z->sub.cdtr.dbitstree.need = 5;
		  huft_build (c,
			      30,
			      0,
			      cpdist,
			      cpdext,
			      &z->sub.cdtr.dbitstree,
			      z->hufts,
			      &f);

#if 0
for (k = 0; k < 32; k++) {
    static const inflate_huft fixed_td[32] = {
	{80,5,1},	{87,5,257},	{83,5,17},	{91,5,4097},
	{81,5,5},	{89,5,1025},	{85,5,65},	{93,5,16385},
	{80,5,3},	{88,5,513},	{84,5,33},	{92,5,8193},
	{82,5,9},	{90,5,2049},	{86,5,129},	{192,5,24577},
	{80,5,2},	{87,5,385},	{83,5,25},	{91,5,6145},
	{81,5,7},	{89,5,1537},	{85,5,97},	{93,5,24577},
	{80,5,4},	{88,5,769},	{84,5,49},	{92,5,12289},
	{82,5,13},	{90,5,3073},	{86,5,193},	{192,5,24577}
	};
    inflate_huft *hufts = z->sub.cdtr.dbitstree.tree;
    if (z->sub.cdtr.dbitstree.need != 5)
	printf ("z->sub.cdtr.dbitstree.need = %u != 9\n",
	 z->sub.cdtr.dbitstree.need);
    if (   hufts[k].exop != fixed_td[k].exop
	|| hufts[k].bits != fixed_td[k].bits
	|| hufts[k].base != fixed_td[k].base) {
	printf ("hufts[%u] = {%u, %u, %u} != fixed_td[%u] = {%u, %u, %u}\n",
		k, hufts[k].exop, hufts[k].bits, hufts[k].base,
		k, fixed_td[k].exop, fixed_td[k].bits, fixed_td[k].base);
	}
    hufts[k] = fixed_td[k];
    }
#endif

		  z->mode = CODES_START;
		  }
		else /* if (utmp == 0) */ {
		  flushbits (z, z->bitk & 7);
		  z->mode++;	/* = STORED_LENS */
		  }
	      continue;
	  case STORED_LENS:
	      z->sub.stored.left = utmp;
	      z->mode++;	/* = STORED_NOTLENS */
	      continue;
	  case STORED_NOTLENS:
	      if (z->sub.stored.left != (utmp ^ 0xFFFF)) {
		  ZDBG (("invalid stored block lengths"));
		  z->mode = BAD;
		  continue;
		  }
	      z->bitb = 0;
	      z->bitk = 0;
	      if (z->sub.stored.left == 0)
		  z->mode = z->last ? DONE : TYPE;
		else
		  z->mode++;	/* = STORED_COPY */
	      continue;
	  case STORED_COPY:
	      if (z->avail_in == 0)
		  return Z_OK;
	      if (z->sub.stored.left > z->avail_in)
		  utmp = z->avail_in;
		else
		  utmp = z->sub.stored.left;
#ifdef GZLIB_USE_INTERMEDIATE_WINDOW
	      if (utmp > m)
		    utmp = m;
	      memcpy_data2window (&z->window[z->write], z->next_in, utmp);
	      z->write += utmp;
	      m -= utmp;
#else
	      if (z->avail_out < utmp) {
		  ZDBG (("%s: not enought memory to load kernel!\r\n",
			__FUNCTION__));
		  return Z_OUPUT_FULL;
		  }
	      memcpy_door_open (z->next_out, z->next_in, utmp);
	      z->next_out += utmp;
	      z->avail_out -= utmp;
	      z->total_out += utmp;
#endif
	      z->next_in += utmp;
	      z->avail_in -= utmp;
	      z->sub.stored.left -= utmp;

	      if (z->sub.stored.left == 0)
		  z->mode = z->last ? DONE : TYPE;
#ifdef GZLIB_USE_INTERMEDIATE_WINDOW
	      break;
#else
	      continue;
#endif


	  case TABLE:
	      z->sub.cdtr.sub.trees.table = utmp;

#ifndef PKZIP_BUG_WORKAROUND
	      if (   (z->sub.cdtr.sub.trees.table & 0x1f) > 29
		  || ((z->sub.cdtr.sub.trees.table >> 5) & 0x1f) > 29) {
		  ZDBG (("too many length or distance symbols"));
		  z->mode = BAD;
		  continue;
		  }
#endif

	      z->sub.cdtr.len = 258 + (z->sub.cdtr.sub.trees.table & 0x1f)
				+ ((z->sub.cdtr.sub.trees.table >> 5) & 0x1f);
	      z->sub.cdtr.sub.trees.index = 0;
	      z->mode++;	/* = BTREE */
	      continue;
	  case BTREE: {
	      static const unsigned char border[] = {
		  16, 17, 18,  0,  8,  7,  9,  6, 10,  5,
		  11,  4, 12,  3, 13,  2, 14,  1, 15, 32,
		  71, 80, 76, 32, 69, 46, 76, 79, 82, 82,
		  65, 73, 78, 46,  0,  1,  3, 67,  0
		  };

	      z->sub.cdtr.blens[border[z->sub.cdtr.sub.trees.index++]] = utmp;
	      if (z->sub.cdtr.sub.trees.index < 4 +
					(z->sub.cdtr.sub.trees.table >> 10))
		  continue;
	      while (z->sub.cdtr.sub.trees.index < sizeof (border) - 20)
		  z->sub.cdtr.blens[border[z->sub.cdtr.sub.trees.index++]] = 0;
	      z->sub.cdtr.curhuft.need = 7;


	      { /* inflate_trees_bits() */
	      unsigned short hn = 0;
	      int r;

	      r = huft_build (z->sub.cdtr.blens,
			      sizeof (border) - 20,
			      sizeof (border) - 20,
			      (unsigned short *) 0,
			      (unsigned char *) 0,
			      &z->sub.cdtr.curhuft,
			      z->hufts,
			      &hn);
	      if (r == Z_DATA_ERROR)
		  ZDBG (("oversubscribed dynamic bit lengths tree"));
		else if (r == Z_BUF_ERROR || z->sub.cdtr.curhuft.need == 0) {
		  ZDBG (("incomplete dynamic bit lengths tree"));
		  r = Z_DATA_ERROR;
		  }

	      if (r != Z_OK) {
		  z->mode = BAD;
		  continue;
		  }
	      }

	      z->sub.cdtr.sub.trees.index = 0;
	      z->mode++;	/* = DTREE1 */
	      }
	      continue;

	  case DTREE1:
	      if (htmp->base < 16) {
		  z->sub.cdtr.blens[z->sub.cdtr.sub.trees.index++] = htmp->base;
		  if (z->sub.cdtr.sub.trees.index >= z->sub.cdtr.len)
		      z->mode = DTREE3;
		  }
		else /* htmp->base == 16..18 */ {
		  if (htmp->base == 18)
		      z->sub.cdtr.sub.trees.nbbit = 7;
		    else
		      z->sub.cdtr.sub.trees.nbbit = htmp->base - 14;
		  z->mode++; /* = DTREE2 */
		  }
	      continue;
	  case DTREE2:
	      if (z->sub.cdtr.sub.trees.nbbit == 7)
		  utmp += 11;
		else
		  utmp += 3;

	      if (   z->sub.cdtr.sub.trees.index + utmp > z->sub.cdtr.len
		  || (   z->sub.cdtr.sub.trees.nbbit == 2
		      && z->sub.cdtr.sub.trees.index < 1)) {
		  ZDBG (("invalid bit length repeat"));
		  z->mode = BAD;
		  continue;
		  }
	      {
	      unsigned short tmp;

	      if (z->sub.cdtr.sub.trees.nbbit == 2)
		  tmp = z->sub.cdtr.blens[z->sub.cdtr.sub.trees.index - 1];
		else
		  tmp = 0;

	      do {
		  z->sub.cdtr.blens[z->sub.cdtr.sub.trees.index++] = tmp;
		  } while (--utmp);
	      }

	      if (z->sub.cdtr.sub.trees.index >= z->sub.cdtr.len)
		  z->mode++; /* = DTREE3 */
		else {
		  z->mode = DTREE1;
		  continue;
		  }
	  case DTREE3:
	      z->sub.cdtr.curhuft.need = 0;

	      /* must be <= 9 for lookahead assumptions: */
	      z->sub.cdtr.lbitstree.need = 9;
	      z->sub.cdtr.dbitstree.need = 6;

	      { /* inflate_trees_dynamic() */
	      unsigned short hn = 0;
	      unsigned short nl = 257 + (z->sub.cdtr.sub.trees.table & 0x1f);
	      int r;

	      /* build literal/length tree */
	      r = huft_build (z->sub.cdtr.blens,
			      nl,
			      257,
			      cplens,
			      cplext,
			      &z->sub.cdtr.lbitstree,
			      z->hufts,
			      &hn);

	      if (r != Z_OK || z->sub.cdtr.lbitstree.need == 0) {
		  if (r == Z_DATA_ERROR)
		      ZDBG (("oversubscribed literal/length tree"));
		    else if (r != Z_MEM_ERROR) {
		      ZDBG (("incomplete literal/length tree"));
		      r = Z_DATA_ERROR;
		      }
		  }

	      if (r == Z_OK) {
		  /* build distance tree */
		  r = huft_build (z->sub.cdtr.blens + nl,
				1 + ((z->sub.cdtr.sub.trees.table >> 5) & 0x1f),
				0,
				cpdist,
				cpdext,
				&z->sub.cdtr.dbitstree,
				z->hufts,
				&hn);
		  if (   r != 0
		      || (   z->sub.cdtr.dbitstree.need == 0
			  && (z->sub.cdtr.sub.trees.table & 0x1f) != 0)) {
		      if (r == Z_DATA_ERROR)
			  ZDBG (("oversubscribed distance tree"));
			else if (r == Z_BUF_ERROR) {
#ifdef PKZIP_BUG_WORKAROUND
			  r = Z_OK;
			  }
#else
			  ZDBG (("incomplete distance tree"));
			  r = Z_DATA_ERROR;
			  }
			else if (r != Z_MEM_ERROR) {
			  ZDBG (("empty distance tree with lengths"));
			  r = Z_DATA_ERROR;
			  }
#endif
		      }

		  }

	      if (r != Z_OK) {
		  z->mode = BAD;
		  continue;
		  }
	      }
	      z->mode++;	/* = CODES_START */
	      continue;

	  case CODES_START:
	      z->sub.cdtr.curhuft = z->sub.cdtr.lbitstree;
	      z->mode++;	/* = CODES_LEN */
	      continue;
	  case CODES_LEN:
	      if (htmp->exop == 0) {
#ifdef GZLIB_USE_INTERMEDIATE_WINDOW
		  /* CODES_LIT: (needs m != 0) */
		  OUTBYTE (&z->window[z->write], htmp->base);
		  z->write += 1;
		  m--;
		  z->mode = CODES_START;
		  break;
#else
		  if (z->avail_out == 0) {
		      ZDBG (("%s: not enought memory to load kernel!\r\n",
				__FUNCTION__));
		      return Z_OUPUT_FULL;
		      }
		  *(z->next_out++) = htmp->base;
		  z->avail_out --;
		  z->total_out ++;
		  z->mode = CODES_START;
#endif
		  }
		else if (htmp->exop & 16) {
		  z->sub.cdtr.sub.copy.get = htmp->exop & 15;
		  z->sub.cdtr.len = htmp->base;
		  z->mode++;	/* = CODES_LENEXT */
		  }
		else if ((htmp->exop & 64) == 0) {
		  z->sub.cdtr.curhuft.need = htmp->exop;
		  z->sub.cdtr.curhuft.tree = htmp + htmp->base;
		  }
		else if (htmp->exop & 32) {
		  /* CODES_END: */
		  if (z->bitk > 7) {
		      z->bitk -= 8;
		      z->avail_in++;
		      z->next_in--;
		      }
		  /* usual way to finish */
		  z->mode = z->last ? DONE : TYPE;
		  }
		else {
		  ZDBG (("invalid literal/length code: 0x%X\r\n", htmp->exop));
		  z->mode = BAD;
		  }
	      continue;
	  case CODES_LENEXT:
	      z->sub.cdtr.len += utmp;
	      z->sub.cdtr.curhuft = z->sub.cdtr.dbitstree;
	      z->mode++;	/* = CODES_DIST */
	      continue;
	  case CODES_DIST:
	      if (htmp->exop & 16) {
		  z->sub.cdtr.sub.copy.get = htmp->exop & 15;
		  z->sub.cdtr.sub.copy.dist = htmp->base;
		  z->mode++;	/* = CODES_DISTEXT */
		  }
		else if ((htmp->exop & 64) == 0) {
		  z->sub.cdtr.curhuft.need = htmp->exop;
		  z->sub.cdtr.curhuft.tree = htmp + htmp->base;
		  }
		else {
		  ZDBG (("invalid distance code"));
		  z->mode = BAD;
		  }
	      continue;
	  case CODES_DISTEXT:
	      z->sub.cdtr.sub.copy.dist += utmp;
	      z->mode++;	/* = CODES_COPY */
	      continue;
	  case CODES_COPY:
#ifdef GZLIB_USE_INTERMEDIATE_WINDOW
	      {
		inline void copy_codes (z_stream *z,
					unsigned short index,
					unsigned short *nb)
		  {
		  unsigned short tmp = ((1U << MAX_WBITS) + z->write - index)
					% (1U << MAX_WBITS);
		  if (*nb > (1U << MAX_WBITS) - tmp)
		      *nb = (1U << MAX_WBITS) - tmp;
		  memcpy_window2window (&z->window[z->write],
					&z->window[tmp],
					*nb);
		  }

	      if (z->sub.cdtr.len > m)
		  utmp = m;
		else
		  utmp = z->sub.cdtr.len;
	      copy_codes (z, z->sub.cdtr.sub.copy.dist, &utmp);
	      z->write += utmp;
	      m -= utmp;
	      z->sub.cdtr.len -= utmp;

	      if (z->sub.cdtr.len == 0)
		  z->mode = CODES_START;
	      }
	      break;
#else
	      if (z->avail_out < z->sub.cdtr.len) {
		  ZDBG (("%s: not enought memory to load kernel!\r\n",
			__FUNCTION__));
		  return Z_OUPUT_FULL;
		  }
	      memcpy_door_open (z->next_out,
				z->next_out - z->sub.cdtr.sub.copy.dist,
				z->sub.cdtr.len);
	      z->next_out += z->sub.cdtr.len;
	      z->avail_out -= z->sub.cdtr.len;
	      z->total_out += z->sub.cdtr.len;
	      z->mode = CODES_START;
	      continue;
#endif


	  case DONE:
#ifdef GZLIB_USE_INTERMEDIATE_WINDOW
	      if (z->write != 0) {
		  m = 0; /* force inflate_flush() */
		  break;
		  }
#else
	      /* This has the advantage to do a basic test of the RAM
		 and checks that A20 is open... but need access to
		 all of the memory at once */
	      z->crc32 = crc32 (crc32 (0, 0, 0),
				z->next_out - z->total_out,
				z->total_out);
#endif
	      z->sub.crcchk.arrayindex = 0;
	      flushbits (z, z->bitk & 7);
	      z->mode++;	/* = CRCCHK */
	      continue;
	  case CRCCHK:
	      z->sub.crcchk.sub.array[z->sub.crcchk.arrayindex++] = utmp;
	      if (z->sub.crcchk.arrayindex < 4)
		  continue;
#ifdef GZLIB_USE_INTERMEDIATE_WINDOW
	      ZWINFREE (z->window);
#endif
	      if (z->sub.crcchk.sub.data.length != z->total_out) {
		  ZDBG (("%s: incorrect uncompressed length: %u != %u\r\n",
			__FUNCTION__,
			z->sub.crcchk.sub.data.length, z->total_out));
		  return Z_BAD_LENGTH;
		  }
	      if (z->sub.crcchk.sub.data.crc32 != z->crc32) {
		  ZDBG (("%s: incorrect crc32: 0x%X != 0x%X\r\n",
			__FUNCTION__,
			z->sub.crcchk.sub.data.crc32, z->crc32));
		  return Z_BAD_CRC;
		  }
	      ZDBG (("%s: uncompress finished successfully.\r\n",
			__FUNCTION__));
	      return Z_STREAM_END;
	  case BAD:
#ifdef GZLIB_USE_INTERMEDIATE_WINDOW
	      ZWINFREE (z->window);
#endif
	      ZDBG (("%s: state BAD !\r\n", __FUNCTION__));
	      return Z_DATA_ERROR;
	}

#ifdef GZLIB_USE_INTERMEDIATE_WINDOW
      if (m == 0) {
	  /* always flush the complete window, but last time: */
	  if (z->write > z->avail_out) {
	      ZDBG (("%s: not enought memory to load kernel!\r\n",
			__FUNCTION__));
	      ZWINFREE (z->window);
	      return Z_OUPUT_FULL;
	      }

	  utmp = treat_data (z, z->next_out, z->window, z->write);
	  if (utmp) {
	      ZDBG (("%s: memcpy error!\r\n", __FUNCTION__));
	      ZWINFREE (z->window);
	      return Z_BASE_MEMCOPY_ERROR + utmp;
	      }

	  z->crc32 = crc32 (z->crc32, z->window, z->write);
	  z->next_out += z->write;
	  z->avail_out -= z->write;
	  z->total_out += z->write;
	  z->write = 0;
	  m = 1U << MAX_WBITS;
	  }
#endif
      }
  }

