/* instboot.c */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <sys/types.h>		/* open() */
#include <sys/stat.h>
#include <fcntl.h>		/* open() */
#include <unistd.h>
#include <stdlib.h>
#include <string.h>		/* strncasecmp() */

#include <sys/ioctl.h>
#include <linux/fd.h>		/* struct floppy_struct */
#include <linux/hdreg.h>	/* struct hd_geometry */

#define INSTBOOT_C	/* do not include incompatible functions */
#include "make.h"
#include "library.h"
#include "boot.h"

typedef enum {
    bios_loader = 0,
    dos_com_loader = 1,
    dos_exe_loader = 2
    } loader_type;

/*
 * Functions:
 */

/*
 * For 1.68M floppy disk, do not try to defragment the floppy: It is
 * specially fragmented because most BIOSes cannot access over the 18th
 * sector, even if the floppy is described to have more. Some other BIOSes
 * will read them but cannot read more than 18 sectors at once.
 * Most such tools to make the file contigous on the magnetic support will
 * not try to touch files with attribute "readonly", "hidden" and "system",
 * but do not play (too much) with fire...
 * The 1.68 Mbytes floppy patch touches chs_chain(), the FAT and the file
 * written (gap added), two two last part are _not_ implemented.
 *
 * Note that if the BIOS driver cannot access over the 18th sector, Gujin
 * will not be able to access those files - it can only be accessed after
 * the Linux kernel has been loaded. That means neither vmlinux nor initrd
 * shall map a sector over sector 18: it is not so interresting.
 * Right now I "#ifdef 0" the chs_chain() patch and "#ifdef 1" a patch which
 * will work only if the BIOS cannot read more than 18 sectors at once,
 * but is able to read all the sectors.
 */
static unsigned
chs_chain (bootloader2_type *bootchain,
	   unsigned         lba,
	   unsigned         nb_sector,
	   unsigned	    dma_frontier,
	   unsigned char    dsk_nbhead,
	   unsigned char    dsk_nbsector,
	   unsigned char    dsk_biosnb,
	   unsigned         nb_sect_per_64K)
  {
  unsigned cpt = 0, tmp;
  unsigned end_lba = lba + nb_sector;
  unsigned start_lba = lba - 2;	/* boot sector + first sector already loaded */

#if 0
  if (dsk_nbhead == 2 && dsk_nbsector > 18)	/* extended floppy size */
      dsk_nbsector = 18;	/* assume that only locally */
#endif

  dma_frontier += lba;
  while (lba < end_lba) {
      unsigned short size = MIN (end_lba - lba,
				 dsk_nbsector - lba % dsk_nbsector);

      if (lba < dma_frontier && lba + size >= dma_frontier)
	  size = dma_frontier - lba;

      if (lba == dma_frontier) /* once every 64 Kb */
	  dma_frontier += nb_sect_per_64K;

      tmp = (lba - start_lba) / nb_sect_per_64K + 1;
      if (lba - start_lba + size > tmp * nb_sect_per_64K)
	  size = tmp * nb_sect_per_64K + start_lba - lba; /* %es ++ */

#if 1
      if (dsk_nbhead == 2 && dsk_nbsector > 18 && size > 18)
	  size = 18;
#endif

      /* Will need to "squale" those two later: */
      bootchain->header.nbword = size;
      bootchain->header.next = cpt + 1;

      bootchain->bootloader2_cmd.int1302.reg_dx
	= (lba / dsk_nbsector % dsk_nbhead) << 8 | dsk_biosnb;
      bootchain->bootloader2_cmd.int1302.reg_ax
	= 0x0200 | size;
      bootchain->bootloader2_cmd.int1302.reg_cx
	=  (lba / dsk_nbsector / dsk_nbhead << 8)
	 | (lba % dsk_nbsector + 1);

      bootchain->bootloader2_cmd.int1302.dummy = 0;
      bootchain->bootloader2_cmd.int1302.will_be_set_to_esbx = 0;

#if 0 /* debug */
      if (cpt == 0)
	  printf ("MBR sector at 0x7C00, first sector at 0x7E00\n");
      printf ("%s: %2u sector at C/H/S: %2u/%u/%-2u loaded at 0x%05X..0x%05X\n",
		__FUNCTION__,
		size,
		lba / dsk_nbsector / dsk_nbhead,
		lba / dsk_nbsector % dsk_nbhead,
		lba % dsk_nbsector + 1,
		(lba - start_lba) * 512 + 0x7C00,
		(lba - start_lba + size) * 512 + 0x7C00);
#endif
      lba += size;
      cpt++;
      bootchain++;
      }
  bootchain[-1].header.next = 0;
  return cpt;
  }

static void
chs_first (unsigned char *file,
	   unsigned      boot1param_off,
	   unsigned	 first_lba,
	   unsigned char dsk_nbhead,
	   unsigned char dsk_nbsector,
	   unsigned char dsk_biosnb)
  {
  boot1param_type *b1p = (boot1param_type *)(file + boot1param_off);

  b1p->bootloader2.bootloader2_cmd.int1302.reg_dx
	= ((first_lba / dsk_nbsector % dsk_nbhead) << 8) | dsk_biosnb;
  b1p->bootloader2.bootloader2_cmd.int1302.reg_ax = 0x201;
  b1p->bootloader2.bootloader2_cmd.int1302.reg_cx
	=  (first_lba / dsk_nbsector / dsk_nbhead << 8)
	 | (first_lba % dsk_nbsector + 1);

  b1p->bootloader2.bootloader2_cmd.int1302.dummy = 0;
  b1p->bootloader2.bootloader2_cmd.int1302.will_be_set_to_esbx = 0x07C00200;
  }

static unsigned char *
next_checksum (unsigned short *ptr,
		bootloader2_type *bootchain,
		unsigned dsk_sectorsize,
		unsigned bootchain_off)
  {
  signed index = 0;

  while (1) {
      /* We are "squaling" those two: */
      /* hope here that dsk_sectorsize % 2 == 0 */
      bootchain[index].header.nbword *= dsk_sectorsize/2;
      bootchain[index].header.checksum = 0;
      if (bootchain[index].header.next == 0)
	  break;
      bootchain[index].header.next *= sizeof (bootloader2_type);
      bootchain[index].header.next += bootchain_off;
      index ++;
      }

  index++;
  while (--index >= 0) {
      unsigned short checksum = 0;
      unsigned cpt = bootchain[index].header.nbword + 1;

      while (--cpt)
	  checksum += *--ptr;
      bootchain[index].header.checksum = - checksum;
      }

  return (unsigned char*)ptr;
  }

static unsigned
first_checksum (unsigned char *file,
		unsigned start_off,
		unsigned boot1param_off,
		unsigned end_off,
		unsigned bootchain_off,
		unsigned deltaseg,
		unsigned patch1adr,
		unsigned patch2adr,
		unsigned patch3adr,
		unsigned patch4adr,
		loader_type loader)
  {
  boot1param_type *b1p = (boot1param_type *)(file + boot1param_off);
  unsigned short *ptr;
  unsigned short checksum;
  unsigned nbword = b1p->bootloader2.header.nbword;

  /* Checksum of the _first_ sector loaded by MBR: */
  ptr = (unsigned short *)(file + 512);
  checksum = 0;
  while (ptr < (unsigned short *)(file + 512 + 2 * nbword))
      checksum -= *ptr++;
  b1p->bootloader2.header.checksum = checksum;
  b1p->bootloader2.header.next = bootchain_off;

  /* Check the checksum of the first sector - in case of bad assembler */
  if (*(unsigned short *)(file+patch2adr) != 0x0840 + deltaseg) {
      printf ("Unrecognised patch pattern 2, abort.\n");
      return 1;
      }
  *(unsigned short *)(file+patch2adr) -= deltaseg;
  ptr = (unsigned short *)(file + start_off);
  checksum = 0;

  while (ptr < &b1p->boot1_checksum)
      checksum -= *ptr++;

  if (b1p->boot1_checksum != checksum)
      printf ("WARNING: checksum for Level 1 is incorrect, it is 0x%X"
	" and should be 0x%X (corrected).\n",
	b1p->boot1_checksum, checksum);
  *(unsigned short *)(file+patch2adr) += deltaseg;
  /* end of the check assembler */

  if (loader != bios_loader) {
      /* DOS bootloader, do the patches: */
      if (file[patch1adr] != 0xEB || file[patch1adr+1] != 0x3C
				  || file[patch1adr+2] != 0x90) {
	  printf ("Unrecognised pattern 1 for patch, abort.\n");
	  return 1;
	  }
      file[patch1adr] = 0xFA;
      file[patch1adr+1] = 0xEB;
      file[patch1adr+2] = 0x3B;

      if (*(unsigned short *)(file+patch2adr) != 0x0840 + deltaseg) {
	  printf ("Unrecognised pattern 2 for patch, abort.\n");
	  return 1;
	  }
      if (loader == dos_com_loader)
	  *(unsigned short *)(file+patch2adr) = 0x0010 + deltaseg;
	else /* (loader == dos_exe_loader) */
	  *(unsigned short *)(file+patch2adr) = deltaseg;

      *(unsigned short *)(file+patch3adr) -= 1; /* will call a ret */

      if (*(unsigned *)(file + patch4adr) != 0xCF1F0761) {
	  printf ("Unrecognised pattern 4 for patch: 0x%X, abort.\n",
			*(unsigned *)(file + patch4adr));
	  return 1;
	  }

      *(unsigned *)(file+patch4adr) = 0x21CD4CB4;
      }

  /* Checksum of the (patched) sector loaded by the BIOS: */
  ptr = (unsigned short *)(file + start_off);
  checksum = 0;

  while (ptr < &b1p->boot1_checksum)
      checksum -= *ptr++;

  ptr ++;	/* the actual checksum */

  while (ptr < (unsigned short *)(file + end_off))
      checksum -= *ptr++;
  b1p->boot1_checksum = checksum;
  return 0;
  }

/*
 * Parameters of the FAT12/FAT16 system:
 */
unsigned inline
nbsectorinroot (unsigned nbrootdirentry, unsigned bytepersector)
  {
  return DIVROUNDUP (nbrootdirentry * sizeof (directory_type),
		     bytepersector);
  }

unsigned inline
nbsectorperfat12 (unsigned nbclusteronfat, unsigned bytepersector)
  {
  /*
   * Clusters descriptor FF0-FFF included, 0 and 1 are
   *  reserved (mediadescriptor), the
   *  boot sector + FAT + root-dir themselves do not need to be coded.
   * FAT12, 12 bits to describe one cluster, multiply by 3/2.
   */
  return DIVROUNDUP (3 * ((0xFF0-1)-2) * nbclusteronfat,
		     2 * 0xFFF * bytepersector);
  }

unsigned inline
nbsectorperfat16 (unsigned nbclusteronfat, unsigned bytepersector)
  {
  return DIVROUNDUP (2LL * ((0xFFF0-1)-2) * nbclusteronfat,
	     0xFFFF * bytepersector);
  }


unsigned inline
nbclusteronfat (unsigned nbtotalsector, unsigned sectorpercluster,
		unsigned bytepersector, unsigned nbfat,
		unsigned nbsectorinroot, unsigned reservedsector)
  {
  unsigned first_estim;

  if (nbtotalsector < 4096)
      first_estim = nbsectorperfat12 (nbtotalsector / sectorpercluster,
					bytepersector);
    else
      first_estim = nbsectorperfat16 (nbtotalsector / sectorpercluster,
					bytepersector);

  return DIVROUNDUP(nbtotalsector - nbfat * first_estim
			 - nbsectorinroot - reservedsector,
		    sectorpercluster);
  }

unsigned inline
getdisksectorgap (unsigned nbfat, unsigned nbsectorperfat,
		  unsigned nbsectorinroot)
  {
  return nbfat * nbsectorperfat + nbsectorinroot;
  }

static unsigned
write_fat_12 (FILE *fout,
		unsigned char dsk_mediadescriptor,
		unsigned short nbfilecluster,
		unsigned short nbsectorperfat,
		unsigned short nbclusteronfat)
  {
  unsigned short value, cpt, nbFATbyte, endnbFATbyte;

  if (fwrite(&dsk_mediadescriptor, 1, 1, fout) != 1)
      return 1;
  value = 0xFFFF;
  if (fwrite(&value, 1, 2, fout) != 2)
      return 1;

  nbFATbyte = 3;
  for (cpt = 0; cpt < nbfilecluster; cpt++) {
      if ((cpt & 0x01) == 0) {
	  value = (cpt+3);
	  if (fwrite (&value, 1, 1, fout) != 1) /* little endian */
	      return 1;
	  value >>= 8;
	  nbFATbyte++;
	  }
	else {
	  value = ((cpt+3) << 4) | (value & 0x0F);
	  if (fwrite (&value, 1, 2, fout) != 2) /* little endian */
	      return 1;
	  value = 0x0FFF;
	  nbFATbyte += 2;
	  }
      }

  /* write the end marker: */
  if (value != 0x0FFF)
      value |= 0xFFF0;
  if (fwrite (&value, 1, 2, fout) != 2) /* little endian */
      return 1;
  nbFATbyte += 2;

  /* Fill end of FAT: */
  endnbFATbyte = 3 * (nbclusteronfat + 6) / 2;
  while (nbFATbyte++ <= endnbFATbyte) {
      unsigned char abyte = 0;
      if (fwrite (&abyte, 1, 1, fout) != 1)
	  return 1;
      }

  /* Fill unused part of FAT: */
  /* If I fill it with 0xFF, I have errors by mtools like:
Bad FAT entry 2848 at 2847
Fat error detected
     Where is the bug ? */
  while (nbFATbyte++ <= 512 * nbsectorperfat) {
      unsigned char abyte = 0;
      if (fwrite (&abyte, 1, 1, fout) != 1)
	  return 1;
      }

  return 0;
  }

static unsigned
write_fat_16 (FILE *fout,
		unsigned char dsk_mediadescriptor,
		unsigned short nbfilecluster,
		unsigned short nbsectorperfat,
		unsigned short nbclusteronfat)
  {
  unsigned short value;
  unsigned nbFATbyte, endnbFATbyte;

  if (fwrite(&dsk_mediadescriptor, 1, 1, fout) != 1)
      return 1;
  value = 0xFFFF;
  if (fwrite(&value, 1, 1, fout) != 1)
      return 1;
  if (fwrite(&value, 1, 2, fout) != 2)
      return 1;

  for (value = 0; value < nbfilecluster; value++) {
      unsigned short tmp = value + 3;
      if (fwrite (&tmp, 1, 2, fout) != 2)
	  return 1;
      }

  /* write the end marker: */
  value = 0xFFF0;
  if (fwrite (&value, 1, 2, fout) != 2)
      return 1;

  /* Fill end of FAT: */
  nbFATbyte = 6 + 2 * nbfilecluster;
  endnbFATbyte = 2 * (nbclusteronfat + 4);
  while (nbFATbyte++ <= endnbFATbyte) {
      unsigned char abyte = 0;
      if (fwrite (&abyte, 1, 1, fout) != 1)
	  return 1;
      }

  /* Fill unused part of FAT: */
  while (nbFATbyte++ <= 512 * nbsectorperfat) {
      unsigned char abyte = 0xFF;
      if (fwrite (&abyte, 1, 1, fout) != 1)
	  return 1;
      }

  return 0;
  }

static unsigned
write_root (FILE *fout, unsigned dsk_nbrootdirentry, unsigned filesize,
		unsigned dsk_nbsector, unsigned dsk_nbcyl)
  {
  directory_type rootdir[dsk_nbrootdirentry];
  const directory_type
     entry1 = {
	/* char name[8] */	{'E', 'L', '-', 'B', 'O', 'O', 'T', '-'},
	/* char extension[3] */ {'V', '0', '5'},  /*VERSION*/
	{
	/* readonly */		0,
	/* hidden */		0,
	/* system */		0,
	/* volumelabel */	1,
	/* subdirectory */	0,
	/* archive */		1,
	/* unused : 2 */	0,
	},
	/* reserved[10] */	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{
	/* seconddiv2 :5 */	0,
	/* minute     :6 */	30,
	/* hour	      :5 */	8,
	/* day	      :5 */	1,	/* 1..31 */
	/* month      :4 */	3,	/* 1..12 */
	/* year	      :7 */	19,	/* 1980 + x */
	},
	/* cluster */		0,
	/* size */		0,
	},

     entry2 = {
	/* char name[8] */	{'B', 'O', 'O', 'T', '-', 'H', 'D', ' '},
	/* char extension[3] */ {'S', 'Y', 'S'},
	{
	/* readonly */		1,
	/* hidden */		1,
	/* system */		1,
	/* volumelabel */	0,
	/* subdirectory */	0,
	/* archive */		0,
	/* unused : 2 */	0,
	},
	/* reserved[10] */	 {0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{
	/* seconddiv2 :5 */	25,
	/* minute     :6 */	34,
	/* hour	      :5 */	12,
	/* day	      :5 */	15,	/* 1..31 */
	/* month      :4 */	3,	/* 1..12 */
	/* year	      :7 */	19,	/* 1980 + x */
	},
	/* cluster */		2,
	/* size */		filesize,
	};

  memset (rootdir, 0, sizeof (rootdir)); /* variable array cannot be inited */
  rootdir[0] = entry1;
  rootdir[1] = entry2;

  if (dsk_nbsector == 36) {
      rootdir[1].name[4] = '2';
      rootdir[1].name[5] = '8';
      rootdir[1].name[6] = '8';
      }
  if (dsk_nbsector == 21) {
      rootdir[1].name[4] = '1';
      rootdir[1].name[5] = '6';
      rootdir[1].name[6] = '8';
      }
  if (dsk_nbsector == 15) {
      rootdir[1].name[4] = '1';
      rootdir[1].name[5] = '2';
      rootdir[1].name[6] = '0';
      }
  if (dsk_nbsector == 18) {
      rootdir[1].name[4] = '1';
      rootdir[1].name[5] = '4';
      rootdir[1].name[6] = '4';
      }
  if (dsk_nbsector == 9) {
      if (dsk_nbcyl == 80) {
	  rootdir[1].name[4] = '7';
	  rootdir[1].name[5] = '2';
	  rootdir[1].name[6] = '0';
	  }
	else {
	  rootdir[1].name[4] = '3';
	  rootdir[1].name[5] = '6';
	  rootdir[1].name[6] = '0';
	  }
      }
  if (fwrite(rootdir, 1, sizeof(rootdir), fout) != sizeof(rootdir))
      return 1;
  return 0;
  }

int main (int argc, char **argv)
  {
  FILE *fin, *fout;
  char *fin_name = 0, *fout_name = 0;
  bootloader2_type *bootchain;
  boot1param_type *boot1param;
  unsigned char *file, tmp;
  struct stat statbuf;

  serialconf_type serialconf = {0};
  struct instboot_info_str *instboot_info;
  struct gujin_param_str *gujin_param;

  char extra_cmdline[sizeof (gujin_param->extra_cmdline)] = "";

  unsigned char  VGA_valid_mode[16] = { [0 ... 15] = 0xFF};
  unsigned short default_video_mode = (unsigned short) -1;

  unsigned nbblock;
  unsigned dma_frontier = (64*1024-(0x7C00+2*512))/512; /* 64 */

  unsigned char *chkadr;
  unsigned boot2_start_lba, filesize;
  unsigned dsk_nbtotalsector, dsk_nbsectorperfat, dsk_nbclusteronfat;

  unsigned short dsk_sectorsize = 512;
  unsigned char dsk_biosnb = 0;
  unsigned char dsk_nbfat = 2;

  /* Default for a 1.44M floppy disk (used for boot.exe chksm): */
  unsigned dsk_nbcyl = 80;
  unsigned char dsk_nbhead = 2, dsk_nbsector = 18;
  unsigned char dsk_nbsectorpercluster = 1;
  unsigned short dsk_nbrootdirentry = 224;
  unsigned char dsk_mediadescriptor = 0xF0;
  unsigned dsk_start = 0;

  unsigned cmdline_no_disk = 1;
  struct gujin_param_attrib gujin_param_attr = {
#define Apply(field, deflt, grey)	field:	deflt,
	APPLY_ATTRIBUTE_ORDER(Apply)
#undef Apply
      };
  loader_type type_of_loader = bios_loader;

  for (tmp = 1; tmp < argc; tmp++) {
      unsigned found = 0;
      unsigned val;

#define SIZESTR(str)	str, sizeof(str)-1

      if (!strncasecmp (argv[tmp], SIZESTR("-v"))) {
	  printf (" %s version 0.5 -- see file gpl.txt for the licence\n",
		  argv[0]); /*VERSION*/
	  return 0;
	  }
      if (   !strncasecmp (argv[tmp], SIZESTR("-h="))
	  || !strncasecmp (argv[tmp], SIZESTR("--head="))) {
	  if (sscanf (argv[tmp], "%*[^=]=%i", &val) != 1) {
	      printf ("Error reading parameter %s, abort.\n", argv[tmp]);
	      return 1;
	      }
	  dsk_nbhead = val;
	  found = 1;
	  }
      if (!strncasecmp (argv[tmp], SIZESTR("--nbfat="))) {
	  if (sscanf (argv[tmp], "%*[^=]=%i", &val) != 1) {
	      printf ("Error reading parameter %s, abort.\n", argv[tmp]);
	      return 1;
	      }
	  if (dsk_nbfat == 0 || dsk_nbfat > 5) {
	      printf ("Error reading parameter %s: 1 <= nb_fat <= 5, abort.\n",
			argv[tmp]);
	      return 1;
	      }
	  dsk_nbfat = val;
	  found = 1;
	  }
      if (   !strncasecmp (argv[tmp], SIZESTR("-c="))
	  || !strncasecmp (argv[tmp], SIZESTR("--cylinder="))) {
	  if (sscanf (argv[tmp], "%*[^=]=%i", &val) != 1) {
	      printf ("Error reading parameter %s, abort.\n", argv[tmp]);
	      return 1;
	      }
	  dsk_nbcyl = val;
	  found = 1;
	  }
      if (   !strncasecmp (argv[tmp], SIZESTR("-s="))
	  || !strncasecmp (argv[tmp], SIZESTR("--sector="))) {
	  if (sscanf (argv[tmp], "%*[^=]=%i", &val) != 1) {
	      printf ("Error reading parameter %s, abort.\n", argv[tmp]);
	      return 1;
	      }
	  dsk_nbsector = val;
	  found = 1;
	  }

      if (!strncasecmp (argv[tmp], SIZESTR("--cmdline="))) {
	  unsigned len = strlen (argv[tmp]) - (sizeof("--cmdline=") - 1);
	  if (len == 0 || len > sizeof (extra_cmdline) - 1) {
	      printf ("Error reading parameter %s, empty or too long, abort.\n", argv[tmp]);
	      return 1;
	      }
	  strncpy (extra_cmdline, argv[tmp] + sizeof("--cmdline=") - 1, sizeof (extra_cmdline));
	  found = 1;
	  }

      if (   !strncasecmp (argv[tmp], SIZESTR("-COM"))
	  || !strncasecmp (argv[tmp], SIZESTR("--serial="))) {
	  unsigned val_speed = 9600, val_nbbit = 8, val_nbstop = 1;
	  char val_parity = 'n';
	  if (sscanf (argv[tmp], "%*[^0-9]%i,%i,%c,%i,%i",
			&val, &val_speed, &val_parity,
			&val_nbbit, &val_nbstop) < 1) {
	      printf ("Error reading parameter %s, abort.\n", argv[tmp]);
	      return 1;
	      }
	  if (val < 1 || val > 4) {
	      printf ("Error no such BIOS COM%d serial port, abort.\n", val);
	      return 1;
	      }
	  if (   val_speed != 9600 && val_speed != 4800
	      && val_speed != 2400 && val_speed != 1200
	      && val_speed !=  600 && val_speed !=  300
	      && val_speed !=  150 && val_speed !=  110
	      ) {
	      printf ("Error no such BIOS baud rate: %d, abort.\n", val_speed);
	      return 1;
	      }
	  if (val_nbbit != 8 && val_nbbit != 7) {
	      printf ("Error no such BIOS bit/char: %d, abort.\n", val_nbbit);
	      return 1;
	      }
	  if (val_parity != 'n' && val_parity != 'e' && val_parity != 'o') {
	      printf ("Error no such BIOS parity: %c, abort.\n", val_parity);
	      return 1;
	      }
	  if (val_nbstop != 1 && val_nbstop != 2) {
	      printf ("Error no such BIOS stop length: %d, abort.\n",
			 val_nbstop);
	      return 1;
	      }

	  if (val_nbbit == 8)
	      serialconf.eightbit = 1;
	    else
	      serialconf.eightbit = 0;
	  serialconf.constant_1 = 1;
	  if (val_nbstop == 2)
	      serialconf.doublestop = 1;
	    else
	      serialconf.doublestop = 0;
	  if (val_parity == 'n')
	      serialconf.parity = 0;
	    else {
	      serialconf.parity = 1;
	      if (val_parity == 'e')
		  serialconf.even = 1;
		else
		  serialconf.even = 0;
	      }
	  switch (val_speed) {
	      case  110: serialconf.speed = B110; break;
	      case  150: serialconf.speed = B150; break;
	      case  300: serialconf.speed = B300; break;
	      case  600: serialconf.speed = B600; break;
	      case 1200: serialconf.speed = B1200; break;
	      case 2400: serialconf.speed = B2400; break;
	      case 4800: serialconf.speed = B4800; break;
	      case 9600: serialconf.speed = B9600; break;
	      }
	  serialconf.constant_0 = 0;
	  serialconf.port = val - 1;
	  found = 1;
	  }
      if (!strncasecmp (argv[tmp], SIZESTR("--sectorpercluster="))) {
	  if (sscanf (argv[tmp], "%*[^=]=%i", &val) != 1) {
	      printf ("Error reading parameter %s, abort.\n", argv[tmp]);
	      return 1;
	      }
	  dsk_nbsectorpercluster = val;
	  found = 1;
	  }
      if (!strncasecmp (argv[tmp], SIZESTR("--rootdirentry="))) {
	  if (sscanf (argv[tmp], "%*[^=]=%i", &val) != 1) {
	      printf ("Error reading parameter %s, abort.\n", argv[tmp]);
	      return 1;
	      }
	  dsk_nbrootdirentry = val;
	  found = 1;
	  }
      if (!strncasecmp (argv[tmp], SIZESTR("--mediadescriptor="))) {
	  if (sscanf (argv[tmp], "%*[^=]=%i", &val) != 1) {
	      printf ("Error reading parameter %s, abort.\n", argv[tmp]);
	      return 1;
	      }
	  dsk_mediadescriptor = val;
	  found = 1;
	  }

#define Apply(field, deflt, grey)	\
    if (!strncasecmp (argv[tmp], SIZESTR("--" # field "="))) {		\
	char val = argv[tmp][sizeof("--" # field "=") - 1];		\
	if (val != '0' && val != '1') {					\
	    printf ("Error reading parameter %s, abort.\n", argv[tmp]);	\
	    return 1;							\
	    }								\
	gujin_param_attr.field = val - '0';				\
	found = 1;							\
	}
	APPLY_ATTRIBUTE_ORDER(Apply)
#undef Apply

      if (   !strncasecmp (argv[tmp], SIZESTR("-d=dos.com"))
	  || !strncasecmp (argv[tmp], SIZESTR("--disk=dos.com"))) {
	  type_of_loader = dos_com_loader;
	  cmdline_no_disk = 0;
	  found = 1;
	  }
	else if (   !strncasecmp (argv[tmp], SIZESTR("-d=dos.exe"))
		 || !strncasecmp (argv[tmp], SIZESTR("--disk=dos.exe"))) {
	  type_of_loader = dos_exe_loader;
	  cmdline_no_disk = 0;
	  found = 1;
	  }
	else if (   !strncasecmp (argv[tmp], SIZESTR("-d="))
		 || !strncasecmp (argv[tmp], SIZESTR("--disk="))) {
	  if (sscanf (argv[tmp], "%*[^=]=%i", &val) != 1) {
	      printf ("Error reading parameter %s, abort.\n", argv[tmp]);
	      return 1;
	      }
	  cmdline_no_disk = 0;
	  switch (val) {
	   /* Standard sector per track: 9 for 360K and 720K, 18 for 1.44M,
	    * 36 for 2.88M, 15 for 1.2M, 21 for 1.68M
	    */
	    case 144:  /* Default for a 1.44M floppy disk: */
	      dsk_nbcyl = 80;
	      dsk_nbhead = 2;
	      dsk_nbsector = 18;
	      dsk_nbsectorpercluster = 1;
	      dsk_nbrootdirentry = 224;
	      dsk_mediadescriptor = 0xF0;
	      break;
	    case 720:  /* Default for a 720K floppy disk: */
	      dsk_nbcyl = 80;
	      dsk_nbhead = 2;
	      dsk_nbsector = 9;
	      dsk_nbsectorpercluster = 2;
	      dsk_nbrootdirentry = 112;
	      dsk_mediadescriptor = 0xF9;
	      break;
	    case 120:  /* Default for a 1.2M floppy disk: */
	      dsk_nbcyl = 80;
	      dsk_nbhead = 2;
	      dsk_nbsector = 15;
	      dsk_nbsectorpercluster = 1;
	      dsk_nbrootdirentry = 224;
	      dsk_mediadescriptor = 0xF9;
	      break;
	    case 360:  /* Default for a 360K floppy disk: */
	      dsk_nbcyl = 40;
	      dsk_nbhead = 2;
	      dsk_nbsector = 9;
	      dsk_nbsectorpercluster = 2;
	      dsk_nbrootdirentry = 112;
	      dsk_mediadescriptor = 0xFD;
	      break;
	    case 168:  /* Default for a 1.68M floppy disk: */
	      dsk_nbcyl = 80;
	      dsk_nbhead = 2;
	      dsk_nbsector = 21;
	      dsk_nbsectorpercluster = 1;
	      dsk_nbrootdirentry = 224;
	      dsk_mediadescriptor = 0xF0;
	      break;
	    case 288:  /* Default for a 2.88M floppy disk: */
	      dsk_nbcyl = 80;
	      dsk_nbhead = 2;
	      dsk_nbsector = 36;
	      dsk_nbsectorpercluster = 2;
	      dsk_nbrootdirentry = 448;
	      dsk_mediadescriptor = 0xF0;
	      break;
	      }
	  found = 1;
	  }

      if (!strncasecmp (argv[tmp], SIZESTR("--default_video_mode="))) {
	  if (sscanf (argv[tmp], "%*[^=]=%i", &val) != 1) {
	      printf ("Error reading parameter %s, abort.\n", argv[tmp]);
	      return 1;
	      }
	  default_video_mode = val;
	  found = 1;
	  }
      if (!strncasecmp (argv[tmp], SIZESTR("--exclude_video_mode="))) {
	  char str[20] = {}, *strptr = str;
	  if (sscanf (argv[tmp], "%*[^=]=%i%20s", &val, str) < 1) {
	      printf ("Error reading parameter %s, abort.\n", argv[tmp]);
	      return 1;
	      }
	  if (val >= 0x80) {
	      printf ("Error reading parameter %s, out of range: "
			"0x%X, abort.\n", argv[tmp], val);
	      return 1;
	      }
	  VGA_valid_mode[val/8] &= ~(1 << ((val) % 8));
	  while (*strptr) {
	      if (*strptr == ',') {
	  unsigned other;
		  if (sscanf (++strptr, "%i", &other) < 1) {
		      printf ("Error reading parameter %s after comma, "
				"abort.\n", argv[tmp]);
		      return 1;
		      }
		  if (other >= 0x80) {
		      printf ("Error reading parameter %s, out of range: "
				"0x%X, abort.\n", argv[tmp], other);
		      return 1;
		      }
	  VGA_valid_mode[other/8] &= ~(1 << ((other) % 8));
	  }
	else if (*strptr == '-') {
	  unsigned other;
		  if (sscanf (++strptr, "%i", &other) < 1) {
		      printf ("Error reading parameter %s after minus, "
				"abort.\n", argv[tmp]);
		      return 1;
		      }
		  if (other <= val) {
		      printf ("Error reading parameter %s, "
				"invalid range, abort.\n", argv[tmp]);
		      return 1;
		      }
		  if (other >= 0x80) {
		      printf ("Error reading parameter %s, "
				"out of range: 0x%X, abort.\n",
				argv[tmp], other);
		      return 1;
		      }
		  while (val <= other) {
		      VGA_valid_mode[val/8] &= ~(1 << ((val) % 8));
		      val ++;
		      }
		  }
		else {
		  printf ("Error reading parameter %s, not comma or minus"
			  " separator, abort.\n", argv[tmp]);
		  return 1;
		  }
	      while (   (*strptr >= '0' && *strptr <= '9')
		     || (*strptr >= 'a' && *strptr <= 'f')
		     || (*strptr >= 'A' && *strptr <= 'F')
		     || *strptr == 'x' || *strptr == 'X'
		     )
		  strptr++;
	      }
	  found = 1;
	  }

      if (found == 0) {
	  if (fin_name == 0 && argv[tmp][0] != '-') {
	      fin_name = argv[tmp];
	      found = 1;
	      continue;
	      }
	  if (fout_name == 0 && argv[tmp][0] != '-') {
	      fout_name = argv[tmp];
	      found = 1;
	      continue;
	      }
	  if (   strncasecmp (argv[tmp], SIZESTR("--help"))
	      && strncasecmp (argv[tmp], SIZESTR("-h")))
	      printf ("Error reading parameter \"%s\", abort.\n", argv[tmp]);
	  printf ("%s transform a \"boot\" file into a bootable floppy disk\n"
		  " or DOS executable for the Gujin bootloader.\n", *argv);
	  printf (" USAGE: %s [param] infile outfile  "
		  "(param is case insensitive)\n", *argv);
	  printf (
  " where [param] can be: (<d> represent a number with default d)\n"
  " -v  to display version    --help  to display this help\n"
  " --cmdline=\"append to command line for all kernels\"\n"
  " -COM[1-4][,<9600>[,<n>[,<8>[,<1>]]]]] or --serial=COM[]\n"
  "   to setup a serial interface (VT100+) instead of the PC screen.\n"
  " --default_video_mode=<>, --exclude_video_mode=<>[,<>][,<>-<>] intuitive.\n"
  " --refuse_8x14font=[0,1] : mostly modes 0xF and 0x10 (640x350), if BIOS buggy.\n"
  "   same pattern \"--xxx=[0,1]\" for all other attributes, see boot.h.\n"
  " -d=[360|120|720|<144>|168|244|dos.exe|dos.com] or --disk=[]\n"
  "   describes the size of the floppy (needed if outfile is not a\n"
  "   device) or ask for a DOS bootloader. You can overwrite:\n"
  "   -h=<> or --head=<>     to set number of heads\n"
  "   -c=<> or --cylinder=<> to set number of cylinders\n"
  "   -s=<> or --sector=<>   to set number of sectors\n"
  " --sectorpercluster=<1> --rootdirentry=<224> --mediadescriptor=<0xF0>\n"
  " --nbfat=<2> (shall be in between 1 and 5, read the docs!)\n"
//" -fs=[FAT12|FAT16] or --filesystem=[] to set the filesystem\n"
		);
	  return 1;
	  }

#undef SIZESTR
      }

  if (fin_name == 0 || fout_name == 0) {
      printf ("Parameter error: no input file or output file/device.\n");
      printf (" type \"%s -h\" or \"%s --help\" for help.\n", *argv, *argv);
      return 1;
      }

#if 0 /* using getmntent() would be better */
  {
  FILE *fmount = fopen ("/proc/mounts", "r");
  unsigned pos = 0, fout_len = strlen (fout_name);

  if (fmount == NULL) {
      printf ("Info: No '/proc/mounts' found on this system.\n");
      }
    else {
      int val;
      while ((val = getc (fmount)) != EOF) {
	  if (pos < fout_len) {
	      if (fout_name[pos] == val)
	  pos ++;
	else
	  pos = fout_len + 1;
	      if (pos == fout_len)
	  break;
	      }
	  if (val == '\n')
	      pos = 0;
	  }
      fclose (fmount);
      if (pos == fout_len) {
	  printf ("File/device \"%s\" seems to be mounted.\n", fout_name);
	  return 1;
	  }
      }
  }
#endif

  if (stat(fin_name, &statbuf) != 0) {
      printf ("Cannot stat input file: \"%s\"\n", fin_name);
      return 1;
      }
  if (statbuf.st_size <= 512 || statbuf.st_size >= 3*64*1024) {
      printf ("Input file %s invalid size: %d\n",
		 fin_name, (unsigned)statbuf.st_size);
      return 1;
      }
  filesize = DIVROUNDUP (statbuf.st_size, 512) * 512;
  if ((file = malloc (filesize)) == NULL) {
      printf ("Cannot malloc memory for buffer.\n");
      return 1;
      }
  memset (file, 0, filesize);
  if ((fin = fopen (fin_name, "r")) == NULL) {
      printf ("cannot open input file: \"%s\"\n", fin_name);
      return 1;
      }

  if (fread(file, 1, statbuf.st_size, fin) != statbuf.st_size) {
      printf ("cannot read the complete array (hardware error ?)\n");
      return 1;
      }
/*  if (!feof(fin)) { */
  if (fread(&tmp, 1, 1, fin) != 0) {
      printf ("File not at end after reading array (concurrent access ?)\n");
      return 1;
      }
  if (fclose (fin) == EOF) {
      printf ("cannot close input file: %s\n", fin_name);
      return 1;
      }

  /* Initialise pointers, using hardcoded address of address
	of instboot_info: */
  instboot_info = *(struct instboot_info_str **) (file + 0x200);

  if ((unsigned)instboot_info >= statbuf.st_size
				 - sizeof(struct instboot_info_str)) {
      printf ("instboot_info offset = 0x%X, probably not a gujin file,"
		" abort.\n",
	      (unsigned) instboot_info);
      return 1;
      }

  *((unsigned *)&instboot_info) += (unsigned)file;

  if (   instboot_info->start_adr != 0x3E
      || instboot_info->bootend_adr != 0x1B6
      || instboot_info->boot1param_adr != 0x196) {
      printf ("instboot_info: start_adr = 0x%X, bootend_adr = 0x%X, "
		"boot1param_adr = 0x%X,\nprobably not a gujin file, abort.\n",
	      instboot_info->start_adr,
	      instboot_info->bootend_adr,
	      instboot_info->boot1param_adr);
      return 1;
      }
  if (   instboot_info->patch1adr != 0
      || instboot_info->patch2adr >= 0x200
      || instboot_info->patch3adr >= 0x200
      || instboot_info->patch4adr >= 0x200) {
      printf ("instboot_info: patch1adr = 0x%X, patch2adr = 0x%X, "
	      "patch3adr = 0x%X, patch4adr = 0x%X\n"
	      "probably not a gujin file, abort.\n",
	      instboot_info->patch1adr,
	      instboot_info->patch2adr,
	      instboot_info->patch3adr,
	      instboot_info->patch4adr);
      return 1;
      }
  if (instboot_info->gujin_param_adr >= 0x400) {
      printf ("instboot_info: gujin_param_adr = 0x%X, will not work, abort.\n",
	      instboot_info->gujin_param_adr);
      return 1;
      }
  gujin_param = (struct gujin_param_str *)(file
					   + instboot_info->gujin_param_adr);
  if (gujin_param->magic != 16980327) {
      printf ("gujin_param magic value not found, abort.\n");
      return 1;
      }
  if (gujin_param->version != 0x0005) { /*VERSION*/
      printf ("gujin_param: incompatible version, abort.\n");
      return 1;
      }
  if (   instboot_info->bootchain_adr <= 0x200
      || instboot_info->bootchain_adr >= 0x400 - sizeof (bootloader2_type)) {
      printf ("instboot_info: bootchain at address 0x%X, "
		"will not work, abort.\n",
		instboot_info->bootchain_adr);
      return 1;
      }
  if (   instboot_info->bootchain_nbblock < 2
      || instboot_info->bootchain_nbblock > 100) {
      printf ("instboot_info: bootchain_nbblock = %d, unrealistic, abort.\n",
		instboot_info->bootchain_nbblock);
      return 1;
      }
  {
  unsigned short read_disk_adr;
  unsigned char ret_instr;
  read_disk_adr = *(unsigned short *)((unsigned)file
					+ instboot_info->patch3adr);
  read_disk_adr += instboot_info->patch3adr + 2; /* relative addr */
  ret_instr = *(unsigned char *)((unsigned)file + read_disk_adr - 1);
  if (ret_instr != 0xC3) {
      printf ("instboot_info: read_disk function at 0x%X, "
		"instruction before is not \"retw\" (0x%X), abort.\n",
		read_disk_adr, ret_instr);
      return 1;
      }
  }

  /* look in boot.ld for the following: */
  if (instboot_info->extraseg != 0) {
      printf ("Info: this configuration leaves %d bytes for code,\n",
	0x10000 - (instboot_info->extraseg << 4));
      printf ("Info: and leaves %d bytes for code extra segment.\n",
		0x10000 - ((instboot_info->deltaseg
			    - instboot_info->extraseg) << 4));
      }
    else if (instboot_info->deltaseg != 0)
      printf ("Info: this configuration leaves %d bytes for code.\n",
	0x10000 - (instboot_info->deltaseg << 4));
  printf ("Info: this configuration leaves %d bytes for stack%s.\n",
		0x10000 - instboot_info->edata,
		gujin_param->attrib.big_malloc? "" : " + heap");

  {
  int fout_desc;
  struct stat fout_stat;

  if (   stat(fout_name, &fout_stat) == 0
//    && (S_ISBLK(fout_stat.st_rdev) || S_ISCHR(fout_stat.st_rdev))
      && !S_ISREG(fout_stat.st_rdev) ) {
      /* O_EXCL on Linux means fail if already mounted.
	 At least when supermount is not used... */
      if ((fout_desc = open(fout_name, O_WRONLY | O_EXCL)) == -1) {
	  printf ("cannot open output device: \"%s\" "
		  "(already mounted?)\n", fout_name);
	  printf ("Hint: Try on a command line: \"umount /mnt/floppy\"\n");
	  perror ("ERROR");
	  return 1;
	  }

      if ((fout = fdopen (fout_desc, "w")) == NULL) {
	  printf ("cannot open output device: \"%s\"\n", fout_name);
	  return 1;
	  }
      }
    else {
      if ((fout = fopen (fout_name, "w")) == NULL) {
	  printf ("cannot open output file: \"%s\"\n", fout_name);
	  return 1;
	  }
      }
  }

  if (cmdline_no_disk) {
      struct floppy_struct fd_param;
      struct hd_geometry   hd_param;

      if (ioctl (fileno(fout), FDGETPRM, &fd_param) >= 0) {
	  dsk_nbcyl = fd_param.track;
	  dsk_nbhead = fd_param.head;
	  dsk_nbsector = fd_param.sect;
	  }
	else if (ioctl (fileno(fout), HDIO_GETGEO, &hd_param) >= 0) {
	  dsk_nbcyl = hd_param.cylinders;
	  dsk_nbhead = hd_param.heads;
	  dsk_nbsector = hd_param.sectors;
	  dsk_mediadescriptor = 0xF8;
	  dsk_start = hd_param.start;
	  dsk_nbrootdirentry = 512;
	  dsk_nbsectorpercluster = 1;
	  while (nbclusteronfat (dsk_nbcyl * dsk_nbhead * dsk_nbsector,
	dsk_nbsectorpercluster,
	512,
	dsk_nbfat,
	nbsectorinroot (dsk_nbrootdirentry, 512),
	1) > 0xFFFF)
	      dsk_nbsectorpercluster *= 2;
	  printf (" dsk_nbcyl = %u, dsk_nbhead = %u, dsk_nbsector = %u, "
		  " dsk_mediadescriptor = %u, dsk_start = %u, "
		  " dsk_nbrootdirentry = %u, "
		  " dsk_nbsectorpercluster = %u.\n",
	  dsk_nbcyl, dsk_nbhead, dsk_nbsector,
	  dsk_mediadescriptor, dsk_start,
	  dsk_nbrootdirentry, dsk_nbsectorpercluster);
	  printf ("\n\n ABORT now, this is work in progess and would"
	  " ERASE your hard drive partition!\n");
	  return 1;
	  }
	else {
	  printf ("WARNING: Cannot check floppy disk geometry!\n");
	  }
      if (dsk_start == 0)
	  printf ("Info: configuring for floppy C/H/S = %u/%u/%u, i.e. ",
			dsk_nbcyl, dsk_nbhead, dsk_nbsector);
	else
	  printf ("Info: configuring for partition C/H/S = %u/%u/%u, "
		  "start=%u ", dsk_nbcyl, dsk_nbhead, dsk_nbsector, dsk_start);
      if (dsk_nbsector == 36) {
	  dsk_nbsectorpercluster = 2;
	  dsk_nbrootdirentry = 448;
	  dsk_mediadescriptor = 0xF0;
	  printf ("2.88M\n");
	  }
	else if (dsk_nbsector == 21) {
	  dsk_nbsectorpercluster = 1;
	  dsk_nbrootdirentry = 224;
	  dsk_mediadescriptor = 0xF0;
	  printf ("1.68M\n");
	  }
	else if (dsk_nbsector == 15) {
	  dsk_nbsectorpercluster = 1;
	  dsk_nbrootdirentry = 224;
	  dsk_mediadescriptor = 0xF9;
	  printf ("1.2M\n");
	  }
	else if (dsk_nbsector == 18) {
	  dsk_nbsectorpercluster = 1;
	  dsk_nbrootdirentry = 224;
	  dsk_mediadescriptor = 0xF0;
	  printf ("1.44M\n");
	  }
	else if (dsk_nbsector == 9) {
	  if (dsk_nbcyl == 80) {
	      dsk_nbsectorpercluster = 2;
	      dsk_nbrootdirentry = 112;
	      dsk_mediadescriptor = 0xF9;
	      printf ("720K\n");
	      }
	    else if (dsk_nbcyl == 40) {
	      dsk_nbsectorpercluster = 2;
	      dsk_nbrootdirentry = 112;
	      dsk_mediadescriptor = 0xFD;
	      printf ("360K\n");
	      }
	    else
	      printf ("unknow !\n");
	  }
	else
	  printf ("extended > 1.44M\n");
      }
  dsk_nbtotalsector = dsk_nbcyl * dsk_nbhead * dsk_nbsector;
  dsk_nbclusteronfat = nbclusteronfat (dsk_nbtotalsector,
				dsk_nbsectorpercluster,
				512,
				dsk_nbfat,
				nbsectorinroot (dsk_nbrootdirentry, 512),
				1);
  if (dsk_nbtotalsector <= 0xFFF)
      dsk_nbsectorperfat = nbsectorperfat12 (dsk_nbclusteronfat, 512);
    else
      dsk_nbsectorperfat = nbsectorperfat16 (dsk_nbclusteronfat, 512);

  boot2_start_lba = getdisksectorgap (dsk_nbfat, dsk_nbsectorperfat,
			nbsectorinroot (dsk_nbrootdirentry, 512));

  if (dsk_start != 0) {
      printf ("dsk_nbtotalsector = %u, dsk_nbclusteronfat = %u, "
	      "dsk_nbsectorpercluster = %u, "
	      "FAT%u, dsk_nbsectorperfat = %u, boot2_start_lba = %u\n",
		dsk_nbtotalsector, dsk_nbclusteronfat,
		dsk_nbsectorpercluster,
		(dsk_nbtotalsector < 4096) ? 12 : 16,
		dsk_nbsectorperfat, boot2_start_lba);
      exit(1);
      }

  if (dsk_nbrootdirentry * sizeof(directory_type) % 512) {
      printf ("nbrootdirentry invalid !\n");
      return 1;
      }

  boot1param = (boot1param_type *)(file + instboot_info->boot1param_adr);
  bootchain = (bootloader2_type *)(file + instboot_info->bootchain_adr);

  if (serialconf.constant_1) {
      boot1param->serialconf = serialconf;
      printf ("Info: configuring for serial interface on COM%u, %u,%c,%u,%u.\n",
		serialconf.port + 1,
		((unsigned short []){110,150,300,600,1200,2400,4800,9600})
							[serialconf.speed],
		serialconf.parity? (serialconf.even ? 'e' : 'o') : 'n',
		serialconf.eightbit? 8 : 7,
		serialconf.doublestop? 2 : 1
		);
      }

  /* Setup gujin_param before checksuming : */
  if (strlen (extra_cmdline)) {
      strncpy (gujin_param->extra_cmdline, extra_cmdline,
		sizeof (gujin_param->extra_cmdline)- 1);
      printf ("Info: add on command line for all kernel:\"%s\"\n",
		gujin_param->extra_cmdline);
      }
  gujin_param->default_video_mode = default_video_mode;
#define Apply(field, deflt, grey)	\
	gujin_param->attrib.field = gujin_param_attr.field;
  APPLY_ATTRIBUTE_ORDER(Apply)
#undef Apply

  if (type_of_loader != bios_loader) { /* DOS bootloader */
      gujin_param->attrib.read_only_param = 1;
      if (gujin_param->attrib.bios_only) {
	  printf ("ERROR: compilation options includes (SETUP & BIOS_ONLY)\n"
		  " so cannot produce a DOS executable. Recompile without.\n");
	  return 1;
	  }
      }
  if (gujin_param->attrib.vga_support) {
      unsigned char i;
      for (i = 0; i < nbof(VGA_valid_mode); i++)
	  gujin_param->VGA_valid_mode[i] = VGA_valid_mode[i];
      }
    else {
      unsigned char i;
      for (i = 0; i < nbof(VGA_valid_mode); i++)
	  if (VGA_valid_mode[i] != 0xFF)
	      break;
      if (i < nbof(VGA_valid_mode))
	  printf ("Warning: ignoring excluded video modes "
		  "(VGA_SUPPORT not compiled in).\n");
      }
  if (type_of_loader != bios_loader) {
      /* set the '+' as the display char */
//      *gujin_param->dot_msg = '+'; // I better like sigma for checksum:
	*gujin_param->dot_msg = 228;
      }


  /* Chain blocks and checksum : */
  nbblock = chs_chain (bootchain,
			boot2_start_lba + 2,
			filesize / dsk_sectorsize - 2,
			dma_frontier,
			dsk_nbhead, dsk_nbsector, dsk_biosnb,
			64 * 1024 / dsk_sectorsize);
  if (nbblock > instboot_info->bootchain_nbblock) {
      printf ("cannot fit %u blocks in %u, increase bootchain[] size "
		"in boot.c!\n",
		nbblock, instboot_info->bootchain_nbblock);
      return 1;
      }
    else
      printf ("Info: using %u element (max %u), i.e %u bytes for bootchain[]\n",
		nbblock, instboot_info->bootchain_nbblock,
		nbblock * sizeof(bootchain[0]));

  chkadr = next_checksum ((unsigned short *)(file + filesize),
				bootchain, dsk_sectorsize,
				instboot_info->bootchain_adr);

  if (chkadr != file + 2*512) {
      printf ("Checksum for chain does not cover all of boot2!\n");
      return 1;
      }

  chs_first (file, instboot_info->boot1param_adr, boot2_start_lba + 1,
	     dsk_nbhead, dsk_nbsector, dsk_biosnb);

  boot1param->bootloader2.header.nbword = 512 / 2;

  if (first_checksum (file,
			instboot_info->start_adr,
			instboot_info->boot1param_adr,
			instboot_info->bootend_adr,
			instboot_info->bootchain_adr,
			instboot_info->deltaseg,
			instboot_info->patch1adr,
			instboot_info->patch2adr,
			instboot_info->patch3adr,
			instboot_info->patch4adr,
			type_of_loader) != 0) {
      printf ("Error in first_checksum()!\n");
      return 1;
      }

  /* Set up floppy first sector information, not covered by checksum: */
  /* Set it also for DOS bootloaders to write "Gujin1" and "Gujin2" which
     are displayed when software start, and also for checksum.
     if (type_of_loader == bios_loader) */ {
      bootbefore_type *bb = (bootbefore_type *)(file + 3);
      unsigned short i;

      for (i = 0; i < sizeof(bb->String); i++)
	  bb->String[i] = "Gujin1\0\0"[i];	/* HAS TO be zero terminated */
      bb->Bytepersector = 512; /* it can provide zero terminator of previous
				string if multiple of 256 in little endian... */
      bb->Sectorpercluster = dsk_nbsectorpercluster;
      bb->Reservedsector = 1;	/* not usefull or not working ? */
      bb->NbFAT = dsk_nbfat;
      bb->NbRootdirentry = dsk_nbrootdirentry;
      if (dsk_nbtotalsector < 0x10000) {
	  bb->NbTotalsector = dsk_nbtotalsector;
	  bb->NbTotalsector2 = 0;
	  }
	else {
	  bb->NbTotalsector = 0;
	  bb->NbTotalsector2 = dsk_nbtotalsector;
	  }
      bb->Mediadescriptor = dsk_mediadescriptor;
      bb->NbSectorperFAT = dsk_nbsectorperfat;
      bb->NbSectorpertrack = dsk_nbsector;
      bb->NbHead = dsk_nbhead;
      bb->NbHiddensector = 0;	/* not usefull or not working for floppy ? */
      bb->PhysicaldriveNb = 0;
      bb->Reserved = 0;
      bb->Signaturebyte0x29 = 0x29;
      bb->VolumeserialNb = 0x39394C45;
      for (i = 0; i < sizeof(bb->Volumelabel); i++)
	  bb->Volumelabel[i] = "gujin2\0\0\0\0\0"[i]; /* zero terminated */
      if (dsk_nbtotalsector < 4096)
	  for (i = 0; i < sizeof(bb->FileSysName); i++)
	      bb->FileSysName[i] = "FAT12   "[i];   /* 8 char blank padded */
	else
	  for (i = 0; i < sizeof(bb->FileSysName); i++)
	      bb->FileSysName[i] = "FAT16   "[i];   /* 8 char blank padded */
      }

  /*
   * Everything is set up, begin the write:
   */
  if (type_of_loader == dos_com_loader && statbuf.st_size >= 64*1024)
      printf ("\nWARNING: Generating DOS boot.com file "
	      "bigger than 64 Kbytes!\n");

  if (type_of_loader == dos_exe_loader) {
      struct dos_exe_str {
	  unsigned char  exe_signature[2];
	  unsigned short image_length_mod_512;
	  unsigned short image_length_divroundup_512;
	  unsigned short relocation_nb;
	  unsigned short header_size_div_16;
	  unsigned short min_added_memory_div_16;
	  unsigned short max_added_memory_div_16;
	  unsigned short stack_seg_offset;
	  unsigned short init_sp;
	  unsigned short checksum; /* ~SIGMA(words16bit) */
	  unsigned short init_ip;
	  unsigned short code_seg_offset;
	  unsigned short relocation_offset;
	  unsigned short overlay_number; /* 28 bytes up to (including) there */
//#define LONG_EXE_HEADER /* some OS do not load it else ? */
#ifdef LONG_EXE_HEADER
	  unsigned	 relocation_table[(512 - 28) / 4];
#else
	  unsigned	 relocation_table[1];  /* to have a total of 32 bytes */
#endif
	  } __attribute__ ((packed)) dos_exe = {
	  exe_signature:		{'M', 'Z'},
	  /* The file created is rounded up to 512 bytes due to checksums: */
	  image_length_mod_512:		0,
	  /* I do not understand _why_ I need +1 here on a real DOS6.20: */
	  image_length_divroundup_512:	filesize/512 + 1,
	  relocation_nb:		0,
	  header_size_div_16:		sizeof (struct dos_exe_str) / 16,
	  /*
	   * min_added_memory_div_16: FIXME if BIG_MALLOC
	   * or ZLIB_BIG_MALLOC are defined, now set to 128K total
	   * It should be, without BIG_MALLOC nor ZLIB_BIG_MALLOC:
	   * (64*1024 - instboot_info->sbss) / 16, / * size of BSS * /
	   */
	  min_added_memory_div_16:	128 * 1024 / 16,
	  max_added_memory_div_16:	0xFFFF,
	  stack_seg_offset:		instboot_info->deltaseg,
	  init_sp:			0,
	  checksum:			0, /* no need to set it */
	  init_ip:			0, /* instboot_info->start_adr, */
	  code_seg_offset:		0, /* %cs = program load address */
	  relocation_offset:		offsetof (struct dos_exe_str,
							 relocation_table),
	  overlay_number:		0
	  };

      if (fwrite(&dos_exe, 1, sizeof (dos_exe), fout) != sizeof (dos_exe)) {
	  printf ("Cannot write the EXE descriptor block (drive full or "
		  "Read Only?)\n");
	  return 1;
	  }
      }

  if (fwrite(file, 1, 512, fout) != 512) {
      printf ("cannot write the first sector (drive full or Read Only?)\n");
      return 1;
      }

  if (type_of_loader == bios_loader) {
      unsigned short nbfilecluster = DIVROUNDUP((filesize - 512) / 512,
						dsk_nbsectorpercluster) - 1;
      unsigned (*write_fat) (FILE *fout, unsigned char dsk_mediadescriptor,
			     unsigned short nbfilecluster,
			     unsigned short nbsectorperfat,
			     unsigned short nbclusteronfat);
      unsigned short nbfat;

      if (dsk_nbtotalsector < 4096)
	  write_fat = write_fat_12;
	else
	  write_fat = write_fat_16;

      for (nbfat = 1; nbfat <= dsk_nbfat; nbfat++) {
	  if ((*write_fat) (fout, dsk_mediadescriptor, nbfilecluster,
				dsk_nbsectorperfat, dsk_nbclusteronfat) != 0) {
	      printf ("cannot write the FAT number %u (drive full?)\n",
			nbfat);
	      return 1;
	      }
	  }

      if (write_root(fout, dsk_nbrootdirentry, filesize - 512,
			dsk_nbsector, dsk_nbcyl) != 0) {
	  printf ("cannot write the ROOT directory (drive full?)\n");
	  return 1;
	  }
      }

  if (fwrite(file + 512, 1, filesize - 512, fout) != filesize - 512) {
      printf ("cannot write the complete array (drive full?)\n");
      return 1;
      }
  if (fclose (fout) == EOF) {
      printf ("cannot close output file: %s\n", fout_name);
      return 1;
      }
  return 0;
  }
