/* libextra.c */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"
#if SETUP & EXTRA_SEGMENT	/* before other includes */
#define CALLING_FROM_EXTRASEG
#endif

#include "library.h"
#include "libextra.h"
#include "debug.h"
#include "bios.h"	/* _BIOS_nbtick(), _BIOS_equipment_flags() */
#include "boot.h"	/* BOOT1_COM_port() */
#include "messages.h"
#include "user.h"

static inline void init_processor_str (struct processor_str *proc)
  {
  if (!is_80486()) {
      proc->family = 3;
      proc->brand = proc_nocpuid;
      proc->stdflags = 0;
      proc->extraflags = 0;
      }
    else if (!has_cpuid ()) {
      proc->family = 4;
      proc->brand = proc_nocpuid;
      proc->stdflags = 0;
      proc->extraflags = 0;
      }
    else {
      extern const union cpuid_string cpuid_string_array[8];
      unsigned short i;
      union cpuid_string cpuidstr;
      int cpuid_max = cpuid0 (&cpuidstr);

      if ((cpuid_max & 0xFFFFFF00) == 0x00000500) {
	  /* pre-B0 step Intel P5 processors  */
	  proc->brand = proc_intel;
	  proc->family = 5;
	  cpuid_max &= 0xFF;
	  }
	else {
	  for (i = 0; i < nbof (cpuid_string_array); i++)
	    if (   cpuidstr.number.str1 == cpuid_string_array[i].number.str1
		&& cpuidstr.number.str2 == cpuid_string_array[i].number.str2
		&& cpuidstr.number.str3 == cpuid_string_array[i].number.str3) {
		proc->brand = proc_intel + i;
		break;
		}
	  if (i == nbof (cpuid_string_array))
	      proc->brand = proc_unknown;
	  }

      if (cpuid_max != 0) {
	  union flags_tfms_union {
	      struct cpuid_str data;
	      struct cpuid_extended_str edata;
	      struct {
		  unsigned tfms; /* type/family/model/stepping */
		  unsigned flags;
		  } fields;
	      } flags_tfms;
	  int cpuid_emax;

	  flags_tfms.data = cpuid1();
	  proc->family = flags_tfms.data.family;
	  proc->stdflags = flags_tfms.fields.flags;

	  cpuid_emax = cpuid8000 (&cpuidstr);
	  if (cpuid_emax != 0) {
	      flags_tfms.edata = cpuid8001();
	      proc->extraflags = flags_tfms.fields.flags;
	      if (cpuid_emax > 4)
		  strcpy (proc->longname, cpuid_estring (
			(union cpuid_extended_string *)proc->longname));
		else
		  proc->longname[0] = '\0';
	      }
	    else {
	      proc->extraflags = 0;
	      proc->longname[0] = '\0';
	      }
	  }
      }
  }

static inline void display_processor (struct processor_str *processor)
  {
#if !(SETUP & QUITE)
  static const char * const brand[] = {
      "NoCpuId", "Intel", "AMD", "UMC", "Cyrix",
      "NexGen", "IDT", "Rise", "Transmeta", "Unknown"
      };

  PRINTF (PROCESSOR_MSG, brand[processor->brand], processor->family);
  if (processor->calibrate_rdtsc != 0)
      PRINTF (AT_FREQUENCY, (processor->calibrate_rdtsc + 500000) / 1000000);
  if ((processor->stdflags & 0x1) == 0)
      PRINT (WITHOUT_FPU);
#endif
  }

GENERATE_VOID_CODE2EXTRA_STUB(detect_processor,
				struct processor_str *proc)

SEGMENT_EXTRA(detect_processor) void
detect_processor (struct processor_str *processor)
  {bound_stack();{
  init_processor_str (processor);

  if (   processor->family <= 4 /* unknown, 386, 486 */
      && _BIOS_equipment_flags().coprocessor80x87installed)
      processor->stdflags |= 0x1;	/* set the FPU flag */

  if (processor->stdflags & 0x10 || processor->extraflags & 0x10) {
      /* Has rdtsc assembly function */
      unsigned nbtick, calibrate_rdtsc;
      unsigned long long rdtsc_start = rdtsc(); /* start it just in case... */

      calibrate_rdtsc = 0x100000;
      nbtick = _BIOS_nbtick();
      while (_BIOS_nbtick() == nbtick && --calibrate_rdtsc) ;

      if (calibrate_rdtsc == 0) {
	  DBG (("Ticks are not increased - cannot measure Up freq.\r\n"));
	  processor->calibrate_rdtsc = 0;
	  }
	else if (rdtsc_start == rdtsc()) {
	  DBG (("RDTSC is not increased - cannot measure Up freq.\r\n"));
	  processor->calibrate_rdtsc = 0;
	  }
	else {
	  unsigned rem, div = 55;

	  nbtick = _BIOS_nbtick();
	  while (_BIOS_nbtick() == nbtick) ;
	  rdtsc_start = rdtsc();
	  nbtick += 10;	/* incremented every 55 ms */
	  while (_BIOS_nbtick() <= nbtick) ; /* 550 ms */
	  calibrate_rdtsc = ull_div_ul ((rdtsc() - rdtsc_start) * (1000 / 10),
					div, &rem);
	  processor->calibrate_rdtsc = calibrate_rdtsc;
	  }
      }
    else
      processor->calibrate_rdtsc = 0;

  display_processor (processor);
  }}

#if USER_SUPPORT != 0
GENERATE_CODE2EXTRA_STUB(get_number, unsigned,
				unsigned deflt, unsigned max)

SEGMENT_EXTRA(get_number) unsigned
get_number (unsigned deflt, unsigned max)
  {
  unsigned short key, index;
  unsigned current = deflt;
  char tmp[16], *ptr;

  static const unsigned short numpad_digit_keycode[] = {
	NUMPAD_DIGIT_KEYCODE(0),	NUMPAD_DIGIT_KEYCODE(1),
	NUMPAD_DIGIT_KEYCODE(2),	NUMPAD_DIGIT_KEYCODE(3),
	NUMPAD_DIGIT_KEYCODE(4),	NUMPAD_DIGIT_KEYCODE(5),
	NUMPAD_DIGIT_KEYCODE(6),	NUMPAD_DIGIT_KEYCODE(7),
	NUMPAD_DIGIT_KEYCODE(8),	NUMPAD_DIGIT_KEYCODE(9)
	};
  static const unsigned short digit_keycode[] = {
	DIGIT_KEYCODE(0),	DIGIT_KEYCODE(1),
	DIGIT_KEYCODE(2),	DIGIT_KEYCODE(3),
	DIGIT_KEYCODE(4),	DIGIT_KEYCODE(5),
	DIGIT_KEYCODE(6),	DIGIT_KEYCODE(7),
	DIGIT_KEYCODE(8),	DIGIT_KEYCODE(9)
	};

  unsigned short esc_keycode, backspace_keycode1, backspace_keycode2,
		 return_keycode1, return_keycode2;

  if (BOOT1_COM_port() >= 0) {
      esc_keycode = '\033';
      backspace_keycode1 = '\010';
      backspace_keycode2 = '\177';
      return_keycode1 = '\r';
      return_keycode2 = '\n';
      }
    else {
      esc_keycode = ESC_KEYCODE();
      backspace_keycode1 = BACKSPACE_KEYCODE();
      backspace_keycode2 = BACKSPACE_KEYCODE();
      return_keycode1 = RETURN_KEYCODE();
      return_keycode2 = KP_RETURN_KEYCODE();
      }

  for (;;) {
      if (current > max)
	  current %= 10; /* keep last digit entered (restart entry) */
      sprintf (tmp, "%u", current);
      print (tmp);
      print (" \010"); /* if number has reduced in size */
      key = UI_getkey (30 * TICKS_PER_SECOND);
      if (key == esc_keycode || key == TIMEOUT_KEYCODE())
	  return deflt;
      if (key == return_keycode1 || key == return_keycode2)
	  return current;
      ptr = tmp;
      while (*ptr)
	  *ptr++ = '\010'; /* ^H, backspace for each digit printed */
      print (tmp);
      if (key == backspace_keycode1 || key == backspace_keycode2) {
	  current /= 10;
	  continue;
	  }
      for (index = 0; index < nbof(numpad_digit_keycode); index++)
	  if (key == numpad_digit_keycode[index]) /* ext keyboard */
	      break;
      if (index < nbof(numpad_digit_keycode)) {
	  current = (current * 10) + index;
	  continue;
	  }
      for (index = 0; index < nbof(digit_keycode); index++)
	  if (key == digit_keycode[index]) /* std keyboard */
	      break;
      if (index < nbof(numpad_digit_keycode)) {
	  current = (current * 10) + index;
	  continue;
	  }
      if (key >= '0' && key <= '9') { /* serial */
	  current = (current * 10) + key - '0';
	  continue;
	  }
      }
  }
#endif

/**
 ** Main variable declaration, init to BSS:
 **/
struct processor_str processor;

