/* library.c */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"
#include "library.h"
#include "debug.h"
#if USER_SUPPORT != 0
#include "user.h"	/* UI_putstr */
#endif
#include "boot.h"	/* BOOT1_DOS_running() */
#include "bios.h"	/* _BIOS_getBaseMemory() */
#include "dos.h"	/* DOS_SetInterruptVector() */
#include "messages.h"	/* MSG_EXITING */

/*
 * Define the timer interrupt we manage, may be 0x1C or 0x08,
 * but keep it to 0x08:
 */
#ifdef TREAT_TIMER
#define TIMER_IRQ 0x08
#endif

/*
 * Define if we try to use the BIOS functions to try to
 * enable A20. It is not tested because I do not have such a BIOS,
 * and anybody has already reported it working.
 */
//#define BIOS_ENABLE_A20

/*
 * This is the address we use to detect that A20 is enabled or
 * disabled, by reading/writing at 0x0000:A20_TEST_ADDR and
 * reading at 0xFFFF:A20_TEST_ADDR+0x10
 */
//#define A20_TEST_ADDR 0x007C
#define A20_TEST_ADDR 0x200

/**
 ** Some standart functions:
 **/

const union cpuid_string cpuid_string_array[8] = {
		{ string: "GenuineIntel" },
		{ string: "AuthenticAMD" },
		{ string: "UMC UMC UMC " },
		{ string: "CyrixInstead" },
		{ string: "NexGenDriven" },
		{ string: "CentaurHauls" },
		{ string: "RiseRiseRise" },
		{ string: "GenuineTMx86" }
		};

static const char itoa_array_maj[] = "0123456789ABCDEF";

#ifdef PRINTF_LOWERCASE_HEXADECIMAL
static const char itoa_array_min[] = "0123456789abcdef";
#endif

/*
 * More or less subroutine of printf() rewrite, limited to stack growing
 * down, and limited to formats '%%', '%c', '%s', '%d', '%x', '%X', '%u',
 * without size like '%5.5d'. "%D"/"%U" is treated as "%d"/"%u" but adds
 * comma every 3 digit like: 1,000,000
 *
 * There is an added format: '%$' - it is there to limit the
 * size of the string written - it should point to the last char
 * which can be written, i.e. the last possible '\0' char.
 *
 * Used also in user.c::BOOT1_printf
 */
char *prtf (char *buffer, char *endbuffer, const char *format, void **param)
  {bound_stack();{
  char *dest = buffer;
  const char *itoa_array;

  endbuffer -= 1;	/* for the ending zero */
  while (*format && dest < endbuffer) {
      if (*format != '%')
	  *dest++ = *format++;
	else
	  switch (*++format) {
	    default :
	      *dest++ = '%';
	      if (dest < endbuffer)
		  *dest++ = *format++;
	      break;
	    case '$':
	      endbuffer = *(char **)param++;
	      format++;
	      break;
	    case '%':
	      format++;
	      /* fallthrough */
	    case '\0':
	      *dest++ = '%';
	      break;
	    case 'c':
	      *dest++ = *(char *)param++;
	      format++;
	      break;
	    case 's': {
	      char *cptr = (char *)*param++;
	      while (*cptr != '\0' && dest < endbuffer)
		  *dest++ = *cptr++;
	      }
	      format++;
	      break;
	    case 'D':
	    case 'd':
	      if (*(int *)param < 0) {
		  *(int *)param = -*(int *)param;
		  *dest++ = '-';
		  }
	      if (dest >= endbuffer)
		  break;
	      /* fallthrough */
	    case 'U':
	    case 'u': {
	      char buff[10+3], *buffptr = buff; /* max 32bit: 10digit+3comma */
	      unsigned todisplay = *(unsigned *)param++, tmp, cpt = 0;
	      for (;;) {
		  tmp = todisplay/10;
		  *buffptr++ = itoa_array_maj[todisplay - 10*tmp];
		  todisplay = tmp;
		  if (todisplay == 0)
		      break;
		  if (*format == 'U' || *format == 'D')
		      if (++cpt % 3 == 0)
		   *buffptr++ =',';
		  }
	      while (--buffptr >= buff && dest < endbuffer)
		  *dest++ = *buffptr;
	      }
	      format++;
	      break;
	    case 'x':
#ifdef PRINTF_LOWERCASE_HEXADECIMAL
	      itoa_array = itoa_array_min;
	      goto common_hex;
#endif
	    case 'X':
	      itoa_array = itoa_array_maj;
	      goto common_hex;

	    common_hex:	{
	      char buff[8], *buffptr = buff;
	      unsigned todisplay = *(unsigned *)param++;
	      do {
		  *buffptr++ = itoa_array[todisplay % 16];
		  todisplay /= 16;
		  } while (todisplay != 0);
	      while (--buffptr >= buff && dest < endbuffer)
		  *dest++ = *buffptr;
	      }
	      format++;
	      break;
	    }
      }
  *dest = '\0';

  return buffer;
  }}

#if USER_SUPPORT != 0
#define PRINTSTRING	UI_putstr
#else
#define PRINTSTRING	BOOT1_putstr
#endif

GENERATE_EXTRA2CODE_STUB(printf, int, const char *format, ...)
int printf (const char *format, ...)
  {bound_stack();{
  char buffer[256];

  PRINTSTRING (prtf (buffer, buffer + sizeof(buffer) - 1, format,
		  (void **)((unsigned)&format + sizeof(const char *)) ));
  return 0;
  }}

#if !(SETUP & QUITE)
GENERATE_EXTRA2CODE_STUB(verbose_printf, int, const char *format, ...)
int verbose_printf (const char *format, ...)
  {bound_stack();{
  if (get_gujin_param_attrib().verbose) {
      char buffer[256];

      PRINTSTRING (prtf (buffer, buffer + sizeof(buffer) - 1, format,
		  (void **)((unsigned)&format + sizeof(const char *)) ));
      }
  return 0;
  }}
#endif

GENERATE_VOID_EXTRA2CODE_STUB(print, const char *msg)
void print (const char *msg)
  {
  PRINTSTRING (msg);
  }

#if !(SETUP & QUITE)
GENERATE_VOID_EXTRA2CODE_STUB(verbose_print, const char *msg)
void verbose_print (const char *msg)
  {
  if (get_gujin_param_attrib().verbose)
      PRINTSTRING (msg);
  }
#endif

GENERATE_VOID_EXTRA2CODE_STUB(puts, const char *msg)
void puts (const char *msg)
  {
  PRINTSTRING (msg);
  PRINTSTRING ("\r\n");
  }

#if !(SETUP & QUITE)
GENERATE_VOID_EXTRA2CODE_STUB(verbose_puts, const char *msg)
void verbose_puts (const char *msg)
  {
  if (get_gujin_param_attrib().verbose) {
      PRINTSTRING (msg);
      PRINTSTRING ("\r\n");
      }
  }
#endif

GENERATE_EXTRA2CODE_STUB(sprintf, int, char *dest, const char *format, ...)
int sprintf (char *dest, const char *format, ...)
  {
  prtf (dest, (char *)UINT_MAX, format,
		 (void **)((unsigned)&format + sizeof(const char *)));
  return strlen (dest);
  }

GENERATE_VOID_EXTRA2CODE_STUB(BIOS_killme, void)
__attribute__ ((noreturn))
void BIOS_killme (void)
  {
  DBG (("\r\n%s called: restoring vectors", __FUNCTION__));
#ifdef TREAT_EXCEPTION
  restore_vectors ();	/* cannot be called from EXTRASEG */
#endif
  puts (MSG_EXITING);

  DBG_END();
  if (BOOT1_DOS_running()) {
      DBG ((" Trying to stop 'exe' executable"));
      _DOS_stop_exe ();
      DBG ((" Trying to stop 'com' executable"));
      _DOS_stop_com ();
      DBG ((" not working!"));
      }
    else
      DBG ((" not in DOS environment"));
  DBG ((" Trying to stop DOSEMU"));
  _DOSEMU_stop ();
  DBG ((" Trying a segment violation"));
  _BIOS_sigsegv ();
  DBG ((" Trying the HLT instruction"));
  _BIOS_halt ();
  DBG ((" starting infinite loop"));
  while (1) ;
  }

/**
 ** Exception related things:
 **/

#ifdef TREAT_EXCEPTION
/*
 * "bound" checking and exceptions related things:
 */

#if DEBUG & DEBUG_ADDR
long_bound_type data_limit = {
    low_limit: (unsigned) _srodata,
    high_limit: 0xFFFF		/* <= high_limit + sizeof operation */
    };

volatile const char *BoundMsg; /* init at 0 : BSS */
#endif /* DEBUG_ADDR */

#if DEBUG & DEBUG_STACK
/* This cannot be initialised at link time because fields are 16bits: */
short_bound_type stack_limit;   /* stack < 32 Kbytes */
#endif /* DEBUG_STACK */

static struct {
    unsigned initialised;
    unsigned divide_error,
	     instruction_invalid,
	     stack_segment_fault,
	     general_protection;
#if !(SETUP & BIOS_ONLY)
    unsigned  control_break;
#endif
#if DEBUG & (DEBUG_STACK|DEBUG_ADDR)
    unsigned  bound_error;
#endif
#ifdef TREAT_COUNTER_TRACE_FLAG
    unsigned  debug;
#endif
    } savedvector; /* init at 0 : BSS */

static farptr __attribute__ (( regparm(1) ))
get_vector (unsigned irq)
  {
  if (BOOT1_DOS_running()) {
#if 0
      farptr vector;
      if (   _DPMI_int0x31_available()
	  && _DPMI_get_real_mode_interrupt_vector (irq, &vector) == 0)
	  return vector;
#endif
      return DOS_GetInterruptVector (irq);
      }
    else
      return peekl (4*irq);
  }

static farptr
set_vector_farptr (unsigned irq, farptr fct_faradr)
  {
  if (BOOT1_DOS_running()) {
      farptr returned = get_vector (irq);
#if 0
      if (   !_DPMI_int0x31_available()
	  || _DPMI_set_real_mode_interrupt_vector (irq, fct_faradr) != 0)
#endif
	  DOS_SetInterruptVector (irq, fct_faradr);
      return returned;
      }
    else
      return xchgl (4*irq, fct_faradr);
  }


static unsigned set_vector (unsigned irq, void *fctadr)
  {
  return set_vector_farptr (irq, code_adr (fctadr));
  }

/*
 * Setup vectors, only once if restart:
 * Called from assembly, do not prototype it.
 */
void setup_vectors (void)
  {
  extern unsigned char  divide_error[], instruction_invalid[],
			stack_segment_fault[], general_protection[];

  if (savedvector.initialised++ != 0)
      return;
  savedvector.divide_error = set_vector (0, divide_error);
  savedvector.instruction_invalid = set_vector (6, instruction_invalid);
  savedvector.stack_segment_fault = set_vector (12, stack_segment_fault);
  savedvector.general_protection = set_vector (13, general_protection);
#ifdef TREAT_TIMER
  {
  extern unsigned char  timer_handler_relay[], timer_handler_oldptr[];
  codeseg_pokel (timer_handler_oldptr, get_vector(TIMER_IRQ));
  set_vector (TIMER_IRQ, timer_handler_relay);
  }
#endif
#ifdef TREAT_COUNTER_TRACE_FLAG
  {
  extern unsigned char  TFflagHandler[];
  savedvector.debug = set_vector (1, TFflagHandler);
  }
#endif
#if DEBUG & DEBUG_STACK
  /*
   * Initialise the "bound" 16 bits (signed) stack limits:
   */
  if ((unsigned)_edata > 0x8000)
      stack_limit.low_limit = (unsigned short)_edata + 2 + stack_safety;
    else
      stack_limit.low_limit = 0x8000 + 2 + stack_safety;
  stack_limit.high_limit = 0xFFFC;
#endif
#if DEBUG & (DEBUG_STACK|DEBUG_ADDR)
  {
  extern unsigned char  bound_error[];
  savedvector.bound_error = set_vector (5, bound_error);
  }
#endif
#if !(SETUP & BIOS_ONLY)
  {
  extern unsigned char  control_break[];
  /* Disable managment by DOS of ^break/^C because we cannot stop
     current software as easily (use a setjmp? how about himem?) */
  savedvector.control_break = set_vector (0x23, control_break);
  }
#endif
  }

GENERATE_VOID_EXTRA2CODE_STUB(restore_vectors, void)
void restore_vectors (void)
  {
  if (savedvector.initialised-- == 0)
      return;

  set_vector_farptr (0, savedvector.divide_error);
  set_vector_farptr (6, savedvector.instruction_invalid);
  set_vector_farptr (12, savedvector.stack_segment_fault);
  set_vector_farptr (13, savedvector.general_protection);
#ifdef TREAT_TIMER
  {
  extern unsigned char timer_handler_oldptr[];
  set_vector_farptr (TIMER_IRQ, codeseg_peekl (timer_handler_oldptr));
  }
#endif
#ifdef TREAT_COUNTER_TRACE_FLAG
  set_vector_farptr (1, savedvector.debug);
#endif
#if DEBUG & (DEBUG_STACK|DEBUG_ADDR)
  set_vector_farptr (5, savedvector.bound_error);
#endif
#if !(SETUP & BIOS_ONLY)
  set_vector_farptr (0x23, savedvector.control_break);
#endif
  }

static void puthex (const char *msg, unsigned hex, unsigned crlf)
  {
  char buff[8+2+1], *buffptr = &buff[nbof(buff) - 1];

  *buffptr-- = '\0';
  do {
      *buffptr-- = itoa_array_maj[hex % 16];
      hex /= 16;
      } while (hex != 0);
  *buffptr-- = 'x',
  *buffptr = '0';
  print (msg);
  if (crlf)
      puts (buffptr);
    else
      print (buffptr);
  }

#if BUILD_TESTER()
volatile unsigned segment_door_key; /* init at 0 : BSS */
#endif
#ifdef USE_SETJUMP_KEY
volatile unsigned setjump_key; /* init at 0 : BSS */
setjmp_buf setjump_buffer;
#endif

/* Special calling convention:
 *   DO NOT MODIFY parameters here, if you want to return!
 */
void __attribute__ ((cdecl, regparm (0)
#if !BUILD_TESTER()
	, noreturn
#endif
	))
exceptions (unsigned edi, unsigned esi, unsigned ebp, unsigned initsp,
	    unsigned ebx, unsigned edx, unsigned ecx, unsigned eax,
	    unsigned ds,  unsigned es,
	    unsigned vector,
	    unsigned cs_ip, unsigned flags)
  {
#if DEBUG & DEBUG_STACK
  unsigned short save_stack_low_limit, newlimit;
  unsigned save_esp;
#endif
  const char *string;

#if BUILD_TESTER()
  /* Process immediately the "detect segment door" open/close
     if the key is here: (#GP or #SS or #NM) */
  if (   segment_door_key == 0x76913010
      && (vector == 13 || vector == 12 || vector == 7)) {
      unsigned i = 10;
      farptr codeptr = (farptr) cs_ip;

      segment_door_key = -segment_door_key; /* only once */
      while (--i && peekb (codeptr++) != 0x90) ; /* look for nop */
      if (i != 0) {
	  *(volatile unsigned *)&cs_ip = codeptr;
	  return;
	  }
      }
#endif

#ifdef USE_SETJUMP_KEY
  /* Now rescue the ISO9001 video BIOSes which executes invalid instructions
	 or divisions by zero when trying to set invalid modes */
  if (setjump_key == 0x45746965 && setjump_buffer.cs != 0) {
      setjump_key = -setjump_key;
      longjmp (&setjump_buffer);
      }
#endif

#if DEBUG & DEBUG_STACK
  save_esp = getesp();
  /* leave some space to treat the interrupt: */
  save_stack_low_limit = stack_limit.low_limit;
#if SETUP & BIG_MALLOC
  newlimit = (unsigned short)_edata +2;
#else
  {
  extern unsigned end_malloc;
  newlimit = end_malloc + 2;
  }
#endif
  if (newlimit < 0x8002)
      newlimit = 0x8002;
  stack_limit.low_limit = newlimit;  /* remove safety */
  asm ("" : : : "memory");	/* force update of memory */
#endif /* DEBUG_STACK */

#if USER_SUPPORT != 0
  UI_setcursor (0, 0);
  UI_setfgcolor (UI.stdcolor[red]);
  UI_setbgcolor (UI.stdcolor[lightgray]);
#endif

  switch (vector) {
    case 0:
	string = "DIVISION BY ZERO";
	break;
    case 5:
#if (DEBUG & DEBUG_ADDR) && (DEBUG & DEBUG_STACK)
	string = "BOUND: STACK OVERFLOW OR POINTER OUT OF RANGE";
#elif DEBUG & DEBUG_ADDR
	string = "BOUND: POINTER OUT OF RANGE";
#elif DEBUG & DEBUG_STACK
	string = "BOUND: STACK OVERFLOW";
#else
	string = "BOUND: Who has executed this instruction?";
#endif
	break;
    case 6:
	string = "INVALID INSTRUCTION";
	break;
    case 12:
	string = "STACK SEGMENT FAULT";
	break;
    case 13:
	string = "GENERAL PROTECTION";
	break;
    default:
	string = "??UNKNOWN??";
	break;
    }
  DBG (("\r\n"));
  DBG ((string));
  DBG ((" at address 0x%X\r\n", cs_ip));
  print (string);
  puthex (" at address ", cs_ip, 1);

#if !(SETUP & QUITE)
  if (vector == 5) {
#if DEBUG & DEBUG_STACK
      puthex (" stack limit low: ", save_stack_low_limit, 0);
      puthex (" stack limit high: ", (unsigned short)stack_limit.high_limit, 0);
      puthex (" value:", save_esp, 1);
#endif /* DEBUG_STACK */

#if DEBUG & DEBUG_ADDR
      puthex (" data limit low: ", data_limit.low_limit, 0);
      puthex (" data limit high: ", data_limit.high_limit, 1);

      if (BoundMsg != 0 && strlen((char *)BoundMsg) < 80) {
	  print ("Bound Msg: ");
	  puts ((char *)BoundMsg);
	  }
#endif /* DEBUG_ADDR */
      }
    else if (vector == 6 || vector == 13) {
      puthex ("invalid code (check corruption with boot.asm): ",
		peekb(cs_ip), 0);
      puthex (", ", peekb(cs_ip+1), 0);
      puthex (", ", peekb(cs_ip+2), 0);
      puthex (", ", peekb(cs_ip+3), 0);
      puthex (", ", peekb(cs_ip+4), 0);
      puthex (", ", peekb(cs_ip+5), 1);
      }
#if 0
    else if (vector == 13) {
      /* There is no error code in real mode ! */
      puthex (" CR2 ", get_CR2(), 1); // seems not filled in real mode
      }
#endif

  puthex (" eax = ", eax, 0);
  puthex (" ebx = ", ebx, 0);
  puthex (" ecx = ", ecx, 0);
  puthex (" edx = ", edx, 1);

//  puthex (" initsp = ", initsp, 0);
  puthex (" edi = ", edi, 0);
  puthex (" esi = ", esi, 0);
  puthex (" ebp = ", ebp, 0);
  puthex (" flags = ", flags & 0xFFFF, 1);

  puthex (" cs = ", getcs(), 0);
  puthex (" ds = ", ds & 0x0000FFFF, 0);
  puthex (" es = ", es & 0x0000FFFF, 0);
  puthex (" ss = ", getss(), 0);
  puthex (" esp = ", getesp(), 1);

#if 0
  {
  farptr esp = (((unsigned)getss()) << 16) + getesp();
  puthex (" *(esp+00) = ", peekl(esp + 0x00), 0);
  puthex (" *(esp+04) = ", peekl(esp + 0x04), 0);
  puthex (" *(esp+08) = ", peekl(esp + 0x08), 0);
  puthex (" *(esp+0C) = ", peekl(esp + 0x0C), 1);
  puthex (" *(esp+10) = ", peekl(esp + 0x10), 0);
  puthex (" *(esp+14) = ", peekl(esp + 0x14), 0);
  puthex (" *(esp+18) = ", peekl(esp + 0x18), 0);
  puthex (" *(esp+1C) = ", peekl(esp + 0x1C), 1);
  puthex (" *(esp+20) = ", peekl(esp + 0x20), 0);
  puthex (" *(esp+24) = ", peekl(esp + 0x24), 0);
  puthex (" *(esp+28) = ", peekl(esp + 0x28), 0);
  puthex (" *(esp+2C) = ", peekl(esp + 0x2C), 1);
  puthex (" *(esp+30) = ", peekl(esp + 0x30), 0);
  puthex (" *(esp+34) = ", peekl(esp + 0x34), 0);
  puthex (" *(esp+38) = ", peekl(esp + 0x38), 0);
  puthex (" *(esp+3C) = ", peekl(esp + 0x3C), 1);
  }
#endif
#endif /* QUITE */

  enable_interrupt();	/* control alt del */
  {
  unsigned short vers, patch;
  if (_DOSEMU_check (&vers, &patch) == 0) {
      BOOT1_putstr ("Stoping DOSEMU...");
      _DOSEMU_stop ();
      BOOT1_putstr (" did not work!\r\n");
      }
//    else
      while (!BOOT1_DOS_running()) ; /* loop if not in DOS */
  }
  BIOS_killme() ;
  }

#endif /* TREAT_EXCEPTION */

#if !(SETUP & QUITE)
const char *const mem_name[] = {
    "E801", "AMI", "Phoenix", "8800", "nvram"
    };
#endif

GENERATE_EXTRA2CODE_STUB(detect_memory, enum detect_memory_enum,
			unsigned *basemem, unsigned *extendmem)
enum detect_memory_enum
detect_memory (unsigned *basemem, unsigned *extendmem)
  {
  unsigned short low16Mb_inKb, high16Mb_in64Kb, tmp;
  unsigned ebdaseg;

  DBG (("%s: ", __FUNCTION__));
  *basemem = _BIOS_getBaseMemory ();
  if (*basemem == 0) {
      DBG (("[_BIOS_getBaseMemory returns zero, use NVRAM] "));
      *basemem = ((unsigned short)getnvram(0x16) << 8) + getnvram(0x15);
      }
  DBG (("basemem %u KB, ", *basemem));
  if (!BOOT1_DOS_running()) {
      ebdaseg = _BIOS_getEBDA_segment();
      DBG (("EBDA: at segment 0x%X; ", ebdaseg));
      if (ebdaseg != 0 && *basemem > ebdaseg * 16) {
	  *basemem = ebdaseg * 16;
	  DBG ((" [correcting basemem = %u Kb] ", *basemem));
	  }
      }

  DBG (("_BIOS_getExtendedE801Memory: "));
  if (_BIOS_getExtendedE801Memory (&low16Mb_inKb, &high16Mb_in64Kb) == 0) {
      *extendmem = 64U * high16Mb_in64Kb + low16Mb_inKb;
      DBG (("OK: low16Mb: %u Kb, high16Mb: %u * 64Kb, total: %u Kb.\r\n",
	low16Mb_inKb, high16Mb_in64Kb, *extendmem));
      return mem_E801;
      }
    else
      DBG (("failed, try _BIOS_getAMIExtendedMemory: "));
  *extendmem = _BIOS_getAMIExtendedMemory();
  if (*extendmem != 0) {
      DBG (("OK: total: %u Kb.\r\n", *extendmem));
      return mem_AMI;
      }
    else
      DBG (("failed, try _BIOS_getPhoenixExtendedMemory: "));
  *extendmem = _BIOS_getPhoenixExtendedMemory();
  if (*extendmem != 0) {
      DBG (("OK: total: %u Kb.\r\n", *extendmem));
      return mem_Phoenix;
      }
    else
      DBG (("failed, try _BIOS_getExtendedMemory: "));
  if (_BIOS_getExtendedMemory (&tmp) == 0 && tmp != 0) {
      *extendmem = tmp;
      DBG (("OK: total: %u Kb.\r\n", *extendmem));
      return mem_8800;
      }
    else
      DBG (("failed, get from NVRAM: "));
  /* also 0x31/0x30:
   * extended memory size found above 1 megabyte during POST
   * Using NVRAM here is far to be compatible...
   */
  tmp = getnvram(0x18);
  *extendmem = (tmp << 8) + getnvram(0x17);
  DBG (("OK: total: %u Kb.\r\n", *extendmem));
  return mem_nvram;
  }

#if ASSEMBLY_TYPE == ASSEMBLY_DSES
int initial_open_door (void)
  {
  /* malloc is still not initialised here => no printf */
  if (BOOT1_DOS_running()) {
      unsigned short handle;

      if (   peekw_doorclosed (0x67*4) != 0
	  && DOS_OpenFile ("EMMXXXX0", DOS_open_read_only, &handle) == 0) {
	  farptr entry;
	  unsigned short errorcode;

	  DOS_CloseFile (handle, &errorcode);
	  BOOT1_putstr ("EMM386: ");
	  entry = EMM386_check();
	  if (entry != 0) {
	      unsigned char state = EMM386_get_state(entry) & 0x03;
	      if (state == 1)
		  BOOT1_putstr ("OFF");
		else {
		  if (state & 2)
		      BOOT1_putstr ("AUTO/");
		  if (state == 0)
		      BOOT1_putstr ("ON");
		   else
		      BOOT1_putstr ("OFF");
		  BOOT1_putstr (", switch to OFF: ");
		  if (EMM386_set_state (entry, 0x01) != 0x01)
		      BOOT1_putstr ("failed!");
		    else
		      BOOT1_putstr ("OK");
		  }
	      BOOT1_putstr ("\r\n");
	      }
	    else
	      BOOT1_putstr ("not found\r\n");
      }

      if (segment_door (1) != 0) { /* before the following xchgl() */
	  BOOT1_putstr (
	    "ERROR opening segment doors, either recompile without\r\n"
	    "define 'BIG_MALLOC' nor define 'ASSEMBLY_DSES', or disable\r\n"
	    "the memory manager EMM386/QEMM/Windows (even DOSEMU).\r\n");
	  _BIOS_wait (4 * 1000 * 1000); /* 4 seconds */
	  return 1;
	  }
      }
    else {
      if (segment_door (1) != 0) { /* before the following xchgl() */
	  BOOT1_putstr (
	    "ERROR opening segment doors, sorry please reboot,\r\n"
	    " and recompile without define 'BIG_MALLOC' nor\r\n"
	    " define 'ASSEMBLY_DSES', or disable the memory\r\n"
	    " manager EMM386/QEMM/Windows (even DOSEMU).\r\n");
	  while (1) ;
	  }
      }
  return 0;
  }
#endif /* ASSEMBLY_TYPE == ASSEMBLY_DSES */

/**
 ** A20 related stuff:
 **/

#define A20_TEST_LOW_ADDR	A20_TEST_ADDR
#define A20_TEST_HIGH_ADDR	(0xFFFF0000 + A20_TEST_ADDR + 0x10)

static void flush_8042 (unsigned timeout)
  {
  unsigned char val, data = 0;
#if DEBUG & DEBUG_A20
  unsigned short data_read = 0;
  unsigned copy_timeout = timeout;

  ADBG ((" [%s: ", __FUNCTION__));
#endif

  do {
      val = inb (0x64);	/* 1 ms per inb/outb */
      if (val & 1) {
	  short_delay (10);
	  data = inb (0x60);	/* so 2 ms per loop if val & 1 */
#if DEBUG & DEBUG_A20
	  data_read++;
#endif
	  }
	else if ((val & 2) == 0)
	  break;
      short_delay (10);
      } while (--timeout);

#if DEBUG & DEBUG_A20
  if (timeout == 0)
      ADBG ((" TIMEOUT, please report!"));
    else
      ADBG ((" done in %u loops", copy_timeout - timeout));
  if (data_read)
      ADBG ((" [inb (0x60) read % times, last 0x%X] ", data_read, data));
  ADBG ((", inb (0x64) = 0x%X] ", val));
#endif
  }

GENERATE_EXTRA2CODE_STUB(get_A20, unsigned, unsigned timeout)
unsigned get_A20 (unsigned timeout)
  {
  flush_8042 (500);
  outb (0x64, 0xD0);	/* write to port: read I/O port */
  while (((inb (0x64) & 1) == 0) && --timeout) ;
  if (timeout == 0)
      return 0xFFFFFFFFU;
  return inb(0x60) & 0x01;
  }

static inline unsigned enable_A20_with_BIOS (void)
  {
#ifdef BIOS_ENABLE_A20
  union {
      unsigned short      all;
      struct BIOS_A20_str bitfield;
      } switch_type;

  ADBG ((" [%s: ", __FUNCTION__));
  switch_type.bitfield = BIOS_A20_supported();
  ADBG (("BIOS returns 0x%X", switch_type.all));

  if (switch_type.all != 0) {
      unsigned short low, high;
      unsigned short tmp = BIOS_enable_A20();

      ADBG ((" enable -> 0x%X ", tmp));
      ADBG ((" is_enabled -> 0x%X ", BIOS_is_A20_enabled()));

      low = peekw_doorclosed (A20_TEST_LOW_ADDR);
      high = peekw_doorclosed (A20_TEST_HIGH_ADDR);

      ADBG ((", testing: "));
      if (low != high) {
	  ADBG (("OK] "));
	  return 0; /* working */
	  }
      pokew_doorclosed (A20_TEST_LOW_ADDR, ~low);
      if (high == peekw_doorclosed (A20_TEST_HIGH_ADDR)) {
	  pokew_doorclosed (A20_TEST_LOW_ADDR, low);
	  ADBG (("OK] "));
	  return 0; /* working */
	  }
      }
    else
      ADBG (("not supported by BIOS"));

  ADBG ((", Keyb "));
#else /* BIOS_ENABLE_A20 */
  ADBG (("BIOS not compiled in, Keyb "));
#endif /* BIOS_ENABLE_A20 */
  return 1;
  }

GENERATE_EXTRA2CODE_STUB(enable_A20, unsigned, unsigned timeout)
unsigned enable_A20 (unsigned timeout)
  {
  unsigned short saved, cpt;
  unsigned char val_io92;

#if DEBUG & DEBUG_A20
  unsigned copy_timeout = timeout;
  val_io92 = 0; /* warning */
#endif

  ADBG ((" [%s: ", __FUNCTION__));

  saved = peekw_doorclosed (A20_TEST_LOW_ADDR);

  if (peekw_doorclosed (A20_TEST_HIGH_ADDR) != saved && get_A20 (1000) != 0) {
      /* already enabled and not with the "fast A20" I/O 0x92 */
      ADBG (("already enabled] "));
#if defined (DEBUG) && ((DEBUG & DEBUG_OUTPUT) == SCREEN) && (USER_SUPPORT != 0)
      ADBG ((" PRESS A KEY TO CONTINUE ", UI_getkey(0)));
#endif
      return 0;
      }

  if (enable_A20_with_BIOS() == 0)
      return 0;

  /* Shall we try that first outb(0x64, x):
   DDh	sngl	disable address line A20 (HP Vectra only???)
		default in Real Mode
   DFh	sngl	enable address line A20 (HP Vectra only???)

     Shall we try then:
   outb (0x329, 0xdf or 0xdd)
     Shall we try then:
   outb (0xe0, 0xdd) and outb (0xe0, 0xdf) instead of 0x60
     How about:
   outb (0x3f20, 0x80) / outb (0x3f00, 0x10)
     or:
   outb (0x3f00, 0) / outb (0x3f00, 0x10)
     While we are here:
   outb (0x65, inb(0x65) & 0xfb) and wait bit 0x04 of inb(0x65)
   */

/* All the ADBG(get_A20) report A20 already open here,
   probably it is open by the "copy to extended memory"
   BIOS service... this part needs cleanning! */
  ADBG ((" [get_A20(1000) 1: 0x%X] ", get_A20 (1000)));

  if (inb(0x64) & 1) {
      ADBG ((" reset kbd because read data present: "));
      flush_8042 (500);
      outb (0x64, 0xAD);	/* disable keyboard */
      flush_8042 (500);
      outb (0x64, 0xAE);	/* enable keyboard */
      ADBG (("OK. "));
      }
  flush_8042 (500);	/* 1 second each if kbd abscent (read 0xFF) */
  outb (0x64, 0xD1);	/* write to port */
  flush_8042 (500);
  outb (0x60, 0xDF);	/* set A20 active, would be 0xDD to close it */
  flush_8042 (500);
  outb (0x64, 0xFF);	/* black magic */
  flush_8042 (500);

  ADBG ((" [get_A20(1000) 2: 0x%X] ", get_A20 (1000)));

  cpt = 2;
  for (;;) {
      unsigned short tmp;

      while (   --timeout
	     && (tmp = peekw_doorclosed  (A20_TEST_LOW_ADDR))
			== peekw_doorclosed (A20_TEST_HIGH_ADDR))
	  pokew_doorclosed (A20_TEST_LOW_ADDR, ~tmp);

      if (timeout != 0 || --cpt == 0)
	  break;
      val_io92 = inb(0x92);
      short_delay (10);
      ADBG (("did not work, try outb(0x92, 0x%X | 2): ", val_io92));
      val_io92 |= 2;	/* used for debug */
      outb (0x92, val_io92);
      timeout = 100;	/* said to be "fast A20" */
      }

  pokew_doorclosed (A20_TEST_LOW_ADDR, saved);

  ADBG ((" [get_A20(1000) 3: 0x%X] ", get_A20 (1000)));

  if (timeout == 0)
      ADBG ((" failed] "));
    else {
      if (cpt == 0) {
	  ADBG ((" used outb (0x92, 0x%X) ", val_io92));
			/* said to be "fast A20": */
	  ADBG ((" done, in %u loops] ", 100 - timeout));
	  }
	else
	  ADBG ((" done, in %u loops] ", copy_timeout - timeout));
      }
#if defined (DEBUG) && ((DEBUG & DEBUG_OUTPUT) == SCREEN) && (USER_SUPPORT != 0)
  ADBG ((" PRESS A KEY TO CONTINUE ", UI_getkey(0)));
#endif

  return !timeout; /* return (timeout == 0)? 1 : 0 */
  }

/*
 * Hem ... that's mine, it may work!
 *
 * The segment door opening/closing: I hate windows -:)
 * This stuff is not new, but this way to use it is GPL.
 */
GENERATE_EXTRA2CODE_STUB(segment_door, unsigned, unsigned open)
unsigned segment_door (unsigned open)
  {
  dataseg_type opendoor[] __attribute__ (( aligned(8) )) = {
    {}, /* segment 0 is the invalid one */
    {
    cste_10: 2,
    present: 1,
    writable: 1,
    accessed: 1,        /* 486+: do not update it if it is already set */
    limit: 0xFFFF,
    limit_msb: 0xF,
    granularity: 1	/* limit = 0xFFFFF * 4 Kb = 4 Gb */
    },
    {
    cste_10: 2,
    present: 1,
    writable: 1,
    accessed: 1,        /* 486+: do not update it if it is already set */
    limit: 0xFFFF,
    limit_msb: 0,
    granularity: 0
    }
    };
  gdt_idt_param_type gdtptr __attribute__ (( aligned(16) )) = {
    limit: sizeof (opendoor) - 1, /* -1 : see Intel doc */
    base: farptr2linear (stack_adr (opendoor))
    };
  register unsigned save_ds_es, save_fs_gs, save_ss, selector;

  if (open && enable_A20 (0x10000) != 0)
      return 1;

  disable_interrupt();
  set_gdt (&gdtptr);

  save_ds_es = getds() | (getes() << 16);
  save_fs_gs = getfs() | (getgs() << 16);
  save_ss = getss();

  selector = open? 0x0008 : 0x0010;

  if (protected_mode() != 0) {
      enable_interrupt();
      return 2;
      }

  setgs (selector);
  setfs (selector);
  setes (selector);
  setds (selector);
  setss (selector); /* -fomit-frame-pointer, "cmpl $0x7f2bb701,0x0(%ebp)" */

#if defined (__i386__)
  /* update accessed bit on 386: read %%seg:(0) */
  update_ss();
  update_gs();
  update_fs();
  update_es();
  update_ds();
#endif

  while (real_mode () != 0) ;

  setgs (save_fs_gs >> 16);
  setes (save_ds_es >> 16);
  setds (save_ds_es);
  setfs (save_fs_gs);
  setss (save_ss);

  enable_interrupt();

  return 0;
  }
