/* make.h */
#ifndef MAKE_H
#define MAKE_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/**
 ** This file contains only PREPROCESSOR directives, so can be included
 ** in files not written in C, like the linker configuration file.
 ** It is the only file included in ".h" files, and does basic checking
 ** of the configuration.
 **/

/**
 ** Now we define the relation in between Makefile string-variables
 ** and CPP number-variables - in short do not touch them.
 **/

/*
 * The setup configuration:
 */
#define ASSEMBLY_NONE     0x000000  /* use C equivalent */
#define ASSEMBLY_CISC     0x000001  /* complex addressing mode OK */
#define ASSEMBLY_RISC     0x000002  /* more instructions but simpler ones */
#define ASSEMBLY_DSES     0x000003  /* use only C pointers refering to %ds
					and %es, do not use %fs nor %gs */

#define CODE_SEGMENT      0x000010  /* use a different segment for code */
#define EXTRA_SEGMENT     0x000020  /* use an extra segment for code */

#define BIG_MALLOC        0x000100  /* have a bigger pool for malloc */
#define GC_SECTION        0x000200  /* use garbage collection of sections */
#define UPDATE_BOOTPARAM  0x001000  /* update gujin parameters on disk */
#define MINIDEBUG         0x002000  /* just few debug fcts on '1'..'9' */
#define BIOS_ONLY         0x004000  /* disable boot.exe/boot.com */
#define USE_INT1587       0x008000  /* to copy to high memory */

#define APPLICATION       0x000000  /* build application */
#define TESTER            0x800000  /* build tester */

#define QUITE             0x400000  /* Be really quite */

#define ASSEMBLY_TYPE        (SETUP & 0x0F)
#define BUILD_APPLICATION()  (!(SETUP & TESTER))
#define BUILD_TESTER()       (SETUP & TESTER)

#if (SETUP & EXTRA_SEGMENT) && !(SETUP & CODE_SEGMENT)
#error EXTRA_SEGMENT without CODE_SEGMENT
#endif

#if (SETUP & BIG_MALLOC) && (SETUP & USE_INT1587)
#error BIG_MALLOC and USE_INT1587 do not like each other in SETUP
#endif

#if !(SETUP & QUITE)
#define PRINTF(format...)	verbose_printf (format)
#define PRINT(msg)		verbose_print (msg)
#define PUTS(msg)		verbose_puts (msg)
#else
#define PRINTF(format...)	/* */
#define PRINT(msg)		/* */
#define PUTS(msg)		/* */
#endif

#if (SETUP & MINIDEBUG) && (SETUP & QUITE)
#error Undefine MINIDEBUG when defining QUITE on SETUP
#endif

#if (ASSEMBLY_TYPE == ASSEMBLY_DSES) && (SETUP & USE_INT1587)
#warning USE_INT1587 will not be used with ASSEMBLY_DSES
#endif

#if BUILD_TESTER()
#if SETUP & UPDATE_BOOTPARAM
#error The tester shall not be built with UPDATE_BOOTPARAM
#endif
#if SETUP & MINIDEBUG
#error The tester shall not be built with MINIDEBUG
#endif
#endif

/*
 * The user interface configuration:
 */
#define VGA_SUPPORT     (1<< 0)
#define VGA_EXTENDED    (1<< 1) // modes > 0x13 can be probed/used
#define VGA_MEMORY      (1<< 2) // to be implemented: direct memory access
#define VGA_USEINT1013  (1<< 3) // use _VGA_writestring, reduces code size
#define VGA_INT1013_BG  (1<< 4) // use background color for 256 color modes
#define VGA_FREQUENCY   (1<< 5) // try to set refresh frequencies of screen

#define VESA_WINDOW     (1<< 8)  // i.e. VESA 1 compatible
#define VESA_STRANGE_WINDOW (1<< 9)// window without 2^x bytes granularity, slow
#define VESA_WINFUNCTION (1<< 10) // optimise by using fct instead interrupt
#define VESA_HARDWINDOW (1<< 11) // analyse first code to change window(s)
#define VESA_2WINDOWS   (1<< 12) // use separate read & write window if present
#define VESA_LINEAR     (1<< 13) // i.e. VESA 2 only
#define VESA_EDID       (1<< 14) // interogate screen with EDID
#define VESA_PMINFO     (1<< 15) // Get the I/O ports used by the graphic card
#define VESA_RECALC	(1<< 16) // recalc nb of visible lines if EGA compat
#define VESA_4BPP_EGA   (1<< 17) // 4 Bits Per Pixel
#define VESA_4BPP_TEXT  (1<< 18) // text modes, 16 colors
#define VESA_8BPP       (1<< 19) // 8 Bits Per Pixel
#define VESA_16BPP      (1<< 20) // 15 or 16 Bits Per Pixel
#define VESA_24BPP      (1<< 21) // 24 Bits Per Pixel
#define VESA_32BPP      (1<< 22) // 32 Bits Per Pixel
#define VESA_SUPPORT    (VESA_4BPP_EGA|VESA_4BPP_TEXT|VESA_8BPP \
	    |VESA_16BPP|VESA_24BPP|VESA_32BPP)

#define SERIAL_VT100    (1<< 24)
#define SERIAL_VT420    (1<< 25)
#define SERIAL_SUPPORT  (SERIAL_VT100|SERIAL_VT420)

#define BIOS_MOUSE_SUPPORT   (1<< 28) // mostly PS2 mouses
#define SERIAL_MOUSE_SUPPORT (1<< 29) // serial 2/3 buttons mouse
#define JOYSTICK_SUPPORT     (1<< 30) // the standart BIOS one

#if (USER_SUPPORT & VGA_MEMORY) && (USER_SUPPORT & VGA_USEINT1013)
#error Cannot define VGA_MEMORY and VGA_USEINT1013 at the same time.
#endif

#if (USER_SUPPORT & VESA_HARDWINDOW) && !(USER_SUPPORT & VESA_WINFUNCTION)
#error Cannot define VESA_HARDWINDOW without VESA_WINFUNCTION
#endif

/*
 * The disk subsystem configuration:
 */
// The way to read sectors on permanent storage device:
#define BIOS_SUPPORT    0x0000001
#define EBIOS_SUPPORT   0x0000002
#define IDE_SUPPORT     0x0000004
#define DOS_SUPPORT	0x0000008
#define CDBIOS_SUPPORT	0x0000010	// not implemented now
#define ATAPI_SUPPORT	0x0000020	// not implemented now

// What is probed for the boot menu:
#define E2FS_PROBE	0x0000100
#define FAT12_PROBE	0x0000200
#define FAT16_PROBE	0x0000400
#define FAT32_PROBE	0x0000800	// not implemented now
#define ISOFS_PROBE	0x0001000	// not implemented now
#define BOOTSECT_PROBE  0x0002000

// Each of them reduce code size when defined:
#define NO_FLOPPY_SUPPORT		0x0010000
#define PROBE_ONLY_BIOSBOOTDEVICE	0x0020000
#define NO_PROBE_PARTITIONS		0x0040000

#if DISK_SUPPORT & PROBE_ONLY_BIOSBOOTDEVICE
#if DISK_SUPPORT & (IDE_SUPPORT | ATAPI_SUPPORT)
#warning IDE_SUPPORT and ATAPI_SUPPORT not usefull when PROBE_ONLY_BIOSBOOTDEVICE
#endif
#if !(DISK_SUPPORT & (BIOS_SUPPORT | EBIOS_SUPPORT))
#error PROBE_ONLY_BIOSBOOTDEVICE needs either BIOS_SUPPORT or EBIOS_SUPPORT
#endif
#endif

/*
 * The LANGUAGE configuration:
 */
#define ENGLISH		1
#define FRENCH		2
#define DUTCH		3
#define SPANISH		4
#define ITALIAN		5
#define PORTUGESE	6
#define SWEDISH         7
#define FINNISH         8
// to be corrected and completed!

/*
 * The debug subsystem configuration:
 */

/* debug options: */
#define DEBUG_STACK     0x000100  /* is a standalone - default output */
#define DEBUG_ADDR      0x000200  /* is a standalone - default output */
#define DEBUG_UTIL      0x000400
#define DEBUG_MALLOC    0x000800
#define DEBUG_BOOT      0x001000
#define DEBUG_DISK      0x002000
#define DEBUG_VIDEO     0x004000
#define DEBUG_SERIAL    0x008000
#define DEBUG_MOUSE     0x010000
#define DEBUG_LOADER	0x020000
#define DEBUG_CACHE	0x040000
#define DEBUG_IOCHANNEL	0x080000
#define DEBUG_A20	0x100000

#define DEBUG_OPTION    0xFFFC00

/* debug output: */
#define COM1            0x01
#define COM2            0x02
#define COM3            0x03
#define COM4            0x04
#define LPT1            0x80
#define LPT2            0x81
#define LPT3            0x82
#define LPT4            0x83
#define SCREEN          0xF0
#define DOSFILE         0xF8

#define DEBUG_OUTPUT    0x00FF

/**
 ** Now are defines to select the more logical configuration:
 ** They can be modified if you need something special.
 **/

/*
 * For boot.ld:
 */
#if BUILD_APPLICATION() && (DISK_SUPPORT & E2FS_PROBE)
#define INCLUDE_E2FS
#endif

/*
 * For user.c/mouse.c:
 */

#if USER_SUPPORT & SERIAL_SUPPORT
/*
 * This one is probably only for fun: implement the mouse
 * on a VT terminal...
 */
#define MOUSE_ON_VT
#endif /* SERIAL_SUPPORT */

/*
 * For disk.c:
 */

#if !(DISK_SUPPORT & PROBE_ONLY_BIOSBOOTDEVICE)
/*
 * We need that because the "ide=0x..." Linux command line
 * parameter needs to be in lowercase. see library.c.
 */
#define PRINTF_LOWERCASE_HEXADECIMAL

/*
 * For the TINY configuration, we can display only the BIOS
 * number to know which disk we are using (the boot disk),
 * but we need more information for the full system, to
 * display clearly IDE drives. (do we link in sprintf?)
 */
#define DISK_DISPLAY_FULLNAME

/*
 * Use malloc for storing informations of disks, partitions, freelist,
 * those structure are quite big and it is better to not define them
 * in a structure.
 */
#define DISK_USE_MALLOC

/*
 * Use malloc for storing where are the file fragments on disk
 * (LBA, nbsector)[] and not an array of FAT_MAX_FRAGMENT size.
 * Without any *_USE_MALLOC, the memory allocation function are not
 * needed: nice for the TINY configuration.
 */
#define UTIL_USE_MALLOC

#endif

/*
 * For boot.c:
 */

// Has to be tested before enabling:
//#define WRITE_ENABLE

#if BUILD_TESTER()
#define TREAT_COUNTER_TRACE_FLAG
#endif

#if BUILD_TESTER() || (USER_SUPPORT & VGA_SUPPORT)
#define TREAT_TIMER
#endif

#if BUILD_TESTER() || (USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT))
#define USE_SETJUMP_KEY
#endif

#if BUILD_TESTER() \
    || (DEBUG & (DEBUG_STACK | DEBUG_ADDR)) \
    || (USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT))
#define TREAT_EXCEPTION
#else
// We disable it for tiny configuration to save space, just re-enable
// if you have a problem (to treat invalid instructions, crashes...)
//#define TREAT_EXCEPTION
#endif

/*
 * For gzlib.c:
 */
#if BUILD_APPLICATION() && (DISK_SUPPORT & E2FS_PROBE) && !(SETUP & BIG_MALLOC)
/* Use the gzlib window trick when we cannot do anything else: */
#define GZLIB_BIG_MALLOC
#endif

/*
 * For vmlinuz.c:
 */
#if !(SETUP & BIOS_ONLY)
/*
 * By definning this, the "root=" kernel parameter is given
 * on the command line, which is IHMO cleanner, else it is
 * given on the "param" area around 0x90000 linear.
 * This command line "root=" parameter is supressed if it is
 * already present in the GZIP comment field.
 */
#define ROOT_ON_COMMAND_LINE
#endif

/*
 * Linux-2.2.* do not print the command line, 2.4.* does it.
 * Shall Gujin print it?
 */
#if USER_SUPPORT != 0
#define PRINT_COMMAND_LINE
#endif

/*
 * For malloc.c:
 */
#if SETUP & EXTRA_SEGMENT
/*
 * moves 1.5Kb of code in extraseg or in codeseg, if you need this space:
 */
#define MALLOC_IN_EXTRASEG
#endif

#endif /* MAKE_H */
