/* malloc.c */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"
#if (SETUP & EXTRA_SEGMENT) && defined (MALLOC_IN_EXTRASEG)
#define CALLING_FROM_EXTRASEG	/* before other includes */
#define FCT_PREFIX(fct)	SEGMENT_EXTRA(fct)
#else
#define FCT_PREFIX(fct)	/* */
#endif

#include "library.h"
#include "debug.h"
#include "malloc.h"

/*
 * Malloc matter: over-simplified !
 * I do not think it is SMP safe -:)
 *
 * The fact that it is "over-simplified" does not mean that
 * it is not covered by GPL: if you ever include [a part of]
 * this code, or even [a part of] its design into a private software,
 * the resulting code immediately goes to GPL and you have to
 * distribute the complete source code of your product.
 */

unsigned end_malloc;
static struct malloc_array_str {
    allocmem *adr;	/* strictly increasing */
    signed   size;	/* if < 0, not in use */
    } malloc_array[MALLOC_ARRAY_SIZE]; /* init at 0 : BSS */

#if (__GNUC__ <= 2) && (__GNUC_MINOR__ <= 97)
FCT_PREFIX(malloc)
#else
FCT_PREFIX(malloc) __attribute__ ((malloc))
#endif
allocmem *
malloc (unsigned size)
  {
  unsigned i;
  unsigned size_quantum = DIVROUNDUP (size, MALLOC_QUANTUM);

  MDBG (("MALLOC %u: ", size));

  if (size == 0) {
      DBG ((" [malloc returns 0 because 0 bytes requested] "));
      MDBG (("repeat size: zero!\r\n"));
      return 0;
      }

  for (i = 0; i < nbof (malloc_array); i++) {
      if (malloc_array[i].size >= 0) {
	  if (malloc_array[i].size == 0)
	      break;	 /* end of used array */
	  }
	else {
	  unsigned nbquantum = -malloc_array[i].size / MALLOC_QUANTUM;

	  if (   size_quantum <= nbquantum
	      && size_quantum + MALLOC_WASTE >= nbquantum) {
	      malloc_array[i].size = -malloc_array[i].size;
	      MDBG (("affecting to block %u of %u (= 0x%X) bytes"
			" at address 0x%X\r\n",
			i, malloc_array[i].size, malloc_array[i].size,
			malloc_array[i].adr));
	      return malloc_array[i].adr;
	      }
	  }
      }

  if (i == nbof (malloc_array)) {
      DBG ((" [malloc returns 0] "));
      MDBG (("no more space in array!\r\n"));
      return 0;
      }

#if SETUP & BIG_MALLOC
  if (end_malloc + size_quantum * MALLOC_QUANTUM > MALLOC_BASE + MALLOC_SIZE) {
      DBG ((" [malloc returns 0] "));
      MDBG (("no more space in the pool (requested: %u, free: %u) !\r\n",
		size_quantum * MALLOC_QUANTUM,
		MALLOC_BASE + MALLOC_SIZE - end_malloc));
      return 0;
      }
#else
  if (end_malloc + size_quantum * MALLOC_QUANTUM > getesp() - stack_safety) {
      DBG ((" [malloc returns 0] "));
      MDBG (("no more space before stack (requested: %u, free: %u) !\r\n",
		size_quantum * MALLOC_QUANTUM,
		getesp() - stack_safety - end_malloc));
      return 0;
      }
#endif

  malloc_array[i].size = size_quantum * MALLOC_QUANTUM;
  malloc_array[i].adr = (unsigned *)end_malloc;
  end_malloc += malloc_array[i].size;

#if SETUP & BIG_MALLOC
#if DEBUG & DEBUG_ADDR
  data_limit.high_limit = end_malloc;
#endif
#else  /* BIG_MALLOC */
#if DEBUG & DEBUG_STACK
  if (end_malloc + 2 + stack_safety > 0x8000)
      stack_limit.low_limit = end_malloc + 2 + stack_safety;
#endif
#endif /* BIG_MALLOC */
  MDBG (("allocated new block No %u of %d (= 0x%X) bytes at 0x%X\r\n",
	i, malloc_array[i].size, malloc_array[i].size, malloc_array[i].adr));
  return malloc_array[i].adr;
  }

FCT_PREFIX(free) void
free (allocmem *ptr)
  {
  unsigned i, j;

  MDBG (("FREE addr 0x%X: ", ptr));

  for (i = 0; i < nbof (malloc_array); i++)
      if (   malloc_array[i].adr == ptr
	  && malloc_array[i].size > 0)
	  break;
  if (i == nbof (malloc_array)) {
      MDBG (("invalid address!\r\n"));
      return;
      }

  j = i++;
  while (i < nbof (malloc_array)) {
      if (malloc_array[i].size > 0)
	  break;
      i++;
      }

  malloc_array[j].size = - malloc_array[j].size;

  if (i != nbof (malloc_array)) {
      MDBG (("block %u size %u is not the last block.\r\n",
		j, - malloc_array[j].size));
      return;
      }

#if SETUP & BIG_MALLOC
  MDBG (("last block %u size %u freed to pool",
	j, - malloc_array[j].size));
#else
  MDBG (("last block %u size %u freed to stack",
		j, - malloc_array[j].size));
#endif

  while (j != 0 && malloc_array[j-1].size < 0)
      j --;

  end_malloc = (unsigned) malloc_array[j].adr;
  MDBG ((" [new end_malloc 0x%X] ", end_malloc));

  while (j < nbof (malloc_array)) {
      if (malloc_array[j].size < 0) {
	  MDBG ((", block %u size %u",
			j, -malloc_array[j].size));
#if !(SETUP & BIG_MALLOC) /* clear the pool: unused stack */
	  memset (malloc_array[j].adr, 0, -malloc_array[j].size);
#endif
	  malloc_array[j].adr = 0;
	  malloc_array[j].size = 0;
	  }
      j++;
      }

  MDBG (("\r\n"));

#if SETUP & BIG_MALLOC
#if DEBUG & DEBUG_ADDR
  data_limit.high_limit = end_malloc;
#endif
#else /* BIG_MALLOC */
#if DEBUG & DEBUG_STACK
  if (end_malloc + 2 + stack_safety > 0x8000)
      stack_limit.low_limit = end_malloc + 2 + stack_safety;
#endif
#endif /* BIG_MALLOC */
  }

FCT_PREFIX(realloc) allocmem *
realloc (allocmem *ptr, unsigned size)
  {
  unsigned i, j;
  unsigned size_quantum;

  MDBG (("REALLOC 0x%X size: %d: ", ptr, size));

  if (size == 0) {
      DBG ((" [realloc returns 0] "));
      MDBG (("repeat size: zero!\r\n"));
      return 0;
      }

  if (ptr == 0) {
      MDBG ((" [realloc called with address 0] "));
      return malloc (size);
      }

  for (i = 0; i < nbof (malloc_array); i++)
      if (   malloc_array[i].adr == ptr
	  && malloc_array[i].size > 0)
	  break;

  if (i == nbof (malloc_array)) {
      DBG ((" [realloc returns 0] "));
      MDBG (("invalid address!\r\n"));
      return 0;
      }

  if ((signed) size <= malloc_array[i].size) {
      MDBG (("size preallocated sufficient.\r\n"));
      return ptr;
      }

  j = i++;

  while (i < nbof (malloc_array)) {
      if (malloc_array[i].size > 0)
	  break;
      i++;
      }

  if (i != nbof (malloc_array)) {
      unsigned *tmp;
      MDBG (("is not the last block.\r\n"));

      tmp = malloc (size);
      if (tmp == 0) {
	  MDBG (("REALLOC returns zero\r\n"));
	  return 0;
	  }
      memcpy (tmp, ptr, MIN ((unsigned)malloc_array[j].size, size));
      free (malloc_array[j].adr);
      return tmp;
      }

  MDBG (("is the last block.\r\n"));
  size_quantum = ROUNDUP (size, MALLOC_QUANTUM);

#if SETUP & BIG_MALLOC
  if (end_malloc + size_quantum - malloc_array[j].size
	> MALLOC_BASE + MALLOC_SIZE) {
      DBG ((" [realloc returns 0] "));
      MDBG (("last block, no more space in the pool!\r\n"));
      return 0;
      }
#else
  if (end_malloc + size_quantum - malloc_array[j].size
	> getesp() - stack_safety) {
      DBG ((" [realloc returns 0] "));
      MDBG (("last block, no more space before stack!\r\n"));
      return 0;
      }
#endif

  end_malloc += size_quantum - malloc_array[j].size;
  malloc_array[j].size = size_quantum;

#if SETUP & BIG_MALLOC
#if DEBUG & DEBUG_ADDR
  data_limit.high_limit = end_malloc;
#endif
#else  /* BIG_MALLOC */
#if DEBUG & DEBUG_STACK
  if (end_malloc + 2 + stack_safety > 0x8000)
      stack_limit.low_limit = end_malloc + 2 + stack_safety;
#endif
#endif /* BIG_MALLOC */

  return ptr;
  }

#if SETUP & MINIDEBUG
FCT_PREFIX(report_malloc) void
report_malloc (void)
  {
  unsigned short cpt, total = 0;

  for (cpt = 0; cpt < nbof(malloc_array) && malloc_array[cpt].size; cpt++) {
      printf ("block %u address 0x%X size %d\r\n",
		cpt, malloc_array[cpt].adr, malloc_array[cpt].size);
      if (malloc_array[cpt].size > 0)
	  total += malloc_array[cpt].size;
      }
  printf ("Total allocated: %u, free memory: %u, total stack used: %u Kb\r\n",
		total, freemem(), usedstack() / 1024);
  }
#endif /* MINIDEBUG */

/**
 ** Extra segment stuff:
 **/
#ifdef MALLOC_IN_EXTRASEG
#define GENERATE_STUB		GENERATE_CODE2EXTRA_STUB
#define GENERATE_VOID_STUB	GENERATE_VOID_CODE2EXTRA_STUB
#else
#define GENERATE_STUB		GENERATE_EXTRA2CODE_STUB
#define GENERATE_VOID_STUB	GENERATE_VOID_EXTRA2CODE_STUB
#endif /* MALLOC_IN_EXTRASEG */

GENERATE_STUB(malloc, allocmem *, unsigned size)
GENERATE_VOID_STUB(free, allocmem *ptr)
GENERATE_STUB(realloc, allocmem *, allocmem *ptr, unsigned size)
#if SETUP & MINIDEBUG
GENERATE_VOID_STUB(report_malloc, void)
#endif /* MINIDEBUG */
