/* mouse.c */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"
#include "library.h"
#include "libextra.h"	/* joystick_present() */
#include "bios.h"
#include "boot.h"	/* BOOT1_COM_port() */
#include "debug.h"
#include "malloc.h"
#include "user.h"
#include "mouse.h"

static const union mouse_cursor cursor_mask[SIZEOF_CURSOR_MASK] = {
#define _ 0
#define X 1
	{{X,_,_,_,_,_,_,_}},
	{{X,X,_,_,_,_,_,_}},
	{{X,X,X,_,_,_,_,_}},
	{{X,X,X,X,_,_,_,_}},
	{{X,X,X,X,X,_,_,_}},
	{{X,X,X,X,X,X,_,_}},
	{{X,X,X,X,X,X,X,_}},
	{{X,X,X,X,X,X,X,X}},
	{{X,X,X,X,X,_,_,_}},
	{{X,X,_,X,X,_,_,_}},
	{{X,_,_,_,X,X,_,_}},
	{{_,_,_,_,X,X,_,_}},
	{{_,_,_,_,_,X,X,_}},
	{{_,_,_,_,_,X,X,_}}
#undef X
#undef _
	};

union mouse_field_union {
    unsigned		 color;
    struct attribute_str attr;
    };

/**
 ** The part managing the "clickable fields":
 **/

/* MOUSE_print_field helpers: */
static void
drawbutton (coord    upperleft, coord  lowerright,
	    unsigned ulcolor, unsigned lrcolor,
	    unsigned fillcolor, unsigned drawbackground)
  {bound_stack();{
  coord xy;
  unsigned short delta;

  /* Enlarge the button by few pixels: */
  upperleft.x -= 2;
  lowerright.x += 2;

  xy = upperleft;
  delta = lowerright.x - upperleft.x - 2;
  /* top line */
  UI_plotHline (xy, lowerright.x, ulcolor);

  /* do not change VESA1 banks too often */
  while (++(xy.y) < lowerright.y - 1) {
      UI_setpixel (xy, ulcolor);
      xy.x++;
      if (drawbackground)
	  UI_plotHline (xy, lowerright.x - 1, fillcolor);
      xy.x += delta;
      UI_setpixel (xy, lrcolor);
      xy.x -= delta + 1;
      }

  /* bottom line */
  UI_plotHline (xy, lowerright.x, lrcolor);
  }}

static inline union mouse_field_union
prepare_display_mouse_field (struct mousefield_str *field)
  {
  union mouse_field_union returned;

  if (UI.attributes.valid.transparent) {
      unsigned bgcolor, ulcolor, lrcolor;

      returned.attr = UI.attributes.current;
      UI_setattribute (attr_transparent);

      if (field->attr.checkbox) {
	  if (field->attr.greyed)
	      bgcolor = CHECKBOX_INACTVE_BG;
	    else
	      bgcolor = CHECKBOX_BG;
	  }
	else if (UI.parameter.nbcolor <= 16)
	  bgcolor = UI.bgcolor;
	else if (field->attr.mainfield)
	  bgcolor = BOTTOM_MENU_BUTTON_BG;
	else
	  bgcolor = TOP_MENU_BUTTON_BG;
      if (field->attr.pressed == 0) {
	  ulcolor = UI.stdcolor[white];
	  lrcolor = UI.stdcolor[black];
//	  ulcolor = UI.stdcolor[lightgray];
//	  lrcolor = UI.stdcolor[darkgray];
	  }
	else {
	  ulcolor = UI.stdcolor[black];
	  lrcolor = UI.stdcolor[white];
//	  ulcolor = UI.stdcolor[darkgray];
//	  lrcolor = UI.stdcolor[lightgray];
	  }
      drawbutton (field->upperleft, field->lowerright,
		  ulcolor, lrcolor, bgcolor, 1);
      }
    else if (UI.attributes.valid.underline) {
      returned.attr = UI.attributes.current;
      UI_setattribute (attr_underline);
      }
    else if (UI.attributes.valid.brighter) {
      returned.attr = UI.attributes.current;
      UI_setattribute (attr_brighter);
      }
    else {
      returned.color = UI.fgcolor;
      UI_setfgcolor (MOUSE_SELECTABLE_CHARS);
      }
  return returned;
  }

static inline void
end_display_mouse_field (union mouse_field_union saved)
  {
  if (   UI.attributes.valid.transparent
      || UI.attributes.valid.underline
      || UI.attributes.valid.brighter)
      UI_setattribute (saved.attr);
    else
      UI_setfgcolor (saved.color);
  }

static inline unsigned
mouse_treat_string (char *dest, const char *src,
		    unsigned maxlen, unsigned active)
  {
  char *initdest = dest;

  for (;;) {
      if (!active && *src == '<')
	  *dest = '(';
	else if (!active && *src == '>')
	  *dest = ')';
	else if (--maxlen == 0) {
	  dest[-1] = '.';
	  dest[-2] = '.';
	  dest[-3] = '.';
	  *dest++ = '\0';
	  break;
	  }
	else if (*src == '\0') {
	  if (!active)
	      *dest++ = ' ';
	  *dest++ = '\0';
	  break;
	  }
	else
	  *dest = *src;
      src++;
      dest++;
      }
  return dest - 1 - initdest;
  }

/* Shall this function also manage the screen scrolling ? */
void MOUSE_print_field (const char *str,
			struct mousefieldattr_str attr,
			unsigned short key,
			unsigned short row,
			unsigned short col)
  {bound_stack();{
  struct mousefield_str *field = 0;
  union mouse_field_union saved = {}; /* warning */
  unsigned endpos;
  char displayed[UI.parameter.nbcol];

  /* We should not have \r, \n, \t, \b ... in str! */
  if (str == 0 || *str == '\0') {
      PDBG ((" [%s: add an empty field!] ", __FUNCTION__));
      return;
      }

  if (Mouse.nb_field != 0 && Mouse.fields == 0)
      PDBG ((" [%s inconsistent Mouse structure] ", __FUNCTION__));

  endpos = col + mouse_treat_string (displayed, str,
			UI.parameter.nbcol - col, attr.active);

  /* get memory for new field, realloc 1st param can be 0: */
  if (   attr.active
      && (field = realloc (Mouse.fields, (Mouse.nb_field + 1)
				* sizeof (Mouse.fields[0]))) != 0) {
      Mouse.fields = field;
      field = &Mouse.fields[Mouse.nb_field];
      Mouse.nb_field += 1;

      field->attr = attr;
      field->keypress = key;
      field->str = str;
      field->upperleft.x = col * UI.parameter.charwidth;
      field->upperleft.y = row * UI.parameter.charheigh;
      field->lowerright.x = endpos * UI.parameter.charwidth;
      field->lowerright.y = (row+1) * UI.parameter.charheigh;

      PDBG ((" [adding mouse field {%u,%u},{%u,%u} ] ",
		field->upperleft.x, field->upperleft.y,
		field->lowerright.x, field->lowerright.y));

      saved = prepare_display_mouse_field (field);
      }
  UI_putstr (displayed);
  if (field != 0) {
      end_display_mouse_field (saved);
      UI_setcursor (row, endpos + 1);
      }
  }}

/* Removes part/all the clickable fields: */
void MOUSE_reset_field (struct mousefieldattr_str attr)
  {bound_stack();{
  if (Mouse.nb_field != 0 && Mouse.fields == 0)
      PDBG ((" [%s inconsistent Mouse structure] ", __FUNCTION__));

  if (Mouse.fields != 0) {
      struct mousefield_str array[Mouse.nb_field];
      unsigned nbstaying, cpt;

      for (nbstaying = cpt = 0; cpt < Mouse.nb_field; cpt ++) {
	  if (Mouse.fields[cpt].attr.upperfield && attr.upperfield)
	      continue;
	  if (Mouse.fields[cpt].attr.mainfield && attr.mainfield)
	      continue;
	  array[nbstaying++] = Mouse.fields[cpt];
	  }
      free (Mouse.fields);
      Mouse.fields = 0; /* needed because of the realloc(0, x) */
      Mouse.nb_field = nbstaying;
      if (nbstaying != 0) {
	  Mouse.fields = malloc (nbstaying * sizeof (Mouse.fields[0]));
	  if (Mouse.fields == 0) {
	      PDBG ((" [%s failed malloc] ", __FUNCTION__));
	      Mouse.nb_field = 0;
	      return;
	      }
	  for (cpt = 0; cpt < nbstaying; cpt ++)
	      Mouse.fields[cpt] = array[cpt];
	  }
      }
  }}

/**
 ** The display mouse cursor function:
 **/
static inline coord calculate_new_pos(int deltax, int deltay)
  {
  coord newpos;

  deltax += Mouse.position.x;
  if (deltax < 0)
      newpos.x = 0;
    else if (deltax > Mouse.maximum.x)
      newpos.x = Mouse.maximum.x;
    else
      newpos.x = deltax;

  deltay = Mouse.position.y - deltay;
  if (deltay < 0)
      newpos.y = 0;
    else if (deltay > Mouse.maximum.y)
      newpos.y = Mouse.maximum.y;
    else
      newpos.y = deltay;

  return newpos;
  }

static inline unsigned
get_color_and_field (coord pos, unsigned *color,
		     struct mouse_button_str newbutton)
  {
  struct mousefield_str *fieldptr;
  unsigned field;

  for (field = 0, fieldptr = Mouse.fields;
       field < Mouse.nb_field;
       field++, fieldptr++) {
      if (   pos.x >= fieldptr->upperleft.x
	  && pos.x <  fieldptr->lowerright.x
	  && pos.y >= fieldptr->upperleft.y
	  && pos.y <  fieldptr->lowerright.y)
	  break;
      }

  if (field < Mouse.nb_field) {
      *color = MOUSE_INFIELD_COLOR;
      if (!newbutton.left && !newbutton.right && !newbutton.middle)
	  field = 0xFFFFFFFF; /* do not treat it now */
      }
    else {
      if (newbutton.middle)
	  *color = MOUSE_OUTFIELD_COLOR_MIDDLEBUTTON;
	else if (newbutton.top)
	  *color = MOUSE_OUTFIELD_COLOR_TOPBUTTON;
	else if (newbutton.left && newbutton.right)
	  *color = MOUSE_OUTFIELD_COLOR_TWOBUTTON;
	else if (newbutton.left)
	  *color = MOUSE_OUTFIELD_COLOR_LEFTBUTTON;
	else if (newbutton.right)
	  *color = MOUSE_OUTFIELD_COLOR_RIGHTBUTTON;
	else
	  *color = MOUSE_DEFAULT_COLOR;
      }

  return field;
  }

static inline unsigned
button_equal (struct mouse_button_str a, struct mouse_button_str b)
  {
  union {
      struct mouse_button_str fields;
      unsigned char all;
      } ba = { a }, bb = { b };
  return ba.all == bb.all;
  }

static unsigned
MOUSE_move (int deltax, int deltay, struct mouse_button_str newbutton)
  {bound_stack();{
  coord newpos = calculate_new_pos (deltax, deltay);
  unsigned color;
  unsigned field = get_color_and_field (newpos, &color, newbutton);

  if (UI.parameter.attr.isgraphic) {
      if (   newpos.x != Mouse.position.x
	  || newpos.y != Mouse.position.y
	  || !button_equal (Mouse.button, newbutton)) {
	  unsigned *arrayptr;
	  union mouse_cursor mask;
	  const union mouse_cursor *maskptr;
	  const union mouse_cursor *const
			endmask = &cursor_mask[nbof (cursor_mask)];
	  coord xy;

	  if (!Mouse.button.hidden) {
	      arrayptr = Mouse.savearray;
	      for  (maskptr = cursor_mask, xy.y = Mouse.position.y;
		    maskptr < endmask;
		    maskptr++, xy.y++) {
		  for (mask.group = 1, xy.x = Mouse.position.x;
		       mask.group != 0;
		       mask.group <<= 1, xy.x++) {
		      if (maskptr->group & mask.group)
			  UI_setpixel (xy, *arrayptr++);
		      }
		  }
	      }

	  /* Before re-printing the mouse: */
	  if (   field < Mouse.nb_field
	      && UI.attributes.valid.transparent) {
	      unsigned ulcolor, lrcolor;

	      if (Mouse.fields[field].attr.pressed) {
		  ulcolor = UI.stdcolor[white];
		  lrcolor = UI.stdcolor[black];
//		  ulcolor = UI.stdcolor[lightgray];
//		  lrcolor = UI.stdcolor[darkgray];
		  }
		else {
		  ulcolor = UI.stdcolor[black];
		  lrcolor = UI.stdcolor[white];
//		  ulcolor = UI.stdcolor[darkgray];
//		  lrcolor = UI.stdcolor[lightgray];
		  }
	      drawbutton (Mouse.fields[field].upperleft,
			  Mouse.fields[field].lowerright,
			  ulcolor, lrcolor, 0, 0);
	      }

	  if (!newbutton.hidden) {
	      arrayptr = Mouse.savearray;
	      for (maskptr = cursor_mask, xy.y = newpos.y;
		   maskptr < endmask;
		   maskptr++, xy.y++) {
		  for (mask.group = 1, xy.x = newpos.x;
		       mask.group != 0;
		       mask.group <<= 1, xy.x++) {
		      if (maskptr->group & mask.group) {
			  *arrayptr++ = UI_getpixel (xy);
			  UI_setpixel (xy, color);
			  }
		      }
		  }
	      }
	  }
      }
    else {
      coord oldM = {
	  x: Mouse.position.x / UI.parameter.charwidth,
	  y: Mouse.position.y / UI.parameter.charheigh
	  }, newM = {
	  x: newpos.x / UI.parameter.charwidth,
	  y: newpos.y / UI.parameter.charheigh
	  };

      if (   oldM.x != newM.x
	  || oldM.y != newM.y
	  || !button_equal (Mouse.button, newbutton)) {
#ifdef MOUSE_ON_VT
	  if (BOOT1_COM_port() >= 0) {
	      /* nothing to clear the mouse */
	      if (!newbutton.hidden)
		  SERIAL_setcursor (newM.y, newM.x);
		else
		  SERIAL_setcursor (Mouse.curs_row, Mouse.curs_col);
	      }
	    else
#endif
	      {
	      unsigned newpixel;

	      if (!Mouse.button.hidden)
		  UI_setpixel (oldM, *Mouse.savearray);
	      *Mouse.savearray = UI_getpixel (newM);
#if USER_SUPPORT & VGA_SUPPORT
	      if (UI.parameter.nbcolor == 2)
		  newpixel = '\333'; /* background ignored */
		else if (UI.parameter.nbcolor == 0) {
		  if (color == MOUSE_DEFAULT_COLOR)
		      newpixel = 0x0700 | (unsigned char) '\333';
		    else
		      newpixel = 0x0F00 | (unsigned char) '\333';
		  }
		else
#endif
		  newpixel = (color << 12) | (*Mouse.savearray & 0xFFF);
	      if (!newbutton.hidden)
		  UI_setpixel (newM, newpixel);
	      }
	  }
      }

  Mouse.button = newbutton;
  Mouse.position = newpos;
  return field;
  }}

/**
 ** The pooling functions:
 **/
static inline unsigned
MOUSE_PS2_poller (signed *deltax, signed *deltay,
		  struct mouse_button_str *newbutton)
  {
#if USER_SUPPORT & BIOS_MOUSE_SUPPORT
  PDBG (("PS2: "));

  if (!Mouse.data.PS2.reservedbit) {
      PDBG (("invalid format or no report\r\n"));
      return 2;
      }

  if (Mouse.data.PS2.x_overflow || Mouse.data.PS2.y_overflow) {
      PDBG (("overflow\r\n"));
      return 3;
      }

  *newbutton = (struct mouse_button_str) {
	left: Mouse.data.PS2.left_button,
	right: Mouse.data.PS2.right_button,
	middle: Mouse.data.PS2.middle_button
	};
  *deltax = Mouse.data.PS2.x_data;
  *deltay = Mouse.data.PS2.y_data;

  Mouse.data.PS2 = (struct PS2_struct) {};

  return 0;
#else
  return 1;
#endif
  }

#if USER_SUPPORT & SERIAL_MOUSE_SUPPORT
static inline void
MOUSE_STD_report (signed *deltax, signed *deltay,
		  struct mouse_button_str *newbutton,
		  unsigned char middleval)
  {
  union {
      struct {
	  unsigned char lsb : 6;
	  unsigned char msb : 2;
	  } fields;
      signed char all;
      } tmp;

  tmp.fields.msb = Mouse.data.serial.received.stddata.xmsb;
  tmp.fields.lsb = Mouse.data.serial.received.stddata.x;
  *deltax = tmp.all;
  tmp.fields.msb = Mouse.data.serial.received.stddata.ymsb;
  tmp.fields.lsb = Mouse.data.serial.received.stddata.y;
  *deltay = - tmp.all;
  *newbutton = (struct mouse_button_str) {
	left: Mouse.data.serial.received.stddata.left_button,
	right: Mouse.data.serial.received.stddata.right_button,
	middle: middleval
	};
  }

static inline unsigned
MOUSE_STD_poller (signed *deltax, signed *deltay,
		  struct mouse_button_str *newbutton,
		  char received)
  {
  unsigned char middleval = 0; /* warning */
  PDBG (("%s receive char 0x%X: ", __FUNCTION__, received));

  /* Then we have a resync bit: */
  if (received & 0x40) {
      PDBG (("resync\r\n"));
      if (Mouse.data.serial.cpt == 3 && Mouse.button.middle) {
	  /* we've lost the release event, generate exactly
		the previous report with middle button up: */
	  PDBG (("generating lost frame: "));
	  MOUSE_STD_report (deltax, deltay, newbutton, 0);
	  Mouse.data.serial.received.bytes[0] = received;
	  Mouse.data.serial.cpt = 1;
	  return 0;
	  }
      Mouse.data.serial.cpt = 0;
      }
    else if (Mouse.data.serial.cpt >= 3) {
      if (Mouse.data.serial.cpt > 3 || (received != 0x20 && received != 0)) {
	  Mouse.data.serial.cpt++;
	  PDBG (("rejected\r\n"));
	  return 1;
	  }
      if (received == 0) {
	  PDBG (("Logitech extension middle button up\r\n"));
	  middleval = 0;
	  }
	else {
	  PDBG (("Logitech extension middle button down\r\n"));
	  if ((Mouse.type & MOUSE_SERIAL_MASK) == MOUSE_SERIAL_COM1) {
	      PDBG (("So switching to serial 3 buttons.\r\n"));
	      Mouse.type &= ~MOUSE_SERIAL_MASK;
	      Mouse.type |= MOUSE_SERIAL3_COM1;
	      }
	  middleval = 1;
	  }
      }
  Mouse.data.serial.received.bytes[Mouse.data.serial.cpt++] = received;

  if (Mouse.data.serial.cpt == 3 + (middleval | Mouse.button.middle)) {
      PDBG (("treating frame: "));
      MOUSE_STD_report (deltax, deltay, newbutton, middleval);
      return 0;
      }
    else {
      PDBG (("waiting end of frame\r\n"));
      return 1;
      }
  }

static inline unsigned
MOUSE_MM_poller (signed *deltax, signed *deltay,
		 struct mouse_button_str *newbutton,
		 char received)
  {
  PDBG (("%s receive char 0x%X: ", __FUNCTION__, received));
  /* Then we have to wait a pattern for first byte: */
  if (Mouse.data.serial.cpt == 0 && (received & 0xF8) != 0x80) {
      PDBG (("rejected \r\n"));
      return 1;
      }
  Mouse.data.serial.received.bytes[Mouse.data.serial.cpt++] = received;
  if (Mouse.data.serial.cpt >= 5) {
      PDBG (("treating frame: "));
      Mouse.data.serial.cpt = 0;
      *newbutton = (struct mouse_button_str) {
	left: !Mouse.data.serial.received.mmdata.left_button,
	right: !Mouse.data.serial.received.mmdata.right_button,
	middle: !Mouse.data.serial.received.mmdata.middle_button
	};
      /* small acceleration: * 2 */
      *deltax = (int)Mouse.data.serial.received.mmdata.x
		+ 2 * Mouse.data.serial.received.mmdata.deltax;
      *deltay = (int)Mouse.data.serial.received.mmdata.y
		+ 2 * Mouse.data.serial.received.mmdata.deltay;
      }
    else {
      PDBG (("waiting end of frame\r\n"));
      return 1;
      }
  return 0;
  }
#endif /* SUPPORT & SERIAL_MOUSE_SUPPORT */

static inline unsigned
MOUSE_SERIAL_poller (signed *deltax, signed *deltay,
		     struct mouse_button_str *newbutton)
  {
#if USER_SUPPORT & SERIAL_MOUSE_SUPPORT
    if (SERIAL_DATA_RECEIVED (Mouse.data.serial)) {
      unsigned char received = SERIAL_DATA_GET (Mouse.data.serial);

      if ((Mouse.type & MOUSE_SERIAL_MASK) == MOUSE_SERIALM_COM1) {
	  if (MOUSE_MM_poller (deltax, deltay, newbutton, received))
	      return 1;
	  }
	else {
	  if (MOUSE_STD_poller (deltax, deltay, newbutton, received))
	      return 1;
	  }
      return 0;
      }
#endif
  return 1;
  }

static inline unsigned
MOUSE_JOYSTICK_poller (signed *deltax, signed *deltay,
			struct mouse_button_str *newbutton)
  {
#if USER_SUPPORT & JOYSTICK_SUPPORT
  unsigned nbtick;
  unsigned short Ax, Ay, Bx, By;
  struct JoystickButton_str button;

  if (   !get_gujin_param_attrib().enable_joystick
      || !joystick_present())
      return 1;

  nbtick = _BIOS_nbtick();
  if (Mouse.joystick.last_sample_tick == nbtick)
      return 1; /* only one valid sample per tick */
  if (JoystickPosition (&Ax, &Ay, &Bx, &By) != 0) {
      Mouse.joystick.last_sample_tick = nbtick;
      return 1; /* Joystick no supported by BIOS or abscent */
      }
  if (JoystickButton (&button) != 0) {
      Mouse.joystick.last_sample_tick = nbtick;
      return 1; /* Joystick no supported by BIOS */
      }
  /* We have a real data, calibrate: */
  if (Mouse.joystick.Xmin == 0 || Ax < Mouse.joystick.Xmin)
      Mouse.joystick.Xmin = Ax;
  if (Mouse.joystick.Xmax == 0 || Ax > Mouse.joystick.Xmax)
      Mouse.joystick.Xmax = Ax;
  if (Mouse.joystick.Ymin == 0 || Ay < Mouse.joystick.Ymin)
      Mouse.joystick.Ymin = Ay;
  if (Mouse.joystick.Ymax == 0 || Ay > Mouse.joystick.Ymax)
      Mouse.joystick.Ymax = Ay;

  if (   Mouse.joystick.Xmax - Mouse.joystick.Xmid < 8
      || Mouse.joystick.Xmid - Mouse.joystick.Xmin < 8
      || Mouse.joystick.Ymax - Mouse.joystick.Ymid < 8
      || Mouse.joystick.Ymid - Mouse.joystick.Ymin < 8)
      return 1; /* not enough calibrated */

  if (Mouse.joystick.accel < 3 * 18) /* 3 seconds */
      Mouse.joystick.accel++; /* even for next return, no changes */

  /* Divide per 4 to remove noise: */
#if 0
  *deltax = (Ax - (Mouse.joystick.Xmin + Mouse.joystick.Xmax) / 2 + 2) / 4;
  *deltay = (Ay - (Mouse.joystick.Ymin + Mouse.joystick.Ymax) / 2 + 2) / 4;
#else
  *deltax = (Ax - Mouse.joystick.Xmid + 2) / 4;
  *deltay = (Ay - Mouse.joystick.Ymid + 2) / 4;
#endif
  *deltay = -*deltay;
  if (   *deltax == 0
      && *deltay == 0
      && Mouse.joystick.lastbutton.button_A == button.button_A
      && Mouse.joystick.lastbutton.button_B == button.button_B
      && Mouse.joystick.lastbutton.button_C == button.button_C
      && Mouse.joystick.lastbutton.button_D == button.button_D)
      return 1;

  if (Mouse.joystick.accel > 4) {
      *deltax /= Mouse.joystick.accel / 4;
      *deltay /= Mouse.joystick.accel / 4;
      Mouse.joystick.accel -= 4;
      }
  Mouse.joystick.last_sample_tick = nbtick;
  Mouse.joystick.lastbutton = button;
  *newbutton = (struct mouse_button_str) {
	left:   !button.button_C,
	right:  !button.button_B,
	middle: !button.button_A,
	top:    !button.button_D
	};
  return 0;
#endif
  return 1;
  }

static unsigned MOUSE_poll (void)
  {bound_stack();{
  signed deltax, deltay;
  struct mouse_button_str newbutton;
  unsigned field;

  if (Mouse.type == MOUSE_NONE)
      return 0xFFFFFFFF;

  if (Mouse.maximum.x == 0)
      return 0xFFFFFFFF;

  PDBG (("%s: ", __FUNCTION__));
  if (Mouse.nb_field != 0 && Mouse.fields == 0)
      PDBG ((" [%s inconsistent Mouse structure] ", __FUNCTION__));

  if (MOUSE_JOYSTICK_poller (&deltax, &deltay, &newbutton) != 0) {
      if (Mouse.type == MOUSE_PS2) {
	  if (MOUSE_PS2_poller (&deltax, &deltay, &newbutton))
	      return 0xFFFFFFFF;
	  }
	else if (MOUSE_SERIAL_poller (&deltax, &deltay, &newbutton))
	  return 0xFFFFFFFF;
      }

  PDBG (("deltax = %d, deltay = %d\r\n", deltax, deltay));
  if ((field = MOUSE_move (deltax, deltay, newbutton)) < Mouse.nb_field) {
      return Mouse.fields[field].keypress;
      }
    else
      return 0xFFFFFFFF;
  }}

/**
 ** The start/stop:
 **/

static void MOUSE_start (void)
  {bound_stack();{
  unsigned char status;

  PDBG ((" [%s ", __FUNCTION__));
  if (UI.parameter.charwidth == 0 || UI.parameter.charheigh == 0) {
      PDBG (("Cannot start mouse, charwidth = %u, charheigh = %u\r\n",
		UI.parameter.charwidth, UI.parameter.charheigh));
      return;
      }

#ifdef MOUSE_ON_VT
  if (BOOT1_COM_port() >= 0) {
      if (SERIAL_getcursor (&Mouse.curs_row, &Mouse.curs_col) != 0) {
	  PDBG ((" disabling mouse on VT due to error in getcursor] "));
	  return;
	  }
      }
#endif

  if (UI.parameter.attr.isgraphic) {
      Mouse.maximum.x = UI.parameter.width - (8 * sizeof(cursor_mask[0].group));
      Mouse.maximum.y = UI.parameter.heigh - nbof (cursor_mask);
      }
    else {
      Mouse.maximum.x = UI.parameter.width * UI.parameter.charwidth - 1;
      Mouse.maximum.y = UI.parameter.heigh * UI.parameter.charheigh - 1;
      }

  PDBG (("max.x = %u, max.y = %u", Mouse.maximum.x, Mouse.maximum.y));
  if (Mouse.position.x || Mouse.position.y) {
      Mouse.position.x *= UI.parameter.charwidth;
      Mouse.position.x += UI.parameter.charwidth/2;
      Mouse.position.y *= UI.parameter.charheigh;
      Mouse.position.y += UI.parameter.charheigh/2;
      if (Mouse.position.x > Mouse.maximum.x - UI.parameter.charwidth/2)
	  Mouse.position.x = Mouse.maximum.x - UI.parameter.charwidth/2;
      if (Mouse.position.y > Mouse.maximum.y - UI.parameter.charheigh/2)
	  Mouse.position.y = Mouse.maximum.y - UI.parameter.charheigh/2;
      }
    else
      Mouse.position = Mouse.maximum;
  Mouse.button = (struct mouse_button_str){ hidden: 1 };

  PDBG ((", enabling mouse driver: "));
  if (Mouse.type != MOUSE_PS2 || ((status = _PS_enable(1)) == 0)) {
      MOUSE_move (0, 0, (struct mouse_button_str) {});
      if (Mouse.type != MOUSE_PS2)
	  PDBG (("serial] "));
	else
	  PDBG (("PS2 enable OK] "));
      }
    else {
      PDBG (("PS2 enable error, status = 0x%X] ", status));
      Mouse.maximum = ((coord) {0,0});
      }
  }}

static void MOUSE_stop (void)
  {bound_stack();{
  unsigned char status;

  PDBG ((" [%s] ", __FUNCTION__));
  if (Mouse.maximum.x == 0)
      return;

  MOUSE_move (0, 0, (struct mouse_button_str) { hidden: 1 });

  Mouse.position.x /= UI.parameter.charwidth;
  Mouse.position.y /= UI.parameter.charheigh;
  /* will keep mouse at bottom right if it is there */
  if (Mouse.position.x >= UI.parameter.nbcol - 2)
      Mouse.position.x = 0xFFFFU / 8; /* over limits */
  if (Mouse.position.y >= UI.parameter.nbrow - 2)
      Mouse.position.y = 0xFFFFU / 16; /* over limits */

  Mouse.maximum = ((coord) {0,0});

  PDBG ((" [disabling mouse driver: "));
  if (Mouse.type == MOUSE_PS2) {
      if ((status = _PS_enable(0)) != 0)
	  PDBG (("error, status = 0x%X] ", status));
	else
	  PDBG (("OK] "));
      }
    else
      PDBG ((" serial] "));
  }}

/**
 ** The initialisation:
 **/
static inline unsigned
MOUSE_PS2_init (void)
  {
#if USER_SUPPORT & BIOS_MOUSE_SUPPORT
  struct PS_resetID mouseconf;
  unsigned char status;
  extern char PS2_mouse_callback[];

  PDBG (("Mouse: "));
  if ((status = _PS_reset(&mouseconf)) != 0) {
      PDBG (("reset error, status = 0x%X\r\n", status));
      return 1;
      }
  PDBG (("device ID: 0x%X, reset value: 0x%X",
	mouseconf.device_id, mouseconf.reset_value));

  if ((status = _PS_initialise(3)) != 0) {
      PDBG (("init error, status = 0x%X\r\n", status));
      return 2;
      }

  if ((status = _PS_set_handler(code_adr(PS2_mouse_callback))) != 0) {
      PDBG (("set driver error, status = 0x%X\r\n", status));
      return 3;
      }

  if ((status = _PS_set_sampling(5 /* 100/s */)) != 0)
      PDBG (("set sampling error, status = 0x%X\r\n", status));
  if ((status = _PS_set_resolution(3 /* 8/mm */)) != 0)
      PDBG (("set resolution error, status = 0x%X\r\n", status));

  PDBG (("\r\n"));

  Mouse.type = MOUSE_PS2;
  asm volatile (" dataPS2 = %c0 ": : "i" (&Mouse.data.PS2));
  return 0;
#else /* BIOS_MOUSE_SUPPORT */
  return 1;
#endif /* BIOS_MOUSE_SUPPORT */
  }

#if USER_SUPPORT & BIOS_MOUSE_SUPPORT
asm (
"PS2_mouse_callback:							\n"
"	pushl	%ds							\n"
"	pushl	%eax							\n"
"	movw	%cs,%ax							\n"
"	addw	$deltaseg,%ax	# not necessary if no CODE_SEGMENT	\n"
"	movw	%ax,%ds							\n"
"	movl	12(%esp),%eax						\n"
"	movl	%eax,dataPS2						\n"
"	movl	16(%esp),%eax						\n"
"	movl	%eax,dataPS2 + 4					\n"
"	popl	%eax							\n"
"	popl	%ds							\n"
"	lretw								\n"
	);
#endif /* BIOS_MOUSE_SUPPORT */

static inline unsigned
MOUSE_SERIAL_init (unsigned nb_serial_port)
  {
#if USER_SUPPORT & SERIAL_MOUSE_SUPPORT
  unsigned short cpt, last;
  /* inline */ const struct serial_setup_str
      mouse_std_setup = {
		word_length: serial_7bit,
		stop_bit: serial_1stopbit,
		parity: serial_none,
		speed: B1200b
		},
      mouse_mm_setup = {
		word_length: serial_8bit,
		stop_bit: serial_1stopbit,
		parity: serial_none,
		speed: B1200b
		};
  /* inline */ const struct serial_modem_str
		setmodem = { dtr: 1, rts: 1 },
		restmodem = { dtr: 1, rts: 0 },
		clearmodem = { dtr: 0, rts: 0 };

  if (nb_serial_port > 4)
      last = 4; /* due to the inb(), due to BIOS hardware flow control */
    else
      last = nb_serial_port;

  for (cpt = 0; cpt < last; cpt++) {
      unsigned short ioaddr = peekw (0x00400000 + (2*cpt));
      unsigned char  i, j, k, array[5];
      _BIOS_serialstatus status;

      if ((unsigned short)(BOOT1_COM_port()) == cpt) {
	  PDBG (("Do not check COM%u, it is stdout\r\n", cpt));
	  continue;
	  }
#if (DEBUG & DEBUG_OUTPUT) == COM1
      if (cpt == 0) {
	  PDBG (("Do not check COM1, it is stderr\r\n"));
	  continue;
	  }
#elif (DEBUG & DEBUG_OUTPUT) == COM2
      if (cpt == 1) {
	  PDBG (("Do not check COM2, it is stderr\r\n"));
	  continue;
	  }
#elif (DEBUG & DEBUG_OUTPUT) == COM3
      if (cpt == 2) {
	  PDBG (("Do not check COM3, it is stderr\r\n"));
	  continue;
	  }
#elif (DEBUG & DEBUG_OUTPUT) == COM4
      if (cpt == 3) {
	  PDBG (("Do not check COM4, it is stderr\r\n"));
	  continue;
	  }
#endif

      PDBG (("Checking serial mouse on COM%u, I/O address 0x%X: ",
		cpt+1, ioaddr));

#if 0
      {
      /* loopback => no more power supply to mouse */
      const struct serial_modem_str resetmodem = {
	  dtr: 0, rts: 0, loopback: 1
	  };
      SETMODEM (cpt, ioaddr, resetmodem);
      _BIOS_wait (50000); /* 50 ms */
      SETMODEM (cpt, ioaddr, clearmodem);
      _BIOS_wait (50000); /* 50 ms */
      }
#endif
#if 1
      /* Oh, less than 50 ms would be enough, but you know,
	 micro$oft has also made hardware - and on a 16650... */
      /* Maybe have a look at APM->serial or _BIOS_initserial() before ? */
      SETMODEM (cpt, ioaddr, setmodem);
      _BIOS_wait (400000); /* 400 ms */
#endif
      _BIOS_initserial (cpt, mouse_std_setup);
      SETMODEM (cpt, ioaddr, restmodem);
      _BIOS_wait (100000); /* 100 ms */

      status = _BIOS_getserialstatus (cpt);

      if (status.receive_data_ready) {
	  volatile unsigned char tmp __attribute__ ((unused)) = inb (ioaddr);
	  PDBG (("0x%X, ", tmp));
	  }

      PDBG ((" CTS=%u, DSR=%u, RI=%u, CD=%u; ",
		status.al.modem.cts, status.al.modem.dsr,
		status.al.modem.ri, status.al.modem.cd));

      SETMODEM (cpt, ioaddr, setmodem);

      for (i = 30, j = 0; i != 0 && j < sizeof (array); --i) {
	  _BIOS_wait (9000); /* 10 bits at 1200 bauds: 8.3 ms => 9 ms */
	  status = _BIOS_getserialstatus(cpt);
	  if (status.receive_data_ready) {
	      array[j] = inb (ioaddr);
	      PDBG (("0x%X, ", array[j]));
	      if (array[j])	/* some UARTs, when break received */
		j++;
	      }
	  }

      status = _BIOS_getserialstatus (cpt);
      PDBG ((" DTR and RTS now set; CTS=%u, DSR=%u, RI=%u, CD=%u; ",
		status.al.modem.cts, status.al.modem.dsr,
		status.al.modem.ri, status.al.modem.cd));

      k = 0;
      while (j != 0) {
	  if (j >= 2 && array[j-2] == 'M' && array[j-1] == '3') {
	      PDBG (("serial 3 button mouse\r\n"));
	      Mouse.type = MOUSE_SERIAL3_COM1 + cpt;
	      Mouse.data.serial.comIOaddr = ioaddr;
	      Mouse.data.serial.comport = cpt;
	      return 0;
	      }
	  if (array[j-1] == 'M') {
	      PDBG (("serial std mouse\r\n"));
	      Mouse.type = MOUSE_SERIAL_COM1 + cpt;
	      Mouse.data.serial.comIOaddr = ioaddr;
	      Mouse.data.serial.comport = cpt;
	      return 0;
	      }
	  if (array[j-1] == 'H') {
	      _BIOS_initserial (cpt, mouse_mm_setup);
	      PDBG (("serial MM mouse\r\n"));
	      Mouse.type = MOUSE_SERIALM_COM1 + cpt;
	      Mouse.data.serial.comIOaddr = ioaddr;
	      Mouse.data.serial.comport = cpt;
	      return 0;
	      }
	  /* Leave this debug message for some time: */
	  if (k == 0) {
	      PRINTF (" [on COM%u, %u bytes: 0x%X", cpt+1, j, array[k++]);
	      i = j;
	      while (--i)
		  PRINTF (", 0x%X", array[k++]);
	      PRINTF ("] ");
	      }
	  j--;
	  }
      PDBG (("no mouse there\r\n"));
      SETMODEM (cpt, ioaddr, clearmodem);
      }
#endif /* SERIAL_MOUSE_SUPPORT */
  return 1;
  }

unsigned MOUSE_init (unsigned nb_serial_port)
  {bound_stack();{
  Mouse.maximum = ((coord) {0,0});

#ifndef MOUSE_ON_VT
  if (BOOT1_COM_port() >= 0) {
      PDBG (("MOUSE on VT not compiled-in\r\n"));
      return 2;
      }
#endif

  if (MOUSE_PS2_init() == 0 || MOUSE_SERIAL_init(nb_serial_port) == 0) {
      UI.mouse.start = MOUSE_start;
      UI.mouse.stop = MOUSE_stop;
      UI.mouse.poll = MOUSE_poll;
      return 0;
      }
  PDBG (("MOUSE initialisation failed\r\n"));
  return 1;
  }}


/**
 ** The global variable inited to zero because in BSS:
 **/
struct mouse_interface_str Mouse;
