/* tester.c  => linked to main.c */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"
#include "library.h"
#include "boot.h"
#include "debug.h"
#include "user.h"	/* define UI which is the user interface */
#include "disk.h"
#include "bios.h"	/* _DOSEMU_check() */
#include "malloc.h"	/* init_malloc(), report_malloc() */
#include "dos.h"	/* DOS_version() */
#include "xms.h"

#include "vesabios.h"	/* _VESA_GetPMinterface() and _VESA_setmode() */
#include "vgabios.h"	/* _VGA_get_display_combination_code() */

#if USER_SUPPORT == 0
#define USER_ACTIVE() 0
#define VGA_ACTIVE()  0
#define VESA_ACTIVE() 0
#endif

/********************************************************
 * Few "helper" functions, they are inline, so do not take any
 * space if they are not called: (do not call them more than once)
 *  ---
 *  PROBLEM with GCC: their strings are _not_ removed
 *  even if those functions are not inserted anywhere...
 *  ---
 */

#if USER_SUPPORT & VGA_SUPPORT
static inline void listmode (struct user_interface_str *ui)
  {
  unsigned short start = 0, cpt, firsttime = 1,
		 valid = VIDEO_mode_is_valid(0);

  for (cpt = 1; cpt <= VIDEO_mode_max(); cpt++) {
      if (   cpt == VIDEO_mode_max()
	  || VIDEO_mode_is_valid(cpt) != valid) {
	  if (valid == 0)
	      start = cpt;
	    else {
	      if (!firsttime)
		  printf (", ");
	      firsttime = 0;
	      printf ((start < cpt-2)
			? "0x%X..0x%X"
			: (start == cpt-2)? "0x%X, 0x%X" : "0x%X",
			start, cpt-1);
	      }
	  valid = !valid;
	  }
      }
  if (firsttime)
      printf ("none.\r\n");
    else
      printf (".\r\n");
  }
#endif /* VGA_SUPPORT */

static inline void list_disks (void)
  {
  unsigned short number;
  char disk_name[SIZE_DISKNAME];


  for (number = 0; number < DI.nbdisk; number++) {
      unsigned long long size = DI.param[number].nbtotalsector
				* DI.param[number].bytepersector;
      diskname (&DI.param[number], disk_name);
      printf ("    disk %u: %s, size %u %s, %u partitions\r\n",
		number,
		disk_name,
		(int)((size > 9*1024*1024) ? size/(1024*1024)
					   : size/1024),
		(size > 9*1024*1024) ? "Mb" : "Kb",
		DI.param[number].nbpartition
		);
      }
  }

static inline const char *const PartName (unsigned char parttype)
  {
  const char *partname;

#define P(ident, nb, str) if (parttype == nb) partname = str; else
	  PARTITIONLIST()	partname = "unknown";
#undef P
  return partname;
  }

#if USER_SUPPORT & VESA_HARDWINDOW
static inline void
display_hardwin (unsigned short *offsetarray, struct hardwindow_str *hardwindow)
  {
  unsigned short i, j;

  if (offsetarray)
      printf ("Win0 [%u]: ", offsetarray[0]);
    else
      printf ("AllWin: ");
  if (hardwindow == 0) {
      printf ("NO hardwin.\r\nEnd.");
      return;
      }
  for (i = j = 0;; i++) {
      if (   hardwindow[i].instruction == 0
	  && hardwindow[i].data == 0
	  && hardwindow[i].port == 0) {
	  j++;
	  if (offsetarray == 0 || offsetarray[j+1] == 0) {
	      /* end array kept, end marker written after */
	      printf ("\r\nEnd.");
	      break;
	      }
	    else
	      printf ("\r\nWin%u [%u]: ", j, offsetarray[j]);
	  }
	else {
	  unsigned short instruction = hardwindow[i].instruction;
	  unsigned char tmp;

#ifdef COMPRESS_HARDWIN
	  const char *str = 0;

	  if (instruction != 0xEF66 && (tmp = (instruction >> 8)) != 0) {
	      if (tmp == 1)
		  str = "(page & 0x0F)";
		else if (tmp == 3)
		  str = "((page & 0x0F) << 4)";
		else if (tmp == 19)
		  str = "page";
		else if (tmp == 20)
		  str = "(page >> 4)";
		else if (tmp == 28)
		  str = "(page << 2)";
		else if (tmp == 67)
		  str = "((page & 0xF0) >> 2)";
#ifdef COMPRESS_EXTENDED
		else if ((tmp == 37) && ((instruction & 0xFF) == 0xEE))
		  str = "((inb(0x%X) & 0xF0) ^ (page & 0x0F))";
		else if ((tmp == 11) && ((instruction & 0xFF) == 0xEE))
		  str = "((inb(0x%X) & 0x0F) ^ ((page & 0x0F) << 4))";
#endif
		else
		  str = "ERROR ";
	      instruction &= 0xFF;
	      }
#endif /* COMPRESS_HARDWIN */

	  if (instruction == 0xEF66)
	      tmp = 'l';
	    else if (instruction == 0xEE)
	      tmp = 'b';
	    else if (instruction == 0xEF)
	      tmp = 'w';
	    else
	      tmp = '?';

	  printf ("out%c (0x%X, 0x%X",
		  tmp, hardwindow[i].port, hardwindow[i].data);
#ifdef COMPRESS_HARDWIN
	  if (str) {
	      printf (" ^ ");
	      if (tmp == 'b')
#ifdef COMPRESS_EXTENDED
		  printf (str, hardwindow[i].port);
#else
		  printf (str);
#endif
		else
		  printf ("(%s << 8)", str);
	      }
#endif /* COMPRESS_HARDWIN */
	  printf ("); ");
	  }
      }
  }
#endif /* VESA_HARDWINDOW */

static inline void dump_registers (struct registers *regs)
  {
  printf ("Dumping registers found by boot1:\r\n"
	  " eax = 0x%X ebx = 0x%X ecx = 0x%X\r\n"
	  " edx = 0x%X esi = 0x%X edi = 0x%X ebp = 0x%X\r\n"
	  "  ds = 0x%X  es = 0x%X  fs = 0x%X  gs = 0x%X\r\n"
	  "  ip = 0x%X  cs = 0x%X flags = 0x%X  initsp = 0x%X\r\n\r\n",
		regs->eax, regs->ebx, regs->ecx, regs->edx,
		regs->esi, regs->edi, regs->ebp,
		regs->ds, regs->es, regs->fs, regs->gs,
		regs->ip, regs->cs, regs->flags, regs->initsp);

  printf ("Current cs = 0x%X%c  ds = 0x%X, es = 0x%X, ss = 0x%X, sp = 0x%X\r\n"
	  "  fs = 0x%X, gs = 0x%X, freemem 0x%X, "
	  "stack used: %u Kb / total: %u Kb.\r\n"
	  " _etext=0x%X%c"
#if SETUP & EXTRA_SEGMENT
	  " _eextra=0x%X;"
#endif
	  " _end=0x%X, _sbss=0x%X, _edata=0x%X\r\n",
		getcs(),
#if SETUP & CODE_SEGMENT
		';',
#else
		',',
#endif
		getds(), getes(), getss(), getesp(),
		getfs(), getgs(), freemem(),
		usedstack () / 1024,
		totalstack () / 1024,
		(unsigned)_etext,
#if SETUP & CODE_SEGMENT
		';',
#else
		',',
#endif
#if SETUP & EXTRA_SEGMENT
		(unsigned)_eextra,
#endif
		(unsigned)_end,
		(unsigned)_sbss, (unsigned)_edata );
  }

static inline void display_param (void)
  {
  struct video_parameter_str *str = &UI.parameter;

  printf ("UI: monitor:0x%X, ", UI.monitor);
#if USER_SUPPORT & VESA_EDID
  printf ("%u cm x %u cm, ", UI.edid.Hsize_cm, UI.edid.Vsize_cm);
#endif
  printf ("fgcolor:%u, bgcolor:%u,\r\n", UI.fgcolor, UI.bgcolor);

#if USER_SUPPORT & VGA_FREQUENCY
  printf ("frequency: %u, ", UI.frequency);
#endif
#if USER_SUPPORT & VESA_SUPPORT
  printf ("VESA %u.%u, ",
	  UI.info.vesa.version >> 8,
	  UI.info.vesa.version & 0xFF);
#endif
  printf ("attr current:0x%X, attr valid:0x%X\r\n",
	  *(unsigned char*)&UI.attributes.current,
	  *(unsigned char*)&UI.attributes.valid);

  printf ("UI.param: id: 0x%X, %s mode, base_address: 0x%X\r\n",
	   str->identification,
	   str->attr.isgraphic? "graphic" : "text",
	   str->base_address);

  printf ("size: %ux%u in %u colors, i.e. %ux%u of %ux%u chars\r\n",
	  str->width, str->heigh,
	  str->nbcolor,
	  str->nbcol, str->nbrow,
	  str->charwidth, str->charheigh);

  printf ("red: %u<<%u, green: %u<<%u, blue: %u<<%u",
	  str->red_mask_size, str->red_field_pos,
	  str->green_mask_size, str->green_field_pos,
	  str->blue_mask_size, str->blue_field_pos
	  );

#if USER_SUPPORT & VESA_SUPPORT
  printf (", font: 0x%X\r\n", (unsigned)str->font);

  if (VESA_ACTIVE()) {
      printf ("linelength: %u, winadr: 0x%X, winsize: 0x%X, EGA compat: %s\r\n",
	str->linelength, str->winadr, str->winsize,
	str->EGA_compatible? "yes" : "no"
	);

#if USER_SUPPORT & VESA_WINDOW
      printf ("NbWin = %u, winNo = %u, wingranul: 0x%X",
	      str->NbWin, str->winNo,
#if USER_SUPPORT & VESA_STRANGE_WINDOW
	      str->wingranul);
#else
	      1 << str->wingranul);
#endif
#if USER_SUPPORT & VESA_2WINDOWS
      printf (", readdelta = 0x%X, winreadadr = 0x%X\r\n",
	      str->readdelta, str->winreadadr);
#else
      printf ("\r\n");
#endif
#endif /* VESA_WINDOW */

      printf ("memtype: %u, bit/pixel: %u, nbplane: %u\r\n",
	      str->memtype, str->bitperpixel, str->nbplane);

#if USER_SUPPORT & VESA_PMINFO
      if (   UI.ProtectedModeIO[0] != 0xFFFF
	  || UI.ProtectedModeMem[0].address != 0xFFFFFFFFU) {
	  unsigned i = 0;

	  printf ("IO ports used: %s",
		  (UI.ProtectedModeIO[i] == 0xFFFF)? "none" : "");
	  while (   UI.ProtectedModeIO[i] != 0xFFFF
		 && i < nbof (UI.ProtectedModeIO)) {
	      printf ("0x%X, ", UI.ProtectedModeIO[i]);
	      i++;
	      }
	  i = 0;
	  printf ("\r\nMemory used: %s",
		    (UI.ProtectedModeMem[i].address == 0xFFFFFFFFU)
			? "none" : "");
	  while (UI.ProtectedModeMem[i].address != 0xFFFFFFFFU
		 && i < nbof (UI.ProtectedModeMem)) {
	      printf ("0x%X+0x%X, ", UI.ProtectedModeMem[i].address,
				     UI.ProtectedModeMem[i].length);
	      i++;
	      }
	  printf ("\r\n");
	  }
#endif /* VESA_PMINFO */
      }
#endif /* VESA_SUPPORT */

  printf ("\r\n");
#define ATTR2BYTE(A)    (*(unsigned char *)&(A))
  if (ATTR2BYTE(UI.attributes.valid) != ATTR2BYTE(no_attribute)) {
#undef ATTR2BYTE
      UI_setattribute (no_attribute);
      printf ("normal");
      if (UI.attributes.valid.underline) {
	  printf (" | ");
	  UI_setattribute (attr_underline);
	  printf ("underline");
	  UI_setattribute (no_attribute);
	  }
      if (UI.attributes.valid.reverse) {
	  printf (" | ");
	  UI_setattribute (attr_reverse);
	  printf ("reverse");
	  UI_setattribute (no_attribute);
	  }
      if (UI.attributes.valid.blink) {
	  printf (" | ");
	  UI_setattribute (attr_blink);
	  printf ("blink");
	  UI_setattribute (no_attribute);
	  }
      if (UI.attributes.valid.brighter) {
	  printf (" | ");
	  UI_setattribute (attr_brighter);
	  printf ("brighter");
	  UI_setattribute (no_attribute);
	  }
      if (UI.attributes.valid.darker) {
	  printf (" | ");
	  UI_setattribute (attr_darker);
	  printf ("darker");
	  UI_setattribute (no_attribute);
	  }
      if (UI.attributes.valid.transparent) {
	  printf (" |  ~-_\010\010\010\010");
	  UI_setattribute (attr_transparent);
	  printf ("transparent");
	  UI_setattribute (no_attribute);
	  }
      printf ("\r\n");
      }

  if (str->nbcolor > 2) {
      unsigned save_bgcolor, save_fgcolor;

      save_fgcolor = UI.fgcolor;
      save_bgcolor = UI.bgcolor;

      UI_setbgcolor (UI.stdcolor[black]);
      printf ("bg=black");
      UI_setfgcolor (UI.stdcolor[white]);
      printf (" WHITE");
      UI_setfgcolor (UI.stdcolor[yellow]);
      printf (" YELLOW");
      UI_setfgcolor (UI.stdcolor[lightmagenta]);
      printf (" LIGHTMAGENTA");
      UI_setfgcolor (UI.stdcolor[lightred]);
      printf (" LIGHTRED");
      UI_setfgcolor (UI.stdcolor[lightcyan]);
      printf (" LIGHTCYAN");
      UI_setfgcolor (UI.stdcolor[lightgreen]);
      printf (" LIGHTGREEN");
      UI_setfgcolor (UI.stdcolor[lightblue]);
      printf (" LIGHTBLUE");
      if (str->nbcol <= 80) {
	  UI_setbgcolor (save_bgcolor);
	  UI_setfgcolor (save_fgcolor);
	  printf ("\r\n");	/* scroll => fill line with bgcolor */
	  UI_setbgcolor (UI.stdcolor[black]);
	  }
      UI_setfgcolor (UI.stdcolor[darkgray]);
      printf (" DARKGRAY");
      UI_setfgcolor (UI.stdcolor[lightgray]);
      printf (" LIGHTGRAY");
      UI_setfgcolor (UI.stdcolor[brown]);
      printf (" BROWN");
      UI_setfgcolor (UI.stdcolor[magenta]);
      printf (" MAGENTA");
      UI_setfgcolor (UI.stdcolor[red]);
      printf (" RED");
      UI_setfgcolor (UI.stdcolor[cyan]);
      printf (" CYAN");
      UI_setfgcolor (UI.stdcolor[green]);
      printf (" GREEN");
      UI_setfgcolor (UI.stdcolor[blue]);
      printf (" BLUE");

      UI_setbgcolor (save_bgcolor);
      UI_setfgcolor (save_fgcolor);

      UI_setfgcolor (UI.stdcolor[black]);
      printf ("\r\nfg=black");
      UI_setbgcolor (UI.stdcolor[white]);
      printf (" WHITE");
      UI_setbgcolor (UI.stdcolor[yellow]);
      printf (" YELLOW");
      UI_setbgcolor (UI.stdcolor[lightmagenta]);
      printf (" LIGHTMAGENTA");
      UI_setbgcolor (UI.stdcolor[lightred]);
      printf (" LIGHTRED");
      UI_setbgcolor (UI.stdcolor[lightcyan]);
      printf (" LIGHTCYAN");
      UI_setbgcolor (UI.stdcolor[lightgreen]);
      printf (" LIGHTGREEN");
      UI_setbgcolor (UI.stdcolor[lightblue]);
      printf (" LIGHTBLUE");
      if (str->nbcol <= 80) {
	  UI_setbgcolor (save_bgcolor);
	  UI_setfgcolor (save_fgcolor);
	  printf ("\r\n");	/* scroll => fill line with bgcolor */
	  UI_setfgcolor (UI.stdcolor[black]);
	  }
      UI_setbgcolor (UI.stdcolor[darkgray]);
      printf (" DARKGRAY");
      UI_setbgcolor (UI.stdcolor[lightgray]);
      printf (" LIGHTGRAY");
      UI_setbgcolor (UI.stdcolor[brown]);
      printf (" BROWN");
      UI_setbgcolor (UI.stdcolor[magenta]);
      printf (" MAGENTA");
      UI_setbgcolor (UI.stdcolor[red]);
      printf (" RED");
      UI_setbgcolor (UI.stdcolor[cyan]);
      printf (" CYAN");
      UI_setbgcolor (UI.stdcolor[green]);
      printf (" GREEN");
      UI_setbgcolor (UI.stdcolor[blue]);
      printf (" BLUE");

      UI_setbgcolor (save_bgcolor);
      UI_setfgcolor (save_fgcolor);
      printf ("\r\n");
      }
  }

static inline void display_pattern (void)
  {
  unsigned i;
  char tmp[2] = { '.', 0 };

  UI_setcursor (0, 0);
  printf ("\r\n mode : 0x%X, size: %ux%u, %s\r\n",
	  UI.parameter.identification, UI.parameter.nbcol, UI.parameter.nbrow,
	  UI.parameter.attr.isgraphic? "graphic" : "text");
  printf (" color: %u*%ux%u, char: %ux%u",
	  UI.parameter.nbcolor, UI.parameter.width, UI.parameter.heigh,
	  UI.parameter.charwidth, UI.parameter.charheigh);

#if USER_SUPPORT & VGA_FREQUENCY
  if (UI.frequency)
      printf (", %u Hz", UI.frequency);
#endif

  for (i = 0; i < 256; i++) {
      if (i % 16 == 0) {
	  UI_setcursor (5 + i/16, 0);
	  printf ("0x%X", i);
	  if (i < 16)
	      printf ("0 ");
	    else
	      printf (" ");
	  }
      if ((BOOT1_COM_port() >= 0)
			? ((i >= 0x20) && (i < 0x80 || i >= 0xA0))
			: (i != '\t' && i != '\n' && i != '\r'
				     && i != '\007' && i != '\010') )
	  tmp[0] = i;
	else
	  tmp[0] = '.';
      UI_putstr (tmp);
      }

  if (UI.parameter.nbcolor != 0) {
      unsigned save_bgcolor, save_fgcolor;

      save_fgcolor = UI.fgcolor;
      save_bgcolor = UI.bgcolor;
      for (i = 0; i < 16; i++) {
	  unsigned j;

	  UI_setcursor (i+5, 23);
	  if (i < UI.parameter.nbcolor)
	      UI_setfgcolor (i);
	    else
	      UI.fgcolor = i;

	  for (j = 0; j < 16; j++) {
	      if (j < UI.parameter.nbcolor)
		  UI_setbgcolor (j);
		else
		  UI.bgcolor = j;
	      printf ("@");	// printf("\333");
	      }
	  }
      UI_setbgcolor (save_bgcolor);
      UI_setfgcolor (save_fgcolor);
      /* force it (for VT420 terminal) : */
      UI.fgcolor = save_fgcolor;
      UI.bgcolor = save_bgcolor;
      }
    else {
      unsigned savedcolor = UI.fgcolor;	/* attributes, not for serial */
      for (i = 0; i < 256; i++) {
	  UI_setcursor (i/16+5, 23+i%16);
	  UI.fgcolor = i;
	  printf ("@");
	  }
      UI.fgcolor = savedcolor;
      }

  /* Screen is now filled up to character 39, i.e. 320 pixel */

  if (UI.parameter.attr.isgraphic) {
      if (UI.parameter.width < 600) {
	  unsigned y, cpt;
	  for (y = 0; y < 16; y++)
	      for (cpt = 0; cpt < UI.parameter.charheigh; cpt++) {
		  coord xy = {
		      x: 79 * UI.parameter.charwidth / 2,
		      y: (5+y) * UI.parameter.charheigh + cpt
		      };
		  UI_setpixel (xy, y+cpt);
		  }
	  }
	else {
	  unsigned x, y;
	  for (x = 0; x < 256; x++)
	      for (y = 0; y < 16U * UI.parameter.charheigh; y++) {
		  coord xy = {
		      x: 320 + x,
		      y: 5 * UI.parameter.charheigh + y
		      };
		  UI_setpixel (xy, (y << 8) | x);
		  }
	  }
      }

#if 0
  if (UI.parameter.nbcolor == 1 << 15) { /* what does setting 16th bit do ? */
      unsigned x;
      for (x = 0; x < 256; x++)
	  UI_setpixel (320 + x, 5*UI.parameter.charheigh - 5,
				x | 0x8000);
      }
#endif

  UI_setcursor (22, 0);
  }

static inline void detect_serial_timeout (void)
  {
  unsigned char number, tmp, max;

  DBG (("Timeout for "));
  max = _BIOS_equipment_flags().nb_serial_port;
  printf ("%u serial port%s", max, (max > 1) ? "s" : "");

  for (number = 0; number < max; number++) {
      tmp = peekb (0x0040007C + number);
      DBG (("COM%u: 0x%X", number+1, tmp));
      if (tmp == 0 || tmp > 5) {
	  DBG ((" set to 1"));
	  pokeb (0x0040007C + number, 1);
	  }
      if (number < max - 1)
	  DBG ((", "));
	else
	  DBG (("\r\n"));
      }
  }

static inline void detect_environment (void)
  {
  unsigned short dosemu_version, dosemu_patchlevel, dos_version;
  const char *msg = "?????";
  char cpuid_max = 0;

  if (!is_80486())
      msg = "80386";
    else if (!has_cpuid ())
      msg = "80486 w/o CPUID";
    else {
      static const struct {
	  union cpuid_string cstr;
	  const char *msg;
	  } cpu_data[] = {
	  { cstr: { string: "GenuineIntel" } , msg: "Intel" },
	  { cstr: { string: "AuthenticAMD" } , msg: "AMD" },
	  { cstr: { string: "UMC UMC UMC " } , msg: "UMC" },
	  { cstr: { string: "CyrixInstead" } , msg: "Cyrix" },
	  { cstr: { string: "NexGenDriven" } , msg: "NexGen" },
	  { cstr: { string: "CentaurHauls" } , msg: "IDT" },
	  { cstr: { string: "RiseRiseRise" } , msg: "Rise" },
	  { cstr: { string: "GenuineTMx86" } , msg: "TRANSMETA" }
	  };
      union cpuid_string cpuidstr;
      unsigned short i;

      cpuid_max = cpuid0 (&cpuidstr);

      for (i = 0; i < nbof (cpu_data); i++)
	  if (   cpuidstr.number.str1 == cpu_data[i].cstr.number.str1
	      && cpuidstr.number.str2 == cpu_data[i].cstr.number.str2
	      && cpuidstr.number.str3 == cpu_data[i].cstr.number.str3) {
	      msg = cpu_data[i].msg;
	      break;
	      }
      if (i == nbof (cpu_data))
	  cpuid_max = 0;
      }

  if (cpuid_max > 0) {
      struct cpuid_str CpuId = cpuid1();
      printf ("%s family %d processor", msg, CpuId.family);
      }
    else
      printf ("%s processor", msg);

  if (!_BIOS_equipment_flags().coprocessor80x87installed)
      printf (" w/o FPU");

  if (BOOT1_DOS_running()) {
      printf (" on DOS");
      dos_version = DOS_version ();
      printf (" %u.%u", dos_version & 0xFF, dos_version >> 8);
      }
    else
      printf (" on BIOS");

  if (_DOSEMU_check (&dosemu_version, &dosemu_patchlevel) == 0)
      printf (", DOSEMU: V%u.%u-%u",
		dosemu_version >> 8,
		dosemu_version & 0xFF, dosemu_patchlevel);

  if (BOOT1_COM_port() < 0) {
      if (_BIOS_is_extended_keyboard()) {
	  if (_BIOS_is_122key_keyboard())
	      printf (", 122 key");
	    else
	      printf (", ext");
	  }
	else
	  printf (", std");
      printf (" keyboard");
      }
  printf (".\r\n");

  detect_serial_timeout ();

  printf (".\r\n");
  }

static inline unsigned largest_free_himem (void)
  {
  farptr entrypoint;

  DBG ((" [XMS: "));
  entrypoint = _XMS_get_entry_point ();
  DBG (("(at 0x%X) ", entrypoint));
  if ((entrypoint >> 16) != 0) { /* segment not null */
      unsigned short version_bcd, revision, flags, free, largest;
      unsigned char  error1;

      version_bcd = _XMS_get_version (entrypoint, &revision, &flags);
      DBG (("v%u.%u, rev 0x%X, HMA %s; ",
		version_bcd >> 8, version_bcd & 0xFF,
		revision, (flags & 1)? "present" : "abscent"));
      free = _XMS_query_free_memory (entrypoint, &largest, &error1);
      DBG (("free mem: %u Kb, largest: %u Kb, error: 0x%X; ",
		free, largest, error1));
      if ((version_bcd >> 8) >= 3) {
	  unsigned largest_KB, highest_memory_address, total_KB_free;
	  unsigned char error2;

	  total_KB_free = _XMSv3_query_free_memory (entrypoint,
			&largest_KB, &highest_memory_address, &error2);
	  DBG (("V3: total free: %u Kb, largest: %u Kb,"
		" highest address: 0x%X, error: 0x%X",
		total_KB_free, largest_KB,
		highest_memory_address, error2));
	  if (error2 == 0) {
	      DBG (("use V3 %u Kb] ", largest_KB));
	      return largest_KB;
	      }
	    else if (error1 == 0) {
	      DBG (("V3 error happend, use short value: %u Kb] ",
			largest));
	      return largest;
	      }
	    else {
	      DBG (("two error happened, returns 0] "));
	      return 0;
	      }
	  }
	else {
	  DBG (("old version, use short value: %u Kb] ", largest));
	  return largest;
	  }
      }
    else {
      DBG (("invalid address] "));
      return 0;
      }
  }

#if USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT)
static inline void detect_screen(void)
  {
  const char *msg;

  printf ("Screen interface: ");
  UI.monitor = BIOS_probemonitor();
  if (UI.monitor == 0)
      msg = "Unknown screen.\r\n";
    else if (UI.monitor == 1)
      msg = "B&W screen.\r\n";
    else if (UI.monitor == 2)
      msg = "Monochrome screen.\r\n";
    else if (UI.monitor == 3)
      msg = "Color screen.\r\n";
    else
      msg = "Color screen type 0x%X.\r\n";
  printf (msg, UI.monitor);

#if USER_SUPPORT & VESA_EDID
  if (UI.edid.horizontal_resolution != 0) {
      printf ("        "
	      "EDID: %ux%u @ %u Hz on a %u inches viewable monitor."
	      "\r\n",
		UI.edid.horizontal_resolution,
		UI.edid.vertical_resolution,
		UI.edid.frequency,
		(UI.edid.Vsize_cm + 3 * UI.edid.Hsize_cm / 4 + 2) / 3
		);
      }
#endif /* VESA_EDID */
  }
#endif /* VGA_SUPPORT | VESA_SUPPORT */

/********************************************************
 * Few functions to test different parts:
 */

/*
 * to test segment door closing / opening:
 */
#define FCT_DOOR_OPEN(seg) \
    static unsigned door_open_##seg (void) {	\
	segment_door_key = 0x76913010;		\
	trydoor_peek_##seg ();			\
	if (segment_door_key == 0x76913010) {	\
	    segment_door_key = 0;		\
	    return 1; /* is open */		\
	    }					\
	  else {				\
	    segment_door_key = 0;		\
	    return 0; /* is closed */		\
	    }					\
	}

FCT_DOOR_OPEN(cs)
FCT_DOOR_OPEN(ds)
FCT_DOOR_OPEN(es)
FCT_DOOR_OPEN(fs)
FCT_DOOR_OPEN(gs)
FCT_DOOR_OPEN(ss)

/*
 * to force stack underflow or test pointer checking :
 */
static void test_bound_stack (void (*fct) (void *))
  {bound_stack();{
  if (fct)
      (*(void (*)(void *)) fct) (fct);
  }}

static void test_bound_data (unsigned addr)
  {
  checkptr ((unsigned char *)addr, "Bound data test");
  }

#if SETUP & EXTRA_SEGMENT
/*
 * Test extrasegment handling:
 */
static void putstr (const char *str) /* not inline, codeseg */
  {
  UI_putstr (str);
  }

static void __attribute__ ((section (".Xextra2text") ))
extraseg_puts (const char *str)
  {
  EXTRASEG_REVERSE_STUB (putstr);
  }

static void __attribute__ ((section (".Xextra2text")))
test_intersegment_call (const char *param)
  {
  extraseg_puts (param);
  }

static void __attribute__ ((section (".Xtext2extra")))
codeseg_test_intersegment_call (const char *param)
  {
  EXTRASEG_STUB (test_intersegment_call);
  }
#define test_intersegment_call codeseg_test_intersegment_call
#else
static void test_intersegment_call (const char *param) { }
#endif /* EXTRA_SEGMENT */

/*
 * Test longjmp/setjmp:
 */
static void testjump (void)
  {
  static setjmp_buf buff;

  printf ("At entry:\r\n"
	  "eax = 0x%X, ebx = 0x%X, ecx = 0x%X, edx = 0x%X, \r\n"
	  "esi = 0x%X, edi = 0x%X, ebp = 0x%X, esp = 0x%X, \r\n"
	  "cs = 0x%X, ds = 0x%X, es = 0x%X, fs = 0x%X, gs = 0x%X, ss = 0x%X\r\n"
	  "flags = 0x%X\r\n",
	  geteax(), getebx(), getecx(), getedx(),
	  getesi(), getedi(), getebp(), getesp(),
	  getcs(), getds(), getes(), getfs(), getgs(), getss(),
	  getflags());

  if (setjmp (&buff) == 0) {
      printf ("\r\nHas set the jump buffer to:\r\n"
	      "eax = %s, ebx = 0x%X, ecx = %s, edx = %s, \r\n"
	      "esi = 0x%X, edi = 0x%X, ebp = 0x%X, esp = 0x%X, \r\n"
	      "cs = 0x%X, ds = 0x%X, es = 0x%X, fs = 0x%X, gs = 0x%X, ss = 0x%X\r\n"
	      "flags = 0x%X ; ip = 0x%X\r\n",
		"------", buff.ebx, "------", "------",
		  buff.esi, buff.edi, buff.ebp, buff.sp,
		  buff.cs, buff.ds, buff.es, buff.fs, buff.gs, buff.ss,
		  buff.flags, buff.ip);

      setes (0x0003);
      setfs (0x0004);
      setflags (0);

      printf ("\r\nHas modified few registers, now they are:\r\n"
	      "eax = 0x%X, ebx = 0x%X, ecx = 0x%X, edx = 0x%X, \r\n"
	      "esi = 0x%X, edi = 0x%X, ebp = 0x%X, esp = 0x%X, \r\n"
	      "cs = 0x%X, ds = 0x%X, es = 0x%X, fs = 0x%X, gs = 0x%X, ss = 0x%X\r\n"
	      "flags = 0x%X\r\n\r\n",
	      geteax(), getebx(), getecx(), getedx(),
	      getesi(), getedi(), getebp(), getesp(),
	      getcs(), getds(), getes(), getfs(), getgs(), getss(),
	      getflags());

      printf ("I will jump, hopes that the rope is strong enough\r\n");
      longjmp (&buff);
      printf ("\r\nYou should never read this message!\r\n");
      }
  printf ("\r\nBack to life after jumping, with:\r\n"
	  "eax = 0x%X, ebx = 0x%X, ecx = 0x%X, edx = 0x%X, \r\n"
	  "esi = 0x%X, edi = 0x%X, ebp = 0x%X, esp = 0x%X, \r\n"
	  "cs = 0x%X, ds = 0x%X, es = 0x%X, fs = 0x%X, gs = 0x%X, ss = 0x%X\r\n"
	  "flags = 0x%X\r\n\r\n",
	  geteax(), getebx(), getecx(), getedx(),
	  getesi(), getedi(), getebp(), getesp(),
	  getcs(), getds(), getes(), getfs(), getgs(), getss(),
	  getflags());
  }

/*
 * Test timer:
 */
static __attribute__ ((cdecl, regparm (0)))
void timer_handler (void) {
    extern setjmp_buf setjump_buffer;

    longjmp (&setjump_buffer);
    }

static void timertest (void)
  {
  extern setjmp_buf setjump_buffer;
  extern unsigned char timer_handler_ptr[], timer_counter[];

  codeseg_pokel (timer_handler_ptr, (unsigned)timer_handler);
  codeseg_pokel (timer_counter, 16); /* one second timeout */
  if (setjmp (&setjump_buffer) != 0) {
      codeseg_pokel (timer_handler_ptr, 0);
      UI_putstr ("timer expired\r\n");
      return;
      }
  UI_putstr ("waiting timer... ");
  for (;;);
  }

/*
 * To get the port address and value of VESA
 * graphic page switching code...
 */
//#define EXTENDED_RECORDER	/* just to play, do not work enough... */

static struct recorder_str {
    unsigned nbinstruction;
#ifdef EXTENDED_RECORDER
    unsigned short nbint;
    unsigned short nbpushf;
    unsigned short nbpopf;
    unsigned short nbiret;
    unsigned short nblock;
    unsigned short nbidt;
    unsigned short nblmsw;
    unsigned short nbcr0;
#endif
    /* : previous fields are assembly offset hardcoded */
    unsigned short nbwrite;
    unsigned short arraysize;
    struct recorder_array_str {
	unsigned short instruction;
	unsigned short addr;
	unsigned       value;
	} *array;
    } __attribute__ ((packed)) recorder asm ("recorder");

static void __attribute__ ((cdecl, regparm (0), unused))
			 // asm ("recorder_register")
recorder_register (const unsigned edx, const unsigned eax, unsigned bx)
	/* "const" is written in purpose ! */
  {
  if (recorder.nbwrite < recorder.arraysize) {
      recorder.array[recorder.nbwrite].instruction = (unsigned short) bx;
      recorder.array[recorder.nbwrite].addr = (unsigned short) edx;
      recorder.array[recorder.nbwrite].value = eax;
      }
  recorder.nbwrite++;
  }

asm (
".text						\n"
"	.align 0x20				\n"
"recorder_handler:				\n"
"	pushl	%ds				\n"
"	pushl	%ebx				\n"
"	movw	%cs,%bx				\n"
"	addw	$deltaseg,%bx			\n"
"	movw	%bx,%ds				\n"
"	incl	recorder+0			\n"
"	movw	%sp,%bx				\n"
"	movw	%ss:10(%bx),%bx			\n"
"	movw	%bx,%ds				\n"
"	movw	%sp,%bx				\n"
"	movw	%ss:8(%bx),%bx			\n"
"	movw	(%bx),%bx	# %ds is actually interrupted software %cs	\n"
"	cmpb	$0xEE,%bl			\n"
"	je	recorder_match			\n"
"	cmpb	$0xEF,%bl			\n"
"	je	recorder_match			\n"
"	cmpb	$0x6E,%bl			\n"
"	je	recorder_match			\n"
"	cmpb	$0x6F,%bl			\n"
"	je	recorder_match			\n"
"	cmpw	$0xEF66,%bx			\n"
"	je	recorder_match			\n"
"	# still no 'rep out*' tested, in fact all other prefixes.	\n"
"	"
#ifdef EXTENDED_RECORDER
"	"
"	cmpb	$0xF0,%bl			\n"
"	je	recorder_lock			\n"
"	cmpb	$0xCD,%bl			\n"
"	je	recorder_int			\n"
"	cmpb	$0x9C,%bl			\n"
"	je	recorder_pushf			\n"
"	cmpb	$0x9D,%bl			\n"
"	je	recorder_popf_iret		\n"
"	cmpb	$0xCF,%bl			\n"
"	je	recorder_popf_iret		\n"
"	cmpw    $0x010F,%bx			\n"
"	je	recorder_idt_gdt		\n"
"	cmpw    $0x200F,%bx			\n"
"	je	recorder_set_cr			\n"
"	cmpw    $0x220F,%bx			\n"
"	je	recorder_get_cr			\n"
"	cmpb	$0x17,%bl			\n"
"	je	recorder_popss			\n"
"	cmpb	$0x16,%bl			\n"
"	je	recorder_pushss			\n"
"	andw	$0xF8FF,%bx			\n"
"	cmpw	$0xD08E,%bx			\n"
"	je	recorder_movtoss		\n"
"	cmpw	$0xD08C,%bx			\n"
"	je	recorder_movfromss		\n"
"	"
#endif
"	"
"recorder_end:					\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	iretw					\n"
"						\n"
"recorder_match:				\n"
"	pushl	%ecx				\n"
"	pushl	%ebx				\n"
"	pushl	%eax				\n"
"	pushl	%edx				\n"
"	movw	%cs,%bx				\n"
"	addw	$deltaseg,%bx			\n"
"	movw	%bx,%ds			# may need to set %es also...	\n"
"	calll	recorder_register		\n"
"	popl	%edx				\n"
"	popl	%eax				\n"
"	popl	%ebx				\n"
"	popl	%ecx				\n"
"	jmp	recorder_end			\n"
"	"
#ifdef EXTENDED_RECORDER
"	"
"						\n"
"recorder_lock:					\n"
"	mov	%cs,%bx				\n"
"	add	$deltaseg,%bx			\n"
"	mov	%bx,%ds				\n"
"	incw	recorder+12			\n"
"	mov	%sp,%bx				\n"
"	incw	%ss:8(%bx)			\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	jmp	recorder_handler		\n"
"						\n"
"recorder_dummyss:				\n"
"	.byte	0, 0				\n"
"						\n"
"recorder_popss:				\n"
"#	mov	%cs,%bx				\n"
"#	add	$deltaseg,%bx			\n"
"#	mov	%bx,%ds				\n"
"#	incw	recorder+xx			\n"
"	pushl	%ebp				\n"
"	movw	%sp,%bp				\n"
"	movw	18(%bp),%bx			\n"
"	movw	%bx,%cs:recorder_dummyss	\n"
"	incw	12(%bp)				\n"
"	movl	14(%bp),%ebx			\n"
"	movl	%ebx,16(%bp)			\n"
"	movl	10(%bp),%ebx			\n"
"	movl	%ebx,12(%bp)			\n"
"	movl	6(%bp),%ebx			\n"
"	movl	%ebx,8(%bp)			\n"
"	movw	4(%bp),%bx			\n"
"	movw	%bx,6(%bp)			\n"
"	popl	%ebp				\n"
"	add	$2,%sp				\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	jmp	recorder_handler		\n"
"						\n"
"recorder_pushss:				\n"
"#	mov	%cs,%bx				\n"
"#	add	$deltaseg,%bx			\n"
"#	mov	%bx,%ds				\n"
"#	incw	recorder+xx			\n"
"	pushw	%bx	# reserve space		\n"
"	pushl	%ebp				\n"
"	movw	%sp,%bp				\n"
"	incw	14(%bp)				\n"
"	movl	6(%bp),%ebx			\n"
"	movl	%ebx,4(%bp)			\n"
"	movl	10(%bp),%ebx			\n"
"	movl	%ebx,8(%bp)			\n"
"	movl	14(%bp),%ebx			\n"
"	movl	%ebx,12(%bp)			\n"
"	movw	18(%bp),%bx			\n"
"	movw	%bx,16(%bp)			\n"
"	movw	%cs:recorder_dummyss,%bx	\n"
"	movw	%bx,18(%bp)			\n"
"	popl	%ebp				\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	jmp	recorder_handler		\n"
"						\n"
"recorder_movtoss:	# mov (addr),%ss not handled and crashes					\n"
"#	pushw	%ds				\n"
"#	mov	%cs,%bx				\n"
"#	add	$deltaseg,%bx			\n"
"#	mov	%bx,%ds				\n"
"#	incw	recorder+xx			\n"
"#	popw	%ds				\n"
"	movw    %sp,%bx				\n"
"	addw	$2,%ss:8(%bx)			\n"
"	movw    %ss:8(%bx),%bx			\n"
"	dec	%bx				\n"
"	movb    (%bx),%bl			\n"
"	andb	$0x07,%bl			\n"
"	cmpb    $0,%bl				\n"
"	jne     1f				\n"
"	movw	%ax,%cs:recorder_dummyss	\n"
"	1:					\n"
"	cmpb    $1,%bl				\n"
"	jne     1f				\n"
"	movw	%cx,%cs:recorder_dummyss	\n"
"	1:					\n"
"	cmpb    $2,%bl				\n"
"	jne     1f				\n"
"	movw	%dx,%cs:recorder_dummyss	\n"
"	1:					\n"
"	cmpb    $4,%bl				\n"
"	jne     1f				\n"
"	movw	%sp,%cs:recorder_dummyss	\n"
"	1:					\n"
"	cmpb    $5,%bl				\n"
"	jne     1f				\n"
"	movw	%bp,%cs:recorder_dummyss	\n"
"	1:					\n"
"	cmpb    $6,%bl				\n"
"	jne     1f				\n"
"	movw	%si,%cs:recorder_dummyss	\n"
"	1:					\n"
"	cmpb    $7,%bl				\n"
"	jne     1f				\n"
"	movw	%di,%cs:recorder_dummyss	\n"
"	1:					\n"
"	cmpb    $3,%bl				\n"
"	jne     1f				\n"
"	popl	%ebx				\n"
"	movw	%bx,%cs:recorder_dummyss	\n"
"	popl	%ds				\n"
"	jmp	recorder_handler		\n"
"	1:					\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	jmp	recorder_handler		\n"
"						\n"
"recorder_movfromss:				\n"
"#	pushw	%ds				\n"
"#	mov	%cs,%bx				\n"
"#	add	$deltaseg,%bx			\n"
"#	mov	%bx,%ds				\n"
"#	incw	recorder+xx			\n"
"#	popw	%ds				\n"
"	movw    %sp,%bx				\n"
"	addw	$2,%ss:8(%bx)			\n"
"	movw    %ss:8(%bx),%bx			\n"
"	dec	%bx				\n"
"	movb    (%bx),%bl			\n"
"	andb	$0x07,%bl			\n"
"	cmpb    $0,%bl				\n"
"	jne     1f				\n"
"	movw	%cs:recorder_dummyss,%ax	\n"
"	1:					\n"
"	cmpb    $1,%bl				\n"
"	jne     1f				\n"
"	movw	%cs:recorder_dummyss,%cx	\n"
"	1:					\n"
"	cmpb    $2,%bl				\n"
"	jne     1f				\n"
"	movw	%cs:recorder_dummyss,%dx	\n"
"	1:					\n"
"	cmpb    $4,%bl				\n"
"	jne     1f				\n"
"	movw	%cs:recorder_dummyss,%sp	\n"
"	1:					\n"
"	cmpb    $5,%bl				\n"
"	jne     1f				\n"
"	movw	%cs:recorder_dummyss,%bp	\n"
"	1:					\n"
"	cmpb    $6,%bl				\n"
"	jne     1f				\n"
"	movw	%cs:recorder_dummyss,%si	\n"
"	1:					\n"
"	cmpb    $7,%bl				\n"
"	jne     1f				\n"
"	movw	%cs:recorder_dummyss,%di	\n"
"	1:					\n"
"	cmpb    $3,%bl				\n"
"	jne     1f				\n"
"	popl	%ebx				\n"
"	movw	%cs:recorder_dummyss,%bx	\n"
"	popl	%ds				\n"
"	jmp	recorder_handler		\n"
"	1:					\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	jmp	recorder_handler		\n"
"						\n"
"recorder_pushf:				\n"
"	mov	%cs,%bx				\n"
"	add	$deltaseg,%bx			\n"
"	mov	%bx,%ds				\n"
"	incw	recorder+6			\n"
"	pushw	%bx	# reserve space		\n"
"	pushl	%ebp				\n"
"	movw	%sp,%bp				\n"
"	movl	6(%bp),%ebx			\n"
"	movl	%ebx,4(%bp)			\n"
"	movl	10(%bp),%ebx			\n"
"	movl	%ebx,8(%bp)			\n"
"	movl	14(%bp),%ebx			\n"
"	movl	%ebx,12(%bp)			\n"
"	movw	18(%bp),%bx			\n"
"	movw	%bx,16(%bp)			\n"
"	andw	$~0x100,18(%bp)			\n"
"	incw	12(%bp)				\n"
"	popl	%ebp				\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	jmp	recorder_handler		\n"
"						\n"
"recorder_popf_iret:				\n"
"	pushl	%ebp				\n"
"	movw	%sp,%bp				\n"
"	addw	$18,%bp				\n"
"	movw	%sp,%bx				\n"
"	movw	%ss:12(%bx),%bx			\n"
"	cmpb	$0xCF,(%bx)			\n"
"	jne	recorder_popf			\n"
"recorder_iret:					\n"
"	mov	%cs,%bx				\n"
"	add	$deltaseg,%bx			\n"
"	mov	%bx,%ds				\n"
"	decw	recorder+8			\n"
"	incw    recorder+10			\n"
"	movw	2(%bp),%bx			\n"
"	movw	%bx,%ds				\n"
"	movw	(%bp),%bx			\n"
"	decw	%bx				\n"
"	addw	$4,%bp				\n"
"recorder_popf:					\n"
"	pushw	%ds				\n"
"	pushw	%bx				\n"
"	mov     %cs,%bx				\n"
"	add     $deltaseg,%bx			\n"
"	mov     %bx,%ds				\n"
"	incw    recorder+8			\n"
"	popw	%bx				\n"
"	popw	%ds				\n"
"	orw	$0x100,(%bp)			\n"
"	addw	$2,%bp				\n"
"	incw	%bx				\n"
"	cmpb	$0x9D,(%bx)			\n"
"	je	recorder_popf			\n"
"	cmpb	$0xCF,(%bx)			\n"
"	je	recorder_iret			\n"
"	popl	%ebp				\n"
"	jmp	recorder_end			\n"
"						\n"
"recorder_int:					\n"
"	pushw	%bx				\n"
"	mov     %cs,%bx				\n"
"	add     $deltaseg,%bx			\n"
"	mov     %bx,%ds				\n"
"	incw    recorder+4			\n"
"	popw	%bx				\n"
"	shrw	$8,%bx				\n"
"	shlw	$2,%bx				\n"
"	pushw	$0				\n"
"	popw	%ds				\n"
"	movl	%ds:(%bx),%ebx			\n"
"	pushl	%ebx	# reserve space		\n"
"	pushl	%ebp				\n"
"	mov	%sp,%bp				\n"
"	addw	$2,16(%bp)			\n"
"	andw	$~0x300,20(%bp)	# clear flags TF, IF [, AC]	\n"
"	xchgl	%ebx,12(%bp)			\n"
"	xchgl	%ebx,8(%bp)			\n"
"	xchgl	%ebx,4(%bp)			\n"
"	pushf					\n"
"	popw	%bx				\n"
"	orb	$1,%bh				\n"
"	pushw	%bx				\n"
"	popfw					\n"
"	popl	%ebp				\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	lretw					\n"
"						\n"
"	.align 4				\n"
"recorder_savedidt:				\n"
"	.zero 6					\n"
"						\n"
"	.align 4				\n"
"recorder_dummyidt:				\n"
"	.zero 6					\n"
"						\n"
"						\n"
"recorder_idt_gdt:				\n"
"	movw	%sp,%bx				\n"
"	movw	%ss:8(%bx),%bx			\n"
"	movb	2(%bx),%bl			\n"
"	andb	$0x28,%bl			\n"
"	cmpb	$0x08,%bl			\n"
"	jne	recorder_notlsidt		\n"
"	pushw	%ds				\n"
"	mov     %cs,%bx				\n"
"	add     $deltaseg,%bx			\n"
"	mov     %bx,%ds				\n"
"	incw    recorder+14			\n"
"	popw	%ds				\n"
"	movw	%sp,%bx				\n"
"	pushl	%eax				\n"
"	movw	%ss:8(%bx),%ax			\n"
"	addw	$3,%ss:8(%bx)			\n"
"	movw	%ax,%bx				\n"
"	movb	2(%bx),%al			\n"
"	movb	%al,%cs:recorder_patch		\n"
"	movb	$0x90,%ah			\n"
"	andb	$0xC6,%al			\n"
"	cmpb	$0x06,%al			\n"
"	je	1f				\n"
"	andb	$0xC0,%al			\n"
"	jz	2f				\n"
"	1:					\n"
"	movb	3(%bx),%ah			\n"
"	pushl   %ebx				\n"
"	movw    %sp,%bx				\n"
"	incw    %ss:16(%bx)			\n"
"	popl    %ebx				\n"
"	2:					\n"
"	movb	%ah,%cs:recorder_patch+1	\n"
"	movb	$0x90,%ah			\n"
"	cmpb	$0x80,%al			\n"
"	je	3f				\n"
"	movb	4(%bx),%ah			\n"
"	pushl   %ebx				\n"
"	movw    %sp,%bx				\n"
"	incw    %ss:16(%bx)			\n"
"	popl    %ebx				\n"
"	3:					\n"
"	movb	%ah,%cs:recorder_patch+2	\n"
"	popl	%eax				\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	sidtw	recorder_savedidt		\n"
"	lidtw	recorder_dummyidt		\n"
"	.byte	0x0F				\n"
"	.byte	0x01				\n"
"recorder_patch:				\n"
"	nop					\n"
"	nop					\n"
"	nop					\n"
"	sidtw	recorder_dummyidt		\n"
"	lidtw	recorder_savedidt		\n"
"	jmp	recorder_handler		\n"
"						\n"
"recorder_notlsidt:				\n"
"	movw	%sp,%bx				\n"
"	movw	%ss:8(%bx),%bx			\n"
"	movb	2(%bx),%bl			\n"
"#	andb	$0x38,%bl	# lmsw (addr) not recognised	\n"
"#	cmpb	$0x30,%bl			\n"
"	andb	$0xF8,%bl			\n"
"	cmpb	$0xF0,%bl			\n"
"	jne	recorder_notlmsw		\n"
"	pushw	%ds				\n"
"	mov	%cs,%bx				\n"
"	add	$deltaseg,%bx			\n"
"	mov	%bx,%ds				\n"
"	incw	recorder+16			\n"
"	popw	%ds				\n"
"	movw	%sp,%bx				\n"
"	movw	%ss:8(%bx),%bx			\n"
"	movb	2(%bx),%bl			\n"
"	andb	$0x07,%bl			\n"
"	movb	%bl,%cs:recorder_markercr0	\n"
"	movw	%sp,%bx				\n"
"	addw	$3,%ss:8(%bx)			\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	cmpb	$0,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%cs:recorder_dummycr0,%ax	\n"
"	1:					\n"
"	cmpb	$1,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%cs:recorder_dummycr0,%cx	\n"
"	1:					\n"
"	cmpb	$2,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%cs:recorder_dummycr0,%dx	\n"
"	1:					\n"
"	cmpb	$3,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%cs:recorder_dummycr0,%bx	\n"
"	1:					\n"
"	cmpb	$4,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	mov	%cs:recorder_dummycr0,%sp	\n"
"	1:					\n"
"	cmpb	$5,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%cs:recorder_dummycr0,%bp	\n"
"	1:					\n"
"	cmpb	$6,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%cs:recorder_dummycr0,%si	\n"
"	1:					\n"
"	cmpb	$7,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%cs:recorder_dummycr0,%di	\n"
"	1:					\n"
"	jmp	recorder_handler		\n"
"recorder_notlmsw:				\n"
"#	cmpb	$0x20,%bl	# smsw (addr) not recognised	\n"
"	cmpb	$0xE0,%bl			\n"
"	jne	recorder_notsmsw		\n"
"	pushw	%ds				\n"
"	mov     %cs,%bx				\n"
"	add     $deltaseg,%bx			\n"
"	mov     %bx,%ds				\n"
"	incw    recorder+16			\n"
"	popw	%ds				\n"
"	movw	%sp,%bx				\n"
"	movw	%ss:8(%bx),%bx			\n"
"	movb	2(%bx),%bl			\n"
"	andb	$0x07,%bl			\n"
"	movb	%bl,%cs:recorder_markercr0	\n"
"	movw	%sp,%bx				\n"
"	addw	$3,%ss:8(%bx)			\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	cmpb	$0,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%ax,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$1,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%cx,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$2,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%dx,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$3,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%bx,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$4,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%sp,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$5,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%bp,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$6,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%si,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$7,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movw	%di,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	jmp	recorder_handler		\n"
"recorder_notsmsw:				\n"
"	jmp	recorder_end			\n"
"						\n"
"						\n"
"recorder_dummycr0:				\n"
"	.zero 4					\n"
"recorder_markercr0:				\n"
"	.byte 0					\n"
"						\n"
"recorder_set_cr:				\n"
"	movw	%sp,%bx				\n"
"	movw	%ss:8(%bx),%bx			\n"
"	movb	2(%bx),%bl			\n"
"	testb	$0x80,%bl			\n"
"	jz	recorder_not_cr0		\n"
"	testb	$0x40,%bl			\n"
"	jz	recorder_not_cr0		\n"
"	testb	$0x28,%bl			\n"
"	jnz	recorder_not_cr0		\n"
"	andb	$0x07,%bl			\n"
"	movb	%bl,%cs:recorder_markercr0	\n"
"	mov     %cs,%bx				\n"
"	add     $deltaseg,%bx			\n"
"	mov     %bx,%ds				\n"
"	incw    recorder+18			\n"
"	movw	%sp,%bx				\n"
"	addw	$3,%ss:8(%bx)			\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	cmpb	$0,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%cs:recorder_dummycr0,%eax	\n"
"	1:					\n"
"	cmpb	$1,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%cs:recorder_dummycr0,%ecx	\n"
"	1:					\n"
"	cmpb	$2,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%cs:recorder_dummycr0,%edx	\n"
"	1:					\n"
"	cmpb	$3,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%cs:recorder_dummycr0,%ebx	\n"
"	1:					\n"
"	cmpb	$4,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%cs:recorder_dummycr0,%esp	\n"
"	1:					\n"
"	cmpb	$5,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%cs:recorder_dummycr0,%ebp	\n"
"	1:					\n"
"	cmpb	$6,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%cs:recorder_dummycr0,%esi	\n"
"	1:					\n"
"	cmpb	$7,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%cs:recorder_dummycr0,%edi	\n"
"	1:					\n"
"	jmp	recorder_handler		\n"
"						\n"
"recorder_not_cr0:				\n"
"	jmp	recorder_end			\n"
"						\n"
"recorder_get_cr:				\n"
"	movw	%sp,%bx				\n"
"	movw	%ss:8(%bx),%bx			\n"
"	movb	2(%bx),%bl			\n"
"	testb	$0x80,%bl			\n"
"	jz	recorder_not_cr0		\n"
"	testb	$0x40,%bl			\n"
"	jz	recorder_not_cr0		\n"
"	testb	$0x28,%bl			\n"
"	jnz	recorder_not_cr0		\n"
"	andb	$0x07,%bl			\n"
"	movb	%bl,%cs:recorder_markercr0	\n"
"	mov     %cs,%bx				\n"
"	add     $deltaseg,%bx			\n"
"	mov     %bx,%ds				\n"
"	incw    recorder+18			\n"
"	movw	%sp,%bx				\n"
"	addw	$3,%ss:8(%bx)			\n"
"	popl	%ebx				\n"
"	popl	%ds				\n"
"	cmpb	$0,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%eax,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$1,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%ecx,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$2,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%edx,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$3,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%ebx,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$4,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%esp,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$5,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%ebp,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$6,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%esi,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	cmpb	$7,%cs:recorder_markercr0	\n"
"	jne	1f				\n"
"	movl	%edi,%cs:recorder_dummycr0	\n"
"	1:					\n"
"	jmp	recorder_handler		\n"
"	"
#endif
	);

static void recorder_test (unsigned param)
  {
  extern unsigned char recorder_handler[];
  unsigned i;
  struct interrupt_str {
      unsigned inter[256];
      } interrupt_table;
  gdt_idt_param_type idt_param, save_idt;
  struct recorder_array_str recorder_array[22] = {};

  idt_param.base = farptr2linear (stack_adr (&interrupt_table));
  idt_param.limit = sizeof (interrupt_table) -1;

  recorder = ((struct recorder_str) {});
  recorder.array = recorder_array;
  recorder.arraysize = nbof (recorder_array);

  for (i = 0; i < nbof (interrupt_table.inter); i++)
      interrupt_table.inter[i] = peekl (i * 4);
  interrupt_table.inter[1] = code_adr (recorder_handler);

  get_idt (&save_idt);
  set_idt (&idt_param);

#ifdef EXTENDED_RECORDER
  {
  extern gdt_idt_param_type recorder_dummyidt asm ("recorder_dummyidt");
  extern CR0_type recorder_dummycr0 asm ("recorder_dummycr0");
  extern unsigned short recorder_dummyss asm ("recorder_dummyss");

  recorder_dummyidt = save_idt;
  recorder_dummycr0 = get_CR0();
  recorder_dummyss = getss();

//  recorder_dummycr0.protected_mode = 1; /* lets say that... */
  }
#endif

  setTFflag();
  /* tests here */
#if USER_SUPPORT & VESA_WINFUNCTION
  if (UI.parameter.addrpositionfarfct) {
      unsigned status, dummydx;
      asm volatile (
	" lcallw *%a4 # _VESA_setwin"
	: "=a" (status), "=d" (dummydx)
	: "b" (UI.parameter.winNo & 0x00FF), /* %bh = 0 */
	  "d" (param),
	  "pr" (&UI.parameter.addrpositionfarfct),
	  "a" (0x4F05)
	);
      }
    else
#endif /* VESA_WINFUNCTION */
      {
      switch (param) {
	case 0: {	/* that seems to work... */
#ifdef EXTENDED_RECORDER
	  static gdt_idt_param_type idt_tmp;
	  asm volatile (" sidtw %0 " : : "m" (idt_tmp) : "memory");
	  asm volatile (" pushfw ; popfw ");
	  asm volatile (" mov %%cr0,%%eax " : : : "eax");
	  asm volatile (" lock nop ");
	  asm volatile (" lmsw %%ax" : : : "eax");
	  asm volatile (" pushw %ss ; popw %ss ");
	  asm volatile (" pushw %ax ; mov %ss,%ax ; mov %ax,%ss ; popw %ax ");
	  asm volatile (
"		pushfw		\n"
"		pushw   %cs	\n"
"		pushw   $next	\n"
"		iret		\n"
"		next:		\n"
		);
#endif /* EXTENDED_RECORDER */
	  }
	  break;
	case 1:
	  asm volatile ( " int $0x10 # writechar+attr "
		: /* no output */
		: "a" (0x0900 | '\333'), "c" (1),
		  "b" (((unsigned short)0 << 8) | 7)
		);
	  break;
	case 2:
	  asm volatile ( " int $0x10 # writechar+color "
		: /* no output */
		: "a" (0x0A00 | '\333'), "c" (1),
		  "b" (((unsigned short)0 << 8) | 7)
		);
	  break;
	case 3: {
	  unsigned char page = 0, attr = 7, row = 4*param,
			col = 0, update_cursor = 0;
	  asm volatile (
"	      xchgl   %%ebp,%4			\n"
"	      int     $0x10 # writestring	\n"
"	      xchgl   %%ebp,%4			\n"
	      : /* no output */
	      : "c" (1),
		"b" (((unsigned short)page << 8) | attr),
		"d" (((unsigned short)row << 8) | col),
		"a" (0x1300 | !!update_cursor),
		"r" ("\333")  /* in fact %es:%bp */
	      );
	  }
	  break;
	}
      }

  /* end test */
  clearTFflag();
  asm volatile ("nop");
  set_idt (&save_idt);
  if (   param != 0
#if USER_SUPPORT & VESA_WINFUNCTION
      || UI.parameter.addrpositionfarfct
#endif /* VESA_WINFUNCTION */
      )
      UI_setmode (UI.parameter.identification); /* reset it ! */

  printf ("recorder(param: %u): %u IO ports written (limit %u)"
	  " on %u instructions\r\n"
#ifdef EXTENDED_RECORDER
	  "%u soft interrupt, %u [ls]idt, %u [ls]msw, %u cr0, %u pushf, "
	  "%u popf, %u iret, %u lock.\r\n"
#endif
	  , param, recorder.nbwrite, recorder.arraysize, recorder.nbinstruction
#ifdef EXTENDED_RECORDER
	  , recorder.nbint, recorder.nbidt, recorder.nblmsw, recorder.nbcr0,
	   recorder.nbpushf, recorder.nbpopf, recorder.nbiret, recorder.nblock
#endif
	   );

  for (i = 0; i < recorder.nbwrite && i < recorder.arraysize; i++) {
      const char *mnemonic = "???";

      if (recorder.array[i].instruction == 0xEF66) {
	  mnemonic = "outl eax=";
	  }
	else {
	  unsigned char inst = recorder.array[i].instruction & 0xFF;
	  recorder.array[i].value &= 0xFFFF;

	  if ((inst & 0x7F) == 0x6E) {
	      recorder.array[i].value &= 0xFF;
	      if (inst == 0xEE)
		  mnemonic = "outb al=";
		else if (inst == 0x6E)
		  mnemonic = "outsb al=";
	      }
	    else if (inst == 0xEF)
	      mnemonic = "outw ax=";
	    else if (inst == 0x6F)
	      mnemonic = "outsw ax=";
	  }

      printf (" %s=0x%X, dx=0x%X\r\n",
		mnemonic,
		recorder.array[i].value,
		recorder.array[i].addr);
      }
  printf ("end.\r\n\r\n");
  }

/********************************************************
 * The real test software, independant of graphic and disk...
 */
static unsigned gethexa (void)
  {
  unsigned number = 0;
  unsigned char str[2] = {0, 0}, zero_displayed = 0;

  for (;;) {
      unsigned char got = UI_getkey(0) & 0x00FF;

      if (got == 'q' || got == 'Q')
	  got = 'A';	/* french keyboard */

      if (got >= 'a' && got <= 'f')
	  got = got - 'a' + 'A';

      /* a switch takes too much space... */
      if (got == '0' && number == 0) {
	  if (zero_displayed)
	      continue;
	  zero_displayed = 1;
	  str[0] = got;
	  printf (str);
	  }
	else if (got >= '0' && got <= '9') {
	  number = (number << 4) | (got - '0');
	  str[0] = got;
	  printf (str);
	  }
	else if (got >= 'A' && got <= 'F') {
	  number = (number << 4) | (got - 'A' + 10);
	  str[0] = got;
	  printf (str);
	  }
	else if (got == '\010' || got == '\177') {	/* Delete for VTxx0 */
	  if (number == 0) {
	      if (zero_displayed)
	  zero_displayed = 0;
	else
	  continue;
	      }
	  number >>= 4;
	  printf ("\010 \010");
	  }
	else if (got == '\r' || got == '\n')
	  return number;
	else if (got == '\033')
	  return -1;
	else
	  printf ("\007");
      }
  }

static void dump_disk (unsigned cpt)
  {
  unsigned i;
  struct diskparam_str *dps = &DI.param[cpt];
  char disk_name[SIZE_DISKNAME];

  diskname (dps, disk_name);
  printf ("\r%s", disk_name);

  if (dps->ebiosversion != 0) {
      printf (" [EBIOS V0x%X", dps->ebiosversion);
      if (dps->ideIOadr != 0 && dps->irq != 0)
	  printf (" I/O: 0x%X, IRQ: %u, %s] ", dps->ideIOadr, dps->irq,
		    dps->ismaster ? "master" : "slave");
	else
	  printf ("] ");
      }
  if (dps->access == dos_part)
      printf (", SectorPerCluster: %u, ClustersOnDisk: %u, "
	      "MediaDescriptorByte: 0x%X",
		dps->nbsectorpertrack, dps->nbcylinder, dps->nbhead);
    else
      printf (", C/H/S: %u/%u/%u",
		dps->nbcylinder, dps->nbhead, dps->nbsectorpertrack);
  printf (", total sector: %u, sector size: %u",
		(unsigned)dps->nbtotalsector, /* long long */
		dps->bytepersector);

  if (!dps->bootable) {
      printf (", Disk not bootable.\r\n");
      return;
      }

  printf (", bootable, %u partitions:", dps->nbpartition);
  for (i = 0; i < dps->nbpartition; i++) {
      printf (" %u/OS%u", dps->partition[i].misc.order,
		dps->partition[i].OSnumber);
      if (   dps->partition[i].start == 0
	  && dps->partition[i].length == 0
	  && dps->partition[i].type == EMPTY) {
	  printf (": [empty]");
	  continue;
	  }
      if (dps->partition[i].active == part_active)
	  printf (": [active, ");
	else if (dps->partition[i].active == part_inactive)
	  printf (": [inactive, ");
	else {
	  printf (" : [invalid]");
	  continue;
	  }
      printf ("%u/%u, %s]",
		dps->partition[i].start,
		dps->partition[i].length,
		PartName (dps->partition[i].type));
      }
  printf ("\r\n");

#if defined (NB_FREELIST) && (NB_FREELIST != 0)
  if (dps->nbfreelist != 0) {
      printf ("Free list: ");
      for (i = 0; i < dps->nbfreelist; i++)
	  printf (" [%u sectors at %u] ",
			dps->freelist[i].length, dps->freelist[i].start);
      printf ("\r\n");
      }
#endif /* NB_FREELIST */
  }

static unsigned tester_probe_mode (unsigned first, unsigned last)
  {
  unsigned cpt, returned = 0;
  char buffer[(UI.parameter.nbrow?: 25) * 60];
  char *ptrbuf, *end = buffer + sizeof(buffer);
  struct video_parameter_str param;
#if USER_SUPPORT & VGA_SUPPORT
  unsigned lastvalid = first - 1;
#endif

  DBG (("\r\n\r\n"));
  ptrbuf = buffer + sprintf (buffer, "%$\r\nValid video modes:\r\n", end);
  for (cpt = first; cpt <= last; cpt++) {
#if USER_SUPPORT & VGA_SUPPORT
      if (cpt < 0x100) {
	  if (!VIDEO_mode_is_valid (cpt))
	      continue;
	  if (lastvalid != cpt - 1)
	      DBG (("mode 0x%X to 0x%X refused from validmode bitfield\r\n",
			lastvalid + 1, cpt - 1));
	  lastvalid = cpt;
	  }
#endif

      if (UI_getsize (cpt, &param) != 0)
	  continue;
      returned ++;
      ptrbuf += sprintf (ptrbuf, "%$ mode 0x%X: text %ux%u, graphic "
				 "%u*%ux%u, char %ux%u\r\n", end,
			param.identification, param.nbcol, param.nbrow,
			param.nbcolor, param.width, param.heigh,
			param.charwidth, param.charheigh
			);
      if (ptrbuf - buffer > (signed) sizeof(buffer) - 80) {
	  ptrbuf += sprintf (ptrbuf, "%$ [incomplete] ", end);
	  DBG ((" [Abort, string will overflow] "));
	  break;
	  }
      }
#if USER_SUPPORT & VGA_SUPPORT
  if (lastvalid != cpt - 1)
      DBG (("mode 0x%X to 0x%X refused from validmode bitfield\r\n",
		lastvalid + 1, cpt - 1));
#endif
  UI_putstr (buffer);
  return returned; /* number of valid modes */
  }

#if USER_SUPPORT & (BIOS_MOUSE_SUPPORT | SERIAL_MOUSE_SUPPORT)
static void mouse_error (unsigned char status)
  {
  if (status & 0x80)
      UI_putstr ("carry error, ");
  status &= ~0x80;
  if (status == PS_successfull)
      UI_putstr ("\r\n");
    else if (status == PS_invalid_fct)
      UI_putstr ("invalid function\r\n");
    else if (status == PS_invalid_input)
      UI_putstr ("invalid input\r\n");
    else if (status == PS_interface_error)
      UI_putstr ("interface error \r\n");
    else if (status == PS_need_resend)
      UI_putstr ("need resend\r\n");
    else if (status == PS_no_device_handler)
      UI_putstr ("no device handler\r\n");
    else
      UI_putstr ("unknown\r\n");
  }

asm (
"									\n"
"	.GLOBAL mouse_handler, mouse_handler_relay			\n"
"mouse_handler_relay:							\n"
"	pushw	%gs							\n"
"	pushw	%fs							\n"
"	pushw	%ds							\n"
"	pushw	%es							\n"
"	pushl	%edx							\n"
"	pushl	%ecx							\n"
"	pushl	%eax							\n"
"									\n"
"	movl	24(%esp),%ecx						\n"
"	movl	28(%esp),%edx						\n"
"									\n"
"	movw	%cs,%ax							\n"
"	addw	$deltaseg,%ax	# not necessary if no CODE_SEGMENT	\n"
"	movw	%ax,%ds							\n"
"	movw	%ax,%es							\n"
"									\n"
"	# If a DOS bootloader is built, we have to switch stack here	\n"
"	# because:							\n"
"	# - we may overflow the local DOS stack				\n"
"	# - we assume %ss == %ds,					\n"
"	#   the compiler will generate mov (%ebp),%ax			\n"
"	mousestacksize = 200    # (50-1) 32 bit words			\n"
"	.comm   mouse_stack, mousestacksize, 4				\n"
"									\n"
"	cli			# our temporary stack is small...	\n"
"									\n"
"	movw	%ss,mouse_stack+2					\n"
"	movw	%sp,mouse_stack						\n"
"	movw	%ax,%ss							\n"
"	movw	$mouse_stack+mousestacksize,%sp				\n"
"	nop								\n"
"									\n"
"	pushl	%edx							\n"
"	pushl	%ecx							\n"
"	calll	mouse_handler						\n"
"									\n"
"	lss	mouse_stack,%sp						\n"
"									\n"
"	sti			# our temporary stack is small...	\n"
"									\n"
"	popl	%eax							\n"
"	popl	%ecx							\n"
"	popl	%edx							\n"
"	popw	%es							\n"
"	popw	%ds							\n"
"	popw	%fs							\n"
"	popw	%gs							\n"
"	lretw								\n"
"									\n"
	);

static void mousetest (void)
  {
  struct PS_resetID mouseconf;
  struct PS_status psstat;
  unsigned char status, device, resol, sample;
  extern char mouse_handler_relay[];

  if ((status = _PS_reset(&mouseconf)) != 0) {
      mouse_error (status);
      return;
      }
  printf ("mouse device ID: 0x%X, reset value: 0x%X",
	mouseconf.device_id, mouseconf.reset_value);

  if ((status = _PS_initialise(3)) != 0) {
      mouse_error (status);
      return;
      }
  printf (", initialised (packet 3)\r\n");

  if ((status = _PS_get_type(&device)) != 0) {
      mouse_error (status);
      return;
      }
  printf ("device: 0x%X, ", device);

  if ((status = _PS_device_status(&psstat, &resol, &sample)) != 0) {
      mouse_error (status);
      return;
      }
  printf ("resolution: %u, sample rate: %u, "
	  "right button: %s, left button: %s, %s scaling\r\n",
	  resol, sample,
	  psstat.right_button_pressed? "down" : "up",
	  psstat.left_button_pressed? "down" : "up",
	  psstat.twoTOone_scaling? "2:1" : "1:1"
	  );

#if 0
  printf ("set 2:1 scaling, ");
  if ((status = _PS_set2to1scaling(1)) != 0) {
      mouse_error (status);
      return;
      }
#endif

  printf ("disabling");
  if ((status = _PS_enable(0)) != 0) {
      mouse_error (status);
      return;
      }
  printf (", set driver");
  if ((status = _PS_set_handler(code_adr(mouse_handler_relay))) != 0) {
      mouse_error (status);
      return;
      }
  printf (", enabling");
  if ((status = _PS_enable(1)) != 0) {
      mouse_error (status);
      return;
      }
  {
  unsigned char sampling = 5, resol = 1, key;
  for (;;) {
      printf ("\r\nPress '+', '-', '*', '/' or ESC or move mouse...\r\n\r\n");
      key = UI_getkey(0);
      if (key == '\033')
	  break;
      if (key == '+' || key == '-') {
	  if (key == '-' && sampling > 0)
	      sampling--;
	  if (key == '+' && sampling < 6)
	      sampling++;
	  _PS_enable(0);
	  if ((status = _PS_set_sampling(sampling)) != 0) {
	      mouse_error (status);
	      break;
	      }
	  _PS_enable(1);
	  printf ("\nsampling set to %u\r\n\r\n", sampling);
	  }
      if (key == '*' || key == '/') {
	  if (key == '/' && resol > 0)
	      resol--;
	  if (key == '*' && resol < 3)
	      resol++;
	  _PS_enable(0);
	  if ((status = _PS_set_resolution(resol)) != 0) {
	      mouse_error (status);
	      break;
	      }
	  _PS_enable(1);
	  printf ("\nresolution set to %u\r\n\r\n", resol);
	  }
      }
  }
  printf ("\r\ndisabling mouse driver ");
  if ((status = _PS_enable(0)) != 0) {
      mouse_error (status);
      return;
      }
  printf ("OK.\r\n");
  }

/*
 * This function is called with an alternate
 * stack, it cannot use printf() because of the
 * check_stack() when (DEBUG & DEBUG_STACK)
 */
__attribute__ ((cdecl, regparm (0)))
void mouse_handler (unsigned msbword3_lsbword4,
	    unsigned msbword1_lsbword2)
  {
  unsigned short word1 = msbword1_lsbword2 >> 16;
  unsigned short word2 = msbword1_lsbword2 & 0xFFFF;
  unsigned short word3 = msbword3_lsbword4 >> 16;

  struct mouse_status_str {
      unsigned short left_button_pressed        : 1;
      unsigned short right_button_pressed       : 1;
      unsigned short reserved1                  : 1;
      unsigned short reserved2                  : 1;
      unsigned short x_negative                 : 1;
      unsigned short y_negative                 : 1;
      unsigned short x_overflow                 : 1;
      unsigned short y_overflow                 : 1;
      } __attribute__ ((packed))
	 mouse_status = *(struct mouse_status_str *)&word1;
  signed char deltax = word2 & 0xFF;
  signed char deltay = word3 & 0xFF;

  static unsigned short mouseXpos, mouseYpos;
  unsigned short word4 = msbword3_lsbword4 & 0xFFFF;
  void puthex (unsigned short hex) {
      char buff[5];

      buff[4] = 0;
      buff[3] = "0123456789ABCDEF"[hex % 16]; hex /= 16;
      buff[2] = "0123456789ABCDEF"[hex % 16]; hex /= 16;
      buff[1] = "0123456789ABCDEF"[hex % 16]; hex /= 16;
      buff[0] = "0123456789ABCDEF"[hex % 16];
      UI_putstr (buff);
      }

  UI_putstr ("  word1: "); puthex (word1);
  UI_putstr ("  word2: "); puthex (word2);
  UI_putstr ("  word3: "); puthex (word3);
  UI_putstr ("  word4: "); puthex (word4);

  if (mouse_status.x_overflow || mouse_status.y_overflow) {
      UI_putstr (" overflow !\r\n");
      return;
      }
  if (!mouse_status.reserved2 || mouse_status.reserved1) {
      UI_putstr (" invalid format !\r\n");
      return;
      }

  if (mouse_status.x_negative ^ (deltax < 0))
      deltax = -deltax;

  if (mouse_status.y_negative ^ (deltay < 0))
      deltay = -deltay;

  mouseXpos += deltax;
  mouseYpos += deltay;

#if DEBUG & DEBUG_STACK
  UI_putstr ("; ");
  puthex (mouseXpos);
  UI_putstr (", ");
  puthex (mouseYpos);
  UI_putstr (", ");
  UI_putstr (mouse_status.left_button_pressed? "down" : "up  ");
  UI_putstr (", ");
  UI_putstr (mouse_status.right_button_pressed? "down" : "up  ");
  UI_putstr ("         \r");
#else
  printf ("; %u, %u, %s, %s         \r",
	  mouseXpos,
	  mouseYpos,
	  mouse_status.left_button_pressed? "down" : "up  ",
	  mouse_status.right_button_pressed? "down" : "up  ");
#endif
  }
#endif

static unsigned menu (struct registers *regs)
  {
  unsigned start, stop, number, partition, currmode;
  unsigned char key;
  volatile int divide_per_0 = 0;

  for (;;) {
      if (!USER_ACTIVE())
	  UI_putstr ("\r"
		"You are still in the \"basic\" level, i.e. no knowlegde of screen\r\n"
		" size, no cursor movement, no probing...\r\n\r\n");

		/*0000000001111111111222222222233333333334*/
		/*1234567890123456789012345678901234567890*/
      UI_putstr (	"\r"
		"Select your choice by pressing one of:\r\n"
		"  0: display registers                \r\n"
		"  1: display disks found              \r\n"
		"  2: reboot from a drive/partition    \r\n"
		"  3: probe video modes                \r\n"
		"  4: clear screen                     \r\n"
		"  5: switch to VGA / SERIAL interface \r\n"
#if USER_SUPPORT & VESA_SUPPORT
		"  6: switch to VESA interface         \r\n"
#endif
		"  7: shift previous lines (graphic)   \r\n"
		"  8: Measures (ONLY pentium+, rdtsc)  \r\n"
		"  9: display few horizontal lines     \r\n"
		"  A..M: internal tests (will crash).  \r\n"
		"  -: display 'param' structure        \r\n"
		"  .: open/close segment doors (VESA2) \r\n"
		"  + or , : select one display mode    \r\n"
		"  / , *: restart with/without error   \r\n"
		"\r\n");
      key = UI_getkey(0) & 0x00FF;

      /* Those commands do not need a initialised interface: */
      switch (key) {
	case '0':
	  dump_registers (regs);
	  if ((getsw() & 1) == 0) { /* not in virtual 8086 mode ! */
	      printf ("doors: gs ");
	      printf (door_open_gs()? "open" : "closed");
	      printf (", fs ");
	      printf (door_open_fs()? "open" : "closed");
	      printf (", es ");
	      printf (door_open_es()? "open" : "closed");
	      printf (", ds ");
	      printf (door_open_ds()? "open" : "closed");
	      printf (", cs ");
	      printf (door_open_cs()? "open" : "closed");
	      printf (", ss ");
	      printf (door_open_ss()? "open" : "closed");
	      printf ("\r\n");
	      }
	  printf ("\r\nPress a key to continue...");
	  UI_getkey(0);
	  continue;
	case '1':
	  printf ("Found %u valid disks:\r\n\r\n", DI.nbdisk);
	  for (number = 0; number < DI.nbdisk; number ++) {
	      dump_disk (number);
	      printf ("\r\nPress a key to continue...");
	      UI_getkey(0);
	      }
	  continue;
	case '2':
	  list_disks();
	  do {
	      printf ("\r\nType the disk you want to boot from (0..%u),"
			" or %u to \"fail\" this boot: ",
			DI.nbdisk - 1, DI.nbdisk);
	      number = gethexa();
	      } while (number != (unsigned) -1 && number > DI.nbdisk);
	  if (number == (unsigned) -1)
	      continue;
	  if (number == DI.nbdisk) {
	      printf ("\r\nInvoking the possible \"failed\" boot handler...\r\n");
	      _BIOS_failedboot();
	      continue;
	      }
	  printf ("\r\n\r\nDisk description:\r\n");
	  dump_disk (number);
	  for (;;) {
	      unsigned i;

	      printf ("\r\nType the OS<number> you want to boot from, or 0 for MBR: ");
	      if ((partition = gethexa()) == (unsigned) -1)
		  break;
	      if (partition == 0) {
		  partition = DI.param[number].nbpartition;
		  break;
		  }
	      for (i = 0; i < DI.param[number].nbpartition; i++)
		  if (partition == DI.param[number].partition[i].OSnumber) {
		      partition = i;
		      break;
		      }
	      if (i != DI.param[number].nbpartition)
		  break;
	      printf (" NOT FOUND!\r\n");
	      }
	  if (partition == (unsigned) -1)
	      continue;
	  if (partition != DI.param[number].nbpartition) {
	      start = DI.param[number].partition[partition].start;
	      printf ("\r\n\r\nPartition type 0x%X\r\n\r\n",
			DI.param[number].partition[partition].type);
	      }
	    else {
	      start = 0;
	      printf ("\r\n\r\nBooting from MBR (sector 0)\r\n\r\n");
	      }
#if USER_SUPPORT & VGA_SUPPORT
	  if (BOOT1_COM_port() < 0)
	      VGA_init(); /* force mode 3 if current mode is VESA */
#endif
	  if (   DI.param[number].access == bios_chs
	      || DI.param[number].access == ebios_lba) {
	      printf ("setting edx to 0x%X.\r\n", DI.param[number].disknb);
	      regs->edx = DI.param[number].disknb; /* %dh bit 5 = 0 */
	      }
#ifdef TREAT_EXCEPTION
	  restore_vectors ();
#endif
	  printf ("reading one sector at lba=%d on disk 0x%X and jumping...\r\n\r\n",
			start, DI.param[number]);
	  return DISK_readsector (&DI.param[number], start, 1, 0x00007C00);
	case 'A': case 'a':
	  divide_per_0 = 100 / divide_per_0;
	  continue;
	case 'B': case 'b':
	  asm volatile (" ud2 ");	/* THE undefined instruction */
	  continue;
	case 'C': case 'c':
	  test_bound_stack ((void*)&test_bound_stack);
	  continue;
	case 'D': case 'd':
	  test_bound_data (0);	/* at least this address is protected */
	  continue;
	case 'E': case 'e': {
	  unsigned dummy = 0x10004;
	  asm volatile (" movl %%ds:(%%eax),%%eax " : "+a" (dummy));
	  }
	  continue;
	case 'F': case 'f': {
	  unsigned dummy = 0x10004;
	  asm volatile (" movl %%ss:(%%eax),%%eax " : "+a" (dummy));
	  }
	  continue;
	case 'L': case 'l':
	  test_intersegment_call ("extraseg works\r\n");
	  testjump ();
	  continue;
	case 'M': case 'm':
#if USER_SUPPORT & (BIOS_MOUSE_SUPPORT | SERIAL_MOUSE_SUPPORT)
	  mousetest ();
#else
	  printf ("mouse not compiled in!\r\n");
#endif
	  break;
	case 'T': case 't':
	  timertest();
	  break;
	case '.':
	  if (door_open_gs()) {
	      if (segment_door(0) != 0)
		  printf ("closing doors failed\r\n");
		else
		  printf ("closing doors OK\r\n");
	      }
	    else {
	      if (segment_door(1) != 0)
		  printf ("opening doors failed\r\n");
		else
		  printf ("opening doors OK\r\n");
	      }
	  continue;
	case '/':
#ifdef TREAT_EXCEPTION
	  restore_vectors ();
#endif
	  return 1;
	case '*':
#ifdef TREAT_EXCEPTION
	  restore_vectors ();
#endif
	  return 0;
	case '-':
	  display_param();
	  printf ("\r\nPress a key to continue...");
	  UI_getkey(0);
	  continue;
	case '5' :
	  if (BOOT1_COM_port() < 0) {
#if USER_SUPPORT & VGA_SUPPORT
	      unsigned mode = 0, tmp;
	      if (!VGA_ACTIVE()) {
		  if (VGA_init() != 0) {
		      printf ("no VGA card found\r\n");
		      continue;
		      }
		  mode = UI_getmode () & 0x7FFF;
		  }
		else
		  printf ("Already in VGA, active mode 0x%X\r\n",
			  UI_getmode () & 0x7FFF);
	      printf ("\r\nVGA card present, modes: ");
	      listmode (&UI);
	      printf ("switch to VGA mode (from previous list): 0x");
	      number = gethexa();
	      if (number != (unsigned) -1) {
		  if ((tmp = UI_setmode (number)) != 0)
		      printf ("\r\nError 0x%X switching to mode 0x%X\r\n\r\n",
				tmp, number);
		  }
		else {
		  if (mode != 0)
		      if ((tmp = UI_setmode (mode)) != 0)
		   printf ("\r\nError 0x%X initialising to mode 0x%X\r\n\r\n",
			    tmp, number);
		  }
#else
	      printf ("VGA support not compiled in!\r\n");
#endif
	      }
	    else {
#if USER_SUPPORT & SERIAL_SUPPORT
//	      UI_init (1);
	      printf (" Terminal [");
	      if (!USER_ACTIVE() && SERIAL_init() != 0) {
		  printf ("Switch to SERIAL failed!]\r\n");
		  continue;
		  }
	      printf ("%s] ", UI.info.serial.terminalname);
	      switch (UI.monitor) {
		case 0: printf ("Unknown"); break;
		case 1: printf ("B&W"); break;
		case 2: printf ("Monochrome"); break;
		case 3: printf ("Color"); break;
		}
	      printf (" with valid modes: 0..%u", UI.info.serial.max_mode);
	      printf ("\r\nswitch to SERIAL mode (from previous list): 0x");
	      number = gethexa();
	      if (number == (unsigned) -1)
		  printf ("\r\nStaying in current mode, at your own risks,\r\n"
			    "  Colors and attributes NOT initialised.\r\n\r\n");
		else
		  UI_setmode (number);
#else
	      printf ("SERIAL support not compiled in!\r\n");
#endif
	      }
	  continue;
#if USER_SUPPORT & VESA_SUPPORT
	case '6' : {
	  if (BOOT1_COM_port() >= 0) {
	      printf ("VESA on a serial line?\r\n\r\n");
	      continue;
	      }
	  if (!VESA_ACTIVE()) { /* Now in a VGA or uninitialised mode */
	      if (VESA_init() != 0) {
		  printf ("Switch to VESA failed!\r\n");
		  continue;
		  }
	      UI_clearscreen();
	      }
	    else
	      printf ("Already in VESA, active mode 0x%X\r\n",
			UI_getmode () & 0x7FFF);

	  printf ("%s card, VESA version %u.%u.",
			UI.info.vesa.cardname,
			UI.info.vesa.version >> 8,
			UI.info.vesa.version & 0xFF);
	  if (tester_probe_mode (0x100, 0x17F) == 0) {
	      printf ("\r\n Argggg - none of the modes are valid...\r\n"
		    "\r\n\007Press a key to continue...");
	      UI_getkey(0);
	      UI_init (0);
	      continue;
	      }

	  do {
#if (USER_SUPPORT & VESA_LINEAR) && (USER_SUPPORT & VESA_WINDOW)
	      if ((UI.info.vesa.version >> 8) >= 2)
		  printf ("\r\n (add 0x4000 for linear frambuffer mode)");
#endif
	      printf ("\r\nswitch to VESA mode (from previous list): 0x");
	      number = gethexa();
	      if (VESA_ACTIVE() && number == (unsigned) -1)
		  break;
	      } while (    (number & ~0x4000) < 0x100
			|| (number & ~0x4000) > 0x17F);
	  printf ("\r\n\r\n");
	  if (number == (unsigned) -1)
	      continue;
	  if (UI_setmode (number) != 0)
	      printf ("Setmode failed!\r\n");
	  }
	  continue;
#endif

#if 1
	case '=': {
	  char test[20] = { 0 };

	  printf ("strlen (\"%s\") = %u.\r\n", test, strlen(test));

	  printf ("strcpy (\"%s\", \"1234567XX\") = ", test);
	  strcpy (test, "1234567XX");
	  printf ("\"%s\".\r\n", test);

	  printf ("strcpy (\"%s\", \"1234567\") = ", test);
	  strcpy (test, "1234567");
	  printf ("\"%s\".\r\n", test);

	  printf ("strlen (\"%s\") = %u.\r\n", test, strlen(test));
	  printf ("strchr (\"%s\", '1') = %u.\r\n", test, strchr (test, '1'));
	  printf ("strchr (\"%s\", '3') = %u.\r\n", test, strchr (test, '3'));
	  printf ("strchr (\"%s\", '7') = %u.\r\n", test, strchr (test, '7'));
	  printf ("strchr (\"%s\", '8') = %u.\r\n", test, strchr (test, '8'));

#ifdef TEST_LIBINCL_H
	  printf ("strchr (\"%s\", '3') = '%c'\r\n",
			test, *strchr (test, '3'));
	  printf ("strchr (\"%s\", 'X') = 0x%X\r\n",
			test, strchr (test, 'X'));
	  printf ("strchr (\"%s\", '7') = 0x%X\r\n",
			test, strchr (test, '7'));
	  printf ("strchr (\"%s\", '\\0') = 0x%X\r\n",
			test, strchr (test, '\0'));
#endif

	  printf ("strncpy (\"%s\", \"123456789\", 6) = ", test);
	  strncpy (test, "123456790", 6);
	  printf ("\"%s\".\r\n", test);

	  printf ("strncmp (\"%s\", \"123450\", 5) = %d.\r\n",
			test, strncmp (test, "123450", 5));
	  printf ("strncmp (\"%s\", \"123450\", 6) = %d.\r\n",
			test, strncmp (test, "123450", 6));
	  printf ("strncmp (\"%s\", \"123457\", 6) = %d.\r\n",
			test, strncmp (test, "123457", 6));

	  printf ("strcmp (\"1234\", \"1234\") = %d\r\n",
			strcmp ("1234", "1234"));
	  printf ("strcmp (\"1234\", \"123456\") = %d\r\n",
			strcmp ("1234", "123456"));
	  printf ("strcmp (\"123456\", \"1234\") = %d\r\n",
			strcmp ("123456", "1234"));
	  printf ("strcmp (\"aaaab\", \"aaaac\") = %d\r\n",
	  strcmp ("aaaab", "aaaac"));
	  printf ("stricmp (\"aAaab\", \"aaaab\") = %d\r\n",
	  stricmp ("aAaab", "aaaab"));
	  printf ("stricmp (\"aAaab\", \"aaaac\") = %d\r\n",
	  stricmp ("aAaab", "aaaac"));

	  printf ("strcat (\"%s\", \"CC\") = ", test);
	  strcat (test, "CC");
	  printf ("\"%s\".\r\n", test);


	  strcpy (test, "-------");
	  printf ("\r\nmemset (\"%s\" + 1, 'Y', 3) = ", test);
	  memset (test + 1, 'Y', 3);
	  printf ("\"%s\".\r\n", test);

	  printf ("memcpy (\"%s\" + 1, \"XX\", 2) = ", test);
	  memcpy (test + 1, "XX", 2);
	  printf ("\"%s\".\r\n", test);

	  printf ("memeql (\"abc\", \"abc\", 3) = %u.\r\n",
		   memeql ("abc", "abc", 3));
	  printf ("memeql (\"abc\", \"abd\", 3) = %u.\r\n",
		   memeql ("abc", "abd", 3));

#ifdef TEST_LIBINCL_H
	  strcpy (test, "-------XX");
	  printf ("memmove (\"%s\" +1, \"%s\" +7, 2) = ", test, test);
	  memmove (test +1, test +7, 2);
	  printf ("\"%s\".\r\n", test);

	  printf ("memmove (\"%s\" +4, \"%s\" +1, 2) = ", test, test);
	  memmove (test +4, test +1, 2);
	  printf ("\"%s\".\r\n", test);
#endif

	  printf ("\r\nPress a key to continue...");
	  UI_getkey(0);
	  }
	  break;
#endif
	case '3': case '4': case '7':
	case '8': case '9': case ',': case '+':
	case 'G': case 'H': case 'I': case 'J': case 'K':
	case 'g': case 'h': case 'i': case 'j': case 'k':
	  /* those commands needs an initialised interface */
	  break;
	default:
	  continue;
	}

      if (!USER_ACTIVE()) {
	  printf (" Initialise the interface first!\r\n\r\n");
	  continue;
	  }

       /* Those commands do not need a initialised interface: */
      switch (key) {
	case '3':
	  /*
	   * Everex/Viewpoint %ah=0, %al=0x70, %bl=extended mode,
	   *  if %bl = 0, 640x480x16 => should at least work
	   *  (we could have scaned %bl = 0..0x7F)
	   * Paradise VGA, AT&T VDC600 %al = 0x7E or 0x7F:
	   *  extended functions, skip the autodetection for all video cards
	   * Strange noise for modes >= 0x70 on one board.
	   */
	  if (BOOT1_COM_port() < 0) {
#if USER_SUPPORT & VESA_SUPPORT
	      if (VESA_ACTIVE())
		  UI_putstr (
			/*0000000001111111111222222222233333333334*/
			/*1234567890123456789012345678901234567890*/
			" It is safe to probe VESA modes,    \r\n"
			"from 0x100 to 0x17F.                \r\n"
			);
		else
#endif
		  UI_putstr (
			/*0000000001111111111222222222233333333334*/
			/*1234567890123456789012345678901234567890*/
			" It is quite safe to probe video modes  \r\n"
			"up to 0x13, but you can probe up to 7F. \r\n"
			" You can turn OFF your monitor for these\r\n"
			"tests, a beep will be emitted at end.   \r\n"
			" This can take few minutes on slow PC,  \r\n"
			"but if CAPS LOCK led is no more working,\r\n"
			"your VIDEO BIOS is seriously buggy...   \r\n"
			" Then, just reboot your PC, the faulty  \r\n"
			"mode will never be tryed again.         \r\n"
			"As usual, TRY AT YOUR OWN RISKS !\r\n"
			"---------------------------------\r\n"
			);
	      }
	    else {
	      UI_putstr (
			" If you have a really VT420 compatible  \r\n"
			"display, you can probe up to 5, else    \r\n"
			"limit the probe from 0 to 1. If you have\r\n"
			"a non VT100 compatible terminal, I am   \r\n"
			"really sorry!                           \r\n"
			);
	      }

	  do {
	      printf ("\r\nFirst mode to probe (ESC to abort, <= 17F): 0x");
	      number = gethexa();
	      if (number == (unsigned) -1)
		  break;
	      } while (number > 0x17F);
	  if (number == (unsigned) -1)
	      break;
	  start = number;
	  do {
	      printf ("\r\nLast mode to probe (ESC to abort, >= 0x%X): 0x",
			start);
	      number = gethexa();
	      if (number == (unsigned) -1)
		  break;
	      } while (number < start);
	  if (number == (unsigned) -1)
	      break;
	  stop = number;
	  number = tester_probe_mode (start, stop);
	  if (VGA_ACTIVE()) {
	      void VGA_VESA_setup_color(void);
	      unsigned cpt;

	      /* probing modes in VGA may destroy palette */
	      VGA_VESA_setup_color ();
	      for (cpt = number + 2; cpt < UI.parameter.nbrow; cpt++)
		   printf ("\r\n"); /* erase the dirty chars generated */
	      UI_setcursor (number + 2, 0);
	      }
	  if (number == 0)
	      printf ("\r\n No valid mode found!");
	  printf ("\r\n\007Press a key to continue...");
	  UI_getkey(0);
	  break;
	case '4':
	  UI_clearscreen();
	  break;
	case '7': {
	  unsigned y, yend;
	  unsigned char row, col;
	  signed x;

	  if (!UI.parameter.attr.isgraphic)
	      break;
	  UI_getcursor (&row, &col);
	  yend = (row - 1) * (UI.parameter.charheigh? UI.parameter.charheigh : 8);
	  for (y = 0; y < yend; y++) {
	      for (x = (UI.parameter.nbcol - 1) * UI.parameter.charwidth - 1;
		   --x >= 0;) {
		  coord xy = {x: x, y: y}, xydest = xy;

		  xydest.x += UI.parameter.charwidth;
		  UI_setpixel (xydest, UI_getpixel (xy));
		  }
	      for (x = 0; x < UI.parameter.charwidth; x++) {
		  coord xy = {x: x, y: y};
		  UI_setpixel (xy, UI.bgcolor);
		  }
	      }
	  }
	  break;
	case '8': {
	  unsigned short x, y, xstart, tmp;
	  unsigned char saved_row, saved_col;
	  unsigned long long rdtsc_start;
	  unsigned total_rdtsc, nbtick, timeout;
	  unsigned total_rdtsc_clearscreen, total_rdtsc_scrolling;
	  unsigned calibrate_rdtsc;
	  unsigned char ioval;

	  setTFflagCounter (0);
	  setTFflag();
	  UI_clearscreen();
	  clearTFflag();

	  UI_putstr ("\r\n\r\n\r\n\r\n\r\n\r\n"
		" Will use assembly instruction 'rdtsc',\r\n"
		"   I hope you have it (i.e. at least Pentium) !\r\n");

	  disable_interrupt();
	  rdtsc_start = rdtsc();
	  UI_clearscreen();
	  total_rdtsc = rdtsc() - rdtsc_start;
	  enable_interrupt();
	  total_rdtsc_clearscreen = total_rdtsc;

	  UI_putstr ("\r\n\r\n"
  "   Got clearscreen result, now measuring scrolling and shifting\r\n\r\n"
  "  Number of instruction is excluding time spend in interrupts,\r\n"
  "  so is only really valid in VESA2 (linear framebuffer) modes,\r\n"
  "  or VESA1 when the positionning _function_ is used (curently ");
#if USER_SUPPORT & VESA_WINDOW
#if USER_SUPPORT & VESA_WINFUNCTION
	  if (UI.parameter.winsize >= 0x100000)
	      printf ("VESA2).\r\n\r\n");
	    else if (UI.parameter.addrpositionfarfct != 0)
	      printf ("true).\r\n\r\n");
	    else
#endif
	      printf ("false).\r\n\r\n");
#else
	  if (VESA_ACTIVE())
	      printf ("VESA2).\r\n\r\n");
	    else
	      printf ("VGA).\r\n\r\n");
#endif

	  tmp = getTFflagCounter();
	  printf ("  clear screen: %U instructions in %U cycles\r\n"
		  "     i.e. an average of %u cycles / instructions\r\n",
		  tmp, total_rdtsc, total_rdtsc / tmp);
	  UI_getcursor (&saved_row, &saved_col);

	  setTFflagCounter (0);
	  UI_setcursor (UI.parameter.nbrow-1, 0);
	  setTFflag();
	  printf ("\r\n");
	  clearTFflag();

	  UI_setcursor (UI.parameter.nbrow-1, 0);
	  disable_interrupt();
	  rdtsc_start = rdtsc();
	  printf ("\r\n");
	  total_rdtsc_scrolling = rdtsc() - rdtsc_start;
	  enable_interrupt();

	  UI_setcursor (saved_row - 2, 0);
	  tmp = getTFflagCounter();
	  printf ("  scrolling: %U instructions in %U cycles\r\n"
		  "     i.e. an average of %u cycles / instructions\r\n",
	  tmp, total_rdtsc_scrolling, total_rdtsc_scrolling / tmp);

	  xstart = (UI.parameter.nbcol - 1) * UI.parameter.charwidth - 1;

	  setTFflagCounter (0);
	  setTFflag();
	  for (y = 0; y < UI.parameter.charheigh; y++) {
	      for (x = xstart; x > 0;) {
		  coord xy = {x: --x, y: y},
			xydest = {x: x + UI.parameter.charwidth, y: y};
		  UI_setpixel (xydest, UI_getpixel (xy));
		  }
	      }
	  clearTFflag();

	  disable_interrupt();
	  rdtsc_start = rdtsc();
	  for (y = 0; y < UI.parameter.charheigh; y++) {
	      for (x = xstart; x > 0;) {
		  coord xy = {x: --x, y: y},
			xydest = {x: x + UI.parameter.charwidth, y: y};
		  UI_setpixel (xydest, UI_getpixel (xy));
		  }
	      }
	  total_rdtsc = rdtsc() - rdtsc_start;
	  enable_interrupt();

	  tmp = getTFflagCounter();
	  printf ("  shifting: %U instructions in %U cycles\r\n"
		  "     i.e. an average of %u cycles / instructions\r\n\r\n",
	  tmp, total_rdtsc, total_rdtsc / tmp);

	  nbtick = _BIOS_nbtick();
	  printf ("Calibrating rtdsc: "
#undef LONG_REPORT
#ifdef LONG_REPORT
		  "_BIOS_nbtick() = 0x%X ... ", nbtick
#endif
		 );
	  while (_BIOS_nbtick() == nbtick) ;
	  rdtsc_start = rdtsc();
	  nbtick += 10;	/* incremented every 55 ms */
	  while (_BIOS_nbtick() <= nbtick) ;
#if 0
	  calibrate_rdtsc = (rdtsc() - rdtsc_start) * 1000 / 55 / 10;
#elif 1
	  { unsigned rem, div = 55;
	  calibrate_rdtsc = ull_div_ul ((rdtsc() - rdtsc_start) * (1000 / 10),
					div, &rem);
	  }
#else
	 /* nbticks/sec = processor frequency < 2^32 = 4,294,967,295
		but do not multiply by 1000 first ! */
	  calibrate_rdtsc = (unsigned)(rdtsc() - rdtsc_start) / 55 / 10 * 1000;
#endif

#ifdef LONG_REPORT
	  printf ("0x%X, rdtsc = %U ticks/sec\r\n",
			 _BIOS_nbtick(), calibrate_rdtsc);
#else
	  printf ("%U ticks/sec\r\n", calibrate_rdtsc);
#endif

	  {
	  unsigned size = (-1 + my_ffs (UI.parameter.nbcolor))
			  * UI.parameter.width
			  * UI.parameter.heigh / 8;
	  /* The read bandwidth is approximate */
	  printf ("Video memory: %U bytes, "
		  "bandwidth: write= %u Mb/s, read= %u Mb/s.\r\n",
		  size,
		  calibrate_rdtsc / total_rdtsc_clearscreen
				  * size / 1024 / 1024,
		  calibrate_rdtsc / (total_rdtsc_scrolling
				     - total_rdtsc_clearscreen)
				  * size / 1024 / 1024
		  );
	  }

	  printf ("Checking vertical refresh bit: ");
	  timeout = 100000;
	  ioval = inb (0x3DA) & 0x08;	/* 0x3BA for MDA */
	  while (--timeout && ioval == (inb (0x3DA) & 0x08)) ;
	  if (!timeout) {
	      printf ("Error\r\n");
	      break;
	      }
	  ioval = inb (0x3DA) & 0x08;
	  while (ioval == (inb (0x3DA) & 0x08)) ;
	  rdtsc_start = rdtsc();
	  while (ioval != (inb (0x3DA) & 0x08)) ;
	  while (ioval == (inb (0x3DA) & 0x08)) ;
	  total_rdtsc = rdtsc() - rdtsc_start;

	  printf ("OK, done in %U cycles, refresh = %u Hz\r\n",
		total_rdtsc, (calibrate_rdtsc+total_rdtsc/2)/total_rdtsc);

	  printf ("Checking horizontal refresh bit: ");
	  timeout = 100000;
	  ioval = inb (0x3DA) & 0x01;  /* 0x3BA for MDA */
	  while (--timeout && ioval == (inb (0x3DA) & 0x01)) ;
	  if (!timeout) {
	      printf ("Error\r\n");
	      break;
	      }
	  ioval = inb (0x3DA) & 0x01;
	  while (ioval == (inb (0x3DA) & 0x01)) ;
	  rdtsc_start = rdtsc();
	  while (ioval != (inb (0x3DA) & 0x01)) ;
	  while (ioval == (inb (0x3DA) & 0x01)) ;
	  total_rdtsc = rdtsc() - rdtsc_start;

	  printf ("OK, done in %U cycles, Hfreq = %u KHz\r\n\r\n",
		total_rdtsc, (calibrate_rdtsc+total_rdtsc/2)/total_rdtsc
				 / 1000);
	  printf ("\r\nPress a key to continue...");
	  UI_getkey(0);
	  }
	  break;
	case '9': {
	  unsigned size, cpt;

	  for (cpt = 0; cpt < 16; cpt++)
	      for (size = 0; size < 20; size++) {
		  coord xy = {x: 40 - size, y: 3 + size + cpt};
		  UI_plotHline (xy, 40 + size - 1, UI.stdcolor[cpt]);
		  }
	  for (cpt = 0; cpt < 16; cpt++)
	      for (size = 0; size < 20; size++) {
		  coord xy = {x: 82 - size + cpt, y: 3 + size + cpt};
		  UI_plotHline (xy, 82 + size + cpt - 1,
					 UI.stdcolor[cpt]);
		  }
	  for (size = 0; size < 40; size++) {
	      coord xy = {x: 99 - size, y: 80 + size };
	      UI_plotHline (xy, 99 + size - 1, UI.stdcolor[magenta]);
	      }
	  printf ("\r\nPress a key to continue...");
	  UI_getkey(0);
	  }
	  break;
	case '+':
	case ',':
	  printf ("Current mode: ");
	  currmode = UI_getmode () & ~0x8000;
	  if (currmode & 0x4000)
	      printf ("0x%X (linear)\r\n", currmode & ~0x4000);
	    else
	      printf ("0x%X\r\n", currmode);
	  if (!VESA_ACTIVE()) {
	      if (BOOT1_COM_port() >= 0)
		  printf ("SERIAL");
		else
		  printf ("VGA");
	      }
	    else
	      printf ("VESA");
	  printf (" mode selected (ESC to abort");
#if (USER_SUPPORT & VESA_LINEAR) && (USER_SUPPORT & VESA_WINDOW)
	  if ((UI.info.vesa.version >> 8) >= 2 && VESA_ACTIVE())
	      printf (", add 0x4000 for linear mode");
#endif
	  printf ("): 0x");
	  if ((number = gethexa()) != (unsigned) -1) {
	      unsigned tmp;
	      if ((tmp = UI_setmode (number)) != 0) {
		  if (tmp > 2 && tmp < 0x100)
		      UI_setmode (currmode);
		  printf ("\r\nError 0x%X: video mode switch failed\r\n\r\n",
				 tmp);
		  }
		else {
		  display_pattern ();
		  printf ("\r\nPress return to keep that mode");
		  if ((UI_getkey(0) & 0x00FF) != '\r')
		      UI_setmode (currmode);
		  }
	      }
	  break;
	case 'G': case 'g': {
	  unsigned short i;
	  coord xy;
	  unsigned start_tick, total_tick;
	  unsigned char thekey;

	  if (!UI.parameter.attr.isgraphic) {
	      printf ("\r 'G' is just displaying the full screen\r\n"
		      " with the 16 colors in graphic modes.\r\n\r\n");
	      break;
	      }

	  for (i = 0; i < nbof (UI.stdcolor); i++) {
	      start_tick = _BIOS_nbtick();
	      for (xy.y = 0; xy.y < UI.parameter.heigh; xy.y++) {
		  if (key == 'g') {
		      for (xy.x = 0; xy.x < UI.parameter.width; xy.x++)
			  UI_setpixel (xy, UI.stdcolor[i]);
		      }
		    else {
		      xy.x = 0;
		      UI_plotHline (xy, UI.parameter.width,
						UI.stdcolor[i]);
		      }
		  }
	      total_tick = _BIOS_nbtick() - start_tick;

	      UI_setcursor (0, 0);
	      printf ("It took %u ticks. "
		      "Press a key to continue, ESC to abort", total_tick);
	      thekey = UI_getkey(0);
	      if (thekey == '\033' || thekey == 0 || thekey == '\003')
		  break;
	      }
	  }
	  break;
	case 'H':
	case 'h':
	  recorder_test (0);
	  break;
	case 'I':
	case 'i':
	  recorder_test (1);
	  break;
	case 'J':
	case 'j':
	  recorder_test (2);
	  break;
	case 'K':
	case 'k':
	  recorder_test (3);
	  break;
	}
      }
  }

__attribute__ ((cdecl, regparm (0)))
int main (struct registers *regs)
  {
  struct gujin_param_attrib gujin_attr;
  int tmp;
  unsigned basemem, extendmem, freemem;
  enum detect_memory_enum index;

  init_malloc();
  DBG_INIT();

  UI_init (FIRST_BOOT1_COM_port() >= 0);

  printf ("Compilation options: USER=0x%X, DISK=0x%X, DEBUG=0x%X,"
	  " SETUP=0x%X\r\n",
	  USER_SUPPORT, DISK_SUPPORT, DEBUG, SETUP);

  gujin_attr = get_gujin_param_attrib ();

  /* Check that - if error there will be memory corruptions... */
#if (USER_SUPPORT & VGA_SUPPORT)
  if (!gujin_attr.vga_support)
      printf ("Panic: vga_support attributes not in sync!\r\n");
#else
  if (gujin_attr.vga_support)
      printf ("Panic: vga_support attributes not in sync!\r\n");
#endif

  if (!BOOT1_DOS_running())
      gujin_attr.read_only_param = 0;
  set_gujin_param_attrib (gujin_attr);

  detect_environment (); /* before serial init, if serial interface */

  if (BOOT1_DOS_running()) {
      freemem = largest_free_himem();
      if (freemem)
	  printf ("Himem maxfree: %u Kb, ", freemem);
      }
  index = detect_memory (&basemem, &extendmem);
  printf ("Base memory: %u Kb, extended memory: %u Kb [%s]\r\n",
	   basemem, extendmem, mem_name[index]);

#if ASSEMBLY_TYPE == ASSEMBLY_DSES /* else done in install_exceptions */
  /*
   * We just warn the user here, and try to disable
   * all this stuff at load time:
   */
  if (BOOT1_DOS_running()) {
      farptr entry;
      unsigned short handle;

      if (DOS_loaded_HIGH())
	  BOOT1_putstr ( "WARNING: DOS is loaded HIGH!\r\n");
      if (SMARTDRV_control(0, SMARTDRV_get_info).abscent == 0)
	  BOOT1_putstr ( "WARNING: SMARTDRV is loaded!\r\n");

      entry = peekl (0x67*4);
      if (   entry != 0
	  && DOS_OpenFile ("EMMXXXX0", DOS_open_read_only, &handle) == 0) {
	  unsigned short errorcode;

	  DOS_CloseFile (handle, &errorcode);
	  printf ("EMM386 (vector 0x%X): ", entry);
	  entry = EMM386_check();
	  if (entry != 0) {
	      unsigned char state = EMM386_get_state(entry) & 0x03;
	      if (state == 1)
		  printf ("OFF.");
		else {
		  if (state & 2)
		      printf ("AUTO/");
		  if ((state & 1) == 0)
		      printf ("ON");
		    else
		      printf ("OFF");
#if 0
		  printf (", BIOS enable A20: 0x%X", BIOS_enable_A20());
#endif
		  printf (", switch to OFF: ");
		  if (EMM386_set_state (entry, 0x01) != 0x01)
		      BOOT1_putstr ("failed!");
		    else
		      BOOT1_putstr ("OK.");
		  }
	      BOOT1_putstr ("\r\n");
	      }
	    else
	      printf ("not found.\r\n");
	  }
      }
#endif

#if USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT)
  if (BOOT1_COM_port() < 0)
      detect_screen();
#endif

  printf ("\r\nDetecting disks:\r\n");
  probedisk(0, gujin_attr);
  printf ("Have found %d disks:\r\n", DI.nbdisk);
  list_disks();

  printf ("\r\nPress a key to continue...\r");
  UI_getkey(0);

  tmp = menu (regs);

#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
  if (BOOT1_COM_port() < 0) {
      _VGA_autoload_default_palette (1);
      if (VESA_ACTIVE() && !UI.parameter.BIOS_compatible) {
	  /* some BIOS need VESA before VGA setmode: */
	  if (_VESA_setmode (3) != 0)
	      VDBG ((" [Last _VESA_setmode failed!] "));
	  _VGA_setmode (3);
	  }
	else {
	  unsigned char nbcol, page;
	  _VGA_setmode (_VGA_getmode (&nbcol, &page) & 0x7F);
	  }
      }
#endif

#ifdef TREAT_EXCEPTION
  restore_vectors();
#endif
  DBG_END();
  return tmp;
  }

