/* util.h */
#ifndef UTIL_H
#define UTIL_H

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"

/*
 * recognise only "PATHNAME", without pattern matching,
 * because the software can only use _one_ directory per fs:
 */
#define PATHNAME	"boot"

/*
 * At least one of LINUZNAME1, LINUZNAME2, LINUZNAME3, KERNELNAME has
 * to be defined.
 * This bootloader will recognise "LINUZNAME[123]*" if they are defined,
 * and treat them as usual Linux kernel files, and recognise also
 * "KERNELNAME*" the same way, but it has to be a compressed binary image
 * linked (and starting) at linear address 0x100000, with parameters in
 * the comment field of the gzip structure, so without the "vmlinuz"
 * header.
 */
#define LINUZNAME1		"vmlinuz"
#define LINUZNAME2		"bzImage"
#define LINUZNAME3		"zImage"
#define KERNELNAME		"kernel-"

/*
 * If, in the same directory of "vmlinuz*", there is at least
 * one file named "initrd*", then an initial RAMdisk is loaded.
 *
 * The file loaded is the "initrd*" with the nearest ending,
 * i.e. if you have those files in /boot:
 *  vmlinuz-2.2.0.img   zImage-3.test   bzImage-2.4 kernel
 *  initrd     initrd-2.gz   initrd-2.2.gz
 *
 * Then "vmlinuz-2.2.0.img" will load "initrd-2.2.gz"
 *      "zImage-3.test"     will load "initrd"
 *      "bzImage-2.4"       will load "initrd-2.gz"
 *      "kernel"            will load "initrd"
 */
#define INITRDNAME	"initrd"

/*
 * path + filename max len:
 * Defined quite low to have malloced blocks of 64 bytes,
 * so better memory usage.
 */
#define NAME_LENGTH	52

/*
 * max nb of way to boot system:
 * The structure is malloced, but this is just defined
 * as security - for instance is there were a loop.
 */
#define NB_DESC_ARRAY	40

/*
 * max number of fragment for a FAT file - if the file
 * is not fragmented, it has one fragment whatever its size.
 * Undefine it to have malloc/realloc/free system - so the
 * limit is the available memory in data/stack segment.
 * The array is made of "struct sector_chain_str { unsigned lba, nb; };",
 * so eight bytes per fragment. Need to be bigger than 2 for the
 * end marker.
 * Keep it quite high in case of stripped RAID disk?
 */
#ifndef UTIL_USE_MALLOC
#define FAT_MAX_FRAGMENT	100
#endif

/*
 * for DOS & e2fs load file:
 */
#define DEFAULT_BUFFER_SIZE (4 * 1024)

/*
 * Should work the same with this defined, and saves a lot of
 * code space because e2fs_system_file_load() is using the same
 * e2fs library function than e2fs_analyse(), i.e. ext2fs_block_iterate2()
 *
 * Its main problem is that there is more memory malloced - and that
 * space is already small. The E2FSCACHE is not needed here because
 * it does not save _any_ read from HD.
 */
#define NEW_E2FS_READ

#ifndef NEW_E2FS_READ
/*
 * Probably only needed for kernels on a e2fs floppy.
 * Considering access pattern I have seen, the best
 * is E2FSCACHE = 2 or 3 - keep it as low as posible
 * to safe malloced memory.
 * With this code, only the first filesystem open
 * is cached (until it is closed).
 * Block number is also limited to an unsigned,
 * should be unsigned long long with 4 Kb blocks
 * and disks > 33.5 Tb.
 * Define to zero or undefine it to disable.
 * Do not hesitate to compile with DEBUG_CACHE and improve
 * the "aging" system.
 */
#ifndef E2FSCACHE	/* could be defined in Makefile/command line */
#define E2FSCACHE 3
#endif
#endif /* NEW_E2FS_READ */

/*
 * The structure where we store all the bootable systems we have found:
 */
extern struct boot_system_str {
    unsigned short nb, nb_initrd;
    struct desc_str {
	unsigned inode;
	unsigned filesize;
	unsigned char disk;
	unsigned char partition;	/* 0xFF for MBR */
	unsigned char name_offset;	/* to after the pattern */
	unsigned char in_slashboot	 : 1;
	unsigned char is_kernel		 : 1; /* else initrd */
	unsigned char has_vmlinuz_header : 1;
	char filename[NAME_LENGTH];
#ifdef UTIL_USE_MALLOC
	} *desc;
#else
	} desc[NB_DESC_ARRAY];
#endif
    } boot_system;

/*
 * Prototypes and intersegment stuff:
 */
struct system_load_str; /* main.c do not include vmlinuz.h */

#if (SETUP & EXTRA_SEGMENT) && !defined (CALLING_FROM_EXTRASEG)
#define ASMNAME(x)	asm (STRING(codeseg_ ## x))
#else
#define ASMNAME(x)	/* call in same segment */
#endif

void disk_analyse (struct gujin_param_attrib attr) ASMNAME(disk_analyse);

unsigned system_file_load (unsigned index, struct system_load_str *loader)
		ASMNAME(system_file_load);

#undef ASMNAME

#endif /* UTIL_H */
