/* vmlinuz.c */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"
#if SETUP & EXTRA_SEGMENT	/* before other includes */
#define CALLING_FROM_EXTRASEG
#endif

#include "library.h"
#include "boot.h"
#include "debug.h"
#include "user.h"
#include "ide.h"	/* _IDE_guess_altadr() */
#include "disk.h"
#include "bios.h"	/* vmlinuz_EGA() */
#include "mouse.h"
#include "dos.h"	/* SMARTDRV_control() */
#include "xms.h"
#include "vesabios.h"	/* _VESA_GetPMinterface() */
#include "vgabios.h"	/* _VGA_get_display_combination_code() */
#include "util.h"	/* struct desc_str */
#include "vmlinuz.h"
#include "messages.h"

// binutils bug:
// the "vmlinuz.lis" file seems trucated around definition
// of variable "speed_array.1138"

static const union header_signature_union header_signature = {
  { 'H', 'd', 'r', 'S' }
  };

SEGMENT_EXTRA(vmlinuz_EGA) static inline void
vmlinuz_EGA (struct screen_info *screen_info)
  {
  union EGA_bx_union EGA_bx;
  struct EGA_cx_str EGA_cx;

  ZDBG (("%s: ", __FUNCTION__));
  _EGA_getdisplay (&EGA_bx, &EGA_cx);
  ZDBG ((" (_EGA_getdisplay: memory %u, IOaddr %u, "
	 "switchs: 1:%u 2:%u 3:%u 4:%u)",
	 EGA_bx.enums.memory,
	 EGA_bx.enums.IOaddr,
	 EGA_cx.switch1off,
	 EGA_cx.switch2off,
	 EGA_cx.switch3off,
	 EGA_cx.switch4off));
  screen_info->orig_video_ega_bx = EGA_bx.IOaddr_memory;
  screen_info->orig_video_isVGA = 0;
  if (   EGA_bx.enums.IOaddr == EGA_COLOR_3Dx
      || EGA_bx.enums.IOaddr == EGA_MONO_3Bx) {
      unsigned char active_dcc, alternate_dcc;

      if (_VGA_get_display_combination_code (&active_dcc,
					     &alternate_dcc) == 0) {
	  ZDBG ((", VGA detected (active DCC: 0x%X, alternate DCC: 0x%X)",
			active_dcc, alternate_dcc));
	  screen_info->orig_video_isVGA = 1;
	  }
	else
	  ZDBG ((", EGA detected"));

      screen_info->orig_video_lines = peekb (0x00400084) + 1;
      screen_info->orig_font_heigh = peekw (0x00400085);
      }
    else { /* EGA_bx.enums.IOaddr unchanged => CGA/MDA/HGA */
      ZDBG ((", non EGA/VGA detected"));
      screen_info->orig_video_lines = 25;
      screen_info->orig_font_heigh = 0; /* invalid */
      }
  ZDBG ((", lines=%u, font heigh=%u.\r\n",
	 screen_info->orig_video_lines,
	 screen_info->orig_font_heigh));
  }

SEGMENT_EXTRA(vmlinuz_getcursor) static inline void
vmlinuz_getcursor (struct screen_info *screen_info)
  {
#if 1 /* what is the best ? IHMO using the BIOS functions */
  unsigned char row, col;
  unsigned char cursorstart, cursorstop;

  _VGA_getcursor (screen_info->orig_video_page,
		  &row, &col,
		  &cursorstart, &cursorstop);
  ZDBG (("_VGA_getcursor: row %u, col %u, cursorstart %u, cursorstop %u\r\n",
		row, col, cursorstart, cursorstop));
  screen_info->orig_x = col;
  screen_info->orig_y = row;
#else
  unsigned char row, col;
  unsigned ptr = 0x00400050 + 2 * screen_info->orig_video_page;

  col = peekb (ptr);
  row = peekb (ptr+1);
  ZDBG (("BIOS probe (0x%X): row %u, col %u\r\n", ptr, row, col));
  screen_info->orig_x = col;
  screen_info->orig_y = row;
#endif
  }

SEGMENT_EXTRA(vmlinuz_getVGAmode) static inline void
vmlinuz_getVGAmode (struct screen_info *screen_info)
  {
  unsigned char mode, nbcol, page;

  mode = _VGA_getmode (&nbcol, &page);
  ZDBG (("_VGA_getmode mode: 0x%X, nbcol: %u, page: %u; ", mode, nbcol, page));
  screen_info->orig_video_mode = mode & 0x7F;
  screen_info->orig_video_cols = nbcol;
  screen_info->orig_video_page = page; /* not tested if != 0 */
  }

#if USER_SUPPORT != 0
SEGMENT_EXTRA(vmlinuz_VGA) static inline void
vmlinuz_VGA (struct screen_info *screen_info, unsigned row)
  {
  ZDBG (("%s: ", __FUNCTION__));
  vmlinuz_getVGAmode (screen_info);
  ZDBG (("setting cursor at beginning of line %u.\r\n", row));
  screen_info->orig_x = 0;
  screen_info->orig_y = row;
  }
#else
SEGMENT_EXTRA(vmlinuz_VGA) static inline void
vmlinuz_VGA (struct screen_info *screen_info)
  {
  ZDBG (("%s: ", __FUNCTION__));
  vmlinuz_getVGAmode (screen_info);
  vmlinuz_getcursor (screen_info);
  /* We try not to erase the text written later by Gujin: */
  screen_info->orig_x = 0;
  if (screen_info->orig_y + 4 < screen_info->orig_video_lines)
      screen_info->orig_y += 4;
    else
      screen_info->orig_y = screen_info->orig_video_lines - 1;
  }
#endif

SEGMENT_EXTRA(vmlinuz_VESA) static inline void
vmlinuz_VESA (struct screen_info *screen_info, unsigned mode)
  {
  VESA_modeinfo_type modeinfo;

  ZDBG (("%s: ", __FUNCTION__));
  if (_VESA_getmodeinfo (&modeinfo, mode & 0x01FF) == 0) {
      ZDBG (("(_VESA_getmodeinfo OK) "));
      if (modeinfo.MemoryModel == mem_text && modeinfo.NumberOfPlanes == 4) {
	  if (modeinfo.BytesPerScanLine != 2 * modeinfo.XResolution) {
	      ZDBG (("(correcting linelength: old=%u, new=%u) ",
			modeinfo.BytesPerScanLine, 2 * modeinfo.XResolution));
	      modeinfo.BytesPerScanLine = 2 * modeinfo.XResolution;
	      }
	  if (modeinfo.BitsPerPixel != 4) {
	      ZDBG (("(correcting BitsPerPixel: old=%u, new=4) ",
			modeinfo.BitsPerPixel));
	      modeinfo.BitsPerPixel = 4;
	      }
	  }
	else {
	  /* no text modes with linear framebuffer */
	  if (modeinfo.ModeAttributes.linear_framebuffer_supported) {
	      ZDBG (("(enable VESA framebuffer) "));
	      screen_info->orig_video_isVGA = 0x23;
	      screen_info->lfb_base   = modeinfo.PhysBasePtr;
	      }
	  }
      screen_info->lfb_linelength = modeinfo.BytesPerScanLine;
      screen_info->lfb_width  = modeinfo.XResolution;
      screen_info->lfb_height = modeinfo.YResolution;
      screen_info->lfb_depth  = modeinfo.BitsPerPixel;
      screen_info->lfb_pages  = modeinfo.NumberOfImagePages;

#if 0 // When GCC will optimise that pattern ?
      screen_info->red_size   = modeinfo.RedMaskSize;
      screen_info->red_pos    = modeinfo.RedFieldPosition;
      screen_info->green_size = modeinfo.GreenMaskSize;
      screen_info->green_pos  = modeinfo.GreenFieldPosition;
      screen_info->blue_size  = modeinfo.BlueMaskSize;
      screen_info->blue_pos   = modeinfo.BlueFieldPosition;
      screen_info->rsvd_size  = modeinfo.RsvdMaskSize;
      screen_info->rsvd_pos   = modeinfo.RsvdFieldPosition;
#else // this saves 72 bytes.
      memcpy (&screen_info->red_size, &modeinfo.RedMaskSize, 8);
#endif
      ZDBG (("OK.\r\n"));
      }
    else
      ZDBG (("_VESA_getmodeinfo failed.\r\n"));
  }

SEGMENT_EXTRA(vmlinuz_VIDEO) static inline void
vmlinuz_VIDEO (struct screen_info *screen_info
#if USER_SUPPORT != 0
		, unsigned row
#endif
		)
  {
  VESA_VbeInfoBlock infoblock;

  ZDBG (("%s: ", __FUNCTION__));

  vmlinuz_EGA (screen_info);

#if USER_SUPPORT != 0
  vmlinuz_VGA (screen_info, row);
#else
  vmlinuz_VGA (screen_info);
#endif

  ZDBG (("\r\n%s: orig_*: video_ega_bx: 0x%X, video_lines: %u, video_cols: %u, "
	 "font_heigh: %u, video_mode: 0x%X, video_page: %u, x: %u, y: %u, "
	 "isVGA: 0x%X\r\n",
	__FUNCTION__,
	screen_info->orig_video_ega_bx,	screen_info->orig_video_lines,
	screen_info->orig_video_cols,	screen_info->orig_font_heigh,
	screen_info->orig_video_mode,	screen_info->orig_video_page,
	screen_info->orig_x,		screen_info->orig_y,
	screen_info->orig_video_isVGA
	));

  /* normaly only for graphic modes: */
  ZDBG (("_VESA_getinfo: "));
  if (_VESA_getinfo (&infoblock) == 0) {
      unsigned short mode, len;
      farptr addr;

      screen_info->lfb_size = infoblock.TotalMemory;
      ZDBG (("OK, Total Video Memory = 0x%X * 64Kb", infoblock.TotalMemory));
      ZDBG ((", _VESA_getmode: "));
      if (_VESA_getmode (&mode) == 0) {
	  ZDBG (("OK, mode = 0x%X\r\n", mode));
	  vmlinuz_VESA (screen_info, mode);
	  }
	else
	  ZDBG (("failed.\r\n"));
      /* we could put "mode" in "in_out_info.vid_mode", but... */

      ZDBG (("_VESA_GetPMinterface: "));
      if (_VESA_GetPMinterface (&addr, &len) == 0) {
	  ZDBG (("OK, addr = 0x%X, len = 0x%X\r\n", addr, len));
	  screen_info->vesapm_seg = addr >> 16;
	  screen_info->vesapm_off = addr & 0xFFFF;
	  }
	else
	  ZDBG (("failed.\r\n"));
      }
    else
      ZDBG (("failed.\r\n"));

  ZDBG (("%s: lfb_*: isVGA: 0x%X, linelength: %u, "
	 "width: %u, height: %u, depth: %u, pages: %u, "
	 "base: 0x%X, lfb_size = %u * 64 Kb, "
	 "red: %u<<%u, green: %u<<%u, blue: %u<<%u, rsvd: %u<<%u, "
	 "vesapm_seg: 0x%X, vesapm_off: 0x%X\r\n",
	__FUNCTION__,
	screen_info->orig_video_isVGA,
	screen_info->lfb_linelength,
	screen_info->lfb_width,		screen_info->lfb_height,
	screen_info->lfb_depth,		screen_info->lfb_pages,
	screen_info->lfb_base,		screen_info->lfb_size,
	screen_info->red_size,		screen_info->red_pos,
	screen_info->green_size,	screen_info->green_pos,
	screen_info->blue_size,		screen_info->blue_pos,
	screen_info->rsvd_size,		screen_info->rsvd_pos,
	screen_info->vesapm_seg,	screen_info->vesapm_off
	));
  }

SEGMENT_EXTRA(vmlinuz_EXT_MEM_K) static inline unsigned short
vmlinuz_EXT_MEM_K (void)
  {
  unsigned short EXT_MEM_K;

  if (_BIOS_getExtendedMemory (&EXT_MEM_K) != 0 || EXT_MEM_K == 0) {
      ZDBG (("(%s: _BIOS_getExtendedMemory error, using NVRAM)",
		__FUNCTION__));
      EXT_MEM_K = (((unsigned short)getnvram(0x18)) << 8) + getnvram(0x17);
      }
    else
      ZDBG (("(%s: _BIOS_getExtendedMemory OK)", __FUNCTION__));
  return EXT_MEM_K;
  }

SEGMENT_EXTRA(vmlinuz_APM) static inline void
vmlinuz_APM (struct apm_bios_info *apm_bios_info)
  {
  ZDBG (("%s: try APM_installation_check: ", __FUNCTION__));
  if (APM_installation_check (&apm_bios_info->version,
			      &apm_bios_info->flags) == 0) {
      ZDBG (("OK, APM version 0x%X, flags 0x%X, APM32 connection ",
		apm_bios_info->version, apm_bios_info->flags));
      APM_disconnect (); /* ignore possible error */
      if ((apm_bios_info->flags & 2) == 0) {
	  ZDBG (("not supported.\r\n"));
	  }
	else if (APM_connect_32bit(&apm_bios_info->cseg,
				   &apm_bios_info->offset,
				   &apm_bios_info->cseg_16,
				   &apm_bios_info->dseg,
				   &apm_bios_info->cseg_len,
				   &apm_bios_info->dseg_len) != 0) {
	  apm_bios_info->flags &= ~2; /* remove 32bits support */
	  ZDBG (("failed.\r\n"));
	  }
	else {
	  ZDBG (("OK, re-check: "));
	  /*
	   * Redo the installation check as the 32 bit connect
	   * modifies the flags returned on some BIOSs :
	   * (from linux/arch/i386/boot/setup.S)
	   */
	  if (APM_installation_check (&apm_bios_info->version,
				      &apm_bios_info->flags) != 0) {
	      apm_bios_info->flags &= ~2;
	      APM_disconnect ();
	      ZDBG (("failed.\r\n"));
	      }
	    else
	      ZDBG (("OK.\r\n"));
	  }
      }
    else
      ZDBG (("no APM BIOS.\r\n"));
  }

SEGMENT_EXTRA(vmlinuz_DISKS) static inline void
vmlinuz_DISKS (struct drive_info *drive1, struct drive_info *drive2)
  {
  unsigned nbsect;
  unsigned char status;

  if (_BIOSDISK_gettype (0x80, &nbsect,	&status) == 0)
      lmemcpy (drive1, peekl (0x41 * 4), sizeof (struct drive_info));
    else
      ZDBG ((" [no BIOS disk 0x80] "));

  if (_BIOSDISK_gettype (0x81, &nbsect,	&status) == 0)
      lmemcpy (drive2, peekl (0x46 * 4), sizeof (struct drive_info));
    else
      ZDBG ((" [no BIOS disk 0x81] "));
  }

SEGMENT_EXTRA(vmlinuz_CONFIG) static inline void
vmlinuz_CONFIG (struct sys_desc_table_info *sys_desc_table)
  {
  farptr adr;

  ZDBG (("%s: try _BIOS_getConfiguration: ", __FUNCTION__));
  adr = _BIOS_getConfiguration();
  if (adr != 0) {
      ZDBG (("OK, at 0x%X.\r\n", adr));
      lmemcpy (sys_desc_table, adr, sizeof (struct sys_desc_table_info));
      }
    else
      ZDBG (("failed.\r\n"));
  }

SEGMENT_EXTRA(vmlinuz_IN_OUT) static inline void
vmlinuz_IN_OUT (struct in_out_info *in_out,
		unsigned short minor_major_root,
		unsigned initrd_load_address,
		unsigned initrd_length,
		unsigned linadr_cmdline)
  {
  in_out->header_signature.longword = header_signature.longword;
  in_out->header_version_number = 0x0202;
  in_out->cmd_line_ptr = linadr_cmdline;
  in_out->type_of_loader.type = GUJIN;
  in_out->type_of_loader.version = 0;
  in_out->root.minor_major = minor_major_root;
  in_out->root_flags.read_only = 1;
  in_out->ramdisk_image = initrd_load_address;
  in_out->ramdisk_size  = initrd_length;
  }

SEGMENT_EXTRA(vmlinuz_E820) static inline void
vmlinuz_E820 (struct e820map_info *e820map,
	      unsigned e820map_nb,
	      unsigned char *nb_E820_entries)
  {
  unsigned cont_val = 0, tmp,
	   elemsize = sizeof (struct e820map_info),
	   bufsize = e820map_nb * elemsize;
  void *ptr = e820map;

  ZDBG (("%s: try _BIOS_QueryMemoryMap (bufsize %u at 0x%X): ",
		__FUNCTION__, bufsize, ptr));
  while (   (tmp = _BIOS_QueryMemoryMap (&cont_val, &bufsize, &ptr)) == 0
	 && cont_val != 0
	 && bufsize >= elemsize)
      ZDBG ((" (ret 0x%X, cont_val %u, bufsize %u, quering next memory map) ",
		tmp, cont_val, bufsize));
  ZDBG ((" (end with ret 0x%X, cont_val %u, bufsize %u) ",
		tmp, cont_val, bufsize));
  *nb_E820_entries = (e820map_nb * elemsize - bufsize) / elemsize;

  ZDBG (("%u E820_entries: ", *nb_E820_entries));
  for (cont_val = 0; cont_val < *nb_E820_entries; cont_val++)
      ZDBG ((" {addr: 0x%X, size: 0x%X, type: 0x%X} ",
		(unsigned)e820map[cont_val].addr,
		(unsigned)e820map[cont_val].size,
		(unsigned)e820map[cont_val].type));
  ZDBG ((".\r\n"));
  }

/*
 * Command line stuff:
 */
SEGMENT_EXTRA(vmlinuz_console_CMDLINE) static inline char *
vmlinuz_console_CMDLINE (char *ptr)
  {
  static const unsigned short speed_array[] = {
      110, 150, 300, 600, 1200, 2400, 4800, 9600
      };

  ptr += sprintf (ptr, " console=ttyS%u,%u%c%u",
		BOOT1_COM_port(),
		speed_array[BOOT1_COM().speed],
		BOOT1_COM().parity ? (BOOT1_COM().even? 'e' : 'o') : 'n',
		BOOT1_COM().eightbit? 8 : 7
		);
  return ptr;
  }

SEGMENT_EXTRA(vmlinuz_ide_CMDLINE) static inline char *
vmlinuz_ide_CMDLINE (char *ptr)
  {
  unsigned short cpt;

  /* We let the Linux driver guess the IRQ used... */
  for (cpt = 0;
       cpt < nbof (DI.IDE_IObase_found) && DI.IDE_IObase_found[cpt];
       cpt++)
      ptr += sprintf (ptr, " ide%u=0x%x,0x%x",
		cpt, DI.IDE_IObase_found[cpt],
		_IDE_guess_altadr (DI.IDE_IObase_found[cpt]));
  return ptr;
  }

SEGMENT_EXTRA(vmlinuz_mouse_CMDLINE) static inline char *
vmlinuz_mouse_CMDLINE (char *ptr)
  {
#if USER_SUPPORT & (BIOS_MOUSE_SUPPORT | SERIAL_MOUSE_SUPPORT)
  enum MOUSE_type mouse;

  if ((mouse = MOUSE_gettype()) != MOUSE_NONE) {
      ptr = STRINIT (ptr, " mouse=");
      if (mouse == MOUSE_PS2)
	  ptr = STRINIT (ptr, "/dev/psaux");
	else {
	  ptr = STRINIT (ptr, "/dev/ttyS");
	  *ptr++ = '0' + (mouse & ~MOUSE_SERIAL_MASK);
	  *ptr++ = ',';
	  if ((mouse & MOUSE_SERIAL_MASK) == MOUSE_SERIAL_COM1)
	      *ptr++ = '2';
	    else if ((mouse & MOUSE_SERIAL_MASK) == MOUSE_SERIAL3_COM1)
	      *ptr++ = '3';
	    else
	      *ptr++ = 'M';
	  *ptr = '\0';
	  }
      }
#endif /* MOUSE_SUPPORT */
  return ptr;
  }

SEGMENT_EXTRA(vmlinuz_general_variables_CMDLINE) static inline char *
vmlinuz_general_variables_CMDLINE (char *ptr)
  {
#if USER_SUPPORT != 0
  /* those are already given in vmlinuz_VIDEO(), but for serial modes */
  ptr += sprintf (ptr, " COLS=%u LINES=%u",
		  UI.parameter.nbcol, UI.parameter.nbrow);
  if (BOOT1_COM_port() < 0)
      ptr += sprintf (ptr, " NumLock=%s",
		_BIOS_get_shift_flags().num_lock_active? "on" : "off");
#endif
  return ptr;
  }

SEGMENT_EXTRA(vmlinuz_CMDLINE) static inline void
vmlinuz_CMDLINE (char *command_line, unsigned orig_video_isVGA,
		 char *cmdline_extraparam, farptr extra_cmdline
#ifdef ROOT_ON_COMMAND_LINE
		 , unsigned short root_filesystem
#endif
		 )
  {
  char *cmdptr = command_line;
  unsigned short tmp;

  cmdptr = STRINIT (cmdptr, "linux");

  if (BOOT1_COM_port() >= 0)
      cmdptr = vmlinuz_console_CMDLINE (cmdptr);
    else if (orig_video_isVGA == 0x23)
      cmdptr = STRINIT (cmdptr, " video=vesa");

  cmdptr = vmlinuz_ide_CMDLINE (cmdptr);
  cmdptr = vmlinuz_mouse_CMDLINE (cmdptr);
  cmdptr = vmlinuz_general_variables_CMDLINE (cmdptr);
#ifdef LANG_CMDLINE
  cmdptr = STRINIT (cmdptr, " "LANG_CMDLINE);
#endif

  /* Add nearby end the Gujin gujin_param.extra_cmdline field: */
  if (peekb (extra_cmdline)) {
      *cmdptr++ = ' ';
      for (;;) {
	  *cmdptr = peekb (extra_cmdline++);
	  if (*cmdptr)
	      cmdptr++;
	    else
	      break;
	  }
      }

  /* add at end the params in the comment part of the GZIP file: */
  if ((tmp = strlen (cmdline_extraparam)) != 0) {
      *cmdptr++ = ' ';
      memcpy (cmdptr, cmdline_extraparam, tmp);
      cmdptr += tmp;
      }

#ifdef ROOT_ON_COMMAND_LINE
  if (root_filesystem) {
      const char *ptr = command_line;
      /* first look if there is already a "root=" given: */
      while (*ptr) {
	  if (memeql (ptr, "root=", sizeof ("root=") - 1))
	      break;
	  ptr++;
	  }
      if (*ptr == 0) {
	  /* "root=" parameter not already given */
	  cmdptr = STRINIT (cmdptr, " root=/dev/hd");
	  *cmdptr++ = root_filesystem >> 8; /* i.e. 'a' */
	  *cmdptr++ = root_filesystem; /* i.e. '1' */
	  *cmdptr = '\0';
	  }
      }
#endif

  }

/*
 * The L-user PC is a new-BIOS IDE-only PC, with only one distribution (one
 * root filesystem), and that function will guess the root filesystem when
 * the default root in vmlinuz is either zero or SCSI like defaut kernels
 * in most distributions. It may fail (more than one E2FS root filesystem
 * containing a "/boot" directory) - so you would better use another way...
 */
#if DISK_SUPPORT & IDE_SUPPORT
SEGMENT_EXTRA(root_filesystem) static inline unsigned short
root_filesystem (struct desc_str *system_desc, unsigned short vmlinuz_root)
  {
  unsigned short number;
  struct diskparam_str *dsk = &DI.param[system_desc->disk];
  unsigned short IDEcounter;

  if (vmlinuz_root && (vmlinuz_root >> 8) != 0x08) {
      ZDBG (("%s: not guessing root, already inited to 0x%X.\r\n",
		__FUNCTION__, vmlinuz_root));
#ifdef ROOT_ON_COMMAND_LINE
      return 0;
#else
      return vmlinuz_root;
#endif
      }

  if (   system_desc->in_slashboot
      && dsk->partition[system_desc->partition].type == Linux) {
      /*
       * Guess the root is the partition containing /boot/vmlinuz:
       */
      number = dsk->partition[system_desc->partition].OSnumber;
      }
    else {
      unsigned char probeall;

      /*
       * No path in vmlinuz file, get the _last_ "probably root"
       * partition of the disk:
       */
      if (vmlinuz_root != 0) {
	  /* if it is maybe a SCSI system, we should not touch the value: */
	  unsigned char cpt_bios = 0, cpt_ide = 0, flag_unknown = 0, cpt;

	  for (cpt = 0; cpt < DI.nbdisk; cpt++) {
	      if (DI.param[cpt].disknb <= 3)
		  continue; /* floppies do not count */
	      if (   DI.param[cpt].access == bios_chs
		  || DI.param[cpt].access == ebios_lba) {
		  cpt_bios++;
		  if (DI.param[cpt].ideIOadr == 0) /* maybe SCSI */
		      flag_unknown = 1;
		  }
		else
		  cpt_ide++;
	      }

	  if (flag_unknown && (cpt_bios > cpt_ide)) {
	      ZDBG (("%s: not guessing root, maybe SCSI: root stay 0x%X.\r\n",
			__FUNCTION__, vmlinuz_root));
#ifdef ROOT_ON_COMMAND_LINE
	      return 0;
#else
	      return vmlinuz_root;
#endif
	      }
	  }
      number = 0;
      probeall = 1;
      for (;;) {
	  unsigned char cpt;

	  for (cpt = 0; cpt < dsk->nbpartition; cpt++)
	      if (   dsk->partition[cpt].type == Linux
		  && dsk->partition[cpt].misc.maybe_root)
		  number = dsk->partition[cpt].OSnumber;
	  if (number != 0 && dsk->ideIOadr != 0)
	      break;
	  ZDBG (("%s: did not find a useable root filesystem on this disk.\r\n",
			__FUNCTION__));
	  if (probeall) {
	      probeall = 0;
	      dsk = &DI.param[0];
	      }
	    else if (++dsk >= &DI.param[DI.nbdisk]) {
	      ZDBG (("%s: did not find any useable root filesystem.\r\n",
			__FUNCTION__));
#ifdef ROOT_ON_COMMAND_LINE
	      return 0;
#else
	      return vmlinuz_root;
#endif
	      }
	  }
      }

  for (IDEcounter = 0; IDEcounter < nbof (DI.IDE_IObase_found); IDEcounter++) {
      if (DI.IDE_IObase_found[IDEcounter] == 0) {
	  IDEcounter = nbof (DI.IDE_IObase_found);
	  break;
	  }
      if (DI.IDE_IObase_found[IDEcounter] == dsk->ideIOadr)
	  break;
      }

#ifdef ROOT_ON_COMMAND_LINE
  /* number is two chars: MSB is the letter and LSB is the number
     of /dev/hd<letter><number> : */
  number += '0';
  if (IDEcounter < nbof (DI.IDE_IObase_found))
      number |= ('a' + 2 * IDEcounter) << 8;
    else {
      ZDBG (("%s: do not know ideIOadr of this disk.\r\n", __FUNCTION__));
      return 0;
      }
  if (!dsk->ismaster)
      number += 0x100;
#else
  /*
   * look files linux/init/main.c structure "root_dev_names"
   * to get the relation hd? <-> number. It is _not_ the
   * MAJOR() of the block device, which you can get from
   * "stat /dev/hda" or simply "ls -l /dev/hd[a-h]",
   * but for hd[ab] !
   */

  if (IDEcounter < nbof (DI.IDE_IObase_found)) {
      static const unsigned char root_array[] = {
	  0x03, 0x16, 0x21, 0x22, 0x38, 0x39, 0x58, 0x59, 0x5A, 0x5B
	  };
      number |= ((unsigned short)root_array[IDEcounter]) << 8;
      }
    else {
      ZDBG (("%s: do not know ideIOadr of this disk.\r\n", __FUNCTION__));
      return vmlinuz_root;
      }
  if (!dsk->ismaster)
      number |= 0x40;
#endif

  return number;
  }
#else /* DISK_SUPPORT & IDE_SUPPORT */
SEGMENT_EXTRA(root_filesystem) static inline unsigned short
root_filesystem (struct desc_str *system_desc, unsigned short vmlinuz_root)
  {
#ifdef ROOT_ON_COMMAND_LINE
  return 0;
#else
  return vmlinuz_root;
#endif
  }
#endif /* DISK_SUPPORT & IDE_SUPPORT */

SEGMENT_EXTRA(linux_set_params) static inline void
linux_set_params (struct desc_str *system_desc,
		  farptr          params_farptr,
		  unsigned short  minor_major_root,
		  unsigned        extendmem,
		  unsigned        initrd_load_address,
		  unsigned        initrd_length,
		  char		  *cmdline_extraparam,
		  farptr	  extra_cmdline)
  {
  struct linux_param LnxParam = {};
#ifndef ROOT_ON_COMMAND_LINE
  unsigned rootfs;
#endif

  test_linux_param_struct();

  ZDBG (("\r\n\r\n%s:\r\n", __FUNCTION__));
  _BIOS_set_repeat_rate_delay (0, 0); /* max repeat rate, min delay */

#if USER_SUPPORT != 0 /* else UI does not exists */
  /* leave few lines to display Gujin messages: */
  {
  unsigned char row, col;

  UI_getcursor (&row, &col);
  vmlinuz_VIDEO (&LnxParam.screen_info, row + 4);
  }
#else
  vmlinuz_VIDEO (&LnxParam.screen_info);
#endif

  LnxParam.screen_info.EXT_MEM_K = vmlinuz_EXT_MEM_K();

  /* old Command Line magic for 2.2- kernels: */
  LnxParam.screen_info.CL_MAGIC = 0xA33F;
  LnxParam.screen_info.CL_OFFSET = offsetof (struct linux_param,
						command_line);
  LnxParam.ALT_MEM_K = extendmem;
  ZDBG (("EXT_MEM_K = 0x%X, ALT_MEM_K = 0x%X, "
	 "CL_MAGIC = 0x%X, CL_OFFSET = 0x%X\r\n",
		LnxParam.screen_info.EXT_MEM_K,
		LnxParam.ALT_MEM_K,
		LnxParam.screen_info.CL_MAGIC,
		LnxParam.screen_info.CL_OFFSET));

  vmlinuz_APM (&LnxParam.apm_bios_info);
  /* APM32 connect could switch to protected mode and close doors? */

  vmlinuz_DISKS (&LnxParam.drive1, &LnxParam.drive2);
  vmlinuz_CONFIG (&LnxParam.sys_desc_table);
  vmlinuz_IN_OUT (&LnxParam.in_out,
#ifdef ROOT_ON_COMMAND_LINE
		  minor_major_root,
#else
		  rootfs = root_filesystem (system_desc, minor_major_root),
#endif
		  initrd_load_address,
		  initrd_length,
		  /* New Command Line magic for 2.4+ kernels: */
		  farptr2linear(params_farptr)
		   + offsetof(struct linux_param, command_line)
		  );
  vmlinuz_E820 (LnxParam.e820map,
		nbof (LnxParam.e820map),
		&LnxParam.nb_E820_entries);
  vmlinuz_CMDLINE (LnxParam.command_line,
		   LnxParam.screen_info.orig_video_isVGA,
		   cmdline_extraparam,
		   extra_cmdline
#ifdef ROOT_ON_COMMAND_LINE
		   , root_filesystem (system_desc, minor_major_root)
#endif
		   );

#ifdef PRINT_COMMAND_LINE
#ifndef ROOT_ON_COMMAND_LINE
  PRINTF (ROOT_X, rootfs);
#endif

  printf (COMMAND_LINE_S, LnxParam.command_line);
#endif /* PRINT_COMMAND_LINE */

  smemcpy (params_farptr, &LnxParam, sizeof(LnxParam));
  }

#ifdef INITRDNAME
/*
 * Select the closest match between initrd* and vmlinuz*
 * at the end of filename, returns the index (zero is
 * none because initrd's are at end of list):
 */
GENERATE_CODE2EXTRA_STUB(initrd_file, unsigned,	unsigned index)

SEGMENT_EXTRA(initrd_file) unsigned initrd_file (unsigned index)
  {
  unsigned short cpt, len, match_len = 0;
  struct desc_str *elem, *vmlin, *match_elem = 0;

  vmlin = &boot_system.desc[index];
  elem = &boot_system.desc[boot_system.nb]; /* first initrd */

  ZDBG (("%s: vm_ptr: %s compared to ", __FUNCTION__,
		&vmlin->filename[vmlin->name_offset]));

  for (cpt = 0; cpt < boot_system.nb_initrd; cpt++, elem++) {
      if (   elem->disk == vmlin->disk
	  && elem->partition == vmlin->partition
	  && elem->in_slashboot == vmlin->in_slashboot) {
	  char *in_ptr = &elem->filename[elem->name_offset],
	       *vm_ptr = &vmlin->filename[vmlin->name_offset];
	  len = 0;
	  ZDBG (("\"%s\"", in_ptr));
	  for (;;) {
	      if (*vm_ptr == 0 && *in_ptr == 0) {
		  len = sizeof (boot_system.desc->filename);
		  break; /* exact match */
		  }
	      if (*vm_ptr != *in_ptr || *vm_ptr == 0)
		  break;
	      len++;
	      vm_ptr++;
	      in_ptr++;
	      }
	  ZDBG (("=%u, ", len));
	  if (match_elem != 0 && len < match_len)
	      continue;
	  match_len = len;
	  match_elem = elem;
	  }
      }
  if (match_elem) {
      ZDBG ((" => best match %s, len %u.\r\n",
		 match_elem->filename, match_len));
      return match_elem - boot_system.desc;
      }
    else {
      ZDBG ((" => no initrd found.\r\n"));
      return 0;
      }
  }
#endif /* INITRDNAME */

/*
 * This is the (only) DOS incompatible function if you are running
 * "boot.exe" (and segment doors closed, i.e. without BIG_MALLOC) -
 * The operating system (if there is one) will stop the execution of
 * the process at least at the "protected_mode()" function call.
 */
SEGMENT_EXTRA(vmlinuz_start) static unsigned
vmlinuz_start (unsigned loadaddress, unsigned kernel_size,
	       farptr params_farptr, unsigned linadr_togpl,
	       unsigned loadadr, unsigned runadr)
  {bound_stack();{
  static const seg_type PMseg[] __attribute__ (( aligned(8) )) = {
    {data: {}}, /* segment 0 is the invalid one */
    {data: {}}, /* segment 0x08 : unused */
    {code: {	/* segment 0x10 : __KERNEL_CS */
    limit: 0xFFFF,
    accessed: 1,        /* 486+: do not update it if it is already set */
    readable: 1,
    cste_11: 3,
    present: 1,
    limit_msb: 0xF,
    deflt: 1,
    granularity: 1	/* limit = 0xFFFFF * 4 Kb = 4 Gb */
    }},
    {data: {	/* segment 0x18 : __KERNEL_DS */
    limit: 0xFFFF,
    accessed: 1,	/* 486+: do not update it if it is already set */
    writable: 1,
    cste_10: 2,
    present: 1,
    limit_msb: 0xF,
    big: 1,
    granularity: 1
    }}
    };
  gdt_idt_param_type gdt = {
      limit: sizeof (PMseg) - 1, /* -1 : see Intel doc */
      base: farptr2linear (data_adr (PMseg))
      };
  unsigned short irqmask;

  if (BOOT1_DOS_running()) {
      unsigned short handle;
      farptr EMM386_entry;
      if (   peekl (0x67*4) != 0
	  && DOS_OpenFile ("EMMXXXX0", DOS_open_read_only, &handle) == 0) {
	  unsigned short errorcode;

	  DOS_CloseFile (handle, &errorcode);
	  if ((EMM386_entry = EMM386_check()) != 0) {
	      ZDBG ((" [switch EMM386 OFF] "));
	      EMM386_set_state (EMM386_entry, 0x01);
	      }
	  }
      DOS_ResetDisk();
      }

#if !(SETUP & USE_INT1587)
  if (BOOT1_DOS_running() && loadaddress != 0x100000)
#endif
      {
      /* HIMEM and INT1587 may enable it only with I/O 0x92 */
      unsigned enable_A20 (unsigned timeout);
      if (enable_A20 (0x10000) != 0)
	  return 1;
      }

  puts (STARTING_KERNEL);
  ZDBG (("Starting kernel at 0x%X\r\n", loadaddress));

#ifdef TREAT_EXCEPTION
  restore_vectors();
#endif
  DBG_END();

  disable_interrupt();
  disable_nmi();
  reset_387();

  irqmask = ((unsigned short)inb (0xA1) << 8) | inb (0x21);

  if (params_farptr == 0x90000000) {
      /* This is no more needed after 2.4 kernel ... */
				/* initialization sequence */
      outb (0x20, 0x11);	/* send it to 8259A-1 */
      outb (0xA0, 0x11);	/* and to 8259A-2+ */
      outb (0x21, 0x20);	/* start of hardware int's (0x20)+ */
      outb (0xA1, 0x28);	/* start of hardware int's 2 (0x28) */
      outb (0x21, 0x04);	/* 8259-1 is master */
      outb (0xA1, 0x02);	/* 8259-2 is slave */
      outb (0x21, 0x01);	/* 8086 mode for both */
      outb (0xA1, 0x01);
      }

  outb (0x21, 0xFB);	/* mask IRQ1, IRQ3..IRQ7 */
  outb (0xA1, 0xFF);	/* mask IRQ8..IRQ15 */
  setfs(getcs());

  if (protected_mode() != 0) {
      outb (0xA1, irqmask >> 8);
      outb (0x21, irqmask);
      enable_interrupt();
      return 2;
      }

  set_gdt (&gdt);

  /* As input also: GDT segment __KERNEL_DS = 0x18 is an
     unlimited R/W data segment starting at zero */
  asm volatile (
"	movl	%%edx,%%ebp	# GCC features				\n"
"	movl	%%eax,%%fs:2f+2	# %%fs == %%cs				\n"
"	movw	$0x18,%%ax						\n"
"	movw	%%ax,%%es						\n"
"									\n"
"	# If DOS=HIGH or HIMEM, the following crash their own memory:	\n"
"	cld								\n"
"	cmp	%%esi,%%edi						\n"
"	jl	1f							\n"
"	lea	(%%edi,%%ecx,4),%%edi					\n"
"	lea	(%%esi,%%ecx,4),%%esi					\n"
"	std								\n"
"	1:								\n"
"	rep movsl %%es:(%%esi),%%es:(%%edi)				\n"
"									\n"
"	mov	%%cr0,%%eax						\n"
"	and	$0x00000001,%%eax					\n"
"	mov	%%eax,%%cr0						\n"
"	xorl	%%eax,%%eax						\n"
"	mov	%%eax,%%cr3						\n"
"	mov	%%eax,%%cr4						\n"
"	pushl	%%eax							\n"
"	popfl								\n"
"	mov	%%eax,%%dr0						\n"
"	mov	%%eax,%%dr1						\n"
"	mov	%%eax,%%dr2						\n"
"	mov	%%eax,%%dr3						\n"
"	mov	%%eax,%%dr4						\n"
"	mov	%%eax,%%dr5						\n"
"	mov	%%eax,%%dr6						\n"
"	mov	%%eax,%%dr7						\n"
"									\n"
"	movl	%%ebx,%%esi	# moved to %%esi			\n"
"	xorl	%%ebx,%%ebx	# Flag to indicate an initial boot	\n"
"	cmpl	%%ebp,%%ebx						\n"
"	jne	1f							\n"
"	xorl	%%eax,%%eax						\n"
"	xorl	%%edi,%%edi						\n"
"	movw	%%cs,%%di						\n"
"	shll	$4,%%edi						\n"
"	.align 4							\n"
"	start_protected:						\n"
"	movl	$start_protected,%%ecx					\n"
"	shrl	$2,%%ecx						\n"
"	rep stosl %%eax,%%es:(%%edi)					\n"
"	addl	$(end_protected - start_protected),%%edi		\n"
"	movw	%%ds,%%cx	# ecx is null				\n"
"	shll	$4,%%ecx						\n"
"	addl	%0,%%ecx						\n"
"	subl	%%edi,%%ecx						\n"
"	shrl	$2,%%ecx						\n"
"	rep stosl %%eax,%%es:(%%edi)					\n"
"	addl	%1,%%edi						\n"
"	movw	%%ss,%%cx	# ecx is null				\n"
"	addw	$0x1000,%%cx						\n"
"	shll	$4,%%ecx						\n"
"	subl	%%edi,%%ecx						\n"
"	shrl	$2,%%ecx						\n"
"	rep stosl %%eax,%%es:(%%edi)					\n"
"	1:								\n"
"	xorl	%%edi,%%edi						\n"
"	2:								\n"
"	ljmpl	$0x10, $0x100000 # self-modifying code: NOT IN ROM!	\n"
"	.align 4							\n"
"	end_protected:							\n"
	: /* no output -:) at least in the current world */
	: "i" (PMseg), "i" (sizeof (PMseg)),
	  "d" (linadr_togpl), "a" (runadr),
	  "S" (loadaddress), "D" (loadadr), "c" ((kernel_size + 3) / 4),
	  "b" (farptr2linear(params_farptr)) // moved to %%esi
	);
  return 0;
  }}

GENERATE_VOID_CODE2EXTRA_STUB(treat_gzip_comment,
	struct system_load_str *loader, const char *ptr)

SEGMENT_EXTRA(treat_gzip_comment) void
treat_gzip_comment (struct system_load_str *loader, const char *ptr)
  {
  static const unsigned char
      comment1[] = "This software is distributed under the GPL license.",
      comment2[] = "This software is not used outside the copyright"
			" owner company.",
      comment3[] = "This software has been authorised in writing"
			" by Etienne LORRAIN.";

  while (*ptr == ' ') ptr++;

  if (memeql (ptr, comment1, sizeof (comment1) - 1)) {
      puts ("");
      puts (comment1);
      ptr += sizeof (comment1) - 1;
      loader->license_compliant = 1;
      }
    else if (memeql (ptr, comment2, sizeof (comment2) - 1)) {
      puts ("");
      puts (comment2);
      ptr += sizeof (comment2) - 1;
      loader->license_compliant = 1;
      }
    else if (memeql (ptr, comment3, sizeof (comment3) - 1)) {
      ptr += sizeof (comment3) - 1;
      loader->license_compliant = 1;
      }
    else
      loader->license_compliant = 0;

  strncpy (loader->cmdline_extraparam, ptr,
		sizeof (loader->cmdline_extraparam));

  for (;;) {
      unsigned index, value = 0;

      while (*ptr == ' ' || *ptr == '\t' || *ptr == '\n' || *ptr == '\r')
	  ptr++;
      if (*ptr == '\0')
	  break;

/* TODO? minRAM=5Mb , minfreq=100MHz, buildroot=/dev/hda5 */

#if 1
      {
      static const char *const kern_param[] = {
/* loadadr=0x00100000 : the load address */
	"loadadr=0x",
/* runadr=0x00100000 : the run address */
	"runadr=0x",
/* if bit (1 << [34567]) set, 80[34567]86 excluded (see vmlinuz.h) */
/* All (i386 compatible) processors: 0x000000F8 */
	"maskcpu=0x",
/* maskflags=0x...  : request those bits set, cpuid 0x00000001 */
/* example TSC required: flagmask=0x00000010 */
	"maskflags=0x",
/* maskextra=0x... : request those bits set, cpuid 0x80000001 */
	"maskextra=0x",
/* Which video mode is managed by the kernel (see vmlinuz.h) */
	"maskvesa=0x",
/* Which video resolution is refused by the kernel (see vmlinuz.h) */
	"maskresolution=0x"
	};
      for (index = 0; index < nbof (kern_param); index++) {
	  unsigned len = strlen (kern_param[index]);
	  if (memeql (ptr, kern_param[index], len)) {
	      ptr += len;
	      break;
	      }
	  }
      if (index >= nbof (kern_param)) {
	  /* Ignore that */
	  while (*ptr != '\0' && *ptr != ' '  && *ptr != '\t'
			      && *ptr != '\n' && *ptr != '\r')
	      ptr++;
	  continue;
	  }
      }
#else
      if (STRBEGINEQL (ptr, "loadadr=0x")) {
	  index = 0;
	  /* loadadr=0x00100000 : the load address */
	  ptr += sizeof ("loadadr=0x") - 1;
	  }
      if (STRBEGINEQL (ptr, "runadr=0x")) {
	  index = 1;
	  /* runadr=0x00100000 : the run address */
	  ptr += sizeof ("runadr=0x") - 1;
	  }
	else if (STRBEGINEQL (ptr, "maskcpu=0x")) {
	  index = 2;
	  /* if bit (1 << [34567]) set, 80[34567]86 excluded (see vmlinuz.h) */
		/* All (i386 compatible) processors: 0x000000F8 */
	  ptr += sizeof ("maskcpu=0x") - 1;
	  }
	else if (STRBEGINEQL (ptr, "maskflags=0x")) {
	  index = 3;
	  /* maskflags=0x...  : request those bits set, cpuid 0x00000001 */
		/* example TSC required: flagmask=0x00000010 */
	  ptr += sizeof ("maskflags=0x") - 1;
	  }
	else if (STRBEGINEQL (ptr, "maskextra=0x")) {
	  index = 4;
	  /* maskextra=0x... : request those bits set, cpuid 0x80000001 */
	  ptr += sizeof ("maskextra=0x") - 1;
	  }
	else if (STRBEGINEQL (ptr, "maskvesa=0x")) {
	  index = 5;
	  /* Which video mode is managed by the kernel */
	  ptr += sizeof ("maskvesa=0x") - 1;
	  }
	else if (STRBEGINEQL (ptr, "maskresolution=0x")) {
	  index = 6;
	  /* Which video resolution is refused by the kernel */
	  ptr += sizeof ("maskresolution=0x") - 1;
	  }
	else {
	  /* Ignore that */
	  while (*ptr != '\0' && *ptr != ' '  && *ptr != '\t'
			      && *ptr != '\n' && *ptr != '\r')
	      ptr++;
	  continue;
	  }
#endif
      for (;;) {
	  if (*ptr >= '0' && *ptr <= '9') {
	      value <<= 4;
	      value += *ptr++ - '0';
	      }
	    else if (*ptr >= 'A' && *ptr <= 'F') {
	      value <<= 4;
	      value += 10 + *ptr++ - 'A';
	      }
	    else if (*ptr >= 'a' && *ptr <= 'f') {
	      value <<= 4;
	      value += 10 + *ptr++ - 'a';
	      }
	    else
	      break;
	  }
      loader->comment.array[index] = value;
      }
  }

GENERATE_VOID_CODE2EXTRA_STUB(treat_gzip_name,
	struct system_load_str *loader, const char *ptr)

SEGMENT_EXTRA(treat_gzip_name) void
treat_gzip_name (struct system_load_str *loader, const char *ptr)
  {
  }

#if defined (LINUZNAME1) || defined (LINUZNAME2) || defined (LINUZNAME3)

SEGMENT_EXTRA(vmlinuz_first_params) static inline unsigned
vmlinuz_first_params (struct system_load_str *loader,
		      struct linux_param *param,
		      unsigned firstblocklen)
  {
  char *version_string;
  /* the loaded file is at least bigger than firstblocklen: */
  if (firstblocklen <= offsetof(struct linux_param, command_line)) {
      ZDBG ((" [vmlinuz filesize too small: firstblocklen %u, min %u] ",
		firstblocklen, offsetof(struct linux_param, command_line)));
      return 1;
      }

  if (param->in_out.header_signature.longword != header_signature.longword) {
      ZDBG ((" [%s: bad vmlinuz signature 0x%X] ",
		__FUNCTION__,
		param->in_out.header_signature.longword));
      return 2;
      }

  ZDBG ((" [setup_sects=%u, minor_root=0x%X, major_root=0x%X, "
		"kernel_version: %s] ",
	 param->in_out.setup_sects,
	 param->in_out.root.fields.minor_root,
	 param->in_out.root.fields.major_root,
	 (char *)param + 512 + param->in_out.kernel_version));

  version_string = (char *)param + 512 + param->in_out.kernel_version;
  ZDBG (("\r\nLoading kernel: %s", version_string));

  if (   version_string[0] >= '2'
      && version_string[1] == '.'
      && version_string[2] >= '4'
      && version_string[3] == '.'
      && version_string[4] >= '1') {
      /* added in between 2.4.0-test1 and 2.4.0-test10 */
      loader->params_farptr = 0x80000000;
      }
    else
      loader->params_farptr = 0x90000000;
  ZDBG ((", Linux param at 0x%X.\r\n", loader->params_farptr));

  if (   param->in_out.code32_start != 0x100000
      && param->in_out.code32_start != 0x1000) {
      ZDBG ((" [code32_start != 0x100000, I did not plan that entry value:"
		" 0x%X:0x%X] ",
		param->in_out.start_sys_seg, param->in_out.code32_start));
      return 3;
      }
  loader->minor_major_root = param->in_out.root.minor_major;

  return 0;
  }

GENERATE_CODE2EXTRA_STUB(vmlinuz_header_treat, unsigned,
	struct system_load_str *loader,
	unsigned char *buffer,
	unsigned buflen)

SEGMENT_EXTRA(vmlinuz_header_treat) unsigned
vmlinuz_header_treat (struct system_load_str *loader,
			unsigned char *buffer,
			unsigned buflen)
  {bound_stack();{
  if (loader->gzip_start) { /* no vmlinuz header */
      loader->gzip_start = buffer;
      return 0;
      }
  if (loader->treat_start_bytes == 0) { /* first call */
      struct linux_param *param = (struct linux_param *)buffer;
      unsigned setup_sects = param->in_out.setup_sects, tmp;

      /* hopes that buflen is sufficient: */
      if ((tmp = vmlinuz_first_params (loader, param, buflen)) != 0)
	  return tmp;

      setup_sects += 1; /* bootsector */
      loader->treat_start_bytes = (-512 * setup_sects) - 1; /* finish at -1 */
      ZDBG ((" [treat_start_bytes = %d] ", loader->treat_start_bytes));
      }
  if (loader->treat_start_bytes < 0) {
      if (-loader->treat_start_bytes >= (int)buflen)
	  loader->treat_start_bytes += buflen;
	else {
	  /*
	   * look for assembly instruction "lss SYMBOL_NAME(start_stack),%esp"
	   * in the 50 first protected mode instruction bytes:
	   *   (see linux/arch/i386/boot/compressed/head.S)
	   */
	  unsigned char *ptr, *end;
	  unsigned addr;

	  loader->treat_start_bytes++; /* the -1 to finnish at -1 */
	  loader->treat_start_bytes = -loader->treat_start_bytes; /* stage 2 */
	  ptr = buffer + loader->treat_start_bytes;
	  end = ptr + 50;
	  while (ptr < end) {
	      if (ptr[0] == 0x0F && ptr[1] == 0xB2 && ptr[2] == 0x25)
		  break;
	      ptr++;
	      }
	  /* get symbol value, remove 0x100000 because of the link address,
		if it was compiled to be loaded high, or remove 0x1000.  */
	  addr = (*(unsigned *)&ptr[3]);
	  if (addr & 0x100000)
	      addr -= 0x100000;
	    else
	      addr -= 0x1000;

	  loader->treat_start_bytes += addr;
	  if (ptr < end && loader->treat_start_bytes < 30 * 1024) {
	      ZDBG ((" [gzip header in %u bytes] ", loader->treat_start_bytes));
	      }
	    else {
	      if (ptr < end)
		  ZDBG ((" [gzip header offset too big] "));
		else
		  ZDBG ((" [did not find \"lss start_stack,%esp\"] "));
	      return 0x0F;
	      }
	  }
      }
  if (loader->treat_start_bytes > (int)buflen) {
      loader->treat_start_bytes -= buflen;
      return 0;	/* need more bytes */
      }
    else if (loader->treat_start_bytes > 0) {
      /*
       * Search the 0x8B, 0x1F signature - knowing that different linker
       * versions will align or not align section .data
       * (see linux/arch/i386/boot/compressed/Makefile, target piggy.o).
       * The first 4-bytes integer will be the length of the compressed
       * file - but that is not really interresting now...
       */
      unsigned char *ptr = buffer + loader->treat_start_bytes,
		    *end = ptr + 50;
      while (ptr < end) {
	  if (ptr[0] == 0x1F && ptr[1] == 0x8B)
	      break;
	  ptr++;
	  }
      if (ptr < end) {
	  ZDBG ((" [gzip signature: 0x%X] ", *(unsigned short *) ptr));
	  loader->gzip_start = ptr;
	  return 0; /* normal end, now treat gzlib */
	  }
	else {
	  ZDBG ((" [gzip signature not found] "));
	  return 0x0E;
	  }
      }
    else
      return 0; /* GCC happy */
  }}

#endif /* LINUZNAME* */

/**
 ** System loading:
 **/
GENERATE_VOID_CODE2EXTRA_STUB(menu_load_system,
	unsigned system_index,
	unsigned extendmem,
	unsigned freemem,
	farptr   extra_cmdline)

SEGMENT_EXTRA(menu_load_system) void
menu_load_system (unsigned system_index,
		  unsigned extendmem,
		  unsigned freemem,
		  farptr   extra_cmdline,
		  farptr   instboot_info,
		  unsigned short DefaultTextVideoMode,
		  struct gujin_param_attrib gujin_attr)
  {
  unsigned returned;
  struct system_load_str  kloader, iloader;
  struct gpl_compliant_str togpl = {kloader: &kloader, iloader: &iloader};
  const farptr XMSentry = BOOT1_DOS_running() ? _XMS_get_entry_point () : 0;

  if (XMSentry) {
      unsigned char error;

      kloader.XMSentry = iloader.XMSentry = XMSentry;
      ZDBG ((" [HIMEM: present, "));
#if 0 /* A20 seems to be enabled even before, so no need for that: */
      if (_XMS_query_A20 (XMSentry, &error))
	  ZDBG (("[init: A20 enabled, error 0x%X] ", error));
	else
	  ZDBG (("[init: A20 disabled, error 0x%X] ", error));

      if (_XMS_global_enable_A20 (XMSentry, &error)) {
	  ZDBG (("cannot disable A20-HMA, error 0x%X!] ", error));
	  print (ERROR_XMS_ENABLE_A20);
	  return;
	  }

      if (_XMS_local_enable_A20 (XMSentry, &error)) {
	  ZDBG (("cannot enable A20, error 0x%X!] ", error));
	  print (ERROR_XMS_ENABLE_A20);
	  return;
	  }
#endif
      kloader.KBalloced = 512;	/* initial kernel load size */
      kloader.max_size = freemem * 1024;
      kloader.load_address = 0;	/* offset in handle */
      kloader.handle = _XMS_allocate_memory (XMSentry,
					     kloader.KBalloced,
					     &error);
      if (kloader.handle == 0) {
	  ZDBG (("cannot allocate %u Kb for kernel, error 0x%X!] ",
		 kloader.KBalloced, error));
#if !(SETUP & BIOS_ONLY) /* unused string removal bug of GCC */
	  printf (ERROR_XMS_ALLOCATE, error);
#endif
	  return;
	  }
      ZDBG (("unsing handle %u] ", kloader.handle));
      }
    else {
      kloader.XMSentry = iloader.XMSentry = 0;
      ZDBG ((" [do not allocate HIMEM memory] "));
#if (ASSEMBLY_TYPE != ASSEMBLY_DSES) && !(SETUP & USE_INT1587)
      if (segment_door (1) != 0) {
	  ZDBG ((" [cannot open doors, aborting] "));
	  printf (ERROR_SEGMENT_DOOR);
	  return;
	  }
#endif
      kloader.max_size = extendmem * 1024;
      kloader.load_address = 0x100000;
      }

#if defined (LINUZNAME1) || defined (LINUZNAME2) || defined (LINUZNAME3)
  kloader.treat_start_bytes = 0;
  if (boot_system.desc[system_index].has_vmlinuz_header)
      kloader.gzip_start = (void *) 0;
    else {
      kloader.gzip_start = (void *) 1;
      kloader.params_farptr = 0x80000000;
      }
#else
  kloader.params_farptr = 0x80000000;
#endif
  kloader.comment.elem.late_relocation_address = 0x100000;
  kloader.comment.elem.runadr = 0x100000;
  kloader.cmdline_extraparam[0] = '\0';

  ZDBG (("%s: loading kernel %s, max size %u, at 0x%X, ",
	__FUNCTION__,
	boot_system.desc[system_index].filename,
	kloader.max_size,
	kloader.load_address));

  printf (LOADING_KERNEL_S, boot_system.desc[system_index].filename);
  returned = system_file_load (system_index, &kloader);

  ZDBG (("result: 0x%X (size: %u compressed, %u uncompressed bytes, "
	 "minor_major_root: 0x%X) (reloc adr 0x%X, run adr 0x%X).\r\n",
	 returned,
	 kloader.compressed_size,
	 kloader.uncompressed_size,
	 kloader.minor_major_root,
	 kloader.comment.elem.late_relocation_address,
	 kloader.comment.elem.runadr));

  /* TODO: relocation & initrd size & position */
  if ((kloader.comment.elem.late_relocation_address & ~0xC0000000)
	+ kloader.uncompressed_size > 1024 * extendmem) {
      ZDBG (("Will not be able to relocate %u Kbytes at address 0x%X "
		"because only %u Kbytes memory available!\r\n",
		kloader.uncompressed_size / 1024,
		kloader.comment.elem.late_relocation_address,
		extendmem));
      returned = 0xFF;
      }

  if (XMSentry) {
      unsigned char error;

      if (returned != 0) {
	  ZDBG (("FAILED Loading kernel error 0x%X\r\n", returned));
#if !(SETUP & BIOS_ONLY) /* unused string removal bug of GCC */
	  printf (ERROR_LOADING_FILE, returned);
#endif
	  goto free_kernel_memory;
	  }
      kloader.load_address = _XMS_lock (XMSentry, kloader.handle, &error);
      if (kloader.load_address == 0) {
	  ZDBG (("FAILED Lock kernel error 0x%X\r\n", error));
#if !(SETUP & BIOS_ONLY) /* unused string removal bug of GCC */
	  printf (ERROR_XMS_LOCK, error);
#endif
	  goto free_kernel_memory;
	  }
      ZDBG (("Locked at 0x%X, error 0x%X\r\n", kloader.load_address, error));
      }
    else {
      if (returned != 0) {
	  ZDBG (("FAILED Loading kernel error 0x%X\r\n", returned));
	  printf (ERROR_LOADING_FILE, returned);
	  return;
	  }
      }

#ifdef INITRDNAME
  if ((togpl.initrd_index = initrd_file (system_index)) != 0) {
      if (XMSentry) {
	  unsigned char error;

	  iloader.KBalloced = 64;	/* initial initrd load size */
	  iloader.max_size = (freemem - kloader.KBalloced) * 1024;
	  iloader.handle = _XMS_allocate_memory (XMSentry,
						iloader.KBalloced,
						&error);
	  if (iloader.handle == 0) {
	      ZDBG (("cannot allocate %u Kb for initrd, error 0x%X!] ",
			iloader.KBalloced, error));
#if !(SETUP & BIOS_ONLY) /* unused string removal bug of GCC */
	      printf (ERROR_XMS_ALLOCATE, error);
#endif
	      goto unlock_kernel_memory;
	      }
	  }
	else {
	  /* Load the initrd just after kernel, at a round addr: */
	  /* Seems that it is assumed there is > 1Mb free memory
		over the last kernel address by vmlinuz: */
	  iloader.load_address = ROUNDUP(kloader.load_address
					 + kloader.uncompressed_size
					 + 0x100000,
					 0x100000);
	  if (iloader.load_address + 64 * 1024 /* min size of initrd */
		> 1024 * extendmem) {
	      print (ERROR_NOT_ENOUGH_RAM);
	      return;
	      }
	  iloader.max_size = 1024 * extendmem - iloader.load_address;
	  }

#if defined (LINUZNAME1) || defined (LINUZNAME2) || defined (LINUZNAME3)
      iloader.gzip_start = (void *) 1; /* no vmlinuz header */
#endif
      ZDBG (("Loading initial RAMdisk \"%s\" at 0x%X, max size %u, ",
		boot_system.desc[togpl.initrd_index].filename,
		iloader.load_address, iloader.max_size));

      printf (LOADING_INITRD_S, boot_system.desc[togpl.initrd_index].filename);
      returned = system_file_load (togpl.initrd_index, &iloader);

      ZDBG (("result: 0x%X (size: %u compressed, %u uncompressed bytes).\r\n",
		returned, iloader.compressed_size, iloader.uncompressed_size));

      if (XMSentry) {
	  unsigned char error;

	  if (returned != 0) {
	      ZDBG (("FAILED loading initrd error 0x%X\r\n", returned));
#if !(SETUP & BIOS_ONLY) /* unused string removal bug of GCC */
	      printf (ERROR_LOADING_FILE, returned);
#endif
	      goto free_initrd_memory;
	      }
	  iloader.load_address = _XMS_lock (XMSentry, iloader.handle, &error);
	  if (iloader.load_address == 0) {
	      ZDBG (("FAILED Locking initrd error 0x%X, ", error));
#if !(SETUP & BIOS_ONLY) /* unused string removal bug of GCC */
	      printf (ERROR_XMS_LOCK, error);
#endif
	      goto free_initrd_memory;
	      }
	  ZDBG (("Locked at 0x%X, error 0x%X, ",
		 iloader.load_address, error));
	  }
	else if (returned != 0) {
	  ZDBG (("FAILED loading initrd\r\n"));
	  printf (ERROR_LOADING_FILE, returned);
	  return;
	  }
      }
    else
#endif /* INITRDNAME */
      {
      togpl.initrd_index = 0;
      /* no initrd to linux_set_params() */
      iloader.load_address = 0;
      iloader.uncompressed_size = 0;
      }

#if USER_SUPPORT != 0
  {
  /* init all of them only for warnings: */
  unsigned short current_mode = 0;
  unsigned fgcolor = 0, bgcolor = 0;
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
  unsigned char in_vesa = 0;
#endif

  if (kloader.license_compliant) {
      togpl.signature = 0x01031967;
      togpl.version = 0x100;
      togpl.feature = 1;
      togpl.instboot = instboot_info;
      togpl.UI = farptr2linear(data_adr(&UI));
      togpl.DI = farptr2linear(data_adr(&DI));
      togpl.boot_system = farptr2linear(data_adr(&boot_system));
      }

  if (BOOT1_COM_port() < 0) {
      current_mode = UI_getmode() & ~0x8000;
      fgcolor = UI.fgcolor;
      bgcolor = UI.bgcolor;
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
      in_vesa = VESA_ACTIVE();
#endif
      /*
       * Arggg... I do not know if this kernel has been compiled with
       * framebuffer enabled, and which kind of framebuffer (VESA, ATI...)
       * Starting a kernel in a VESA graphic mode when VESA framebuffer
       * is not compiled-in result in a kernel booting correctly but
       * the display is very strange...
       * I let the user decide on the configuration if it is safe
       * to run the kernel in graphic (VESA framebuffer stuff
       * for the current mode/BPP).
       */
      if (UI.parameter.attr.isgraphic && gujin_attr.force_textmode) {
	  unsigned short tmp = DefaultTextVideoMode;
	  unsigned char cpt;

	  if (tmp != 0xFFFF) {
	      ZDBG (("Searching default text mode 0x%X\r\n", tmp));
	      for (cpt = 0; cpt < UI.nbmode; cpt++)
		  if (((UI.mode[cpt].number ^ tmp) & 0x3FF) == 0)
		      break;
	      if (!UI.mode[cpt].text)
		  ZDBG (("Default text mode 0x%X is NOT text!\r\n", tmp));
	      }
	  if (tmp == 0xFFFF || cpt >= UI.nbmode) {
	      ZDBG (("Searching a text mode\r\n"));
	      for (cpt = 0; cpt < UI.nbmode; cpt++)
		  if (UI.mode[cpt].text)
		      break;
	      }
	  if (cpt >= UI.nbmode) {
	      ZDBG (("did not find any text mode, setting mode 3.\r\n"));
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
	      if (in_vesa) {
		  if (VGA_init())
		      ZDBG (("Panic: cannot switch to VGA!\r\n"));
		  }
#endif
	      if (UI_setmode (3 | 0x8000))
		  ZDBG (("Panic: cannot switch to text mode 3!\r\n",
				UI.mode[cpt].number | 0x8000));
	      }
	    else {
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
	      if (UI.mode[cpt].vesa) {
		  if (!in_vesa) {
		      if (VESA_init())
			  ZDBG (("Panic: cannot switch to VESA!\r\n"));
		      }
		  }
		else {
		  if (in_vesa) {
		      if (VGA_init())
			  ZDBG (("Panic: cannot switch to VGA!\r\n"));
		      }
		  }
#endif
	      if (UI_setmode (UI.mode[cpt].number | 0x8000))
		  ZDBG (("Panic: cannot switch to text mode 0x%X!\r\n",
				UI.mode[cpt].number | 0x8000));
	      }
	  UI_setfgcolor (UI.stdcolor[lightgray]);
	  UI_setbgcolor (UI.stdcolor[black]);
	  UI_clearscreen();
	  PUTS (SWITCHED_TO_TEXT_MODE);
	  }
	else {
	  unsigned char row, col, cpt;

	  UI_setfgcolor (UI.stdcolor[lightgray]);
	  UI_setbgcolor (UI.stdcolor[black]);
	  UI_getcursor (&row, &col);
	  for (cpt = row; cpt < UI.parameter.nbrow; cpt++)
	      puts (""); /* scroll once */
	  UI_setcursor (row - 1, col);
	  }
      }
  /* _no_ brace close */
#endif

  linux_set_params (&boot_system.desc[system_index],
		    kloader.params_farptr,
		    kloader.minor_major_root,
		    extendmem,
		    iloader.load_address,
		    iloader.uncompressed_size,
		    kloader.cmdline_extraparam,
		    extra_cmdline);

  if (BOOT1_DOS_running()) {
      ZDBG ((" [DISK cache "));
      if (SMARTDRV_control(0, SMARTDRV_get_info).abscent == 0) {
	  SMARTDRV_flush_buffer();
	  ZDBG ((" flushed] "));
	  }
	else
	  ZDBG ((" abscent] "));
      }

  if ((returned = setjmp (&togpl.jmpbuf)) == 0) {
      returned = vmlinuz_start (kloader.load_address,
				kloader.uncompressed_size,
				kloader.params_farptr,
				(kloader.license_compliant)
				    ? farptr2linear(stack_adr(&togpl)) : 0,
				kloader.comment.elem.late_relocation_address,
				kloader.comment.elem.runadr
				);
      }

  /* If we come back here, there is an error... */

#if USER_SUPPORT != 0
  if (BOOT1_COM_port() < 0) {
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
      if (in_vesa && !VESA_ACTIVE())
	  VESA_init();
#endif
      if (current_mode != (UI_getmode() & ~0x8000))
	  UI_setmode (current_mode);
      UI_setfgcolor (fgcolor);
      UI_setbgcolor (bgcolor);
      }
  }  /* the long waited closing brace */
#endif

  if (returned == 1)
      puts (ERROR_STARTUP_ENABLE_A20);
    else if (returned == 2)
      puts (ERROR_SWITCH_PROTECTED_MODE);
  puts ("");

  if (XMSentry) {
      unsigned char error;

#ifdef INITRDNAME
      if (iloader.handle) {
	  /* there was an initrd */
	  if (_XMS_unlock(XMSentry, iloader.handle, &error))
	      ZDBG ((" [FAILED unlock initrd handle 0x%X: 0x%X] ",
			iloader.handle, error));
free_initrd_memory:
	  if (_XMS_free_memory(XMSentry, iloader.handle, &error))
	      ZDBG ((" [FAILED free initrd handle 0x%X: 0x%X] ",
			iloader.handle, error));
	  }

unlock_kernel_memory:
      if (_XMS_unlock(XMSentry, kloader.handle, &error))
	  ZDBG ((" [FAILED unlock kernel handle 0x%X: 0x%X] ",
			kloader.handle, error));
#endif /* INITRDNAME */

free_kernel_memory:
      if (_XMS_free_memory(XMSentry, kloader.handle, &error))
	  ZDBG ((" [FAILED free kernel handle 0x%X: 0x%X] ",
			kloader.handle, error));
      }
  }

