/* user.c */

/* Gujin is a bootloader, it loads a Linux kernel from cold boot or DOS.
 * Copyright (C) 1999-2001 Etienne Lorrain, fingerprint (2D3AF3EA):
 *   2471 DF64 9DEE 41D4 C8DB 9667 E448 FF8C 2D3A F3EA
 * E-Mail: etienne.lorrain@masroudeau.com
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License,
 * version 2, as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "make.h"

#if USER_SUPPORT != 0

#include "library.h"
#include "libextra.h"	/* kbd_use_ext_fct() */
#include "boot.h"
#include "user.h"
#include "gmem.h"
#include "bios.h"
#include "debug.h"
#include "malloc.h"
#include "dos.h"	/* DOS_[GS]etInterruptVector() */
#include "messages.h"	/* VIDEO_CARD_CHANGED */
#include "vgabios.h"	/* _VGA_get_display_combination_code, monitor freq */

#if USER_SUPPORT & VESA_SUPPORT
#include "vesabios.h"
#endif

#if USER_SUPPORT & (VESA_4BPP_TEXT | VESA_4BPP_EGA | VESA_8BPP | VESA_RECALC)
#include "ega.h"
#endif

/*
 * Few basic macros:
 */
#define ATTR2BYTE(A)	(*(unsigned char *)&(A))

#if USER_SUPPORT & VESA_4BPP_EGA

#if EGA_DEFAULT_READ_MODE == 0
#define EGA_READ_PLANE(plane)	_EGA_select_read_plane (plane)
#define EGA_READ_INIT()	/* */
#elif EGA_DEFAULT_READ_MODE == 1
#define EGA_READ_PLANE(plane)	_EGA_select_read_ignore (1 << plane)
#define EGA_READ_INIT()	do { \
    _EGA_read_compare (0x0F);						\
    _EGA_select_read_ignore (0);	/* ignore if bit cleared */	\
    } while (0)
#else
#error EGA_DEFAULT_READ_MODE should be defined to 0 or 1.
#endif

#if USER_SUPPORT & VESA_2WINDOWS
#define GXCHGB(xy, color) do { \
    volatile unsigned char dummyread;                   \
    dummyread = gpeekb (VESA_readpixeloffset (xy));     \
    gpokeb (VESA_pixeloffset (xy), color);              \
    } while (0)
#else
#define GXCHGB(xy, color) \
    gxchgb (VESA_pixeloffset (xy), color)
#endif

#endif /* VESA_4BPP_EGA */

#if USER_SUPPORT & VESA_STRANGE_WINDOW
#define DIV_WINDOW_GRANUL(off) \
	((off) / UI.parameter.wingranul)
#define ROUND_WINDOW_GRANUL(off) \
	((off) / UI.parameter.wingranul * UI.parameter.wingranul)
#define DISPLAY_WINDOW_GRANUL(str) \
	((str)->wingranul)
#define INIT_WINDOW_GRANUL(size, param) \
	param->wingranul = (size)
#else
#define DIV_WINDOW_GRANUL(off)  \
	((off) >> UI.parameter.wingranul)
#define ROUND_WINDOW_GRANUL(off) \
	((off) & (0xFFFFFFFFU << UI.parameter.wingranul))
#define DISPLAY_WINDOW_GRANUL(str) \
	(1U << (str)->wingranul)
#define INIT_WINDOW_GRANUL(size, param) \
	param->wingranul = (my_ffs (size) - 1)
#endif /* VESA_STRANGE_WINDOW */

/*
 * This was a try, it does not work, so undef it:
 */
#undef VGA16_XOR_TRICK

/*
 * The default "common library" functions:
 */
static unsigned
COMMON_setfgcolor (unsigned color)
  {
  if (   color > UI.parameter.nbcolor
      || UI.parameter.nbcolor == 0) {
      VDBG ((" [%s: color too high: %u] ", __FUNCTION__, color));
      return 1;
      }
  UI.fgcolor = color;
  return 0;
  }

static unsigned
COMMON_setbgcolor (unsigned color)
  {
  if (   color > UI.parameter.nbcolor
      || UI.parameter.nbcolor == 0) {
      VDBG ((" [%s: color too high: %u] ", __FUNCTION__, color));
      return 1;
      }
  UI.bgcolor = color;
  return 0;
  }

static unsigned
COMMON_error (void)
  {
  DBG (("ERROR: %s called.\r\n", __FUNCTION__));
  return 1;
  }

#if USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT)
static unsigned
COMMON_setcursor (unsigned char row, unsigned char col)
  {
  if (row > UI.parameter.nbrow) {
      VDBG ((" [%s to row %u impossible, max = %u] ",
	    __FUNCTION__, row, UI.parameter.nbrow));
      return 1;
      }

  if (col > UI.parameter.nbcol) {
      VDBG ((" [%s to col %u impossible, max = %u] ",
	    __FUNCTION__, col, UI.parameter.nbcol));
      return 1;
      }

  UI.parameter.row = row;
  UI.parameter.col = col;
  return 0;
  }

#if (USER_SUPPORT & VESA_SUPPORT) || !(USER_SUPPORT & VGA_USEINT1013)
static unsigned
COMMON_getcursor (unsigned char *row, unsigned char *col)
  {
  checkptr (row, "COMMON_getcursor invalid 1st parameter.");
  checkptr (col, "COMMON_getcursor invalid 2nd parameter.");

  // could use EGA registers, but _we_ are maintaining the cursor
  // position.
  *row = UI.parameter.row;
  *col = UI.parameter.col;
  return 0;
  }
#endif /* VESA_SUPPORT || !VGA_USEINT1013 */
#endif /* VGA_SUPPORT || VESA_SUPPORT */

static void add_valid_mode (struct videomode_str *mode)
  {bound_stack();{
  unsigned short i;

  if (UI.mode == 0 && UI.nbmode != 0)
      VDBG (("ERROR: UI.mode == 0 with UI.nbmode = %u\r\n", UI.nbmode));

  for (i = 0; i < UI.nbmode; i++) {
      if (UI.mode[i].number == mode->number) {
	  if (UI.mode[i].vesa && !mode->vesa)
	      return;
	  VDBG ((" [replace in UI.mode index %u with mode 0x%X] ",
		 i, mode->number));
	  UI.mode[i] = *mode;
	  return;
	  }
      }

  /* get memory for new mode, realloc can be called with 1st param == 0 */
  {
  struct videomode_str *tmp;

  tmp = realloc (UI.mode, (UI.nbmode+1) * sizeof (struct videomode_str));
  if (tmp == 0) {
      VDBG ((" [%s failed realloc] ", __FUNCTION__));
      return;
      }
  UI.mode = tmp;
  }

  if (!mode->vesa) {
      /* VGA modes: add at end */
      UI.mode[UI.nbmode] = *mode;
      }
    else {
      unsigned short j;

      /* VESA modes: add at beginning, but keep same order */
      for (i = 0; i < UI.nbmode && UI.mode[i].vesa; i++) ;

      /* copy from high to low memory, no memcpy! */
      for (j = UI.nbmode; j > i; j--)
	  UI.mode[j] = UI.mode[j-1];

      UI.mode[i] = *mode;
      }
  UI.nbmode++;
  }}

static void remove_valid_mode (unsigned short number)
  {bound_stack();{
  unsigned short i;

  VDBG ((" [remove_valid_mode mode 0x%X", number));
  for (i = 0; i < UI.nbmode; i++) {
      if (UI.mode[i].number == number) {
	  while (i < UI.nbmode - 1) {
	      UI.mode[i] = UI.mode[i+1];
	      i++;
	      }
	  UI.nbmode--;
	  VDBG ((" done] "));
	  return;
	  }
      }
  VDBG ((" not found] "));
  }}

/* This is called in the menu */
void reset_low_valid_mode (unsigned up_to)
  {bound_stack();{
  unsigned short i;

  for (i = 0; i < up_to; i++)
      remove_valid_mode (i);
  }}

#if USER_SUPPORT & VESA_HARDWINDOW
static void free_hardwin (void)
  {
  if (UI.parameter.hardwindow != 0) {
      free (UI.parameter.hardwindow);
      UI.parameter.hardwindow = 0;
      }
  if (UI.parameter.offsetarray != 0) {
      free (UI.parameter.offsetarray);
      UI.parameter.offsetarray = 0;
      }
#if USER_SUPPORT & VESA_2WINDOWS
  if (UI.parameter.readhardwindow != 0) {
      free (UI.parameter.readhardwindow);
      UI.parameter.readhardwindow = 0;
      }
  if (UI.parameter.readoffsetarray != 0) {
      free (UI.parameter.readoffsetarray);
      UI.parameter.readoffsetarray = 0;
      }
#endif /* VESA_2WINDOWS */
  }

static unsigned
set_hardwin (struct hardwindow_str *hardwin,
		unsigned short *offsetarray,
		unsigned offset)
  {
  unsigned short page = DIV_WINDOW_GRANUL(offset);
  unsigned short enabled;

  if (hardwin == 0)
      return 1;

  if (offsetarray != 0)
      hardwin = &hardwin[offsetarray[page]];

  enabled = interrupt_enabled();
  if (enabled)
      disable_interrupt ();

  while (hardwin->instruction != 0) {
      if (hardwin->instruction == 0xEF66)
	  outl (hardwin->port, hardwin->data);
	else {
	  unsigned short data;
	  unsigned char comp = hardwin->instruction >> 8,
			inst = hardwin->instruction & 0xFF;

	  if (inst != 0xEE && inst != 0xEF) {
	      if (enabled)
		  enable_interrupt();
	      VDBG ((" [%s invalid instruction!] ", __FUNCTION__));
	      return 2;
	      }

	  if (comp == 0)
	      data = hardwin->data;
#ifdef COMPRESS_HARDWIN
	    else if (   comp == 1  || comp == 3  || comp == 19
		     || comp == 67 || comp == 20 || comp == 28
#ifdef COMPRESS_EXTENDED
		     || comp == 37 || comp == 11
#endif /* COMPRESS_EXTENDED */
		     ) {
	      unsigned short pg = page;
	      unsigned tmp = hardwin->data;

#ifdef COMPRESS_EXTENDED
	      if (inst == 0xEE) {
		  if (comp == 37) {
		      tmp ^= inb (hardwin->port) & 0xF0;
		      comp = 1;
		      }
		  if (comp == 11) {
		      tmp ^= inb (hardwin->port) & 0x0F;
		      comp = 3;
		      }
		  }
#endif /* COMPRESS_EXTENDED */

	      if (comp == 1 || comp == 3)
		  pg &= 0x0F;

	      if (comp == 3)
		  pg <<= 4;
		else if (comp == 28)
		  pg <<= 2;
		else if (comp == 67)
		  pg = (pg & 0xF0) >> 2;
		else if (comp == 20)
		  pg >>= 4;

	      if (inst == 0xEE)
		  data = tmp ^ pg;
		else
		  data = tmp ^ (pg << 8);
	      }
#endif  /* COMPRESS_HARDWIN */
	    else {
	      if (enabled)
		  enable_interrupt();
	      VDBG ((" [%s invalid byte!] ", __FUNCTION__));
	      return 3;
	      }

	  if (inst == 0xEE)
	      outb (hardwin->port, data);
	    else
	      outw (hardwin->port, data);
	  }
      hardwin++;
      }
  if (enabled)
      enable_interrupt();
  return 0;
  }
#endif /* VESA_HARDWINDOW */

/*
 * The default functions, at startup:
 */
static unsigned
DEFLT_setfgcolor (unsigned color)
  {
  if (COMMON_setfgcolor (color))
      return 1;
  BOOT1_setcolor (color);	/* ONLY graphic BIOS modes */
  return 0;
  }

/*
 * The "common VESA/VGA BIOS library" functions:
 */
static unsigned BIOS_getkey (unsigned timeout)
  {bound_stack();{
  unsigned returned;

  if (timeout)
      timeout += _BIOS_nbtick();

  if (UI.mouse.start)
      UI.mouse.start();
  for (;;) {
#if BUILD_APPLICATION()
      if (kbd_use_ext_fct()) {
	  if (_BIOS_extkbd_checkkey()) {
	      returned = _BIOS_extkbd_getkey ();
	      break;
	      }
	  }
	else
#endif
	  {
	  if (_BIOS_checkkey()) {
	      returned = _BIOS_getkey ();
	      break;
	      }
	  }
      if (timeout && (_BIOS_nbtick() >= timeout)) {
	  returned = TIMEOUT_KEYCODE();
	  break;
	  }
      if (UI.mouse.poll != 0) {
	  returned = UI.mouse.poll();
	  if (returned != 0xFFFFFFFF)
	      break;
	  }
      }
  if (UI.mouse.stop)
      UI.mouse.stop();

//VDBG ((" [%s : 0x%X] ", __FUNCTION__, returned));

  if (   (returned & 0xFF) == 3 /* ^C */
      || (returned & 0xFF) == 4 /* ^D, PuTTY do not transmit ^C */
	) {
      VDBG (("\r\nTrying to stop '.com', stop DOSEMU, "
	     "generate DOSEMU exception, halt\r\n"));
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
      _VGA_autoload_default_palette (1);
      if (VESA_ACTIVE() && !UI.parameter.BIOS_compatible) {
	  /* some BIOS need VESA before VGA setmode: */
	  if (_VESA_setmode (3) != 0)
	      VDBG ((" [Last _VESA_setmode failed!] "));
	  _VGA_setmode (3);
	  }
#endif
      BIOS_killme();
      }

  return returned;
  }}

#if USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT)
/*
 * The initial probing for VGA/VESA cards:
 */
unsigned BIOS_probemonitor (void)
  {bound_stack();{
  unsigned char alternate_display, active_display;
  union EGA_bx_union EGA_bx;
  struct EGA_cx_str EGA_cx;

  VDBG (("%s: ", __FUNCTION__));

#if USER_SUPPORT & VESA_EDID
  {
  VBE_EDID_type info = {};
  static const struct edid_str extablished_edid[] = {
	{ 720, 400, 70 },	{ 720, 400, 88 },
	{ 640, 480, 60 },	{ 640, 480, 67 },
	{ 640, 480, 72 },	{ 640, 480, 75 },
	{ 800, 600, 56 },	{ 800, 600, 60 },
	{ 800, 600, 72 },	{ 800, 600, 75 },
	{ 832, 624, 75 },
	{ 1024, 768, 87 },	/* interlaced */
	{ 1024, 768, 60 },	{ 1024, 768, 70 },
	{ 1024, 768, 75 },
	{ 1280, 1024, 75}
	};

//#define VDBG(X) printf X

  VDBG ((" [EDID "));
  /* Try it twice, sometimes _EDID_read() is not working the first time: */
  if (_EDID_detect () != 0) {
      VDBG (("not supported] "));
      }
    else if (_EDID_read (&info) != 0 && _EDID_read (&info) != 0) {
      VDBG (("read failed] "));
      }
    else {
      int i;
      unsigned short mask;

      VDBG (("%u.%u: screen 0x%X, size %u cm x %u cm, "
	     "established timming 1:0x%X 2:0x%X, manufacturer timming mask: 0x%X, ",
	     info.version, info.revision, info.monitor_model,
	     info.max_horizontal_size_cm, info.max_vertical_size_cm,
	     *(unsigned char *) &info.established_timming,
	     *((unsigned char *) &info.established_timming + 1),
	     info.manufacturer_timming));

#if USER_SUPPORT & VGA_FREQUENCY
      {
      unsigned char maxHz640 = 0, maxHz800 = 0,
		    maxHz1024 = 0, maxHz1280 = 0;
      unsigned short maxVertical = 0;

      if (info.established_timming._1280_1024_75Hz)
	  maxVertical = 1024;
	else if (info.established_timming._1024_768_i87Hz)
	  maxVertical = 768;
	else if (info.established_timming._800_600_56Hz)
	  maxVertical = 600;

      if (info.established_timming._640_480_72Hz)
	  maxHz640 = 72;

      if (info.established_timming._800_600_72Hz)
	  maxHz800 = 72;
	else if (info.established_timming._800_600_60Hz)
	  maxHz800 = 60;

      if (info.established_timming._1024_768_75Hz)
	  maxHz1024 = 75;
	else if (info.established_timming._1024_768_70Hz)
	  maxHz1024 = 70;
	else if (info.established_timming._1024_768_60Hz)
	  maxHz1024 = 60;

      if (info.established_timming._1280_1024_75Hz)
	  maxHz1280 = 75;

      VDBG ((" [Declaring CIRRUS maxHlines %u, maxVfreq 640x480 = %u Hz,"
	     " 800x600 = %u Hz, 1024x768 = %u Hz, 1280x1024 = %u Hz]",
	     maxVertical, maxHz640, maxHz800, maxHz1024, maxHz1280));

      _CIRRUS_set_monitor_type (maxHz640, maxHz800, maxHz1024, maxHz1280, maxVertical);
      VDBG (("] "));
      }
#endif

      for (mask = 0x8000, i = 15; mask != 0; mask >>= 1, i--)
	  if (mask & *(unsigned short *) &info.established_timming)
	      break;
      if (i >= 0)
	  UI.edid = extablished_edid[i]; /*  clears the structure */

      UI.edid.Hsize_cm = info.max_horizontal_size_cm;
      UI.edid.Vsize_cm = info.max_vertical_size_cm;

      for (i = 0; i < (signed) nbof(info.standard_timming); i++) {
	  unsigned short hres, vres, freq;

	  if (info.standard_timming[i].resolution <= 1)
	      continue;
	  hres = (info.standard_timming[i].resolution + 31) * 8;
	  switch (info.standard_timming[i].aspect_ratio) {
	    case 3: vres = hres * 56 / 100; break;
	    case 2: vres = hres * 80 / 100; break;
	    case 1:
	    default: vres = hres * 75 / 100; break;
	    }
	  freq = info.standard_timming[i].vertical_refresh_freq + 60;

	  VDBG (("mode %u: horizontal=%u, vertical=%u, "
		 "refresh=%u Hz, aspect=%u; ",
		i, hres, vres, freq, info.standard_timming[i].aspect_ratio));

	  if (   hres > UI.edid.horizontal_resolution
	      && freq > 70) { /* else we do not care */
	      UI.edid.horizontal_resolution = hres;
	      UI.edid.vertical_resolution = vres;
	      UI.edid.frequency = freq;
	      }
	  }
#if 0
      if (maxHfreq) {
	  enum _CIRRUS_set_horizontal_frequency tmp;

	  if (maxHfreq < 35)
	      tmp = _CIRRUS_standard_VGA;
	    else if (maxHfreq == 35)
	      tmp = _CIRRUS_SuperVGA;
	    else if (maxHfreq < 37)
	      tmp = _CIRRUS_extended_SuperVGA;
	    else if (maxHfreq < 48)
	      tmp = _CIRRUS_multi_frequency;
	    else if (maxHfreq < 56)
	      tmp = _CIRRUS_extended_multi_frequency;
	    else if (maxHfreq < 64)
	      tmp = _CIRRUS_super_multi_frequency;
	    else
	      tmp = _CIRRUS_extended_super_multi_frequency;
	  VDBG ((" [Declaring CIRRUS monitor type, maxHfreq=%u, nb=%u] ",
		 maxHfreq, (unsigned)tmp));
	  _CIRRUS_set_horizontal_frequency (tmp);
	  }
#endif

      VDBG (("]\r\n"));
      if (info.DPMS.RBGcolor)
	  return info.monitor_model;
	else
	  return 2;
      }
  }
#endif /* VESA_EDID */

//#define VDBG(X) /**/

  if (_VGA_get_display_combination_code (&active_display,
					 &alternate_display) == 0) {
      VDBG (("INT10/1A00 returns 0x%X, i.e. ", active_display));
#if 0 /* assembly too big */
      switch (active_display) {
	case 0: /* no display */
	  VDBG (("no display. "));
	  break;
	default:
	  VDBG (("unknown. "));
	  break;
	case 1: /* mono adapter & mono display */
	  VDBG (("B&W display.\r\n"));
	  return 1;
	case 5: /* EGA mono */
	case 7: /* VGA mono */
	case 0x0B: /* MCGA mono */
	  VDBG (("monochrome display.\r\n"));
	  return 2;
	case 2: /* CGA color */
	case 4: /* EGA color */
	case 6: /* PGA color */
	case 8: /* VGA color */
	case 0x0A: /* MCGA digital color */
	case 0x0C: /* MCGA analog color */
	  VDBG (("color display.\r\n"));
	  return 3;
	}
#else
      if (active_display == 0)
	  VDBG (("no display. "));
	else if (active_display == 3 || active_display == 9 || active_display > 0x0C)
	  VDBG (("unknown. "));
	else if (active_display == 1) {
	  VDBG (("B&W display.\r\n"));
	  return 1;
	  }
	else if (active_display & 1) {
	  VDBG (("monochrome display, forced to color.\r\n"));
	  _VGA_set_display_combination_code (active_display + 1,
						alternate_display);
	  _VGA_enable_gray_scale_summing (0);
	  return 3;
	  }
	else {
	  VDBG (("color display.\r\n"));
	  return 3;
	  }
#endif
      }
  VDBG (("INT10/1A00 not supported, "));

  _EGA_getdisplay (&EGA_bx, &EGA_cx);

  VDBG (("INT10/1200 reports bx = 0x%X, cx = 0x%X, i.e. ",
	EGA_bx.IOaddr_memory, *(unsigned *)&EGA_cx));

  if (EGA_bx.enums.IOaddr == EGA_COLOR_3Dx) {
      VDBG (("color display.\r\n"));
      return 3;
      }
    else if (EGA_bx.enums.IOaddr == EGA_MONO_3Bx) {
      VDBG (("B&W display.\r\n"));
      return 1;
      }
    else {
      VDBG (("error.\r\n"));
      return 0;
      }
  }}

static const struct DAC_str VGA_VESA_default_colordac16[16] = {
#define H 63
#define M 2*H/3
#define L H/3
#define X 4*H/9
#define Y H/4
    {red: 0, green: 0, blue: 0},	/* black */
    {red: 0, green: 0, blue: M},	/* blue */
    {red: 0, green: M, blue: 0},	/* green */
    {red: 0, green: M, blue: M},	/* cyan */
    {red: M, green: 0, blue: 0},	/* red */
    {red: M, green: 0, blue: M},	/* magenta */
    {red: X, green: Y, blue: Y},	/* brown */
    {red: M, green: M, blue: M},	/* light gray */
    {red: L, green: L, blue: L},	/* dark gray */
    {red: 0, green: 0, blue: H},	/* light blue */
    {red: 0, green: H, blue: 0},	/* light green */
    {red: 0, green: H, blue: H},	/* light cyan */
    {red: H, green: 0, blue: 0},	/* light red */
    {red: H, green: 0, blue: H},	/* light magenta */
    {red: H, green: H, blue: M},	/* yellow */
    {red: H, green: H, blue: H}		/* white */
#undef H
#undef M
#undef L
#undef X
#undef Y
	};

unsigned __attribute__ ((const))
VESA_color (unsigned char _red, unsigned char _green, unsigned char _blue)
//  {bound_stack();{
  {{
  /* red, green, blue as 6 bit DAC, 0..63 */
  unsigned tmpred =   _red   << 2;
  unsigned tmpgreen = _green << 2;
  unsigned tmpblue =  _blue  << 2;

  tmpred   >>= (8 - UI.parameter.red_mask_size);
  tmpgreen >>= (8 - UI.parameter.green_mask_size);
  tmpblue  >>= (8 - UI.parameter.blue_mask_size);

  return   (tmpred << UI.parameter.red_field_pos)
	 | (tmpgreen << UI.parameter.green_field_pos)
	 | (tmpblue << UI.parameter.blue_field_pos);
  }}

static void VGA_VESA_set_DAC_332 (void) /* 332 palette, 2 bits for blue */
  {bound_stack();{
  struct DAC_str array[256], *ptr = array;
  unsigned char R, G, B;
  /* The linear form wastes bits, because nothing can be seen at bottom */
  /* To have some gray, we need R == G == B */
  const unsigned char ladder4[] = {0,     23,     39,         63};
  const unsigned char ladder8[] = {0, 15, 23, 31, 39, 47, 55, 63};

  for (B = 0; B < 4; B++) /* blue is MSB, to easy write in octal */
      for (G = 0; G < 8; G++)
	  for (R = 0; R < 8; R++) {
	      ptr->red = ladder8[R];
	      ptr->green = ladder8[G];
	      ptr->blue = ladder4[B];
	      ptr++;
	      }
  UI_setDACblock (0, 256, array);
  }}


#if BUILD_APPLICATION()
static
#endif
void VGA_VESA_setup_color (void)
  {bound_stack();{
  unsigned i;

  /* REALTEK QUADTEL BIOS version 6.C01, seems to overwrite return address */
  extern volatile unsigned setjump_key;
  extern setjmp_buf setjump_buffer;

  setjump_key = 0x45746965;
  if (setjmp (&setjump_buffer) != 0) {
      VDBG ((" EXCEPTION RECEIVED! ] "));
      return;
      }

  switch (UI.parameter.nbcolor) {
    case 0: /* no colors, attributes */
#if 0
	/* would be nice to have the same
	   attributes in text & graphic, but... */
      _VGA_setpalette (0, 0x0);
      for (i = 1; i < 8; i++)
	 _VGA_setpalette (i, 0x8);
      _VGA_setpalette (8, 0x10);
      for (i = 9; i < 16; i++)
	 _VGA_setpalette (i, 0x18);
#endif

      if (UI.parameter.attr.isgraphic) {
	  VDBG ((" [no palette, graphic attributes] "));
	  for (i = black; i <= white; i++)
	      UI.stdcolor[i] = 0x85; /* mouse button pressed */
	  /* 0x80: flash when on active field */
	  /* 0x04: highlight */
	  UI.stdcolor[black] = 0;
	  UI.stdcolor[white] = 1;
	  UI.stdcolor[brown] = 1; /* mouse no button pressed */
	  }
	else
	  {
	  VDBG ((" [no palette, text attributes] "));
	  for (i = black; i <= white; i++)
	     UI.stdcolor[i] = 0x07;
	  /* standart no_attribute: from 0x02 to 0x08 */
	  UI.stdcolor[black] = 0;
	  UI.stdcolor[brown] = 0x06; /* mouse no button pressed:
					MOUSE_DEFAULT_COLOR */
	  }
      break;
    case 1 ... 2:
      VDBG ((" [no palette <= 2 colors] "));
      UI.stdcolor[0] = 0;
      for (i = 1; i < 16; i++)
	  UI.stdcolor[i] = 1;
      break;
    case 3 ... 15:
      VDBG ((" [no palette < 16 colors] "));
      for (i = 0; i < 16; i++)
	  UI.stdcolor[i] = i % 4; /* ??? */
      break;
    case 16:
      VDBG ((" [setting palette 16 colors"));
      for (i = 0; i < 16; i++) {
	   UI.stdcolor[i] = i;
#if USER_SUPPORT & VGA_SUPPORT
	   _VGA_setpalette (i, i);	/* ATI bios, when probing VGA */
#endif
	   }
      UI_setDACblock (0, 16, VGA_VESA_default_colordac16);
      VDBG (("] "));
      break;
    case 17 ... 256:
      VDBG ((" [setting palette 256 colors"));
      VGA_VESA_set_DAC_332 ();
      VDBG (("] "));

      UI.parameter.red_mask_size = 3;
      UI.parameter.red_field_pos = 0;
      UI.parameter.green_mask_size = 3;
      UI.parameter.green_field_pos = 3;
      UI.parameter.blue_mask_size = 2;
      UI.parameter.blue_field_pos = 6;
#if 0
      /* organised 2 MSB:blue, 3middle:green, 3LSB:red */
			 /* octal! BGR*/
      UI.stdcolor[black] =        0000;
      UI.stdcolor[blue] =         0200;
      UI.stdcolor[green] =        0040;
      UI.stdcolor[cyan] =         0240;
      UI.stdcolor[red] =          0004;
      UI.stdcolor[magenta] =      0204;
      UI.stdcolor[brown] =        0044;
      UI.stdcolor[lightgray] =    0244;
      UI.stdcolor[darkgray] =     0122;
      UI.stdcolor[lightblue] =    0300;
      UI.stdcolor[lightgreen] =   0070;
      UI.stdcolor[lightcyan] =    0370;
      UI.stdcolor[lightred] =     0007;
      UI.stdcolor[lightmagenta] = 0307;
      UI.stdcolor[yellow] =       0277;
      UI.stdcolor[white] =        0377;
      break;
#else
      /* fall through */
#endif
    default: /* more than 256 colors */
      VDBG ((" [init base color, >= 256 color] "));
      for (i = 0; i < 16; i++)
	   UI.stdcolor[i] = VESA_color (VGA_VESA_default_colordac16[i].red,
					VGA_VESA_default_colordac16[i].green,
					VGA_VESA_default_colordac16[i].blue);
      break;
    }

  setjump_key = 0;
  }}

#endif	/* VGA_SUPPORT | VESA_SUPPORT */

/*
 * The "VGA BIOS library" functions:
 */
#if USER_SUPPORT & VGA_SUPPORT
static unsigned
VGA_getmode (void)
//  {bound_stack();{
  {{
  unsigned char nbcol, page;
  char result;

  result = _VGA_getmode (&nbcol, &page);
  if (result < 0)
      return 0x8000 | (result & 0x7F);
    else
      return result;
  }}

static void
VGA_setattribute (struct attribute_str attr)
  {
  ATTR2BYTE(attr) &= ATTR2BYTE(UI.attributes.valid);

  if (ATTR2BYTE(UI.attributes.valid) != 0) {
      if (UI.parameter.attr.isgraphic) {
	  /* only monochrome graphic (0x0F) mode in VGA */
	  UI.bgcolor = 0;
	  if (attr.blink)
	      UI.fgcolor = 4; /* also highlight... */
	    else if (attr.brighter)
	      UI.fgcolor = 5;
	    else
	      UI.fgcolor = 1;
	  }
	else {
	  /* only monochrome text (0x07) mode in VGA */
	  UI.bgcolor = 0;
	  if (attr.underline)
	      UI.fgcolor = 0x01;
	    else if (attr.reverse)
	      UI.fgcolor = 0x10;
	    else
	      UI.fgcolor = 0x07;	/* see also init of stdcolor array */
	  if (attr.blink)
	      UI.fgcolor |= 0x80;
	  if (attr.brighter)
	      UI.fgcolor |= 0x08;
	  }
      }
  UI.attributes.current = attr;
  }

/* Cursor movement: home at (0,0) */
static unsigned
VGA_setcursor (unsigned char row, unsigned char col)
  {
  if (COMMON_setcursor (row, col) == 0)
      _VGA_setcursor (0, row, col);
  return 0;
  }

#if USER_SUPPORT & VGA_USEINT1013
static unsigned
VGA_getcursor (unsigned char *row, unsigned char *col)
  {bound_stack();{
  unsigned char cursorstart, cursorend;

  checkptr (row, "VGA_getcursor invalid 1st parameter.");
  checkptr (col, "VGA_getcursor invalid 2nd parameter.");

  /* the BIOS _is_ maintainning the cursor position */
  _VGA_getcursor (0, row, col, &cursorstart, &cursorend);
  return 0;
  }}

static inline void
VGA_set_color_and_page (unsigned char *color, unsigned char *page)
  {
   if (UI.parameter.attr.isgraphic || UI.parameter.nbcolor == 2) {
#if USER_SUPPORT & VGA_INT1013_BG
      /* 256 colors do not completely work with INT10, %ah=0x13 */
      if (UI.parameter.nbcolor == 256)
	  *page = UI.bgcolor; /* used as background color in 256 color modes */
	else
#endif
	  *page = 0;
#ifdef VGA16_XOR_TRICK
	  /* use XOR trick for VGA graphic 16 colors */
      if (UI.parameter.nbcolor == 16)
	  *color = 0x80 | (UI.fgcolor ^ UI.bgcolor);
	else
#endif
	  *color = UI.fgcolor;
      }
    else {
      *page = 0;
      *color = (UI.bgcolor << 4) | UI.fgcolor;
      }

  }

static inline void
VGA_writechar (char character, unsigned char page, unsigned char attr,
		unsigned short nbtimes)
  {
  _VGA_writechar_attr (character, page, attr, nbtimes);
  }

static inline void
VGA_readchar (char *character, unsigned char *attr)
  {
  _VGA_readchar (character, 0, attr);
  }

static void
VGA_putstr (const char *str)
  {bound_stack();{
  const char *ptr = str;
  unsigned char color, page;

  VGA_set_color_and_page (&color, &page);

  /* str pointer has to be in %es segment ! */
  /* _VGA_writestring do not clear the end of line when printing "\r\n",
     so do it ourself : */

  do {
      unsigned char row, col;
      while (*ptr != '\0' && *ptr != '\r') ptr ++;
      if (ptr - str != 0) {
	  unsigned char cursorstart, cursorend;

	  _VGA_getcursor (0, &row, &col, &cursorstart, &cursorend);
	  _VGA_writestring (str, ptr - str, page, color, 1, row, col);
	  }
      if (*ptr == '\r') {
	  unsigned char cursorstart, cursorend;

	  _VGA_getcursor (0, &row, &col, &cursorstart, &cursorend);
	  if (col < UI.parameter.nbcol)
	      VGA_writechar (' ', page, color, UI.parameter.nbcol - col);
	  str = ptr++;
	  }
      } while (*ptr != 0);
  }}

#else  /* VGA_USEINT1013 */

#define VGA_getcursor COMMON_getcursor

#if USER_SUPPORT & VGA_MEMORY
static inline void
VGA_writechar (char character, unsigned char page, unsigned char attr,
			unsigned short nbtimes)
  {
  unsigned offset = (UI.parameter.row * UI.parameter.nbcol
			+ UI.parameter.col) * 2;
  nbtimes ++;
  while (--nbtimes) {
      gpokew (offset, ((unsigned short)attr << 8) | (character & 0xFF));
      offset += 2;
      }
  }

static inline void
VGA_readchar (char *character, unsigned char *attr)
  {
  unsigned offset = (UI.parameter.row * UI.parameter.nbcol
			+ UI.parameter.col) * 2;
  unsigned short tmp = gpeekw (offset);

  *character = tmp & 0xFF;
  *attr = tmp >> 8;
  }

static inline void
VGA_scroll (int nblines)
  {
  unsigned delta = UI.parameter.nbcol * 2 * nblines, cpt;

  if (nblines > 0)
      for (cpt = delta;
	   cpt < UI.parameter.nbrow * UI.parameter.nbcol * 2U;
	   cpt += 2)
	  gpokew (cpt - delta, gpeekw (cpt));
    else /* untested */
      for (cpt = UI.parameter.nbrow * UI.parameter.nbcol * 2U;
	   cpt >= delta;
	   cpt -= 2)
	  gpokew (cpt, gpeekw (cpt - delta));
  }

static inline void
VGA_set_color_and_page (unsigned char *color, unsigned char *page)
  {
  *page = 0;
  *color = (UI.bgcolor << 4) | UI.fgcolor;
  }

#else /* VGA_MEMORY */
static inline void
VGA_writechar (char character, unsigned char page, unsigned char attr,
		unsigned short nbtimes)
  {
  /* For the people measuring the quality of software
     by grep-ing the bad words: fuck fuck.
     Sorry, I do not have a very extended vocabulary,
     at least in this language!
   */

#if 1
  _VGA_writechar_attr (character, page, attr, nbtimes);
#elif 0
  _VGA_writechar_color (character, page, attr, nbtimes);
#elif 0
  /* problem of scroll when writing the leftmost lowest char: */
  unsigned char row = UI.parameter.row, col = UI.parameter.col;
  while (nbtimes--)
      _VGA_writestring (&character, 1, page, attr, 0, row, col++);
#endif
  }

static inline void
VGA_readchar (char *character, unsigned char *attr)
  {
  _VGA_readchar (character, 0, attr);
  }

static inline void
VGA_scroll (int nblines)
  {
/*
 * Scrolling the screen is not so often working,
 * it is better to do the alternate method...
 * Scrolling the screen by writing '\n' on the last line
 * of the screen is not working in DOSEMU over telnet.
 */
  unsigned short version, patchlevel;
  if (_DOSEMU_check(&version, &patchlevel) == 0) {
      unsigned char filler = 0;
      /* We cannot use the "filler" parameter here because in 256 colors,
	 we do not know if the BIOS is able to manage the "page" parameter
	 as the background color - so then we do not know if the filler
	 should be the background color or zero for graphic 256 color modes. */
      _VGA_scroll (0, 0, UI.parameter.nbrow - 1, UI.parameter.nbcol - 1,
		filler, nblines);
      }
    else {
      while (--nblines >= 0) {
	  _VGA_writestring ("\n", 1, 0, 0, 0, UI.parameter.nbrow-1, 0);
	  }
      }
  }

static inline void
VGA_set_color_and_page (unsigned char *color, unsigned char *page)
  {
   if (UI.parameter.attr.isgraphic || UI.parameter.nbcolor == 2) {
      if (UI.parameter.nbcolor == 256)
	  *page = UI.bgcolor; /* used as background color in 256 color modes */
	else
	  *page = 0;
      *color = UI.fgcolor;
      }
    else {
      *page = 0;
      *color = (UI.bgcolor << 4) | UI.fgcolor;
      }

  }

#endif /* VGA_MEMORY */

static void
VGA_putstr (const char *str)
  {bound_stack();{
  unsigned char color, page;

  VGA_set_color_and_page (&color, &page);

  while (*str != '\0') {
      switch (*str) {
	case '\007': /* ^G */
	  BOOT1_putstr ("\007");
	  break;
	case '\010':
	  if (UI.parameter.col > 0)
	      UI.parameter.col -= 1;
	  break;
	default:
	  VGA_writechar (*str, page, color, 1);
	  if (++UI.parameter.col >= UI.parameter.nbcol) {
	      UI.parameter.col = 0;	/* automatic CRLF */
	case '\n':  /* it is valid in C */
	      if (++UI.parameter.row >= UI.parameter.nbrow) {
		  UI.parameter.row--;
		  VGA_scroll (1);
	case '\r': /* it is valid in C */
		  VGA_writechar (' ', page, color,
			UI.parameter.nbcol - UI.parameter.col);
		  UI.parameter.col = 0;
		  }
	      }
	  break;
	}
      _VGA_setcursor (0, UI.parameter.row, UI.parameter.col);
      str++;
      }
  }}
#endif /* VGA_USEINT1013 */

static void FASTCALL
VGA_graphic_setpixel (coord xy, unsigned color)
  {
  _VGA_setpixel (xy.x, xy.y, color, 0);
  }

static unsigned FASTCALL
VGA_graphic_getpixel (coord xy)
  {
  return _VGA_getpixel (xy.x, xy.y, 0);
  }

static void
VGA_graphic_plotHline (coord xy, unsigned short xend, unsigned color)
  {
  while (xy.x < xend) {
      _VGA_setpixel (xy.x, xy.y, color, 0);
      xy.x++;
      }
  }

static void FASTCALL
VGA_text_setpixel (coord xy, unsigned color)
  {
  unsigned char row, col;

  VGA_getcursor (&row, &col);
  VGA_setcursor (xy.y, xy.x);
  VGA_writechar (color & 0xFF, 0, (color >> 8) & 0xFF, 1);
  VGA_setcursor (row, col);
  }

static unsigned FASTCALL
VGA_text_getpixel (coord xy)
  {
  unsigned char row, col, letter, attr;

  VGA_getcursor (&row, &col);
  VGA_setcursor (xy.y, xy.x);
  VGA_readchar(&letter, &attr);
  VGA_setcursor (row, col);
  return ((unsigned)attr << 8) | letter;
  }

static void
VGA_text_plotHline (coord xy, unsigned short xend, unsigned color)
  {
  unsigned char row, col;

  VGA_getcursor (&row, &col);
  VGA_setcursor (xy.y, xy.x);
  VGA_writechar (color & 0xFF, 0, (color >> 8) & 0xFF, xend - xy.x);
  VGA_setcursor (row, col);
  }

static void
VGA_clearscreen (void)
  {
/*
 * The scroll method to clear the screen is only working
 * when the scroll BIOS is working, which is not so usual...
 */
#if 0
  unsigned char color, page;

  VGA_set_color_and_page (&color, &page);

  _VGA_scroll (0, 0, UI.parameter.nbrow-1, UI.parameter.nbcol-1,
	color, 0);
#else
  signed short cpt;
  unsigned char color, page;

  VGA_set_color_and_page (&color, &page);

  for (cpt = 0; cpt < UI.parameter.nbrow; cpt++) {
      VGA_setcursor (cpt, 0);
      VGA_writechar (' ', page, color, UI.parameter.nbcol);
      }

  if (UI.parameter.heigh % UI.parameter.charheigh != 0) {
      /* 800x600 with background color and 8x16 chars, half
	 a line left. hopes that will not crash anything */
      VGA_setcursor (cpt, 0);
      VGA_writechar (' ', page, color, UI.parameter.nbcol);
      }
#endif
  VGA_setcursor (0, 0);
  }

static unsigned
VGA_setDACblock (unsigned short startindex, unsigned short nb,
		 const struct DAC_str *palette)
  {bound_stack();{
  VGA_RGB vga_palette[3*nb];
  unsigned cpt;

  for (cpt = 0; cpt < nb; cpt++) { /* convert */
      vga_palette[cpt].red = palette[cpt].red;
      vga_palette[cpt].green = palette[cpt].green;
      vga_palette[cpt].blue = palette[cpt].blue;
      }

  _VGA_setDACarray (startindex, nb, vga_palette);
  return 0;
  }}

__attribute__ ((cdecl, regparm (0))) static
void VGA_timer_handler (const unsigned ip_flags,
			const unsigned seggsfs,
			const unsigned segdses,
			const unsigned edx,
			const unsigned ecx,
			const unsigned eax)
  {
  extern setjmp_buf setjump_buffer;
  extern volatile unsigned setjump_key;

  /*
   * Now rescue the ISO9001 video BIOSes which executes infinite loops
   * when trying to set invalid modes, but only when they do not change
   * the interrupt table position (lidt assembly instruction), or when
   * they do it interrupt disabled, else there is no hopes to get back
   * the control of the PC (unless they also execute an invalid instruction).
   */
  VDBG (("timeout for VGA operation.\r\n"));
  if (setjump_key == 0x45746965 && setjump_buffer.cs != 0)
      longjmp (&setjump_buffer);
  VDBG (("ERROR: did not longjmp in VGA_timer_handler.\r\n"));
  }

/*
 * Use out of line versions of these functions because they are
 * a bit long:
 */
static void VGA_safe_setmode (char mode)
  {
  _VGA_safe_setmode (mode);
  }

static void
VGA_safe_writechar_attr (char character, unsigned char page,
			 unsigned char attr)
  {
  _VGA_safe_writechar_attr (character, page, attr, 1);
  }

#if !(USER_SUPPORT & VGA_MEMORY)
static void
VGA_safe_readchar (char *character, unsigned char *attr)
  {
  _VGA_safe_readchar (character, 0, attr);
  }

/* we do a "_VGA_safe_setcursor (0, 0, 1);", and
 * to correctly detect "graphicbgcolor", we should
 * have coordinate correctly chosen here, whatever
 * the font heigh:
 */
#define COORD_X		10 /* on the second char of the screen */
#define COORD_Y		6  /* on the first line, font heigh >= 8 */
#define COORD_PAGE	0
#define TESTCOLOR	3

static unsigned char
VGA_safe_getpixel (void)
  {
  return _VGA_safe_getpixel (COORD_X, COORD_Y, COORD_PAGE);
  }

static void
VGA_safe_setpixel (unsigned char color)
  {
  _VGA_safe_setpixel (COORD_X, COORD_Y, color, COORD_PAGE);
  }

#undef COORD_X
#undef COORD_Y
#undef COORD_PAGE

#endif

static unsigned
VGA_getsize (unsigned mode, struct video_parameter_str *param)
  {bound_stack();{
  volatile unsigned char oldmode; /* because setjmp() */
  unsigned char newmode, nbcol, page;
  unsigned error = 0;
  functionality_table_type table = {0};
  unsigned char cursorstart, cursorend;
  extern volatile unsigned setjump_key; /* library.c, exceptions */
  extern unsigned char timer_handler_ptr[], timer_counter[];

  VDBG (("%s mode 0x%X: ", __FUNCTION__, mode));

  checkptr (param, "VGA_getsize invalid parameter.");

  memset (param, 0, sizeof(*param));
  param->identification = mode & 0x7F;

  if (!VIDEO_mode_is_valid (param->identification)) {
      VDBG (("refused from validmode bitfield.\r\n"));
      return 0x100;
      }
  VIDEO_mode_invalidate(param->identification); /* may be revalidated later */
#if SETUP & UPDATE_BOOTPARAM
  BOOT1_update_bootparam (); /* if the PC crash, it will not be tried again */
#endif
  _VGA_getcursor (0, &param->row, &param->col, &cursorstart, &cursorend);

  oldmode = _VGA_getmode (&nbcol, &page) & 0x7F;
  if (oldmode != param->identification) {
      extern setjmp_buf setjump_buffer;

      setjump_key = 0x45746965;
      codeseg_pokel (timer_handler_ptr, (unsigned)VGA_timer_handler);
      codeseg_pokel (timer_counter, 4*18); /* 4 second timeout */
      if (setjmp (&setjump_buffer) != 0) {
	  setjump_key = 0;
//	  if (_VIDEO_cold_initialise() != 0)
//	      VGA_safe_setmode (oldmode);
	  /* may reset more parameters: */
	  _VGA_safe_autoload_default_palette (1);
	  VGA_safe_setmode (oldmode | 0x80);
	  VGA_VESA_setup_color();
	  _VGA_blinking (0); /* it is re-enabled by INT10/setmode */
	  VDBG (("have hit an exception, like division by 0"
		 " or invalid instruction, or timeout.\r\n"));
	  codeseg_pokel (timer_handler_ptr, 0);
	  codeseg_pokel (timer_counter, 0);
	  return 0x800;
	  }
       /* do not reset palette in the following setmode (can be long): */
      if (!_VGA_autoload_default_palette (0))
	  VDBG ((" [autoload_default_palette not supported!] "));

      /*
       * Ater next _VGA_setmode, we have to be very carefull,
       * because we are maybe in an invalid mode - where
       * some VGA primitives modify unusual registers or even memory!
       * (old S3 Elsa Winner card) - use the VGA_safe_*() functions.
       */
      VGA_safe_setmode (mode | 0x80); /* try to not clear screen */
      newmode = _VGA_safe_getmode (&nbcol); /* will check nbcol later */

      if (((unsigned)newmode & 0x7F) != param->identification) {
	  VDBG (("_VGA_setmode (0x%X) and _VGA_getmode (0x%X) differ.\r\n",
		param->identification, newmode & 0x7F));
	  error = 0x101;
	  }
      }

#if 0	/* to test that setjmp/longjmp rescue is working */
  asm volatile (
"	cmpl	$2,%0		\n"
"	jne	1f		\n"
"	xor     %%ax,%%ax	\n"
"	mov     %%ax,%%ds	\n"
"	mov     %%ax,%%es	\n"
"	ud2			\n"
"	1:			\n"
"	cmpl	$1,%0		\n"
"	je	1b		\n"
	: : "g" (mode));
#endif

  if (   error == 0
      && (   _VGA_get_functionality_info (&table) != 0
	  || param->identification != table.current_mode
	  || nbcol != table.nbcolumn)) {
      VDBG (("wrong _VGA_get_functionality_info.\r\n"));
      error = 0x102;
      }

  if (error == 0 && table.nonVGAmodes.adapter_required) {
      VDBG (("needs an adapter.\r\n"));
      error = 0x103;
      }

  if (error == 0) {
#if USER_SUPPORT & VGA_MEMORY
      /* This accepts only TEXT modes: */
      _VGA_safe_setcursor (0, 0, 0);
      VGA_safe_writechar_attr ('@', 0, 7);
      if (peekb (0xB8000000U) != '@')
	  error = 0x104;
      VGA_safe_writechar_attr ('!', 0, 7);
      if (peekb (0xB8000000U) != '!')
	  error = 0x105;
      param->base_address = 0xB8000;
      param->attr.isgraphic = 0;
      param->attr.graphicbgcolor = 0;
#else
      unsigned char charread, charsaved, attrread, attrsaved;

      /* Detect (in a BIOS way) if this is a graphic mode:
       * By writing a char in the background color (which
       * is 0 in graphic modes, at least if page == 0) - If we
       * can read it back, it is a text mode.
       * Take care, with 2 color modes, the writing color
       * is ignored (always set to 1) so check attribute also.
       * This can crash VIDEO BIOS of very old cards.
       */
      _VGA_safe_setcursor (0, 0, 1);
      VDBG (("["));
      VGA_safe_readchar (&charsaved, &attrsaved); /* not change cursor */
      VDBG (("G"));
      VGA_safe_writechar_attr ('@', 0, 0); /* not change/update cursor */
      VDBG (("?"));
      VGA_safe_readchar (&charread, &attrread);
      if (charread != '@') {
	  /* It is a graphic mode, but check that s/getpixel are working,
	     else treat it as a text mode: */
	  unsigned char color = VGA_safe_getpixel ();
	  VGA_safe_setpixel (0);
	  if (   (VGA_safe_setpixel (0), VGA_safe_getpixel () != 0)
	      || (VGA_safe_setpixel (1), VGA_safe_getpixel () != 1)) {
	      VDBG (("y, but no s/getpixel()] "));
	      VGA_safe_setpixel (color);
	      param->base_address = 0xA0000;
	      param->attr.isgraphic = 0;
	      param->attr.graphicbgcolor = 0;
	      }
	    else {
	      VDBG (("y] "));
	      param->base_address = 0xA0000;
	      param->attr.isgraphic = 1;
	      /* Is it graphic and can it draw bgcolor? maybe */
	      if (table.nbcolor == 256) {
		  VDBG (("[bgcolor"));
		  /* not change/update cursor, but write in page
		     TESTCOLOR, will it change bgcolor of page 0 ? */
		  VGA_safe_writechar_attr (' ', TESTCOLOR, 0);
		  VDBG (("? "));
		  color = VGA_safe_getpixel ();
		  VDBG (("read %u ", color));
		  if (color == TESTCOLOR) {
		      param->attr.graphicbgcolor = 1;
		      VDBG (("yes] "));
		      }
		    else {
		      param->attr.graphicbgcolor = 0;
		      VDBG (("no] "));
		      }
		  }
#ifdef VGA16_XOR_TRICK
		else if (table.nbcolor == 16)
		  param->attr.graphicbgcolor = 1;
#endif
		else
		  param->attr.graphicbgcolor = 0;
	      }
	  }
	else if (attrread != 0) {
	  VDBG (("y - 2 colors] "));
	  table.nbcolor = 2;
	  param->base_address = 0xA0000;
	  param->attr.isgraphic = 1;
	  /* Is it graphic and can it draw bgcolor? no */
	  param->attr.graphicbgcolor = 0;
	  }
	else {
	  VDBG (("n] "));
	  param->base_address = 0xB8000;
	  param->attr.isgraphic = 0;
	  /* Is it graphic and can it draw bgcolor? no */
	  param->attr.graphicbgcolor = 0;
	  }
      VGA_safe_writechar_attr (charsaved, 0, attrsaved);

      if (   table.nbcolor != 0
	  && (   table.nbcolor > 256
	      || table.nbcolor != (1 << (my_ffs (table.nbcolor) - 1)))) {
#if 1
	  /*
	   * I am fed up. The VGA BIOS should be publically described,
	   * even its latest revision... There are hidden fields here.
	   * I will not sign a NDA to know what it is.
	   */
	  error = 0x107;
#else
	  /* Another VIDEO BIOS which is lying...
	     and the number of color cannot be safely detected. */
	  VDBG ((" [BIOS say nbcolor = %u, set it to ", table.nbcolor));
	  if (!param->attr.isgraphic) {
	      VDBG (("text16] "));
	      table.nbcolor = 16;
	      }
	    else {
	      unsigned savecolor, color;

	      VDBG ((" <getpixel"));
	      savecolor = VGA_safe_getpixel ();
	      VDBG ((", setpixel"));
	      VGA_safe_setpixel (2);
	      VDBG ((", getpixel"));
	      color = VGA_safe_getpixel ();

	      if (color != 2) {
		  VDBG (("> (2->%u) 2] ", color));
		  table.nbcolor = 2;
		  }
		else {
		  VDBG ((", setpixel"));
		  VGA_safe_setpixel (17);
		  VDBG ((", getpixel"));
		  color = VGA_safe_getpixel ();
		  if (color != 17) {
		      VDBG (("> (17->%u) 16] ", color));
		      table.nbcolor = 16;
		      }
		    else {
		      VDBG (("> 256] "));
		      table.nbcolor = 256;
		      }
		  }
	      VDBG ((" <setpixel"));
	      VGA_safe_setpixel (savecolor);
	      VDBG ((">"));
	      }
#endif
	  }
#endif /* VGA_MEMORY */
      }

  if (oldmode != param->identification) {
      VGA_safe_setmode (oldmode | 0x80);
      VGA_VESA_setup_color();
       /* do reset palette in other setmode: */
      if (!_VGA_autoload_default_palette (1))
	  VDBG ((" [autoload_default_palette not supported!] "));
      _VGA_blinking (0); /* it is re-enabled by INT10/setmode */
//      _VGA_set_block_specifier (0);
      }

  _VGA_setpage (0); /* necessary ? */
  _VGA_setcursor (0, param->row, param->col);	/* force it */

  if (error != 0) {
      setjump_key = 0;
      codeseg_pokel (timer_handler_ptr, 0);
      codeseg_pokel (timer_counter, 0);
      return error;
      }

  param->nbcol = table.nbcolumn;
  {
  unsigned short version, patchlevel;
  if (_DOSEMU_check(&version, &patchlevel) == 0) { /* bug in DOSEMU, see
      http://marc.theaimsgroup.com/?l=linux-msdos&m=98830103525815&w=2
      */
      table.nbrow = peekb (0x00400084) + 1;
      }
  }
  param->nbrow = table.nbrow;
  param->nbcolor = table.nbcolor;
  VDBG (("%ux%u, %u colors/attributes, regenlen: %u,"
	 " regenstart: %u, nbscanline: %u, %u bytes/char,"
	 " primarycharblock = %u, secondarycharblock = %u, ",
	table.nbcolumn, table.nbrow, table.nbcolor,
	table.regenbufferlen, table.regenbufferstart,
	table.nbscanline, table.nbbyteperchar,
	table.primarycharblock, table.secondarycharblock));

  param->charwidth = 8;	/* nothing else on BIOS */
  param->charheigh = table.nbbyteperchar;

  if (param->attr.isgraphic) {
      param->heigh = table.nbrow * param->charheigh;
      param->width = table.nbcolumn * param->charwidth;
      if (param->heigh == 592 && param->charheigh == 16)
	  param->heigh = 600;
      VDBG (("graphic, width = %u, heigh = %u", param->width, param->heigh));
      }
    else {
      param->heigh = table.nbrow;	/* needed for text mouse */
      param->width = table.nbcolumn;
      VDBG (("text, width = %u, heigh = %u", param->width, param->heigh));
      }
  VIDEO_mode_revalidate(param->identification);
  setjump_key = 0;
  codeseg_pokel (timer_handler_ptr, 0);
  codeseg_pokel (timer_counter, 0);

#ifndef KEEP_ALL_VGA_MODE
  if (param->identification <= 0x37 || param->nbcolor > 4)
#endif
      {
      struct videomode_str videomode;

      videomode.number = param->identification;
      videomode.width = param->width;
      videomode.heigh = param->heigh;
      videomode.text = !param->attr.isgraphic;
      videomode.vesa = 0;
      /*
       * NOTE: some video card have an INCOMPLETE or CORRUPTED 8x14 font!
       * at least one has this font up to the char 'z', no more !
       */
      videomode.font8x14 = (param->charheigh == 14);
      if (param->nbcolor)
	  videomode.bpp = my_ffs (param->nbcolor) - 1;
	else
	  videomode.bpp = 0;
      videomode.attr = param->attr.isgraphic;
      add_valid_mode (&videomode);
      }

#if SETUP & UPDATE_BOOTPARAM
  BOOT1_update_bootparam (); /* no PC crash, OK */
#endif
  VDBG ((", end getsize.\r\n"));
  return 0;
  }}

static unsigned VGA_setmode (unsigned mode);

static const struct user_interface_fct_str VGA_UI_FCT = {
    getsize:		VGA_getsize,
    setmode:		VGA_setmode,
    getmode:		VGA_getmode,
    clearscreen:	VGA_clearscreen,

    setcursor:		VGA_setcursor,
    getcursor:		VGA_getcursor,
    setfgcolor:		COMMON_setfgcolor,
    setbgcolor:		COMMON_setbgcolor,
    setattribute:	VGA_setattribute,

    putstr:		VGA_putstr,

    setpixel:		VGA_text_setpixel,
    getpixel:		VGA_text_getpixel,
    plotHline:		VGA_text_plotHline,

    setDACblock:	VGA_setDACblock,
    };

static unsigned
VGA_setmode (unsigned mode)
  {bound_stack();{
  struct video_parameter_str param;
  unsigned saved_mode = mode, tmp;

  mode &= ~0x8080;

  if (mode > 0x7F) {
      VDBG (("%s called with mode 0x%X !\r\n", __FUNCTION__, mode));
      return 1;
      }
  if ((tmp = VGA_getsize (mode, &param)) != 0)
      return tmp;

#if USER_SUPPORT & VESA_HARDWINDOW
  free_hardwin();
#endif

  UI.parameter = param;
  UI.function = VGA_UI_FCT;
  if (UI.parameter.attr.isgraphic) {
      UI.function.setpixel = VGA_graphic_setpixel;
      UI.function.getpixel = VGA_graphic_getpixel;
      UI.function.plotHline = VGA_graphic_plotHline;
      }
  _VGA_setmode (mode | 0x80);	/* do not clearscreen now */

#if 0 /* This does not help my failing board (AT3D), modes 0x0F and 0x10 */
  if (UI.parameter.charheigh == 14) {
      /* Some new boards forget to do so, when they have a false
	 8x16 to map to 8x14: */
      _VGA_loadgraphic_ROM8x14 (25);
      }
#endif

  _VGA_setcursor (0, UI.parameter.row, UI.parameter.col); /* but redo that */

#if ASSEMBLY_TYPE == ASSEMBLY_DSES
  UI.parameter.base_address -= getds() << 4;
#else
#if USER_SUPPORT & VGA_MEMORY
  setgs (UI.parameter.base_address >> 4);
#endif
#endif

  VGA_VESA_setup_color ();

#if USER_SUPPORT & VGA_FREQUENCY
  VDBG ((" [get refresh frequency: "));
  UI.frequency = _CT_get_refresh_rate (mode);
  if (UI.frequency == 0)
      UI.frequency = _ET4000_get_refresh_rate (UI.parameter.heigh);
  VDBG (("%u Hz] ", UI.frequency));
#endif

#if USER_SUPPORT & VGA_MEMORY
  if (UI.parameter.nbcolor != 0) {
      UI.parameter.nbcolor = 16;
      UI.attributes.valid = no_attribute;
      UI.attributes.current = no_attribute;
      UI_setfgcolor (UI.stdcolor[blue]);
      UI_setbgcolor (UI.stdcolor[lightgray]);
      _VGA_blinking (0); /* disable it */
//      _VGA_set_block_specifier (0);
      }
    else {
      UI_setfgcolor (1);
      UI.bgcolor = 0;
      UI.attributes.valid = egatxt_attribute;
      UI_setattribute (no_attribute);
      }
#else /* VGA_MEMORY */
  if (UI.parameter.nbcolor > 2) {
      _VGA_blinking (0); /* disable it */
//      _VGA_set_block_specifier (0);
      UI.attributes.valid = no_attribute;
      UI.attributes.current = no_attribute;

#ifndef VGA16_XOR_TRICK
      if (UI.parameter.attr.isgraphic && UI.parameter.nbcolor != 256) {
	  /* graphic modes with > 1 page: BIOS do not have background color */
#else
      if (UI.parameter.attr.isgraphic && UI.parameter.nbcolor < 16) {
	  /* use XOR trick for VGA graphic 16 colors */
#endif
	  UI_setfgcolor (UI.stdcolor[cyan]);
	  UI_setbgcolor (UI.stdcolor[black]);
//	  _VGA_setbackground (UI.stdcolor[lightgray]); /* only whole screen */
	  }
	else {
	  UI_setfgcolor (UI.stdcolor[blue]);
	  UI_setbgcolor (UI.stdcolor[lightgray]);
	  }
      }
    else {
      if (UI.parameter.nbcolor == 0) {
	  if (UI.parameter.attr.isgraphic)
	      UI.attributes.valid = egagfx_attribute;
	    else
	      UI.attributes.valid = egatxt_attribute;
	  UI_setattribute (no_attribute);
	  }
      UI_setfgcolor (UI.stdcolor[white]);
      UI_setbgcolor (UI.stdcolor[black]);
      }
#endif /* VGA_MEMORY */

  if ((saved_mode & 0x8080) == 0)
      UI_clearscreen (); /* with the right color */

  VDBG (("success VGA_setmode\r\n"));
  return 0;
  }}

unsigned VGA_init (void)
  {bound_stack();{
  functionality_table_type		tmptable;
  static_functionality_table_type	tmpstatictable;
  unsigned i, extended_inited;
  int compaq_tmp;

  VDBG (("%s: ", __FUNCTION__));

  if (_VGA_get_functionality_info (&tmptable) != 0) {
      VDBG (("FAILED, functionality not supported.\r\n"));
      return 1;
      }

#if 0 /* I do not have a card which support this, so cannot test */
  VDBG (("DMQS: "));
  if ((i = _XGA_get_SVGA_DMQS_data_length()) != 0) {
      union {
	  char array[i];
	  struct DMQS_data_str data;
	  } buffer;
      if (_XGA_get_SVGA_DMQS_data (&buffer.data) != 0)
	  VDBG (("get data failed\r\n"));
	else {
	  struct DMQS_data_str *ptr = &buffer.data;
	  unsigned short XGA_cpt = 0;
	  while (1) {
	      char *msg = "unknown";
	      if (ptr->winsize == window_64K)
		  msg = "64 Kb";
		else if (ptr->winsize == window_1M)
		  msg = "1 Mb";
		else if (ptr->winsize == window_2M)
		  msg = "2 Mb";
		else if (ptr->winsize == window_4M)
		  msg = "4 Mb";
	      PRINTF ("XGA DMQS[%u]: %u Kb of video RAM at 0x%X,0000"
			" (window %s, monitor 0x%X)\r\n",
			  XGA_cpt++,
			  ptr->memory_div256K * 256,
			  ptr->winadr_div0x10000,
			  msg,
			  ptr->monitor_id);
	      if (ptr->DMQS_next_instance_offset == 0)
		  break;
	      ((char *)ptr) += ptr->DMQS_next_instance_offset;
	      if ((char *)ptr - (char *)&buffer >= (signed)i)
		  break;
	      }
	  }
      }
    else {
      VDBG (("none\r\n"));
      PRINTF ("no XGA DMQS\r\n");
      }
#endif

  VDBG (("VGA detected, static table at 0x%X",
	 tmptable.static_functionality_table));

  if (tmptable.static_functionality_table == 0) { /* DOSEMU w/o console */
      unsigned char nbcol, page;

      for (i = 0; i <= VIDEO_mode_max(); i++)
	  VIDEO_mode_invalidate (i);
      VIDEO_mode_revalidate (_VGA_getmode (&nbcol, &page));
      }
    else {
      for (i = 0; i < sizeof (static_functionality_table_type); i += 4)
	  *(unsigned *)((unsigned char *)&tmpstatictable + i) =
		peekl (tmptable.static_functionality_table + i);

      /*
       * Should we enable detection of modes 0x13..0x37 & 0x38..0x7F ?
       * Most VIDEO BIOS fill "mode_supported" only for first 0x13 modes.
       * This should be only usefull when there is no VESA BIOS; and VESA
       * modes are now only mapped in the VGA number to get "clearscreen"
       * compatibility working (get video mode number, set the same
       * video mode with bit 7 cleared).
       * BUT there is also some interresting text modes, which are
       * usually not mapped in VESA (132 columns, more than 25 lines...).
       * Clearly, some VIDEO BIOS will crash when setting/using some
       * of those extended modes.
       */
      VDBG ((", modes:"));
      extended_inited = (tmpstatictable.mode_supported[2] & 0xF0) != 0;
      for (i = 0; i < nbof(tmpstatictable.mode_supported); i++) {
	  VDBG ((" 0x%X,", tmpstatictable.mode_supported[i]));
	  if (i > 2)
	      extended_inited |= tmpstatictable.mode_supported[i] != 0;
	  }
      if (extended_inited)
	  extended_inited = 0x7F;
	else
	  extended_inited = 0x13;

      for (i = 0; i <= extended_inited; i++) {
#if 0
	  if ((tmpstatictable.mode_supported[i/8] & (1 << (i % 8))) == 0)
	      VIDEO_mode_invalidate (i); /* do not revalidate others */
#else
	  unsigned char shift;
	  unsigned short offset = divrem8 (i, &shift);

	  if ((tmpstatictable.mode_supported[offset] & (1 << shift)) == 0)
	      VIDEO_mode_invalidate (i); /* do not revalidate others */
#endif
	  }

      if ((compaq_tmp = _COMPAQ_getvideoramaddr()) >= 0) {
	  VDBG (("COMPAQ video RAM addr: 0x%X\r\n", compaq_tmp * 1024 * 1024));
	  VDBG (("COMPAQ Qvision: disable mode 0x6A.\r\n"));
	  /* This mode hard crashes at first VGA_clearscreen(),
	     and anyway is said to be 100x75 4 BPP text */
	  VIDEO_mode_invalidate (0x6A);
	  }
	else if (_PARADISE_installation_check() != -1) {
	  VDBG (("Paradise: disable mode 0x5B.\r\n"));
	  /* This mode crashes at VGA_getsize() (lidt done ?) */
	  VIDEO_mode_invalidate (0x5B);
	  }

#if SETUP & UPDATE_BOOTPARAM
      BOOT1_update_bootparam (); /* save all unsupported modes once */
#endif
      } /* DOSEMU */

#if USER_SUPPORT & VESA_SUPPORT
  if (VESA_ACTIVE() && !UI.parameter.BIOS_compatible) {
      /* some BIOS need VESA before VGA setmode: */
      if (_VESA_setmode (3) != 0)
	  VDBG ((" [Initial VESA_setmode in VGA_init failed!] "));
      if (VGA_setmode (3) != 0) /* for instance VGA_MEMORY */
	  VDBG ((" [Initial VGA_setmode failed!] "));
      }
#endif

  UI.function.getsize = VGA_getsize;
  UI.function.setmode = VGA_setmode;
  UI.function.getmode = VGA_getmode;
  VDBG (("\r\n"));

  return 0;
  }}

#if SETUP & UPDATE_BOOTPARAM
void setup_gujin_param_vga_mode (void)
  {
  __attribute__ ((unused)) struct gujin_param_str *dummy;
  unsigned i, index = 0, max = nbof (dummy->vga_mode), test;
  struct vga_mode_str mode;

  for (test = 0; test < 4; test++) {
      for (i = 0; i < UI.nbmode && index < max; i++) {
	  if (UI.mode[i].vesa)
	      continue;
	  if (   (test == 0) ? (UI.mode[i].text && UI.mode[i].bpp == 4)
	      : ((test == 1) ? (UI.mode[i].bpp == 8)
	      : ((test == 2) ? (!UI.mode[i].text && UI.mode[i].bpp == 4)
	      : /*(test == 3) ?*/ (UI.mode[i].bpp < 4)))
	      ) {
	      /* save all text mode first */
	      mode.isgraphic  = !UI.mode[i].text;
	      mode.number     = UI.mode[i].number;
	      mode.bpp        = UI.mode[i].bpp;
	      mode.font8x14   = UI.mode[i].font8x14;
	      if (mode.isgraphic)
		  mode.width_div8 = UI.mode[i].width / 8;
		else
		  mode.width_div8 = UI.mode[i].width;
	      mode.heigh = UI.mode[i].heigh;
	      VGA_set_saved_mode (index++, mode);
	      }
	  }
      }
  while (index < max)
      VGA_set_saved_mode (index++, ((struct vga_mode_str) {0}));
  }

void restore_gujin_param_vga_mode (void)
  {
  struct videomode_str mode = {vesa: 0};
  __attribute__ ((unused)) struct gujin_param_str *dummy;
  unsigned i, max = nbof (dummy->vga_mode);
  struct vga_mode_str savedmode;

  for (i = 0; i < max; i++) {
      VGA_get_saved_mode (i, &savedmode);
      mode.number = savedmode.number;
      mode.heigh =  savedmode.heigh;
      mode.text =   !savedmode.isgraphic;
      if (savedmode.isgraphic)
	  mode.width = savedmode.width_div8 * 8;
	else
	  mode.width = savedmode.width_div8;
      mode.bpp =    savedmode.bpp;
      mode.attr = (savedmode.bpp == 0);
      mode.font8x14 = savedmode.font8x14;
      if (mode.width == 0 || mode.heigh == 0)
	  break;
      add_valid_mode (&mode);
      }
  }
#endif /* UPDATE_BOOTPARAM */


#endif /* VGA_SUPPORT */

/*
 * The "VESA BIOS library" functions:
 */
#if USER_SUPPORT & VESA_SUPPORT
static unsigned
VESA_getmode (void)
//  {bound_stack();{
  {{
  unsigned short mode;

  if (_VESA_getmode (&mode) != 0)
      VDBG (("Error in %s, mode read 0x%X. ", __FUNCTION__, mode));
    else if ((mode & ~0xC000U) == UI.parameter.identification) {
      if (UI.parameter.winsize == VESA2_MARKER)
	  mode |= 0x4000; /* false VESA2 */
	else
	  mode &= ~0x4000;
      }
  return mode & ~0x8000;
  }}

/* Cursor movment: home at (0,0) */
static unsigned
VESA_setcursor (unsigned char row, unsigned char col)
  {
  if (COMMON_setcursor (row, col) != 0)
      return 1;

#if USER_SUPPORT & VESA_4BPP_TEXT
  if (   UI.parameter.memtype == mem_text
      && UI.parameter.EGA_compatible)
      _EGA_setcursor (row * UI.parameter.nbcol + col);
#endif

  return 0;
  }

#if USER_SUPPORT & VESA_WINDOW

static unsigned	FASTCALL VESA_pixeloffset (coord xy);

#if USER_SUPPORT & VESA_2WINDOWS
static unsigned	FASTCALL
VESA_changewin (unsigned offset)	/* NEVER called if linear */
  {
#if USER_SUPPORT & VESA_HARDWINDOW
  if (set_hardwin (UI.parameter.hardwindow, UI.parameter.offsetarray,
		   offset) != 0)
#endif
      {
      if (_VESA_setwin (DIV_WINDOW_GRANUL(offset), UI.parameter.winNo) != 0) {
	  VDBG (("%s error at 0x%X wingranul = %u (writewin: %u)\r\n",
	  __FUNCTION__, DIV_WINDOW_GRANUL(offset),
	  UI.parameter.wingranul, UI.parameter.winNo));
	  }
      }

  return ROUND_WINDOW_GRANUL(offset);
  }

static unsigned	FASTCALL
VESA_changereadwin (unsigned offset)
  {
#if USER_SUPPORT & VESA_HARDWINDOW
  if (set_hardwin (UI.parameter.readhardwindow, UI.parameter.readoffsetarray,
		   offset) != 0)
#endif
      {
      if (_VESA_setwin (DIV_WINDOW_GRANUL(offset), UI.parameter.winNo ^ 1) != 0) {
	  VDBG (("%s error at 0x%X wingranul = %u (writewin: %u ^ 1)\r\n",
		__FUNCTION__, DIV_WINDOW_GRANUL(offset),
		UI.parameter.wingranul, UI.parameter.winNo));
	  }
      }

  return ROUND_WINDOW_GRANUL(offset);
  }

static unsigned	FASTCALL
VESA_readpixeloffset (coord xy)
  {
  unsigned offset;

  if (UI.parameter.NbWin <= 1) /* only one read/write window or VESA2 */
      return VESA_pixeloffset (xy);

  offset = xy.y * UI.parameter.linelength
	   + (xy.x * UI.parameter.bitperpixel) / 8;

  /* unsigned calculus, still working when overflow: */
  if (UNUSUAL(offset - UI.parameter.winreadadr >= UI.parameter.winsize))
      UI.parameter.winreadadr = VESA_changereadwin (offset);
  return offset - UI.parameter.winreadadr;
  }

#else /* VESA_2WINDOWS */

static unsigned	FASTCALL
VESA_changewin (unsigned offset)
  {
#if USER_SUPPORT & VESA_HARDWINDOW
  if (set_hardwin (UI.parameter.hardwindow, UI.parameter.offsetarray, offset)
		!= 0)
#endif
      {
      unsigned char tmp = UI.parameter.winNo;

      if (_VESA_setwin (DIV_WINDOW_GRANUL(offset), tmp) != 0) {
	  VDBG (("%s error at 0x%X wingranul = %u (NbWin: %u, 1)\r\n",
	 __FUNCTION__, DIV_WINDOW_GRANUL(offset),
	 UI.parameter.wingranul, UI.parameter.NbWin));
	  }

      if (UI.parameter.NbWin == 2) {
	  if (_VESA_setwin (DIV_WINDOW_GRANUL(offset), (tmp ^ 1)) != 0) {
	      VDBG (("%s error at 0x%X wingranul = %u (NbWin: %u, 2)\r\n",
	     __FUNCTION__, DIV_WINDOW_GRANUL(offset),
	     UI.parameter.wingranul, UI.parameter.NbWin));
	      }
	  }
      }

  return ROUND_WINDOW_GRANUL(offset);
  }

#define VESA_readpixeloffset	VESA_pixeloffset

#endif /* VESA_2WINDOWS */

static unsigned	FASTCALL
VESA_pixeloffset (coord xy)
  {
  unsigned offset = xy.y * UI.parameter.linelength
		    + (xy.x * UI.parameter.bitperpixel) / 8;

  /* unsigned calculus, still working when overflow: */
  if (UNUSUAL(offset - UI.parameter.winadr >= UI.parameter.winsize))
      UI.parameter.winadr = VESA_changewin (offset);
  return offset - UI.parameter.winadr;
  }

static unsigned __attribute__ ((const))
VESA_max_size (unsigned offset_start, unsigned offset_end, unsigned winsize)
  {
#if 0 // bad assembly: two tests generated (with gcc-2.96)
  if (winsize == 0x10000) { /* nearly always for VESA1 */
      if (offset_end / 0x10000 != offset_start / 0x10000)
	  return 0x10000 - (offset_start % 0x10000);
      }
//    else if (winsize != VESA2_MARKER) {
    else if (winsize < 0x10000) {
      unsigned short us_winsize = (unsigned short)winsize;

      if (offset_end / us_winsize != offset_start / us_winsize)
	  return us_winsize - (offset_start % us_winsize);
      }
#else
  if (winsize <= 0x10000) {
      if (winsize == 0x10000) {
	  if (offset_end / 0x10000 != offset_start / 0x10000)
	      return 0x10000 - (offset_start % 0x10000);
	  }
	else {
#if 0
	  unsigned short us_winsize = (unsigned short)winsize;

	  if (offset_end / us_winsize != offset_start / us_winsize)
	      return us_winsize - (offset_start % us_winsize);
#else
	  unsigned short rem1, rem2;

	  if (   ul_div_us (offset_end, (unsigned short)winsize, &rem1)
	      != ul_div_us (offset_start, (unsigned short)winsize, &rem2))
	      return winsize - rem2;
#endif
	  }
      }
#endif
  return offset_end - offset_start;
  }

#else /* VESA_WINDOW */

static unsigned	inline VESA_changewin (unsigned offset)
  {
  return UI.parameter.winadr;
  }

#define VESA_readpixeloffset	VESA_pixeloffset

static unsigned	inline
VESA_pixeloffset (coord xy)
  {
  return xy.y * UI.parameter.linelength
       + (xy.x * UI.parameter.bitperpixel) / 8
       - UI.parameter.winadr;
  }

static unsigned inline __attribute__ ((const))
VESA_max_size (unsigned offset_start, unsigned offset_end, unsigned winsize)
  {
  return offset_end - offset_start;
  }

#endif /* VESA_WINDOW */

#if USER_SUPPORT & VESA_4BPP_EGA
static void FASTCALL
EGA_setpixel (coord xy, unsigned color)
  {
  _EGA_write_bitmask (0x80 >> (xy.x % 8));
  /* read to load the latch for the 7 other pixels: */
  /* Default: write mode 2 */
  GXCHGB (xy, color);
  }

static unsigned FASTCALL
EGA_getpixel (coord xy)
//  {bound_stack();{
  {{
  unsigned offset = VESA_readpixeloffset (xy);
  unsigned char mask = 0x80 >> (xy.x % 8), color = 0, plane;

  plane = UI.parameter.nbplane;
  while (plane--) {
      volatile unsigned char tmp;

      EGA_READ_PLANE (plane);
      tmp = gpeekb (offset);
      if (tmp & mask)
	  color |= 1 << plane;
      }

  return color;
  }}
#endif /* VESA_4BPP_EGA */

#if USER_SUPPORT & VESA_8BPP
static void FASTCALL
VESA_setpixel_color8bit (coord xy, unsigned color)
  {
  gpokeb (VESA_pixeloffset (xy), color);
  }

static unsigned FASTCALL
VESA_getpixel_color8bit (coord xy)
  {
  return gpeekb (VESA_readpixeloffset (xy));
  }
#endif /* VESA_8BPP */

#if USER_SUPPORT & VESA_16BPP
static void FASTCALL
VESA_setpixel_color16bit (coord xy, unsigned color)
  {
  gpokew (VESA_pixeloffset (xy), color);
  }

static unsigned FASTCALL
VESA_getpixel_color16bit (coord xy)
  {
  return gpeekw (VESA_readpixeloffset (xy));
  }
#endif /* VESA_16BPP */

#if USER_SUPPORT & VESA_24BPP
static void FASTCALL
VESA_setpixel_color24bit (coord xy, unsigned color)
//  {bound_stack();{
  {{
  unsigned pixoffset = VESA_pixeloffset (xy);

  pixoffset &= ~3;

  switch (xy.x & 3) {
    case 0:
      gpokew (pixoffset, color);
      gpokeb (pixoffset+2, color >> 16);
      break;
    case 1:
#if USER_SUPPORT & VESA_WINDOW
      if (UNUSUAL(pixoffset + 4 == UI.parameter.winsize)) {
	  gpokeb (pixoffset+3, color);
	  pixoffset += 4;
	  UI.parameter.winadr = VESA_changewin (pixoffset + UI.parameter.winadr);
	  gpokew (0, color >> 8);
	  }
	else
#endif /* VESA_WINDOW */
	  {
	  gpokeb (pixoffset+3, color);
	  gpokew (pixoffset+4, color >> 8);
	  }
      break;
    case 2:
#if USER_SUPPORT & VESA_WINDOW
      if (UNUSUAL(pixoffset + 4 == UI.parameter.winsize)) {
	  gpokew (pixoffset+2, color);
	  pixoffset += 4;
	  UI.parameter.winadr = VESA_changewin (pixoffset + UI.parameter.winadr);
	  gpokeb (0, color >> 16 );
	  }
	else
#endif /* VESA_WINDOW */
	  {
	  gpokew (pixoffset+2, color);
	  gpokeb (pixoffset+4, color >> 16);
	  }
      break;
    case 3:
      gpokeb (pixoffset+1, color);
      gpokew (pixoffset+2, color>>8);
      break;
    }
  }}

static unsigned FASTCALL
VESA_getpixel_color24bit (coord xy)
//  {bound_stack();{
  {{
  unsigned pixoffset = VESA_readpixeloffset (xy);

  pixoffset &= ~3;
  switch (xy.x & 3) {
    case 0:
      return gpeekl (pixoffset) & 0x00FFFFFFU;
    case 1: {
      unsigned returned;
#if USER_SUPPORT & VESA_WINDOW
      if (UNUSUAL(pixoffset + 4 == UI.parameter.winsize)) {
	  returned = gpeekb (pixoffset + 3);
	  pixoffset += 4;

#if USER_SUPPORT & VESA_2WINDOWS
	  if (UI.parameter.NbWin == 2)
	      UI.parameter.winreadadr = VESA_changereadwin (pixoffset + UI.parameter.winreadadr);
	    else
#endif
	      UI.parameter.winadr = VESA_changewin (pixoffset + UI.parameter.winadr);
	  returned |= gpeekw (0) << 8;
	  }
	else
#endif /* VESA_WINDOW */
	  {
	  returned = gpeekb (pixoffset + 3);
	  returned |= gpeekw (pixoffset + 4) << 8;
	  }
      return returned;
      }
    case 2: {
      unsigned returned;
#if USER_SUPPORT & VESA_WINDOW
      if (UNUSUAL(pixoffset + 4 == UI.parameter.winsize)) {
	  returned = gpeekw (pixoffset + 2);
	  pixoffset += 4;
#if USER_SUPPORT & VESA_2WINDOWS
	  if (UI.parameter.NbWin == 2)
	      UI.parameter.winreadadr = VESA_changereadwin (pixoffset + UI.parameter.winreadadr);
	    else
#endif
	      UI.parameter.winadr = VESA_changewin (pixoffset + UI.parameter.winadr);
	  returned |= gpeekb (0) << 16;
	  }
	else
#endif /* VESA_WINDOW */
	  {
	  returned = gpeekw (pixoffset + 2);
	  returned |= gpeekb (pixoffset + 4) << 16;
	  }
      return returned;
      }
    case 3:
      return gpeekl (pixoffset) >> 8;
    }
  return 0; /* will not happen often ! */
  }}
#endif /* VESA_24BPP */

#if USER_SUPPORT & VESA_32BPP
static void FASTCALL
VESA_setpixel_color32bit (coord xy, unsigned color)
  {
  gpokel (VESA_pixeloffset (xy), color);
  }

static unsigned FASTCALL
VESA_getpixel_color32bit (coord xy)
  {
  return gpeekl (VESA_readpixeloffset (xy));
  }
#endif /* VESA_32BPP */

static void
VESA_setattribute (struct attribute_str attr)
  {
  if (~ATTR2BYTE(UI.attributes.valid) & ATTR2BYTE(attr))
      ATTR2BYTE(attr) &= ATTR2BYTE(UI.attributes.valid);
  UI.attributes.current = attr;
  }

static void
GRAPHIC_drawchar (unsigned char C)
  {bound_stack();{
  coord xy = {
      x: UI.parameter.col * UI.parameter.charwidth,
      y: UI.parameter.row * UI.parameter.charheigh
      };
  unsigned short charheigh = UI.parameter.charheigh + 1;
  farptr font = UI.parameter.font;
  unsigned fgcolor, bgcolor;

  if (UI.parameter.charwidth >= 32)
      VDBG (("ERROR: charwidth too high in %s (%u).\r\n",
		__FUNCTION__, UI.parameter.charwidth));

  if (UI.attributes.current.reverse) {
      fgcolor = UI.bgcolor;
      bgcolor = UI.fgcolor;
      }
    else {
      fgcolor = UI.fgcolor;
      bgcolor = UI.bgcolor;
      }

  font += C * UI.parameter.charwidth/8
	    * UI.parameter.charheigh;	/* 'C' is unsigned */

  while	(--charheigh != 0) {
      unsigned mask = 1 << (UI.parameter.charwidth - 1),
	draw = peekl (font);

      if (UI.attributes.current.underline && charheigh == 2)
	  draw = 0xFFFFFFFFU;

      while (mask != 0) {
	  if (draw & mask)
	      UI_setpixel (xy, fgcolor);
	    else if (!UI.attributes.current.transparent)
	      UI_setpixel (xy, bgcolor);
	  mask >>= 1;
	  xy.x++;
	  }
      xy.x -= UI.parameter.charwidth;
      font += UI.parameter.charwidth / 8;
      xy.y++;
      }
  }}

#if USER_SUPPORT & VESA_4BPP_TEXT
static void VESA_drawchar (unsigned char C)
  {
  if (UI.parameter.memtype == mem_text) {
      coord xy = {x: UI.parameter.col, y: UI.parameter.row};
      UI_setpixel (xy,	(UI.fgcolor << 8)
				| (UI.bgcolor << 12) | C);
      }
    else
      GRAPHIC_drawchar (C);
  }
#else
static void inline VESA_drawchar (unsigned char C)
  {
  GRAPHIC_drawchar (C);
  }
#endif /* VESA_4BPP_TEXT */

#if USER_SUPPORT & VESA_4BPP_EGA
static void
EGA_plotHline (coord xy, unsigned short xend, unsigned color)
  {bound_stack();{
  if (xend < xy.x) {
      VDBG (("%s called with xy.x = %u and xend = %u",
	__FUNCTION__, xy.x, xend));
      return;
      }
#ifdef VERY_SLOW
  while (xy.x < xend) {
      UI_setpixel (xy, color);
      xy.x++;
      }
#else /* VERY_SLOW */
  {
  const unsigned mask1 = 0x07, /* first align on 8 bits */
		 mask2 = 0x1F; /* then align on 32 bits */
  unsigned offset, offsetend;
//  CR0_type CR0 = ({ force_cache (EGA_plotHline, sizeof(EGA_plotHline)); disable_cache (); });

  if (color > UI.parameter.nbcolor)
      VDBG (("ERROR: %s called with color=%u\r\n", __FUNCTION__, color));

  /* Default: write mode 2 */
  if ((xy.x & ~mask1) == (xend & ~mask1)) {
      _EGA_write_bitmask (   (0xFF >> (xy.x & mask1))
			  & ~(0xFF >> (xend & mask1)  ));
      GXCHGB (xy, color);
      xy.x = xend;
//      enable_cache (CR0);
      return;
      }
    else {
      coord xyend;
      if ((xy.x & mask1) != 0 && xy.x < xend) {
	  _EGA_write_bitmask (0xFF >> (xy.x & mask1));
	  GXCHGB (xy, color);
	  xy.x = (xy.x + mask1) & ~mask1;
	  }
      _EGA_write_bitmask (0xFF);
      while ((xy.x & mask2) != 0 && xy.x < (xend & ~7)) {
	  gpokeb (VESA_pixeloffset (xy), color);
	  xy.x += 8;
	  }

      if ((xend & mask1) != 0 && xy.x < xend) {
	  _EGA_write_bitmask (~(0xFF >> (xend & mask1)));
	  xyend.y = xy.y;
	  xyend.x = xend;
	  GXCHGB (xyend, color);
	  xend = xend & ~mask1;
	  }
      _EGA_write_bitmask (0xFF);
      while ((xend & mask2) != 0 && xy.x < xend) {
	  xend -= 8;
	  xyend.y = xy.y;
	  xyend.x = xend;
	  gpokeb (VESA_pixeloffset (xyend), color);
	  }
      }

  offset = (UI.parameter.linelength * xy.y) + (xy.x >> 3);
  offsetend = (UI.parameter.linelength * xy.y) + (xend >> 3);

  color |= color << 8;
  color |= color << 16;

  while (offset < offsetend) {
      unsigned size = VESA_max_size (offset, offsetend, UI.parameter.winsize);

      if (   offset - UI.parameter.winadr >= UI.parameter.winsize
	  || offset - UI.parameter.winadr + size >= UI.parameter.winsize)
	  UI.parameter.winadr = VESA_changewin (offset); /* wingranul */
      offset = rep_gpokel (size / 4, offset, UI.parameter.winadr, color);
      }
//  enable_cache (CR0);
  }
#endif /* VERY_SLOW */
  }}
#endif /* VESA_4BPP_EGA */

#if USER_SUPPORT & VESA_24BPP
static void
VESA_24BPP_plotHline (coord xy, unsigned short xend, unsigned color)
  {bound_stack();{
  if (xend < xy.x) {
      VDBG (("%s called with xy.x = %u and xend = %u",
	__FUNCTION__, xy.x, xend));
      return;
      }
#ifdef VERY_SLOW
  while (xy.x < xend) {
      UI_setpixel (xy, color);
      xy.x++;
      }
#else /* VERY_SLOW */
  {
  unsigned offset, offsetend;
//  CR0_type CR0 = ({ force_cache (VESA_24BPP_plotHline, sizeof(VESA_24BPP_plotHline)); disable_cache (); });

  if (color > UI.parameter.nbcolor)
      VDBG (("ERROR: %s called with color=%u\r\n", __FUNCTION__, color));

  offset = (UI.parameter.linelength * xy.y) + (xy.x * 3);
  offsetend = (UI.parameter.linelength * xy.y) + (xend * 3);

  while (offset < offsetend) {
      unsigned size = VESA_max_size (offset, offsetend, UI.parameter.winsize);

      if (   offset - UI.parameter.winadr >= UI.parameter.winsize
	  || offset - UI.parameter.winadr + size >= UI.parameter.winsize)
	  UI.parameter.winadr = VESA_changewin (offset); /* wingranul */
      offset = rep_g3pokel (size, offset, UI.parameter.winadr, color);
      }
//  enable_cache (CR0);
  }
#endif /* VERY_SLOW */
  }}
#endif /* VESA_24BPP */

#if USER_SUPPORT & (VESA_8BPP | VESA_16BPP | VESA_32BPP)
static void
VESA_plotHline (coord xy, unsigned short xend, unsigned color)
  {bound_stack();{
  if (xend < xy.x) {
      VDBG (("%s called with xy.x = %u and xend = %u",
	__FUNCTION__, xy.x, xend));
      return;
      }
#ifdef VERY_SLOW
  while (xy.x < xend) {
      UI_setpixel (xy, color);
      xy.x++;
      }
#else /* VERY_SLOW */
  {
  unsigned offset, offsetend;
  unsigned char mask, shift, cptshift;
//  CR0_type CR0 = ({ force_cache (VESA_plotHline, sizeof(VESA_plotHline)); disable_cache (); });

  if (   color > UI.parameter.nbcolor
#if USER_SUPPORT & VESA_4BPP_TEXT
      && UI.parameter.memtype != mem_text
#endif
      )
      VDBG (("ERROR: %s called with color=%u\r\n", __FUNCTION__, color));

  /* FIXME: This is too generic and could even handle 1, 2, 4 BPP,
     but we do not need that here: those VESA modes are refused. */
  shift = my_ffs (UI.parameter.bitperpixel) - 1; /* 8BPP => shift = 3 */
  mask = 0x1F >> shift;

  while ((xy.x & mask) != 0 && xy.x < xend) {
      UI_setpixel (xy, color);
      xy.x++;
      }
  while ((xend & mask) != 0 && xy.x < xend) {
      coord xyend = {x: --xend, y: xy.y};
      UI_setpixel (xyend, color);
      }

  for (cptshift = shift; cptshift <= 4; cptshift++)
      color |= color << (1 << cptshift);

  offset = UI.parameter.linelength * xy.y + ((xy.x << shift) >> 3);
  offsetend = UI.parameter.linelength * xy.y + ((xend << shift) >> 3);

  while (offset < offsetend) {
      unsigned size = VESA_max_size (offset, offsetend, UI.parameter.winsize);

      if (   offset - UI.parameter.winadr >= UI.parameter.winsize
	  || offset - UI.parameter.winadr + size >= UI.parameter.winsize)
	  UI.parameter.winadr = VESA_changewin (offset); /* wingranul */
      offset = rep_gpokel (size / 4, offset, UI.parameter.winadr, color);
      }
//  enable_cache (CR0);
  }
#endif /* VERY_SLOW */
  }}
#endif /* (VESA_8BPP | VESA_16BPP | VESA_32BPP) */

static void
VESA_scroll (signed int charheigh) /* works now only with charheigh > 0 */
  {bound_stack();{
#ifdef VERY_SLOW
  coord xy = {y: 0}, xydest;
  do {
      xydest.x = xy.x = UI.parameter.width;
      xydest.y = xy.y + charheigh;
      do {
	  xy.x--;
	  xydest.x--;
	  UI_setpixel (xy, UI_getpixel (xydest));
	  } while (xy.x != 0);
      } while (++xy.y < UI.parameter.heigh - charheigh);
#else

  unsigned offsetsrc = UI.parameter.linelength * charheigh,
	   offsetdst = 0,
	   offsetend = UI.parameter.linelength * UI.parameter.heigh;

//  CR0_type CR0 = ({ force_cache (VESA_scroll, sizeof(VESA_scroll)); disable_cache (); });

#if USER_SUPPORT & VESA_4BPP_EGA
  unsigned char plane;
  unsigned save_src, save_dst, save_end;
  struct EGA_mode_str EGA_mode = {
      write_source : cpuORsetreset,
      enable_tristate : 0,
      read_mode : EGA_DEFAULT_READ_MODE,
      enable_odd_even_addr : 0,
      CGA_shift_register : 0,
      enable_256_color : 0,
      reserved : 0
      };

//  bound_stack();	/* failure treated here in write mode 2 ! */

  if (UI.parameter.memtype == mem_EGA) {
      _EGA_read_write_mode (EGA_mode);      /* write mode 0 */
      _EGA_write_bitmask (0xFF);
      }

  if (   UI.parameter.nbplane != 1
      && UI.parameter.memtype != mem_EGA
      && UI.parameter.memtype != mem_text)
      VDBG (("ERROR %s: nbplane = %u, memtype = %u! ",
	     __FUNCTION__, UI.parameter.nbplane,
	     (unsigned)UI.parameter.memtype));

  /*
   * Limit the copy size of each plane because of the visual effect:
   */
  save_end = offsetend;
  while (offsetsrc < save_end) {
      if (UI.parameter.memtype == mem_EGA)
	  offsetend = MIN (save_end, offsetsrc + 8*1024);

      plane = UI.parameter.nbplane;
      save_src = offsetsrc;
      save_dst = offsetdst;
      while (plane--) {
	  if (UI.parameter.memtype == mem_EGA) {
	      EGA_READ_PLANE (plane);
	      _EGA_enable_plane (1 << plane);
	      offsetsrc = save_src;
	      offsetdst = save_dst;
	      }
#else
//  bound_stack();
#endif

  while (offsetsrc < offsetend) {
      unsigned size, hastochangewin;
      unsigned winadr, winreadadr;

      size = VESA_max_size (offsetsrc, offsetend, UI.parameter.winsize);
      size = VESA_max_size (offsetdst, offsetdst+size, UI.parameter.winsize);

      if (size % 4 != 0 || size == 0 || size > UI.parameter.winsize)
	  VDBG (("ERROR %s: size = %u! ", __FUNCTION__, size));

#if USER_SUPPORT & VESA_2WINDOWS
      if (UI.parameter.NbWin == 2) {
	  if (   offsetsrc - UI.parameter.winreadadr >= UI.parameter.winsize
	      || offsetsrc - UI.parameter.winreadadr + size >= UI.parameter.winsize)
	      UI.parameter.winreadadr = VESA_changereadwin (offsetsrc);
	  winreadadr = UI.parameter.winreadadr;

	  if (   offsetdst - UI.parameter.winadr >= UI.parameter.winsize
	      || offsetdst - UI.parameter.winadr + size >= UI.parameter.winsize)
	      UI.parameter.winadr = VESA_changewin (offsetdst); /* wingranul */
	  winadr = UI.parameter.winadr;

	  hastochangewin = 0;
	  }
	else
#endif
	  {
	  if (   offsetsrc - UI.parameter.winadr >= UI.parameter.winsize
	      || offsetsrc - UI.parameter.winadr + size >= UI.parameter.winsize)
	      UI.parameter.winadr = VESA_changewin (offsetsrc);	/* wingranul */

	  winadr = winreadadr = UI.parameter.winadr;
	  hastochangewin = (   offsetdst - UI.parameter.winadr >= UI.parameter.winsize
			    || offsetdst - UI.parameter.winadr + size >= UI.parameter.winsize);
	  }

#if USER_SUPPORT & VESA_WINDOW
      if (hastochangewin && UI.parameter.NbWin != 0) {
	  /* We have to go through a buffer */
	  unsigned treated = MIN (size, 1024) / 4;
	  unsigned buffer[256]; //treated]; /* get an align of 4 */

	  offsetsrc = winadr + rep_movel_fromg (buffer, offsetsrc - winadr, treated);

//	  VDBG (("#0x%X ", treated*4));

	  UI.parameter.winadr = VESA_changewin (offsetdst); /* wingranul, hastochangewin is true */

	  winadr = UI.parameter.winadr;
	  offsetdst = winadr + rep_movel_tog (offsetdst - winadr, buffer, treated);
	  }
	else
#endif
	  {
	  /* We can do here a direct copy, i.e. without buffer */
//	  VDBG (("=0x%X ", size));

	  rep_movel_gtog (offsetdst - winadr, offsetsrc - winreadadr, size/4);
	  offsetsrc += size;
	  offsetdst += size;
	  }
      }

#if USER_SUPPORT & VESA_4BPP_EGA
	  } /* while (plane--) */
      }  /* while (offsetsrc < save_end) */

  if (UI.parameter.memtype == mem_EGA) {
      _EGA_enable_plane (0x0F);
      EGA_mode.write_source = cpuASsetreset; /* Default: write mode 2 */
      _EGA_read_write_mode (EGA_mode);
      }
#endif
//  enable_cache (CR0);
#endif /* VERY_SLOW */
  }}

static void
VESA_clearscreen (void)
  {
  unsigned fillcolor;
  coord xy = {x: 0, y: 0} ;

  if (UI.parameter.memtype == mem_text)
      fillcolor = (UI.bgcolor << 12) | (UI.fgcolor << 8) | ' ';
    else
      fillcolor = UI.bgcolor;

  while (xy.y < UI.parameter.heigh) {
      UI_plotHline (xy, UI.parameter.width, fillcolor);
      xy.y++;
      }

  VESA_setcursor (0, 0);
  }

static void
VESA_putstr (const char *str)
  {
  while	(*str) {
      unsigned nblines = 1;

      switch (*str) {
	case '\007': /* ^G */
	  BOOT1_putstr ("\007");
	  break;
	case '\010': /* ^H */
	  if (UI.parameter.col > 0)
	      UI.parameter.col -= 1;
//	  VESA_drawchar (' ');	// for attribute "transparent"
	  break;
	default:
	  VESA_drawchar (*str);
	  if (++UI.parameter.col >= UI.parameter.nbcol) {
	      UI.parameter.col = 0;	/* automatic CRLF */
	case '\n': /* it is valid in C */
	      if (++UI.parameter.row >= UI.parameter.nbrow) {
		  /* estimate how many lines to scroll on this string */
		  const char *tmp = str + 1;
		  unsigned heigh;

		  while (*tmp != 0)
		      if (*tmp++ == '\n')
			  nblines ++;
		  if (UI.parameter.memtype == mem_text)
		      heigh = nblines;
		    else
		      heigh = nblines * UI.parameter.charheigh;
		  VESA_scroll (heigh);
		  UI.parameter.row = UI.parameter.nbrow - nblines;

	case '\r': /* it is valid in C */ {
		  coord xy = {x: UI.parameter.col, y: UI.parameter.row};
		  unsigned short end_y = UI.parameter.row + nblines, width;
		  unsigned fillcolor;

		  if (UI.parameter.memtype == mem_text) {
		      fillcolor = (UI.bgcolor << 12) | (UI.fgcolor << 8) | ' ';
		      width = UI.parameter.nbcol;
		      }
		    else {
		      xy.x *= UI.parameter.charwidth;
		      xy.y *= UI.parameter.charheigh;
		      end_y *= UI.parameter.charheigh;
		      fillcolor = UI.bgcolor;
		      width = UI.parameter.width;
		      }
		  while (xy.y < end_y) {
		      UI_plotHline (xy, width, fillcolor);
		      xy.y++;
		      }
		  }
		  UI.parameter.col = 0;
		  }
	      }
	  break;
	}
      str++;
      }
#if USER_SUPPORT & VESA_4BPP_TEXT
  VESA_setcursor (UI.parameter.row, UI.parameter.col); // display cursor
#endif
  }

static unsigned
VESA_setDACblock (unsigned short startindex, unsigned short nb,
		  const struct DAC_str *palette)
  {
#if USER_SUPPORT & (VESA_4BPP_TEXT | VESA_4BPP_EGA | VESA_8BPP)
  if (_VESA_setDACarray (startindex, nb, 0, (VESA_RGB *)palette) != 0)
      _EGA_setDACarray (startindex, nb, palette);
#endif /* VESA_4BPP_TEXT | VESA_4BPP_EGA | VESA_8BPP */
  return 0;
  }

#if USER_SUPPORT & VESA_WINDOW
static unsigned
VESA_setup_window (struct video_parameter_str *param, VESA_modeinfo_type *modeinfo)
  {bound_stack();{
  if (modeinfo->WinGranularity == 0 && modeinfo->MemoryModel != mem_text) {
      VDBG ((" [modeinfo.WinGranularity == 0, force linear graphic mode] \r\n"));
      return 0x300;
      }

#if USER_SUPPORT & VESA_2WINDOWS
  param->readdelta = 0; /* HAS TO BE null if NbWin < 2 */
#endif

  if (   !modeinfo->WinAAttributes.win_writable
      && !modeinfo->WinAAttributes.win_readable)
      modeinfo->WinAAttributes.win_exist = 0;

  if (   !modeinfo->WinBAttributes.win_writable
      && !modeinfo->WinBAttributes.win_readable)
      modeinfo->WinBAttributes.win_exist = 0;

  if (!modeinfo->WinBAttributes.win_exist) {
      if (!modeinfo->WinAAttributes.win_exist) {
	  VDBG ((" [no window exists] "));
	  return 0x301;
	  }
      if (!modeinfo->WinAAttributes.win_writable) {
	  VDBG ((" [only window A not writeable] "));
	  return 0x302;
	  }
      if (!modeinfo->WinAAttributes.win_readable) {
	  VDBG ((" [only window A not readable] "));
	  return 0x303;
	  }
      param->base_address = modeinfo->WinASegment << 4;
      param->winNo = 0;
      VDBG ((" [using winA] "));
      param->NbWin = 1;
      }
    else if (!modeinfo->WinAAttributes.win_exist) {
      if (!modeinfo->WinBAttributes.win_writable) {
	  VDBG ((" [only window B not writeable] "));
	  return 0x304;
	  }
      if (!modeinfo->WinBAttributes.win_readable) {
	  VDBG ((" [only window B not readable] "));
	  return 0x305;
	  }
      param->base_address = modeinfo->WinBSegment << 4;
      param->winNo = 1;
      VDBG ((" [using winB] "));
      param->NbWin = 1;
      }
    else {
#if USER_SUPPORT & VESA_2WINDOWS
      if (   modeinfo->WinAAttributes.win_writable
	  && modeinfo->WinBAttributes.win_readable) {
	  VDBG ((" [using winA write / winB read] "));
	  param->winNo = 0; /* winA as write only */
	  param->base_address = modeinfo->WinASegment << 4;
	  param->readdelta = ((unsigned)modeinfo->WinBSegment
				- modeinfo->WinASegment) << 4;
	  }
	else if (   modeinfo->WinBAttributes.win_writable
	 && modeinfo->WinAAttributes.win_readable) {
	  VDBG ((" [using winA read / winB write] "));
	  param->winNo = 1; /* winB as write only */
	  param->base_address = modeinfo->WinBSegment << 4;
	  param->readdelta = ((unsigned)modeinfo->WinASegment
				- modeinfo->WinBSegment) << 4;
	  }
	else {
	  VDBG ((" [either winA & winB read only or winA & winB write only] "));
	  return 0x306;
	  }
#else
      if (modeinfo->WinASegment != modeinfo->WinBSegment) {
	  if (   modeinfo->WinAAttributes.win_writable
	      && modeinfo->WinAAttributes.win_readable) {
	      VDBG ((" [using small winA read/write] "));
	      param->base_address = modeinfo->WinASegment << 4;
	      param->winNo = 0;
	      }
	    else if (   modeinfo->WinBAttributes.win_writable
	     && modeinfo->WinBAttributes.win_readable) {
	      VDBG ((" [using small winB read/write] "));
	      param->base_address = modeinfo->WinBSegment << 4;
	      param->winNo = 1;
	      }
	    else {
	      VDBG (("unsupported winA seg != winB seg"));
	      return 0x307;
	      }
	  }
	else {
	  param->base_address = modeinfo->WinASegment << 4;
	  param->winNo = 0;
	  VDBG (("using winA AND winB "));
	  }
#endif
      param->NbWin = 2;
      }
  param->winsize = modeinfo->WinSize * 1024;
  return 0;
  }}
#else  /* VESA_WINDOW */
static unsigned inline
VESA_setup_window (struct video_parameter_str *param, VESA_modeinfo_type *modeinfo)
  {
  return -1;
  }
#endif /* VESA_WINDOW */

#if USER_SUPPORT & VESA_LINEAR
static unsigned
VESA_setup_linear (struct video_parameter_str *param,
		   VESA_modeinfo_type *modeinfo)
  {bound_stack();{

/* FIXME: "S3 Incorporated. 86C775/86C785" report:
   VESA2 call successfull, version: 2.0, capabilities: 0,
	videoram: 32 x 64 Kbytes, name:[S3 Incorporated. 86C775/86C785], modes:
	... mode 0x104: graphic 1024x768, 4 bpp ...
	VESA_getsize mode 0x4104: graphic 1024x768, 4 bpp, local char: 8x16,
	winA = 7, winB = 0, wingranul = 64 Kb, size = 64 Kb winA start = 0xA000,
	winB start = 0xA000 using linear framebuffer at 0xF0000000,
	MemoryModel 0x3, BytesPerScanLine = 128,
	BitsPerPixel 4 / NumberOfPlanes 4, NumberOfImagePages 4,
	red mask/off: 0/0, green mask/off: 0/0, blue mask/off: 0/0
	get VGA font farptr 0xC0006AE0, nbrow = 63.
	[VESA_setmode: EGA_s/getpixel]  [setting palette 16 colors] ,
	success VESA_setmode
	But does not work (linear mode MemoryModel 0x3).
	Works nicely in mode 0x4106: graphic 1280x1024, 4 bpp because
	refuses linear then.
  AllWin: outb (0x3C4, 0x4); outw (0x3D4, 0x4838); outw (0x3D4, 0xA539);
	  outw (0x3D4, 0x6A ^ (page << 2));
  Shall we disable linear mode in MemoryModel 0x3 ?
*/
  if ((unsigned)modeinfo->BytesPerScanLine * modeinfo->YResolution
	 > modeinfo->WinSize * 1024U) {
      if (!modeinfo->ModeAttributes.linear_framebuffer_supported) {
	  VDBG (("no linear buffer supported.\r\n"));
	  return 0x201;
	  }
      if (getsw() & 1) {
	  VDBG (("Cannot use linear framebuffer, "
		  "processor not in real mode.\r\n"));
	  return 0x202; /* Under DOS, maybe do something else... */
	  }
      if (modeinfo->PhysBasePtr <= 0x100000) {
	  VDBG (("graphic linear framebuffer in the first megabyte! (0x%X)\r\n",
		 modeinfo->PhysBasePtr));
	  return 0x203;
	  }
      if (modeinfo->MemoryModel == 3) {
	  VDBG (("Do not know how to handle 4 BPP with linear video memory\r\n"));
	  return 0x204;
	  }
      param->base_address = 0xA8000; /* != A0000 to know linear addressing */
      }
    else {
      /*
       * Support text modes if only VESA_LINEAR is defined, normally
       * they are not maped with (modeinfo->PhysBasePtr == 0xB800), also
       * 800x600x16 can be supported because it fits in one 64 Kb window,
       * normally modeinfo->ModeAttributes.linear_framebuffer_supported
       * is false. Some 320x240*256 mode can be there too.
       */
      VDBG ((" [VESA2 mapping linear framebuffer to window (<64K mode)] "));
      param->base_address = modeinfo->PhysBasePtr =
		((unsigned)(modeinfo->WinASegment
				 ? modeinfo->WinASegment
				 : modeinfo->WinBSegment)) << 4;
      }
#if USER_SUPPORT & VESA_WINDOW
  param->NbWin = 0;
  param->winNo = 0;
  param->wingranul = 0;

#if USER_SUPPORT & VESA_2WINDOWS
  param->readdelta = 0; /* HAS TO BE null if NbWin < 2 */
#endif

#endif
  param->winsize = VESA2_MARKER;
  param->winadr = param->base_address - modeinfo->PhysBasePtr;

  VDBG (("using linear framebuffer at 0x%X, ", modeinfo->PhysBasePtr));
  return 0;
  }}
#else  /* VESA_LINEAR */
static unsigned
VESA_setup_linear (struct video_parameter_str *param, VESA_modeinfo_type *modeinfo)
  {
  return -1;
  }
#endif /* VESA_LINEAR */

static unsigned
VESA_getsize (unsigned mode, struct video_parameter_str *param)
  {bound_stack();{
  VESA_modeinfo_type modeinfo;
  unsigned saved_mode = mode & ~0x8000;

#if 1
  /* Alliance Semiconductor PMAT3D 4 Mbytes VESA 1.2 probing mode <= 0x7F,
     division by 0 */
  extern volatile unsigned setjump_key;
  extern setjmp_buf setjump_buffer;

  VDBG (("%s mode 0x%X: ", __FUNCTION__, mode));

  setjump_key = 0x45746965;
  if (setjmp (&setjump_buffer) != 0) {
      setjump_key = 0;
      VDBG (("have hit an exception, like division by 0"
	" or invalid instruction, or timeout.\r\n"));
      return 0x700;
      }
#else
  VDBG (("%s mode 0x%X: ", __FUNCTION__, mode));
#endif

  checkptr (param, "VESA_getsize invalid parameter.\r\n");

  memset (param, 0, sizeof(*param));
  mode &= ~0x8000;
  mode &= ~0x4000;
  param->identification = mode;

  if (_VESA_getmodeinfo (&modeinfo, mode) != 0) {
      VDBG (("_VESA_getmodeinfo error.\r\n"));
      return 0x101;
      }

#if 1
  setjump_key = 0;
#endif

  if (   modeinfo.ModeAttributes.optional_info_available == 0
      || modeinfo.ModeAttributes.mode_supported == 0) {
      VDBG (("_VESA_getmodeinfo: no optional info / unsupported (0x%X)\r\n",
	    *(unsigned short*) &modeinfo.ModeAttributes));
      return 0x102;
      }

  VDBG (("%s %ux%u, %u bpp, local char: %ux%u, ",
	 (modeinfo.ModeAttributes.graphic_mode)? "graphic" : "text",
	 modeinfo.XResolution, modeinfo.YResolution,
	 modeinfo.BitsPerPixel,
	 modeinfo.XCharSize, modeinfo.YCharSize));

  if (modeinfo.ModeAttributes.graphic_mode) {
      /* Refuse now a video mode if there is clearly not enought
	 memory to map it - it should have the unsupported bit,
	 but... */
      if ((unsigned)modeinfo.XResolution * modeinfo.YResolution
		* modeinfo.BitsPerPixel / (8 * 64 * 1024)
		 > UI.info.vesa.memory_64Kb) {
	  VDBG (("_VESA_getmodeinfo: no enough memory on board!\r\n"));
	  return 0x103;
	  }
      }

  VDBG (("winA = %u, winB = %u, wingranul = %u Kb, size = %u Kb "
	"winA start = 0x%X, winB start = 0x%X ",
	 *(unsigned char *)&modeinfo.WinAAttributes,
	 *(unsigned char *)&modeinfo.WinBAttributes,
	 modeinfo.WinGranularity, modeinfo.WinSize,
	 modeinfo.WinASegment, modeinfo.WinBSegment));

#if USER_SUPPORT & VESA_4BPP_TEXT
  if (modeinfo.MemoryModel == mem_text && modeinfo.NumberOfPlanes == 4) {
      if (modeinfo.BytesPerScanLine != 2 * modeinfo.XResolution) {
	  VDBG ((" [correcting linelength: old=%u, new=%u] ",
			modeinfo.BytesPerScanLine, 2 * modeinfo.XResolution));
	  modeinfo.BytesPerScanLine = 2 * modeinfo.XResolution;
	  }
      if (modeinfo.WinASegment != 0xB800) {
	  VDBG ((" [correcting WinASegment: old=0x%X, new=0x%X] ",
		modeinfo.WinASegment, 0xB800));
	  modeinfo.WinASegment = 0xB800;
	  }
      if (modeinfo.BitsPerPixel != 4) {
	  VDBG ((" [correcting BitsPerPixel: old=%u, new=4] ",
		 modeinfo.BitsPerPixel));
	  modeinfo.BitsPerPixel = 4;
	  }
      }
#endif /* VESA_4BPP_TEXT */

  if (mode == saved_mode) {
      unsigned tmp = VESA_setup_window (param, &modeinfo);
#if BUILD_APPLICATION()
      if (tmp != 0) {
	  VDBG ((" [window failed, try linear] "));
	  tmp = VESA_setup_linear (param, &modeinfo);
	  }
#endif
      if (tmp != 0) {
	  VDBG ((" cannot acces video memory.\r\n"));
	  return tmp;
	  }
      }
    else {
      unsigned tmp = VESA_setup_linear (param, &modeinfo);
#if BUILD_APPLICATION()
      if (tmp != 0) {
	  VDBG ((" [linear failed, try window] "));
	  tmp = VESA_setup_window (param, &modeinfo);
	  }
#endif
      if (tmp != 0) {
	  VDBG ((" cannot acces video memory.\r\n"));
	  return tmp;
	  }
      }

#if USER_SUPPORT & VESA_WINDOW
  if (param->winsize != VESA2_MARKER) {
#if USER_SUPPORT & VESA_WINFUNCTION
      VDBG (("addr positionning fct = 0x%X ", modeinfo.WinFuncPtr));
      param->addrpositionfarfct = modeinfo.WinFuncPtr;
#endif

      INIT_WINDOW_GRANUL (modeinfo.WinGranularity * 1024, param);
      if (DISPLAY_WINDOW_GRANUL(param) != modeinfo.WinGranularity * 1024U) {
	  VDBG ((" [Strange VESA window granularity: %u bytes, "
		 "recompile with VESA_STRANGE_WINDOW defined!]\r\n",
		    modeinfo.WinGranularity * 1024));
	  return 0x104;
	  }

      }
#endif

  /* for offsets calculus, our bitperpixel is not necessarily linked to
     the number of colors (in case of different color planes): */
  VDBG (("MemoryModel 0x%X, BytesPerScanLine = %u, "
	 "BitsPerPixel %u / NumberOfPlanes %u, NumberOfImagePages %u, ",
	  modeinfo.MemoryModel, modeinfo.BytesPerScanLine,
	  modeinfo.BitsPerPixel, modeinfo.NumberOfPlanes,
	  modeinfo.NumberOfImagePages));

  param->memtype = modeinfo.MemoryModel;
  param->linelength = modeinfo.BytesPerScanLine;
  param->nbimage = modeinfo.NumberOfImagePages;

  if ( 0
#if USER_SUPPORT & VESA_4BPP_TEXT
      || (modeinfo.BitsPerPixel == 4 && modeinfo.MemoryModel == mem_text)
#endif
#if USER_SUPPORT & VESA_4BPP_EGA
      || (modeinfo.NumberOfPlanes == 4 && modeinfo.BitsPerPixel == 4 )
#endif
#if USER_SUPPORT & VESA_8BPP
      || (modeinfo.NumberOfPlanes <= 1 && modeinfo.BitsPerPixel == 8 )
#endif
#if USER_SUPPORT & VESA_16BPP
      || (modeinfo.NumberOfPlanes <= 1 && modeinfo.BitsPerPixel == 16)
      || (modeinfo.NumberOfPlanes <= 1 && modeinfo.BitsPerPixel == 15)
#endif
#if USER_SUPPORT & VESA_24BPP
      || (modeinfo.NumberOfPlanes <= 1 && modeinfo.BitsPerPixel == 24)
#endif
#if USER_SUPPORT & VESA_32BPP
      || (modeinfo.NumberOfPlanes <= 1 && modeinfo.BitsPerPixel == 32)
#endif
     ) {
      if (param->memtype == mem_text)
	  param->bitperpixel = 16;    /* scroll, plotHline (16 instead of 4) */
	else if (modeinfo.BitsPerPixel == 15)
	  param->bitperpixel = 16;
	else if (modeinfo.BitsPerPixel == 4)
	  param->bitperpixel = 1;
	else
	  param->bitperpixel = modeinfo.BitsPerPixel;
      if (modeinfo.BitsPerPixel > 4)
	  param->nbplane = 1;
	else
	  param->nbplane = 4;
      }
    else {
      VDBG (("%u bpp & %u planes not supported or not compiled-in!\r\n",
		modeinfo.BitsPerPixel, modeinfo.NumberOfPlanes));
      return 0x105;
      }

  VDBG (("red mask/off: %u/%u, green mask/off: %u/%u, blue mask/off: %u/%u ",
	modeinfo.RedMaskSize,   modeinfo.RedFieldPosition,
	modeinfo.GreenMaskSize, modeinfo.GreenFieldPosition,
	modeinfo.BlueMaskSize,  modeinfo.BlueFieldPosition));

  param->red_mask_size = modeinfo.RedMaskSize;
  param->red_field_pos = modeinfo.RedFieldPosition;
  param->green_mask_size = modeinfo.GreenMaskSize;
  param->green_field_pos = modeinfo.GreenFieldPosition;
  param->blue_mask_size = modeinfo.BlueMaskSize;
  param->blue_field_pos = modeinfo.BlueFieldPosition;

  if (   modeinfo.RedMaskSize != 0
      && modeinfo.GreenMaskSize != 0
      && modeinfo.BlueMaskSize != 0) {
      param->nbcolor = 1 << (modeinfo.RedMaskSize + modeinfo.GreenMaskSize
					+ modeinfo.BlueMaskSize);
      }
    else
      param->nbcolor = 1 << modeinfo.BitsPerPixel;

  param->attr.isgraphic = modeinfo.ModeAttributes.graphic_mode;
  /*
   * Is it graphic and can it draw bgcolor? Lets say yes
   * if it is graphic and has more than 16 colors - I do not
   * have less than 16 colors graphic mode anyway...
   */
  param->attr.graphicbgcolor = param->attr.isgraphic;
  param->width = modeinfo.XResolution;	/* set it also for text/scroll */
  param->heigh = modeinfo.YResolution;	/* set it also for text/scroll */
  if (modeinfo.ModeAttributes.mode_not_VGA_compatible)
      param->EGA_compatible = 0;
    else
      param->EGA_compatible = 1;
  if (modeinfo.ModeAttributes.BIOS_output_supported)
      param->BIOS_compatible = 1;
    else
      param->BIOS_compatible = 0;

#if USER_SUPPORT & VESA_4BPP_TEXT
  if (param->memtype == mem_text) {
      param->font = 0;
      param->nbcol = param->width;
      param->nbrow = param->heigh;
      param->charwidth = modeinfo.XCharSize;
//      param->width *= param->charwidth;
      param->charheigh = modeinfo.YCharSize;
//      param->heigh *= modeinfo.YCharSize;
      VDBG (("using hardware font.\r\n"));
      }
    else
#endif
      {
      unsigned short byteperchar;
      unsigned char nbrow;

      VDBG (("get VGA font "));
      if (param->heigh >= 350) {
	  param->font = _VGA_getfont (6, &nbrow, &byteperchar); /* 8x16 */
	  param->charheigh = 16;
	  if (param->heigh == 350 && !get_gujin_param_attrib().refuse_8x14font) {
	      unsigned short byteperchar14;
	      unsigned char nbrow14;
	      farptr font = _VGA_getfont (2, &nbrow14, &byteperchar14); /* 8x14 */

	      VDBG (("(font 8x16 at 0x%X, font 8x14 at 0x%X nbrow %u) ",
			param->font, font, nbrow14));
	      if (   font >= 0xC0000000 /* PC real mode address */
		  && font != param->font
		  && nbrow14 == (350 / 14) - 1) {
		  /* This font does seem to exist: */
		  param->font = font;
		  param->charheigh = 14;
		  }
	      }
	  }
	else {
	  param->font = _VGA_getfont (3, &nbrow, &byteperchar); /* 8x8 */
	  param->charheigh = 8;
	  }
      param->charwidth = 8;
      param->nbcol = param->width / param->charwidth;
      param->nbrow = param->heigh / param->charheigh;
      VDBG (("farptr 0x%X, nbrow = %u.\r\n", param->font, nbrow));
      }

  {
  struct videomode_str videomode;

  videomode.number = param->identification;
  videomode.width = param->width;
  videomode.heigh = param->heigh;
  videomode.bpp = get_bpp (param);
  videomode.text = !param->attr.isgraphic;
  videomode.attr = param->attr.isgraphic;
  videomode.vesa = 1;
  videomode.font8x14 = (param->charheigh == 14);
  add_valid_mode (&videomode);
  }

  return 0;
  }}

static unsigned VESA_setmode (unsigned mode);

static const struct user_interface_fct_str VESA_UI_FCT = {
    getsize:		VESA_getsize,
    setmode:		VESA_setmode,
    getmode:		VESA_getmode,
    clearscreen:	VESA_clearscreen,

    setcursor:		VESA_setcursor,
    getcursor:		COMMON_getcursor,
    setfgcolor:		COMMON_setfgcolor,
    setbgcolor:		COMMON_setbgcolor,
    setattribute:	VESA_setattribute,

    putstr:		VESA_putstr,

    setpixel:		(typeof (UI.function.setpixel)) COMMON_error,
    getpixel:		(typeof (UI.function.getpixel)) COMMON_error,
    plotHline:		(typeof (UI.function.plotHline)) COMMON_error,

    setDACblock:	VESA_setDACblock,
    };

#if USER_SUPPORT & VESA_HARDWINDOW
/*
 * Before continuing reading that, I hope you have already read the
 * General Public License (GPL) version 2 under which this software
 * is released.
 * You can only use an extract of this code OR an idea used in this
 * code OR an algorithm present here if your software is released
 * under the GPL license. In fact, your software will be immediately
 * placed under GPL (free and unlimited source code available)
 * if you use, even by linking only (statically or dynamically),
 * anything written here - or in any other file of Gujin.
 */
static struct hard_recorder_str {
    unsigned short        nbwrite;
    unsigned short	  maxwrite;
    struct hardwindow_str *array;
    } __attribute__ ((packed)) hard_recorder;

static void __attribute__ ((cdecl, regparm (0), unused))
hard_recorder_register (const unsigned edx, const unsigned eax, unsigned bx)
	/* "const" is written in purpose ! */
  {
  if (hard_recorder.nbwrite < hard_recorder.maxwrite) {
      hard_recorder.array[hard_recorder.nbwrite].instruction = bx;
      hard_recorder.array[hard_recorder.nbwrite].port = edx;
      hard_recorder.array[hard_recorder.nbwrite].data = eax;
      }
  hard_recorder.nbwrite++;
  }

asm (
".text									\n"
"	.align 0x20							\n"
"hard_recorder_handler:							\n"
"	pushl	%ds							\n"
"	pushl	%ebx							\n"
"	movw	%sp,%bx							\n"
"	movw	%ss:10(%bx),%bx						\n"
"	movw	%bx,%ds							\n"
"	movw	%sp,%bx							\n"
"	movw	%ss:8(%bx),%bx						\n"
"	movw	(%bx),%bx	# %bx now contains the opcode		\n"
"	cmpb	$0xEE,%bl						\n"
"	je	hard_recorder_match					\n"
"	cmpb	$0xEF,%bl						\n"
"	je	hard_recorder_match					\n"
"	cmpw	$0xEF66,%bx						\n"
"	je	hard_recorder_match					\n"
"hard_recorder_end:							\n"
"	popl	%ebx							\n"
"	popl	%ds							\n"
"	iretw								\n"
"hard_recorder_match:							\n"
"	pushl	%ecx							\n"
"	pushl	%ebx							\n"
"	pushl	%eax							\n"
"	pushl	%edx							\n"
"	movw	%cs,%bx							\n"
"	addw	$deltaseg,%bx						\n"
"	movw	%bx,%ds							\n"
"	calll	hard_recorder_register					\n"
"	popl	%edx							\n"
"	popl	%eax							\n"
"	popl	%ebx							\n"
"	popl	%ecx							\n"
"	jmp	hard_recorder_end					\n"
	);

#ifdef COMPRESS_HARDWIN
static unsigned
hardwindow_compress (struct hardwindow_str *long_array,
		     struct hardwindow_str *end,
		     unsigned firstwin_last)
  {bound_stack();{
  unsigned short page_counter;
  struct hardwindow_str compressed_array[firstwin_last + 1],
			*array, *compress;
  unsigned char tmp;

  VDBG ((" [hardwindow_compress "));
  memcpy (compressed_array, long_array, sizeof (compressed_array));

  compress = compressed_array;
  array = &long_array[firstwin_last + 1];
  page_counter = 1;
  while (array <= end) {
      if (compress->instruction == 0xEF66) {
	  if (array->instruction != 0xEF66) /* outl in ".code16" */
	      VDBG ((" failed, different instructions on outl] "));
	  }
	else {
	  if ((compress->instruction & 0xFF) != array->instruction)
	      VDBG ((" failed, different instructions on outw or outb] "));
	  }

      if (array->port != compress->port) {
	  VDBG ((" failed, different ports] "));
	  break;
	  }

      if (array->instruction == 0xEF66) { /* outl */
	  if (array->data != compress->data) {
	      VDBG ((" failed, outl with different data] "));
	      break;
	      }
	  VDBG (("l"));
	  }
	else if (compress->instruction == 0) {
	  if (array->instruction != 0) {
	      VDBG ((" failed, different number of instructions] "));
	      break;
	      }
	  }
	else if ((tmp = (compress->instruction >> 8)) == 0) { /* simple */
	  if (array->data == compress->data) {
	      if (compress->instruction == 0xEE)
		  VDBG (("b"));
		else if (compress->instruction == 0xEF)
		  VDBG (("w"));
		else {
		  VDBG (("?"));
		  break;
		  }
	      }
	    else {
	      unsigned char xor_data;

	      /* Try to find a pattern */
	      if (compress->instruction == 0xEE)
		  xor_data = array->data ^ compress->data;
		else
		  xor_data = (array->data ^ compress->data) >> 8;

	      if (xor_data == page_counter) {
		  compress->instruction |= 19 << 8;
		  VDBG (("C"));
		  }
#if 0 /* start as the previous compression */
		else if (xor_data == (page_counter & 0x0F)) {
		  compress->instruction |= 1 << 8;
		  VDBG (("D"));
		  }
#endif
		else if (xor_data == ((page_counter & 0x0F) << 4)) {
		  compress->instruction |= 3 << 8;
		  VDBG (("E"));
		  }
		else if (xor_data == (page_counter >> 4)) {
		  compress->instruction |= 20 << 8;
		  VDBG (("F"));
		  }
		else if (xor_data == (page_counter << 2)) {
		  compress->instruction |= 28 << 8;
		  VDBG (("G"));
		  }
		else if (xor_data == ((page_counter & 0xF0) >> 2)) {
		  compress->instruction |= 67 << 8;
		  VDBG (("H"));
		  }
		else {
		  VDBG ((" failed, cannot compress] "));
		  break;
		  }
	      }
	  }
	else if (tmp == 1 || tmp == 3 || tmp == 19 || tmp == 67
			  || tmp == 20 || tmp == 28) {
	  unsigned short page = page_counter;

	  if (tmp == 1 || tmp == 3)
	      page &= 0x0F;

	  if (tmp == 3)
	      page <<= 4;
	    else if (tmp == 28)
	      page <<= 2;
	    else if (tmp == 67)
	      page = (page & 0xF0) >> 2;
	    else if (tmp == 20)
	      page >>= 4;

	  if (array->instruction == 0xEE) {
	      if (array->data != (compress->data ^ page)) {
		  if (   tmp == 19
		      && (array->data == (compress->data ^ (page & 0x0F)))) {
		      compress->instruction &= 0x1FF;
		      VDBG (("D"));
		      }
		    else {
		      VDBG ((" failed, outb compression1 differ] "));
		      break;
		      }
		  }
		else
		  VDBG (("*"));
	      }
	    else if (array->instruction == 0xEF) {
	      if (array->data != (compress->data ^ (page << 8))) {
		  if (   tmp == 19
		      && (array->data == (compress->data ^ ((page & 0x0F) << 8)))) {
		      compress->instruction &= 0x1FF;
		      VDBG (("D"));
		      }
		    else {
		      VDBG ((" failed, outw compression1 differ] "));
		      break;
		      }
		  }
		else
		  VDBG (("*"));
	      }
	    else {
	      VDBG ((" failed, unknown compressed1 instruction!] "));
	      break;
	      }
	  }
	else {
	  VDBG ((" failed, unknown compression method!] "));
	  }

      if (compress->instruction == 0) {
	  page_counter ++;
	  compress = compressed_array;
	  VDBG ((","));
	  }
	else
	  compress++;
      array ++;
      }
  if (array <= end)
      return 0; /* compression impossible with this intelligence level
			(sorry, it's mine !) */

  VDBG ((" OK] "));
  memcpy (long_array, compressed_array, sizeof (compressed_array));
  return firstwin_last;
  }}
#else /* COMPRESS_HARDWIN */
#define hardwindow_compress(a, b, c) (0) /* failed */
#endif /* COMPRESS_HARDWIN */

static unsigned hardwindow_init (struct hardwindow_str **hardwindow,
				 unsigned short **offsetarray,
				 unsigned char	winNo)
  {bound_stack();{
  extern unsigned char hard_recorder_handler[];
  unsigned tmp, saved_interrupt_vector;
  unsigned short *curr_offset_ptr, nbwindow, max_array, i, j;
  /* This array has to be long, nbwindow is a byte: max 256, mostly
     when the VESA1 window granularity is 4 Kbytes... 12 Kbytes */
  struct hardwindow_str long_array[256 * 6];

  nbwindow = DIVROUNDUP ((unsigned) UI.parameter.linelength * UI.parameter.heigh,
			 DISPLAY_WINDOW_GRANUL(&UI.parameter));
  /* offsetarray[0] = 0, offsetarray[nbwin-1] = normal,
     offsetarray[nbwin] = last used long_array index,
     offsetarray[nbwin+1] = end marker = 0 */
  VDBG ((" malloc1 %u bytes for %u windows ", 2 * nbwindow + 4, nbwindow));
  *offsetarray = malloc (2 * nbwindow + 4);
  if (*offsetarray == 0) {
      *hardwindow = 0;
      VDBG (("FAILED!] "));
      return 1;
      }
  curr_offset_ptr = *offsetarray;
  VDBG (("OK, "));

  if (nbwindow != 1) { /* remove possible optimisation */
      _VESA_setwin (1, winNo);
#if !(USER_SUPPORT & VESA_2WINDOWS)
      if (UI.parameter.NbWin == 2)
	  _VESA_setwin (1, winNo ^ 1);
#endif
      }

 if (BOOT1_DOS_running()) {
      saved_interrupt_vector = DOS_GetInterruptVector (1);
      DOS_SetInterruptVector (1, code_adr (hard_recorder_handler));
      }
    else {
      saved_interrupt_vector = peekl (4 * 1);
      pokel (4 * 1, code_adr (hard_recorder_handler));
      }

  hard_recorder.maxwrite = nbof (long_array) / nbwindow;
  *curr_offset_ptr = 0;
  for (i = 0; i < nbwindow; i++) {
      hard_recorder.array = &long_array[*curr_offset_ptr];
      hard_recorder.nbwrite = 0;
      VDBG ((" set win 0x%X ", i));
      setTFflag();
      _VESA_setwin (i, winNo);
#if !(USER_SUPPORT & VESA_2WINDOWS)
      if (UI.parameter.NbWin == 2)
	  _VESA_setwin (i, winNo ^ 1); /* concatenate ! */
#endif
      clearTFflag();
      if (hard_recorder.nbwrite == 0) {
	  VDBG ((" did not work, nothing recorded!] "));
	  free (*offsetarray);
	  *offsetarray = 0;
	  *hardwindow = 0;
	  return 3;
	  }
	else if (hard_recorder.nbwrite >= hard_recorder.maxwrite) {
	  VDBG ((" OVERFLOW %d out* !] ", hard_recorder.nbwrite));
	  free (*offsetarray);
	  *offsetarray = 0;
	  *hardwindow = 0;
	  return 4;
	  }
	else
	  VDBG ((" %d out*, ", hard_recorder.nbwrite));
      hard_recorder.array[hard_recorder.nbwrite] = ((struct hardwindow_str) {});
      *(curr_offset_ptr+1) = *curr_offset_ptr + hard_recorder.nbwrite + 1;
      curr_offset_ptr++; /* GCC bug ? */
      if (*curr_offset_ptr >= nbof (long_array) - 1 - hard_recorder.maxwrite) {
	  VDBG ((" WILL OVERFLOW TEMPORARY ARRAY!" ));
	  free (*offsetarray);
	  *offsetarray = 0;
	  *hardwindow = 0;
	  return 5;
	  }
      }
  _VESA_setwin (0, winNo);
#if !(USER_SUPPORT & VESA_2WINDOWS)
  if (UI.parameter.NbWin == 2)
      _VESA_setwin (0, winNo ^ 1);
#endif

  if (BOOT1_DOS_running())
      DOS_SetInterruptVector (1, saved_interrupt_vector);
    else
      pokel (4 * 1, saved_interrupt_vector);

  max_array = *curr_offset_ptr - 1;
  *++curr_offset_ptr = 0;	/* keep end array, write end marker after */

  /* just cleanning the array before trying to compress: */
  {
  struct hardwindow_str *ptr = long_array;

  for (i = 0; i <= max_array; i++, ptr++) {
      if ((ptr->instruction & 0xFF) == 0xEF) {
	  ptr->instruction = 0xEF; /* outw */
	  ptr->data &= 0xFFFF;
	  }
	else if ((ptr->instruction & 0xFF) == 0xEE) {
	  ptr->instruction = 0xEE; /* outb */
	  ptr->data &= 0xFF;
	  }
      }
  }

  if ((tmp = hardwindow_compress (long_array, &long_array[max_array],
					(*offsetarray)[1] - 1)) != 0) {
      free (*offsetarray);
      *offsetarray = 0;
      max_array = tmp;
      }

  tmp = sizeof (struct hardwindow_str) * (1 + max_array);
  VDBG ((" OK, malloc2 %u bytes ", tmp));
  *hardwindow = malloc (tmp);
  if (*hardwindow == 0) {
       VDBG ((" FAILED!] "));
       if (*offsetarray) {
	   free (*offsetarray);
	   *offsetarray = 0;
	   }
       return 2;
       }

  if (*offsetarray)
      VDBG (("OK,\r\nWin0 [%u]: ", (*offsetarray)[0]));
    else
      VDBG (("OK,\r\nAllWin: "));
  for (i = 0, j = 0; i <= max_array; i++) {
      (*hardwindow)[i] = long_array[i];

      if ((*hardwindow)[i].instruction == 0xEF66)
	  VDBG (("outl (0x%X, 0x%X); ",
		 (*hardwindow)[i].port, (*hardwindow)[i].data));
#ifdef COMPRESS_HARDWIN
	else if ((tmp = ((*hardwindow)[i].instruction >> 8)) != 0) {
	  if (((*hardwindow)[i].instruction & 0xFF) == 0xEE)
	      VDBG (("outb "));
	    else if (((*hardwindow)[i].instruction & 0xFF) == 0xEF)
	      VDBG (("outw "));
	    else
	      VDBG (("ERROR "));
	  VDBG (("(0x%X, ", (*hardwindow)[i].port));
	  if (tmp == 1)
	      VDBG (("0x%X ^ (page & 0x0F)); ", (*hardwindow)[i].data));
	    else if (tmp == 3)
	      VDBG (("0x%X ^ ((page & 0x0F) << 4));", (*hardwindow)[i].data));
	    else if (tmp == 19)
	      VDBG (("0x%X ^ page);", (*hardwindow)[i].data));
	    else if (tmp == 20)
	      VDBG (("0x%X ^ (page >> 4));", (*hardwindow)[i].data));
	    else if (tmp == 28)
	      VDBG (("0x%X ^ (page << 2));", (*hardwindow)[i].data));
	    else if (tmp == 67)
	      VDBG (("0x%X ^ ((page & 0xF0) >> 2));", (*hardwindow)[i].data));
#ifdef COMPRESS_EXTENDED
	    else if (tmp == 37 && (((*hardwindow)[i].instruction & 0xFF) == 0xEE))
	      VDBG (("0x%X ^ ((inb(0x%X) & 0xF0) ^ (page & 0x0F)));", (*hardwindow)[i].data, (*hardwindow)[i].port));
	    else if (tmp == 11 && (((*hardwindow)[i].instruction & 0xFF) == 0xEE))
	      VDBG (("0x%X ^ ((inb(0x%X) & 0x0F) ^ ((page & 0x0F) << 4)));", (*hardwindow)[i].data, (*hardwindow)[i].port));
#endif
	    else
	      VDBG (("ERROR"));
	    }
#endif /* COMPRESS_HARDWIN */
	else if ((*hardwindow)[i].instruction == 0xEE)
	  VDBG (("outb (0x%X, 0x%X); ",
		 (*hardwindow)[i].port, (*hardwindow)[i].data));
	else if ((*hardwindow)[i].instruction == 0xEF)
	  VDBG (("outw (0x%X, 0x%X); ",
		 (*hardwindow)[i].port, (*hardwindow)[i].data));
	else if (   (*hardwindow)[i].instruction == 0
		 && (*hardwindow)[i].data == 0
		 && (*hardwindow)[i].port == 0) {
	  if (i == max_array)
	      VDBG (("\r\nEnd."));
	    else {
	      j++;
	      VDBG (("\r\nWin%u [%u]: ", j, (*offsetarray)[j]));
	      }
	  }
	else
	  VDBG (("ERROR; "));
      }

  VDBG ((" OK"));
  return 0;
  }}
#endif /* VESA_HARDWINDOW */

#if (USER_SUPPORT & VESA_HARDWINDOW) && (USER_SUPPORT & VESA_2WINDOWS)
static inline unsigned
compare_hardwin (struct hardwindow_str *hardwindow1,
		 unsigned short *offsetarray1,
		 struct hardwindow_str *hardwindow2,
		 unsigned short *offsetarray2)
  {
  unsigned short nb, end;

  if ((offsetarray1 == 0) ^ (offsetarray2 == 0)) {
      VDBG ((" [%s compression difference] ", __FUNCTION__));
      return 0xFFFFFFFFU;
      }
  if (offsetarray1) { /* not compressed */
      unsigned short *ptr = offsetarray1 + 1; /* start index of second page */

      while (*ptr) ptr++; /* go to the end */
      end = *--ptr; /* that is the last index of hardwin used */
      VDBG ((" [%s uncompressed total used %u] ", __FUNCTION__, end));
      }
    else { /* compressed */
      struct hardwindow_str *ptr = hardwindow1;

      while (ptr->instruction) ptr++;
      end = ptr - hardwindow1;
      VDBG ((" [%s compressed total used %u] ", __FUNCTION__, end));
      }

    VDBG ((" ["));
    for (nb = 0; nb < end; nb++) {
      VDBG (("%d, ", nb));
      if (   hardwindow1->instruction != hardwindow2->instruction
	  || hardwindow1->port !=        hardwindow2->port
	  || hardwindow1->data !=        hardwindow2->data)
	  break;
      hardwindow1++;
      hardwindow2++;
      }
    VDBG (("] "));

  if (nb < end) {
      VDBG ((" [%s dependancy detected at %u] ", __FUNCTION__, nb));
      return nb + 1; /* can have a dependancy at offset 0 */
      }
    else {
      VDBG ((" [%s no dependancy] ", __FUNCTION__));
      return 0;
      }
  }

static inline void
initialise_second_hardwindow (unsigned short nbwindow)
  {
  /* _VESA_setwin (0, UI.parameter.winNo); already done */
  if (hardwindow_init(&UI.parameter.readhardwindow,
		      &UI.parameter.readoffsetarray,
		      UI.parameter.winNo ^ 1) != 0)
      VDBG (("FAILED hardwin 2"));
    else if (nbwindow >= 2) {
      struct hardwindow_str *hardwindow2;
      unsigned short *offsetarray2;

      _VESA_setwin (1, UI.parameter.winNo);
      if (hardwindow_init(&hardwindow2, &offsetarray2,
			  UI.parameter.winNo ^ 1) != 0)
	  VDBG (("FAILED 2nd try hardwin 2"));
	else {
	  unsigned short pos;
	  while ((pos = compare_hardwin (UI.parameter.readhardwindow,
					 UI.parameter.readoffsetarray,
					 hardwindow2,
					 offsetarray2)) != 0) {
	      VDBG (("HARDWINDOW of two window inter-dependant "
		     "at %u: inst1=0x%X, data1=0x%X, addr1=0x%X ; "
		     "inst2=0x%X, data2=0x%X, addr2=0x%X! "
		     "readoffsetarray=0x%X, offsetarray=0x%X; "
		     "1st window: instruction=0x%X, port=0x%X ",
		     pos - 1,
		     UI.parameter.readhardwindow[pos-1].instruction,
		     UI.parameter.readhardwindow[pos-1].data,
		     UI.parameter.readhardwindow[pos-1].port,
		     hardwindow2[pos-1].instruction,
		     hardwindow2[pos-1].data,
		     hardwindow2[pos-1].port,
		     UI.parameter.readoffsetarray,
		     UI.parameter.offsetarray,
		     UI.parameter.hardwindow[pos-1].instruction,
		     UI.parameter.hardwindow[pos-1].port));
#ifdef COMPRESS_EXTENDED
	      if (   UI.parameter.readoffsetarray == 0
		  && UI.parameter.offsetarray == 0
		  && UI.parameter.hardwindow[pos-1].port
				== UI.parameter.readhardwindow[pos-1].port) {
		  if (   UI.parameter.hardwindow[pos-1].instruction == 0x3EE
		      && (   UI.parameter.readhardwindow[pos-1].instruction == 0x1EE
			  || UI.parameter.readhardwindow[pos-1].instruction == 0x13EE) /* 0d19 */) {
		      VDBG ((" simple HARDWINDOW inter-dependancy fixed 1"));
		      UI.parameter.readhardwindow[pos-1].instruction = 0x25EE; /* 0d37 */
		      hardwindow2[pos-1].instruction = 0x25EE; /* 0d37 */
		      UI.parameter.hardwindow[pos-1].instruction |= 0x800; /* 0d11 */
		      UI.parameter.readhardwindow[pos-1].data = 0;
		      hardwindow2[pos-1].data = 0;
		      UI.parameter.hardwindow[pos-1].data = 0;
		      continue;
		      }
		  if (    UI.parameter.readhardwindow[pos-1].instruction == 0x3EE
		      && (   UI.parameter.hardwindow[pos-1].instruction == 0x1EE
			  || UI.parameter.hardwindow[pos-1].instruction == 0x13EE) /* 0d19 */) {
		      VDBG ((" simple HARDWINDOW inter-dependancy fixed 2"));
		      UI.parameter.readhardwindow[pos-1].instruction |= 0x800; /* 0d11 */
		      hardwindow2[pos-1].instruction |= 0x800; /* 0d11 */
		      UI.parameter.hardwindow[pos-1].instruction = 0x25EE; /* 0d37 */
		      UI.parameter.readhardwindow[pos-1].data = 0;
		      hardwindow2[pos-1].data = 0;
		      UI.parameter.hardwindow[pos-1].data = 0;
		      continue;
		      }
		  }
#endif
	      free_hardwin (); /* disable read and write hardwindow */
	      break;
	      }
	  if (offsetarray2)
	      free (offsetarray2);
	  free (hardwindow2);
	  }
      _VESA_setwin (0, UI.parameter.winNo);
      }
  }
#endif /* VESA_HARDWINDOW */

static unsigned
VESA_setmode (unsigned mode)
  {bound_stack();{
  struct video_parameter_str param;
  unsigned tmp, clearscreen;

  if (mode & 0x8000)
      clearscreen = 0;
    else
      clearscreen = 1;
  mode &= ~0x8000;

  if ((tmp = VESA_getsize (mode, &param)) != 0)
      return tmp;

  mode &= ~0x4000;

#if USER_SUPPORT & VESA_LINEAR
  if (param.winsize == VESA2_MARKER) {
#if ASSEMBLY_TYPE != ASSEMBLY_DSES
      if (param.winadr != 0 /* false VESA2 */ && segment_door (1) != 0) {
	  VDBG (("GS door does not want to open!\r\n"));
#if !(USER_SUPPORT & VESA_WINDOW)
	  remove_valid_mode (mode);
#endif
	  return 3;
	  }
#endif /* ASSEMBLY_TYPE != ASSEMBLY_DSES */

      if (param.winadr != 0 /* false VESA2 */) {
	  if (_VESA_setmode (mode | 0x4000) != 0) {
#if !(USER_SUPPORT & VESA_WINDOW)
	      remove_valid_mode (mode);
#endif
	      return 4;
	      }
	  }
	else {
	  if (_VESA_setmode (mode & ~0x4000) != 0) { /* VESA2 < 64K mode */
#if !(USER_SUPPORT & VESA_WINDOW)
	      remove_valid_mode (mode);
#endif
	      return 5;
	      }
	  _VESA_setwin (0, 1);
	  _VESA_setwin (0, 0);
	  }
#if USER_SUPPORT & VESA_HARDWINDOW
      free_hardwin();
#endif
      UI.parameter = param;
      }
    else
#endif /* VESA_LINEAR */
      {
#if USER_SUPPORT & VESA_WINDOW
      unsigned short nbwindow;
#if USER_SUPPORT & VESA_WINFUNCTION
      unsigned saved = UI.parameter.addrpositionfarfct;
#endif
      if (_VESA_setmode (mode & ~0x4000) != 0) {
	  VDBG (("VESA1: cannot switch to mode 0x%X!\r\n", mode & ~0x4000));
#if USER_SUPPORT & VESA_LINEAR
	  if ((UI.info.vesa.version >> 8) < 2) /* else can try VESA2 also */
#endif
	      remove_valid_mode (mode);
	  return 6;
	  }
#if USER_SUPPORT & VESA_WINFUNCTION
      UI.parameter.addrpositionfarfct = 0;
#endif
      if (param.NbWin >= 1 && param.memtype != mem_text)
	  if (_VESA_setwin (0, param.winNo) != 0) {
	      VDBG (("VESA1: cannot setup window for "
			"write/readwrite win!\r\n"));
#if USER_SUPPORT & VESA_WINFUNCTION
	      UI.parameter.addrpositionfarfct = saved;
#endif
	      _VESA_setmode (UI.parameter.identification);
#if USER_SUPPORT & VESA_LINEAR
	      if ((UI.info.vesa.version >> 8) < 2) /* else can try VESA2 also */
#endif
		  remove_valid_mode (mode);
	      return 7;
	      }
      if (param.NbWin == 2 && param.memtype != mem_text)
	  if (_VESA_setwin (0, param.winNo ^ 1) != 0) {
	      VDBG (("VESA1: cannot setup window for read win!\r\n"));
#if USER_SUPPORT & VESA_WINFUNCTION
	      UI.parameter.addrpositionfarfct = saved;
#endif
	      _VESA_setmode (UI.parameter.identification);
#if USER_SUPPORT & VESA_LINEAR
	      if ((UI.info.vesa.version >> 8) < 2) /* else can try VESA2 also */
#endif
		  remove_valid_mode (mode);
	      return 8;
	      }
#if USER_SUPPORT & VESA_HARDWINDOW
      free_hardwin();
#endif

      UI.parameter = param;
      UI.parameter.winadr = 0;

      nbwindow = DIVROUNDUP ((unsigned) UI.parameter.linelength	* UI.parameter.heigh,
				DISPLAY_WINDOW_GRANUL(&UI.parameter));

      VDBG ((" [access: "));
      if (nbwindow == 0)
	  VDBG (("nbwindow null] "));
#if USER_SUPPORT & VESA_WINFUNCTION
	else if (UI.parameter.addrpositionfarfct == 0)
	  VDBG (("function pointer null] "));
#endif
	else {
#if USER_SUPPORT & VESA_HARDWINDOW
	  if (get_gujin_param_attrib().enable_VESA_hardwin) {
	      if (hardwindow_init(&UI.parameter.hardwindow,
				  &UI.parameter.offsetarray,
				  UI.parameter.winNo) != 0)
		  VDBG (("FAILED hardwin 1"));
#if USER_SUPPORT & VESA_2WINDOWS
		else if (UI.parameter.NbWin == 2)
		  initialise_second_hardwindow (nbwindow);
#endif /* VESA_2WINDOWS */
	      VDBG (("] "));
	      }
	    else
	      VDBG (("enable_VESA_hardwin gujin param attribute not set, "));
#else
	  VDBG (("function pointer not null] "));
#endif /* VESA_HARDWINDOW */
	  }

#else /* VESA_WINDOW */
      return 9;
#endif /* VESA_WINDOW */
      }

#if ASSEMBLY_TYPE == ASSEMBLY_DSES
  UI.parameter.base_address -= getds() << 4;
#else
  setgs (UI.parameter.base_address >> 4);
#endif

#if USER_SUPPORT & VGA_FREQUENCY
  VDBG ((" [get refresh frequency: "));
  {
  unsigned char nbcol, page; /* Yahoo: get VGA equivalent mode number! */
  UI.frequency = _CT_get_refresh_rate (_VGA_getmode (&nbcol, &page));
  }
  if (UI.frequency == 0)
      UI.frequency = _ET4000_get_refresh_rate (UI.parameter.heigh);
  VDBG (("%u Hz] ", UI.frequency));
#endif

  UI.function = VESA_UI_FCT;

#if USER_SUPPORT & VESA_4BPP_EGA
  if (UI.parameter.memtype == mem_EGA) {
      const struct EGA_mode_str EGA_mode = {
	  write_source : cpuASsetreset,	/* write mode 2 as default */
	  enable_tristate : 0,
	  read_mode : EGA_DEFAULT_READ_MODE,
	  enable_odd_even_addr : 0,
	  CGA_shift_register : 0,
	  enable_256_color : 0,
	  reserved : 0
	  };

      VDBG ((" [%s: EGA_s/getpixel] ", __FUNCTION__));
      UI.function.setpixel = EGA_setpixel;
      UI.function.getpixel = EGA_getpixel;
      UI.function.plotHline = EGA_plotHline;
      _EGA_read_write_mode (EGA_mode);
      _EGA_write_setup (EGA_overwride, 0);
      _EGA_set_reset_enable (0);	/* write (mode 0) comes from CPU */
      EGA_READ_INIT();
      }
    else
#endif
      {
      if (UI.parameter.bitperpixel == 32) {
#if USER_SUPPORT & VESA_32BPP
	  UI.function.setpixel = VESA_setpixel_color32bit;
	  UI.function.getpixel = VESA_getpixel_color32bit;
	  UI.function.plotHline = VESA_plotHline;
#endif
	  }
	else if (UI.parameter.bitperpixel == 24) {
#if USER_SUPPORT & VESA_24BPP
	  UI.function.setpixel = VESA_setpixel_color24bit;
	  UI.function.getpixel = VESA_getpixel_color24bit;
	  UI.function.plotHline = VESA_24BPP_plotHline;
#endif
	  }
	else if (UI.parameter.bitperpixel == 16) {
#if USER_SUPPORT & VESA_16BPP
	  UI.function.setpixel = VESA_setpixel_color16bit;
	  UI.function.getpixel = VESA_getpixel_color16bit;
	  UI.function.plotHline = VESA_plotHline;
#endif
	  }
	else {
#if USER_SUPPORT & VESA_8BPP
	  UI.function.setpixel = VESA_setpixel_color8bit;
	  UI.function.getpixel = VESA_getpixel_color8bit;
	  UI.function.plotHline = VESA_plotHline;
#endif
	  }
      }

  VGA_VESA_setup_color ();

  UI_setbgcolor (UI.stdcolor[lightgray]);
  UI_setfgcolor (UI.stdcolor[blue]);

#if USER_SUPPORT & VESA_4BPP_TEXT
  if (UI.parameter.memtype == mem_text) {
      UI.attributes.valid = no_attribute;
      if (UI.parameter.EGA_compatible)
	  _EGA_blinking (0); /* No way to do so by VESA interface */
      }
    else
#endif
      UI.attributes.valid = graphic_attribute;

  UI.attributes.current = no_attribute;

#if USER_SUPPORT & VESA_RECALC
  /*
   * EGA VERTICAL_DISPLAY_END field is only 10 bits, so max 1024 lines.
   * This will not work for "monochrome" VESA modes, i.e. I/O port
   * at 0x3B4/0x3B5 - I do not have such a hardware.
   * We do not modify the "overflow" register (2 most significant bits of
   * reg 0x07), because then we have to unprotect write access (reg 0x11),
   * and this register is maybe not readable.
   * We just modify the register if the buggy BIOS has programmed an
   * erroneous value not too far away (hope this reg is readable on
   * buggy BIOSes, else no correction).
   */
  if (   UI.parameter.EGA_compatible
      && UI.parameter.nbrow * UI.parameter.charheigh <= 1024) {
      unsigned char tmp = _EGA_CRTR_in(0x12);
      unsigned short maxline;
      short error;

      if (UI.parameter.font != 0)
	  maxline = UI.parameter.heigh - 1; // to be used all the time
	else
	  maxline = UI.parameter.charheigh * UI.parameter.nbrow - 1;
      error = (maxline & 0xFF) - tmp;

  if (error != 0 && ABS(error) <= 4) {
      VDBG ((" [VESA_RECALC: old reg 0x12: 0x%X, maxline = %u, error = %d] ",
		tmp, maxline, error));
      _EGA_CRTR_out (0x12, maxline & 0xFF);
      }
  }
#endif

  if (clearscreen)
      UI_clearscreen();

  VDBG ((", success VESA_setmode\r\n"));
  return 0;
  }}

unsigned VESA_init (void)
  {bound_stack();{
  VESA_VbeInfoBlock vesainfo;
  unsigned short mode, i, first_init;
  const char signature[4] = { 'V', 'E', 'S', 'A' };

#if USER_SUPPORT & VESA_LINEAR
  const char signature_vesa2[4] = { 'V', 'B', 'E', '2' };

  VDBG (("VESA2 call "));
  *(unsigned *)vesainfo.VbeSignature = *(unsigned *)signature_vesa2;

  if (   _VESA_getinfo (&vesainfo) != 0
      || *(unsigned *)(vesainfo.VbeSignature) != *(unsigned *)signature) {
      VDBG (("failed, trying VESA1, "));
#else
  VDBG (("VESA1 call "));
      {
#endif
      *(unsigned *)(vesainfo.VbeSignature) = *(unsigned *)signature;
      if (   _VESA_getinfo (&vesainfo) != 0
	  || *(unsigned *)(vesainfo.VbeSignature) != *(unsigned *)signature) {
	  VDBG (("failed\r\n"));
	  UI.info.vesa.version = 0;
	  memset (UI.info.vesa.cardname, 0, sizeof (UI.info.vesa.cardname));
	  return 1;
	  }
      }

  VDBG (("successfull"
	", version: %u.%u, capabilities: %u, videoram: %u x 64 Kbytes, ",
	vesainfo.VbeVersion >> 8, vesainfo.VbeVersion & 0xFF,
	*(unsigned *)&vesainfo.Capabilities,
	vesainfo.TotalMemory));

  first_init = (UI.info.vesa.version == 0)? 1 : 0;
  UI.info.vesa.version = vesainfo.VbeVersion;
  UI.info.vesa.memory_64Kb = vesainfo.TotalMemory;

  for (i = 0; i < sizeof (UI.info.vesa.cardname) - 1; i++) {
      UI.info.vesa.cardname[i] = peekb (vesainfo.OemStringPtr + i);
      if (UI.info.vesa.cardname[i] == 0)
	  break;
      }
  UI.info.vesa.cardname[i] = '\0';

  if (!strncmp (UI.info.vesa.cardname,
		"\007Copyright 1988-1991 TRIDENT MICROSYSTEM", 39)) {
      /* remove 16 bits modes - they are far to work */
      unsigned short trident_mode[0x11], j = 0;
      strcpy (UI.info.vesa.cardname, "TRIDENT MICROSYSTEM corrected");
      for (i = 0;
	   (mode = peekw(vesainfo.VideoModePtr + i*sizeof(short))) != 0xFFFF;
	   i++)
	  if (mode < 0x110)
	      trident_mode[j++] = mode;
      trident_mode[j] = 0xFFFF;
      vesainfo.VideoModePtr = data_adr (trident_mode);
      /*
       * There is a palette problem here - seems that the 256 color palette
       * is always active, and trying to program the 32000 color with the
       * VGA or EGA primitives do not work: those modes are not usefull.
       */
      VDBG ((" [Trident 8900 modes with more than 8 bpp removed, don't work] "));
#if USER_SUPPORT & VGA_SUPPORT
      /*
       * This mode may work, but in fact, after 20 minutes - the "clear screen"
       * operation was still not finished... even on my P120 it is quite slow.
       */
      VIDEO_mode_invalidate (0x60);
      VDBG ((" [Trident 8900 disable VGA mode 0x60] "));
#endif
      }

  VDBG (("name:[%s], modes:", UI.info.vesa.cardname));
  for (i = 0;
       (mode = peekw(vesainfo.VideoModePtr + i*sizeof(short))) != 0xFFFF;
       i++) {
      if (first_init) {
	  struct video_parameter_str param;

	  VDBG (("\r\n0x%X ", mode));
	  /*
	   * This will call add_valid_mode():
	   * Take care, some boards have two times the same VESA number
	   * in the list "vesainfo.VideoModePtr" !
	   */
	  if (VESA_getsize (mode, &param) == 0)
	      VDBG (("OK, "));
	    else
	      VDBG (("ERROR, "));
	  }
	else {
	  /*
	   * Some VESA video mode may have been disabled after a failure
	   * of VESA_setmode(), do not revalidate them here...
	   */
	  VDBG (("0x%X, ", mode));
	  }
      }
  VDBG (("\r\n"));

#if USER_SUPPORT & VESA_PMINFO
  {
  farptr PMdata_addr, base;
  unsigned short j, PMdata_len, tmp;

  tmp = _VESA_GetPMinterface (&PMdata_addr, &PMdata_len);
  VDBG (("[get PM data: 0x%X, PMdata_addr = 0x%X, PMdata_len = 0x%X: ",
		tmp, PMdata_addr, PMdata_len));
  if (   tmp
      || (unsigned long)PMdata_addr <= 0xA0000000U
      || (unsigned long)PMdata_addr > 0xFFFFFFFFU - 10
      || PMdata_len <= 10) {
      VDBG (("failed] "));
      UI.ProtectedModeIO[0] = 0xFFFF;
      UI.ProtectedModeMem[0].address = 0xFFFFFFFF;
      UI.ProtectedModeMem[0].length = 0;
      }
    else {
      base = peekw (PMdata_addr + 6);
      VDBG (("Offsets = %u, %u, %u, I/O at %u; ",
		peekw (PMdata_addr),
		peekw (PMdata_addr+2),
		peekw (PMdata_addr+4),
		base ));
      VDBG (("I/O: "));
      if (base < 8) {
	  VDBG (("abscent] "));
	  UI.ProtectedModeIO[0] = 0xFFFF;
	  UI.ProtectedModeMem[0].address = 0xFFFFFFFFU;
	  UI.ProtectedModeMem[0].length = 0;
	  }
	else {
	  base += PMdata_addr;
	  j = 0;
	  do {
	      tmp = peekw (base + 2*j);
	      VDBG (("%x, ", tmp));
#if 0 /* It is too late to enable it here, do it in /etc/dosemu.conf */
 /* In /etc/dosemu.conf, add for Voodoo3 (VESA page switching at 0x142C):
  $_ports = "0x3C8 0x3C9 0x3D4 0x3D5 0x3DA range 0x142C,0x142F" */
	      {
	      unsigned short version, patchlevel;

	      if (_DOSEMU_check (&version, &patchlevel) == 0) {
		  VDBG (("Info: enabling DOSEMU I/O access for 0x%X\r\n",
				tmp));
		  _DOSEMU_enable_ioports (tmp, 1);
		  }
	      }
#endif
	      if (j >= lastof (UI.ProtectedModeIO)) {
		  if (j == lastof (UI.ProtectedModeIO)) {
		      VDBG ((" [UI.ProtectedModeIO array too short] "));
		      UI.ProtectedModeIO[j] = 0xFFFF;
		      }
		  }
		else
		  UI.ProtectedModeIO[j] = tmp;
	      j++;
	      } while (tmp != 0xFFFF && 2*j <= PMdata_len);
	  if (tmp != 0xFFFF && j <= lastof (UI.ProtectedModeIO)) {
	      UI.ProtectedModeIO[j] = 0xFFFF;
	      }
	  VDBG (("MEM: "));
	  if (2*j <= PMdata_len) {
	      base += 2*j;
	      PMdata_len -= 2*j;
	      j = 0;
	      do {
		  UI.ProtectedModeMem[j].address = peekl (base + 6*j);
		  UI.ProtectedModeMem[j].length = peekw (base + 6*j+4);
		  VDBG (("%x+%x, ",
			UI.ProtectedModeMem[j].address,
			UI.ProtectedModeMem[j].length));
		  j++;
		  } while (   (UI.ProtectedModeMem[j-1].address & 0xFFFF) != 0xFFFF
			   && j < nbof (UI.ProtectedModeMem)
			   && 6*j <= PMdata_len);
	      if (j == nbof (UI.ProtectedModeMem))
		  VDBG ((" [UI.ProtectedModeMem array too short] "));
	      UI.ProtectedModeMem[j-1].address = 0xFFFFFFFFU;
	      }
	    else {
	      VDBG (("over PMdata_len limit "));
	      UI.ProtectedModeMem[0].address = 0xFFFFFFFFU;
	      }
	  VDBG (("] \r\n"));
	  }
      }
  }
#endif /* VESA_PMINFO */

  if (_VESA_getmode (&mode) != 0)
      VDBG (("Initial VESA_getmode failed.\r\n"));
#if 0
    else
      if (VESA_setmode (mode | 0x8000) != 0)	/* Voodoo: clear screen */
	  VDBG (("Initial VESA_setmode failed.\r\n"));
#endif

  /* If VESA_setmode() failed, it does not matter, stay with previous UI */
  UI.function.getsize = VESA_getsize;
  UI.function.setmode = VESA_setmode;

  return 0;
  }}
#endif /* VESA_SUPPORT */

/*
 * The default "serial BIOS library" functions, only for
 * DEC VT100+ compatible terminal.
 */
static unsigned SERIAL_getc (char *C)
  {bound_stack();{
  _BIOS_serialstatus ret;
  unsigned int cpt = 3; /* cpt * BIOS serial timeout */
  unsigned port = BOOT1_COM_port();

  checkptr (C, "SERIAL_getc invalid parameter.");

  do {
      ret = _BIOS_getserial (port);
      } while ((ret.timeout) && (--cpt != 0));

  *C = ret.al.charread;
  if (ret.timeout)
      return 1;
    else
      return 0;
  }}

static inline unsigned CSI_tilda (unsigned short nb)
  {
  if (nb >= 11 && nb <= 15)
      return FCT_KEYCODE(nb-11+1);	/* F1..F5 */
    else if (nb >= 17 && nb <= 21)
      return FCT_KEYCODE(nb-17+6);	/* F6..F10 */
    else if (nb >= 23 && nb <= 26)
      return FCT_KEYCODE(nb-23+11);	/* F11..F14 */
    else if (nb == 28)
      return FCT_KEYCODE(1);		/* Help == F1 */
    else if (nb == 29)
      return FCT_KEYCODE(5);		/* Do == F5 (PF1..PF4) */
    else if (nb >= 31 && nb <= 34)
      return FCT_KEYCODE(nb-31+17);	/* F15..F18 */
    else if (nb == 6)
      return PAGEDOWN_KEYCODE();
    else if (nb == 5)
      return PAGEUP_KEYCODE();
    else if (nb == 1)
      return HOME_KEYCODE(); /* Find, on a PC */
    else if (nb == 4)
      return END_KEYCODE(); /* Select, on a PC */
    else
      return '~';
  }

static unsigned SERIAL_getkey (unsigned timeout)
  {bound_stack();{
  unsigned returned;

  if (timeout)
      timeout += _BIOS_nbtick();

  if (UI.mouse.start)
      UI.mouse.start();
    for (;;) {
      if (_BIOS_getserialstatus(BOOT1_COM_port()).receive_data_ready) {
	  unsigned char tmp;

	  /* Ignore ^S, ^Q, i.e. XOFF XON, sorry: */
	  while (   SERIAL_getc(&tmp) != 0
		 || tmp == '\023' || tmp == '\021') {
	      _BIOS_wait (SERIAL_DELAY);
	      continue;
	      }
	  if (   tmp == (unsigned char)'\233' /* CSI */
	      || tmp == (unsigned char)'\217' /* SS3 */
	      || (   tmp == '\033' && SERIAL_getc(&tmp) == 0
		  && (tmp == '[' || tmp == 'O'))) {
	      if (tmp == (unsigned char)'\233' || tmp == '[') {
		  unsigned short nb = 0;

		  /* TREAT CSI or ESC [ */
		  while (SERIAL_getc(&tmp) == 0 && tmp >= '0' && tmp <= '9')
		      nb = (nb * 10) + tmp - '0';
		  if (tmp == '~')
		      returned = CSI_tilda(nb);
		    else if (tmp == 'A')
		      returned = UP_KEYCODE();
		    else if (tmp == 'B')
		      returned = DOWN_KEYCODE();
		    else if (tmp == 'C')
		      returned = RIGHT_KEYCODE();
		    else if (tmp == 'D')
		      returned = LEFT_KEYCODE();
		    else if (tmp == '[') { /* ESC [ [ A..E  (PC equivalent) */
		      if (SERIAL_getc(&tmp) == 0  && (tmp >= 'A' && tmp <= 'E'))
			  returned = FCT_KEYCODE(tmp - 'A' + 1);
			else
			  returned = tmp;
		      }
		    else
		      returned = tmp;
		  }
		else {
		  /* TREAT SS3 or ESC O */
		  if (SERIAL_getc(&tmp) == 0  && (tmp >= 'P' && tmp <= 'S'))
		      returned = FCT_KEYCODE(tmp - 'P' + 1); /* PF1..PF4 */
		    else
		      returned = tmp;
		  }
	      }
	    else
	      returned = tmp;
	  break;
	  }
      if (timeout && (_BIOS_nbtick() >= timeout)) {
	  returned = TIMEOUT_KEYCODE();
	  break;
	  }
      if (UI.mouse.poll != 0) {
	  returned = UI.mouse.poll();
	  if (returned != 0xFFFFFFFF)
	      break;
	  }
      }
  if (UI.mouse.stop)
      UI.mouse.stop();

//VDBG ((" [%s : 0x%X] ", __FUNCTION__, returned));

  if (returned == 3) { /* ^C */
      SDBG (("\r\nTrying to stop '.com', stop DOSEMU, "
	     "generate DOSEMU exception, halt\r\n"));
      BIOS_killme();
      }

  return returned;
  }}

#if USER_SUPPORT & SERIAL_SUPPORT
static int BOOT1_printf (const char *format, ...)
  {bound_stack();{
  char buffer[256];

#if 0 // DEBUG & DEBUG_SERIAL
  char *tmp;
  unsigned short cpt;
  static unsigned report = 0;

  tmp = prtf (buffer, buffer + sizeof(buffer) - 1, format,
	  (void **)((unsigned)&format + sizeof(const char *)) );
  for (cpt = 0; tmp[cpt] != 0; cpt++) {
      if ((tmp[cpt] & 0x80) | (tmp[cpt] & 0xFF))
	  report = 10;
      if (report) {
	  if ((tmp[cpt] & 0xFF) < 0x20)
	      SDBG ((" {0x%X} ", tmp[cpt] & 0xFF));
	    else
	      SDBG ((" {%c} ", tmp[cpt] & 0xFF));
	  report --;
	  }
      }
  BOOT1_putstr (tmp);
#else
  BOOT1_putstr (prtf (buffer, buffer + sizeof(buffer) - 1, format,
	  (void **)((unsigned)&format + sizeof(const char *)) ));
#endif
  return 0;
  }}

static void SERIAL_setup_color (void)
  {bound_stack();{
  static const unsigned char SERIAL_color[16] = {
      /* Serial terminal with linux colors: */
      0,	/* black */
      4,	/* blue  */
      2,	/* green */
      6,	/* cyan  */
      1,	/* red   */
      5,	/* magenta */
      3,	/* brown */
	/* light colors only for foreground, not background. */
      7 + 8,	/* lightgray */
      7,	/* darkgray */
      4 + 8,	/* lightblue */
      2 + 8,	/* lightgreen */
      6 + 8,	/* lightcyan */
      1 + 8,	/* lightred */
      5 + 8,	/* lightmagenta */
      3 + 8,	/* yellow */
      8		/* white */
      };
  static const unsigned char DEC_VT_color[16] = {
      0,	/* black */
      1,	/* blue  */
      2,	/* green */
      3,	/* cyan  */
      4,	/* red   */
      5,	/* magenta */
      6,	/* brown */
      7,	/* lightgray */
      8,	/* darkgray */
      9,	/* lightblue */
      10,	/* lightgreen */
      11,	/* lightcyan */
      12,	/* lightred */
      13,	/* lightmagenta */
      14,	/* yellow */
      15	/* white */
      };
  unsigned i;
  const char *color;

  if (UI.monitor == 4 || UI.monitor == 3)	/* VT[2345]40 or VT[234]30 */
      color = DEC_VT_color;
    else if (UI.monitor == 0)	/* Linux terminal (color?) */
      color = SERIAL_color;
    else {			/* B&W display, even if it is green! */
      for (i = 0; i < nbof (UI.stdcolor); i++)
	  UI.stdcolor[i] = i & 1;
      return;
      }

  for (i = 0; i < nbof (UI.stdcolor); i++)
      UI.stdcolor[i] = color[i];
  }}

static unsigned SERIAL_waitchar (unsigned char waited)
  {bound_stack();{
  int cpt = 0, status, nb = 0;
  unsigned char C;

  while (1) {
      status = SERIAL_getc(&C);
      if (status == 0 && C == waited)
	  return 0;
      if (status == 0 && C == '\033' && waited >= 0x80 && waited <= 0x9F) {
	  /* special VT 8->7bits conversion */
#define CHAR_CSI	'\233'
#define CHAR_DCS	'\220'
#define CHAR_ST		'\234'
	  waited -= 0x40;
	  continue;
	  }
      SDBG ((" %s, waiting 0x%X", __FUNCTION__, waited));
      if (status != 0 || cpt++ > 10) {
	  SDBG ((", %s received.\r\n", (nb != 0)? "other char" : "nothing"));
	  return 1;
	  }
      if (status == 0) {
	  nb++;
	  SDBG ((", get: 0x%X\r\n", C));
	  }
      }
  }}

static unsigned SERIAL_waitreportNUM (unsigned int *val, unsigned char separator)
  {bound_stack();{
  unsigned int cpt = 0, tmp = 0;
  unsigned char C;

  while (cpt++ < 10) {
      if (SERIAL_getc(&C) != 0) {
	  SDBG ((" %s wait number, nothing received.\r\n", __FUNCTION__));
	  continue;
	  }
      if (C >= '0' && C <= '9')
	  tmp = (tmp * 10) + (C - '0');
	else if (C == separator)
	  break;
	else if (C == '\033' && separator >= 0x80 && separator <= 0x9F) {
	  separator -= 0x40;
	  continue;
	  }
	else
	  SDBG ((" %s, not a digit nor 0x%X but 0x%X\r\n",
		__FUNCTION__, separator, C));
      }
  if (cpt >= 10) {
      SDBG ((" %s aborted.\r\n", __FUNCTION__));
      return 1;
      }
  if (val)
      *val = tmp;
  return 0;
  }}

static unsigned SERIAL_getVTreport (unsigned feature, int *result)
  {bound_stack();{
  unsigned tmp;

  BOOT1_printf ("\033[?%u$p", feature);

  if (SERIAL_waitchar (CHAR_CSI) != 0) {
      SDBG ((" %s aborted on CSI ", __FUNCTION__));
      return 1;
      }
  if (SERIAL_waitchar ('?') != 0) {
      SDBG ((" %s aborted on '?' ", __FUNCTION__));
      return 1;
      }
  if (SERIAL_waitreportNUM (&tmp, ';') != 0 || tmp != feature) {
      SDBG ((" %s bad 'feature' response ", __FUNCTION__));
      return 1;
      }
  if (SERIAL_waitreportNUM (&tmp, '$') != 0) {
      SDBG ((" %s bad 'value' response ", __FUNCTION__));
      return 1;
      }
  if (SERIAL_waitchar ('y') != 0) {
      SDBG ((" %s aborted on trailer ", __FUNCTION__));
      return 1;
      }
  if (result)
      *result = tmp;
  return 0;
  }}

static unsigned SERIAL_getcontrol (const char *feature, int *result)
  {bound_stack();{
  unsigned int tmp;

  BOOT1_printf ("\033P$q%s\033\\", feature); /* DCS $ q <for instance "*|"> ST */

  if (SERIAL_waitchar (CHAR_DCS) != 0) {
      SDBG ((" %s aborted on DCS ", __FUNCTION__));
      return 1;
      }
  if (SERIAL_waitreportNUM (&tmp, '$') != 0) {
      SDBG ((" %s bad format ", __FUNCTION__));
      return 1;
      }
  if (tmp != 0) {
      SDBG ((" %s 'invalid request': 0x%X ", __FUNCTION__, tmp));
//      return 1;	// my VT420 always set this bit, why?
      }
  if (SERIAL_waitchar ('r') != 0) {
      SDBG ((" %s aborted on 'r' ", __FUNCTION__));
      return 1;
      }
  if (SERIAL_waitreportNUM (&tmp, feature[0]) != 0) {
      SDBG ((" %s bad 'value/terminator' response ", __FUNCTION__));
      return 1;
      }
  /* Control Fuction are either one or two char wide */
  if (feature[1] != 0 && SERIAL_waitchar (feature[1]) != 0) {
      SDBG ((" %s bad terminator2 response ", __FUNCTION__));
      return 1;
      }
  if (SERIAL_waitchar (CHAR_ST) != 0) {
      SDBG ((" %s bad terminator ", __FUNCTION__));
      return 1;
      }
  if (result)
      *result = tmp;
  return 0;
  }}

static void
SERIAL_putstr (const char *str)
  {bound_stack();{
  /* const */ char *ptr = (char *) str;

  /*
   * terminal do not clear the end of line when printing "\r\n",
   *  so do it ourself:
   * Cursor is entirely managed by terminal.
   */

  do {
      while (*ptr != '\0' && *ptr != '\r') ptr ++;
      /* str pointer has to be in %es segment ! */
      if (ptr - str != 0) {
	  char save = *ptr;
	  *ptr = '\0';		/* there is no protected memory... */
	  BOOT1_putstr (str);	/* but this has to be finished at return */
	  *ptr = save;
	  }
      if (*ptr == '\r') {
	  if (UI.monitor != 0) {
	      BOOT1_putstr ("\033[K"); /* clear end of line */
	      _BIOS_wait (SERIAL_DELAY);
	      }
	  str = ptr++;
	  }
      } while (*ptr != 0);

  }}

static void
SERIAL_clearscreen (void)
  {
  BOOT1_putstr ("\033[H\033[J");
  _BIOS_wait (SERIAL_DELAY);
  }

static unsigned SERIAL_getmode (void)
  {bound_stack();{
  unsigned int tmp;
  unsigned mode;

  if (SERIAL_getVTreport (3, &tmp) != 0) {
      SDBG ((" %s aborted on columns\r\n", __FUNCTION__));
      return 0;	/* return mode 0 if error */
      }
  if (tmp == 1 || tmp == 3) /* set or permanently set */
      mode = 1;
    else
      mode = 0;

  if (UI.monitor <= 1)
      return mode; /* would still work, but too long a timeout */

  if (SERIAL_getcontrol ("*|", &tmp) != 0) {
      SDBG ((" %s aborted on lines\r\n", __FUNCTION__));
      return mode;	/* return 0 or 1 if partial error */
      }
  switch (tmp) {
    case 24: break;
    case 36: mode += 2; break;
    case 48: mode += 4; break;
    default:
      SDBG ((" %s invalid lines\r\n", __FUNCTION__));
      break;
    }
  return mode; /* always return screen cleared on last mode set */
  }}

#ifndef MOUSE_ON_VT
static
#endif /* Also called from mouse.c: */
unsigned SERIAL_getcursor (unsigned char *row, unsigned char *col)
  {bound_stack();{
  unsigned int tmprow, tmpcol;

  checkptr (row, "SERIAL_getcursor invalid 1st parameter.");
  checkptr (col, "SERIAL_getcursor invalid 2nd parameter.");

  BOOT1_putstr ("\033[6n");
  if (SERIAL_waitchar (CHAR_CSI) != 0) {
      SDBG ((" %s aborted on CSI ", __FUNCTION__));
      return 1;
      }
  if (SERIAL_waitreportNUM (&tmprow, ';') != 0) {
      SDBG ((" %s aborted on number 1 ", __FUNCTION__));
      return 1;
      }
  if (SERIAL_waitreportNUM (&tmpcol, 'R') != 0) {
      SDBG ((" %s aborted on number 2 ", __FUNCTION__));
      return 1;
      }
  *row = (unsigned char)tmprow - 1;	/* HOME on VT at (1,1) */
  *col = (unsigned char)tmpcol - 1;
  return 0;
  }}

#ifndef MOUSE_ON_VT
static
#endif /* Also called from mouse.c: */
unsigned SERIAL_setcursor (unsigned char row, unsigned char col)
  {
#if 1
  BOOT1_printf ("\033[%u;%uH", row+1, col+1);   /* HOME on VT at (1,1) */
  _BIOS_wait (SERIAL_DELAY);
#else
  char tmpstr[10], *ptr = tmpstr;

  row++; col++; /* HOME on VT at (1,1) */
  *ptr++ = '\033';
  *ptr++ = '[';
  if (row >= 10)
      *ptr++ = '0' + (row / 10);
  *ptr++ = '0' + (row % 10);
  *ptr++ = ';';
  if (col >= 100)
      *ptr++ = '0' + (col / 100);
  if (col >= 10)
      *ptr++ = '0' + ((col / 10) % 10);
  *ptr++ = '0' + (col % 10);
  *ptr++ = 'H';
  *ptr = '\0';
  BOOT1_putstr (tmpstr);
#endif
  return 0;
  }

static void
SERIAL_setattribute (struct attribute_str attr)
  {bound_stack();{
  unsigned char old = ATTR2BYTE(UI.attributes.current);

  ATTR2BYTE(attr) &= ATTR2BYTE(UI.attributes.valid);

  if (   ATTR2BYTE(attr) == 0
      || ((ATTR2BYTE(attr) & old) != old)) {
      /* force if 0, or if some attr are removed */
      /* "\033[21m" do not work on VT100 */
      BOOT1_putstr ("\033[m");
      old = 0;
      }
  UI.attributes.current = attr;

  ATTR2BYTE(attr) &= ~old;

  if (ATTR2BYTE(attr) != 0) {
#if 0
      unsigned char first = 1;
      BOOT1_putstr ("\033[");
      if (attr.underline) {
	  BOOT1_putstr ("4");
	  first = 0;
	  }
      if (attr.reverse) {
	  if (!first)
	      BOOT1_putstr (";");
	  first = 0;
	  BOOT1_putstr ("7");
	  }
      if (attr.blink) {
	  if (!first)
	      BOOT1_putstr (";");
	  first = 0;
	  BOOT1_putstr ("5");
	  }
      if (attr.brighter) {
	  if (!first)
	      BOOT1_putstr (";");
	  BOOT1_putstr ("1");
	  }
      BOOT1_putstr ("m");
#else
      char tmp[11], *ptr = tmp;

      *ptr++ = '\033'; *ptr++ = '[';
      if (attr.underline)
	  *ptr++ = '4';
      if (attr.reverse) {
	  if (ptr != &tmp[1])
	      *ptr++ = ';';
	  *ptr++ = '7';
	  }
      if (attr.blink) {
	  if (ptr != &tmp[1])
	      *ptr++ = ';';
	  *ptr++ = '5';
	  }
      if (attr.brighter) {
	  if (ptr != &tmp[1])
	      *ptr++ = ';';
	  *ptr++ = '1';
	  }
      *ptr++ = 'm'; *ptr++ = '\0';
      BOOT1_putstr (tmp);
#endif
      }
  _BIOS_wait (SERIAL_DELAY);
  }}

static unsigned SERIAL_setfgcolor (unsigned color)
  {
  if (UI.monitor == 1) /* B&W */
      return 0;

  /* How to test validity of the color ? */
  UI.fgcolor = color;
  if (UI.monitor != 0) { /* DEC VT terminal */
      /* "\033[1m" for extra bright foreground */
      if (UI.fgcolor > 8)
	  BOOT1_printf ("\033[1;%um", 30 + UI.fgcolor - 8);
	else
	  BOOT1_printf ("\033[0;%um", 30 + UI.fgcolor);
      }
    else
      BOOT1_printf ("\033[%um", 30 + UI.fgcolor);
  _BIOS_wait (SERIAL_DELAY);

  return 0;
  }

static unsigned SERIAL_setbgcolor (unsigned color)
  {
  if (UI.monitor == 1) /* B&W */
      return 0;

  /* How to test validity of the color ? */
  UI.bgcolor = color;
  BOOT1_printf ("\033[%um", 40 + UI.bgcolor);
  _BIOS_wait (SERIAL_DELAY);
  return 0;
  }

static void FASTCALL SERIAL_setpixel (coord xy, unsigned color)
  {
  SDBG ((" %s called! ", __FUNCTION__));
  }

static unsigned FASTCALL SERIAL_getpixel (coord xy)
  {
  SDBG ((" %s called! ", __FUNCTION__));
  return 1;
  }

static void
SERIAL_plotHline (coord xy, unsigned short xend, unsigned color)
  {
  SDBG ((" %s called! ", __FUNCTION__));
  }

static unsigned SERIAL_setDACblock (unsigned short startindex, unsigned short nb,
				    const struct DAC_str *palette)
  {
  SDBG ((" %s called! ", __FUNCTION__));
  return 1;
  }

static unsigned
SERIAL_getsize (unsigned mode, struct video_parameter_str *param)
  {
  memset (param, 0, sizeof(*param));

  param->identification = mode;
  param->attr.isgraphic = 0;
  param->attr.graphicbgcolor = 0;

#ifdef MOUSE_ON_VT
  param->charwidth = 8;
  param->charheigh = 16;
#endif

  if (mode & 1)
      param->nbcol = 132;
    else
      param->nbcol = 80;
  param->width = param->nbcol; /* mouse and change_mode() */

  if (mode >= 2 && UI.monitor == 0)
      return 0x100;	/* not a VT, but may accept 80/132 cols */

  switch (mode >> 1) {
    default:
    case 0: param->nbrow = 24; break;
    case 1: param->nbrow = 36; break;
    case 2: param->nbrow = 48; break;
    }
  param->heigh = param->nbrow; /* mouse and change_mode() */

  if (UI.monitor <= 1) /* B&W display */
      param->nbcolor = 0; /* with attributes */
    else
      param->nbcolor = 16; /* nbof (SERIAL_color) */
  return 0;
  }

static unsigned SERIAL_probe (void)
  {bound_stack();{
  unsigned char C, nb1, nb2, cpt, returned;

  UI.monitor = 0;

  strcpy (UI.info.serial.terminalname, "VT000");

  if (_BIOS_getserialstatus(BOOT1_COM_port()).receive_data_ready) {
      SDBG (("\r\nFlushing serial line: "));
      cpt = 30; /* max char flushed */
      while (--cpt) {
	  if (SERIAL_getc (&C) != 0)
	      break;
	  SDBG (("0x%X ", C));
	  }
      }

  SDBG (("\r\nSending Primary DA request to terminal, i.e. CSI c: "));
  BOOT1_printf ("\033[c");

  returned = 0x10;
  if (SERIAL_getc (&C) != 0) {
      SDBG (("no answer.\r\n"));
      return returned;
      }
  returned++;
  if (C != '\033' && C != 0233 /* octal! */) {
      SDBG (("char received: 0x%X, abort.\r\n", C));
      return returned;
      }
  returned++;
  if (C != 0233 /* octal! */) {
      if (SERIAL_getc (&C) != 0) {
	  SDBG (("no answer after ESC.\r\n"));
	  return returned;
	  }
      if (C != '[') {
	  SDBG (("no '[' after ESC.\r\n"));
	  return returned;
	  }
      }

  returned++;
  if (SERIAL_getc (&C) != 0) {
      SDBG (("nothing after 'ESC ['\r\n"));
      return returned;
      }
  returned++;
  if (C != '?') {
      SDBG (("not '?' after 'ESC [': 0x%X\r\n", C));
      return returned;
      }

  returned++;
  if (SERIAL_getc (&C) != 0) {
      SDBG (("nothing after 'ESC [ ?'\r\n"));
      return returned;
      }
  returned++;
  if (C < '0' || C > '9') {
      SDBG (("not a digit after 'ESC [ ?': 0x%X\r\n", C));
      return returned;
      }
  nb1 = C - '0';

  returned++;
  if (SERIAL_getc (&C) != 0) {
      SDBG (("nothing after 'ESC [ ? digit1'\r\n"));
      return returned;
      }
  returned++;
  if ((C < '0' || C > '9') && (C != ';') && (C != 'c')) {
      SDBG (("no digit, nor ';' or 'c' after 'ESC [ ? digit1': 0x%X\r\n", C));
      return returned;
      }
  if (C >= '0' && C <= '9')
      nb1 = 10 * nb1 + C - '0';

  switch (nb1) {
    case 1 :
      SDBG (("VT100.\r\n"));
      UI.info.serial.terminalname[2] = '1';
      break;
    case 6 :
      SDBG (("VT102.\r\n"));
      UI.info.serial.terminalname[2] = '1';
      UI.info.serial.terminalname[4] = '2';
      break;
    case 62:
      SDBG (("VT2*0.\r\n"));
      UI.info.serial.terminalname[2] = '2';
      break;
    case 63:
      SDBG (("VT3*0.\r\n"));
      UI.info.serial.terminalname[2] = '3';
      break;
    case 64:
      SDBG (("VT4*0.\r\n"));
      UI.info.serial.terminalname[2] = '4';
      break;
    case 65:
      SDBG (("VT5*0.\r\n"));
      UI.info.serial.terminalname[2] = '5';
      break;
    default:
      SDBG (("unknown: %u.\r\n", nb1));
      break;
    }

  while (C != 'c' && SERIAL_getc (&C) == 0) {
      SDBG ((" flushing char 0x%X", C));
      }

  SDBG (("\r\nSending Secondary DA request to terminal, i.e. CSI > c: "));
  BOOT1_printf ("\033[>c");

  returned = 0x20;
  if (SERIAL_getc (&C) != 0) {
      SDBG (("no answer.\r\n"));
      return returned;
      }
  returned++;
  if (C != '\033' && C != 0233 /* octal! */) {
      SDBG (("char received: 0x%X, abort.\r\n", C));
      return returned;
      }
  returned++;
  if (C != 0233 /* octal! */) {
      if (SERIAL_getc (&C) != 0) {
	  SDBG (("no answer after ESC.\r\n"));
	  return returned;
	  }
      if (C != '[') {
	  SDBG (("no '[' after ESC.\r\n"));
	  return returned;
	  }
      }

  returned++;
  if (SERIAL_getc (&C) != 0) {
      SDBG (("nothing after 'ESC ['\r\n"));
      return returned;
      }
  returned++;
  if (C != '>') {
      SDBG (("not '>' after 'ESC [': 0x%X\r\n", C));
      return returned;
      }

  returned++;
  if (SERIAL_getc (&C) != 0) {
      SDBG (("nothing after 'ESC [ >'\r\n"));
      return returned;
      }
  returned++;
  if (C < '0' || C > '9') {
      SDBG (("not a digit after 'ESC [ >': 0x%X\r\n", C));
      return returned;
      }
  nb2 = C - '0';

  returned++;
  if (SERIAL_getc (&C) != 0) {
      SDBG (("nothing after 'ESC [ > digit1'\r\n"));
      return returned;
      }
  returned++;
  if ((C < '0' || C > '9') && (C != ';') && (C != 'c')) {
      SDBG (("no digit, nor ';' or 'c' after 'ESC [ > digit1': 0x%X\r\n", C));
      return returned;
      }
  if (C >= '0' && C <= '9')
      nb2 = 10 * nb2 + C - '0';

  switch (nb2) {
    case 24:
    case 41: /* B & W display */
      SDBG (("VT*20 (%u).\r\n", nb2));
      UI.monitor = 1;
      if (UI.info.serial.terminalname[2] != '1')
	  UI.info.serial.terminalname[3] = '2';
      break;
    case 18: /* grayscale display */
      SDBG (("VT*30.\r\n"));
      UI.monitor = 2;
      UI.info.serial.terminalname[3] = '3';
      break;
    case 19: /* color display */
      SDBG (("VT*40.\r\n"));
      UI.monitor = 3;
      UI.info.serial.terminalname[3] = '4';
      break;
    default:
      SDBG (("unknown: %u.\r\n", nb2));
      break;
    }

  while (C != 'c' && SERIAL_getc (&C) == 0) {
      SDBG ((" flushing char 0x%X", C));
      }
  returned = 0;

  {
  /* Set the two default mode, hope that is 80x24 and 132x24 */
  struct videomode_str mode = {
	number: 0, width: 80, heigh: 24,
	bpp: 0,	text: 1, attr: 1
	};

  if (UI.mode == 0 && UI.nbmode != 0)
      VDBG (("ERROR: UI.mode == 0 with UI.nbmode = %u\r\n", UI.nbmode));
  while (UI.nbmode != 0)
      remove_valid_mode (UI.mode[0].number);

  if (nb2 == 18 || nb2 == 19)
      mode.bpp = 4;
  add_valid_mode (&mode);
  mode.number = 1;
  mode.width = 132;
  add_valid_mode (&mode);

  /* everything passed, we can enable extended modes: */
  if (nb1 >= 64) {	/* VT400+ family */
      mode.number = 2;
      mode.width = 80;
      mode.heigh = 36;
      add_valid_mode (&mode);
      mode.number = 3;
      mode.width = 132;
      add_valid_mode (&mode);
      mode.number = 4;
      mode.width = 80;
      mode.heigh = 48;
      add_valid_mode (&mode);
      mode.number = 5;
      mode.heigh = 48;
      mode.width = 132;
      add_valid_mode (&mode);
      UI.info.serial.max_mode = 5;
      }
    else
      UI.info.serial.max_mode = 1;
  }

  return returned;
  }}

static unsigned
SERIAL_setmode (unsigned mode)
  {bound_stack();{
  struct video_parameter_str param;
  unsigned lines;
  unsigned saved_mode = mode, tmp;

  mode &= ~0x8000;

  if (mode > UI.info.serial.max_mode) {
      SDBG (("%s: mode 0x%X cannot be supported by this terminal (max %u)\r\n",
	    __FUNCTION__, mode, UI.info.serial.max_mode));
      return 1;
      }

  if ((tmp = SERIAL_getsize (mode, &param)) != 0)
      return tmp;
  UI.parameter = param;
#if USER_SUPPORT & VGA_FREQUENCY
  UI.frequency = 0;
#endif
  SERIAL_setup_color ();

  if (mode & 1)
      BOOT1_putstr ("\033[?3h");	/* 132 column */
    else
      BOOT1_putstr ("\033[?3l");	/* 80 column */
  _BIOS_wait (SERIAL_DELAY);

  switch (mode >> 1) {
    case 0: lines = 24; break;
    case 1: lines = 36; break;
    case 2: lines = 48; break;
      break;
    default:
      return 3;
    }

  BOOT1_printf ("\033[%ut", lines);	/* X lines/page */
  _BIOS_wait (SERIAL_DELAY);
  BOOT1_printf ("\033[%u*|", lines);	/* X lines/screen */
  _BIOS_wait (SERIAL_DELAY);
  BOOT1_printf ("\033[1;%ur", lines);	/* scrolling region 1..X */
  _BIOS_wait (SERIAL_DELAY);

  UI.attributes.valid = egatxt_attribute;
  UI.attributes.current = no_attribute;

  if ((saved_mode & 0x8000) == 0)
      SERIAL_clearscreen();
  return 0;
  }}

static const struct user_interface_fct_str SERIAL_UI_FCT = {
    getsize:		SERIAL_getsize,
    setmode:		SERIAL_setmode,
    getmode:		SERIAL_getmode,
    clearscreen:	SERIAL_clearscreen,

    setcursor:		SERIAL_setcursor,
    getcursor:		SERIAL_getcursor,
    setfgcolor:		SERIAL_setfgcolor,
    setbgcolor:		SERIAL_setbgcolor,
    setattribute:	SERIAL_setattribute,

    putstr:		SERIAL_putstr,

    setpixel:		SERIAL_setpixel,
    getpixel:		SERIAL_getpixel,
    plotHline:		SERIAL_plotHline,

    setDACblock:	SERIAL_setDACblock,
    };

unsigned SERIAL_init (void)
  {
  if (SERIAL_probe() == 0) { /* all requests answered */
      UI.function = SERIAL_UI_FCT;
      if (SERIAL_getsize (SERIAL_getmode(), &UI.parameter) != 0)
	  SDBG (("Initial SERIAL_getsize failed!\r\n"));
      UI.attributes.valid = egatxt_attribute;
      UI.attributes.current = no_attribute;
      if (UI.monitor == 1)	/* B&W display - i.e. green/black */
	  UI.fgcolor = 1;
      return 0;
      }
    else {
      unsigned char saverow, savecol, row, col;

      if (SERIAL_getcursor (&saverow, &savecol) != 0) {
	  SDBG (("SERIAL_probe and SERIAL_getcursor failed!\r\n"));
	  return 1;
	  }
      /* try this... */
      SERIAL_setcursor (99, 255);
      if (SERIAL_getcursor (&row, &col) == 0) {
	  SDBG (("SERIAL_probe failed, using SERIAL_getcursor: %ux%u\r\n",
			col, row));
	  UI.function = SERIAL_UI_FCT;
	  UI.attributes.valid = egatxt_attribute;
	  UI.attributes.current = no_attribute;
	  UI.parameter.row = saverow;
	  UI.parameter.col = savecol;
	  UI.parameter.heigh = UI.parameter.nbrow = row;
	  UI.parameter.width = UI.parameter.nbcol = col;
	  }
      SERIAL_setcursor (saverow, savecol);
      return 0;
      }
  }
#endif /* SERIAL_SUPPORT */

/*
 * Init a basic generic system which should work everywhere:
 */
void UI_init (unsigned serial)
  {
  unsigned i;

  memset (&UI, 0, sizeof(UI));

  UI.fgcolor = 7;
  for (i = 0; i < nbof (UI.stdcolor); i++)
      UI.stdcolor[i] = i;
  UI.parameter.nbcolor = 16;

  for (i = 0; i < sizeof (UI.function) / sizeof (void *); i++)
      ((void **)&UI.function) [i] = (void *) COMMON_error;

  UI.function.setfgcolor = DEFLT_setfgcolor; /* VGA: only graphic modes */
  UI.function.setbgcolor = COMMON_setbgcolor;
  UI.function.putstr = BOOT1_putstr;

  if (serial) {
      UI.getkey = SERIAL_getkey;
#if !(USER_SUPPORT & SERIAL_SUPPORT)
      DBG (("WARNING: %s SUPPORT NOT COMPILED IN!\r\n", "SERIAL"));
      printf ("WARNING: %s SUPPORT NOT COMPILED IN!\r\n", "SERIAL");
#endif
      }
    else {
#if (USER_SUPPORT & VESA_SUPPORT) && (USER_SUPPORT & VGA_SUPPORT)
      VESA_VbeInfoBlock vesainfo;
      const char signature[4] = { 'V', 'E', 'S', 'A' };

      DBG (("Initial VESA name check: "));
      *(unsigned *)(vesainfo.VbeSignature) = *(unsigned *)signature;
      if (   _VESA_getinfo (&vesainfo) == 0
	  && *(unsigned *)(vesainfo.VbeSignature) == *(unsigned *)signature) {
	  if (VIDEO_mode_checkname (vesainfo.OemStringPtr) != 0) {
	      VIDEO_mode_initname (vesainfo.OemStringPtr);
	      DBG (("Video card changed, reset VGA parameter.\r\n"));
	      VIDEO_mode_reset();
	      /* BOOT1_update_bootparam() will be called by VGA_init()
		 if not DOSEMU */
	      puts (VIDEO_CARD_CHANGED);
	      }
	    else {
	      DBG (("same VESA card.\r\n"));
	      }
	  }
	else {
	  DBG (("No VESA signature.\r\n"));
	  }
#endif
      UI.getkey = BIOS_getkey;
#if !(USER_SUPPORT & (VGA_SUPPORT | VESA_SUPPORT))
      DBG (("WARNING: %s SUPPORT NOT COMPILED IN!\r\n", "VGA/VESA"));
      printf ("WARNING: %s SUPPORT NOT COMPILED IN!\r\n", "VGA/VESA");
#endif
      }
  }

/*
 * extra segment stuff:
 */
GENERATE_EXTRA2CODE_INDIRECT_STUB(UI_getkey, UI.getkey, unsigned,
					unsigned timeout)
#if USER_SUPPORT & VGA_SUPPORT
GENERATE_EXTRA2CODE_STUB(VGA_init, unsigned, void)
#endif
#if USER_SUPPORT & VESA_SUPPORT
GENERATE_EXTRA2CODE_STUB(VESA_init, unsigned, void)
#endif

GENERATE_EXTRA2CODE_INDIRECT_STUB(UI_setmode, UI.function.setmode,
				unsigned, unsigned mode)
GENERATE_EXTRA2CODE_INDIRECT_STUB(UI_getmode, UI.function.getmode,
				unsigned, void)
GENERATE_VOID_EXTRA2CODE_INDIRECT_STUB(UI_clearscreen,
				UI.function.clearscreen, void)
GENERATE_EXTRA2CODE_INDIRECT_STUB(UI_setcursor, UI.function.setcursor,
				unsigned, unsigned char row, unsigned char col)
GENERATE_EXTRA2CODE_INDIRECT_STUB(UI_getcursor, UI.function.getcursor,
				unsigned, unsigned char *row, unsigned char *col)
GENERATE_EXTRA2CODE_INDIRECT_STUB(UI_setfgcolor, UI.function.setfgcolor,
				unsigned, unsigned color)
GENERATE_EXTRA2CODE_INDIRECT_STUB(UI_setbgcolor, UI.function.setbgcolor,
				unsigned, unsigned color)

/*
 * This is in ".data" or ".common/.bss" segment.
 *
 * It is NOT malloced, to have a linker defined address
 * and so produce faster call than the indirect access:
 * In the assembly file, a field of UI is accessed by:
 *   mov (UI+10),%%eax
 * If it were malloced, it would be:
 *   mov UI,%%eax
 *   mov 10(%%eax),%%eax
 */
struct user_interface_str UI;

#endif /* USER_SUPPORT != 0 */

